/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\Kad.cs
 * PURPOSE: Kademlia related code (1)
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *            Some portions of this file are loosely based on the
 *            GPLed code of iMule v1.4.5.
 *            Copyright (C) 2004-2006 iMule Project.
 *            Copyright (C) 2004-2006 Angel Vidal (Kry).
 *            Copyright (C) 2003 Barry Dunne.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using NB.Content;
using NB.Gui;
using NB.I2P;
using NB.Language;
using Support;

namespace NB.Network.eD2k
{
    enum ClientSoftware : byte
    {
        SO_CDONKEY = 1,
        SO_IMULE = 7,
        SO_NACHTBLITZ = 0xEA,
    }

    enum FileRatings : byte
    {
        NotRated = 0,
        InvalidCorruptFake = 1,
        Poor = 2,
        Fair = 3,
        Good = 4,
        Excellent = 5,
    }

    class KadMap
    {
        // We cannot return more than about 40 sources at once,
        // due to the maximum packet size.
        public const ushort c_MaxSourcesPerFile = byte.MaxValue;
        public const ushort c_MaxFilesPerKeyword = 5000;
        public const ushort c_MaxNotesPerFile = 50;

        private const uint c_MaxSourceLifetime = (12 * Time.c_SecsPerHour);
        private const uint c_MaxKeywordLifetime = (24 * Time.c_SecsPerHour);
        private const uint c_MaxNoteLifetime = (24 * Time.c_SecsPerHour);

        private readonly Kademlia f_Kad;

        private readonly EntryMap<FileID, UserID> f_Sources;
        private readonly EntryMap<KeywordID, FileID> f_Keywords;
        private readonly EntryMap<FileID, ClientID> f_Notes;

        // f_UnicodeSupport affects the creation of new entries.
        private Tag.UnicodeSupport f_UnicodeSupport;

        public KadMap(Kademlia a_Kad)
        {
            f_Kad = a_Kad;

            f_Sources = new EntryMap<FileID, UserID>(c_MaxSourcesPerFile);
            f_Keywords = new EntryMap<KeywordID, FileID>(c_MaxFilesPerKeyword);
            f_Notes = new EntryMap<FileID, ClientID>(c_MaxNotesPerFile);
        }

        public void Clean(Time a_Time)
        {
            f_Sources.Clean(a_Time);
            f_Keywords.Clean(a_Time);
            f_Notes.Clean(a_Time);
        }

        public void GetEntryCounts(out uint a_FileEntryCount,
                                   out uint a_KeywordEntryCount,
                                   out uint a_NoteEntryCount)
        {
            a_FileEntryCount = f_Sources.GetCount();
            a_KeywordEntryCount = f_Keywords.GetCount();
            a_NoteEntryCount = f_Notes.GetCount();
        }

        public uint GetMemorySize()
        {
            return (f_Sources.GetMemorySize() + f_Keywords.GetMemorySize() + f_Notes.GetMemorySize());
        }

        public void SetMaxMemoryConsumption(uint a_MaxMemory)
        {
            uint l_Perc = (a_MaxMemory / 100);
            uint l_Remainder = (a_MaxMemory % 100);

            // That are realistic proportions.
            f_Sources.SetMaxMemorySize((l_Perc * 57) + l_Remainder);
            f_Keywords.SetMaxMemorySize(l_Perc * 42);
            f_Notes.SetMaxMemorySize(l_Perc * 1);
        }

        public void AdjustUnicodeSupport(Tag.UnicodeSupport a_UnicodeSupport)
        {
            // This routine does not affected entries that are already
            // indexed.

            f_UnicodeSupport = a_UnicodeSupport;
        }

        public uint GetFileSourceCount(FileID a_FileID,
                                       out uint a_CompleteSources)
        {
            a_CompleteSources = 0;

            uint l_TotalSources;
            IEnumerable<Entry<FileID, UserID>> l_Enumerable = f_Sources.Enumerable(a_FileID, out l_TotalSources);

            if (null != l_Enumerable)
            {
                foreach (FileEntry.LeanFileEntry l_CurEntry in l_Enumerable)
                {
                    if (true == l_CurEntry.f_Complete)
                    {
                        a_CompleteSources++;
                    }
                }
            }

            return l_TotalSources;
        }

        public List<I2PAddress> GetFileSourceArray(FileID a_FileID,
                                                   ulong a_FileSize)
        {
            // Return can be null, if there are no sources.

            uint l_TotalSources;
            IEnumerable<Entry<FileID, UserID>> l_Enumerable = f_Sources.Enumerable(a_FileID, out l_TotalSources);

            if (null == l_Enumerable)
            {
                return null;
            }

            List<I2PAddress> l_Ls = new List<I2PAddress>((int)Math.Min(byte.MaxValue, l_TotalSources));

            foreach (FileEntry.LeanFileEntry l_CurEntry in l_Enumerable)
            {
                if (0 == l_CurEntry.f_FileSize ||
                    l_CurEntry.f_FileSize != a_FileSize)
                {
                    continue;
                }

                l_Ls.Add(l_CurEntry.GetStreamAddress());

                if (byte.MaxValue <= l_Ls.Count)
                {
                    break;
                }
            }

            return (0 >= l_Ls.Count ? null : l_Ls);
        }

        private uint CalcLifetimeInSecs(uint a_MaxSeconds,
                                        ushort a_MaxMinutes)
        {
            return Math.Min(a_MaxSeconds, (uint)a_MaxMinutes * 60);
        }

        public bool AddFileSource(FileEntry a_Entry,
                                  bool a_NeedFilterCheck,
                                  out int a_Load)
        {
            if (true == a_NeedFilterCheck &&
                true == f_Kad.f_KadPrefs.f_KadFilter.RunKnownFileThroughFilter(f_Kad.f_Inst, a_Entry.f_KeyID, a_Entry.f_FileSize, KademliaFilterScopes.ScopeLocalKademliaIndex))
            {
                a_Load = 0;

                return false;
            }

            return AddInternal<FileID, UserID>(a_Entry.MakeLean(f_UnicodeSupport),
                                               CalcLifetimeInSecs(c_MaxSourceLifetime, a_Entry.f_MaxMinutes),
                                               f_Sources,
                                               out a_Load);
        }

        public bool AddKeyword(KeywordEntry a_Entry,
                               out int a_Load)
        {
            if (true == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(a_Entry.f_FileName, KademliaFilterScopes.ScopeLocalKademliaIndex))
            {
                a_Load = 0;

                return false;
            }

            return AddInternal<KeywordID, FileID>(a_Entry.MakeLean(f_UnicodeSupport),
                                                  CalcLifetimeInSecs(c_MaxKeywordLifetime, a_Entry.f_MaxMinutes),
                                                  f_Keywords,
                                                  out a_Load);
        }

        public bool AddNote(NoteEntry a_Entry,
                            out int a_Load)
        {
            if (true == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(a_Entry.f_FileName, KademliaFilterScopes.ScopeLocalKademliaIndex))
            {
                a_Load = 0;

                return false;
            }

            return AddInternal<FileID, ClientID>(a_Entry.MakeLean(f_UnicodeSupport),
                                                 CalcLifetimeInSecs(c_MaxNoteLifetime, a_Entry.f_MaxMinutes),
                                                 f_Notes,
                                                 out a_Load);
        }

        private bool AddInternal<TKeyID, TRefID>(Entry<TKeyID, TRefID> a_LeanEntry,
                                                 uint a_LifetimeInSecs,
                                                 EntryMap<TKeyID, TRefID> a_Map,
                                                 out int a_Load)
                                                 where TKeyID : IUInt128
                                                 where TRefID : IUInt128
        {
            // a_LeanEntry may be null.

            if (null != a_LeanEntry &&
                0 != a_LifetimeInSecs)
            {
                Time l_Time = f_Kad.f_Inst.GetTime().AddSeconds(a_LifetimeInSecs);
                a_LeanEntry.SetLifetime(l_Time);

                a_Load = a_Map.Set(a_LeanEntry);

                return true;
            }

            a_Load = 0;

            return false;
        }

        public void SendFileResult(FileID a_KeyID,
                                   I2PAddress a_Addr)
        {
            SendResults<FileID, UserID>(a_KeyID, f_Sources.Enumerable(a_KeyID), a_Addr);
        }

        public void SendValidKeywordResult(KeywordID a_KeywordID,
                                           I2PAddress a_Addr)
        {
            SendResults<KeywordID, FileID>(a_KeywordID, f_Keywords.Enumerable(a_KeywordID), a_Addr);
        }

        public void SendValidSearchTermsResult(KeywordID a_KeywordID,
                                               SearchTerm a_SearchTerms,
                                               I2PAddress a_Addr)
        {
            IEnumerable<Entry<KeywordID, FileID>> l_Enumerable = f_Keywords.Enumerable(a_KeywordID);

            if (null == l_Enumerable)
            {
                return;
            }

            List<Entry<KeywordID, FileID>> l_Results = null;

            foreach (KeywordEntry.LeanKeywordEntry l_Entry in l_Enumerable)
            {
                if (true == SearchTermsMatch(l_Entry, a_SearchTerms))
                {
                    if (null == l_Results)
                    {
                        l_Results = new List<Entry<KeywordID, FileID>>(1);
                    }

                    l_Results.Add(l_Entry);
                }
            }

            SendResults<KeywordID, FileID>(a_KeywordID, l_Results, a_Addr);
        }

        public void AddValidKeywordsToSearchCookie(KeywordID a_KeywordID,
                                                   SearchTerm a_SearchTerms,
                                                   GuiSearchCookie a_Cookie)
        {
            IEnumerable<Entry<KeywordID, FileID>> l_Enumerable = f_Keywords.Enumerable(a_KeywordID);

            if (null == l_Enumerable)
            {
                return;
            }

            foreach (KeywordEntry.LeanKeywordEntry l_Entry in l_Enumerable)
            {
                if (true == SearchTermsMatch(l_Entry, a_SearchTerms))
                {
                    TagList l_TagList = l_Entry.f_TagList;

                    a_Cookie.AddFile(f_Kad.f_KadPrefs.f_KadFilter,
                                     l_Entry.f_RefID,
                                     l_Entry.f_FileName,
                                     l_TagList.GetTag(TagCodes.TAG_FILESIZE).GetAsULong(),
                                     (ushort)l_TagList.GetTag(TagCodes.TAG_COMPLETE_SOURCES).GetAsULong(),
                                     (ushort)l_TagList.GetTag(TagCodes.TAG_SOURCES).GetAsULong());
                }
            }

            a_Cookie.MakeResultsAvailableToGui();
        }

        public void SendNoteResult(FileID a_KeyID,
                                   I2PAddress a_Addr)
        {
            SendResults<FileID, ClientID>(a_KeyID, f_Notes.Enumerable(a_KeyID), a_Addr);
        }

        private void SendResults<TKeyID, TRefID>(TKeyID a_KeyID,
                                                 IEnumerable<Entry<TKeyID, TRefID>> a_Results,
                                                 I2PAddress a_Addr)
                                                 where TKeyID : IUInt128
                                                 where TRefID : IUInt128
        {
            // a_Results may be null and should otherwise contain at
            // least one entry.

            if (null == a_Results)
            {
                return;
            }

            uint l_MaxSize = PacketHandler.c_MaxKadMessageSize;
            MemoryIO l_Buffer = MemoryIO.CreateWritable();
            BufferIO.CountWriter l_CountWtr;
            uint l_Count = 0;
            uint l_OldPos = (UInt128.c_ByteSize + sizeof(ushort));

            l_Buffer.WriteUInt128ME(a_KeyID.GetUInt128());
            BufferIO.CountWriter.InitAndMoveBuffer(l_Buffer, sizeof(ushort), out l_CountWtr);
            bool l_RestrictToClassicAddresses = (typeof(UserID) == typeof(TRefID) &&
                                                 true == NetworkManager.ShouldRestrictPeerToClassicAddresses(a_Addr));
            KademliaFilter l_KadFilter = f_Kad.f_KadPrefs.f_KadFilter;

            if (false == l_KadFilter.HasActiveRules(KademliaFilterScopes.ScopeSearchResultsForOthers))
            {
                l_KadFilter = null;
            }

            foreach (Entry<TKeyID, TRefID> l_Entry in a_Results)
            {
                if (true == l_RestrictToClassicAddresses)
                {
                    FileEntry.LeanFileEntry l_LeanEntry = (FileEntry.LeanFileEntry)(object)l_Entry;

                    if (false == l_LeanEntry.f_ClassicAddr)
                    {
                        continue;
                    }
                }

                l_Buffer.WriteUInt128ME(l_Entry.f_RefID.GetUInt128());

                if (typeof(KeywordEntry.LeanKeywordEntry) == l_Entry.GetType())
                {
                    KeywordEntry.LeanKeywordEntry l_LeanEntry = l_Entry as KeywordEntry.LeanKeywordEntry;

                    if (null != l_KadFilter &&
                        true == l_KadFilter.RunFileNameThroughFilter(l_LeanEntry.f_FileName, KademliaFilterScopes.ScopeSearchResultsForOthers))
                    {
                        continue;
                    }

                    uint l_CompletedSources;
                    uint l_Sources = GetFileSourceCount(l_LeanEntry.f_RefID, out l_CompletedSources);
                    l_LeanEntry.TagListToBufferWithCounts(f_UnicodeSupport, l_CompletedSources, l_Sources, l_Buffer);
                }
                else
                {
                    if (null != l_KadFilter)
                    {
                        if (typeof(FileEntry.LeanFileEntry) == l_Entry.GetType())
                        {
                            FileEntry.LeanFileEntry l_LeanEntry = l_Entry as FileEntry.LeanFileEntry;

                            if (true == l_KadFilter.RunKnownFileThroughFilter(f_Kad.f_Inst, l_LeanEntry.f_KeyID, l_LeanEntry.f_FileSize, KademliaFilterScopes.ScopeSearchResultsForOthers))
                            {
                                continue;
                            }
                        }
                        else
                        {
                            NoteEntry.LeanNoteEntry l_LeanEntry = l_Entry as NoteEntry.LeanNoteEntry;

                            if (true == l_KadFilter.RunFileNameThroughFilter(l_LeanEntry.f_FileName, KademliaFilterScopes.ScopeSearchResultsForOthers))
                            {
                                continue;
                            }
                        }
                    }

                    l_Entry.TagListToBuffer(l_Buffer);
                }

                uint l_NewPos = l_Buffer.p_Position;

                if (l_MaxSize < l_NewPos)
                {
                    // If l_Packet size is too big, revert the last
                    // write.

                    l_Buffer.SetLength(l_OldPos);

                    break;
                }

                l_Count++;
                l_OldPos = l_NewPos;
            }

            if (0 == l_Count)
            {
                return;
            }

            l_CountWtr.Finish(l_Count);

            f_Kad.f_PacketHndlr.SendPacket(KadOpcodes.KADEMLIA_SEARCH_RES, a_Addr, l_Buffer);
        }

        public static bool SearchTermsMatch(KeywordEntry a_Entry,
                                            SearchTerm a_SearchTerm)
        {
            return SearchTermsMatch(a_Entry.MakeLean(Tag.UnicodeSupport.NBStyle), a_SearchTerm);
        }

        public static bool SearchTermsMatch(KeywordEntry.LeanKeywordEntry a_Entry,
                                            SearchTerm a_SearchTerm)
        {
            // We compare strings case insensitive.

            if (SearchTermTypes.NAND >= a_SearchTerm.f_Type)
            {
                if (SearchTermTypes.AND == a_SearchTerm.f_Type)
                {
                    return (true == SearchTermsMatch(a_Entry, a_SearchTerm.f_Left) &&
                            true == SearchTermsMatch(a_Entry, a_SearchTerm.f_Right));
                }

                if (SearchTermTypes.OR == a_SearchTerm.f_Type)
                {
                    return (true == SearchTermsMatch(a_Entry, a_SearchTerm.f_Left) ||
                            true == SearchTermsMatch(a_Entry, a_SearchTerm.f_Right));
                }

                return (true == SearchTermsMatch(a_Entry, a_SearchTerm.f_Left) &&
                        false == SearchTermsMatch(a_Entry, a_SearchTerm.f_Right));
            }
            else if (SearchTermTypes.StringsAND == a_SearchTerm.f_Type ||
                     SearchTermTypes.StringsOR == a_SearchTerm.f_Type)
            {
                bool l_AND = (SearchTermTypes.StringsAND == a_SearchTerm.f_Type);

                return SearchManager.ContainsKeywords(a_Entry.f_FileName, a_SearchTerm.f_Strings, l_AND);
            }
            else if (SearchTermTypes.StringsNOT == a_SearchTerm.f_Type)
            {
                // We do this as an inverted OR.

                return !SearchManager.ContainsKeywords(a_Entry.f_FileName, a_SearchTerm.f_Strings, false);
            }
            else if (SearchTermTypes.MetaTag == a_SearchTerm.f_Type)
            {
                Tag l_SearchTag = a_SearchTerm.f_MetaStrTag;

                // The file type is special.
                if (TagCodes.TAG_FILETYPE == l_SearchTag.f_Code)
                {
                    FileType l_RealFileType = FileTypeDetector.GuessFileType(a_Entry.f_FileName);

                    if (true == l_SearchTag.IsStr())
                    {
                        FileType l_ExpectedType;
                        if (true == FileTypeDetector.TryToRecognizeStringFileType(l_SearchTag.GetStr(), out l_ExpectedType))
                        {
                            return (l_ExpectedType == l_RealFileType);
                        }
                    }
                    else if (true == l_SearchTag.IsNumber())
                    {
                        return (l_SearchTag.GetAsULong() == (ulong)l_RealFileType);
                    }

                    return false;
                }

                // The file extension is special.
                if (TagCodes.TAG_FILEFORMAT == l_SearchTag.f_Code)
                {
                    if (true == l_SearchTag.IsStr())
                    {
                        string l_ExpectedExt = l_SearchTag.GetStr();
                        string l_RealExt = LongPath.GetExtension(a_Entry.f_FileName);

                        if (0 == l_ExpectedExt.Length)
                        {
                            return (null == l_RealExt);
                        }

                        return Util.StringsEqualIgnoreCase(l_ExpectedExt, l_RealExt);
                    }

                    return false;
                }

                Tag l_ItemTag = a_Entry.f_TagList.GetTag(l_SearchTag.f_Code);

                if (null != l_ItemTag)
                {
                    if (true == l_SearchTag.IsStr())
                    {
                        if (true == l_ItemTag.IsStr())
                        {
                            return Util.StringsEqual(l_SearchTag.GetStr(), l_ItemTag.GetStr());
                        }
                    }
                    else
                    {
                        return Tag.AreEqualNumberTags(l_SearchTag, l_ItemTag);
                    }
                }

                return false;
            }
            else if (SearchTermTypes.OpNotEqual >= a_SearchTerm.f_Type)
            {
                Tag l_SearchTag = a_SearchTerm.f_MinMaxTag;
                Tag l_ItemTag = a_Entry.f_TagList.GetTag(l_SearchTag.f_Code);

                if (null != l_ItemTag)
                {
                    ulong l_Num1;

                    if (true == l_SearchTag.TryGetAsULong(out l_Num1))
                    {
                        ulong l_Num0;

                        if (true == l_ItemTag.TryGetAsULong(out l_Num0))
                        {
                            return Compare(a_SearchTerm.f_Type, l_Num0.CompareTo(l_Num1));
                        }
                    }
                    else if (true == l_SearchTag.IsFloat() &&
                             true == l_ItemTag.IsFloat())
                    {
                        float l_Float0 = l_ItemTag.GetFloat();
                        float l_Float1 = l_SearchTag.GetFloat();

                        return Compare(a_SearchTerm.f_Type, l_Float0.CompareTo(l_Float1));
                    }
                }
            }

            return false;
        }

        private static bool Compare(SearchTermTypes a_CompareOperation,
                                    int a_Relation)
        {
            switch (a_CompareOperation)
            {
                case SearchTermTypes.OpLessEqual:
                    return (0 >= a_Relation);

                case SearchTermTypes.OpGreaterEqual:
                    return (0 <= a_Relation);

                case SearchTermTypes.OpLess:
                    return (0 > a_Relation);

                case SearchTermTypes.OpGreater:
                    return (0 < a_Relation);

                case SearchTermTypes.OpEqual:
                    return (0 == a_Relation);

                case SearchTermTypes.OpNotEqual:
                    return (0 != a_Relation);

                default: throw (new ApplicationException());
            }
        }

        class EntryMap<TKeyID, TRefID>
                 where TKeyID : IUInt128
                         where TRefID : IUInt128
        {
            // Should be larger than one I2P datagram.
            public const uint c_MinMemorySizeLimit = (64 * 1024);

            public readonly uint f_MaxEntriesPerCategory;
            private readonly Dictionary<TKeyID, RefMap> f_Dic;
            private readonly LinkedList<IDPair> f_Order;
            private readonly Dictionary<IDPair, LinkedListNode<IDPair>> f_OrderDic;

            private uint f_TotalEntries;
            // NOTE: We only account for the net sizes of
            // the contained entries, not for the costs
            // of administration.
            private uint f_MaxMemorySize;
            private uint f_MemorySize;

            public EntryMap(uint a_MaxEntriesPerCategory)
            {
                f_MaxEntriesPerCategory = a_MaxEntriesPerCategory;

                f_Dic = new Dictionary<TKeyID, RefMap>();
                f_Order = new LinkedList<IDPair>();
                f_OrderDic = new Dictionary<IDPair, LinkedListNode<IDPair>>();

                f_MaxMemorySize = c_MinMemorySizeLimit;
            }

            public uint GetCount()
            {
                return f_TotalEntries;
            }

            public void SetMaxMemorySize(uint a_MaxMemorySize)
            {
                f_MaxMemorySize = Math.Max(c_MinMemorySizeLimit, a_MaxMemorySize);

                EnforceMemoryThreshold(f_MaxMemorySize);
            }

            public uint GetMemorySize()
            {
                return f_MemorySize;
            }

            public int GetTotalLoad()
            {
                // We use the memory size (not the count) for
                // this.

                return CalcLoad(f_MemorySize, f_MaxMemorySize);
            }

            public int GetLoad(TKeyID a_KeyID)
            {
                // We use the category's count for this.

                RefMap l_RefMap;
                if (true == f_Dic.TryGetValue(a_KeyID, out l_RefMap))
                {
                    return CalcLoad(l_RefMap.GetCount(), f_MaxEntriesPerCategory);
                }

                return 0;
            }

            private static int CalcLoad(uint a_Count,
                                         uint a_MaxCount)
            {
                ulong l_Dividend = ((ulong)a_Count * 100);

                return (int)(l_Dividend / a_MaxCount);
            }

            public int Set(Entry<TKeyID, TRefID> a_Entry)
            {
                // Returns the load of the category affected.

                uint l_MemoryRequired = a_Entry.GetMemorySize();

                if (f_MaxMemorySize < l_MemoryRequired)
                {
                    // We won't be able to get this one in
                    // (should never happen).

                    return 100;
                }

                EnforceMemoryThreshold(f_MaxMemorySize - l_MemoryRequired);

                RefMap l_RefMap;
                bool l_HasRefMap = f_Dic.TryGetValue(a_Entry.f_KeyID, out l_RefMap);

                if (false == l_HasRefMap)
                {
                    l_RefMap = new EntryMap<TKeyID, TRefID>.RefMap();

                    f_Dic.Add(a_Entry.f_KeyID, l_RefMap);
                }
                else
                {
                    if (true == l_RefMap.Has(a_Entry.f_RefID))
                    {
                        // Prepare re-entry.

                        Remove(a_Entry.f_KeyID, a_Entry.f_RefID, l_RefMap, true);
                    }
                    else
                    {
                        // We are adding this ref id for the first time,
                        // so we might have to make room.

                        if (l_RefMap.GetCount() == f_MaxEntriesPerCategory)
                        {
                            Entry<TKeyID, TRefID> a_OldestEntry = l_RefMap.GetOldest();

                            Remove(a_OldestEntry.f_KeyID, a_OldestEntry.f_RefID, l_RefMap, true);
                        }
                    }
                }

                l_RefMap.Set(a_Entry);

                IDPair l_Pair = new IDPair(a_Entry.f_KeyID, a_Entry.f_RefID);

                LinkedListNode<IDPair> l_Node = f_Order.AddFirst(l_Pair);
                f_OrderDic.Add(l_Pair, l_Node);

                f_TotalEntries++;
                f_MemorySize += l_MemoryRequired;

                return CalcLoad(l_RefMap.GetCount(), f_MaxEntriesPerCategory);
            }

            public void Remove(TKeyID a_KeyID,
                               TRefID a_RefID)
            {
                RefMap l_RefMap;
                if (true == f_Dic.TryGetValue(a_KeyID, out l_RefMap))
                {
                    Remove(a_KeyID, a_RefID, l_RefMap, false);
                }
            }

            private void Remove(TKeyID a_KeyID,
                                TRefID a_RefID,
                                RefMap a_RefMap,
                                bool a_NeverRemoveRefMap)
            {
                // a_RefMap will be removed from f_Dic if it
                // becomes empty.

                uint l_BytesReleased;
                if (true == a_RefMap.Remove(a_RefID, out l_BytesReleased))
                {
                    IDPair l_Pair = new IDPair(a_KeyID, a_RefID);
                    LinkedListNode<IDPair> l_Node = f_OrderDic[l_Pair];

                    f_Order.Remove(l_Node);
                    bool OK = f_OrderDic.Remove(l_Pair);

                    if (false == a_NeverRemoveRefMap &&
                        0 == a_RefMap.GetCount())
                    {
                        OK = f_Dic.Remove(a_KeyID);
                    }

                    f_TotalEntries--;
                    f_MemorySize -= l_BytesReleased;
                }
            }

            private void EnforceMemoryThreshold(uint a_MemoryThreshold)
            {
                while (a_MemoryThreshold < f_MemorySize)
                {
                    if (false == DropOldest())
                    {
                        throw (new ApplicationException());
                    }
                }
            }

            public bool DropOldest()
            {
                if (0 < f_Order.Count)
                {
                    LinkedListNode<IDPair> l_Node = f_Order.Last;
                    RefMap l_RefMap = f_Dic[l_Node.Value.f_KeyID];

                    Remove(l_Node.Value.f_KeyID, l_Node.Value.f_RefID, l_RefMap, false);

                    return true;
                }

                return false;
            }

            public void Clean(Time a_Time)
            {
                List<TKeyID> l_KeyRemList = new List<TKeyID>();
                List<TRefID> l_RefRemList = new List<TRefID>();

                foreach (KeyValuePair<TKeyID, RefMap> l_Pair in f_Dic)
                {
                    foreach (Entry<TKeyID, TRefID> l_Entry in l_Pair.Value.Enumerable())
                    {
                        if (true == l_Entry.p_Lifetime.IsEarlierThan(a_Time))
                        {
                            l_RefRemList.Add(l_Entry.f_RefID);
                        }
                    }

                    foreach (TRefID l_RefID in l_RefRemList)
                    {
                        Remove(l_Pair.Key, l_RefID, l_Pair.Value, true);
                    }

                    l_RefRemList.Clear();

                    if (0 == l_Pair.Value.GetCount())
                    {
                        l_KeyRemList.Add(l_Pair.Key);
                    }
                }

                foreach (TKeyID l_KeyID in l_KeyRemList)
                {
                    bool OK = f_Dic.Remove(l_KeyID);
                }
            }

            public IEnumerable<Entry<TKeyID, TRefID>> Enumerable(TKeyID a_KeyID)
            {
                uint l_Junk;
                return Enumerable(a_KeyID, out l_Junk);
            }

            public IEnumerable<Entry<TKeyID, TRefID>> Enumerable(TKeyID a_KeyID,
                                                                 out uint a_Count)
            {
                // Return can be null, otherwise the enumeration
                // will contain at least one entry.

                RefMap l_RefMap;
                if (true == f_Dic.TryGetValue(a_KeyID, out l_RefMap))
                {
                    a_Count = l_RefMap.GetCount();

                    return l_RefMap.Enumerable();
                }

                a_Count = 0;

                return null;
            }

            private class IDPair : IEquatable<IDPair>
            {
                public readonly TKeyID f_KeyID;
                public readonly TRefID f_RefID;
                private readonly int f_HashCode;

                public IDPair(TKeyID a_KeyID,
                              TRefID a_RefID)
                {
                    f_KeyID = a_KeyID;
                    f_RefID = a_RefID;
                    f_HashCode = (f_KeyID.GetHashCode() ^ f_RefID.GetHashCode());
                }

                public override bool Equals(object obj)
                {
                    // Warn us if someone operates non-generic.
                    throw (new ApplicationException());
                }

                public bool Equals(IDPair a_Other)
                {
                    return (true == this.f_KeyID.GetUInt128().Equals(a_Other.f_KeyID.GetUInt128()) &&
                            true == this.f_RefID.GetUInt128().Equals(a_Other.f_RefID.GetUInt128()));
                }

                public override int GetHashCode()
                {
                    return f_HashCode;
                }
            }

            private class RefMap
            {
                private readonly LinkedList<Entry<TKeyID, TRefID>> f_Entries;
                private readonly Dictionary<TRefID, LinkedListNode<Entry<TKeyID, TRefID>>> f_Dic;

                public RefMap()
                {
                    f_Entries = new LinkedList<Entry<TKeyID, TRefID>>();
                    f_Dic = new Dictionary<TRefID, LinkedListNode<Entry<TKeyID, TRefID>>>();
                }

                public IEnumerable<Entry<TKeyID, TRefID>> Enumerable()
                {
                    return f_Entries;
                }

                public uint GetCount()
                {
                    return (uint)f_Entries.Count;
                }

                public bool Has(TRefID a_RefID)
                {
                    return f_Dic.ContainsKey(a_RefID);
                }

                public void Set(Entry<TKeyID, TRefID> a_Entry)
                {
                    // Adds a new entry or turns an existing entry
                    // into the youngest.

                    LinkedListNode<Entry<TKeyID, TRefID>> l_Node;

                    if (true == f_Dic.TryGetValue(a_Entry.f_RefID, out l_Node))
                    {
                        // We already have an entry of that kind.

                        Remove(l_Node);
                    }

                    l_Node = f_Entries.AddFirst(a_Entry);
                    f_Dic.Add(a_Entry.f_RefID, l_Node);
                }

                public bool Remove(TRefID a_RefID,
                                   out uint a_BytesReleased)
                {
                    LinkedListNode<Entry<TKeyID, TRefID>> l_Node;
                    if (true == f_Dic.TryGetValue(a_RefID, out l_Node))
                    {
                        a_BytesReleased = l_Node.Value.GetMemorySize();

                        Remove(l_Node);

                        return true;
                    }

                    a_BytesReleased = 0;

                    return false;
                }

                private void Remove(LinkedListNode<Entry<TKeyID, TRefID>> a_Node)
                {
                    f_Entries.Remove(a_Node);

                    bool OK = f_Dic.Remove(a_Node.Value.f_RefID);
                }

                public Entry<TKeyID, TRefID> GetOldest()
                {
                    return f_Entries.Last.Value;
                }
            }
        }
    }

    enum SearchTermOps : byte
    {
        Boolean = 0,
        StringsClassic = 1,
        MetaTag = 2,
        MinMax = 3,

        // Own extension types.

        StringsAdvanced = 4,
    }

    enum SearchTermTypes : byte
    {
        AND = 0,
        OR = 1,
        NAND = 2,

        StringsAND = 3, // (x AND y)
        MetaTag = 4,

        // NOTE: e.g. OpGreaterEqual is used to build the
        // following condition:
        // [number from file meta data] >= [number from search term]
        OpEqual = 5,
        OpGreaterEqual = 6,
        OpLessEqual = 7,
        OpGreater = 8,
        OpLess = 9,
        OpNotEqual = 10,

        // Own extension types.

        StringsOR = 11, // (x OR y)
        StringsNOT = 12, // !(x OR y)
    }

    class SearchTerm
    {
        public const byte c_MaxSearchExpressionTreeLevel = 23;

        public readonly SearchTermTypes f_Type;

        public readonly SearchTerm f_Left;
        public readonly SearchTerm f_Right;

        public readonly string[] f_Strings;
        public readonly StringTag f_MetaStrTag;
        public readonly Tag f_MinMaxTag;

        private SearchTerm(SearchTermTypes a_Type,
                           SearchTerm a_Left,
                           SearchTerm a_Right)
        {
            f_Type = a_Type;
            f_Left = a_Left;
            f_Right = a_Right;
        }

        private SearchTerm(SearchTermTypes a_Type,
                           string[] a_Strings,
                           StringTag a_MetaStrTag,
                           Tag a_MinMaxTag)
        {
            f_Type = a_Type;
            f_Strings = a_Strings;
            f_MetaStrTag = a_MetaStrTag;
            f_MinMaxTag = a_MinMaxTag;
        }

        public static SearchTerm CreateBooleanType(SearchTermTypes a_Type,
                                                   SearchTerm a_Left,
                                                   SearchTerm a_Right)
        {
            return new SearchTerm(a_Type, a_Left, a_Right);
        }

        public static SearchTerm CreateStringType(string a_String)
        {
            return CreateStringsType(SearchTermTypes.StringsAND, new string[1] { a_String });
        }

        public static SearchTerm CreateStringsType(SearchTermTypes a_Type,
                                                   string[] a_Strings)
        {
            return new SearchTerm(a_Type, a_Strings, null, null);
        }

        public static SearchTerm CreateMetaStrType(TagCodes a_TagCode,
                                                   string a_Str)
        {
            StringTag l_MetaStrTag = new StringTag(a_TagCode, a_Str);

            return new SearchTerm(SearchTermTypes.MetaTag, null, l_MetaStrTag, null);
        }

        public static SearchTerm CreateMinMaxType(SearchTermTypes a_Type,
                                                  Tag a_MinMaxTag)
        {
            return new SearchTerm(a_Type, null, null, a_MinMaxTag);
        }

        private static SearchTerm BuildChain(bool a_AND,
                                             params SearchTerm[] a_Terms)
        {
            // a_Terms can contain null items.

            SearchTermTypes l_Type = (false == a_AND ? SearchTermTypes.OR : SearchTermTypes.AND);

            List<SearchTerm> l_Ls = new List<SearchTerm>(a_Terms.Length);

            SearchTerm l_Last = null;
            for (int i = 0; i < a_Terms.Length; i++)
            {
                SearchTerm l_Cur = a_Terms[i];

                if (null != l_Cur)
                {
                    if (null == l_Last)
                    {
                        l_Last = l_Cur;
                    }
                    else
                    {
                        l_Ls.Add(new SearchTerm(l_Type, l_Last, l_Cur));

                        l_Last = null;
                    }
                }
            }

            if (null != l_Last)
            {
                l_Ls.Add(l_Last);

                l_Last = null;
            }

            // l_Ls doesn't contain null items (which makes things
            // simple now).

            if (0 < l_Ls.Count)
            {
                l_Last = l_Ls[0];

                for (int i = 1; i < l_Ls.Count; i++)
                {
                    l_Last = new SearchTerm(l_Type, l_Ls[i], l_Last);
                }
            }

            return l_Last;
        }

        private static SearchTerm BuildOrChainFromStringArray(string[] a_Strings)
        {
            SearchTerm[] l_Terms = new SearchTerm[a_Strings.Length];

            for (int i = (a_Strings.Length - 1); 0 <= i; i--)
            {
                l_Terms[i] = SearchTerm.CreateStringType(a_Strings[i]);
            }

            return BuildChain(false, l_Terms);
        }

        public static SearchTerm CreateTree(GuiSearchTerms a_GuiSearchTerms)
        {
            // a_AndKeywords must contain at least one positive
            // keyword. f_OrKeywords and a_NotKeywords may be null.
            // a_Extension may be null, or an empty string to
            // forbid any extension.

            SearchTerm l_Keywords;

            {
                l_Keywords = SearchTerm.CreateStringsType(SearchTermTypes.StringsAND, a_GuiSearchTerms.f_AndKeywords);

                if (null != a_GuiSearchTerms.f_OrKeywords &&
                    0 < a_GuiSearchTerms.f_OrKeywords.Length)
                {
                    SearchTerm l_OrKeywords = BuildOrChainFromStringArray(a_GuiSearchTerms.f_OrKeywords);

                    // A valid search result must at least contain one of
                    // the OR-keywords, so we use AND here to link the
                    // AND-keywords to the OR-keywords.

                    l_Keywords = SearchTerm.CreateBooleanType(SearchTermTypes.AND, l_Keywords, l_OrKeywords);
                }

                if (null != a_GuiSearchTerms.f_NotKeywords &&
                    0 < a_GuiSearchTerms.f_NotKeywords.Length)
                {
                    SearchTerm l_NotKeywords = BuildOrChainFromStringArray(a_GuiSearchTerms.f_NotKeywords);

                    l_Keywords = SearchTerm.CreateBooleanType(SearchTermTypes.NAND, l_Keywords, l_NotKeywords);
                }
            }

            SearchTerm l_Extension = null;

            if (null != a_GuiSearchTerms.f_Extension)
            {
                l_Extension = SearchTerm.CreateMetaStrType(TagCodes.TAG_FILEFORMAT, a_GuiSearchTerms.f_Extension);
            }

            SearchTerm l_FileType = null;

            if (FileType.FT_ANY != a_GuiSearchTerms.f_FileType)
            {
                string l_TypeStr = FileTypeDetector.FileTypeToNetworkString(a_GuiSearchTerms.f_FileType);

                l_FileType = SearchTerm.CreateMetaStrType(TagCodes.TAG_FILETYPE, l_TypeStr);
            }

            SearchTerm l_FileSize = null;

            {
                SearchTerm l_MinFileSize = null;
                SearchTerm l_MaxFileSize = null;

                if (0 < a_GuiSearchTerms.f_MinFileSize)
                {
                    SearchTermTypes l_Op = (a_GuiSearchTerms.f_MinFileSize == a_GuiSearchTerms.f_MaxFileSize ?
                                           SearchTermTypes.OpEqual : SearchTermTypes.OpGreaterEqual);

                    l_MinFileSize = SearchTerm.CreateMinMaxType(l_Op, new NumberTag(TagCodes.TAG_FILESIZE, a_GuiSearchTerms.f_MinFileSize));
                }

                if (ulong.MaxValue > a_GuiSearchTerms.f_MaxFileSize &&
                    a_GuiSearchTerms.f_MaxFileSize > a_GuiSearchTerms.f_MinFileSize)
                {
                    l_MaxFileSize = SearchTerm.CreateMinMaxType(SearchTermTypes.OpLessEqual, new NumberTag(TagCodes.TAG_FILESIZE, a_GuiSearchTerms.f_MaxFileSize));
                }

                l_FileSize = BuildChain(true, l_MinFileSize, l_MaxFileSize);
            }

            return BuildChain(true, l_Keywords, l_Extension, l_FileType, l_FileSize);
        }

        public static bool ToBuffer(BufferIO a_Buffer,
                                    SearchTerm a_SearchTerms)
        {
            // Return can be null if the nesting is too deep
            // (in this case the file position is invalid).

            return ToBuffer(a_Buffer, a_SearchTerms, 0);
        }

        private static bool ToBuffer(BufferIO a_Buffer,
                                     SearchTerm a_SearchTerms,
                                     int a_Level)
        {
            if (c_MaxSearchExpressionTreeLevel < a_Level)
            {
                return false;
            }

            a_Level++;

            if (SearchTermTypes.NAND >= a_SearchTerms.f_Type)
            {
                a_Buffer.WriteByte((byte)SearchTermOps.Boolean);
                a_Buffer.WriteByte((byte)a_SearchTerms.f_Type);

                if (false == ToBuffer(a_Buffer, a_SearchTerms.f_Left, a_Level) ||
                    false == ToBuffer(a_Buffer, a_SearchTerms.f_Right, a_Level))
                {
                    return false;
                }
            }
            else if (SearchTermTypes.StringsAND == a_SearchTerms.f_Type ||
                     SearchTermTypes.StringsOR == a_SearchTerms.f_Type ||
                     SearchTermTypes.StringsNOT == a_SearchTerms.f_Type)
            {
                if (SearchTermTypes.StringsAND == a_SearchTerms.f_Type)
                {
                    a_Buffer.WriteByte((byte)SearchTermOps.StringsClassic);
                }
                else
                {
                    a_Buffer.WriteByte((byte)SearchTermOps.StringsAdvanced);
                    a_Buffer.WriteByte((byte)a_SearchTerms.f_Type);
                }

                a_Buffer.WriteUTF8StringUInt16Len(string.Join(" ", a_SearchTerms.f_Strings));
            }
            else if (SearchTermTypes.MetaTag == a_SearchTerms.f_Type)
            {
                a_Buffer.WriteByte((byte)SearchTermOps.MetaTag);

                StringTag l_MetaTag = a_SearchTerms.f_MetaStrTag;

                a_Buffer.WriteUTF8StringUInt16Len(l_MetaTag.GetStr());
                a_Buffer.ZeroBytes(sizeof(ushort));
                a_Buffer.WriteByte((byte)l_MetaTag.f_Code);
            }
            else if (SearchTermTypes.OpEqual <= a_SearchTerms.f_Type)
            {
                a_Buffer.WriteByte((byte)SearchTermOps.MinMax);
                a_Buffer.WriteByte((byte)(a_SearchTerms.f_Type - SearchTermTypes.OpEqual));
                // Unicode support doesn't matter here. We are writing
                // numbers.
                Tag.WriteTagContext l_WriteTagCntxt = new Tag.WriteTagContext(a_Buffer, Tag.UnicodeSupport.None);
                a_SearchTerms.f_MinMaxTag.WriteToBuffer(ref l_WriteTagCntxt);
            }
            else
            {
                throw (new ApplicationException());
            }

            return true;
        }

        public static SearchTerm FromBuffer(BufferIO a_Buffer)
        {
            // Return can be null on error.

            try
            {
                return FromBufferCore(a_Buffer, 0);
            }
            catch (BufferIO.InterpretBufferException)
            {
                return null;
            }
        }

        private static SearchTerm FromBufferCore(BufferIO a_Buffer,
                                                 byte a_Level)
        {
            // Return can be null on error.

            if (c_MaxSearchExpressionTreeLevel < a_Level)
            {
                return null;
            }

            a_Level++;

            SearchTermOps l_Op = (SearchTermOps)a_Buffer.ReadByte();

            SearchTerm l_SearchTerms = null;

            if (SearchTermOps.Boolean == l_Op)
            {
                SearchTermTypes l_BoolOp = (SearchTermTypes)a_Buffer.ReadByte();

                if (SearchTermTypes.NAND >= l_BoolOp)
                {
                    // There has got to be a valid left and
                    // right item.

                    SearchTerm l_Left = FromBufferCore(a_Buffer, a_Level);

                    if (null != l_Left)
                    {
                        SearchTerm l_Right = FromBufferCore(a_Buffer, a_Level);

                        if (null != l_Right)
                        {
                            l_SearchTerms = new SearchTerm(l_BoolOp, l_Left, l_Right);
                        }
                    }
                }
            }
            else if (SearchTermOps.StringsClassic == l_Op)
            {
                l_SearchTerms = ReadStringType(SearchTermTypes.StringsAND, a_Buffer);
            }
            else if (SearchTermOps.StringsAdvanced == l_Op)
            {
                SearchTermTypes l_StrOp = (SearchTermTypes)a_Buffer.ReadByte();

                if (SearchTermTypes.StringsAND == l_StrOp ||
                    SearchTermTypes.StringsOR == l_StrOp ||
                    SearchTermTypes.StringsNOT == l_StrOp)
                {
                    l_SearchTerms = ReadStringType(l_StrOp, a_Buffer);
                }
            }
            else if (SearchTermOps.MetaTag == l_Op)
            {
                string l_Str = a_Buffer.ReadUTF8StringUInt16Len().ToLowerInvariant();
                a_Buffer.JumpAhead(sizeof(ushort));
                TagCodes l_TagCode = (TagCodes)a_Buffer.ReadByte();

                l_SearchTerms = SearchTerm.CreateMetaStrType(l_TagCode, l_Str);
            }
            else if (SearchTermOps.MinMax == l_Op)
            {
                SearchTermTypes l_CompOp = (SearchTermTypes)((byte)SearchTermTypes.OpEqual + a_Buffer.ReadByte());

                if (SearchTermTypes.OpNotEqual >= l_CompOp)
                {
                    Tag.ReadTagContext l_ReadTagCntxt = new Tag.ReadTagContext(false);
                    Tag l_MinMaxTag = Tag.ReadTagFromBuffer(a_Buffer, ref l_ReadTagCntxt);

                    if (null != l_MinMaxTag &&
                        (false == l_MinMaxTag.IsNumber() ||
                         false == l_MinMaxTag.IsFloat()))
                    {
                        l_SearchTerms = SearchTerm.CreateMinMaxType(l_CompOp, l_MinMaxTag);
                    }
                }
            }

            return l_SearchTerms;
        }

        private static SearchTerm ReadStringType(SearchTermTypes a_Type,
                                                 BufferIO a_Buffer)
        {
            string l_Str = a_Buffer.ReadUTF8StringUInt16Len().ToLowerInvariant();
            string[] l_Strings = l_Str.Split(Kademlia.s_InvKadKeywordChars, StringSplitOptions.RemoveEmptyEntries);

            if (0 == l_Strings.Length)
            {
                return null;
            }

            return SearchTerm.CreateStringsType(a_Type, l_Strings);
        }
    }

    class FileEntry : Entry<FileID, UserID>
    {
        // FileID (the id of the file).
        // UserID (the id of the user that published).

        public readonly TagList f_TagList;
        public readonly ushort f_MaxMinutes;

        public readonly I2PAddress f_StreamAddr;
        public readonly bool f_Complete;
        // f_FileSize can be zero.
        public readonly ulong f_FileSize;

        private FileEntry(FileID a_FileID,
                          UserID a_UserID,
                          TagList a_TagList,
                          I2PAddress a_StreamAddr,
                          bool a_Complete,
                          ulong a_FileSize) :
                     base(a_FileID, a_UserID)
        {
            f_TagList = a_TagList;
            f_MaxMinutes = a_TagList.GetAndDeleteMaxLifetime();

            f_StreamAddr = a_StreamAddr;
            f_Complete = a_Complete;
            f_FileSize = a_FileSize;
        }

        public static FileEntry Create(Kademlia a_Kad,
                                       KnownFile a_Known)
        {
            // Return can be null, if the file is not suitable for
            // publishing.

            FileID l_FileID;
            bool l_Completed = (ComplStatus.Complete == a_Known.p_Status);

            if (false == a_Known.GetFileIDVolatile(out l_FileID))
            {
                return null;
            }

            TagList l_TagList = new TagList();
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_SOURCETYPE, 1));
            // NOTE: We don't need to publish our own datagram address,
            // since the receiver knows it.
            l_TagList.AddTag(new AddressTag(TagCodes.TAG_SOURCE_STREAM_DEST, a_Kad.f_Self.p_StreamAddr));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_FLAGS, false == l_Completed ? 0u : 1u));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_FILESIZE, a_Known.f_FileSize));

            return new FileEntry(l_FileID,
                                 a_Kad.f_Inst.f_NetworkMngr.p_OwnPersona.f_UserID,
                                 l_TagList,
                                 a_Kad.f_Self.p_StreamAddr,
                                 l_Completed,
                                 a_Known.f_FileSize);
        }

        public static FileEntry FromTagList(Kademlia a_Kad,
                                            FileID a_FileID,
                                            UserID a_UserID,
                                            TagList a_TagList,
                                            I2PAddress a_ExpectedDatagramAddr)
        {
            if (false == a_UserID.IsValid())
            {
                return null;
            }

            ulong l_Type;
            if (false == a_TagList.TryGetTagAsULong(TagCodes.TAG_SOURCETYPE, out l_Type) ||
                // We accept only type 1.
                1 != l_Type)
            {
                return null;
            }

            I2PAddress l_DatagramAddr;

            if (true == a_TagList.TryGetTagAsAddress(TagCodes.TAG_SOURCE_DATAGRAM_DEST, out l_DatagramAddr))
            {
                if (null != a_ExpectedDatagramAddr &&
                    false == l_DatagramAddr.ValueEquals(a_ExpectedDatagramAddr))
                {
                    return null;
                }

                a_TagList.RemoveTag(TagCodes.TAG_SOURCE_DATAGRAM_DEST);
            }

            I2PAddress l_StreamAddr;

            if (false == a_TagList.TryGetTagAsAddress(TagCodes.TAG_SOURCE_STREAM_DEST, out l_StreamAddr) ||
                // We don't need to check the passed datagram
                // destination, but the stream destination should be
                // checked.
                true == a_Kad.f_Inst.f_NetworkMngr.IsAddressBanned(l_StreamAddr))
            {
                return null;
            }

            uint l_Flags;

            if (false == a_TagList.TryGetTagAsUInt(TagCodes.TAG_FLAGS, out l_Flags))
            {
                return null;
            }

            bool l_Complete = (0 != ((TagFlags)l_Flags & TagFlags.TAG_FLAGS_COMPLETEFILE));
            ulong l_FileSize;

            if (false == a_TagList.TryGetTagAsULong(TagCodes.TAG_FILESIZE, out l_FileSize))
            {
                l_FileSize = 0;
            }
            else if (false == ContentManager.IsValidFileSize(l_FileSize))
            {
                return null;
            }

            return new FileEntry(a_FileID,
                                 a_UserID,
                                 a_TagList,
                                 l_StreamAddr,
                                 l_Complete,
                                 l_FileSize);
        }

        public LeanFileEntry MakeLean(Tag.UnicodeSupport a_UnicodeSupport)
        {
            byte[] l_BinaryTagList = TagListToBytes(f_TagList, a_UnicodeSupport);

            if (null == l_BinaryTagList)
            {
                return null;
            }

            return new LeanFileEntry(this, l_BinaryTagList);
        }

        public class LeanFileEntry : Entry<FileID, UserID>
        {
            public readonly bool f_Complete;
            // f_FileSize can be zero.
            public readonly ulong f_FileSize;
            public readonly bool f_ClassicAddr;
            private readonly byte[] f_BinaryTagList;

            public LeanFileEntry(FileEntry a_Entry,
                                 byte[] a_BinaryTagList) :
                            base(a_Entry.f_KeyID, a_Entry.f_RefID)
            {
                f_Complete = a_Entry.f_Complete;
                f_FileSize = a_Entry.f_FileSize;
                f_ClassicAddr = a_Entry.f_StreamAddr.p_IsClassicAddress;
                f_BinaryTagList = a_BinaryTagList;
            }

            public override uint GetMemorySize()
            {
                return (base.GetMemorySize(2) +
                        (uint)f_BinaryTagList.Length);
            }

            public override void TagListToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteBytes(f_BinaryTagList);
            }

            public I2PAddress GetStreamAddress()
            {
                MemoryIO l_Buffer = MemoryIO.CreateReadable(f_BinaryTagList);
                TagList l_TagList = TagList.ReadListTrusted(l_Buffer, false);

                return l_TagList.GetTag(TagCodes.TAG_SOURCE_STREAM_DEST).GetAddr();
            }
        }
    }

    class KeywordEntry : Entry<KeywordID, FileID>
    {
        // KeywordID (the MD4 of one of the keywords in the file name).
        // FileID (the id of the referenced file).

        public const uint c_MaxPlausibleSources = 1000000;

        public readonly TagList f_TagList;
        public readonly ushort f_MaxMinutes;

        // f_FileName is guaranteed to be legal.
        public readonly string f_FileName;
        public readonly ulong f_FileSize;
        public readonly ushort f_CompleteSources;
        public readonly ushort f_Sources;

        private KeywordEntry(KeywordID a_KeywordID,
                             FileID a_FileID,
                             TagList a_TagList,
                             string a_FileName,
                             ulong a_FileSize,
                             ushort a_CompleteSources,
                             ushort a_Sources) :
                        base(a_KeywordID, a_FileID)
        {
            f_TagList = a_TagList;
            f_MaxMinutes = a_TagList.GetAndDeleteMaxLifetime();

            f_FileName = a_FileName;
            f_FileSize = a_FileSize;
            f_CompleteSources = a_CompleteSources;
            f_Sources = a_Sources;
        }

        public static KeywordEntry Create(Kademlia a_Kad,
                                          KeywordID a_KeywordID,
                                          KnownFile a_Known,
                                          FileID a_FileID,
                                          uint a_MinCompleteSources,
                                          uint a_MinSources)
        {
            // Return can be null, if the file is not suitable for
            // publishing.

            ComplStatus l_Status;
            string l_ParentPath;
            string l_FileName = a_Known.GetFileNameVolatile(false, out l_Status, out l_ParentPath);

            if (0 >= l_FileName.Length)
            {
                l_FileName = FileNames.c_FallbackFileName;
            }

            uint l_ComplSources = 0;
            uint l_Sources = 0;

            if (ComplStatus.Incomplete == l_Status)
            {
                IncomplFileProgress l_IncomplProgress = a_Known.p_IncomplProgressVolatile;

                if (null == l_IncomplProgress)
                {
                    return null;
                }

                if (true == l_IncomplProgress.f_DoneMap.p_AllSet)
                {
                    l_ComplSources++;
                    l_Sources++;
                }
                else if (true == l_IncomplProgress.f_DoneMap.Has8MiBBitmapeD2kRelevance(a_Known.f_FileSize))
                {
                    l_Sources++;
                }
            }
            else if (ComplStatus.Complete == l_Status)
            {
                l_ComplSources++;
                l_Sources++;
            }
            else
            {
                return null;
            }

            KnownFile.ActivityContext l_ActCntxt = a_Known.p_ActCntxt;

            if (null != l_ActCntxt)
            {
                uint l_CompleteNodes = l_ActCntxt.p_CompleteNodes;
                l_ComplSources += l_CompleteNodes;
                l_Sources += (l_CompleteNodes + l_ActCntxt.p_PartialNodes);
            }

            l_ComplSources = Math.Max(l_ComplSources, a_MinCompleteSources);
            l_Sources = Math.Max(l_Sources, a_MinSources);

            TagList l_TagList = new TagList();
            l_TagList.AddTag(new StringTag(TagCodes.TAG_FILENAME, l_FileName));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_FILESIZE, a_Known.f_FileSize));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_COMPLETE_SOURCES, l_ComplSources));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_SOURCES, l_Sources));

            // iMule adds tags like TAG_FILETYPE and TAG_FILEFORMAT,
            // but it just takes them from the file's extension.
            // Since anyone can do that, we don't bother sending them.

            return new KeywordEntry(a_KeywordID,
                                    a_FileID,
                                    l_TagList,
                                    l_FileName,
                                    a_Known.f_FileSize,
                                    (ushort)l_ComplSources,
                                    (ushort)l_Sources);
        }

        public static KeywordEntry FromTagList(Kademlia a_Kad,
                                               KeywordID a_KeywordID,
                                               FileID a_FileID,
                                               TagList a_TagList)
        {
            string l_FileName;

            if (false == a_TagList.TryGetTagAsString(TagCodes.TAG_FILENAME, out l_FileName))
            {
                return null;
            }

            l_FileName = Util.FixUTF8Latin1ConversionErrors(l_FileName);
            l_FileName = FileNames.TryToMakeFileNameLegal(l_FileName);

            if (null == l_FileName)
            {
                // From time to time there are empty file names.

                return null;
            }

            ulong l_FileSize;

            if (false == a_TagList.TryGetTagAsULong(TagCodes.TAG_FILESIZE, out l_FileSize) ||
                false == ContentManager.IsValidFileSize(l_FileSize))
            {
                return null;
            }

            uint l_CompleteSources;
            uint l_Sources;
            if (false == a_TagList.TryGetTagAsUInt(TagCodes.TAG_COMPLETE_SOURCES, out l_CompleteSources) ||
                false == a_TagList.TryGetTagAsUInt(TagCodes.TAG_SOURCES, out l_Sources))
            {
                return null;
            }

            if (0 == l_Sources ||
                c_MaxPlausibleSources < l_Sources ||
                l_Sources < l_CompleteSources)
            {
                return null;
            }

            if (false == a_Kad.f_SearchMngr.VerifyKeyword(a_KeywordID.f_Val, l_FileName))
            {
                return null;
            }

            return new KeywordEntry(a_KeywordID,
                                    a_FileID,
                                    a_TagList,
                                    l_FileName,
                                    l_FileSize,
                                    (ushort)Math.Min(l_CompleteSources, ushort.MaxValue),
                                    (ushort)Math.Min(l_Sources, ushort.MaxValue));
        }

        public LeanKeywordEntry MakeLean(Tag.UnicodeSupport a_UnicodeSupport)
        {
            byte[] l_BinaryTagList = TagListToBytes(f_TagList, a_UnicodeSupport);

            if (null == l_BinaryTagList)
            {
                return null;
            }

            return new LeanKeywordEntry(this, l_BinaryTagList);
        }

        public class LeanKeywordEntry : Entry<KeywordID, FileID>
        {
            public readonly string f_FileName;
            public readonly TagList f_TagList;
            private readonly byte[] f_BinaryTagList;

            public LeanKeywordEntry(KeywordEntry a_Entry,
                                    byte[] a_BinaryTagList) :
                               base(a_Entry.f_KeyID, a_Entry.f_RefID)
            {
                f_FileName = a_Entry.f_FileName;
                f_TagList = a_Entry.f_TagList;
                f_BinaryTagList = a_BinaryTagList;
            }

            public override uint GetMemorySize()
            {
                return (base.GetMemorySize(3) +
                        ((uint)f_FileName.Length * sizeof(char)) +
                        // For f_TagList and f_BinaryTagList together.
                        ((uint)f_BinaryTagList.Length * 2));
            }

            public override void TagListToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteBytes(f_BinaryTagList);
            }

            public void TagListToBufferWithCounts(Tag.UnicodeSupport a_UnicodeSupport,
                                                  uint a_CompleteSources,
                                                  uint a_Sources,
                                                  BufferIO a_DestBuf)
            {
                MemoryIO l_SrcBuf = MemoryIO.CreateReadable(f_BinaryTagList);
                TagList l_TagList = TagList.ReadListTrusted(l_SrcBuf, 0 != ((uint)Tag.UnicodeSupport.ImplicitUTF8 & (uint)a_UnicodeSupport));

                l_TagList.RemoveTag(TagCodes.TAG_COMPLETE_SOURCES);
                l_TagList.RemoveTag(TagCodes.TAG_SOURCES);
                l_TagList.AddTag(new NumberTag(TagCodes.TAG_COMPLETE_SOURCES, Math.Min(ushort.MaxValue, a_CompleteSources)));
                l_TagList.AddTag(new NumberTag(TagCodes.TAG_SOURCES, Math.Min(ushort.MaxValue, a_Sources)));

                l_TagList.WriteTo(a_DestBuf, sizeof(byte), a_UnicodeSupport);
            }
        }
    }

    class NoteEntry : Entry<FileID, ClientID>
    {
        // FileID (the id of the file that is commented on).
        // ClientID (the client id of the client that comments it).

        public readonly TagList f_TagList;
        public readonly ushort f_MaxMinutes;

        public readonly string f_FileName;
        // f_FileSize can be zero.
        public readonly ulong f_FileSize;
        // At least one of the following two must contain
        // nonzero data.
        public readonly FileRatings f_Rating;
        public readonly string f_Description;

        private NoteEntry(FileID a_FileID,
                          ClientID a_ClientID,
                          TagList a_TagList,
                          string a_FileName,
                          ulong a_FileSize,
                          FileRatings a_Rating,
                          string a_Description) :
                     base(a_FileID, a_ClientID)
        {
            f_TagList = a_TagList;
            f_MaxMinutes = a_TagList.GetAndDeleteMaxLifetime();

            f_FileName = a_FileName;
            f_FileSize = a_FileSize;
            f_Rating = a_Rating;
            f_Description = a_Description;
        }

        public static NoteEntry Create(Kademlia a_Kad,
                                       KnownFile a_Known)
        {
            // Return can be null, if the file is not suitable for
            // publishing.

            bool l_HasNote = (FileRatings.NotRated != a_Known.f_OwnRating ||
                              null != a_Known.f_OwnComment);
            FileID l_FileID;

            if (false == l_HasNote ||
                false == a_Known.GetFileIDVolatile(out l_FileID))
            {
                return null;
            }

            string l_FileName = a_Known.GetFileNameVolatile(false);

            if (0 >= l_FileName.Length)
            {
                l_FileName = FileNames.c_FallbackFileName;
            }

            TagList l_TagList = new TagList();
            l_TagList.AddTag(new StringTag(TagCodes.TAG_FILENAME, l_FileName));
            l_TagList.AddTag(new NumberTag(TagCodes.TAG_FILESIZE, a_Known.f_FileSize));

            if (FileRatings.NotRated != a_Known.f_OwnRating)
            {
                l_TagList.AddTag(new NumberTag(TagCodes.TAG_FILERATING, (byte)a_Known.f_OwnRating));
            }

            if (null != a_Known.f_OwnComment)
            {
                l_TagList.AddTag(new StringTag(TagCodes.TAG_DESCRIPTION, a_Known.f_OwnComment));
            }

            return new NoteEntry(l_FileID,
                                 a_Kad.f_Self.f_ClientID,
                                 l_TagList,
                                 l_FileName,
                                 a_Known.f_FileSize,
                                 a_Known.f_OwnRating,
                                 a_Known.f_OwnComment);
        }

        public static NoteEntry FromTagList(Kademlia a_Kad,
                                            FileID a_FileID,
                                            ClientID a_ClientID,
                                            TagList a_TagList)
        {
            string l_FileName;

            if (false == a_TagList.TryGetTagAsString(TagCodes.TAG_FILENAME, out l_FileName))
            {
                return null;
            }

            l_FileName = Util.FixUTF8Latin1ConversionErrors(l_FileName);
            l_FileName = FileNames.TryToMakeFileNameLegal(l_FileName);

            if (null == l_FileName)
            {
                return null;
            }

            ulong l_FileSize;

            if (false == a_TagList.TryGetTagAsULong(TagCodes.TAG_FILESIZE, out l_FileSize))
            {
                l_FileSize = 0;
            }
            else if (false == ContentManager.IsValidFileSize(l_FileSize))
            {
                return null;
            }

            FileRatings l_Rating;
            ulong l_RatingNum;

            if (false == a_TagList.TryGetTagAsULong(TagCodes.TAG_FILERATING, out l_RatingNum))
            {
                l_Rating = FileRatings.NotRated;
            }
            else
            {
                if ((byte)FileRatings.Excellent < l_RatingNum)
                {
                    return null;
                }

                l_Rating = (FileRatings)l_RatingNum;
            }

            string l_Desc;

            if (false == a_TagList.TryGetTagAsString(TagCodes.TAG_DESCRIPTION, out l_Desc) ||
                false == a_Kad.TrimAndVerifyComment(l_Desc, out l_Desc))
            {
                return null;
            }

            if (FileRatings.NotRated == l_Rating &&
                null == l_Desc)
            {
                // At least one of the two must be present.

                return null;
            }

            return new NoteEntry(a_FileID,
                                 a_ClientID,
                                 a_TagList,
                                 l_FileName,
                                 l_FileSize,
                                 l_Rating,
                                 l_Desc);
        }

        public LeanNoteEntry MakeLean(Tag.UnicodeSupport a_UnicodeSupport)
        {
            byte[] l_BinaryTagList = TagListToBytes(f_TagList, a_UnicodeSupport);

            if (null == l_BinaryTagList)
            {
                return null;
            }

            return new LeanNoteEntry(this, l_BinaryTagList);
        }

        public class LeanNoteEntry : Entry<FileID, ClientID>
        {
            public readonly string f_FileName;
            private readonly byte[] f_BinaryTagList;

            public LeanNoteEntry(NoteEntry a_Entry,
                                 byte[] a_BinaryTagList) :
                            base(a_Entry.f_KeyID, a_Entry.f_RefID)
            {
                f_FileName = a_Entry.f_FileName;
                f_BinaryTagList = a_BinaryTagList;
            }

            public override uint GetMemorySize()
            {
                return (base.GetMemorySize(2) +
                        ((uint)f_FileName.Length * sizeof(char)) +
                        (uint)f_BinaryTagList.Length);
            }

            public override void TagListToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteBytes(f_BinaryTagList);
            }
        }
    }

    class Entry<TKeyID, TRefID>
          where TKeyID : IUInt128
          where TRefID : IUInt128
    {
        public readonly TKeyID f_KeyID;
        public readonly TRefID f_RefID;

        private Time f_Lifetime;

        public Time p_Lifetime
        {
            get { return f_Lifetime; }
        }

        protected Entry(TKeyID a_KeyID,
                        TRefID a_RefID)
        {
            f_KeyID = a_KeyID;
            f_RefID = a_RefID;
        }

        public override int GetHashCode()
        {
            return f_KeyID.GetHashCode();
        }

        protected byte[] TagListToBytes(TagList a_TagList,
                                        Tag.UnicodeSupport a_UnicodeSupport)
        {
            // Return can be null.

            byte[] l_BinaryTagList = a_TagList.TagListToBytes(a_UnicodeSupport);

            if (1024 < l_BinaryTagList.Length)
            {
                return null;
            }

            return l_BinaryTagList;
        }

        public virtual uint GetMemorySize()
        {
            throw (new ApplicationException());
        }

        protected uint GetMemorySize(uint a_ExtraPointerSizes)
        {
            // Returns the rough net memory usage of the class.

            return (((uint)IntPtr.Size * (2 + a_ExtraPointerSizes)) +
                    (UInt128.c_ByteSize * 2) +
                    sizeof(ulong));
        }

        public virtual void TagListToBuffer(BufferIO a_Buffer)
        {
            throw (new ApplicationException());
        }

        public void SetLifetime(Time a_Lifetime)
        {
            f_Lifetime = a_Lifetime;
        }
    }

    enum KadParams : byte
    {
        // KADEMLIA (parameter)
        KADEMLIA_FIND_VALUE = 2,
        KADEMLIA_STORE = 4,
        KADEMLIA_FIND_NODE = 11,
    }

    class KadStats
    {
        public uint f_KadRequests;

        public uint f_SourcePublishRequestsRcvd;
        public uint f_KeywordPublishRequestsRcvd;
        public uint f_NotesPublishRequestsRcvd;

        public uint f_PublishRequestsSent;
        public uint f_PublishResponsesRcvd;

        public uint f_KadDatagramsReceived;
        public uint f_KadDatagramsSent;

        public ulong f_KadDataVolumeReceived;
        public ulong f_KadDataVolumeSent;

        public uint f_OddPublishReq;
        public uint f_CorruptValues;
        public uint f_UnprocessableSearchResults;

        public KadStats()
        {
        }

        public void NotifyOddPublishRequest()
        {
            f_OddPublishReq++;
        }

        public void NotifyUntrustedPublishRequest()
        {
        }

        public void NotifyCorruptValues()
        {
            f_CorruptValues++;
        }

        public void NotifyMalformedKadReq()
        {
        }

        public void NotifyRejectedKeywordSearchResult()
        {
        }

        public void NotifyPeerHasMyAddressesWrong()
        {
        }

        public void NotifyUnprocessableSearchResult()
        {
            f_UnprocessableSearchResults++;
        }

        public void NotifyCorruptDataSize()
        {
        }

        public void NotifyReceivedDatagram(uint a_Size)
        {
            f_KadDatagramsReceived++;
            f_KadDataVolumeReceived += a_Size;
        }

        public void NotifySentDatagram(uint a_Size)
        {
            f_KadDatagramsSent++;
            f_KadDataVolumeSent += a_Size;
        }
    }

    class ContactManager
    {
        private readonly Dictionary<ClientID, Contact> f_Contacts;
        private readonly Dictionary<I2PAddress, Contact> f_ContactsByDatagramAddr;
        private uint f_VerifiedCount;

        public uint p_Count
        {
            get { return (uint)f_Contacts.Count; }
        }

        public uint p_VerifiedCount
        {
            get { return f_VerifiedCount; }
        }

        public ContactManager()
        {
            f_Contacts = new Dictionary<ClientID, Contact>();
            f_ContactsByDatagramAddr = new Dictionary<I2PAddress, Contact>();
        }

        public IEnumerable<Contact> Enumerable()
        {
            return f_Contacts.Values;
        }

        public Contact[] GetBootstrapContactsForPeer(bool a_RestrictToClassicAddresses)
        {
            // Copy all verified contacts and randomize the
            // order afterwards, thus we'll have different
            // results for each call.

            Contact[] l_Arr = new Contact[f_VerifiedCount];
            int l_Idx = 0;

            foreach (Contact l_Contact in f_Contacts.Values)
            {
                if (true == a_RestrictToClassicAddresses &&
                    false == l_Contact.p_HasClassicAddresses)
                {
                    continue;
                }

                if (true == l_Contact.p_IsVerified)
                {
                    l_Arr[l_Idx++] = l_Contact;

                    if (l_Idx == f_VerifiedCount)
                    {
                        break;
                    }
                }
            }

            Util.RandomizeArrayOrder<Contact>(l_Arr, 0, l_Idx);

            return l_Arr;
        }

        public void NotifyActivity(Time a_Time,
                                   Contact a_Contact)
        {
            if (false == a_Contact.p_IsVerified)
            {
                f_VerifiedCount++;
            }

            a_Contact.NotifyActivity(a_Time);
        }

        public void Add(Contact a_Contact)
        {
            f_Contacts.Add(a_Contact.f_ClientID, a_Contact);
            f_ContactsByDatagramAddr.Add(a_Contact.p_DatagramAddr, a_Contact);

            if (true == a_Contact.p_IsVerified)
            {
                f_VerifiedCount++;
            }
        }

        public void Remove(Contact a_Contact)
        {
            f_Contacts.Remove(a_Contact.f_ClientID);
            f_ContactsByDatagramAddr.Remove(a_Contact.p_DatagramAddr);

            if (true == a_Contact.p_IsVerified)
            {
                f_VerifiedCount--;
            }
        }

        public void Remove(I2PAddress a_DatagramAddr)
        {
            Contact l_Contact = GetContact(a_DatagramAddr);

            if (null != l_Contact)
            {
                Remove(l_Contact);
            }
        }

        public void Reinitialize(Contact a_Self)
        {
            // Removes any contact that might have any form of
            // conflict with us. Also sets each remaining
            // contact's vitality to alive.

            List<Contact> l_RemLs = null;

            foreach (Contact l_Contact in f_Contacts.Values)
            {
                if (false == a_Self.HasAnyConflict(l_Contact))
                {
                    l_Contact.SetVitality(ContactVitality.Alive);
                }
                else
                {
                    if (null == l_RemLs)
                    {
                        l_RemLs = new List<Contact>(1);
                    }

                    l_RemLs.Add(l_Contact);
                }
            }

            if (null != l_RemLs)
            {
                foreach (Contact l_Contact in l_RemLs)
                {
                    Remove(l_Contact);
                }
            }
        }

        public bool HasContact(ClientID a_ClientID)
        {
            return f_Contacts.ContainsKey(a_ClientID);
        }

        public Contact GetContact(ClientID a_ClientID)
        {
            Contact l_Contact;
            if (true == f_Contacts.TryGetValue(a_ClientID, out l_Contact))
            {
                return l_Contact;
            }

            return null;
        }

        public Contact GetContact(I2PAddress a_DatagramAddr)
        {
            Contact l_Contact;
            if (true == f_ContactsByDatagramAddr.TryGetValue(a_DatagramAddr, out l_Contact))
            {
                return l_Contact;
            }

            return null;
        }

        public bool UpdateAddresses(Contact a_Contact,
                                    I2PAddress a_DatagramAddr,
                                    I2PAddress a_StreamAddr)
        {
            Contact l_ExistingContact = GetContact(a_Contact.p_DatagramAddr);

            if (null == l_ExistingContact)
            {
                f_ContactsByDatagramAddr.Remove(a_Contact.p_DatagramAddr);
                f_ContactsByDatagramAddr.Add(a_DatagramAddr, a_Contact);

                a_Contact.SetDatagramAddr(a_DatagramAddr);
                a_Contact.SetStreamAddr(a_StreamAddr);
            }
            else if (a_Contact == l_ExistingContact)
            {
                a_Contact.SetStreamAddr(a_StreamAddr);
            }
            else
            {
                // Another client has this datagram destination
                // already in use.

                return false;
            }

            return true;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct GuiContactInfo
    {
        public readonly ClientID f_ClientID;
        public readonly Distance f_Distance;
        public readonly Time f_LastActivity;

        public GuiContactInfo(ClientID a_ClientID,
                              Distance a_Distance,
                              Time a_LastActivity)
        {
            f_ClientID = a_ClientID;
            f_Distance = a_Distance;
            f_LastActivity = a_LastActivity;
        }
    }

    enum KademliaStatus : byte
    {
        // No datagram handling at all.
        Disconnected = 0,
        // No confirmed contact so far, but datagrams are transported.
        TryingToConnect = 1,
        // At least one contact has answered.
        Connected = 2,
    }

    class Kademlia
    {
        // Limit is in UTF-16 code units.
        public const ushort c_MaxCommentChars = 96;

        private const ushort c_BootstrapContactThreshold = 32;
        public const ushort c_MaxContactsPerBuffer = 500;
        private const uint c_NodesMagic = 0xE949ED22;
        private const uint c_NodesHeaderSize = 512;

        public static readonly char[] s_InvKadKeywordChars = new char[]
        {
            // Default iMule chars.
            ' ', '(', ')', '[', ']', '{', '}', '<', '>', ',',
            '.', '_', '-', '!', '?', ':', ';', '\\', '/',
            // Chars used by the string.Trim() routine (less space).
            '\x0009', '\x000A', '\x000B', '\x000C', '\x000D',
            '\x0085', '\x00A0', '\x1680', '\x2000', '\x2001',
            '\x2002', '\x2003', '\x2004', '\x2005', '\x2006',
            '\x2007', '\x2008', '\x2009', '\x200A', '\x200B',
            '\x2028', '\x2029', '\x3000', '\xFEFF'
        };

        public readonly Instance f_Inst;

        public KadPrefs f_KadPrefs;
        private string[] f_CommentsFilterArray;
        private KademliaStatus f_Status;
        public readonly Contact f_Self;

        private readonly Dictionary<RoutingZone, RoutingZone> f_TimedZones;

        public readonly KadStats f_Stats;
        public readonly KadMap f_Map;
        public readonly SearchManager f_SearchMngr;
        public readonly ContactManager f_ContactMngr;
        public readonly PublishManager f_PublishMngr;
        public readonly PacketHandler f_PacketHndlr;
        public readonly RoutingZone f_RootZone;

        private Time f_IgnoreGuiTime;

        private Time f_NextBootstrapTime;
        private Time f_NextMapCleanTime;

        private Time f_NextSearchTime;
        private Time f_NextZoneEventsTime;
        private Time f_NextPublishTime;
        private Time f_NextTempContactCleanTime;
        private Time f_NextSelfLookupTime;

        private bool f_Bootstrapping;
        private uint f_GeneralPeersGeneration;

        public KademliaStatus p_KadStatus
        {
            get { return f_Status; }
        }

        public KademliaStatus p_KadStatusVolatile
        {
            get { return Util.VolatileRead(ref f_Status); }
        }

        public Kademlia(Instance a_Inst,
                        ref KadPrefs a_KadPrefs,
                        byte[] a_InitialNodes)
        {
            // a_InitialNodes may be null.

            f_Inst = a_Inst;

            // Must be done early.
            f_KadPrefs = a_KadPrefs;
            f_Self = new Contact(f_KadPrefs.f_ClientID);
            f_TimedZones = new Dictionary<RoutingZone, RoutingZone>();

            f_Stats = new KadStats();
            f_Map = new KadMap(this);
            f_SearchMngr = new SearchManager(this);
            f_ContactMngr = new ContactManager();
            f_PublishMngr = new PublishManager(this);
            // PacketHandler and RoutingZone rely on other
            // initialization to be done before construction.
            f_PacketHndlr = new PacketHandler(this);
            f_RootZone = new RoutingZone(this);

            Time l_Time = f_Inst.GetTime();

            f_NextBootstrapTime = l_Time.AddSeconds(3 * Time.c_Second);
            f_NextZoneEventsTime = l_Time.AddSeconds(1 * Time.c_SecsPerMinute);
            f_NextPublishTime = l_Time;
            f_NextMapCleanTime = l_Time.AddSeconds(Time.c_SecsPerHalfHour);
            f_NextSelfLookupTime = l_Time.AddSeconds(5 * Time.c_SecsPerMinute);

            // When loading nodes, we first use the a_InitialNodes
            // we got from a local file ('nodes.dat'), since they
            // are propably not up to date. After that we load the
            // more recent nodes from our internal cache (if present).

            if (null != a_InitialNodes)
            {
                MemoryIO l_Buffer = MemoryIO.CreateReadable(a_InitialNodes);
                bool l_Succ = LoadNodesFromBuffer(l_Buffer, false);
            }

            StorageStates l_State = LoadNodes();

            if (StorageStates.Success != l_State &&
                StorageStates.NoData != l_State)
            {
                Util.DebugBreak();
            }
        }

        private bool KadOffline()
        {
            return (KademliaStatus.Disconnected == f_Status);
        }

        public void NotifyNewSamConnection(Time a_Time)
        {
            // This routine gets called by the network manager.

            f_Status = KademliaStatus.TryingToConnect;
            f_Self.SetDatagramAddr(f_Inst.f_NetworkMngr.p_DatagramAddr);
            f_Self.SetStreamAddr(f_Inst.f_NetworkMngr.p_StreamAddr);

            f_PacketHndlr.NotifyAddressUpdate();

            f_ContactMngr.Reinitialize(f_Self);
            f_RootZone.Reinitialize();

            // Give the remaining contacts some headstart.
            f_NextTempContactCleanTime = a_Time.AddSeconds(10 * Time.c_SecsPerMinute);
        }

        public void NotifySamConnectionBroken()
        {
            f_SearchMngr.StopAllSearches();
            f_PublishMngr.StopPublishing();

            f_Status = KademliaStatus.Disconnected;
            f_Self.SetDatagramAddr(null);
            f_Self.SetStreamAddr(null);
        }

        public void UpdatePreferences(ref KadPrefs a_KadPrefs)
        {
            f_KadPrefs = a_KadPrefs;
            f_CommentsFilterArray = f_KadPrefs.BuildCommentFilterArray();
            f_Map.SetMaxMemoryConsumption(f_KadPrefs.f_MaxKadMapSize);
            f_Map.AdjustUnicodeSupport(f_KadPrefs.p_PublUnicodeSupport);

            if (KademliaStatus.Connected == f_Status)
            {
                f_PublishMngr.UpdateFlags(false, f_KadPrefs.f_PublFlags);
            }
            else
            {
                f_PublishMngr.StopPublishing();
            }
        }

        public bool TrimAndVerifyComment(string a_Comment,
                                         out string a_TrimmedComment)
        {
            // a_TrimmedComment may reference the same string object as
            // a_Comment on input. a_TrimmedComment can be null on
            // return.

            a_TrimmedComment = a_Comment.Trim();

            if (0 == a_TrimmedComment.Length)
            {
                a_TrimmedComment = null;

                return true;
            }

            return (Kademlia.c_MaxCommentChars >= a_TrimmedComment.Length &&
                    false == KadPrefs.IsCommentFiltered(a_TrimmedComment, f_CommentsFilterArray));
        }

        public void NotifyBannedAddress(I2PAddress a_BannedAddr)
        {
            f_Inst.f_MainForm.AddToLog(f_Inst.f_TextSet.Format(TextKeys.MsgDestBanned, a_BannedAddr.ToShortString()));

            f_Inst.f_Kad.f_ContactMngr.Remove(a_BannedAddr);

            // NOTE: This code will only remove contacts that have the
            // passed destination address as datagram address, while
            // contacts that have the same stream address will survive
            // for now (a timer controlled cleanup scan is used to get
            // rid of such contacts).
            // We will not remove published entries with this address
            // from the kademlia storage (however re-publish attempts with
            // this address (for datagrams or streams) get blocked in the
            // FileEntry class).
        }

        public void Process(Time a_Time)
        {
            UpdateGui(a_Time);

            if (BridgeStatus.Connected != f_Inst.f_NetworkMngr.p_BridgeStatus)
            {
                return;
            }

            if (0 == f_ContactMngr.p_VerifiedCount)
            {
                if (KademliaStatus.TryingToConnect != f_Status)
                {
                    f_Status = KademliaStatus.TryingToConnect;
                    f_PublishMngr.StopPublishing();
                }
            }
            else
            {
                if (KademliaStatus.Connected != f_Status)
                {
                    f_Status = KademliaStatus.Connected;
                    f_PublishMngr.UpdateFlags(true, f_KadPrefs.f_PublFlags);
                }
            }

            if (true == f_NextBootstrapTime.IsEarlierThan(a_Time))
            {
                f_Bootstrapping = Bootstrap(a_Time);

                f_NextBootstrapTime = a_Time.AddSeconds(Time.c_SecsPerMinute + (Time.c_Second * 10));
            }

            if (true == f_NextMapCleanTime.IsEarlierThan(a_Time))
            {
                f_Map.Clean(a_Time);

                f_NextMapCleanTime = a_Time.AddSeconds(Time.c_SecsPerHalfHour);
            }

            while (true)
            {
                Packet l_Packet = f_Inst.f_UpDwnMngr.DequeueDatagram();

                if (null == l_Packet)
                {
                    break;
                }

                if (Protocols.OP_KADEMLIAHEADER == l_Packet.f_Prot)
                {
                    f_PacketHndlr.ProcessPacket(a_Time, l_Packet);
                }
            }

            if (true == f_NextSearchTime.IsEarlierThan(a_Time))
            {
                f_SearchMngr.Process(a_Time);

                f_NextSearchTime = a_Time.AddSeconds(5 * Time.c_Second);
            }

            if (true == f_NextPublishTime.IsEarlierThan(a_Time))
            {
                f_PublishMngr.Process(a_Time);

                f_NextPublishTime = a_Time.AddSeconds(5);
            }

            if (true == f_NextZoneEventsTime.IsEarlierThan(a_Time))
            {
                HandleZoneEvents(a_Time);

                f_NextZoneEventsTime = a_Time.AddSeconds(Time.c_SecsPerMinute);
            }

            if (true == f_NextTempContactCleanTime.IsEarlierThan(a_Time))
            {
                Time l_HalfHourAgo = a_Time.SubSeconds(Time.c_SecsPerHalfHour);

                List<Contact> l_RemoveList = new List<Contact>();

                foreach (Contact l_Contact in f_ContactMngr.Enumerable())
                {
                    if ((false == l_Contact.p_IsInRoutingZone &&
                         false == l_Contact.IsPinned(a_Time) &&
                         false == l_Contact.HadActivitySince(l_HalfHourAgo)) ||
                        true == f_Inst.f_NetworkMngr.IsAddressBanned(l_Contact.p_StreamAddr))
                    {
                        l_RemoveList.Add(l_Contact);
                    }
                }

                foreach (Contact l_Contact in l_RemoveList)
                {
                    f_ContactMngr.Remove(l_Contact);
                }

                f_NextTempContactCleanTime = a_Time.AddSeconds(10 * Time.c_SecsPerMinute);
            }

            if (true == f_NextSelfLookupTime.IsEarlierThan(a_Time))
            {
                f_SearchMngr.FindNode(f_Self.f_ClientID, true);

                f_NextSelfLookupTime = a_Time.AddSeconds(Time.c_SecsPerHalfHour);
            }
        }

        private bool Bootstrap(Time a_Time)
        {
            if (c_BootstrapContactThreshold <= f_ContactMngr.p_VerifiedCount)
            {
                return false;
            }

            if (0 == f_ContactMngr.p_Count)
            {
                // Inform the GUI.

                f_Inst.InvokeMainFormAsync(new ArglessHandler(f_Inst.f_MainForm.NotifyConnectedButWithoutKademilaContacts));
            }
            else
            {
                List<Contact> l_List = f_RootZone.GetBootstrapContactsForMyself(256);

                foreach (Contact l_Contact in l_List)
                {
                    f_PacketHndlr.SendBootstrapRequest(a_Time, l_Contact);
                }
            }

            return true;
        }

#region RoutingZoneEvents

        private void HandleZoneEvents(Time a_Time)
        {
            List<RoutingZone> l_MergeList = new List<RoutingZone>(f_TimedZones.Count);

            foreach (RoutingZone l_Zone in f_TimedZones.Values)
            {
                if (true == l_Zone.f_SmallTime.IsEarlierThan(a_Time))
                {
                    bool l_TryMergeParent;
                    l_Zone.OnSmallTimer(a_Time, f_Bootstrapping, out l_TryMergeParent);
                    l_Zone.f_SmallTime = a_Time.AddSeconds(Time.c_SecsPerMinute + Time.c_SecsPerHalfMinute);

                    if (true == l_TryMergeParent)
                    {
                        l_MergeList.Add(l_Zone);
                    }
                }

                if (true == l_Zone.f_BigTime.IsEarlierThan(a_Time))
                {
                    l_Zone.OnBigTimer();
                    l_Zone.f_BigTime = a_Time.AddSeconds(Time.c_SecsPerHalfHour);
                }
            }

            // We can't merge while processing a timer event,
            // so we try now.

            foreach (RoutingZone a_Zone in l_MergeList)
            {
                RoutingZone.TryMergeParent(a_Zone);
            }
        }

        public void AddEvent(RoutingZone a_Zone)
        {
            Time l_Now = f_Inst.GetTime();
            uint l_Extra = (((5u + a_Zone.f_Level) * Time.c_SecsPerMinute) + (10 * Time.c_Second));

            a_Zone.f_SmallTime = l_Now.AddSeconds(Time.c_SecsPerMinute);
            a_Zone.f_BigTime = l_Now.AddSeconds(l_Extra);

            f_TimedZones[a_Zone] = a_Zone;
        }

        public void RemoveEvent(RoutingZone a_Zone)
        {
            bool OK = f_TimedZones.Remove(a_Zone);
        }

#endregion

        private void UpdateGui(Time a_Time)
        {
            MainForm.FormInput l_Input = f_Inst.f_MainForm.f_Input;
            Time l_NextGuiUpdateTime = Time.FromBinary((ulong)System.Threading.Interlocked.Read(ref l_Input.f_NextNetworkUpdateTime));

            if (false == l_NextGuiUpdateTime.Equals(f_IgnoreGuiTime) &&
                true == l_NextGuiUpdateTime.IsEarlierThan(a_Time))
            {
                MainForm.FormInput.NetworkDataTypes l_Type = Util.VolatileRead(ref l_Input.f_SubscribedNetDataTypes);

                if (0 != (MainForm.FormInput.NetworkDataTypes.Contacts & l_Type))
                {
                    GuiContactInfo[] l_Arr = new GuiContactInfo[f_ContactMngr.p_Count];
                    int l_Idx = 0;

                    foreach (Contact l_Contact in f_ContactMngr.Enumerable())
                    {
                        l_Arr[l_Idx++] = new GuiContactInfo(l_Contact.f_ClientID, l_Contact.f_Distance, l_Contact.GetLastActivityTime());
                    }

                    l_Input.f_ContactInfos = l_Arr;
                    Util.VolatileWrite(out l_Input.f_ContactInfosGeneration, l_Input.f_ContactInfosGeneration + 1);
                }

                if (0 != (MainForm.FormInput.NetworkDataTypes.GeneralPeers & l_Type))
                {
                    uint l_ConnectedUsers;
                    UserContext[] l_Arr = f_Inst.f_UserMngr.BuildArrayForGui(1000, out l_ConnectedUsers);
                    MainForm.FormInput.GeneralPeers l_Peers = new MainForm.FormInput.GeneralPeers(++f_GeneralPeersGeneration, l_ConnectedUsers, l_Arr);
                    Util.VolatileWrite(out l_Input.f_GeneralPeers, l_Peers);
                }

                if (0 != (MainForm.FormInput.NetworkDataTypes.Stats & l_Type))
                {
                    l_Input.f_Stats = f_Inst.f_NetworkMngr.BuildStats();
                    Util.VolatileWrite(out l_Input.f_StatsGeneration, l_Input.f_StatsGeneration + 1);
                }

                f_IgnoreGuiTime = l_NextGuiUpdateTime;
            }
        }

#region ContactHandling

        public Contact TrySetContact(Contact a_Contact,
                                     bool a_Trusted,
                                     out bool a_IsMe)
        {
            // Returns the contact instance (not necessarily
            // a_Contact) that should be used from now on,
            // or null if this contact should not be used
            // at all.

            // NOTE: We intentionally allow several contacts
            // to have the same stream address (we wouldn't
            // know how to find out the real owner).

            Contact l_Existing = f_ContactMngr.GetContact(a_Contact.f_ClientID);

            if (null == l_Existing)
            {
                a_IsMe = IsMyClientID(a_Contact.f_ClientID);

                if (false == a_IsMe &&
                    false == HasAddressConflictWithMe(a_Contact) &&
                    false == a_Contact.AreDestinationsBanned(f_Inst.f_NetworkMngr))
                {
                    Contact l_ConflictingContact = f_ContactMngr.GetContact(a_Contact.p_DatagramAddr);

                    if (null != l_ConflictingContact)
                    {
                        // This datagram address is occupied.

                        if (false == a_Trusted)
                        {
                            if (false == KadOffline() &&
                                false == l_ConflictingContact.p_IsVerified &&
                                true == f_RootZone.WouldLikeToAdd(a_Contact.f_ClientID))
                            {
                                f_PacketHndlr.PingContact(a_Contact);
                            }

                            return null;
                        }
                        else
                        {
                            if (true == l_ConflictingContact.IsPinned(f_Inst.GetTime()))
                            {
                                return null;
                            }

                            if (true == l_ConflictingContact.p_IsInRoutingZone)
                            {
                                bool l_Removed = f_RootZone.Remove(l_ConflictingContact);
                            }

                            f_ContactMngr.Remove(l_ConflictingContact);
                        }
                    }

                    a_Contact.SetVitality(ContactVitality.Alive);

                    bool l_EnteredZone = f_RootZone.Add(a_Contact);

                    f_ContactMngr.Add(a_Contact);

                    return a_Contact;
                }

                return null;
            }

            a_IsMe = false;

            if (false == l_Existing.HasExactSameAddresses(a_Contact))
            {
                if (false == a_Trusted)
                {
                    if (false == KadOffline())
                    {
                        f_PacketHndlr.PingContact(l_Existing);
                    }
                }
                else if (false == HasAddressConflictWithMe(a_Contact) &&
                         false == a_Contact.AreDestinationsBanned(f_Inst.f_NetworkMngr))
                {
                    f_ContactMngr.UpdateAddresses(l_Existing, a_Contact.p_DatagramAddr, a_Contact.p_StreamAddr);
                }
            }
            else
            {
                if (false == l_Existing.p_IsInRoutingZone)
                {
                    bool l_EnteredZone = f_RootZone.Add(l_Existing);
                }
            }

            return l_Existing;
        }

        public StorageStates SaveNodes()
        {
            // Only the NetworkManager class may call this routine
            // from outside.

            return f_Inst.WriteToElement(new ElmKey(Instance.c_NodesElmKey), SaveNodesCore);
        }

        private StorageStates SaveNodesCore(BufferIO a_Buffer)
        {
            BufferIO.CountWriter l_CountWtr;

            a_Buffer.WriteUInt32LE(c_NodesMagic);
            BufferIO.CountWriter.InitAndMoveBuffer(a_Buffer, sizeof(uint), out l_CountWtr);
            a_Buffer.SetPosition(c_NodesHeaderSize);

            ExportNodesToBuffer(a_Buffer, false);
            uint l_BytesWritten = (a_Buffer.p_Position - c_NodesHeaderSize);

            l_CountWtr.Finish(l_BytesWritten);

            return StorageStates.Success;
        }

        private StorageStates LoadNodes()
        {
            return f_Inst.ReadFromElement(new ElmKey(Instance.c_NodesElmKey), LoadNodesCore);
        }

        private StorageStates LoadNodesCore(BufferIO a_Buffer)
        {
            if (c_NodesHeaderSize > a_Buffer.GetRemaining() ||
                c_NodesMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            uint l_BodySize = a_Buffer.ReadUInt32LE();
            a_Buffer.SetPosition(c_NodesHeaderSize);

            if ((sizeof(uint) + (c_MaxContactsPerBuffer * Contact.c_MaximumSize)) < l_BodySize ||
                l_BodySize > a_Buffer.GetRemaining() ||
                false == LoadNodesFromBuffer(a_Buffer, true) ||
                false == a_Buffer.EndReached())
            {
                return StorageStates.Corruption;
            }

            return StorageStates.Success;
        }

        public bool LoadNodesFromBuffer(BufferIO a_Buffer,
                                        bool a_TrustedBuffer)
        {
            if (sizeof(uint) > a_Buffer.GetRemaining())
            {
                return false;
            }

            uint l_Contacts = a_Buffer.ReadUInt32LE();
            ulong l_ExpectedMinSize = ((ulong)l_Contacts * Contact.c_MinimumSize);

            if (a_Buffer.GetRemaining() < l_ExpectedMinSize)
            {
                return false;
            }

            l_Contacts = Math.Min(c_MaxContactsPerBuffer, l_Contacts);

            for (uint i = 0; i < l_Contacts; i++)
            {
                Contact l_Contact;

                switch (Contact.FromBuffer(this, a_Buffer, out l_Contact))
                {
                    case BufferIO.LoadResult.Success:
                        bool l_Junk;
                        TrySetContact(l_Contact, a_TrustedBuffer, out l_Junk);
                        break;

                    case BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood:
                        continue;

                    case BufferIO.LoadResult.Corruption:
                        return false;

                    default: throw (new ApplicationException());
                }
            }

            return true;
        }

        private void ExportNodesToBuffer(BufferIO a_Buffer,
                                         bool a_ForExternalUse)
        {
            bool l_RestrictToClassicAddresses = a_ForExternalUse;
            uint l_Contacts = 0;
            Contact[] l_Arr = new Contact[(int)Math.Min(c_MaxContactsPerBuffer, f_ContactMngr.p_Count + 1)];

            foreach (Contact l_Contact in f_ContactMngr.Enumerable())
            {
                if (false == l_Contact.p_IsInRoutingZone)
                {
                    continue;
                }

                if (true == l_RestrictToClassicAddresses &&
                    false == l_Contact.p_HasClassicAddresses)
                {
                    continue;
                }

                l_Arr[l_Contacts++] = l_Contact;

                if (c_MaxContactsPerBuffer <= l_Contacts)
                {
                    break;
                }
            }

            if (true == a_ForExternalUse)
            {
                bool l_IncludeSelf = (false == KadOffline() &&
                                      64 <= l_Contacts &&
                                      c_MaxContactsPerBuffer > l_Contacts);

                if (true == l_IncludeSelf &&
                    true == l_RestrictToClassicAddresses &&
                    false == f_Self.p_HasClassicAddresses)
                {
                    l_IncludeSelf = false;
                }

                if (true == l_IncludeSelf)
                {
                    int l_SelfIdx = Util.GetRandomIndexFast((int)l_Contacts);
                    Array.Copy(l_Arr, l_SelfIdx, l_Arr, l_SelfIdx + 1, l_Contacts - l_SelfIdx);
                    l_Contacts++;
                    l_Arr[l_SelfIdx] = f_Self;
                }

                // Get rid of any order (for security reasons).
                Util.RandomizeArrayOrder<Contact>(l_Arr, 0, (int)l_Contacts);
            }

            uint l_MaxNeededLength = (a_Buffer.p_Position + sizeof(uint) + (l_Contacts * (false == l_RestrictToClassicAddresses ? Contact.c_MaximumSize : Contact.c_MinimumSize)));

            if (a_Buffer.p_Length < l_MaxNeededLength)
            {
                a_Buffer.SetCapacity(l_MaxNeededLength);
            }

            BufferIO.CountWriter l_CountWtr;
            BufferIO.CountWriter.InitAndMoveBuffer(a_Buffer, sizeof(uint), out l_CountWtr);

            for (uint i = 0; i < l_Contacts; i++)
            {
                l_Arr[i].ToBufferWithSpecificVitality(a_Buffer, ContactVitality.Alive, l_RestrictToClassicAddresses);
            }

            l_CountWtr.Finish(l_Contacts);
        }

        public void GuiAddAddressForBootstrapping(GeneralAPCCookie a_Cookie)
        {
            I2PAddress l_DatagramAddr = (I2PAddress)((object[])a_Cookie.GetArg())[0];

            // We don't know a lot, so just make stuff up. It will
            // be corrected later.
            byte[] l_Bytes = Encoding.ASCII.GetBytes(l_DatagramAddr.ToFullString());
            byte[] l_ID = Support.HashFunctions.MD_4.ComputeHash(l_Bytes, 0, l_Bytes.Length);
            ClientID l_ClientID = new ClientID(UInt128.FromByteArrayME(l_ID, 0));
            Contact l_Contact = new Contact(l_ClientID, CalcDistance(l_ClientID));
            l_Contact.SetDatagramAddr(l_DatagramAddr);
            l_Contact.SetStreamAddr(I2PAddress.BuildRandom());

            bool l_Junk;
            TrySetContact(l_Contact, false, out l_Junk);

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiImportNodes(GeneralAPCCookie a_Cookie)
        {
            byte[] l_Data = a_Cookie.GetArgAsByteArray();
            MemoryIO l_Buffer = MemoryIO.CreateReadable(l_Data);
            bool l_Succ = LoadNodesFromBuffer(l_Buffer, false);

            a_Cookie.SetState(false == l_Succ ? CookieStates.Failure : CookieStates.Success);
        }

        public void GuiExportNodes(GeneralAPCCookie a_Cookie)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritableAndGrowable(c_MaxContactsPerBuffer * Contact.c_MaximumSize);

            ExportNodesToBuffer(l_Buffer, true);

            if (Contact.c_MinimumSize > l_Buffer.p_Position)
            {
                a_Cookie.SetState(CookieStates.Failure);
            }

            a_Cookie.SetReturnObjectAndState(l_Buffer.CopyRange(0, l_Buffer.p_Position), CookieStates.Success);
        }

#endregion

        public bool HasAddressConflictWithMe(Contact a_Contact)
        {
            // May be called even while there is no connection.

            if (true == KadOffline())
            {
                return false;
            }

            return f_Self.HasSameAddressesWithCrossCheck(a_Contact);
        }

        public bool IsMyClientID(ClientID a_ClientID)
        {
            return f_Self.HasClientID(a_ClientID);
        }

        public bool IsMyUserID(UserID a_UserID)
        {
            return f_Inst.f_NetworkMngr.p_OwnPersona.IsMyUserID(a_UserID);
        }

        public Distance CalcDistance(ClientID a_Target)
        {
            return CalcDistance(f_Self.f_ClientID, a_Target);
        }

        public static Distance CalcDistance(ClientID a_ID1,
                                            ClientID a_ID2)
        {
            return new Distance(a_ID1.f_Val.Xor(a_ID2.f_Val));
        }

        public ClientID CalcClientID(Distance a_Dist)
        {
            return new ClientID(f_Self.f_ClientID.f_Val.Xor(a_Dist.f_Val));
        }

        public bool VerifyEndReached(BufferIO a_Buffer)
        {
            return VerifyRemaining(a_Buffer, 0);
        }

        public bool VerifyRemaining(BufferIO a_Buffer,
                                    uint a_ByteCount)
        {
            if (a_Buffer.GetRemaining() != a_ByteCount)
            {
                f_Stats.NotifyCorruptDataSize();

                return false;
            }

            return true;
        }

        public bool VerifyRemainingMin(BufferIO a_Buffer,
                                       uint a_ByteCount)
        {
            if (a_Buffer.GetRemaining() < a_ByteCount)
            {
                f_Stats.NotifyCorruptDataSize();

                return false;
            }

            return true;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct KadPrefs
    {
        public const byte c_PublMinAcceptReqFitness = 4;
        public const byte c_PublMinOwnFitness = 5;
        public const byte c_PublMaxNodesToReach = 16;

        public const uint c_MinKadMapSize = (1024 * 1024);
        public const uint c_MaxKadMapSize = (75 * 1024 * 1024);

        public const char c_CommentsFilterSepChar = '|';
        public const string c_CommentsFilterSepStr = "|";
        public const ushort c_MaxCommentsFilterStrChars = 1024;

        private const PublishFlags c_DefaultPublFlags = PublishFlags.EnableAll;
        private const bool c_DefaultPublUnicodeTextsInNBStyle = false;

        public readonly byte f_PublMinAcceptReqFitness;
        public readonly byte f_PublMinOwnFitness;
        public readonly bool f_PublAcceptOnlyTrustedUsers;
        public readonly uint f_MaxKadMapSize;
        public readonly string f_CommentsFilter;
        public readonly bool f_RandomizeClientID;
        public readonly ClientID f_ClientID;
        // From v1.3.0 on.
        public readonly PublishFlags f_PublFlags;
        // From v1.4.0 on.
        public readonly bool f_PublUnicodeTextsInNBStyle;
        // From v1.4.25 on.
        public readonly KademliaFilter f_KadFilter;

        public Tag.UnicodeSupport p_PublUnicodeSupport
        {
            get { return (false == f_PublUnicodeTextsInNBStyle ? Tag.UnicodeSupport.ImplicitUTF8 : Tag.UnicodeSupport.NBStyle); }
        }

        public KadPrefs(uint a_PublMinAcceptReqFitness,
                        uint a_PublMinOwnFitness,
                        bool a_PublAcceptOnlyTrustedUsers,
                        uint a_MaxKadMapMemSize,
                        string a_CommentsFilter,
                        bool a_RandomizeClientID,
                        ClientID a_ClientID,
                        PublishFlags a_PublFlags,
                        bool a_PublUnicodeTextsInNBStyle,
                        KademliaFilter a_KadFilter)
        {
            f_PublMinAcceptReqFitness = (byte)a_PublMinAcceptReqFitness;
            f_PublMinOwnFitness = (byte)a_PublMinOwnFitness;
            f_PublAcceptOnlyTrustedUsers = a_PublAcceptOnlyTrustedUsers;
            f_MaxKadMapSize = a_MaxKadMapMemSize;
            f_CommentsFilter = a_CommentsFilter;
            f_RandomizeClientID = a_RandomizeClientID;
            f_ClientID = a_ClientID;
            f_PublFlags = a_PublFlags;
            f_PublUnicodeTextsInNBStyle = a_PublUnicodeTextsInNBStyle;
            f_KadFilter = a_KadFilter;
        }

        public static void CreateDefault(out KadPrefs a_Dest)
        {
            a_Dest = new KadPrefs(c_PublMinAcceptReqFitness,
                                  c_PublMinOwnFitness,
                                  false,
                                  30 * 1024 * 1024,
                                  // Default filter string.
                                  "http:|https:|ftp:|www.|.i2p",
                                  true,
                                  new ClientID(UInt128.BuildRandom()),
                                  c_DefaultPublFlags,
                                  c_DefaultPublUnicodeTextsInNBStyle,
                                  KademliaFilter.Create());
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      uint a_Revision,
                                      out KadPrefs a_Dest)
        {
            uint l_MinAcceptPublishReqFitness = a_Buffer.ReadByte();
            uint l_MinOwnPublishingFitness = a_Buffer.ReadByte();
            bool l_OnlyTrustedUsersMayPublish = a_Buffer.ReadBool();
            uint l_MaxKadMapSize = a_Buffer.ReadUInt32LE();
            string l_CommentsFilter = a_Buffer.ReadUTF16LEStringEncLen();
            bool l_RandomizeClientID = a_Buffer.ReadBool();
            // NOTE: If there is no client id, we'll randomize one
            // right here.
            UInt128 l_ClientID = (false == l_RandomizeClientID ? a_Buffer.ReadUInt128LE() : UInt128.BuildRandom());
            PublishFlags l_PublFlags = (1 > a_Revision ? c_DefaultPublFlags : (PublishFlags)a_Buffer.ReadByte());
            bool l_PublUnicodeTextsInNBStyle = (2 > a_Revision ? c_DefaultPublUnicodeTextsInNBStyle : a_Buffer.ReadBool());
            KademliaFilter l_KadFilter;

            if (3 > a_Revision)
            {
                l_KadFilter = KademliaFilter.Create();
            }
            else
            {
                StorageStates l_State = KademliaFilter.CreateFromConfig(a_Buffer, out l_KadFilter);

                if (StorageStates.Success != l_State)
                {
                    a_Dest = new KadPrefs();

                    return false;
                }
            }

            if (RoutingZone.c_MaxPossibleFitness < l_MinAcceptPublishReqFitness ||
                RoutingZone.c_MaxPossibleFitness < l_MinOwnPublishingFitness ||
                c_MinKadMapSize > l_MaxKadMapSize ||
                c_MaxKadMapSize < l_MaxKadMapSize ||
                c_MaxCommentsFilterStrChars < l_CommentsFilter.Length)
            {
                Util.DebugBreak();

                a_Dest = new KadPrefs();

                return false;
            }

            a_Dest = new KadPrefs(l_MinAcceptPublishReqFitness, l_MinOwnPublishingFitness, l_OnlyTrustedUsersMayPublish, l_MaxKadMapSize, l_CommentsFilter, l_RandomizeClientID, new ClientID(l_ClientID), l_PublFlags, l_PublUnicodeTextsInNBStyle, l_KadFilter);

            return true;
        }

        public uint ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteByte(f_PublMinAcceptReqFitness);
            a_Buffer.WriteByte(f_PublMinOwnFitness);
            a_Buffer.WriteBool(f_PublAcceptOnlyTrustedUsers);
            a_Buffer.WriteUInt32LE(f_MaxKadMapSize);
            a_Buffer.WriteUTF16LEStringEncLen(f_CommentsFilter);
            a_Buffer.WriteBool(f_RandomizeClientID);

            if (false == f_RandomizeClientID)
            {
                a_Buffer.WriteUInt128LE(f_ClientID.f_Val);
            }

            a_Buffer.WriteByte((byte)f_PublFlags);
            a_Buffer.WriteBool(f_PublUnicodeTextsInNBStyle);
            f_KadFilter.ToConfigBuffer(a_Buffer);

            // Increase with every change.
            return 3;
        }

        public void DetermineChangeKind(ref KadPrefs a_ChangeTo,
                                        ref PrefChangeKinds a_ChangeKind)
        {
            if (true == Preferences.NeedsReconnOrRestart(a_ChangeKind))
            {
                return;
            }

            if (PrefChangeKinds.ApplicableWithoutReconnect == a_ChangeKind ||
                this.f_PublMinAcceptReqFitness != a_ChangeTo.f_PublMinAcceptReqFitness ||
                this.f_PublMinOwnFitness != a_ChangeTo.f_PublMinOwnFitness ||
                this.f_PublAcceptOnlyTrustedUsers != a_ChangeTo.f_PublAcceptOnlyTrustedUsers ||
                this.f_MaxKadMapSize != a_ChangeTo.f_MaxKadMapSize ||
                false == this.IsEqualCommentFilterArray(a_ChangeTo.f_CommentsFilter) ||
                this.f_PublFlags != a_ChangeTo.f_PublFlags ||
                this.f_PublUnicodeTextsInNBStyle != a_ChangeTo.f_PublUnicodeTextsInNBStyle ||
                false == this.f_KadFilter.Equals(a_ChangeTo.f_KadFilter, false))
            {
                a_ChangeKind = PrefChangeKinds.ApplicableWithoutReconnect;
            }
        }

        public string[] BuildCommentFilterArray()
        {
            return BuildCommentFilterArray(f_CommentsFilter);
        }

        public static string[] BuildCommentFilterArray(string a_CommentsFilter)
        {
            List<string> l_Ls = new List<string>(a_CommentsFilter.Split(new char[] { c_CommentsFilterSepChar }, StringSplitOptions.RemoveEmptyEntries));
            l_Ls.Sort();

            for (int i = (l_Ls.Count - 2); 0 <= i; i--)
            {
                if (l_Ls[i] == l_Ls[i + 1])
                {
                    // Delete redundant entry.

                    l_Ls.RemoveAt(i + 1);
                }
            }

            return l_Ls.ToArray();
        }

        public static bool IsCommentFiltered(string a_CommentStr,
                                             string[] a_CommentsFilterArray)
        {
            foreach (string l_FilterStr in a_CommentsFilterArray)
            {
                if (0 <= Util.IndexOfIgnoreCase(a_CommentStr, l_FilterStr))
                {
                    return true;
                }
            }

            return false;
        }

        private bool IsEqualCommentFilterArray(string a_OtherCommentsFilter)
        {
            string[] l_A = BuildCommentFilterArray(this.f_CommentsFilter);
            string[] l_B = BuildCommentFilterArray(a_OtherCommentsFilter);

            return (l_A.Length == l_B.Length &&
                    string.Join(c_CommentsFilterSepStr, l_A) == string.Join(c_CommentsFilterSepStr, l_B));
        }
    }

    class RoutingZone
    {
        public const byte c_MaxPossibleFitness = UInt128.c_BitSize;
        private const byte c_MaxPossibleLevel = (UInt128.c_BitSize - 1);

        public const byte c_K = RoutingBin.c_MaxEntries;
        public const byte c_HalfK = (c_K / 2);
        private const byte c_NearLevels = 5;

        public readonly Kademlia f_Kad;
        private readonly RoutingZone f_ParentZone;
        public readonly byte f_Level;
        private readonly Distance f_ZoneIndex;
        private readonly bool f_IsNear;

        // For leafes.
        private RoutingBin f_Bin;

        // For non-leafes, in that case both pointers are valid.
        private RoutingZone f_SubZone0;
        private RoutingZone f_SubZone1;

        // Controlled by the Kademlia class.
        public Time f_SmallTime;
        public Time f_BigTime;

        // We require f_Disposed for the TryMergeParent().
        private bool f_Disposed;

        public bool p_IsLeaf
        {
            get { return (null == f_SubZone0); }
        }

        public RoutingZone(Kademlia a_Kad) :
                      this(a_Kad, null, 0, new Distance(UInt128.s_Zero))
        {
        }

        private RoutingZone(Kademlia a_Kad,
                            RoutingZone a_ParentZone,
                            byte a_Level,
                            Distance a_ZoneIndex)
        {
            f_Kad = a_Kad;
            f_ParentZone = a_ParentZone;
            f_Level = a_Level;
            f_ZoneIndex = a_ZoneIndex;
            f_IsNear = (c_NearLevels > f_Level ||
                        true == a_ZoneIndex.f_Val.AreHighBitsClear(c_NearLevels));

            f_Bin = new RoutingBin(this);
            f_Kad.AddEvent(this);
        }

        public override int GetHashCode()
        {
            return f_ZoneIndex.f_Val.GetHashCode();
        }

        public bool IsCompatible(Contact a_Contact)
        {
            return UInt128.AreHighBitsEqual(f_ZoneIndex.f_Val, a_Contact.f_Distance.f_Val, f_Level);
        }

        private RoutingZone GetSubZone(Distance a_Distance)
        {
            if (false == a_Distance.f_Val.GetHighBit(f_Level))
            {
                return f_SubZone0;
            }

            return f_SubZone1;
        }

        public void RetrieveInfo(out RoutingInfo a_Info)
        {
            a_Info = new RoutingInfo();

            RetrieveInfoCore(ref a_Info);

            // We will take a look at the near side of the
            // tree and estimated the number of total nodes
            // in the network from that. Note that the near
            // side (f_IsNear) has no level limit.

            RoutingZone l_Zone = this;
            while (true)
            {
                if (false == l_Zone.p_IsLeaf)
                {
                    l_Zone = l_Zone.f_SubZone0;

                    continue;
                }

                a_Info.f_EstimatedTotalNodes = EstimateCount(l_Zone.f_Level, l_Zone.f_Bin.p_Count);
                a_Info.f_EstimatedTotalVerifiedNodes = EstimateCount(l_Zone.f_Level, l_Zone.f_Bin.CountVerifiedNodes());

                break;
            }

            a_Info.NoDullTheorization();

            if (0 == a_Info.f_EstimatedTotalVerifiedNodes)
            {
                a_Info.f_MaxEstimatedTotalNodes = 0;
            }
            else
            {
                uint l_Levels = (Util.c_BitsPerUInt - Util.GetClearHighBits((uint)a_Info.f_EstimatedTotalVerifiedNodes / RoutingBin.c_MaxEntries));
                a_Info.f_MaxEstimatedTotalNodes = (28 >= l_Levels ? (RoutingBin.c_MaxEntries * (1u << (int)l_Levels)) : uint.MaxValue);
            }
        }

        private void RetrieveInfoCore(ref RoutingInfo a_Info)
        {
            if (false == p_IsLeaf)
            {
                f_SubZone0.RetrieveInfoCore(ref a_Info);
                f_SubZone1.RetrieveInfoCore(ref a_Info);
            }

            a_Info.f_MaxLevel = Math.Max(a_Info.f_MaxLevel, f_Level);
            a_Info.f_Bins++;
            a_Info.f_Nodes += f_Bin.p_Count;
            a_Info.f_VerifiedNodes += f_Bin.CountVerifiedNodes();
        }

        private static uint EstimateCount(uint a_Level,
                                          uint a_N)
        {
            // Calculate (2^(Level - 1) * N).

            if (0 == a_Level ||
                0 == a_N)
            {
                // Only one bin or no nodes at all.

                return a_N;
            }

            uint l_Multiplier = (1u << (int)a_Level);
            uint l_Count = (l_Multiplier * a_N);

            if (0 == l_Multiplier ||
                (l_Count / l_Multiplier) != a_N)
            {
                // Overflow.

                return uint.MaxValue;
            }

            return l_Count;
        }

        public bool Add(Contact a_Contact)
        {
            // It is not necessary to call WouldLikeToAdd()
            // before calling this routine.

            return AddCore(a_Contact);
        }

        private bool AddCore(Contact a_Contact)
        {
            // Returns true if the contact was added.

            if (false == p_IsLeaf)
            {
                return GetSubZone(a_Contact.f_Distance).AddCore(a_Contact);
            }

            // Try to add the contact, if it doesn't work we'll
            // have to split.

            bool l_OK = f_Bin.Add(a_Contact, false);

            if (false == l_OK &&
                true == CanSplit())
            {
                Split();

                l_OK = GetSubZone(a_Contact.f_Distance).AddCore(a_Contact);
            }

            return l_OK;
        }

        public bool Remove(Contact a_Contact)
        {
            // For external use only.

            return RemoveCore(a_Contact);
        }

        private bool RemoveCore(Contact a_Contact)
        {
            if (false == p_IsLeaf)
            {
                return GetSubZone(a_Contact.f_Distance).RemoveCore(a_Contact);
            }

            return f_Bin.Remove(a_Contact, false);
        }

        public bool WouldLikeToAdd(ClientID a_ClientID)
        {
            return WouldLikeToAddCore(a_ClientID, f_Kad.CalcDistance(a_ClientID));
        }

        private bool WouldLikeToAddCore(ClientID a_ClientID,
                                        Distance a_Distance)
        {
            if (false == p_IsLeaf)
            {
                return GetSubZone(a_Distance).WouldLikeToAddCore(a_ClientID, a_Distance);
            }

            if (null != f_Bin.GetContact(a_ClientID))
            {
                return false;
            }

            return (c_K > f_Bin.p_Count ||
                    true == CanSplit());
        }

        public void Reinitialize()
        {
            // Removes any contact that might have a
            // conflict with us.

            if (false == p_IsLeaf)
            {
                // Go through the entire set.

                f_SubZone0.Reinitialize();
                f_SubZone1.Reinitialize();

                return;
            }

            Contact l_Self = f_Kad.f_Self;
            Contact[] l_Arr = f_Bin.f_Entries.p_Arr;
            for (int i = (f_Bin.f_Entries.p_Count - 1); 0 <= i; i--)
            {
                if (true == l_Self.HasAnyConflict(l_Arr[i]))
                {
                    l_Arr[i].TurnRoutingZoneStateOff();
                    f_Bin.f_Entries.DeleteAt(i);
                }
            }
        }

        private void GetAllEntries(List<Contact> a_Result)
        {
            if (false == p_IsLeaf)
            {
                f_SubZone0.GetAllEntries(a_Result);
                f_SubZone1.GetAllEntries(a_Result);
            }
            else
            {
                f_Bin.GetAllEntries(a_Result);
            }
        }

        private void GetEntries(List<Contact> a_Result,
                                uint a_MaxRequired)
        {
            // Returns the contacts from the buckets with the
            // lowest a_Distance first.

            if (a_Result.Count >= a_MaxRequired)
            {
                return;
            }

            if (false == p_IsLeaf)
            {
                f_SubZone0.GetEntries(a_Result, a_MaxRequired);
                f_SubZone1.GetEntries(a_Result, a_MaxRequired);
            }
            else
            {
                f_Bin.GetEntries(a_Result, a_MaxRequired);
            }
        }

        public List<Contact> GetBootstrapContactsForMyself(uint a_MaxRequired)
        {
            List<Contact> l_Results = new List<Contact>((int)a_MaxRequired);
            GetEntries(l_Results, a_MaxRequired);

            return l_Results;
        }

        public void GetClosestTo(ref GetClosestToArgs a_Args)
        {
            // Finds the n contacts that have the most similar
            // client id compared to a_Args.f_Target.

            GetClosestToRecursive(ref a_Args);
        }

        private void GetClosestToRecursive(ref GetClosestToArgs a_Args)
        {
            if (a_Args.f_Results.Count >= a_Args.f_MaxRequired)
            {
                return;
            }

            if (true == p_IsLeaf)
            {
                f_Bin.GetClosestTo(ref a_Args);

                return;
            }

            RoutingZone l_NearSubZone = GetSubZone(a_Args.f_Distance);

            l_NearSubZone.GetClosestToRecursive(ref a_Args);

            if (a_Args.f_Results.Count < a_Args.f_MaxRequired)
            {
                // We need even more entries, so involve the
                // far side.

                RoutingZone l_FarSubZone = (l_NearSubZone == f_SubZone0 ? f_SubZone1 : f_SubZone0);

                l_FarSubZone.GetClosestToRecursive(ref a_Args);
            }
        }

        private bool CanSplit()
        {
            if (false == p_IsLeaf ||
                c_MaxPossibleLevel <= f_Level)
            {
                return false;
            }

            return (true == f_IsNear &&
                    c_K == f_Bin.p_Count);
        }

        private void Split()
        {
            // The caller must have verified that splitting is
            // possible using CanSplit().

            f_SubZone0 = GenerateSubZone(false);
            f_SubZone1 = GenerateSubZone(true);

            Contact[] l_Arr = f_Bin.f_Entries.p_Arr;
            for (int i = (f_Bin.f_Entries.p_Count - 1); 0 <= i; i--)
            {
                bool OK = GetSubZone(l_Arr[i].f_Distance).f_Bin.Add(l_Arr[i], true);
            }

            f_Bin.f_Entries.Clear();
            f_Kad.RemoveEvent(this);
        }

        public static void TryMergeParent(RoutingZone a_Zone)
        {
            // It's crucial to check whether the zone was disposed
            // or not.

            if (false == a_Zone.f_Disposed &&
                null != a_Zone.f_ParentZone &&
                true == a_Zone.f_ParentZone.CanMerge())
            {
                a_Zone.f_ParentZone.Merge();
            }
        }

        private bool CanMerge()
        {
            if (false == p_IsLeaf)
            {
                return (true == f_SubZone0.p_IsLeaf &&
                        true == f_SubZone1.p_IsLeaf &&
                        c_HalfK > (f_SubZone0.f_Bin.p_Count + f_SubZone1.f_Bin.p_Count));
            }
            else if (null != f_ParentZone)
            {
                return f_ParentZone.CanMerge();
            }

            return false;
        }

        private RoutingZone Merge()
        {
            // The caller must have verified that merging is
            // possible using CanMerge(). Returns the deepest
            // zone that could not be merged away.

            if (true == p_IsLeaf)
            {
                return f_ParentZone.Merge();
            }

            CloseSubZone(ref f_SubZone0);
            CloseSubZone(ref f_SubZone1);
            f_Kad.AddEvent(this);

            if (null != f_ParentZone &&
                true == f_ParentZone.CanMerge())
            {
                return f_ParentZone.Merge();
            }

            return this;
        }

        private void CloseSubZone(ref RoutingZone a_SubZone)
        {
            bool l_OK = a_SubZone.f_Bin.CopyTo(ref this.f_Bin);

            f_Kad.RemoveEvent(a_SubZone);

            a_SubZone.f_Disposed = true;
            a_SubZone = null;
        }

        private RoutingZone GenerateSubZone(bool a_Far)
        {
            Distance l_NewZoneIndex = new Distance(f_ZoneIndex.f_Val.SetHighBit(f_Level, a_Far));

            return new RoutingZone(f_Kad, this, (byte)(f_Level + 1), l_NewZoneIndex);
        }

        public bool OnBigTimer()
        {
            // This routine may not call (directly or indirectly)
            // LeafStatusChanged().

            if (true == f_IsNear ||
                c_HalfK >= f_Bin.p_Count)
            {
                RandomLookup();

                return true;
            }

            return false;
        }

        public void OnSmallTimer(Time a_Time,
                                 bool a_SpareDeadContacts,
                                 out bool a_TryMergeParent)
        {
            // This routine may not call (directly or indirectly)
            // LeafStatusChanged().

            a_TryMergeParent = false;

            Time l_HalfHourAgo = a_Time.SubSeconds(Time.c_SecsPerHalfHour);

            Contact[] l_Arr = f_Bin.f_Entries.p_Arr;
            for (int i = (f_Bin.f_Entries.p_Count - 1); 0 <= i; i--)
            {
                if (false == l_Arr[i].WaitingForPong(a_Time))
                {
                    if (false == a_SpareDeadContacts &&
                        ContactVitality.Dead == l_Arr[i].GetVitality())
                    {
                        a_TryMergeParent = true;

                        // We don't need to check if a contact is pinned
                        // here, since we just remove it from the zone
                        // (the contact manager still holds it).
                        l_Arr[i].TurnRoutingZoneStateOff();
                        f_Bin.f_Entries.DeleteAt(i);
                    }
                    else if (false == l_Arr[i].HadActivitySince(l_HalfHourAgo))
                    {
                        f_Kad.f_PacketHndlr.PingContact(a_Time, l_Arr[i]);
                    }
                }
            }
        }

        private void RandomLookup()
        {
            // Look up a random client id fitting into this zone.

            UInt128 l_Random = UInt128.BuildRandom();
            uint l_RandBits = ((uint)c_MaxPossibleLevel - f_Level);

            Distance l_Dist = new Distance(UInt128.CopyPartial(l_Random, f_ZoneIndex.f_Val, l_RandBits));

            f_Kad.f_SearchMngr.FindNode(f_Kad.CalcClientID(l_Dist), false);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct RoutingInfo
        {
            public uint f_MaxLevel;
            public uint f_Bins;
            public uint f_Nodes;
            public uint f_VerifiedNodes;
            public uint f_EstimatedTotalNodes;
            public uint f_EstimatedTotalVerifiedNodes;
            public uint f_MaxEstimatedTotalNodes;

            public void NoDullTheorization()
            {
                // Change the theoretical values to the empirical ones
                // if the theory doesn't fit to what we know already.

                f_EstimatedTotalNodes = Math.Min(f_EstimatedTotalNodes, f_Nodes);
                f_EstimatedTotalVerifiedNodes = Math.Min(f_EstimatedTotalVerifiedNodes, f_VerifiedNodes);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct RoutingBin
        {
            // This number is called 'K' in Kademlia discussion.
            public const byte c_MaxEntries = 10;

            public FlexArray<Contact> f_Entries;

            public uint p_Count
            {
                get { return (uint)f_Entries.p_Count; }
            }

            public RoutingBin(RoutingZone a_Zone)
            {
                FlexArray<Contact>.Create(c_MaxEntries, out f_Entries);
            }

            public uint CountVerifiedNodes()
            {
                uint l_Verified = 0;

                Contact[] l_Arr = f_Entries.p_Arr;
                for (int i = (f_Entries.p_Count - 1); 0 <= i; i--)
                {
                    if (true == l_Arr[i].p_IsVerified)
                    {
                        l_Verified += 1;
                    }
                }

                return l_Verified;
            }

            private int FindContact(ClientID a_ClientID,
                                    out Contact a_Contact)
            {
                // Returns subzero if there is no such contact.

                Contact[] l_Arr = f_Entries.p_Arr;
                for (int i = (f_Entries.p_Count - 1); 0 <= i; i--)
                {
                    if (true == l_Arr[i].HasClientID(a_ClientID))
                    {
                        a_Contact = l_Arr[i];

                        return i;
                    }
                }

                a_Contact = null;

                return (-1);
            }

            public bool Add(Contact a_Contact,
                            bool a_JustRearranging)
            {
                // Returns true if the contact was added.

                Contact l_ExistingContact;
                int l_Idx = FindContact(a_Contact.f_ClientID, out l_ExistingContact);

                if (0 <= l_Idx)
                {
                    return false;
                }

                if (c_MaxEntries <= f_Entries.p_Count)
                {
                    return false;
                }

                l_Idx = f_Entries.ReserveAtEnd();
                f_Entries.p_Arr[l_Idx] = a_Contact;

                if (false == a_JustRearranging)
                {
                    a_Contact.TurnRoutingZoneStateOn();
                }

                return true;
            }

            public bool Remove(Contact a_Contact,
                               bool a_JustRearranging)
            {
                Contact l_ExistingContact;
                int l_Idx = FindContact(a_Contact.f_ClientID, out l_ExistingContact);

                if (0 > l_Idx)
                {
                    return false;
                }

                if (false == a_JustRearranging)
                {
                    a_Contact.TurnRoutingZoneStateOff();
                }

                f_Entries.DeleteAt(l_Idx);

                return true;
            }

            public Contact GetContact(ClientID a_ClientID)
            {
                // Return can be null.

                Contact l_Contact;
                int l_Idx = FindContact(a_ClientID, out l_Contact);

                return (0 > l_Idx ? null : l_Contact);
            }

            public bool CopyTo(ref RoutingBin a_Dest)
            {
                // Copies everything form this to a_Dest.

                int l_DestIdx = a_Dest.f_Entries.p_Count;
                int l_SrcCount = this.f_Entries.p_Count;

                if (c_MaxEntries < (l_DestIdx + l_SrcCount))
                {
                    return false;
                }

                a_Dest.f_Entries.ReserveRange(l_DestIdx, l_SrcCount);

                Contact[] l_DestArr = a_Dest.f_Entries.p_Arr;
                Contact[] l_SrcArr = this.f_Entries.p_Arr;
                for (int i = 0; i < l_SrcCount; i++)
                {
                    l_DestArr[l_DestIdx + i] = l_SrcArr[i];
                }

                return true;
            }

            public void GetAllEntries(List<Contact> a_Result)
            {
                GetEntries(a_Result, uint.MaxValue);
            }

            public void GetEntries(List<Contact> a_Result,
                                   uint a_MaxRequired)
            {
                uint l_Count = Math.Min((uint)f_Entries.p_Count, a_MaxRequired - (uint)a_Result.Count);

                Contact[] l_Arr = f_Entries.p_Arr;
                for (uint i = 0; i < l_Count; i++)
                {
                    a_Result.Add(l_Arr[i]);
                }
            }

            public void GetClosestTo(ref RoutingZone.GetClosestToArgs a_Args)
            {
                Contact[] l_Arr = f_Entries.p_Arr;
                for (int i = (f_Entries.p_Count - 1); 0 <= i; i--)
                {
                    if (true == a_Args.f_RestrictToClassicAddresses &&
                        false == l_Arr[i].p_HasClassicAddresses)
                    {
                        continue;
                    }

                    if (l_Arr[i].GetVitality() <= a_Args.f_MaxType)
                    {
                        if (false == a_Args.f_VerifiedOnly ||
                            true == l_Arr[i].p_IsVerified)
                        {
                            // NOTE: This doesn't necessarily reflect our own
                            // distance to l_Target, since we might do a job
                            // for some else.
                            Distance l_TargDistance = Kademlia.CalcDistance(l_Arr[i].f_ClientID, a_Args.f_Target);

                            if (0 == a_Args.f_MinFitness ||
                                true == l_TargDistance.HasMinFitness(a_Args.f_MinFitness))
                            {
                                if (false == a_Args.f_Results.ContainsKey(l_TargDistance))
                                {
                                    a_Args.f_Results.Add(l_TargDistance, l_Arr[i]);

                                    if (a_Args.f_Results.Count == a_Args.f_MaxRequired)
                                    {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct GetClosestToArgs
        {
            public readonly ClientID f_Target;
            public readonly Distance f_Distance;
            public readonly uint f_MaxRequired;
            public readonly ContactVitality f_MaxType;
            public readonly uint f_MinFitness;
            public readonly bool f_RestrictToClassicAddresses;
            public readonly SortedList<Distance, Contact> f_Results;

            public bool f_VerifiedOnly;

            public GetClosestToArgs(Kademlia a_Kad,
                                    ClientID a_Target,
                                    uint a_MaxRequired,
                                    ContactVitality a_MaxType,
                                    uint a_MinFitness,
                                    bool a_RestrictToClassicAddresses,
                                    SortedList<Distance, Contact> a_Results)
            {
                f_Target = a_Target;
                f_Distance = a_Kad.CalcDistance(a_Target);
                f_MaxRequired = a_MaxRequired;
                f_MaxType = a_MaxType;
                f_MinFitness = a_MinFitness;
                f_RestrictToClassicAddresses = a_RestrictToClassicAddresses;
                f_Results = a_Results;

                f_VerifiedOnly = false;
            }
        }
    }

    [Flags]
    enum PublishFlags : byte
    {
        None = 0,
        EnableFiles = 0x1,
        EnableKeywords = 0x2,
        EnableNotes = 0x4,

        EnableAll = (EnableFiles | EnableKeywords | EnableNotes),
    }

    class PublishCookie : GeneralAPCCookie
    {
        public readonly byte f_SlotIndex;
        private readonly uint f_PubNum;
        public readonly bool f_FilesOnly;

        public readonly List<KnownFile> f_Files;
        public readonly Dictionary<UInt128, List<KnownFile>> f_Keywords;
        public readonly Support.HashFunctions.MD_4 f_MD4;

        public PublishCookie(uint a_SlotIndex,
                             bool a_FilesOnly,
                             ContentManager a_ContentMngr) :
            base(a_ContentMngr.NetworkGetFilesForeD2kPublishing)
        {
            f_SlotIndex = (byte)a_SlotIndex;
            f_PubNum = SlotIndexToPubNum(f_SlotIndex);
            f_FilesOnly = a_FilesOnly;
            f_Files = new List<KnownFile>();
            f_Keywords = new Dictionary<UInt128, List<KnownFile>>();
            f_MD4 = new Support.HashFunctions.MD_4();
        }

        public bool IsRelevant(uint a_HighUInt)
        {
            return ((a_HighUInt & PublishManager.c_Mask) == f_PubNum);
        }

        public bool IsEmpty()
        {
            return (0 == f_Files.Count &&
                    0 == f_Keywords.Count);
        }

        private static uint SlotIndexToPubNum(uint a_SlotIndex)
        {
            return (uint)(a_SlotIndex << (32 - PublishManager.c_Exponent));
        }
    }

    class PublishManager
    {
        public const byte c_Exponent = 6;
        public const ushort c_Slots = (1 << c_Exponent);
        public const uint c_Mask = ~(uint.MaxValue >> c_Exponent);

        private const uint c_WaitSecsAfterBuildingNewConn = (2 * Time.c_SecsPerMinute);
        // Also update the index lifetime in KadMap when changing this.
        private const uint c_PauseTimeInSecs = (5 * Time.c_SecsPerHour);

        private readonly Instance f_Inst;
        private readonly Kademlia f_Kad;

        private PublishFlags f_Flags;
        private bool f_MayPublish;
        private byte f_InitialSlotIndex;
        private Time f_NextRoundStartingTime;
        private bool f_RoundStarted;

        private bool f_PhaseOne;
        private byte f_SlotIndex;
        private PublishCookie f_Cookie;
        private UInt128 f_Key;
        private Datagram[] f_Datagrams;
        private ClientSearch f_Search;

        public PublishManager(Kademlia a_Kad)
        {
            f_Inst = a_Kad.f_Inst;
            f_Kad = a_Kad;
        }

        public void StopPublishing()
        {
            UpdateFlags(false, PublishFlags.None);
        }

        public void UpdateFlags(bool a_ConnJustCreated,
                                PublishFlags a_Flags)
        {
            f_Flags = a_Flags;
            bool l_Enable = (0 != (f_Flags & PublishFlags.EnableAll));

            if (l_Enable == f_MayPublish)
            {
                return;
            }

            f_MayPublish = l_Enable;

            if (false == f_MayPublish)
            {
                // Stop publishing.

                if (true == f_RoundStarted)
                {
                    f_RoundStarted = false;
                    ClearSlotParams();
                    NotifyRoundEnding();
                }
            }
            else
            {
                // Start publishing.

                f_InitialSlotIndex = (byte)(Util.GetRandomByte() % c_Slots);
                SetRoundStartingTime(false == a_ConnJustCreated ? 0 : c_WaitSecsAfterBuildingNewConn);
            }
        }

        private void NotifyRoundEnding()
        {
            ContentManager l_ContentMngr = f_Inst.f_ContentMngr;
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(l_ContentMngr.NetworkReleasePublishingResources);
            l_ContentMngr.EnqueueAPC(l_Cookie);
        }

        private void SetRoundStartingTime(uint a_SecsFromNow)
        {
            if (0 == a_SecsFromNow)
            {
                f_NextRoundStartingTime = new Time();
            }
            else
            {
                f_NextRoundStartingTime = f_Inst.GetTime().AddSeconds(a_SecsFromNow);
            }
        }

        public void TriggerPublishing()
        {
            // This routine can be called at any time. It's effect
            // depends on StartPublishing() and StopPublishing().

            if (true == f_MayPublish &&
                false == f_RoundStarted)
            {
                f_NextRoundStartingTime = new Time();
            }
        }

        public void RetrieveStats(Time a_Time,
                                  ref NetworkManager.StatsBuilder a_Stats)
        {
            if (false == f_MayPublish)
            {
                a_Stats.AddText(TextKeys.StatsTabPubStatus, TextKeys.StatsTabPubPassive);

                return;
            }

            if (false == f_RoundStarted)
            {
                a_Stats.AddText(TextKeys.StatsTabPubStatus, TextKeys.StatsTabPubPassive);

                if (false == f_NextRoundStartingTime.IsEarlierThan(a_Time))
                {
                    long l_SecsUntil = f_NextRoundStartingTime.SecondsSince(a_Time);
                    a_Stats.AddTime(TextKeys.StatsTabPubNextPublishing, l_SecsUntil);
                }
                else
                {
                    a_Stats.AddText(TextKeys.StatsTabPubNextPublishing, TextKeys.StatsTabPubWhenPossible);
                }

                return;
            }

            uint l_SlotsDone;

            if (f_InitialSlotIndex <= f_SlotIndex)
            {
                l_SlotsDone = (uint)(f_SlotIndex - f_InitialSlotIndex);
            }
            else
            {
                l_SlotsDone = (uint)((c_Slots - f_InitialSlotIndex) + f_SlotIndex);
            }

            float l_Progress = (((float)l_SlotsDone / c_Slots) / 2f);

            if (false == f_PhaseOne)
            {
                l_Progress += 0.5f;
            }

            string l_StatStr = f_Inst.f_TextSet.Format(f_Inst[TextKeys.StatsTabPubFormatActiveNDone], l_Progress);
            a_Stats.AddText(TextKeys.StatsTabPubStatus, l_StatStr);
        }

        public void Process(Time a_Time)
        {
            if (false == f_MayPublish)
            {
                return;
            }

            if (false == f_RoundStarted)
            {
                if (false == f_NextRoundStartingTime.IsEarlierThan(a_Time))
                {
                    return;
                }

                f_RoundStarted = true;
                f_PhaseOne = true;
                f_SlotIndex = f_InitialSlotIndex;
            }

            if (null == f_Cookie)
            {
                ContentManager l_ContentMngr = f_Inst.f_ContentMngr;
                f_Cookie = new PublishCookie(f_SlotIndex, f_PhaseOne, l_ContentMngr);
                l_ContentMngr.EnqueueAPC(f_Cookie);
            }

            CookieStates l_ContentMngrState = f_Cookie.p_StateVolatile;

            if (CookieStates.Success == l_ContentMngrState)
            {
                if (null == f_Search &&
                    false == f_Cookie.IsEmpty())
                {
                    f_Key = UInt128.CreateForPublishing(f_Cookie.f_SlotIndex, c_Exponent);
                    f_Datagrams = AddCookieToKadMapsAndBuildDatagrams(f_Cookie);

                    if (null != f_Datagrams)
                    {
                        f_Search = f_Kad.f_SearchMngr.FindNode(new ClientID(f_Key), true);
                    }
                }

                if (null != f_Search &&
                    false == f_Search.IsRunning())
                {
                    // The search has ended, now broadcast the datagrams
                    // nodes with the lowest distance.

                    I2PAddress[] l_Addresses = f_Search.GetBestForPublishing(KadPrefs.c_PublMaxNodesToReach);

                    if (null != l_Addresses)
                    {
                        for (int i = (f_Datagrams.Length - 1); 0 <= i; i--)
                        {
                            f_Kad.f_Stats.f_PublishRequestsSent += f_Kad.f_PacketHndlr.BroadcastPacket(f_Datagrams[i].f_Opcode, l_Addresses, ref f_Datagrams[i].f_Data);
                        }
                    }

                    f_Search = null;
                }

                if (null == f_Search)
                {
                    PrepareNextSlot();
                }
            }
            else if (CookieStates.Pending != l_ContentMngrState)
            {
                if (CookieStates.Failure == l_ContentMngrState)
                {
                    PrepareNextSlot();
                }
                else
                {
                    throw (new ApplicationException());
                }
            }
        }

        private void PrepareNextSlot()
        {
            f_SlotIndex++;

            if (c_Slots == f_SlotIndex)
            {
                f_SlotIndex = 0;
            }

            if (f_InitialSlotIndex == f_SlotIndex)
            {
                if (true == f_PhaseOne)
                {
                    // Beginning phase 2.

                    f_PhaseOne = false;
                }
                else
                {
                    // We've finished this round.

                    SetRoundStartingTime(c_PauseTimeInSecs);
                    f_RoundStarted = false;
                    NotifyRoundEnding();
                }
            }

            ClearSlotParams();
        }

        private void ClearSlotParams()
        {
            f_Cookie = null;
            f_Key = null;
            f_Datagrams = null;
            f_Search = null;
        }

        public void NotifyPublishResponse()
        {
            f_Kad.f_Stats.f_PublishResponsesRcvd++;
        }

        private Datagram[] AddCookieToKadMapsAndBuildDatagrams(PublishCookie a_Cookie)
        {
            // Return value is null if there are no datagrams.

            int l_InitialCapacity = (a_Cookie.f_Files.Count + a_Cookie.f_Keywords.Count);
            DatagramHelper l_DatagramHlpr = new DatagramHelper(PacketHandler.c_MaxKadMessageSize, l_InitialCapacity);
            Tag.WriteTagContext l_WriteTagCntxt = new Tag.WriteTagContext(l_DatagramHlpr.f_Buffer, f_Kad.f_KadPrefs.p_PublUnicodeSupport);

            bool l_IncludeFiles = (0 != (f_Flags & PublishFlags.EnableFiles));
            bool l_IncludeNotes = (false == f_PhaseOne &&
                                   0 != (f_Flags & PublishFlags.EnableNotes));

            if (true == l_IncludeFiles &&
                true == NetworkManager.c_RestrictToClassicAddresses &&
                false == true == f_Kad.f_Self.p_StreamAddr.p_IsClassicAddress)
            {
                l_IncludeFiles = false;
            }

            if (true == l_IncludeFiles ||
                true == l_IncludeNotes)
            {
                foreach (KnownFile l_Known in a_Cookie.f_Files)
                {
                    if (true == l_IncludeFiles)
                    {
                        // Handle source publishing (file id and own user id).

                        FileEntry l_FileEntry = FileEntry.Create(f_Kad, l_Known);

                        if (null != l_FileEntry)
                        {
                            string l_FileName = KademliaFilter.GetSuitableFileNameFromKnownFileVolatile(l_Known);

                            if (false == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(l_FileName, KademliaFilterScopes.ScopeLocalKademliaIndex))
                            {
                                int l_OwnLoad;
                                bool l_Added = f_Kad.f_Map.AddFileSource(l_FileEntry, false, out l_OwnLoad);
                            }

                            if (false == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(l_FileName, KademliaFilterScopes.ScopeOwnPublishingToOthers))
                            {
                                l_DatagramHlpr.Start(KadOpcodes.KADEMLIA_PUBLISH_REQ, l_FileEntry.f_KeyID.f_Val);
                                l_DatagramHlpr.f_Buffer.WriteUInt128ME(l_FileEntry.f_RefID.f_Val);
                                TagList.WriteFileEntryForPublishing(l_DatagramHlpr.f_Buffer, l_FileEntry);
                                l_DatagramHlpr.NotifyAdd();
                                l_DatagramHlpr.End();
                            }
                        }
                    }

                    if (true == l_IncludeNotes)
                    {
                        // Handle note publishing (file id and own client id).

                        NoteEntry l_NoteEntry = NoteEntry.Create(f_Kad, l_Known);

                        if (null != l_NoteEntry)
                        {
                            int l_OwnLoad;
                            bool l_Added = f_Kad.f_Map.AddNote(l_NoteEntry, out l_OwnLoad);

                            if (false == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(l_NoteEntry.f_FileName, KademliaFilterScopes.ScopeOwnPublishingToOthers))
                            {
                                l_DatagramHlpr.Start(KadOpcodes.KADEMLIA_PUB_NOTES_REQ, l_NoteEntry.f_KeyID.f_Val);
                                l_DatagramHlpr.f_Buffer.WriteUInt128ME(l_NoteEntry.f_RefID.f_Val);
                                TagList.WriteNoteEntryForPublishing(ref l_WriteTagCntxt, l_NoteEntry);
                                l_DatagramHlpr.NotifyAdd();
                                l_DatagramHlpr.End();
                            }
                        }
                    }
                }
            }

            bool l_IncludeKeywords = (false == f_PhaseOne &&
                                      0 != (f_Flags & PublishFlags.EnableKeywords));

            if (true == l_IncludeKeywords)
            {
                foreach (KeyValuePair<UInt128, List<KnownFile>> l_Pair in a_Cookie.f_Keywords)
                {
                    // Handle keyword publishing (keyword id and one or several file ids).

                    l_DatagramHlpr.Start(KadOpcodes.KADEMLIA_PUBLISH_REQ, l_Pair.Key);

                    foreach (KnownFile l_Known in l_Pair.Value)
                    {
                        FileID l_FileID;

                        if (false == l_Known.GetFileIDVolatile(out l_FileID))
                        {
                            continue;
                        }

                        uint l_CompleteSources;
                        uint l_Sources = f_Kad.f_Map.GetFileSourceCount(l_FileID, out l_CompleteSources);

                        KeywordEntry l_KeywordEntry = KeywordEntry.Create(f_Kad, new KeywordID(l_Pair.Key), l_Known, l_FileID, l_CompleteSources, l_Sources);

                        if (null == l_KeywordEntry)
                        {
                            continue;
                        }

                        int l_OwnLoad;
                        bool l_Added = f_Kad.f_Map.AddKeyword(l_KeywordEntry, out l_OwnLoad);

                        if (true == f_Kad.f_KadPrefs.f_KadFilter.RunFileNameThroughFilter(l_KeywordEntry.f_FileName, KademliaFilterScopes.ScopeOwnPublishingToOthers))
                        {
                            continue;
                        }

                        l_DatagramHlpr.f_Buffer.WriteUInt128ME(l_KeywordEntry.f_RefID.f_Val);
                        TagList.WriteKeywordEntryForPublishing(ref l_WriteTagCntxt, l_KeywordEntry);
                        l_DatagramHlpr.NotifyAdd();
                    }

                    l_DatagramHlpr.End();
                }
            }

            return l_DatagramHlpr.ToArray();
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct DatagramHelper
        {
            private readonly uint f_MaxDatagramSize;

            public readonly MemoryIO f_Buffer;
            private FlexArray<Datagram> f_Arr;

            private UInt128 f_Key;
            private KadOpcodes f_Opcode;
            private ushort f_BlocksWritten;
            private uint f_OldLength;

            public DatagramHelper(uint a_MaxDatagramSize,
                                  int a_InitialArrayCapacity)
            {
                f_MaxDatagramSize = a_MaxDatagramSize;
                f_Buffer = MemoryIO.CreateWritable();
                // Set the buffer capactiy to the size of one
                // datagram plus something.
                f_Buffer.SetCapacity(f_MaxDatagramSize + (2 * 1024));
                FlexArray<Datagram>.Create(a_InitialArrayCapacity, out f_Arr);

                f_Key = null;
                f_Opcode = KadOpcodes.NONE;
                f_BlocksWritten = 0;
                f_OldLength = 0;
            }

            public void Start(KadOpcodes a_Opcode,
                              UInt128 a_KeyID)
            {
                f_Opcode = a_Opcode;
                f_Key = a_KeyID;
                f_BlocksWritten = 0;
                f_OldLength = 0;

                f_Buffer.SetLength(0);
                f_Buffer.WriteUInt128ME(a_KeyID);

                if (KadOpcodes.KADEMLIA_PUBLISH_REQ == f_Opcode)
                {
                    // Reserve space for the count.

                    f_Buffer.JumpAhead(sizeof(ushort));
                }
            }

            public void NotifyAdd()
            {
                uint l_NewSize = f_Buffer.p_Length;

                if (f_MaxDatagramSize < l_NewSize)
                {
                    // We got beyond the allowed size.

                    uint l_HeaderSize;

                    if (0 == f_BlocksWritten)
                    {
                        // Item too big.

                        l_HeaderSize = (UInt128.c_ByteSize + (KadOpcodes.KADEMLIA_PUBLISH_REQ == f_Opcode ? (uint)sizeof(ushort) : 0u));
                        f_Buffer.SetLength(l_HeaderSize);
                        f_Buffer.JumpToEnd();

                        return;
                    }
                    else
                    {
                        l_HeaderSize = ReleaseCore(f_OldLength);
                        f_Buffer.JettisonRange(l_HeaderSize, f_OldLength - l_HeaderSize);
                        f_Buffer.JumpToEnd();
                    }

                    // Keep only the last block that didn't fit into the
                    // released datagram.

                    f_BlocksWritten = 1;
                    f_OldLength = f_Buffer.p_Length;

                    if (f_MaxDatagramSize < f_OldLength)
                    {
                        f_Buffer.SetLength(l_HeaderSize);
                        f_Buffer.JumpToEnd();

                        f_BlocksWritten = 0;
                        f_OldLength = 0;
                    }
                }
                else
                {
                    f_BlocksWritten++;
                    f_OldLength = l_NewSize;
                }
            }

            public void End()
            {
                if (0 != f_BlocksWritten)
                {
                    ReleaseCore(f_OldLength);
                }
            }

            private uint ReleaseCore(uint a_SizeLimit)
            {
                // Returns the size of the header.

                uint l_HeaderSize = UInt128.c_ByteSize;

                if (KadOpcodes.KADEMLIA_PUBLISH_REQ == f_Opcode)
                {
                    l_HeaderSize += sizeof(ushort);

                    f_Buffer.SetPosition(UInt128.c_ByteSize);
                    f_Buffer.WriteUInt16LE(f_BlocksWritten);
                }

                byte[] l_Data = f_Buffer.CopyRange(0, a_SizeLimit);

                int l_Idx = f_Arr.ReserveAtEnd();
                f_Arr.p_Arr[l_Idx] = new Datagram(f_Opcode, f_Key, l_Data);

                return l_HeaderSize;
            }

            public Datagram[] ToArray()
            {
                // Return value is null if the count is zero.

                return (0 == f_Arr.p_Count ? null : f_Arr.ToArray(true));
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Datagram
        {
            // A template of a datagram packet.

            public readonly KadOpcodes f_Opcode;
            public readonly UInt128 f_Key;
            public ByteArray f_Data;

            public Datagram(KadOpcodes a_Opcode,
                            UInt128 a_Key,
                            byte[] a_Data)
            {
                f_Opcode = a_Opcode;
                f_Key = a_Key;
                f_Data = new ByteArray(a_Data);
            }
        }
    }
}

// End of file.
