/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\Links.cs
 * PURPOSE: Handling of links, collections and OS integration
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using NB.Content;
using NB.Content.HashSets;
using NB.Language;
using NB.Network;
using NB.Network.eD2k;
using NB.Compression;
using Microsoft.Win32;
using Support;
using Support.Encodings;
using Support.HashFunctions;

namespace NB
{
    class FileCollection
    {
        private const uint c_NbcMagic = 0x0043424E;
        private const uint c_FileFormatRevision = 0;
        private const uint c_MaxFilesPerCollection = (8 * 1024);
        private const uint c_MaxUncompressedCollectionFileSize = (10 * 1024 * 1024);
        // Limit is in UTF-16 code units.
        public const byte c_MaxTitleChars = 128;

        private readonly FileInfo[] f_Infos;
        public readonly ulong f_TotalSize;

        private string f_Title;

        public uint p_TotalCount
        {
            get { return (uint)f_Infos.Length; }
        }

        public string p_Title
        {
            get { return f_Title; }
        }

        private FileCollection(string a_Title,
                               FileInfo[] a_Infos,
                               ulong a_TotalSize)
        {
            f_Title = a_Title;
            f_Infos = a_Infos;
            f_TotalSize = a_TotalSize;
        }

        public bool ToBuffer(BufferIO a_Buffer)
        {
            if (null == f_Title)
            {
                return false;
            }

            // Write body.

            int l_Count = f_Infos.Length;
            MemoryIO l_Mem = MemoryIO.CreateWritable();
            l_Mem.WriteUTF8StringEncLen(f_Title);
            l_Mem.WriteUInt32LE((uint)l_Count);

            for (int i = 0; i < l_Count; i++)
            {
                l_Mem.WriteUTF8StringEncLen(f_Infos[i].f_FileName);
                l_Mem.WriteUInt64LE(f_Infos[i].f_FileSize);
                f_Infos[i].f_MasterHashes.ToBuffer(l_Mem);
            }

            if (c_MaxUncompressedCollectionFileSize < l_Mem.p_Length)
            {
                return false;
            }

            // Compress.

            ByteArray l_Data = new ByteArray(l_Mem.ToArray(true));

            return zlib.ToCompressedBuffer(ref l_Data, c_NbcMagic, c_FileFormatRevision, a_Buffer);
        }

        public static LoadResult CreateFromBuffer(BufferIO a_Buffer,
                                                  out FileCollection a_Collection)
        {
            a_Collection = null;

            // Uncompress.

            uint l_ActualRevision;
            ByteArray l_Data;
            zlib.DecompressBufferResults l_Res = zlib.FromCompressedBuffer(a_Buffer, c_NbcMagic, 1, c_MaxUncompressedCollectionFileSize, out l_ActualRevision, out l_Data);

            switch (l_Res)
            {
                case zlib.DecompressBufferResults.Success:
                    break;

                case zlib.DecompressBufferResults.ErrSourceBufferTooShort:
                case zlib.DecompressBufferResults.ErrInvalidMagic:
                    return LoadResult.NoCollection;

                case zlib.DecompressBufferResults.ErrFormatTooAdvanced:
                    return LoadResult.TooAdvanced;

                case zlib.DecompressBufferResults.ErrUncompressedSizeExceedsLimit:
                    return LoadResult.UnableToDecompress;

                case zlib.DecompressBufferResults.ErrCompressedBufferCorrupt:
                    return LoadResult.Corruption;

                default: throw (new UnexpectedStateException());
            }

            // Read body.

            MemoryIO l_Mem = MemoryIO.CreateReadable(ref l_Data);
            string l_Title = l_Mem.ReadUTF8StringEncLen();
            uint l_Count = l_Mem.ReadUInt32LE();

            if (c_MaxTitleChars < l_Title.Length ||
                0 == l_Count ||
                c_MaxFilesPerCollection < l_Count)
            {
                return LoadResult.Corruption;
            }

            FileInfo[] l_Infos = new FileInfo[(int)l_Count];
            ulong l_TotalSize = 0;

            for (int i = ((int)l_Count - 1); 0 <= i; i--)
            {
                string l_FileName = l_Mem.ReadUTF8StringEncLen();
                l_FileName = FileNames.TryToMakeFileNameLegal(l_FileName);

                if (null == l_FileName)
                {
                    return LoadResult.FileNameIncompatible;
                }

                ulong l_FileSize = l_Mem.ReadUInt64LE();

                if (false == ContentManager.IsValidFileSize(l_FileSize))
                {
                    return LoadResult.Corruption;
                }

                MasterHashes l_MasterHashes;

                if (false == MasterHashes.FromBuffer(true, l_Mem, out l_MasterHashes) ||
                    false == l_MasterHashes.p_HaseD2kHash)
                {
                    return LoadResult.Corruption;
                }

                l_Infos[i] = new FileInfo(l_FileName, l_FileSize, ref l_MasterHashes);
                l_TotalSize += l_FileSize;

                if (l_FileSize > l_TotalSize)
                {
                    return LoadResult.Corruption;
                }

                if (1 <= l_ActualRevision)
                {
                    uint l_ExtraSize = l_Mem.Read7BitEncUInt32();

                    if (0 != l_ExtraSize)
                    {
                        if (l_Mem.GetRemaining() < l_ExtraSize)
                        {
                            return LoadResult.Corruption;
                        }

                        l_Mem.JumpAhead(l_ExtraSize);
                    }
                }
            }

            a_Collection = new FileCollection(l_Title, l_Infos, l_TotalSize);

            return LoadResult.Success;
        }

        public static LoadResult CreateFromFileArray(KnownFile[] a_Files,
                                                     out FileCollection a_Collection)
        {
            a_Collection = null;

            int l_Count = a_Files.Length;

            if (0 == l_Count ||
                c_MaxFilesPerCollection < l_Count)
            {
                return (0 == l_Count ? LoadResult.NoFiles : LoadResult.TooManyFiles);
            }

            FileInfo[] l_Infos = new FileInfo[l_Count];
            ulong l_TotalSize = 0;

            for (int i = 0; i < l_Count; i++)
            {
                if (false == a_Files[i].HasTrustedHashes())
                {
                    return LoadResult.NeedMoreHashing;
                }

                string l_FileName = a_Files[i].GetFileNameVolatile(false);
                MasterHashes l_MasterHashes;
                MasterHashes.CloneVolatile(ref a_Files[i].f_MasterHashes, out l_MasterHashes);
                l_Infos[i] = new FileInfo(l_FileName, a_Files[i].f_FileSize, ref l_MasterHashes);
                l_TotalSize += a_Files[i].f_FileSize;
            }

            a_Collection = new FileCollection(string.Empty, l_Infos, l_TotalSize);

            return LoadResult.Success;
        }

        public FileInfo[] ExposeArray()
        {
            return f_Infos;
        }

        public void AdjustEnabledState(bool a_Set)
        {
            for (int i = (f_Infos.Length - 1); 0 <= i; i--)
            {
                f_Infos[i].f_Enabled = a_Set;
            }
        }

        public void SetTitle(string a_Title)
        {
            f_Title = a_Title;
        }

        public enum LoadResult : byte
        {
            Success,

            // When creating a new collection.
            NoFiles,
            TooManyFiles,
            NeedMoreHashing,

            // When loading an existing collection.
            NoCollection,
            TooAdvanced,
            UnableToDecompress,
            FileNameIncompatible,
            Corruption,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct FileInfo
        {
            public readonly string f_FileName;
            public readonly ulong f_FileSize;
            public readonly MasterHashes f_MasterHashes;

            public bool f_Enabled;

            public FileInfo(string a_FileName,
                            ulong a_FileSize,
                            ref MasterHashes a_MasterHashes)
            {
                f_FileName = a_FileName;
                f_FileSize = a_FileSize;
                f_MasterHashes = a_MasterHashes;

                f_Enabled = true;
            }
        }
    }

    class FileLink
    {
        private const string c_MagnetPrefix = "magnet:?";
        private const string c_eD2kPrefix = "ed2k://|";
        private const string c_UrlEncodedeD2kPrefix = "ed2k://%7C";
        private const string c_eD2kSuffix = "|/";

        // This will occupy 64 KiB at most.
        private const uint c_MaxPartHashesToRead = 0x1000;

        // All hashs are exactly 32 chars long in string
        // representation.
        private const byte c_eD2kHashStrLen = (2 * MD_4.c_HashByteSize);
        private const byte c_AICHHashStrLen = ((SHA_1.c_HashByteSize / 5) * 8);
        private const byte c_TAICHHashStrLen = ((Tiger.c_HashByteSize / 3) * 4);

        // f_FileName is guaranteed to be legal.
        public readonly string f_FileName;
        // f_FileSize can exceed the application's limit.
        public readonly ulong f_FileSize;
        public readonly MasterHashes f_MasterHashes;

        private FileLink(string a_FileName,
                         ulong a_FileSize)
        {
            // Do not use this constructor directly.

            f_FileName = a_FileName;
            f_FileSize = a_FileSize;
        }

        public FileLink(string a_FileName,
                        ulong a_FileSize,
                        ref MasterHashes a_MasterHashes) :
                   this(a_FileName, a_FileSize)
        {
            f_MasterHashes = a_MasterHashes;
        }

        public FileLink(string a_FileName,
                        ulong a_FileSize,
                        FileID a_FileID) :
                   this(a_FileName, a_FileSize)
        {
            f_MasterHashes = new MasterHashes(a_FileID);
        }

        public bool HasEnoughForExport()
        {
            return (null != f_FileName &&
                    0 != f_FileSize &&
                    true == f_MasterHashes.p_HaseD2kHash);
        }

        public bool CopyLinkToClipboard(CopyLinkToClipboardOps a_Op)
        {
            StringBuilder l_Bld = new StringBuilder();
            bool l_Succ;

            if (CopyLinkToClipboardOps.eD2kUriPlusHtml >= a_Op)
            {
                bool l_Html = (CopyLinkToClipboardOps.eD2kUriPlusHtml == a_Op);
                l_Succ = WriteeD2kUri(P2PHashSets._ALL_HASHSETS_, l_Html, l_Bld);
            }
            else
            {
                bool l_Html = (CopyLinkToClipboardOps.MagnetUriPlusHtml == a_Op);
                l_Succ = WriteMagnetLink(P2PHashSets._ALL_HASHSETS_, null, l_Html, l_Bld);
            }

            if (false == l_Succ)
            {
                return false;
            }

            Util.SetClipboard(l_Bld.ToString());

            return true;
        }

        public bool WriteMagnetLink(P2PHashSets a_RestrictHashes,
                                    string a_TypeSuffix,
                                    bool a_Html,
                                    StringBuilder a_Bld)
        {
            // a_TypeSuffix should be null for a single file
            // magnet link.

            if (false == HasEnoughForExport() ||
                0 == (a_RestrictHashes & P2PHashSets.eD2kHashSet))
            {
                return false;
            }

            if (true == a_Html)
            {
                BeginHtml(a_Bld);
            }

            a_Bld.Append(c_MagnetPrefix);

            if (null != f_FileName)
            {
                WriteMagnetLinkParam(a_Bld, "dn" + a_TypeSuffix + "=", UrlEncoding.UrlEncode(f_FileName));
            }

            if (0 != f_FileSize)
            {
                WriteMagnetLinkParam(a_Bld, "xl" + a_TypeSuffix + "=", Util.ULongToInvString(f_FileSize));
            }

            if (0 != (a_RestrictHashes & P2PHashSets.eD2kHashSet) &&
                true == f_MasterHashes.p_HaseD2kHash)
            {
                WriteMagnetLinkDigest(a_Bld, a_TypeSuffix, "ed2k", f_MasterHashes.eD2kHashAsStr());
            }

            if (0 != (a_RestrictHashes & P2PHashSets.AICHTree) &&
                true == f_MasterHashes.p_HasAICHHash)
            {
                WriteMagnetLinkDigest(a_Bld, a_TypeSuffix, "aich", f_MasterHashes.AICHHashAsStr());
            }

            if (0 != (a_RestrictHashes & P2PHashSets.TAICHTree) &&
                true == f_MasterHashes.p_HasTAICHHash)
            {
                WriteMagnetLinkDigest(a_Bld, a_TypeSuffix, "taich", f_MasterHashes.TAICHHashAsStr());
            }

            // Remove the last '&'.
            a_Bld.Length--;

            if (true == a_Html)
            {
                EndHtml(a_Bld);
            }

            return true;
        }

        private void WriteMagnetLinkParam(StringBuilder a_Bld,
                                          string a_Prefix,
                                          string a_Suffix)
        {
            a_Bld.Append(a_Prefix);
            a_Bld.Append(a_Suffix);
            a_Bld.Append('&');
        }

        private void WriteMagnetLinkDigest(StringBuilder a_Bld,
                                           string a_TypeSuffix,
                                           string a_HashName,
                                           string a_HashStr)
        {
            a_Bld.Append("xt");
            a_Bld.Append(a_TypeSuffix);
            a_Bld.Append("=urn:");
            a_Bld.Append(a_HashName);
            a_Bld.Append(':');
            a_Bld.Append(a_HashStr);
            a_Bld.Append('&');
        }

        public bool WriteeD2kUri(P2PHashSets a_RestrictHashes,
                                 bool a_Html,
                                 StringBuilder a_Bld)
        {
            if (false == HasEnoughForExport() ||
                0 == (a_RestrictHashes & P2PHashSets.eD2kHashSet))
            {
                return false;
            }

            if (true == a_Html)
            {
                BeginHtml(a_Bld);
            }

            // eMule seems to use HTML encoding for the file
            // name within the URI. We use URL encoding.

            a_Bld.Append(c_eD2kPrefix);
            a_Bld.Append("file|");
            a_Bld.Append(UrlEncoding.UrlEncode(f_FileName));
            a_Bld.Append('|');
            a_Bld.Append(Util.ULongToInvString(f_FileSize));
            a_Bld.Append('|');
            a_Bld.Append(f_MasterHashes.eD2kHashAsStr());
            a_Bld.Append('|');

            if (0 != (a_RestrictHashes & P2PHashSets.AICHTree) &&
                true == f_MasterHashes.p_HasAICHHash)
            {
                a_Bld.Append("h=");
                a_Bld.Append(f_MasterHashes.AICHHashAsStr());
                a_Bld.Append('|');
            }

            if (0 != (a_RestrictHashes & P2PHashSets.TAICHTree) &&
                 true == f_MasterHashes.p_HasTAICHHash)
            {
                a_Bld.Append("th=");
                a_Bld.Append(f_MasterHashes.TAICHHashAsStr());
                a_Bld.Append('|');
            }

            a_Bld.Append('/');

            if (true == a_Html)
            {
                EndHtml(a_Bld);
            }

            return true;
        }

        private void BeginHtml(StringBuilder a_Bld)
        {
            a_Bld.Append("<a href=\"");
        }

        private void EndHtml(StringBuilder a_Bld)
        {
            a_Bld.Append("\">");
            a_Bld.Append(HtmlEncoding.HtmlEncode(f_FileName));
            a_Bld.Append("</a>");
        }

        public static bool IsPotentialMagnetLink(string a_Text)
        {
            return Util.StartsWith(a_Text, c_MagnetPrefix);
        }

        public static bool IsPotentialeD2kUri(string a_Text,
                                              bool a_AllowUrlEncoding)
        {
            return (true == Util.StartsWith(a_Text, c_eD2kPrefix) ||
                    (true == a_AllowUrlEncoding &&
                     true == Util.StartsWith(a_Text, c_UrlEncodedeD2kPrefix)));
        }

        public static ParserResult FromMagnetLink(string a_LinkText,
                                                  out FileLink a_Link)
        {
            a_Link = null;

            if (false == IsPotentialMagnetLink(a_LinkText))
            {
                return ParserResult.NoMagnetLink;
            }

            string[] l_Params = Util.Split(a_LinkText, c_MagnetPrefix.Length, '&');
            string l_FileName = null;
            bool l_HadFileSize = false;
            ulong l_FileSize = 0;
            MasterHashes l_MasterHashes = new MasterHashes();

            foreach (string l_Param in l_Params)
            {
                if (true == Util.StartsWith(l_Param, "xt=urn:"))
                {
                    // One of the file hashs.

                    string l_Urn = l_Param.Substring(7);
                    bool l_OK = true;

                    if (true == Util.StartsWith(l_Urn, "ed2k:"))
                    {
                        l_OK = ParseDigest(l_Urn, 5, P2PHashSets.eD2kHashSet, ref l_MasterHashes);
                    }
                    else if (true == Util.StartsWith(l_Urn, "aich:"))
                    {
                        l_OK = ParseDigest(l_Urn, 5, P2PHashSets.AICHTree, ref l_MasterHashes);
                    }
                    else if (true == Util.StartsWith(l_Urn, "taich:"))
                    {
                        l_OK = ParseDigest(l_Urn, 6, P2PHashSets.TAICHTree, ref l_MasterHashes);
                    }

                    if (false == l_OK)
                    {
                        return ParserResult.FailedToReadHashValue;
                    }
                }
                else if (true == Util.StartsWith(l_Param, "xl="))
                {
                    // The (decimal) file size in bytes.

                    if (true == l_HadFileSize)
                    {
                        return ParserResult.RedundantFileSize;
                    }

                    if (false == Util.TryParseUInt64Inv(l_Param, 3, l_Param.Length - 3, out l_FileSize))
                    {
                        return ParserResult.FailedToReadFileSize;
                    }

                    if (0 == l_FileSize)
                    {
                        return ParserResult.FileSizeIllegal;
                    }

                    l_HadFileSize = true;
                }
                else if (true == Util.StartsWith(l_Param, "dn="))
                {
                    // The http encoded file name.

                    if (null != l_FileName)
                    {
                        return ParserResult.RedundantFileName;
                    }

                    l_FileName = UrlEncoding.UrlDecode(l_Param, 3, l_Param.Length - 3);
                    l_FileName = FileNames.MakeFileNameLegalFailsafe(l_FileName);
                }
                else if (true == Util.StartsWith(l_Param, "xt."))
                {
                    return ParserResult.SeveralFilesNotAllowed;
                }
                // Else we don't know/care what it is.
            }

            if (null == l_FileName)
            {
                return ParserResult.FileNameMissing;
            }

            if (false == l_HadFileSize ||
                false == l_MasterHashes.p_HaseD2kHash)
            {
                return ParserResult.IDorSizeMissing;
            }

            a_Link = new FileLink(l_FileName, l_FileSize, ref l_MasterHashes);

            return ParserResult.Success;
        }

        public static ParserResult FromeD2kUri(string a_LinkText,
                                               out FileLink a_Link)
        {
            string l_DecodedLinkText;

            if (false == Util.StartsWith(a_LinkText, c_UrlEncodedeD2kPrefix))
            {
                l_DecodedLinkText = a_LinkText;
            }
            else
            {
                // This link is URL encoded.

                l_DecodedLinkText = UrlEncoding.UrlDecode(a_LinkText);
            }

            return FromeD2kUriCore(l_DecodedLinkText, out a_Link);
        }

        private static ParserResult FromeD2kUriCore(string a_LinkText,
                                                    out FileLink a_Link)
        {
            a_Link = null;

            if (false == IsPotentialeD2kUri(a_LinkText, false) ||
                false == Util.EndsWith(a_LinkText, c_eD2kSuffix) ||
                a_LinkText.Length < (c_eD2kPrefix.Length + c_eD2kSuffix.Length))
            {
                return ParserResult.NoeD2kLink;
            }

            // There will always be at least one element.
            string[] l_Params = Util.Split(a_LinkText, c_eD2kPrefix.Length, a_LinkText.Length - c_eD2kPrefix.Length - c_eD2kSuffix.Length, '|');

            if ("file" != l_Params[0])
            {
                return ParserResult.OnlyFileAllowed;
            }

            if (4 > l_Params.Length)
            {
                return ParserResult.TooFewParams;
            }

            ulong l_FileSize;

            if (false == Util.TryParseUInt64Inv(l_Params[2], out l_FileSize))
            {
                return ParserResult.FailedToReadFileSize;
            }

            if (0 == l_FileSize)
            {
                return ParserResult.FileSizeIllegal;
            }

            MasterHashes l_MasterHashes = new MasterHashes();

            if (false == ParseDigest(l_Params[3], 0, P2PHashSets.eD2kHashSet, ref l_MasterHashes))
            {
                return ParserResult.FailedToReadHashValue;
            }

            bool l_PartHashChunkFound = false;
            byte[] l_PartHashes = null;

            for (int i = 4; i < l_Params.Length; i++)
            {
                string l_Param = l_Params[i];
                bool l_OK = true;

                if (true == Util.StartsWith(l_Param, "h="))
                {
                    l_OK = ParseDigest(l_Param, 2, P2PHashSets.AICHTree, ref l_MasterHashes);
                }
                else if (true == Util.StartsWith(l_Param, "th="))
                {
                    l_OK = ParseDigest(l_Param, 3, P2PHashSets.TAICHTree, ref l_MasterHashes);
                }
                else if (true == Util.StartsWith(l_Param, "p="))
                {
                    // List of part hashs (semicolon separated).

                    // We know that eMule handles this differently because
                    // it includes two part hashs for a file that is exactly
                    // as large as one part, so we won't copy this strange
                    // behaviour.

                    if (true == l_PartHashChunkFound)
                    {
                        return ParserResult.RedundantPartHashes;
                    }

                    l_PartHashChunkFound = true;
                    bool l_GotPartHashCount = (true == ContentManager.IsValidFileSize(l_FileSize) &&
                                               eD2kHashSet.c_MaxFileSize >= l_FileSize);
                    uint l_PartHashCount = (false == l_GotPartHashCount ? 0 : eD2kHashSet.GeteD2kHashSetElementCount(l_FileSize));

                    // We will simply ignore this data if it's
                    // somehow unexpected.
                    if (true == l_GotPartHashCount &&
                        0 != l_PartHashCount &&
                        c_MaxPartHashesToRead >= l_PartHashCount)
                    {
                        uint l_ExpectedSize = (2 + (l_PartHashCount * MD_4.c_HashByteSize * 2) + (l_PartHashCount - 1));

                        if (l_Param.Length == l_ExpectedSize)
                        {
                            l_PartHashes = new byte[(int)l_PartHashCount * MD_4.c_HashByteSize];

                            int k = 2;
                            for (uint d = 0; d < l_PartHashCount; d++, k += (MD_4.c_HashByteSize * 2))
                            {
                                // Check for delimiter.
                                if (2 < k)
                                {
                                    if (':' != l_Param[k])
                                    {
                                        return ParserResult.FailedToReadPartHashes;
                                    }

                                    k++;
                                }

                                try
                                {
                                    HexEncoding.FromHex(l_Param, k, 2 * MD_4.c_HashByteSize, l_PartHashes, (int)d * MD_4.c_HashByteSize);
                                }
                                catch
                                {
                                    return ParserResult.FailedToReadPartHashes;
                                }
                            }

                            ByteArray l_eD2kBytes;
                            bool l_GotHash = l_MasterHashes.ExposeHash(P2PHashSets.eD2kHashSet, out l_eD2kBytes);

                            if (false == eD2kHashSet.ConfirmeD2kUriPartHashes(ref l_eD2kBytes, l_PartHashes))
                            {
                                return ParserResult.SuppliedPartHashNotMatching;
                            }
                        }
                    }
                }
                else if ("/" == l_Param)
                {
                    // iMule uses this. Just ignore it.

                    continue;
                }
                // Else we don't know/care what it is.

                if (false == l_OK)
                {
                    return ParserResult.FailedToReadHashValue;
                }
            }

            string l_FileName = UrlEncoding.UrlDecode(l_Params[1]);
            l_FileName = FileNames.MakeFileNameLegalFailsafe(l_FileName);

            a_Link = new FileLink(l_FileName, l_FileSize, ref l_MasterHashes);

            return ParserResult.Success;
        }

        private static bool ParseDigest(string a_Src,
                                        int a_SrcIdx,
                                        P2PHashSets a_HashSet,
                                        ref MasterHashes a_MasterHashes)
        {
            if (true == a_MasterHashes.HasAllHashes(a_HashSet))
            {
                // We've already handled this one.

                return false;
            }

            // We know at compile time that all hashs
            // have the same length.
            int l_Chars = c_eD2kHashStrLen;

            if (l_Chars != (a_Src.Length - a_SrcIdx))
            {
                return false;
            }

            try
            {
                switch (a_HashSet)
                {
                    case P2PHashSets.eD2kHashSet:
                        a_MasterHashes.SeteD2kHash(HexEncoding.FromHex(a_Src, a_SrcIdx, l_Chars), 0);
                        break;

                    case P2PHashSets.AICHTree:
                        {
                            byte[] l_Hash = Base32.FromBase32(a_Src, a_SrcIdx, l_Chars);

                            if (null == l_Hash)
                            {
                                return false;
                            }

                            a_MasterHashes.SetAICHHash(l_Hash, 0);
                        }
                        break;

                    case P2PHashSets.TAICHTree:
                        {
                            byte[] l_Hash = Base64.FromBase64(a_Src, a_SrcIdx, l_Chars);

                            if (null == l_Hash)
                            {
                                return false;
                            }

                            a_MasterHashes.SetTAICHHash(l_Hash, 0);
                        }
                        break;

                    default: throw (new ApplicationException());
                }
            }
            catch (FormatException)
            {
                return false;
            }

            return true;
        }

        public void AddAsString(Instance a_Inst,
                                StringBuilder a_Bld,
                                bool a_StartWithLineBreaks)
        {
            if (true == a_StartWithLineBreaks)
            {
                a_Bld.AppendLine();
                a_Bld.AppendLine();
            }

            TextSet l_TextSet = a_Inst.f_TextSet;

            if (null != f_FileName)
            {
                l_TextSet.AddText(a_Bld, TextKeys.FilesTabDetailsFileName, "\"" + f_FileName + "\"");
            }

            if (0 != f_FileSize)
            {
                l_TextSet.AddText(a_Bld, TextKeys.FilesTabDetailsFileSize, a_Inst.f_MainForm.BuildLongSizeStr(f_FileSize));
            }

            if (true == f_MasterHashes.p_HaseD2kHash)
            {
                l_TextSet.AddText(a_Bld, TextKeys.FilesTabDetailsFileID, f_MasterHashes.eD2kHashAsStr());
            }

            if (true == f_MasterHashes.p_HasAICHHash)
            {
                l_TextSet.AddText(a_Bld, TextKeys.FilesTabDetailsAICHHash, f_MasterHashes.AICHHashAsStr());
            }

            if (true == f_MasterHashes.p_HasTAICHHash)
            {
                l_TextSet.AddText(a_Bld, TextKeys.FilesTabDetailsTAICHHash, f_MasterHashes.TAICHHashAsStr());
            }
        }

        public enum CopyLinkToClipboardOps : byte
        {
            // CopyLinkToClipboard() relies on this order.

            eD2kUri = 0,
            eD2kUriPlusHtml = 1,
            MagnetUri = 2,
            MagnetUriPlusHtml = 3,
        }

        public enum ParserResult : byte
        {
            Success = 0,

            NoMagnetLink = 1,
            NoeD2kLink = 2,
            SeveralFilesNotAllowed = 3,
            OnlyFileAllowed = 4,
            FileSizeIllegal = 5,
            FileNameMissing = 6,
            IDorSizeMissing = 7,
            TooFewParams = 8,
            FailedToReadFileSize = 9,
            FailedToReadHashValue = 10,
            FailedToReadPartHashes = 11,
            SuppliedPartHashNotMatching = 12,
            RedundantPartHashes = 13,
            RedundantFileName = 14,
            RedundantFileSize = 15,
        }
    }

    static class WindowsIntegration
    {
        // Versions before 1.4.8 used the first edition of the license
        // and had a value named 'LicenseAccepted'.
        private const string c_LicenseAcceptedValueName = "2ndLicenseAccepted";
        private const string c_LicenseAcceptedValue = "Yes";

        private const string c_RunKeyName = "Software\\Microsoft\\Windows\\CurrentVersion\\Run";
        private const string c_CommandSubKeyName = "shell\\open\\command";

        private const string c_URLProtocol = "URL Protocol";
        private const byte c_eD2kProtIdx = 1;
        private const byte c_MagnetProtIdx = 2;
        private const byte c_NBCollectionIdx = 3;

        private static readonly ClassInfo[] s_Classes;

        static WindowsIntegration()
        {
            s_Classes = new ClassInfo[4];
            s_Classes[0] = new ClassInfo(Software.c_AppName, Software.c_AppName + " File Sharing Client", null, false);
            s_Classes[c_eD2kProtIdx] = new ClassInfo("eD2k", "URL: eD2k Protocol", null, true);
            s_Classes[c_MagnetProtIdx] = new ClassInfo("Magnet", "Magnet URI", "application/x-magnet", true);
            s_Classes[c_NBCollectionIdx] = new ClassInfo(".nbc", Software.c_AppName + " Collection", null, false);
        }

        public static bool IsLicenseAcceptedValueSet()
        {
            return HandleLicenseAcceptedValueAccess(OpType.JustCheck);
        }

        public static bool SetLicenseAcceptedValue()
        {
            return HandleLicenseAcceptedValueAccess(OpType.SetOverwrite);
        }

        private static bool HandleLicenseAcceptedValueAccess(OpType a_OpType)
        {
            // This code won't work if the application runs with
            // restricted rights.

            string l_AppRegPath = "Software\\" + Software.c_AppName;

            try
            {
                // We don't need to dispose this one.
                RegistryKey l_CurUser = Registry.CurrentUser;

                if (OpType.JustCheck == a_OpType)
                {
                    // Returns null if the key doesn't exist.
                    using (RegistryKey l_Key = l_CurUser.OpenSubKey(l_AppRegPath, false))
                    {
                        return (null != l_Key &&
                                true == IsString(l_Key.GetValue(c_LicenseAcceptedValueName), c_LicenseAcceptedValue));
                    }
                }
                else if (OpType.Remove == a_OpType)
                {
                    // Returns null if the key doesn't exist.
                    using (RegistryKey l_Key = l_CurUser.OpenSubKey(l_AppRegPath, true))
                    {
                        if (null != l_Key)
                        {
                            l_Key.DeleteValue(c_LicenseAcceptedValueName, false);
                        }

                        // We will land here even if the value didn't exist.

                        return true;
                    }
                }
                else
                {
                    using (RegistryKey l_Key = CreateKey(l_CurUser, l_AppRegPath))
                    {
                        if (OpType.SetDontOverwrite == a_OpType &&
                            null != l_Key.GetValue(c_LicenseAcceptedValueName))
                        {
                            return false;
                        }

                        SetString(l_Key, c_LicenseAcceptedValueName, c_LicenseAcceptedValue);

                        return true;
                    }
                }
            }
            catch
            {
            }

            return false;
        }

        public static bool IsRegisteredForAutoLaunch()
        {
            return HandleAutoLaunchRegistration(OpType.JustCheck);
        }

        public static bool RegisteredForAutoLaunch(bool a_OverwriteExisting)
        {
            return HandleAutoLaunchRegistration(false == a_OverwriteExisting ? OpType.SetDontOverwrite : OpType.SetOverwrite);
        }

        public static bool UnregisteredForAutoLaunch()
        {
            return HandleAutoLaunchRegistration(OpType.Remove);
        }

        private static bool HandleAutoLaunchRegistration(OpType a_OpType)
        {
            // This code won't work if the application runs with
            // restricted rights.

            string l_AppPath = "\"" + Software.sp_ModuleName + "\"";
            bool l_NeedWriteAcc = (OpType.JustCheck != a_OpType);

            try
            {
                // We don't need to dispose this one.
                RegistryKey l_CurUser = Registry.CurrentUser;

                // Returns null if the key doesn't exist.
                using (RegistryKey l_Key = l_CurUser.OpenSubKey(c_RunKeyName, l_NeedWriteAcc))
                {
                    if (OpType.JustCheck == a_OpType)
                    {
                        return (null != l_Key &&
                                true == IsString(l_Key.GetValue(Software.c_AppName), l_AppPath));
                    }
                    else if (OpType.Remove == a_OpType)
                    {
                        if (null == l_Key)
                        {
                            return true;
                        }

                        l_Key.DeleteValue(Software.c_AppName);

                        return true;
                    }
                    else
                    {
                        if (null == l_Key)
                        {
                            return false;
                        }

                        if (OpType.SetDontOverwrite == a_OpType &&
                            null != l_Key.GetValue(Software.c_AppName))
                        {
                            return false;
                        }

                        SetString(l_Key, Software.c_AppName, l_AppPath);

                        return true;
                    }
                }
            }
            catch
            {
            }

            return false;
        }

        public static bool RegisterApplication()
        {
            return HandleShellIntegration(OpType.SetOverwrite, s_Classes[0]);
        }

        public static void ApplyAssociations(ref AssocPrefs a_AssocPrefs)
        {
            OpType l_AutoLaunchType = (false == a_AssocPrefs.f_LaunchWithOS ? OpType.Remove : OpType.SetOverwrite);

            HandleAutoLaunchRegistration(l_AutoLaunchType);
            HandleShellIntegrationConditional(c_eD2kProtIdx, a_AssocPrefs.f_AssocWitheD2kURIs);
            HandleShellIntegrationConditional(c_MagnetProtIdx, a_AssocPrefs.f_AssocWithMagnetURIs);
            HandleShellIntegrationConditional(c_NBCollectionIdx, a_AssocPrefs.f_AssocWithNBCollections);
        }

        public static void HandleShellIntegrationConditional(int a_ClassIdx,
                                                             bool a_Assoc)
        {
            OpType l_Type = (false == a_Assoc ? OpType.Remove : OpType.SetOverwrite);

            HandleShellIntegration(l_Type, s_Classes[a_ClassIdx]);
        }

        private static bool HandleShellIntegration(OpType a_OpType,
                                                   ClassInfo a_Info)
        {
            string l_KeyName = ("Software\\Classes\\" + a_Info.f_Name);
            string l_AppPath = ("\"" + Software.sp_ModuleName + "\"");
            string l_CommandVal = (l_AppPath + " \"%1\"");

            try
            {
                // We don't need to dispose this one.
                RegistryKey l_CurUser = Registry.CurrentUser;

                if (OpType.JustCheck == a_OpType)
                {
                    // Returns null if the key doesn't exist.
                    using (RegistryKey l_Key = l_CurUser.OpenSubKey(l_KeyName, false))
                    {
                        if (null == l_Key ||
                            (true == a_Info.f_IsProtocol &&
                             false == IsString(l_Key.GetValue(c_URLProtocol), string.Empty)))
                        {
                            return false;
                        }

                        using (RegistryKey l_CommandKey = l_Key.OpenSubKey(c_CommandSubKeyName, false))
                        {
                            if (null == l_CommandKey)
                            {
                                return false;
                            }

                            return IsString(l_CommandKey.GetValue(null), l_CommandVal);
                        }
                    }
                }
                else if (OpType.Remove == a_OpType)
                {
                    // Returns null if the key doesn't exist.
                    using (RegistryKey l_CommandKey = l_CurUser.OpenSubKey(l_KeyName + '\\' + c_CommandSubKeyName, false))
                    {
                        if (null == l_CommandKey ||
                            false == IsString(l_CommandKey.GetValue(null), l_CommandVal))
                        {
                            // We aren't registered.

                            return false;
                        }
                    }

                    l_CurUser.DeleteSubKeyTree(l_KeyName);

                    return true;
                }
                else
                {
                    if (OpType.SetDontOverwrite == a_OpType)
                    {
                        // Returns null if the key doesn't exist.
                        using (RegistryKey l_Key = l_CurUser.OpenSubKey(l_KeyName, false))
                        {
                            if (null != l_Key)
                            {
                                return false;
                            }
                        }
                    }
                    else
                    {
                        try
                        {
                            // Delete what's there.
                            l_CurUser.DeleteSubKeyTree(l_KeyName);
                        }
                        catch
                        {
                        }
                    }

                    using (RegistryKey l_Key = CreateKey(l_CurUser, l_KeyName))
                    {
                        try
                        {
                            if (null != a_Info.f_Description)
                            {
                                SetString(l_Key, null, a_Info.f_Description);
                            }

                            if (null != a_Info.f_Content)
                            {
                                SetString(l_Key, "Content Type", a_Info.f_Content);
                            }

                            if (true == a_Info.f_IsProtocol)
                            {
                                SetString(l_Key, c_URLProtocol, string.Empty);
                            }

                            using (RegistryKey l_DefIconKey = CreateKey(l_Key, "DefaultIcon"))
                            {
                                SetString(l_DefIconKey, null, l_AppPath + ",0");
                            }

                            using (RegistryKey l_CommandKey = CreateKey(l_Key, c_CommandSubKeyName))
                            {
                                SetString(l_CommandKey, null, l_CommandVal);
                            }

                            return true;
                        }
                        catch
                        {
                            // Revert the mess we've done.
                            l_CurUser.DeleteSubKeyTree(l_KeyName);
                        }
                    }
                }
            }
            catch
            {
            }

            return false;
        }

        private static bool IsString(object a_Obj,
                                     string a_TestStr)
        {
            // a_Obj may be null.

            return (null != a_Obj &&
                    typeof(string) == a_Obj.GetType() &&
                    a_TestStr == (string)a_Obj);
        }

        private static RegistryKey CreateKey(RegistryKey a_Parent,
                                             string a_SubPath)
        {
            // Throws an exception on any error.

            RegistryKey l_Key = a_Parent.CreateSubKey(a_SubPath);

            if (null == l_Key)
            {
                // We somehow didn't make it.

                throw (new ApplicationException());
            }

            return l_Key;
        }

        private static string GetString(RegistryKey a_Parent,
                                        string a_ValueName)
        {
            // Throws an exception on any error.

            object l_Obj = a_Parent.GetValue(a_ValueName);

            // If it's not a string, an exception will happen
            // (which is what we want).
            return (string)l_Obj;
        }

        private static void SetString(RegistryKey a_Parent,
                                      string a_ValueName,
                                      string a_ValueData)
        {
            // Throws an exception on any error.

            a_Parent.SetValue(a_ValueName, a_ValueData);
        }

        private enum OpType : byte
        {
            JustCheck = 0,
            SetDontOverwrite = 1,
            SetOverwrite = 2,
            Remove = 3,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ClassInfo
        {
            public readonly string f_Name;
            // f_Description and f_Content can be null
            public readonly string f_Description;
            public readonly string f_Content;
            public readonly bool f_IsProtocol;

            public ClassInfo(string a_Name,
                             string a_Description,
                             string a_Content,
                             bool a_IsProtocol)
            {
                f_Name = a_Name;
                f_Description = a_Description;
                f_Content = a_Content;
                f_IsProtocol = a_IsProtocol;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct AssocPrefs
    {
        // File association and other OS integration.

        public readonly bool f_LaunchWithOS;
        public readonly bool f_AssocWitheD2kURIs;
        public readonly bool f_AssocWithMagnetURIs;
        public readonly bool f_AssocWithNBCollections;

        public AssocPrefs(bool a_LaunchWithOS,
                          bool a_AssocWitheD2kURIs,
                          bool a_AssocWithMagnetURIs,
                          bool a_AssocWithNBCollections)
        {
            f_LaunchWithOS = a_LaunchWithOS;
            f_AssocWitheD2kURIs = a_AssocWitheD2kURIs;
            f_AssocWithMagnetURIs = a_AssocWithMagnetURIs;
            f_AssocWithNBCollections = a_AssocWithNBCollections;
        }

        public static void CreateDefault(out AssocPrefs a_Dest)
        {
            a_Dest = new AssocPrefs(false,
                                    false,
                                    false,
                                    true);
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      uint a_Revision,
                                      out AssocPrefs a_Dest)
        {
            bool l_LaunchWithOS = a_Buffer.ReadBool();
            bool l_AssocWitheD2kURIs = a_Buffer.ReadBool();
            bool l_AssocWithMagnetURIs = a_Buffer.ReadBool();
            bool l_AssocWithNBCollections = a_Buffer.ReadBool();

            a_Dest = new AssocPrefs(l_LaunchWithOS, l_AssocWitheD2kURIs, l_AssocWithMagnetURIs, l_AssocWithNBCollections);

            return true;
        }

        public uint ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteBool(f_LaunchWithOS);
            a_Buffer.WriteBool(f_AssocWitheD2kURIs);
            a_Buffer.WriteBool(f_AssocWithMagnetURIs);
            a_Buffer.WriteBool(f_AssocWithNBCollections);

            // Increase with every change.
            return 1;
        }

        public void DetermineChangeKind(ref AssocPrefs a_ChangeTo,
                                        ref PrefChangeKinds a_ChangeKind)
        {
            if (true == Preferences.NeedsReconnOrRestart(a_ChangeKind))
            {
                return;
            }

            if (PrefChangeKinds.ApplicableWithoutReconnect == a_ChangeKind ||
                this.f_LaunchWithOS != a_ChangeTo.f_LaunchWithOS ||
                this.f_AssocWitheD2kURIs != a_ChangeTo.f_AssocWitheD2kURIs ||
                this.f_AssocWithMagnetURIs != a_ChangeTo.f_AssocWithMagnetURIs ||
                this.f_AssocWithNBCollections != a_ChangeTo.f_AssocWithNBCollections)
            {
                a_ChangeKind = PrefChangeKinds.ApplicableWithoutReconnect;
            }
        }
    }

    class KademliaFilter
    {
        public const uint c_NoRuleID = Util.c_NoIndex;
        private const byte c_FormatRevision = 0;
        private const uint c_ConfigMagic = 0xD3EBE60F;
        // The high byte must be clear.
        private const uint c_FileMagic = 0x00C7836C;
        public const ushort c_MaxRuleCount = 256;
        public const uint c_MaxUncompressedByteSize = ((2 * 1024 * 1024) + (c_MaxRuleCount * ((c_MaxTitleChars + c_MaxRulesChars) * sizeof(char))));

        private const byte c_SequenceCount = 3;
        // Limits are in UTF-16 code units.
        public const byte c_MaxTitleChars = 128;
        public const uint c_MaxRuleSequenceChars = (16u * 1024);
        public const uint c_MaxRulesChars = ((c_SequenceCount * c_MaxRuleSequenceChars) + (c_SequenceCount - 1u));
        public const ushort c_MaxPatternChars = 260;
        public const char c_RuleSequenceSeparatorChar = '|';
        public const char c_RulePatternSeparatorChar = Util.c_Space;

        private bool f_Enabled;
        private KademliaFilterScopes f_DefaultScopes;

        // The key is the rule item id.
        private FlexDictionary<uint, FilterItem> f_Items;
        private uint f_ActiveCount;
        // f_RecentID can be c_NoRuleID.
        private uint f_RecentID;

        public bool p_Enabled
        {
            get { return f_Enabled; }
        }

        public KademliaFilterScopes p_DefaultScopes
        {
            get { return f_DefaultScopes; }
        }

        public uint p_Count
        {
            get
            {
                uint l_Count = (uint)f_Items.p_Count;

                return l_Count;
            }
        }

        private KademliaFilter(bool a_Enabled,
                               KademliaFilterScopes a_DefaultScopes)
        {
            f_Enabled = a_Enabled;
            f_DefaultScopes = a_DefaultScopes;
            FlexDictionary<uint, FilterItem>.Create(out f_Items);
            f_ActiveCount = 0;
            f_RecentID = c_NoRuleID;
        }

        private KademliaFilter(bool a_Enabled,
                               KademliaFilterScopes a_DefaultScopes,
                               ref FlexDictionary<uint, FilterItem> a_Items,
                               uint a_ActiveCount,
                               uint a_RecentID)
        {
            f_Enabled = a_Enabled;
            f_DefaultScopes = a_DefaultScopes;
            a_Items.Clone(out f_Items);
            f_ActiveCount = a_ActiveCount;
            f_RecentID = a_RecentID;
        }

        public static KademliaFilter Create()
        {
            const bool lc_DefaultEnabled = false;

            return new KademliaFilter(lc_DefaultEnabled, KademliaFilterScopes.AllSpecificScopes);
        }

        public KademliaFilter Clone()
        {
            return new KademliaFilter(f_Enabled, f_DefaultScopes, ref f_Items, f_ActiveCount, f_RecentID);
        }

        public bool Equals(KademliaFilter a_Other,
                           bool a_CheckFunctionalEquivalenceOnly)
        {
            bool l_ThisHasActiveRules = this.HasActiveRules();

            if (a_Other.HasActiveRules() != l_ThisHasActiveRules)
            {
                return false;
            }

            if (true == a_CheckFunctionalEquivalenceOnly &&
                false == l_ThisHasActiveRules)
            {
                return true;
            }

            uint l_ThisCount = this.p_Count;

            if (false == a_CheckFunctionalEquivalenceOnly &&
                a_Other.p_Count != l_ThisCount)
            {
                return false;
            }

            if (this.f_Enabled != a_Other.f_Enabled ||
                this.f_DefaultScopes != a_Other.f_DefaultScopes ||
                this.f_ActiveCount != a_Other.f_ActiveCount)
            {
                return false;
            }

            FlexDicItem<uint, FilterItem>[] l_OtherArr;
            List<uint> l_OtherLs = null;

            if (false == a_CheckFunctionalEquivalenceOnly)
            {
                l_OtherArr = a_Other.f_Items.ExposeItemArray();
            }
            else
            {
                FlexDicEnum<uint, FilterItem> l_OtherEnum;
                l_OtherArr = a_Other.f_Items.ExposeItemArray(out l_OtherEnum);
                l_OtherLs = new List<uint>((int)a_Other.f_ActiveCount);

                for (int h; true == l_OtherEnum.GetNextIndex(l_OtherArr, out h); )
                {
                    if (true == l_OtherArr[h].f_Val.p_Enabled)
                    {
                        l_OtherLs.Add((uint)h);
                    }
                }
            }

            FlexDicEnum<uint, FilterItem> l_ThisEnum;
            FlexDicItem<uint, FilterItem>[] l_ThisArr = this.f_Items.ExposeItemArray(out l_ThisEnum);

            for (int i; true == l_ThisEnum.GetNextIndex(l_ThisArr, out i); )
            {
                if (false == a_CheckFunctionalEquivalenceOnly)
                {
                    uint l_OtherIdx;

                    if (false == FindItemByTitle(ref a_Other.f_Items, l_ThisArr[i].f_Val.p_Title, true, out l_OtherIdx) ||
                        false == l_ThisArr[i].f_Val.Equals(ref l_OtherArr[l_OtherIdx].f_Val, false))
                    {
                        return false;
                    }
                }
                else if (true == l_ThisArr[i].f_Val.p_Enabled)
                {
                    bool l_Match = false;

                    for (int k = 0, lim_k = l_OtherLs.Count; k < lim_k; k++)
                    {
                        uint l_OtherIdx = l_OtherLs[k];

                        if (true == l_ThisArr[i].f_Val.Equals(ref l_OtherArr[l_OtherIdx].f_Val, true))
                        {
                            l_OtherLs.RemoveAt(k);
                            l_Match = true;

                            break;
                        }
                    }

                    if (false == l_Match)
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        public void ClearItems()
        {
            f_Items.Clear();
            f_ActiveCount = 0;
            f_RecentID = c_NoRuleID;
        }

        public uint[] CreateArrayOfRuleIDs()
        {
            // Returns null if there are no items.

            return f_Items.ToKeyArray(true);
        }

        public bool HasActiveRules()
        {
            return (true == f_Enabled &&
                    0 != f_ActiveCount);
        }

        public bool HasActiveRules(KademliaFilterScopes a_Scopes)
        {
            // a_Scopes is ignored currently.

            return HasActiveRules();
        }

        public void SetEnabledFlagAndDefaultScopes(bool a_Enabled,
                                                   KademliaFilterScopes a_DefaultScopes)
        {
            f_Enabled = a_Enabled;
            f_DefaultScopes = a_DefaultScopes;
        }

        public bool GetRule(uint a_RuleID,
                            out bool a_Enabled,
                            out string a_Title,
                            out KademliaFilterScopes a_Scopes,
                            out string a_Rules)
        {
            int l_Idx;

            if (true == f_Items.TryGet(a_RuleID, out l_Idx))
            {
                FlexDicItem<uint, FilterItem>[] l_Arr = f_Items.ExposeItemArray();
                a_Enabled = l_Arr[l_Idx].f_Val.p_Enabled;
                a_Title = l_Arr[l_Idx].f_Val.p_Title;
                a_Scopes = l_Arr[l_Idx].f_Val.a_Scopes;
                a_Rules = l_Arr[l_Idx].f_Val.p_Rules;

                return true;
            }

            a_Enabled = false;
            a_Scopes = KademliaFilterScopes.None;
            a_Rules = a_Title = null;

            return false;
        }

        private uint FindUnoccupiedRuleID()
        {
            for (uint i = (f_RecentID + 1); true; i++)
            {
                if (c_NoRuleID == i ||
                    true == f_Items.ContainsKey(i))
                {
                    continue;
                }

                return i;
            }
        }

        public bool TryAddRule(bool a_Enabled,
                               string a_Title,
                               KademliaFilterScopes a_Scopes,
                               string a_Rules,
                               out uint a_RuleID)
        {
            // Returns zero if the proposed title is already present
            // or the item limit is reached.

            uint l_Junk;

            if (c_MaxRuleCount > p_Count &&
                false == FindItemByTitle(ref f_Items, a_Title, false, out l_Junk))
            {
                FilterItem l_RuleItem = new FilterItem(a_Enabled, a_Title, a_Scopes, a_Rules);
                AddRuleCore(ref l_RuleItem, out a_RuleID);

                return true;
            }

            a_RuleID = c_NoRuleID;

            return false;
        }

        private bool TryAddRuleWithAutoGeneratedTitle(ref FilterItem a_RuleItem,
                                                      out uint a_RuleID)
        {
            // Returns zero if creating a rule did not succeed.

            if (c_MaxRuleCount <= p_Count)
            {
                goto FAIL;
            }

            StringBuilder l_TitleBld = new StringBuilder(a_RuleItem.p_Title, c_MaxTitleChars);
            int l_TitleCoreLen = Math.Min(l_TitleBld.Length, c_MaxTitleChars - 12);
            FilterItem l_NewItem = a_RuleItem;

            for (uint i = 2; i < (2 + c_MaxRuleCount); i++)
            {
                if (l_TitleBld.Length > l_TitleCoreLen)
                {
                    l_TitleBld.Length = l_TitleCoreLen;
                }

                l_TitleBld.AppendFormat(Util.s_InvariantCulture, " #{0}", i);

                string l_TitleCandid = l_TitleBld.ToString();
                uint l_Junk;

                if (true == FindItemByTitle(ref f_Items, l_TitleCandid, false, out l_Junk))
                {
                    continue;
                }

                l_NewItem.SetTitle(l_TitleCandid);

                AddRuleCore(ref l_NewItem, out a_RuleID);

                return true;
            }

            FAIL:

            a_RuleID = c_NoRuleID;

            return false;
        }

        private void AddRuleCore(ref FilterItem a_RuleItem,
                                 out uint a_RuleID)
        {
            uint l_NewRuleID = FindUnoccupiedRuleID();
            int l_Idx;
            FlexDicItem<uint, FilterItem>[] l_Arr;
            bool l_Added = f_Items.PrepareAdd(l_NewRuleID, out l_Idx, out l_Arr);

            l_Arr[l_Idx].f_Val = a_RuleItem;

            if (true == a_RuleItem.p_Enabled)
            {
                f_ActiveCount++;
            }

            f_RecentID = a_RuleID = l_NewRuleID;
        }

        public void ToggleRuleEnabledFlag(uint a_RuleID)
        {
            int l_Idx;
            bool l_OK = f_Items.TryGet(a_RuleID, out l_Idx);

            FlexDicItem<uint, FilterItem>[] l_Arr = f_Items.ExposeItemArray();
            bool l_Enabled = !l_Arr[l_Idx].f_Val.p_Enabled;

            l_Arr[l_Idx].f_Val.AdjustEnabledFlag(l_Enabled);
            ModifyActiveCountByOne(l_Enabled);
        }

        public bool TryModifyRule(uint a_RuleID,
                                  bool a_Enabled,
                                  string a_Title,
                                  KademliaFilterScopes a_Scopes,
                                  string a_Rules)
        {
            // There must be an item with the rule id a_RuleID. This
            // routine returns zero if the proposed title is already
            // occupied.

            uint l_Junk;

            if (false == FindItemByTitle(ref f_Items, a_Title, false, a_RuleID, out l_Junk))
            {
                FilterItem l_RuleItem = new FilterItem(a_Enabled, a_Title, a_Scopes, a_Rules);
                ModifyRuleCore(a_RuleID, ref l_RuleItem);

                return true;
            }

            return false;
        }

        private void ModifyRuleCore(uint a_RuleID,
                                    ref FilterItem a_RuleItem)
        {
            int l_Idx;
            bool l_OK = f_Items.TryGet(a_RuleID, out l_Idx);

            ModifyRuleCore(ref f_Items.ExposeItemArray()[l_Idx].f_Val, ref a_RuleItem);
        }

        private void ModifyRuleCore(ref FilterItem a_ExistingRuleItem,
                                    ref FilterItem a_InputRuleItem)
        {
            bool l_OldEnabledState = a_ExistingRuleItem.p_Enabled;
            a_ExistingRuleItem = a_InputRuleItem;

            if (a_InputRuleItem.p_Enabled != l_OldEnabledState)
            {
                ModifyActiveCountByOne(!l_OldEnabledState);
            }
        }

        private void ModifyActiveCountByOne(bool a_Inc)
        {
            if (false == a_Inc)
            {
                f_ActiveCount--;
            }
            else
            {
                f_ActiveCount++;
            }
        }

        public bool TryRemoveRule(uint a_RuleID)
        {
            int l_Idx;

            if (false == f_Items.TryGet(a_RuleID, out l_Idx))
            {
                return false;
            }

            bool l_IsActive = f_Items.ExposeItemArray()[l_Idx].f_Val.p_Enabled;
            bool l_Removed = f_Items.Remove(a_RuleID);

            if (true == l_Removed &&
                true == l_IsActive)
            {
                f_ActiveCount--;
            }

            return l_Removed;
        }

        public bool CheckForConflictingTitles(KademliaFilter a_Other)
        {
            // Returns a nonzero value if a conflict is detected.

            uint l_ThisCount = this.p_Count;
            uint l_OtherCount = a_Other.p_Count;

            if (0 == l_ThisCount ||
                0 == l_OtherCount)
            {
                return false;
            }

            if (l_OtherCount < l_ThisCount)
            {
                return a_Other.CheckForConflictingTitles(this);
            }

            FlexDicEnum<uint, FilterItem> l_ThisEnum;
            FlexDicItem<uint, FilterItem>[] l_ThisArr = this.f_Items.ExposeItemArray(out l_ThisEnum);

            for (int i; true == l_ThisEnum.GetNextIndex(l_ThisArr, out i); )
            {
                uint Junk;

                if (true == FindItemByTitle(ref a_Other.f_Items, l_ThisArr[i].f_Val.p_Title, false, out Junk))
                {
                    return true;
                }
            }

            return false;
        }

        public void MergeRules(KademliaFilter a_Other,
                               CollisionMergeMode a_Mode,
                               out bool a_HadOverflow)
        {
            a_HadOverflow = false;

            if (0 == a_Other.p_Count)
            {
                return;
            }

            if (0 == this.p_Count)
            {
                TakeOverItems(a_Other);

                return;
            }

            bool l_HadOverflow = false;
            FlexDicEnum<uint, FilterItem> l_Enum;
            FlexDicItem<uint, FilterItem>[] l_ImpArr = a_Other.f_Items.ExposeItemArray(out l_Enum);

            for (int i; true == l_Enum.GetNextIndex(l_ImpArr, out i); )
            {
                string l_Title = l_ImpArr[i].f_Val.p_Title;
                uint l_ExistingIdx;

                if (false == FindItemByTitle(ref this.f_Items, l_Title, false, out l_ExistingIdx))
                {
                    if (c_MaxRuleCount <= this.p_Count)
                    {
                        l_HadOverflow = true;
                    }
                    else
                    {
                        uint l_Junk;
                        AddRuleCore(ref l_ImpArr[i].f_Val, out l_Junk);
                    }

                    continue;
                }

                switch (a_Mode)
                {
                    case CollisionMergeMode.IgnoreImported:
                        break;

                    case CollisionMergeMode.OverwriteExisting:
                        this.ModifyRuleCore(ref this.f_Items.ExposeItemArray()[l_ExistingIdx].f_Val, ref l_ImpArr[i].f_Val);
                        break;

                    case CollisionMergeMode.MergeImportedIntoExisting:
                    {
                        bool l_HadMergeOverflow;
                        this.f_Items.ExposeItemArray()[l_ExistingIdx].f_Val.Merge(l_ImpArr[i].f_Val.p_Rules, out l_HadMergeOverflow);

                        if (true == l_HadMergeOverflow)
                        {
                            l_HadOverflow = true;
                        }

                        break;
                    }

                    case CollisionMergeMode.ChangeTitleOfImported:
                    {
                        uint l_Junk;

                        if (false == this.TryAddRuleWithAutoGeneratedTitle(ref l_ImpArr[i].f_Val, out l_Junk))
                        {
                            l_HadOverflow = true;
                        }

                        break;
                    }
                }
            }

            if (true == l_HadOverflow)
            {
                a_HadOverflow = true;
            }
        }

        private void TakeOverItems(KademliaFilter a_Other)
        {
            // Replaces all items of this instance with the items
            // of a_Other.

            a_Other.f_Items.Clone(out this.f_Items);
            this.f_ActiveCount = a_Other.f_ActiveCount;
            this.f_RecentID = a_Other.f_RecentID;
        }

        public static StorageStates CreateFromConfig(BufferIO a_Buffer,
                                                     out KademliaFilter a_Dest)
        {
            // Expect magic, length, checksum and content.

            a_Dest = null;

            if (c_ConfigMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            uint l_EffFormatRevision = a_Buffer.ReadByte();

            if (c_FormatRevision < l_EffFormatRevision)
            {
                return StorageStates.TooAdvanced;
            }

            uint l_ContentLen = a_Buffer.ReadUInt32LE();

            if (0 == l_ContentLen)
            {
                return StorageStates.Corruption;
            }

            uint l_ContentChecksum = a_Buffer.ReadUInt32LE();

            // Now we've reached the position where the content starts.

            object l_Adler32 = Adler32.CreateBoxed();
            a_Buffer.HashRange((IHashFunction)l_Adler32, a_Buffer.p_Position, l_ContentLen);

            if (Adler32.GetChecksumFromBox(l_Adler32) != l_ContentChecksum)
            {
                return StorageStates.Corruption;
            }

            return KademliaFilter.FromBuffer(a_Buffer, l_EffFormatRevision, false, out a_Dest);
        }

        public static StorageStates CreateFromBuffer(BufferIO a_Buffer,
                                                     bool a_ExpectNoFurtherData,
                                                     out KademliaFilter a_Dest)
        {
            uint l_ActualRevision;
            ByteArray l_Data;
            zlib.DecompressBufferResults l_Res = zlib.FromCompressedBuffer(a_Buffer, c_FileMagic, c_FormatRevision, c_MaxUncompressedByteSize, out l_ActualRevision, out l_Data);

            if (zlib.DecompressBufferResults.Success != l_Res)
            {
                a_Dest = null;

                return StorageStates.Corruption;
            }

            MemoryIO l_Mem = MemoryIO.CreateReadable(ref l_Data);

            return FromBuffer(l_Mem, l_ActualRevision, a_ExpectNoFurtherData, out a_Dest);
        }

        private static StorageStates FromBuffer(BufferIO a_Buffer,
                                                uint a_FormatRevision,
                                                bool a_ExpectNoFurtherData,
                                                out KademliaFilter a_Dest)
        {
            // This routine will catch all relevant exceptions
            // regarding a_Buffer.

            StorageStates l_State;

            try
            {
                bool l_Enabled = a_Buffer.ReadBool();
                KademliaFilterScopes l_DefaultScopes = (KademliaFilterScopes)a_Buffer.ReadByte();
                uint l_Count = a_Buffer.Read7BitEncUInt32();

                if (false == AreValidScopes(l_DefaultScopes) ||
                    c_MaxRuleCount < l_Count)
                {
                    l_State = StorageStates.Corruption;

                    goto FAIL;
                }

                uint l_ActiveCount = 0;
                FlexDictionary<uint, FilterItem> l_Items;
                FlexDictionary<uint, FilterItem>.Create((int)l_Count, out l_Items);

                for (uint k = 0; k != l_Count; k++)
                {
                    FilterItem l_Item;
                    uint l_Junk;

                    if (false == FilterItem.FromBuffer(a_Buffer, out l_Item) ||
                        true == FindItemByTitle(ref l_Items, l_Item.p_Title, false, out l_Junk))
                    {
                        l_State = StorageStates.Corruption;

                        goto FAIL;
                    }

                    int l_NewIdx;
                    FlexDicItem<uint, FilterItem>[] l_Arr;
                    bool l_Added = l_Items.PrepareAdd(k, out l_NewIdx, out l_Arr);

                    l_Arr[l_NewIdx].f_Val = l_Item;

                    if (true == l_Item.p_Enabled)
                    {
                        l_ActiveCount++;
                    }
                }

                if (true == a_ExpectNoFurtherData &&
                    false == a_Buffer.EndReached())
                {
                    l_State = StorageStates.Corruption;

                    goto FAIL;
                }

                a_Dest = new KademliaFilter(l_Enabled, l_DefaultScopes, ref l_Items, l_ActiveCount, l_Count);

                return StorageStates.Success;
            }
            catch (BufferIO.InterpretBufferException)
            {
                l_State = StorageStates.Corruption;

                goto FAIL;
            }
            catch (BufferIO.FileAccessException)
            {
                l_State = StorageStates.ReadFailed;

                goto FAIL;
            }

            FAIL:

            a_Dest = null;

            return l_State;
        }

        public void ToConfigBuffer(BufferIO a_Buffer)
        {
            // Write magic, length, checksum and content.

            const byte lc_LenAndChecksumSize = (2 * sizeof(uint));

            // Write magic.
            a_Buffer.WriteUInt32LE(c_ConfigMagic);
            a_Buffer.WriteByte(c_FormatRevision);

            // Write content.
            a_Buffer.JumpAhead(lc_LenAndChecksumSize);
            uint l_ContentStartPos = a_Buffer.p_Position;
            ToBuffer(a_Buffer);
            uint l_EndPos = a_Buffer.p_Position;
            uint l_ContentLen = (l_EndPos - l_ContentStartPos);

            // Calc checksum.
            object l_Adler32 = Adler32.CreateBoxed();
            a_Buffer.HashRange((IHashFunction)l_Adler32, l_ContentStartPos, l_ContentLen);

            // Write length.
            a_Buffer.SetPosition(l_ContentStartPos - lc_LenAndChecksumSize);
            a_Buffer.WriteUInt32LE(l_ContentLen);

            // Write checksum.
            a_Buffer.WriteUInt32LE(Adler32.GetChecksumFromBox(l_Adler32));

            // Jump to end.
            a_Buffer.SetPosition(l_EndPos);
        }

        public bool ExportToBuffer(BufferIO a_Buffer,
                                   uint[] a_RuleIDs)
        {
            // a_RuleIDs may be null. Returns zero if an io error
            // occurred.

            MemoryIO l_Mem = MemoryIO.CreateWritable();
            ToBuffer(l_Mem, a_RuleIDs);

            ByteArray l_Data;
            l_Mem.ExposeRange(0, l_Mem.p_Length, out l_Data);

            return zlib.ToCompressedBuffer(ref l_Data, c_FileMagic, c_FormatRevision, a_Buffer);
        }

        private void ToBuffer(BufferIO a_Buffer)
        {
            ToBuffer(a_Buffer, null);
        }

        private void ToBuffer(BufferIO a_Buffer,
                              uint[] a_RuleIDs)
        {
            // If a_RuleIDs is null, all rules get written.

            uint l_Count = (null == a_RuleIDs ? p_Count : Util.GetArrayLength(a_RuleIDs));

            uint l_MaxRequiredSize = (sizeof(bool) + sizeof(byte) + sizeof(uint));

            if (0 != l_Count)
            {
                FlexDicItem<uint, FilterItem>[] l_ItemArr;

                if (null == a_RuleIDs)
                {
                    FlexDicEnum<uint, FilterItem> l_Enum;
                    l_ItemArr = f_Items.ExposeItemArray(out l_Enum);

                    for (int i; true == l_Enum.GetNextIndex(l_ItemArr, out i); )
                    {
                        l_MaxRequiredSize += l_ItemArr[i].f_Val.CalcMaximumWrittenByteSize();
                    }
                }
                else
                {
                    l_ItemArr = f_Items.ExposeItemArray();

                    foreach (uint l_RuleID in a_RuleIDs)
                    {
                        int l_Idx;
                        bool l_OK = f_Items.TryGet(l_RuleID, out l_Idx);

                        l_MaxRequiredSize += l_ItemArr[l_Idx].f_Val.CalcMaximumWrittenByteSize();
                    }
                }
            }

            a_Buffer.EnsureCapacity(a_Buffer.GetCapacity() + l_MaxRequiredSize);

            a_Buffer.WriteBool(f_Enabled);
            a_Buffer.WriteByte((byte)f_DefaultScopes);
            a_Buffer.Write7BitEncUInt32(l_Count);

            if (0 == l_Count)
            {
                return;
            }

            FlexDicItem<uint, FilterItem>[] l_Arr;

            if (null == a_RuleIDs)
            {
                FlexDicEnum<uint, FilterItem> l_Enum;
                l_Arr = f_Items.ExposeItemArray(out l_Enum);

                for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
                {
                    l_Arr[i].f_Val.ToBuffer(a_Buffer);
                }
            }
            else
            {
                l_Arr = f_Items.ExposeItemArray();

                foreach (uint l_RuleID in a_RuleIDs)
                {
                    int l_Idx;
                    bool l_OK = f_Items.TryGet(l_RuleID, out l_Idx);

                    l_Arr[l_Idx].f_Val.ToBuffer(a_Buffer);
                }
            }
        }

        public bool RunKnownFileThroughFilter(Instance a_Inst,
                                              FileID a_FileID,
                                              ulong a_FileSize,
                                              KademliaFilterScopes a_Scope)
        {
            // If there is no known file like that, this routine will
            // return zero.

            if (false == HasActiveRules(a_Scope))
            {
                return false;
            }

            KnownFileCache l_Cache = a_Inst.f_NetworkMngr.f_KnownFileCache;

            if (null != l_Cache)
            {
                KnownFile l_Known = l_Cache.TryGetFile(a_FileID, a_FileSize);

                if (null != l_Known)
                {
                    return RunKnownFileThroughFilter(l_Known, a_Scope);
                }
            }

            return false;
        }

        public bool RunKnownFileThroughFilter(KnownFile a_Known,
                                              KademliaFilterScopes a_Scope)
        {
            if (false == HasActiveRules(a_Scope))
            {
                return false;
            }

            string l_FileName = GetSuitableFileNameFromKnownFileVolatile(a_Known);

            return RunFileNameThroughFilter(l_FileName, a_Scope);
        }

        public static string GetSuitableFileNameFromKnownFileVolatile(KnownFile a_Known)
        {
            return a_Known.GetFileNameVolatile(false);
        }

        public bool RunFileNameThroughFilter(string a_FileName,
                                             KademliaFilterScopes a_Scope)
        {
            if (false == HasActiveRules(a_Scope))
            {
                return false;
            }

            return RunFileNameThroughFilterCore(a_FileName, a_Scope);
        }

        private bool RunFileNameThroughFilterCore(string a_FileName,
                                                  KademliaFilterScopes a_Scope)
        {
            FilterParams l_Params = new FilterParams(f_DefaultScopes, a_Scope, false, a_FileName, null);

            return (KademliaFilterScopes.None != RunParamsThroughFilter(ref l_Params));
        }

        public KademliaFilterScopes TestIfFileNameGetsFilteredOut(string a_FileName,
                                                                  KademliaFilterScopes a_Scopes)
        {
            // a_Scopes may be a combination of scopes.

            if (false == HasActiveRules(a_Scopes))
            {
                return KademliaFilterScopes.None;
            }

            FilterParams l_Params = new FilterParams(f_DefaultScopes, a_Scopes, true, a_FileName, null);

            return RunParamsThroughFilter(ref l_Params);
        }

        private KademliaFilterScopes RunParamsThroughFilter(ref FilterParams a_Params)
        {
            // Returns KademliaFilterScopes.None if the input did not
            // get filtered out.

            KademliaFilterScopes l_Scopes = KademliaFilterScopes.None;

            FlexDicEnum<uint, FilterItem> l_Enum;
            FlexDicItem<uint, FilterItem>[] l_Arr = f_Items.ExposeItemArray(out l_Enum);

            for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
            {
                if (true == a_Params.CanAbort(l_Scopes))
                {
                    break;
                }

                if (false == l_Arr[i].f_Val.p_Enabled)
                {
                    continue;
                }

                string l_Rules = l_Arr[i].f_Val.p_Rules;

                if (0 == Util.GetStringLength(l_Rules))
                {
                    continue;
                }

                KademliaFilterScopes l_RuleScopes = l_Arr[i].f_Val.GetEffectiveScopes(a_Params.f_DefaultScopes);

                if (KademliaFilterScopes.None == (a_Params.f_TestScopes & l_RuleScopes))
                {
                    continue;
                }

                string[] l_Keywords = a_Params.TryGetKeywords();

                if (null == l_Keywords)
                {
                    break;
                }

                if (true == DoKeywordsMatchRule(l_Keywords, l_Rules))
                {
                    l_Scopes |= l_RuleScopes;
                }
            }

            return l_Scopes;
        }

        private static bool DoKeywordsMatchRule(string[] a_Keywords,
                                                string a_Rules)
        {
            RulesEnumerator l_RulesEnum = new RulesEnumerator(a_Rules);
            bool l_HadANDorORSeq = false;

            for (uint i = 0; c_SequenceCount != i; i++)
            {
                if (0 != i &&
                    false == l_RulesEnum.AdvanceToNextSequence())
                {
                    break;
                }

                if (0 == l_RulesEnum.p_CurrentSequenceLength)
                {
                    continue;
                }

                if (1 >= i)
                {
                    l_HadANDorORSeq = true;
                }
                else if (false == l_HadANDorORSeq)
                {
                    return false;
                }

                bool l_AND = (0 == i);
                bool l_RequiredResult = (2 != i);

                if (DoKeywordsMatchRuleSequence(a_Keywords, l_AND, ref l_RulesEnum) != l_RequiredResult)
                {
                    return false;
                }
            }

            return l_HadANDorORSeq;
        }

        private static bool DoKeywordsMatchRuleSequence(string[] a_Keywords,
                                                        bool a_AND,
                                                        ref RulesEnumerator a_Enum)
        {
            // Expects the a_Keywords to contain at least one element.

            const bool lc_CaseSensitive = false;
            uint l_PatternIdx;
            uint l_PatternLen;

            while (true == a_Enum.TryGetNextPattern(out l_PatternIdx, out l_PatternLen))
            {
                bool l_Match = Util.DoesAnyStringMatchPattern(a_Keywords, a_Enum.f_Rules, l_PatternIdx, l_PatternLen, lc_CaseSensitive);

                if (false == a_AND)
                {
                    if (true == l_Match)
                    {
                        return true;
                    }
                }
                else if (false == l_Match)
                {
                    return false;
                }
            }

            return a_AND;
        }

        private static bool FindItemByTitle(ref FlexDictionary<uint, FilterItem> a_Items,
                                            string a_Title,
                                            bool a_CaseSensitive,
                                            out uint a_Index)
        {
            return FindItemByTitle(ref a_Items, a_Title, a_CaseSensitive, c_NoRuleID, out a_Index);
        }

        private static bool FindItemByTitle(ref FlexDictionary<uint, FilterItem> a_Items,
                                            string a_Title,
                                            bool a_CaseSensitive,
                                            uint a_IgnoreRuleID,
                                            out uint a_Index)
        {
            // Looks for a rule item with the passed title. Any rule
            // item with the rule id a_IgnoreRuleID will be ignored.

            FlexDicEnum<uint, FilterItem> l_Enum;
            FlexDicItem<uint, FilterItem>[] l_Arr = a_Items.ExposeItemArray(out l_Enum);

            for (int i; true == l_Enum.GetNextIndex(l_Arr, out i); )
            {
                if (c_NoRuleID != a_IgnoreRuleID &&
                    a_IgnoreRuleID == l_Arr[i].p_Key)
                {
                    continue;
                }

                if (true == Util.StringsEqual(a_Title, l_Arr[i].f_Val.p_Title, a_CaseSensitive))
                {
                    a_Index = (uint)i;

                    return true;
                }
            }

            a_Index = Util.c_NoIndex;

            return false;
        }

        public static bool IsValidTitle(string a_Title)
        {
            uint l_Len = (uint)a_Title.Length;

            if (0 == l_Len ||
                c_MaxTitleChars < l_Len ||
                false == Util.IsStringUnicodeNormalized(a_Title))
            {
                return false;
            }

            for (uint i = 0, lim = (l_Len - 1); i != l_Len; i++)
            {
                bool l_Brink = (0 == i ||
                                lim == i);

                if (false == IsValidTitleChar(a_Title[(int)i], l_Brink))
                {
                    return false;
                }
            }

            return true;
        }

        private static bool IsValidTitleChar(char a_Char,
                                             bool a_Brink)
        {
            if (Util.c_Space > a_Char)
            {
                return false;
            }

            if (true == a_Brink &&
                Util.c_Space == a_Char)
            {
                return false;
            }

            return true;
        }

        public static bool AreValidScopes(KademliaFilterScopes a_Scopes)
        {
            return (0 == (uint)(a_Scopes & ~(KademliaFilterScopes.AllSpecificScopes | KademliaFilterScopes.UseDefault)));
        }

        public static bool IsDefaultScopesFlagSet(KademliaFilterScopes a_Scopes)
        {
            return (KademliaFilterScopes.None != (a_Scopes & KademliaFilterScopes.UseDefault));
        }

        public static string JoinRuleSequences(string a_AndSeg,
                                               string a_OrSeg,
                                               string a_NotSeg)
        {
            // Each input string may be null or should have a nonzero
            // length otherwise.

            string l_Rules;

            if (null == a_NotSeg)
            {
                if (null == a_OrSeg)
                {
                    l_Rules = (null == a_AndSeg ? string.Empty : a_AndSeg);
                }
                else
                {
                    l_Rules = (a_AndSeg + c_RuleSequenceSeparatorChar + a_OrSeg);
                }
            }
            else
            {
                l_Rules = (a_AndSeg + c_RuleSequenceSeparatorChar + a_OrSeg + c_RuleSequenceSeparatorChar + a_NotSeg);
            }

            return l_Rules;
        }

        public static void SplitRuleSequences(string a_Rules,
                                              out string a_AndSeg,
                                              out string a_OrSeg,
                                              out string a_NotSeg)
        {
            // Each output string can be null on return.

            string[] l_Strs = Util.Split(a_Rules, 0, c_RuleSequenceSeparatorChar);
            a_AndSeg = TryGetStringFromArray(l_Strs, 0);
            a_OrSeg = TryGetStringFromArray(l_Strs, 1);
            a_NotSeg = TryGetStringFromArray(l_Strs, 2);
        }

        private static string TryGetStringFromArray(string[] a_Arr,
                                                    uint a_Idx)
        {
            // This a helper for SplitRuleSequences().

            if (Util.GetArrayLength(a_Arr) > a_Idx)
            {
                string l_Str = a_Arr[a_Idx];

                return (0 == l_Str.Length ? null : l_Str);
            }

            return null;
        }

        private static bool ValidateRules(string a_Rules,
                                          bool a_FixIfBroken,
                                          bool a_CallerReadingBuffer,
                                          out string a_PatchedRules)
        {
            // a_PatchedRules will be set on return if a_FixIfBroken
            // is nonzero, regardless of the input strings correctness.
            // If a_FixIfBroken is zero a_PatchedRules will be null on
            // return in any case.

            StringBuilder l_Bld = null;

            uint l_TotalLen = Util.GetStringLength(a_Rules);
            uint l_Idx = 0;
            uint l_Len = l_TotalLen;

            for (uint i = 0; true; i++)
            {
                if (0 == l_Len)
                {
                    if (null != l_Bld)
                    {
                        goto FAIL_WITH_FIX;
                    }

                    uint l_SurplusLen = CountTrailingSeqSeparators(a_Rules);

                    if (0 == l_SurplusLen)
                    {
                        a_PatchedRules = (false == a_FixIfBroken ? null : a_Rules);

                        return true;
                    }

                    if (false == a_FixIfBroken)
                    {
                        goto FAIL_NO_FIX;
                    }

                    a_PatchedRules = a_Rules.Substring(0, (int)(l_TotalLen - l_SurplusLen));

                    return false;
                }

                int l_SepIdx = a_Rules.IndexOf(c_RuleSequenceSeparatorChar, (int)l_Idx, (int)l_Len);

                if (0 <= l_SepIdx)
                {
                    l_Len = ((uint)l_SepIdx - l_Idx);
                }

                if (null == l_Bld)
                {
                    bool l_Valid = ValidateSequence(a_Rules, l_Idx, l_Len);

                    if (false == l_Valid)
                    {
                        if (false == a_FixIfBroken)
                        {
                            goto FAIL_NO_FIX;
                        }

                        l_Bld = new StringBuilder(a_Rules, 0, (int)l_Idx, (int)Math.Min(c_MaxRulesChars, l_TotalLen));
                    }
                }
                else if (0 != i)
                {
                    l_Bld.Append(c_RuleSequenceSeparatorChar);
                }

                if (null != l_Bld)
                {
                    FixSequence(a_Rules, l_Idx, l_Len, l_Bld);
                }

                bool l_CorruptTail = (0 <= l_SepIdx &&
                                      (c_SequenceCount - 1) <= i);

                if (true == l_CorruptTail)
                {
                    if (null != l_Bld)
                    {
                        goto FAIL_WITH_FIX;
                    }

                    if (false == a_FixIfBroken)
                    {
                        goto FAIL_NO_FIX;
                    }

                    a_PatchedRules = a_Rules.Substring(0, l_SepIdx);

                    return false;
                }

                l_Idx += (l_Len + Util.BoolToByte(0 <= l_SepIdx));

                l_Len = (l_TotalLen - l_Idx);
            }

            FAIL_WITH_FIX:

            RemoveTrailingSeqSeparators(l_Bld);

            a_PatchedRules = l_Bld.ToString();

            return false;

            FAIL_NO_FIX:

            a_PatchedRules = null;

            return false;
        }

        private static uint CountTrailingSeqSeparators(string a_Rules)
        {
            uint l_Count = 0;

            for (uint i = Util.GetStringLength(a_Rules); 0 != i--;)
            {
                if (c_RuleSequenceSeparatorChar != a_Rules[(int)i])
                {
                    break;
                }

                l_Count++;
            }

            return l_Count;
        }

        private static void RemoveTrailingSeqSeparators(StringBuilder a_Bld)
        {
            uint l_Len = (uint)a_Bld.Length;

            for (uint i = 0; 0 != l_Len;)
            {
                if (c_RuleSequenceSeparatorChar == a_Bld[(int)l_Len - 1])
                {
                    l_Len--;

                    if ((c_SequenceCount - 1u) > i++)
                    {
                        continue;
                    }
                }

                break;
            }

            a_Bld.Length = (int)l_Len;
        }

        private static bool ValidateSequence(string a_Str,
                                             uint a_Idx,
                                             uint a_Len)
        {
            if (0 == a_Len)
            {
                return true;
            }

            if (c_MaxRuleSequenceChars < a_Len)
            {
                return false;
            }

            uint l_PrevIdx = Util.c_NoIndex;
            uint l_PrevLen = 0;

            for (uint l_Idx = a_Idx, l_FinalIdx = (a_Idx + a_Len); true; )
            {
                uint l_Len = (l_FinalIdx - l_Idx);
                int l_SepIdx = a_Str.IndexOf(c_RulePatternSeparatorChar, (int)l_Idx, (int)l_Len);

                if (0 <= l_SepIdx)
                {
                    l_Len = ((uint)l_SepIdx - l_Idx);
                }

                if (false == IsValidPattern(a_Str, l_Idx, l_Len))
                {
                    return false;
                }

                if (Util.c_NoIndex != l_PrevIdx &&
                    0 <= Util.CompareStringsOrdinal(a_Str, l_PrevIdx, l_PrevLen, a_Str, l_Idx, l_Len))
                {
                    return false;
                }

                if (0 > l_SepIdx)
                {
                    return true;
                }

                l_PrevIdx = l_Idx;
                l_PrevLen = l_Len;

                l_Idx += (l_Len + 1);
            }
        }

        private static void FixSequence(string a_Str,
                                        uint a_Idx,
                                        uint a_Len,
                                        StringBuilder a_Bld)
        {
            // The fixed sequence will be written to a_Bld.

            if (0 == a_Len)
            {
                return;
            }

            uint l_InitialCapacity = (Math.Min(c_MaxRuleSequenceChars, a_Len) / 10);
            FlexArray<string> l_Ls;
            FlexArray<string>.Create((int)l_InitialCapacity, out l_Ls);
            StringBuilder l_PatternBld = new StringBuilder(c_MaxPatternChars);
            uint l_ListLen = 0;

            for (uint l_Idx = a_Idx, l_FinalIdx = (a_Idx + a_Len); true; )
            {
                uint l_Len = (l_FinalIdx - l_Idx);

                if (0 == l_Len)
                {
                    break;
                }

                int l_SepIdx = a_Str.IndexOf(c_RulePatternSeparatorChar, (int)l_Idx, (int)l_Len);

                if (0 <= l_SepIdx)
                {
                    l_Len = ((uint)l_SepIdx - l_Idx);

                    if (0 == l_Len)
                    {
                        for (l_Idx++; l_Idx != l_FinalIdx; l_Idx++)
                        {
                            if (c_RulePatternSeparatorChar != a_Str[(int)l_Idx])
                            {
                                break;
                            }
                        }

                        continue;
                    }
                }

                bool l_WasWritten;
                string l_NewPattern;

                if (false == IsValidPattern(a_Str, l_Idx, l_Len, l_PatternBld, out l_WasWritten))
                {
                    if (false == l_WasWritten)
                    {
                        l_NewPattern = null;
                    }
                    else
                    {
                        l_NewPattern = l_PatternBld.ToString();

                        l_PatternBld.Length = 0;
                    }
                }
                else
                {
                    l_NewPattern = a_Str.Substring((int)l_Idx, (int)l_Len);
                }

                if (null != l_NewPattern)
                {
                    uint l_NewListLen = (l_ListLen + Util.BoolToByte(0 != l_ListLen) + Util.GetStringLength(l_NewPattern));

                    if (c_MaxRuleSequenceChars < l_NewListLen)
                    {
                        break;
                    }

                    int l_ArrIdx = l_Ls.ReserveAtEnd();
                    l_Ls.p_Arr[l_ArrIdx] = l_NewPattern;
                    l_ListLen = l_NewListLen;
                }

                if (0 > l_SepIdx)
                {
                    break;
                }

                l_Idx += (l_Len + 1);
            }

            l_Ls.Sort(ComparePatternsByRef);

            uint l_Count = (uint)l_Ls.p_Count;
            string[] l_Arr = l_Ls.p_Arr;
            string l_PrevStr = null;

            for (uint k = 0; k != l_Count; k++)
            {
                string l_CurStr = l_Arr[k];

                if (null == l_PrevStr ||
                    false == Util.StringsEqual(l_CurStr, l_PrevStr))
                {
                    if (null != l_PrevStr)
                    {
                        a_Bld.Append(c_RulePatternSeparatorChar);
                    }

                    a_Bld.Append(l_CurStr);

                    l_PrevStr = l_CurStr;
                }
            }
        }

        private static int ComparePatternsByRef(ref string a_StrA,
                                                ref string a_StrB)
        {
            return Util.CompareStringsOrdinal(a_StrA, a_StrB);
        }

        public static bool IsValidPattern(string a_Str,
                                          uint a_Idx,
                                          uint a_Len)
        {
            bool l_Junk;

            return IsValidPattern(a_Str, a_Idx, a_Len, null, out l_Junk);
        }

        public static bool IsValidPattern(string a_Str,
                                          uint a_Idx,
                                          uint a_Len,
                                          StringBuilder a_Bld,
                                          out bool a_WasWritten)
        {
            // a_Bld may be null. If it's not null and a_WasWritten is
            // nonzero on return, then a_Bld was filled with a
            // valid pattern. Returns a value indicating if the input
            // string contains a valid pattern.

            a_WasWritten = false;

            if (0 == a_Len)
            {
                return false;
            }

            bool l_HadError = (c_MaxPatternChars < a_Len);

            if (true == l_HadError &&
                null == a_Bld)
            {
                return false;
            }

            uint l_BldInitialPos = (null == a_Bld ? 0 : (uint)a_Bld.Length);

            for (uint i = a_Idx, lim = (a_Idx + a_Len), k = 0; i != lim; i++)
            {
                char l_Char = a_Str[(int)i];

                bool l_IsLegalChar = (true == Util.IsWildcardChar(l_Char) ||
                                      (true == FileNames.IsLegalFileNameChar(l_Char) &&
                                       false == SearchManager.IsInvalidKeywordChar(l_Char)));

                if (false == l_IsLegalChar)
                {
                    if (false == l_HadError)
                    {
                        if (null == a_Bld)
                        {
                            return false;
                        }

                        l_HadError = true;
                        k = (i - a_Idx);

                        if (0 != k)
                        {
                            a_Bld.Append(a_Str, (int)a_Idx, (int)k);
                        }
                    }
                }
                else if (true == l_HadError &&
                         c_MaxPatternChars > k)
                {
                    k++;
                    a_Bld.Append(l_Char);
                }
            }

            const bool lc_Strict = true;

            if (false == l_HadError)
            {
                bool l_WasGood = Util.IsWildcardPatternOrderly(a_Str, a_Idx, a_Len, lc_Strict, a_Bld);

                if (false == l_WasGood &&
                    null != a_Bld)
                {
                    a_WasWritten = true;
                }

                return l_WasGood;
            }

            uint l_StrLen = ((uint)a_Bld.Length - l_BldInitialPos);

            if (0 == l_StrLen)
            {
                return false;
            }

            string l_Str = a_Bld.ToString((int)l_BldInitialPos, (int)l_StrLen);

            a_Bld.Length = (int)l_BldInitialPos;

            if (true == Util.IsWildcardPatternOrderly(l_Str, 0, l_StrLen, lc_Strict, a_Bld))
            {
                a_Bld.Append(l_Str);
            }

            a_WasWritten = true;

            return false;
        }

        public static string NormalizeAndLowercaseString(string a_RawStr)
        {
            return Util.UnicodeNormalizeString(a_RawStr).ToLowerInvariant();
        }

        public enum CollisionMergeMode : byte
        {
            Undefined = 0,

            IgnoreImported = 1,
            OverwriteExisting = 2,
            MergeImportedIntoExisting = 3,
            ChangeTitleOfImported = 4,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct FilterParams
        {
            public readonly KademliaFilterScopes f_DefaultScopes;
            public readonly KademliaFilterScopes f_TestScopes;
            public readonly bool f_Exhaustive;

            private readonly string f_FileName;
            private string[] f_Keywords;
            private bool f_NoKeywords;

            public FilterParams(KademliaFilterScopes a_DefaultScopes,
                                KademliaFilterScopes a_TestScopes,
                                bool a_Exhaustive,
                                string a_FileName,
                                string[] a_Keywords)
            {
                // Each a_FileName and a_Keywords may be null, but not both.

                f_DefaultScopes = a_DefaultScopes;
                f_TestScopes = (a_TestScopes & KademliaFilterScopes.AllSpecificScopes);
                f_Exhaustive = a_Exhaustive;

                f_FileName = a_FileName;
                f_Keywords = a_Keywords;

                if (null == a_Keywords ||
                    0 < a_Keywords.Length)
                {
                    f_NoKeywords = false;
                }
                else
                {
                    f_Keywords = null;
                    f_NoKeywords = true;
                }
            }

            public bool CanAbort(KademliaFilterScopes a_Scopes)
            {
                // a_Scopes may be KademliaFilterScopes.None.

                return (false == f_Exhaustive ? (KademliaFilterScopes.None != a_Scopes) : ((a_Scopes & f_TestScopes) == f_TestScopes));
            }

            public string[] TryGetKeywords()
            {
                // Return can be null, if there are not keywords.

                if (true == f_NoKeywords)
                {
                    return null;
                }

                if (null == f_Keywords)
                {
                    f_Keywords = SearchManager.IdentifyKeywords(f_FileName, false);

                    if (null == f_Keywords)
                    {
                        f_NoKeywords = true;
                    }
                }

                return f_Keywords;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct FilterItem
        {
            // Don't use the struct default constructor.

            private static uint s_MaxRulesByteSize;

            private bool f_Enabled;
            private string f_Title;
            private KademliaFilterScopes f_Scopes;
            private string f_Rules;

            public bool p_Enabled
            {
                get { return f_Enabled; }
            }

            public string p_Title
            {
                get { return f_Title; }
            }

            public KademliaFilterScopes a_Scopes
            {
                get { return f_Scopes; }
            }

            public string p_Rules
            {
                get { return f_Rules; }
            }

            static FilterItem()
            {
                s_MaxRulesByteSize = Util.CalcStringWorstCaseByteSize(c_MaxRulesChars);
            }

            public FilterItem(bool a_Enabled,
                              string a_Title,
                              KademliaFilterScopes a_Scopes,
                              string a_Rules)
            {
                f_Enabled = a_Enabled;
                f_Title = a_Title;
                f_Scopes = a_Scopes;
                f_Rules = a_Rules;
            }

            public void AdjustEnabledFlag(bool a_Enabled)
            {
                f_Enabled = a_Enabled;
            }

            public void SetTitle(string a_Title)
            {
                f_Title = a_Title;
            }

            public void Merge(string a_OtherRules,
                              out bool a_HadOverflow)
            {
                a_HadOverflow = false;

                if (0 == a_OtherRules.Length)
                {
                    return;
                }

                if (0 == this.f_Rules.Length)
                {
                    this.f_Rules = a_OtherRules;

                    return;
                }

                if (true == Util.StringsEqual(this.f_Rules, a_OtherRules))
                {
                    return;
                }

                RulesEnumerator l_ThisEnum = new RulesEnumerator(this.f_Rules);
                RulesEnumerator l_OtherEnum = new RulesEnumerator(a_OtherRules);

                uint l_Capacity = Math.Min(c_MaxRulesChars, (uint)(this.f_Rules.Length + a_OtherRules.Length));
                StringBuilder l_Bld = new StringBuilder((int)l_Capacity);
                bool l_HadOverflow = false;

                for (uint i = 0; c_SequenceCount != i; i++)
                {
                    if (0 != i)
                    {
                        l_Bld.Append(c_RuleSequenceSeparatorChar);

                        if (false == l_ThisEnum.p_IsEndOfSequencesReached)
                        {
                            l_ThisEnum.AdvanceToNextSequence();
                        }

                        if (false == l_OtherEnum.p_IsEndOfSequencesReached)
                        {
                            l_OtherEnum.AdvanceToNextSequence();
                        }
                    }

                    MergeSequence(ref l_ThisEnum, ref l_OtherEnum, l_Bld, ref l_HadOverflow);
                }

                RemoveTrailingSeqSeparators(l_Bld);

                string l_MergedRules = l_Bld.ToString();

                f_Rules = l_MergedRules;

                if (true == l_HadOverflow)
                {
                    a_HadOverflow = true;
                }
            }

            private static void MergeSequence(ref RulesEnumerator a_AEnum,
                                              ref RulesEnumerator a_BEnum,
                                              StringBuilder a_Bld,
                                              ref bool a_HadOverflow)
            {
                // This is a helper routine for Merge().

                bool l_HasA = (false == a_AEnum.p_IsEndOfSequencesReached &&
                               0 != a_AEnum.p_CurrentSequenceLength);
                bool l_HasB = (false == a_BEnum.p_IsEndOfSequencesReached &&
                               0 != a_BEnum.p_CurrentSequenceLength);

                if (false == l_HasA)
                {
                    if (true == l_HasB)
                    {
                        a_BEnum.AppendFullSequence(a_Bld);
                    }

                    return;
                }

                if (false == l_HasB ||
                    true == Util.StringsEqual(a_AEnum.f_Rules, a_AEnum.p_CurrentSequenceIndex, a_AEnum.p_CurrentSequenceLength, a_BEnum.f_Rules, a_BEnum.p_CurrentSequenceIndex, a_BEnum.p_CurrentSequenceLength))
                {
                    a_AEnum.AppendFullSequence(a_Bld);

                    return;
                }

                uint l_AIdx = Util.c_NoIndex, l_BIdx = Util.c_NoIndex;
                uint l_ALen = 0, l_BLen = 0;
                uint l_MaxBldLen = ((uint)a_Bld.Length + c_MaxRuleSequenceChars);

                for (bool l_NeedSep = false; true;)
                {
                    if (Util.c_NoIndex == l_AIdx &&
                        false == a_AEnum.TryGetNextPattern(out l_AIdx, out l_ALen))
                    {
                        if (Util.c_NoIndex != l_BIdx ||
                            true == a_BEnum.TryGetNextPattern(out l_BIdx, out l_BLen))
                        {
                            AddToBuilder(a_Bld, l_MaxBldLen, a_BEnum.f_Rules, l_BIdx, a_BEnum.GetCurrentSequenceTailLength(l_BIdx), ref l_NeedSep, ref a_HadOverflow);
                        }

                        return;
                    }

                    if (Util.c_NoIndex == l_BIdx &&
                        false == a_BEnum.TryGetNextPattern(out l_BIdx, out l_BLen))
                    {
                        AddToBuilder(a_Bld, l_MaxBldLen, a_AEnum.f_Rules, l_AIdx, a_AEnum.GetCurrentSequenceTailLength(l_AIdx), ref l_NeedSep, ref a_HadOverflow);

                        return;
                    }

                    int l_Res = Util.CompareStringsOrdinal(a_AEnum.f_Rules, l_AIdx, l_ALen, a_BEnum.f_Rules, l_BIdx, l_BLen);
                    bool l_HadOverflow = false;

                    if (0 >= l_Res)
                    {
                        AddToBuilder(a_Bld, l_MaxBldLen, a_AEnum.f_Rules, l_AIdx, l_ALen, ref l_NeedSep, ref l_HadOverflow);
                        l_AIdx = Util.c_NoIndex;

                        if (0 == l_Res)
                        {
                            l_BIdx = Util.c_NoIndex;
                        }
                    }
                    else
                    {
                        AddToBuilder(a_Bld, l_MaxBldLen, a_BEnum.f_Rules, l_BIdx, l_BLen, ref l_NeedSep, ref l_HadOverflow);
                        l_BIdx = Util.c_NoIndex;
                    }

                    if (true == l_HadOverflow)
                    {
                        a_HadOverflow = true;

                        return;
                    }
                }
            }

            private static void AddToBuilder(StringBuilder a_Bld,
                                             uint a_MaxBldLen,
                                             string a_Str,
                                             uint a_Idx,
                                             uint a_Len,
                                             ref bool a_NeedSep,
                                             ref bool a_HadOverflow)
            {
                if (0 == a_Len)
                {
                    return;
                }

                uint l_BldLenAfter = ((uint)a_Bld.Length + Util.BoolToByte(a_NeedSep) + a_Len);
                uint l_EffLen;

                if (l_BldLenAfter <= a_MaxBldLen)
                {
                    l_EffLen = a_Len;
                }
                else
                {
                    a_HadOverflow = true;
                    uint l_Surplus = (l_BldLenAfter - a_MaxBldLen);

                    if (l_Surplus >= a_Len)
                    {
                        return;
                    }

                    uint l_MaxAllowedLen = (a_Len - l_Surplus);

                    int l_LastSepIdx = a_Str.LastIndexOf(c_RulePatternSeparatorChar, (int)(a_Idx + l_MaxAllowedLen - 1), (int)(l_MaxAllowedLen - 1));

                    if (0 > l_LastSepIdx)
                    {
                        return;
                    }

                    l_EffLen = ((uint)l_LastSepIdx - a_Idx);
                }

                if (false == a_NeedSep)
                {
                    a_NeedSep = true;
                }
                else
                {
                    a_Bld.Append(c_RulePatternSeparatorChar);
                }

                a_Bld.Append(a_Str, (int)a_Idx, (int)l_EffLen);
            }

            public bool Equals(ref FilterItem a_Other,
                               bool a_CheckFunctionalEquivalenceOnly)
            {
                return (this.f_Enabled == a_Other.f_Enabled &&
                        this.f_Scopes == a_Other.f_Scopes &&
                        (true == a_CheckFunctionalEquivalenceOnly ||
                         true == Util.StringsEqual(this.f_Title, a_Other.f_Title)) &&
                        true == Util.StringsEqual(this.f_Rules, a_Other.f_Rules));
            }

            public static bool FromBuffer(BufferIO a_Buffer,
                                          out FilterItem a_Dest)
            {
                bool l_Enabled = a_Buffer.ReadBool();
                string l_Title = a_Buffer.ReadStringFlexibelEncLen();
                KademliaFilterScopes l_Scopes = (KademliaFilterScopes)a_Buffer.ReadByte();

                l_Title = Util.TruncateStr(Util.UnicodeNormalizeString(l_Title), false, c_MaxTitleChars);

                if (false == KademliaFilter.IsValidTitle(l_Title) ||
                    false == KademliaFilter.AreValidScopes(l_Scopes))
                {
                    a_Dest = new FilterItem();

                    return false;
                }

                string l_Rules = a_Buffer.ReadStringFlexibelEncLen(s_MaxRulesByteSize);
                l_Rules = NormalizeAndLowercaseString(l_Rules);
                bool l_WasValid = ValidateRules(l_Rules, true, true, out l_Rules);

                a_Dest = new FilterItem(l_Enabled, l_Title, l_Scopes, l_Rules);

                return true;
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteBool(f_Enabled);
                a_Buffer.WriteStringFlexibelEncLen(f_Title);
                a_Buffer.WriteByte((byte)f_Scopes);
                a_Buffer.WriteStringFlexibelEncLen(f_Rules);
            }

            public uint CalcMaximumWrittenByteSize()
            {
                // Returns the worst-case capacity ToBuffer() will
                // occupy.

                uint l_ByteSize = (sizeof(bool) + sizeof(byte) + (2 * sizeof(uint)));
                l_ByteSize += (Util.GetStringLength(f_Title) * sizeof(char));
                l_ByteSize += (Util.GetStringLength(f_Rules) * sizeof(char));

                return l_ByteSize;
            }

            public KademliaFilterScopes GetEffectiveScopes(KademliaFilterScopes a_DefaultScopes)
            {
                KademliaFilterScopes l_RuleScopes;

                if (false == IsDefaultScopesFlagSet(a_DefaultScopes))
                {
                    l_RuleScopes = f_Scopes;

                    if (true == IsDefaultScopesFlagSet(l_RuleScopes))
                    {
                        l_RuleScopes = a_DefaultScopes;
                    }
                }
                else
                {
                    l_RuleScopes = (a_DefaultScopes & KademliaFilterScopes.AllSpecificScopes);
                }

                return l_RuleScopes;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct RulesEnumerator
        {
            // Don't use the struct default constructor.

            public readonly string f_Rules;

            private uint f_SeqIdx;
            private uint f_SeqFinalIdx;
            private uint f_PatternIdx;

            public uint p_CurrentSequenceIndex
            {
                get
                {
                    return f_SeqIdx;
                }
            }

            public uint p_CurrentSequenceLength
            {
                get
                {
                    return (f_SeqFinalIdx - f_SeqIdx);
                }
            }

            public bool p_IsEndOfSequencesReached
            {
                get { return (Util.c_NoIndex == f_SeqIdx); }
            }

            public RulesEnumerator(string a_Rules)
            {
                f_Rules = a_Rules;

                f_SeqIdx = 0;
                f_SeqFinalIdx = TryGetSequenceFinalIndex(f_Rules, 0);
                f_PatternIdx = Util.c_NoIndex;
            }

            public void ResetCurrentSequence()
            {
                // Only resets the pattern enumeration in the current
                // sequence.

                f_PatternIdx = Util.c_NoIndex;
            }

            public uint GetCurrentSequenceTailLength(uint a_StartingOffset)
            {
                return (f_SeqFinalIdx - a_StartingOffset);
            }

            public bool AdvanceToNextSequence()
            {
                // Do not call this routine again once it returned
                // zero.

                ResetCurrentSequence();

                uint l_NextSeqIdx = (f_SeqFinalIdx + 1);
                uint l_NextSeqFinalIdx = TryGetSequenceFinalIndex(f_Rules, l_NextSeqIdx);

                if (Util.c_NoIndex == l_NextSeqFinalIdx)
                {
                    f_SeqFinalIdx = f_SeqIdx = Util.c_NoIndex;

                    return false;
                }

                f_SeqIdx = l_NextSeqIdx;
                f_SeqFinalIdx = l_NextSeqFinalIdx;

                return true;
            }

            public void AppendFullSequence(StringBuilder a_Bld)
            {
                a_Bld.Append(f_Rules, (int)p_CurrentSequenceIndex, (int)p_CurrentSequenceLength);
            }

            public bool TryGetNextPattern(out uint a_Idx,
                                          out uint a_Len)
            {
                uint l_ScanPos;

                if (Util.c_NoIndex == f_PatternIdx)
                {
                    l_ScanPos = f_SeqIdx;
                }
                else
                {
                    l_ScanPos = f_PatternIdx;
                }

                uint l_ScanLen = (f_SeqFinalIdx - l_ScanPos);

                if (0 == l_ScanLen)
                {
                    a_Idx = Util.c_NoIndex;
                    a_Len = 0;

                    return false;
                }

                int l_NextIdx = f_Rules.IndexOf(c_RulePatternSeparatorChar, (int)l_ScanPos, (int)l_ScanLen);
                a_Idx = l_ScanPos;

                if (0 > l_NextIdx)
                {
                    f_PatternIdx = f_SeqFinalIdx;
                    a_Len = l_ScanLen;
                }
                else
                {
                    f_PatternIdx = ((uint)l_NextIdx + 1);
                    a_Len = ((uint)l_NextIdx - l_ScanPos);
                }

                return true;
            }

            private static uint TryGetSequenceFinalIndex(string a_Rules,
                                                         uint a_StartingIdx)
            {
                // a_StartingPos may be outside the index range of
                // a_Rules, in which case Util.c_NoIndex is returned.

                uint l_Len = Util.GetStringLength(a_Rules);

                if (l_Len <= a_StartingIdx)
                {
                    return (l_Len == a_StartingIdx ? l_Len : Util.c_NoIndex);
                }

                int l_NextIdx = a_Rules.IndexOf(c_RuleSequenceSeparatorChar, (int)a_StartingIdx);

                return (0 > l_NextIdx ? l_Len : (uint)l_NextIdx);
            }
        }
    }

    [Flags]
    enum KademliaFilterScopes : byte
    {
        None = 0,

        ScopeLocalKademliaIndex = 0x1,
        ScopeOwnPublishingToOthers = 0x2,
        ScopeSearchResultsForOthers = 0x4,
        ScopeSearchResultsForSelf = 0x8,

        AllSpecificScopes = 0xF,

        UseDefault = 0x80,
    }
}

namespace NB.Import
{
    class LoadMetFilesContext : Gui.AbortableControllerContext
    {
        private const string c_iMuleAppDataSubFolder = "iMule";
        private const string c_iMuleConfigFile = "imule.conf";
        private const string c_iMuleConfigFileTempDirKey = "TempDir";

        private const string c_FileScanPattern = "*.part.met";
        private const string c_WorkerName = "LoadMetWorker";

        private readonly string f_SuggestedPath;

        private ErrorCodes f_ErrCode;
        private string f_PartFilesDir;
        private ImportableFile[] f_Files;
        private ulong f_TotalSize;

        public LoadMetFilesContext(string a_SuggestedPath)
        {
            f_SuggestedPath = a_SuggestedPath;
        }

        public ErrorCodes GetErrorCode()
        {
            return f_ErrCode;
        }

        public ImportableFile[] GetFiles(out string a_PartFilesDir,
                                         out ulong a_TotalSize)
        {
            a_PartFilesDir = f_PartFilesDir;
            a_TotalSize = f_TotalSize;

            return f_Files;
        }

        public void StartAsync()
        {
            Util.StartThread(c_WorkerName, Worker);
        }

        private void Worker()
        {
            // Try f_SuggestedPath directly.

            f_ErrCode = LoadFilesInTempDir(f_SuggestedPath);

            if (ErrorCodes.Success != f_ErrCode &&
                ErrorCodes.AbortedByUser != f_ErrCode)
            {
                // f_SuggestedPath isn't the real temp dir it seems,
                // so find the temp dir first.

                string l_TempDir;
                f_ErrCode = FindTempDirFromBaseDir(out l_TempDir);

                if (ErrorCodes.Success == f_ErrCode)
                {
                    f_ErrCode = LoadFilesInTempDir(l_TempDir);
                }
            }

            base.SetFinishedBase();
        }

        private ErrorCodes LoadFilesInTempDir(string a_TempDir)
        {
            // This routine can be safely called repeatedly so long
            // as ErrorCodes.Success was not returned.

            string[] l_FileNames;

            try
            {
                l_FileNames = Directory.GetFiles(a_TempDir, c_FileScanPattern, SearchOption.TopDirectoryOnly);
            }
            catch (ArgumentNullException l_Exc)
            {
                throw (l_Exc);
            }
            catch
            {
                return ErrorCodes.FailedToScanTempPath;
            }

            int l_RawCount = l_FileNames.Length;

            if (0 >= l_RawCount)
            {
                return ErrorCodes.TempDirContainsNoUseableFiles;
            }

            // Load all files found.

            List<ImportableFile> l_Ls = new List<ImportableFile>(l_RawCount);
            ulong l_SumOfSize = 0;

            for (int i = 0; i < l_RawCount; i++)
            {
                if (true == base.p_Aborted)
                {
                    return ErrorCodes.AbortedByUser;
                }

                ImportableFile l_File;
                ImportableFile.LoadResults l_Res = ImportableFile.Load(l_FileNames[i], out l_File);

                if (ImportableFile.LoadResults.Success == l_Res)
                {
                    l_Ls.Add(l_File);
                    l_SumOfSize += l_File.p_SizeForImportDlg;
                }

                base.SetPercentage((uint)i, (uint)l_RawCount);
            }

            if (0 >= l_Ls.Count)
            {
                return ErrorCodes.TempDirContainsNoUseableFiles;
            }

            f_PartFilesDir = a_TempDir;
            f_Files = l_Ls.ToArray();
            f_TotalSize = l_SumOfSize;

            return ErrorCodes.Success;
        }

        private ErrorCodes FindTempDirFromBaseDir(out string a_TempDir)
        {
            string l_CnfgDir;
            ErrorCodes l_ErrCode = FindiMuleConfigDir(f_SuggestedPath, out l_CnfgDir);

            if (ErrorCodes.Success == l_ErrCode)
            {
                return FindiMuleTempDir(l_CnfgDir, out a_TempDir);
            }

            a_TempDir = null;

            return l_ErrCode;
        }

        public static ErrorCodes FindiMuleConfigDir(string a_BaseDir,
                                                    out string a_iMuleCnfgDir)
        {
            // NOTE: Before iMule 2.3 the config dir was the dir with
            // the main executable, from 2.3 on it's the subfolder
            // 'iMule' in the OS's non-roaming application data folder.

            if (true == File.Exists(LongPath.Combine(a_BaseDir, c_iMuleConfigFile)))
            {
                // The passed dir seems to fit already.

                a_iMuleCnfgDir = a_BaseDir;

                return ErrorCodes.Success;
            }

            string l_Candidate = LongPath.Combine(Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData), c_iMuleAppDataSubFolder);

            if (true == Directory.Exists(l_Candidate))
            {
                // The application data subfolder exists.

                a_iMuleCnfgDir = l_Candidate;

                return ErrorCodes.Success;
            }

            a_iMuleCnfgDir = null;

            return ErrorCodes.ConfigDirNotFound;
        }

        private static ErrorCodes FindiMuleTempDir(string a_iMuleCnfgDir,
                                                   out string a_iMuleTempDir)
        {
            a_iMuleTempDir = null;

            string l_TempDir;
            ErrorCodes l_Code = GetTempDirValueFromiMuleConfigFile(a_iMuleCnfgDir, out l_TempDir);

            if (ErrorCodes.Success != l_Code)
            {
                return l_Code;
            }

            l_TempDir = PathNormalizer.NormalizeStatic(l_TempDir, true);

            if (0 == l_TempDir.Length)
            {
                return ErrorCodes.TempPathUnusedable;
            }

            a_iMuleTempDir = l_TempDir;

            return ErrorCodes.Success;
        }

        private static ErrorCodes GetTempDirValueFromiMuleConfigFile(string a_BaseDir,
                                                                     out string a_ConfigFileValue)
        {
            a_ConfigFileValue = null;

            string l_FileName = LongPath.Combine(a_BaseDir, c_iMuleConfigFile);

            if (false == File.Exists(l_FileName))
            {
                return ErrorCodes.ConfigFileNotFound;
            }

            ulong l_Size;
            StreamReader l_Rdr = Util.OpenTextFile(l_FileName, out l_Size);

            if (null == l_Rdr)
            {
                return ErrorCodes.ConfigFileAccessError;
            }

            ErrorCodes l_ErrCode;

            if ((2 * 1024 * 1024) < l_Size)
            {
                l_ErrCode = ErrorCodes.ConfigFileTooLarge;
            }
            else
            {
                while (true)
                {
                    string l_Line;

                    try
                    {
                        l_Line = l_Rdr.ReadLine();
                    }
                    catch
                    {
                        l_ErrCode = ErrorCodes.ConfigFileAccessError;

                        break;
                    }

                    if (null == l_Line)
                    {
                        // End of file reached.

                        l_ErrCode = ErrorCodes.ConfigFileKeyValuePairMissing;

                        break;
                    }

                    if (0 > Util.IndexOfIgnoreCase(l_Line, c_iMuleConfigFileTempDirKey))
                    {
                        continue;
                    }

                    int l_SepIdx = l_Line.IndexOf('=');

                    if (0 > l_SepIdx ||
                        false == Util.StringsEqualIgnoreCase(c_iMuleConfigFileTempDirKey, l_Line.Substring(0, l_SepIdx).Trim()))
                    {
                        continue;
                    }

                    a_ConfigFileValue = l_Line.Substring(l_SepIdx + 1);
                    l_ErrCode = ErrorCodes.Success;

                    break;
                }
            }

            l_Rdr.Close();

            return l_ErrCode;
        }

        public enum ErrorCodes : byte
        {
            NotFinished = 0,
            AbortedByUser = 1,

            ConfigDirNotFound = 2,
            ConfigFileNotFound = 3,
            ConfigFileAccessError = 4,
            ConfigFileTooLarge = 5,
            ConfigFileKeyValuePairMissing = 6,
            TempPathUnusedable = 7,

            TempDirContainsNoUseableFiles = 8,
            FailedToScanTempPath = 9,

            Success = 10,
        }
    }

    class ImportableFile
    {
        private const uint c_MaxMetFileSize = (2 * 1024 * 1024);
        private const uint c_MaxPartHashesSize = (1024 * 1024);
        private const uint c_eD2kOldPartStyleMagic = 0x01020000;

        // f_FileName is guaranteed to be legal.
        public readonly string f_FileName;
        public readonly ulong f_FileSize;
        // f_FileID will always be valid.
        public readonly FileID f_FileID;
        // f_PartHashes can be null.
        public readonly byte[] f_PartHashes;
        public readonly DownloadControl f_DlControl;
        public readonly Priority f_DlPriority;
        public readonly Priority f_UlPriority;

        public readonly string f_ContentFileName;
        public readonly ulong f_Done;
        public readonly float f_Progress;
        public readonly uint f_Fragments;
        private FlexArray<FileRange> f_GapRanges;

        public bool f_ImportFlag;

        public ulong p_SizeForImportDlg
        {
            get { return f_Done; }
        }

        public bool p_HasPartHashes
        {
            get { return (null != f_PartHashes); }
        }

        private ImportableFile(string a_FileName,
                               ulong a_FileSize,
                               FileID a_FileID,
                               byte[] a_PartHashes,
                               DownloadControl a_DlControl,
                               Priority a_DlPriority,
                               Priority a_UlPriority,
                               string a_ContentFileName,
                               ulong a_Done,
                               uint a_Fragments,
                               ref FlexArray<FileRange> a_GapRanges)
        {
            f_FileName = a_FileName;
            f_FileSize = a_FileSize;
            f_FileID = a_FileID;
            f_PartHashes = a_PartHashes;
            f_DlControl = a_DlControl;
            f_DlPriority = a_DlPriority;
            f_UlPriority = a_UlPriority;

            f_ContentFileName = a_ContentFileName;
            f_Done = a_Done;
            f_Progress = ((float)f_Done / f_FileSize);
            f_Fragments = a_Fragments;
            f_GapRanges = a_GapRanges;
        }

        public bool HasHashSetOrContent()
        {
            return (true == p_HasPartHashes ||
                    0 != f_Done);
        }

        public ulong GetExpectedMinFileSize()
        {
            int l_Count = f_GapRanges.p_Count;

            if (0 < l_Count &&
                f_FileSize == f_GapRanges.p_Arr[l_Count - 1].p_FinalOffset)
            {
                // There's a gap at the very end.

                return f_GapRanges.p_Arr[l_Count - 1].f_StartingOffset;
            }

            return f_FileSize;
        }

        public void ExposeGapRanges(out FlexArray<FileRange> a_GapRanges)
        {
            a_GapRanges = f_GapRanges;
        }

        public static LoadResults Load(string a_FileName,
                                       out ImportableFile a_ImportFile)
        {
            a_ImportFile = null;

            Exception l_Exc;
            FHandle l_SrcFile = FHandle.OpenR(a_FileName, true, FileOptions.SequentialScan, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

            if (null == l_SrcFile)
            {
                return LoadResults.MetFileAccessFailed;
            }

            try
            {
                FileIO l_File = FileIO.CreateReadable(l_SrcFile);

                if (c_MaxMetFileSize < l_File.p_Length)
                {
                    return LoadResults.MetFileTooLarge;
                }

                return LoadCore(l_File, out a_ImportFile);
            }
            catch (BufferIO.InterpretBufferException)
            {
                return LoadResults.MetFileCorrupt;
            }
            catch (BufferIO.FileAccessException)
            {
                return LoadResults.MetFileAccessFailed;
            }
            finally
            {
                l_SrcFile.Close();
            }
        }

        private static LoadResults LoadCore(FileIO a_File,
                                            out ImportableFile a_ImportFile)
        {
            a_ImportFile = null;

            FileVersions l_Ver = (FileVersions)a_File.ReadByte();

            if (FileVersions.Default > l_Ver ||
                FileVersions.LargeFile < l_Ver)
            {
                return LoadResults.StrangeVersion;
            }

            PartFileFormat l_Format = (FileVersions.Splitted == l_Ver ? PartFileFormat.Splitted : PartFileFormat.DefaultOld);

            if (FileVersions.Default == l_Ver)
            {
                a_File.SetPosition(24);
                uint l_Magic = a_File.ReadUInt32LE();

                if (c_eD2kOldPartStyleMagic == l_Magic)
                {
                    l_Format = PartFileFormat.eD2kOld;
                }

                a_File.SetPosition(1);
            }

            FileID l_FileID;
            byte[] l_PartHashes;

            if (PartFileFormat.DefaultOld == l_Format)
            {
                SkipLastDateChanged(a_File);

                if (false == LoadHashSet(a_File, out l_FileID, out l_PartHashes))
                {
                    return LoadResults.ErrLoadingHashSet;
                }
            }
            else
            {
                uint l_Num = a_File.ReadUInt32LE();

                if (0 == l_Num)
                {
                    if (false == LoadHashSet(a_File, out l_FileID, out l_PartHashes))
                    {
                        return LoadResults.ErrLoadingHashSet;
                    }
                }
                else
                {
                    a_File.SetPosition(2);
                    SkipLastDateChanged(a_File);
                    l_FileID = ReadFileID(a_File);
                    l_PartHashes = null;
                }
            }

            uint l_Count = a_File.ReadByte();
            FlexArray<FileRange> l_GapRanges;
            // We may read as from trusted, since we are inside a
            // try-catch block.
            const bool l_ExpectNormalStrsAsUnicode = true;
            TagList l_TagList = TagList.ReadListTrustedWithFileRanges(a_File, l_Count, l_ExpectNormalStrsAsUnicode, out l_GapRanges);

            if (null == l_TagList)
            {
                return LoadResults.ErrLoadingTagList;
            }

            ulong l_FileSize;

            if (false == l_TagList.TryGetTagAsULong(TagCodes.TAG_FILESIZE, out l_FileSize))
            {
                return LoadResults.MissingFileSize;
            }

            if (false == ContentManager.IsValidFileSize(l_FileSize))
            {
                return LoadResults.FileSizeInvalid;
            }

            if (null != l_PartHashes &&
                (eD2kHashSet.GeteD2kHashSetElementCount(l_FileSize) * MD_4.c_HashByteSize) != l_PartHashes.Length)
            {
                return LoadResults.FileSizeAndPartHashesDontMatch;
            }

            string l_PartFileName;

            if (false == l_TagList.TryGetTagAsString(TagCodes.TAG_PARTFILENAME, out l_PartFileName))
            {
                return LoadResults.MissingPartFileName;
            }

            ulong l_Done;
            uint l_Fragments;

            if (false == VerifyAndRepairGapRanges(l_FileSize, ref l_GapRanges, out l_Done, out l_Fragments))
            {
                return LoadResults.CorruptGapRanges;
            }

            string l_FileName;

            if (false == l_TagList.TryGetTagAsString(TagCodes.TAG_FILENAME, out l_FileName))
            {
                l_FileName = FileNames.c_FallbackFileName;
            }
            else
            {
                l_FileName = FileNames.MakeFileNameLegalFailsafe(l_FileName);
            }

            DownloadControl f_DlControl = DownloadControl.DlDisabled;

            ulong l_StatusNum;
            if (true == l_TagList.TryGetTagAsULong(TagCodes.TAG_STATUS, out l_StatusNum) &&
                // Other than zero is paused or stopped.
                0 == l_StatusNum)
            {
                f_DlControl = DownloadControl.DlEnabled;
            }

            Priority l_DlPriority = ReadPriority(l_TagList, TagCodes.TAG_DLPRIORITY);
            Priority l_UlPriority = ReadPriority(l_TagList, TagCodes.TAG_ULPRIORITY);
            a_ImportFile = new ImportableFile(l_FileName, l_FileSize, l_FileID, l_PartHashes, f_DlControl, l_DlPriority, l_UlPriority, l_PartFileName, l_Done, l_Fragments, ref l_GapRanges);

            return LoadResults.Success;
        }

        private static bool VerifyAndRepairGapRanges(ulong a_FileSize,
                                                     ref FlexArray<FileRange> a_GapRanges,
                                                     out ulong a_Done,
                                                     out uint a_Fragments)
        {
            // This routine checks that the constraints of
            // PieceBounds.CreateFromGapRanges() are honored and
            // repairs if possible. Input is expected to be sorted.

            a_Done = a_FileSize;
            a_Fragments = 0;

            int l_Count = a_GapRanges.p_Count;

            if (0 == l_Count)
            {
                return true;
            }

            ulong l_Pos = 0;
            FileRange[] l_Arr = a_GapRanges.p_Arr;
            for (int i = 0; i < l_Count; i++)
            {
                ulong l_Final = l_Arr[i].p_FinalOffset;

                if (l_Arr[i].f_StartingOffset < l_Pos ||
                    l_Final > a_FileSize)
                {
                    return false;
                }

                bool l_RemoveCurrent;

                if (0 == l_Arr[i].f_Length)
                {
                    l_RemoveCurrent = true;
                }
                else
                {
                    if (0 == i ||
                        l_Arr[i].f_StartingOffset > l_Pos)
                    {
                        l_RemoveCurrent = false;
                    }
                    else
                    {
                        // Seamless, so redim the previous element
                        // and remove the current element.

                        bool l_Succ = FileRange.CreateFromUnsafeBounds(l_Arr[i - 1].f_StartingOffset, l_Final, out l_Arr[i - 1]);

                        l_RemoveCurrent = true;
                    }

                    a_Done -= l_Arr[i].f_Length;
                    l_Pos = l_Final;
                }

                if (true == l_RemoveCurrent)
                {
                    a_GapRanges.DeleteAt(i);
                    l_Arr = a_GapRanges.p_Arr;
                    i--;
                    l_Count--;

                    continue;
                }
            }

            if (0 < l_Count)
            {
                // Calculate the number of fragments.

                a_Fragments = (uint)(l_Count - 1);

                if (0 != l_Arr[0].f_StartingOffset)
                {
                    // Initial fragment.

                    a_Fragments++;
                }

                if (l_Arr[l_Count - 1].p_FinalOffset != a_FileSize)
                {
                    // Ending fragment.

                    a_Fragments++;
                }
            }

            return true;
        }

        private static FileID ReadFileID(FileIO a_File)
        {
            return new FileID(a_File.ReadUInt128BE());
        }

        private static bool LoadHashSet(FileIO a_File,
                                        out FileID a_FileID,
                                        out byte[] a_PartHashes)
        {
            // NOTE: We don't know the file size yet when doing this.

            a_FileID = ReadFileID(a_File);
            // eMule seems to use an ushort here.
            ulong l_PartCount = a_File.ReadUInt64LE();

            if (0 == l_PartCount)
            {
                // No part hashes present.

                a_PartHashes = null;

                return true;
            }

            // Check for 32 bit limit to protect the below code.
            if (Math.Max(eD2kHashSet.c_PartLimitFromMaxFileSize, uint.MaxValue / MD_4.c_HashByteSize) < l_PartCount)
            {
                // This file is larger than our maximum.

                a_PartHashes = null;

                return false;
            }

            uint l_PartHashesSize = ((uint)l_PartCount * MD_4.c_HashByteSize);

            if (c_MaxPartHashesSize < l_PartHashesSize)
            {
                // Don't read all of this, because we don't want to
                // buffer this much.

                a_File.JumpAhead(l_PartHashesSize);
                a_PartHashes = null;

                return true;
            }

            a_PartHashes = a_File.ReadBytes(l_PartHashesSize);

            MD_4 l_MD4 = new MD_4();
            l_MD4.Update(a_PartHashes);
            l_MD4.Finish();

            if (false == l_MD4.Verify(a_FileID.f_Val))
            {
                a_PartHashes = null;
            }

            return true;
        }

        private static void SkipLastDateChanged(FileIO a_File)
        {
            a_File.JumpAhead(sizeof(uint));
        }

        private static Priority ReadPriority(TagList a_TagList,
                                             TagCodes a_Code)
        {
            // If no priority is present, the default will be used.

            ulong l_PriorityNum;
            if (false == a_TagList.TryGetTagAsULong(a_Code, out l_PriorityNum) ||
                (byte)Priorities.PowerShare < l_PriorityNum)
            {
                return Priority.s_Default;
            }

            switch ((Priorities)l_PriorityNum)
            {
                case Priorities.Auto:
                    return Priority.s_Default;

                case Priorities.High:
                    return Priority.s_High;

                case Priorities.Normal:
                    return Priority.s_Normal;

                case Priorities.Low:
                    return Priority.s_Low;

                case Priorities.VeryLow:
                    return Priority.s_VeryLow;

                case Priorities.VeryHigh:
                case Priorities.PowerShare:
                    return Priority.s_VeryHigh;

                default: throw (new ApplicationException());
            }
        }

        public enum LoadResults : byte
        {
            Success = 0,

            // General errors.
            MetFileAccessFailed = 1,
            MetFileTooLarge = 2,
            MetFileCorrupt = 3,

            // Specific errors.
            StrangeVersion = 4,
            ErrLoadingHashSet = 5,
            ErrLoadingTagList = 6,
            MissingFileSize = 7,
            FileSizeInvalid = 8,
            FileSizeAndPartHashesDontMatch = 9,
            MissingPartFileName = 10,
            CorruptGapRanges = 11,
        }

        private enum FileVersions : byte
        {
            Default = 0xE0,
            Splitted = 0xE1,
            LargeFile = 0xE2,
        }

        private enum PartFileFormat : byte
        {
            DefaultOld = 1,
            Splitted = 2,
            eD2kOld = 3,
        }

        private enum Priorities : byte
        {
            Low = 0,
            Normal = 1,
            High = 2,
            VeryHigh = 3,
            VeryLow = 4,
            Auto = 5,
            PowerShare = 6,
        }
    }

    class SharedDirsImporter : Gui.AbortableControllerContext
    {
        private const string c_SharedDirsImportWorkerName = "ShrdDirsImportWorker";
        private const string c_SharedDirsSrcFileName = "shareddir.dat";
        private const uint c_MaxSrcFileSize = (2 * 1024 * 1024);

        private readonly SharedDirectories.Dir[] f_InputDirs;
        private readonly string f_SourceFileName;

        private ResultCode f_ResCode;
        private SharedDirectories.Dir[] f_OutputDirs;

        public SharedDirsImporter(SharedDirectories.Dir[] a_InputDirs,
                                  string a_SourceFileName)
        {
            f_InputDirs = a_InputDirs;
            f_SourceFileName = a_SourceFileName;
        }

        private void SetProgress(uint a_Percentage)
        {
            base.SetPercentage(a_Percentage);
        }

        public ResultCode GetResult(out SharedDirectories.Dir[] a_OutputDirs)
        {
            a_OutputDirs = f_OutputDirs;

            return f_ResCode;
        }

        public void StartAsync()
        {
            Util.StartThread(c_SharedDirsImportWorkerName, Worker);
        }

        private void Worker()
        {
            f_ResCode = RunCore();

            base.SetFinishedBase();
        }

        private ResultCode RunCore()
        {
            // The problem here is that we import a list of shared
            // dirs from a client that has no explicit recursion
            // support, so we need to figure out the recursion state by
            // comparing the list of shared dirs with the file system.

            FlexArray<Dir> l_Dirs;
            ResultCode l_ResCode = ReadSrcFile(out l_Dirs);

            if (ResultCode.Success != l_ResCode)
            {
                return l_ResCode;
            }

            SetProgress(20);

            SortAndRemoveDuplicates(ref l_Dirs);

            SetProgress(30);

            l_ResCode = DetermineRecursionStates(ref l_Dirs);

            if (ResultCode.Success != l_ResCode)
            {
                return l_ResCode;
            }

            SetProgress(80);

            BuildSharedDirectories(ref l_Dirs, out f_OutputDirs);

            return ResultCode.Success;
        }

        private ResultCode ReadSrcFile(out FlexArray<Dir> a_Dirs)
        {
            // On success, a_Dirs can contain redundant entries.

            FlexArray<Dir>.Create(out a_Dirs);

            ulong l_Size;
            StreamReader l_Rdr = Util.OpenTextFile(f_SourceFileName, out l_Size);

            if (null == l_Rdr)
            {
                return ResultCode.IOError;
            }

            ResultCode l_ResCode;

            if (c_MaxSrcFileSize < l_Size)
            {
                l_ResCode = ResultCode.SrcFileTooLarge;
            }
            else
            {
                l_ResCode = ReadSrcFileCore(l_Rdr, ref a_Dirs);
            }

            l_Rdr.Close();

            return l_ResCode;
        }

        private ResultCode ReadSrcFileCore(StreamReader a_Rdr,
                                           ref FlexArray<Dir> a_Dirs)
        {
            PathNormalizer l_PathNormalizer;
            PathNormalizer.Create(out l_PathNormalizer);

            while (true)
            {
                if (true == base.p_Aborted)
                {
                    return ResultCode.AbortedByUser;
                }

                string l_Line;

                try
                {
                    l_Line = a_Rdr.ReadLine();
                }
                catch
                {
                    return ResultCode.IOError;
                }

                if (null == l_Line)
                {
                    return (0 >= a_Dirs.p_Count ? ResultCode.EmptyInputFile : ResultCode.Success);
                }

                string l_Root = LongPath.GetPathRoot(l_Line);

                if (null == l_Root)
                {
                    continue;
                }

                string l_Path = l_PathNormalizer.Normalize(l_Line.Trim(), true);

                if (0 == l_Path.Length)
                {
                    continue;
                }

                int l_Idx = a_Dirs.ReserveAtEnd();
                a_Dirs.p_Arr[l_Idx] = new Dir(l_Path);
            }
        }

        private void SortAndRemoveDuplicates(ref FlexArray<Dir> a_Dirs)
        {
            a_Dirs.Sort(CompareDirs);

            Dir[] l_Arr = a_Dirs.p_Arr;
            for (int i = (a_Dirs.p_Count - 2); 0 <= i; i--)
            {
                if (true == FileNames.AreNamesIdentical(l_Arr[i].f_Path, l_Arr[i + 1].f_Path))
                {
                    a_Dirs.DeleteAt(i + 1);
                    l_Arr = a_Dirs.p_Arr;
                }
            }
        }

        private ResultCode DetermineRecursionStates(ref FlexArray<Dir> a_Dirs)
        {
            // The input should be sorted from a call to
            // SortAndRemoveDuplicates().

            int l_HighestIdx = (a_Dirs.p_Count - 1);
            Dir[] l_Arr = a_Dirs.p_Arr;
            for (int i = l_HighestIdx; 0 <= i; i--)
            {
                if (true == base.p_Aborted)
                {
                    return ResultCode.AbortedByUser;
                }

                string[] l_SubDirs;

                try
                {
                    l_SubDirs = Directory.GetDirectories(l_Arr[i].f_Path);
                }
                catch
                {
                    // No access.

                    l_Arr[i].SetState(false, RecursionState.NoRecursion);

                    continue;
                }

                if (0 == l_SubDirs.Length)
                {
                    l_Arr[i].SetState(false, RecursionState.RecursionPossible);

                    continue;
                }

                bool l_FullRecursion = true;

                foreach (string l_SubDir in l_SubDirs)
                {
                    bool l_RecursionAllowed = false;

                    for (int k = l_HighestIdx; i < k; k--)
                    {
                        if (true == FileNames.AreNamesIdentical(l_SubDir, l_Arr[k].f_Path))
                        {
                            if (RecursionState.RecursionPossible == l_Arr[k].f_RecState)
                            {
                                l_RecursionAllowed = true;
                            }

                            break;
                        }
                    }

                    if (false == l_RecursionAllowed)
                    {
                        l_FullRecursion = false;

                        break;
                    }
                }

                l_Arr[i].SetState(true, false == l_FullRecursion ? RecursionState.NoRecursion : RecursionState.RecursionPossible);
            }

            return ResultCode.Success;
        }

        private void BuildSharedDirectories(ref FlexArray<Dir> a_ReadDirs,
                                            out SharedDirectories.Dir[] a_Dest)
        {
            SharedDirectories.Dir[] l_SecondInputDirs = Dir.FlexArrayToSharedDirArr(ref a_ReadDirs);

            int l_Capacity = (f_InputDirs.Length + l_SecondInputDirs.Length);
            FlexArray<SharedDirectories.Dir> l_Arr;
            FlexArray<SharedDirectories.Dir>.Create(l_Capacity, out l_Arr);

            l_Arr.AddArray(f_InputDirs);
            bool l_Unused;
            SharedDirectories.Dir.CopyWithEnforcedIntegrity(l_SecondInputDirs, ref l_Arr, out l_Unused);

            a_Dest = l_Arr.ToArray(true);
        }

        private static int CompareDirs(ref Dir a_A,
                                       ref Dir a_B)
        {
            return a_A.f_Path.Length.CompareTo(a_B.f_Path.Length);
        }

        public static string BuildFullConfigFileName(string a_Folder,
                                                     out bool a_Exists)
        {
            string l_FullFileName = LongPath.Combine(a_Folder, c_SharedDirsSrcFileName);
            a_Exists = File.Exists(l_FullFileName);

            return l_FullFileName;
        }

        public enum ResultCode : byte
        {
            NotFinished = 0,
            AbortedByUser = 1,

            SrcFileTooLarge = 2,
            IOError = 3,
            EmptyInputFile = 4,

            Success = 5,
        }

        private enum RecursionState : byte
        {
            Unknown = 0,
            NoRecursion = 1,
            RecursionPossible = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Dir
        {
            // Don't use the struct default constructor.

            public readonly string f_Path;
            public bool f_HasSubdirs;
            public RecursionState f_RecState;

            public Dir(string a_Path)
            {
                f_Path = a_Path;
                f_HasSubdirs = false;
                f_RecState = RecursionState.Unknown;
            }

            public void SetState(bool a_HasSubdirs,
                                 RecursionState a_RecState)
            {
                f_HasSubdirs = a_HasSubdirs;
                f_RecState = a_RecState;
            }

            public static SharedDirectories.Dir[] FlexArrayToSharedDirArr(ref FlexArray<Dir> a_Input)
            {
                int l_Count = a_Input.p_Count;
                SharedDirectories.Dir[] l_Dest = new SharedDirectories.Dir[l_Count];

                Dir[] l_Arr = a_Input.p_Arr;
                for (int i = 0; i < l_Count; i++)
                {
                    bool l_Recursion = (true == l_Arr[i].f_HasSubdirs &&
                                        RecursionState.RecursionPossible == l_Arr[i].f_RecState);
                    l_Dest[i] = new SharedDirectories.Dir(l_Arr[i].f_Path, l_Recursion, ShareKinds.eD2k_on_I2P);
                }

                return l_Dest;
            }
        }
    }
}

// End of file.
