/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\Network.cs
 * PURPOSE: Network related code (1)
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using NB.Content;
using NB.Gui;
using NB.I2P;
using NB.Language;
using NB.Network.eD2k;
using Support;
using Support.HashFunctions;

namespace NB.Network
{
    class NetworkManager
    {
        private const string c_NetworkThreadName = "Network";

        public const byte c_InvalidRank = 0;
        public const byte c_BestRank = 1;
        public const ushort c_MaxRank = ushort.MaxValue;

        private const uint c_UpdateDlQueueIntervalSecs = 5;

        private const byte c_MaxIncomingConnsPerTurn = 100;

        public const uint c_MaxBytesUploadPerConnPerTurn = ((64 * 1024) / UpDownManager.c_TurnsPerSecond);

        // This flag prevents sending away addresses that are longer
        // than 387 bytes.
        public const bool c_RestrictToClassicAddresses = true;

        public readonly Instance f_Inst;
        private APCQueue<GeneralAPCCookie> f_APCQueue;
        private readonly SamBridgeManager f_Bridge;
        private BridgeStatus f_BridgeStatus;
        // The following three fields won't change while a connection
        // lasts.
        private OwnUserPersona f_OwnEffPersona;
        // The following two fields are null while there is no
        // connection.
        private I2PAddress f_DatagramAddr;
        private I2PAddress f_StreamAddr;

        private SamConnectionCookie f_ConnCookie;
        private bool f_ShutdownInProgress;
        private ShutdownCookie f_ShutdownCookie;

        // The following three may only be used by the
        // network thread (from the DonkeyClient class).
        public readonly RSACryptoServiceProvider f_RSA;
        public readonly SHA_1 f_SHA1;
        public readonly Tiger f_Tiger;
        public readonly Keccak f_SHA3_512;

        private readonly Dictionary<PendingConnContext, PendingConnContext> f_PendingConns;
        private readonly Dictionary<DonkeyClient, DonkeyClient> f_EstablishedConns;
        private readonly Dictionary<KnownFile, KnownFile> f_ActiveFiles;

        private OwnUserPersona f_UpdatedOwnPersona;
        public NetworkPrefs f_NetworkPrefs;
        public PeerInfo f_SelfInfo;
        public byte[] f_HelloPacketBody;
        public byte[] f_HelloAnswerPacketBody;
        public byte[] f_MuleInfoPacketBody;
        private uint f_SuccessfulOutgoingConns;
        private uint f_AcceptedIngoingConns;
        private uint f_ActiveIngoingConns;

        public DownloadQueue f_DlQueue;
        private Time f_NextDlQueueUpdateTime;

        public UploadQueue f_UlQueue;

        private Time f_NextUlTime;
        private uint f_EstablishedConnProcessingGen;
        private uint f_NextDebugConnID;

        private ProcessCPUInfo f_CPUInfo;

        private Time f_NextKnownFileCacheUpdateTime;
        // f_KnownFileCacheCookie and f_KnownFileCache can be null.
        private GeneralAPCCookie f_KnownFileCacheCookie;
        public KnownFileCache f_KnownFileCache;

        public OwnUserPersona p_OwnPersona
        {
            get { return f_OwnEffPersona; }
        }

        public I2PAddress p_DatagramAddr
        {
            // May only be used while a connection is up.

            get
            {
                return f_DatagramAddr;
            }
        }

        public I2PAddress p_StreamAddr
        {
            // May only be used while a connection is up.

            get
            {
                return f_StreamAddr;
            }
        }

        public BridgeStatus p_BridgeStatus
        {
            // The bridge status as seen by the network thread.
            get { return f_BridgeStatus; }
        }

        public uint p_ConnGeneration
        {
            get { return f_Bridge.p_ConnGeneration; }
        }

        public uint p_MaxStreamInQueueSizePerConn
        {
            get { return (10 * 1024 * 1024); }
        }

        public NetworkManager(Instance a_Inst,
                              OwnUserPersona a_OwnPersona,
                              ref NetworkPrefs a_NetworkPrefs)
        {
            f_Inst = a_Inst;
            APCQueue<GeneralAPCCookie>.Create(out f_APCQueue);
            f_Bridge = new SamBridgeManager();
            f_BridgeStatus = BridgeStatus.Disconnected;

            f_RSA = new RSACryptoServiceProvider();
            f_SHA1 = new SHA_1();
            f_Tiger = new Tiger();
            f_SHA3_512 = new Keccak(Keccak.Variant.SHA3_512Bits);

            f_PendingConns = new Dictionary<PendingConnContext, PendingConnContext>();
            f_EstablishedConns = new Dictionary<DonkeyClient, DonkeyClient>();
            f_ActiveFiles = new Dictionary<KnownFile, KnownFile>();

            f_OwnEffPersona = a_OwnPersona;
            f_NetworkPrefs = a_NetworkPrefs;

            DownloadQueue.Create(f_Inst, out f_DlQueue);
            UploadQueue.Create(out f_UlQueue);

            f_NextDebugConnID = 1;
            ProcessCPUInfo.Create(out f_CPUInfo);

            f_Bridge.e_AddressBanned += NotifyBannedAddress;
        }

        public static bool ShouldRestrictPeerToClassicAddresses(I2PAddress a_PeerAddr)
        {
            return (true == c_RestrictToClassicAddresses &&
                    true == a_PeerAddr.p_IsClassicAddress);
        }

        private void NotifyNewSamConnection()
        {
            f_DatagramAddr = f_Bridge.p_DatagramKeyPair.f_PublicKey;
            f_StreamAddr = f_Bridge.p_StreamKeyPair.f_PublicKey;

            f_SelfInfo = new PeerInfo(f_NetworkPrefs.f_UserName, f_OwnEffPersona.f_UserID, f_DatagramAddr, f_NetworkPrefs.f_AllowDataCompression);
            f_HelloPacketBody = f_SelfInfo.ToBuffer(true);
            f_HelloAnswerPacketBody = f_SelfInfo.ToBuffer(false);
            f_MuleInfoPacketBody = BuildMuleInfoPacketBody();

            Time l_Time = f_Inst.GetTime();
            f_Inst.f_UpDwnMngr.NotifyNewSamConnection(l_Time);
            f_Inst.f_Kad.NotifyNewSamConnection(l_Time);

            f_Inst.f_MainForm.AddToLog(f_Inst.f_TextSet.BuildKeyValuePair(TextKeys.MsgYourClientID, f_Inst.f_Kad.f_Self.f_ClientID.f_Val.ToHexString()));
        }

        private void NotifySamConnectionBroken()
        {
            if (null != f_UpdatedOwnPersona)
            {
                f_OwnEffPersona = f_UpdatedOwnPersona;
                f_UpdatedOwnPersona = null;
            }

            f_HelloPacketBody = null;
            f_HelloAnswerPacketBody = null;
            f_MuleInfoPacketBody = null;
            f_SuccessfulOutgoingConns = 0;
            f_AcceptedIngoingConns = 0;

            f_NextUlTime = new Time();
            f_EstablishedConnProcessingGen = 0;

            f_Inst.f_UpDwnMngr.NotifySamConnectionBroken();
            f_Inst.f_Kad.NotifySamConnectionBroken();
        }

        private void KillAllConnections(bool a_DemandedByUser)
        {
            Software.UpdateShutDownState(0x4);

            if (BridgeStatus.Disconnected == f_BridgeStatus)
            {
                return;
            }

            bool l_WasFullyConnected = (BridgeStatus.Connected == f_BridgeStatus);

            if (false == a_DemandedByUser)
            {
                f_Inst.f_MainForm.AddToLog(false == l_WasFullyConnected ?
                                           TextKeys.MsgSAMBridgeConnAttemptFailed :
                                           TextKeys.MsgSAMBridgeConnLost);
            }
            else
            {
                f_Inst.f_MainForm.AddToLog(false == l_WasFullyConnected ?
                                           TextKeys.MsgSAMBridgeConnAbortedAsDemanded :
                                           TextKeys.MsgSAMBridgeConnClosedAsDemanded);
            }

            if (true == l_WasFullyConnected)
            {
                foreach (PendingConnContext l_PendConn in f_PendingConns.Values)
                {
                    l_PendConn.f_Cookie.Close();
                }

                foreach (DonkeyClient l_Clnt in f_EstablishedConns.Values)
                {
                    l_Clnt.Close();
                }
            }

            f_Bridge.Shutdown();
            f_BridgeStatus = BridgeStatus.Disconnected;
            f_ConnCookie.SetBridgeStatus(BridgeStatus.Disconnected);
            f_ConnCookie = null;

            // Call Process() right here to make sure it's done.
            // Since before we reconnect the next time it must be
            // called, otherwise remains of the old session
            // might interfere with the new one.
            Process(f_Inst.GetTime());

            if (true == l_WasFullyConnected)
            {
                NotifySamConnectionBroken();
            }

            Software.UpdateShutDownState(0x8);
        }

        public void AbortPendingConnectionAttempts(UserContext a_UserCntxt)
        {
            foreach (PendingConnContext l_PendConn in f_PendingConns.Values)
            {
                if (l_PendConn.f_UserCntxt == a_UserCntxt)
                {
                    l_PendConn.f_Cookie.Close();
                }
            }
        }

        public SamBridgeManager ExposeBridge()
        {
            return f_Bridge;
        }

        public bool IsAddressBanned(I2PAddress a_Addr)
        {
            return f_Bridge.IsAddressBanned(a_Addr);
        }

        public void Start()
        {
            // Supposed to be called by the GUI thread.

            Util.StartThread(c_NetworkThreadName, NetworkWorker);
        }

        public void GuiBuildConnection(GeneralAPCCookie a_Cookie)
        {
            SamConnectionCookie l_ConnCookie = (SamConnectionCookie)a_Cookie;

            if (BridgeStatus.Disconnected != f_BridgeStatus)
            {
                Util.DebugBreak();

                l_ConnCookie.SetBridgeStatus(BridgeStatus.Disconnected);
            }
            else
            {
                TextSet l_TestSet = f_Inst.f_TextSet;
                System.Net.IPEndPoint l_SamEndPoint = l_ConnCookie.f_SamEndPoint;
                string l_IPAddrStr = l_SamEndPoint.Address.ToString();
                bool l_IsLoopback = System.Net.IPAddress.Loopback.Equals(l_SamEndPoint.Address);

                if (true == l_IsLoopback)
                {
                    l_IPAddrStr = ("localhost (" + l_IPAddrStr + ")");
                }

                string l_SamEndPointStr = l_TestSet.Format(TextKeys.IPEndPointFormat, l_IPAddrStr, l_SamEndPoint.Port);

                f_Bridge.Start(l_ConnCookie);
                f_BridgeStatus = f_Bridge.p_Status;

                f_Inst.f_MainForm.AddToLog(l_TestSet.Format(TextKeys.MsgSAMBridgeConnecting, l_SamEndPointStr));

                if (BridgeStatus.Disconnected == f_BridgeStatus)
                {
                    // Failed fast.

                    l_ConnCookie.SetBridgeStatus(BridgeStatus.Disconnected);

                    f_Inst.f_MainForm.AddToLog(TextKeys.MsgSAMBridgeConnAttemptFailed);
                }
                else
                {
                    f_ConnCookie = l_ConnCookie;
                }
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiKillConnection(GeneralAPCCookie a_Cookie)
        {
            KillAllConnections(true);

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiBeginShutdown(GeneralAPCCookie a_Cookie)
        {
            Software.UpdateShutDownState(0x2);

            KillAllConnections(true);

            f_ShutdownInProgress = true;
            f_ShutdownCookie = (ShutdownCookie)a_Cookie;

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiSaveNetworkData(GeneralAPCCookie a_Cookie)
        {
            StorageStates l_UserCntxtsState = f_Inst.f_UserMngr.SaveContexts();
            StorageStates l_NodesState = f_Inst.f_Kad.SaveNodes();

            StorageStates l_State = StorageStates.Success;
            a_Cookie.SetReturnObjectAndState(l_State, CookieStates.Success);
        }

        public void GuiForcePublishing(GeneralAPCCookie a_Cookie)
        {
            f_Inst.f_Kad.f_PublishMngr.TriggerPublishing();

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiUpdatePreferences(GeneralAPCCookie a_Cookie)
        {
            Preferences l_Prefs = (Preferences)a_Cookie.GetArg();
            OwnUserPersona l_NewPersona = l_Prefs.f_GeneralPrefs.f_OwnPersona;

            if (BridgeStatus.Connected == f_BridgeStatus)
            {
                // We may not modify f_OwnEffPersona while a connection
                // is up.

                f_UpdatedOwnPersona = l_NewPersona;
            }
            else
            {
                f_OwnEffPersona = l_NewPersona;
            }

            f_NetworkPrefs = l_Prefs.f_NetworkPrefs;
            f_UlQueue.NotifyPreferencesUpdated(ref f_NetworkPrefs);
            f_Inst.f_Kad.UpdatePreferences(ref l_Prefs.f_KadPrefs);

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiEnableAccessToFileAndCallBack(GeneralAPCCookie a_Cookie)
        {
            object[] l_Args = a_Cookie.GetArgAsObjArray();
            KnownFile l_Known = (KnownFile)l_Args[0];
            OneArgHandler l_CallbackRoutine = (OneArgHandler)l_Args[1];

            if (false == l_Known.IsActive())
            {
                a_Cookie.SetState(CookieStates.Failure);

                return;
            }

            l_Known.p_ActCntxt.ManageGuiAccessToPeers(true);
            a_Cookie.SetState(CookieStates.Success);

            // Call the GUI.
            f_Inst.f_MainForm.f_Invoker.InvokeAsync(l_CallbackRoutine, l_Known);
        }

        public void GuiResetBans(GeneralAPCCookie a_Cookie)
        {
            f_Bridge.ResetBans();
            f_Inst.f_UserMngr.ResetBans();

            a_Cookie.SetState(CookieStates.Success);
        }

        public string BuildStats()
        {
            StatsBuilder l_Stats = new StatsBuilder(f_Inst);
            Time l_Time = f_Inst.GetTime();
            Kademlia l_Kad = f_Inst.f_Kad;
            float l_ProcessCPULoad;
            ulong l_ProcessCPUTime;
            bool l_GotCPULoad = f_CPUInfo.GetInfo(l_Time, out l_ProcessCPULoad, out l_ProcessCPUTime);
            ulong l_AvailBytes;
            bool l_GotAvailBytes = Software.GetAvailDiskSpace(true, out l_AvailBytes);

            l_Stats.AddCaption(TextKeys.StatsTabApp);
            l_Stats.AddTime(TextKeys.StatsTabAppUptime, l_Time.SecondsSince(Software.s_AppLaunchTime));
            l_Stats.AddSize(TextKeys.StatsTabAppMemAllocated, (ulong)GC.GetTotalMemory(false));
            l_Stats.AddTime(TextKeys.StatsTabAppCPUTime, (long)l_ProcessCPUTime);
            l_Stats.AddText(TextKeys.StatsTabAppCPUUsage, false == l_GotCPULoad ? f_Inst[TextKeys.UnknownValue] : f_Inst.f_TextSet.PercStr(l_ProcessCPULoad));

            if (true == l_GotAvailBytes)
            {
                l_Stats.AddTextDirect("Available Disk Space", f_Inst.f_MainForm.BuildDefaultSizeStr(l_AvailBytes));
            }

            l_Stats.AddLineBreak();

            l_Stats.AddCaption(TextKeys.StatsTabI2P);
            f_Inst.f_UpDwnMngr.RetrieveStats(l_Time, ref l_Stats);
            l_Stats.AddInteger(TextKeys.StatsTabI2PBannedDests, f_Bridge.p_BannedDestinations);

            l_Stats.AddLineBreak();

            l_Stats.AddCaption(TextKeys.StatsTabStreams);
            l_Stats.AddInteger(TextKeys.StatsTabStreamsSuccOutAttempts, f_SuccessfulOutgoingConns);
            l_Stats.AddInteger(TextKeys.StatsTabStreamsAcceptedInConns, f_AcceptedIngoingConns);
            l_Stats.AddInteger(TextKeys.StatsTabStreamsCurOutAttempts, f_PendingConns.Count);
            int l_ActiveConns = f_EstablishedConns.Count;
            string l_ConnCountSuffixStr = Gui.GeneralPeersPanel.BuildConnectionsCountString(f_Inst.f_TextSet, f_ActiveIngoingConns, (uint)l_ActiveConns - f_ActiveIngoingConns);

            if (null == l_ConnCountSuffixStr)
            {
                l_Stats.AddInteger(TextKeys.StatsTabStreamsCurEstablishedConns, l_ActiveConns);
            }
            else
            {
                l_Stats.AddText(TextKeys.StatsTabStreamsCurEstablishedConns, f_Inst.f_TextSet.Format(TextKeys.NameWithCountFormat, Util.NumberToInvString(l_ActiveConns), l_ConnCountSuffixStr));
            }

            {
                uint l_UploadingSlots;
                uint l_TotalSlots;
                uint l_QueueLength;
                f_Inst.f_NetworkMngr.f_UlQueue.GetMicroStats(out l_UploadingSlots, out l_TotalSlots, out l_QueueLength);
                l_Stats.AddTextDirect("Upload slots busy", f_Inst.f_TextSet.NumStr(l_UploadingSlots) + "/" + f_Inst.f_TextSet.NumStr(l_TotalSlots));
                l_Stats.AddTextDirect("Upload queue length", f_Inst.f_TextSet.NumStr(l_QueueLength));
            }

            l_Stats.AddLineBreak();

            l_Stats.AddCaption(TextKeys.StatsTabContacts);
            l_Stats.AddVerInts(TextKeys.StatsTabContactsInContactMngr, l_Kad.f_ContactMngr.p_Count, l_Kad.f_ContactMngr.p_VerifiedCount);
            l_Stats.AddLineBreak();

            RoutingZone.RoutingInfo l_RoutingInfo;
            l_Kad.f_RootZone.RetrieveInfo(out l_RoutingInfo);

            l_Stats.AddCaption(TextKeys.StatsTabKadRouting);
            l_Stats.AddInteger(TextKeys.StatsTabKadRoutingHighestLevel, l_RoutingInfo.f_MaxLevel);
            l_Stats.AddInteger(TextKeys.StatsTabKadRoutingRoutingBins, l_RoutingInfo.f_Bins);
            l_Stats.AddVerInts(TextKeys.StatsTabKadRoutingContactsUsed, l_RoutingInfo.f_Nodes, l_RoutingInfo.f_VerifiedNodes);

            if (10 >= l_RoutingInfo.f_EstimatedTotalVerifiedNodes)
            {
                l_Stats.AddInteger(TextKeys.StatsTabKadRoutingEstimatedNodes, l_RoutingInfo.f_EstimatedTotalVerifiedNodes);
            }
            else
            {
                l_Stats.AddText(TextKeys.StatsTabKadRoutingEstimatedNodes, string.Format("{0} - {1} (\x2248{2})", (l_RoutingInfo.f_MaxEstimatedTotalNodes / 2), l_RoutingInfo.f_MaxEstimatedTotalNodes, l_RoutingInfo.f_EstimatedTotalVerifiedNodes));
            }

            l_Stats.AddLineBreak();

            l_Stats.AddCaption(TextKeys.StatsTabPub);
            f_Inst.f_Kad.f_PublishMngr.RetrieveStats(l_Time, ref l_Stats);
            l_Stats.AddLineBreak();

            uint l_FileEntryCount;
            uint l_KeywordEntryCount;
            uint l_NoteEntryCount;
            l_Kad.f_Map.GetEntryCounts(out l_FileEntryCount, out l_KeywordEntryCount, out l_NoteEntryCount);
            uint l_KadMapMemSize = l_Kad.f_Map.GetMemorySize();

            l_Stats.AddCaption(TextKeys.StatsTabIndex);
            l_Stats.AddInteger(TextKeys.StatsTabIndexFiles, l_FileEntryCount);
            l_Stats.AddInteger(TextKeys.StatsTabIndexKeywords, l_KeywordEntryCount);
            l_Stats.AddInteger(TextKeys.StatsTabIndexNotes, l_NoteEntryCount);
            l_Stats.AddSize(TextKeys.StatsTabIndexNetSize, l_KadMapMemSize);

            return l_Stats.GetStatsAsString();
        }

        public void EnqueueAPC(GeneralAPCCookie a_Cookie)
        {
            f_APCQueue.Enqueue(a_Cookie);
        }

        public bool EnqueueAPCAndWaitAsGUI(GeneralAPCCookie a_Cookie)
        {
            // Do not continue processing if false is returned.

            EnqueueAPC(a_Cookie);

            return a_Cookie.GuiWaitForCompletion();
        }

        private void NetworkWorker()
        {
            GeneralAPCCookie[] l_APCCookies = null;

            while (true)
            {
                if (null == l_APCCookies)
                {
                    l_APCCookies = f_APCQueue.GetNow();
                }

                if (null != l_APCCookies)
                {
                    APCQueue<GeneralAPCCookie>.ExecuteAndClear(ref l_APCCookies);
                }

                if (true == f_ShutdownInProgress &&
                    0 == GetTotalConnectionCount())
                {
                    Software.UpdateShutDownState(0x10);

                    f_Inst.f_UserMngr.PrepareSavingOnShutdown();
                    StorageStates l_UserCntxtsState = f_Inst.f_UserMngr.SaveContexts();
                    StorageStates l_NodesState = f_Inst.f_Kad.SaveNodes();
                    f_DlQueue.Destroy();

                    break;
                }

                Time l_Time = f_Inst.UpdateTime();

                if (null != f_KnownFileCacheCookie)
                {
                    CookieStates l_State = f_KnownFileCacheCookie.p_StateVolatile;

                    if (CookieStates.Pending != l_State)
                    {
                        if (CookieStates.Success == l_State)
                        {
                            f_KnownFileCache = (KnownFileCache)f_KnownFileCacheCookie.GetReturnObject();
                        }
                        else
                        {
                            f_KnownFileCache = null;
                        }

                        f_KnownFileCacheCookie = null;
                    }
                }

                if (l_Time >= f_NextKnownFileCacheUpdateTime &&
                    null == f_KnownFileCacheCookie)
                {
                    const uint lc_IntervalInSecs = 15;
                    f_NextKnownFileCacheUpdateTime = l_Time.AddSeconds(lc_IntervalInSecs);
                    bool l_HasRules = f_Inst.f_Kad.f_KadPrefs.f_KadFilter.HasActiveRules();

                    if (false == l_HasRules)
                    {
                        f_KnownFileCacheCookie = null;
                        f_KnownFileCache = null;
                    }
                    else
                    {
                        f_KnownFileCacheCookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.NetworkGetKnownFileCache);
                        f_Inst.f_ContentMngr.EnqueueAPC(f_KnownFileCacheCookie);
                    }
                }

                if (BridgeStatus.Disconnected != f_BridgeStatus)
                {
                    // NOTE: We may not change f_BridgeStatus before
                    // a call to KillAllConnections(), since it relies
                    // on the old value.

                    f_Bridge.CheckSanity(l_Time);
                    BridgeStatus l_NewStatus = f_Bridge.p_Status;

                    if (BridgeStatus.Connected == f_BridgeStatus)
                    {
                        if (BridgeStatus.Connected != l_NewStatus)
                        {
                            KillAllConnections(false);
                        }
                    }
                    else if (BridgeStatus.TryingToConnect == f_BridgeStatus)
                    {
                        if (BridgeStatus.TryingToConnect == l_NewStatus)
                        {
                            f_Bridge.ContinueStart();
                            l_NewStatus = f_Bridge.p_Status;
                        }

                        if (BridgeStatus.Connected == l_NewStatus)
                        {
                            f_Inst.f_MainForm.AddToLog(TextKeys.MsgSAMBridgeConnSuccess);

                            f_BridgeStatus = BridgeStatus.Connected;
                            f_ConnCookie.SetEffectiveKeyPairs(f_Bridge.p_DatagramKeyPair, f_Bridge.p_StreamKeyPair);
                            f_ConnCookie.SetBridgeStatus(BridgeStatus.Connected);

                            NotifyNewSamConnection();
                        }
                        else if (BridgeStatus.Disconnected == l_NewStatus)
                        {
                            KillAllConnections(false);
                        }
                    }
                }

                Process(l_Time);
                f_Inst.f_Kad.Process(l_Time);

                l_APCCookies = f_APCQueue.WaitForInput(100);
            }

            f_APCQueue.ShutDown();

            f_CPUInfo.Dispose();
            bool l_OK = Util.AdjustSystemSleepPrevention(false);

            f_ShutdownCookie.SetFinished();

            Software.UpdateShutDownState(0x20);
        }

        public void Process(Time a_Time)
        {
            if (f_BridgeStatus == BridgeStatus.Connected)
            {
                ManageIncomingConns(a_Time);
            }

            ManagePendingConns(a_Time);
            ManageConns(a_Time);

            if (BridgeStatus.Connected == f_BridgeStatus &&
                true == f_NextUlTime.IsEarlierThan(a_Time))
            {
                f_Inst.f_UpDwnMngr.Process(a_Time, out f_NextUlTime);
            }

            if (true == f_NextDlQueueUpdateTime.IsEarlierThan(a_Time))
            {
                f_DlQueue.Process(a_Time);

                f_NextDlQueueUpdateTime = a_Time.AddSeconds(c_UpdateDlQueueIntervalSecs);
            }

            f_UlQueue.Process(a_Time);
        }

        public void RegisterFile(KnownFile a_Known)
        {
            // May only get called by the KnownFile class.

            f_ActiveFiles.Add(a_Known, a_Known);
        }

        public void UnregisterFile(KnownFile a_Known)
        {
            // May only get called by the KnownFile class.

            bool l_Removed = f_ActiveFiles.Remove(a_Known);
        }

        public KnownFile[] CopyConnectedFiles()
        {
            return Util.DictionaryToArray(f_ActiveFiles);
        }

        public List<ITransferable> GetStreamsForTransfer(bool a_Ul)
        {
            List<ITransferable> l_List = new List<ITransferable>(f_EstablishedConns.Count);
            Dictionary<KnownFile, KnownFile> l_Files = new Dictionary<KnownFile, KnownFile>(f_ActiveFiles.Count);

            foreach (DonkeyClient l_Clnt in f_EstablishedConns.Values)
            {
                if (false == a_Ul)
                {
                    if (true == l_Clnt.NeedsToReceive())
                    {
                        if (null == l_Clnt.p_DlPeerCntxt ||
                            false == l_Clnt.IsDownloadingFileContent())
                        {
                            // A stream that is not currently involved in
                            // transferring file content.

                            l_List.Add(l_Clnt);
                        }
                        else
                        {
                            // This stream is transferring file content and
                            // must have the file priority and speed limit
                            // applied.

                            KnownFile l_Known = l_Clnt.p_DlPeerCntxt.f_Known;
                            if (false == l_Files.ContainsKey(l_Known))
                            {
                                l_Files.Add(l_Known, l_Known);
                            }
                        }
                    }
                }
                else
                {
                    // Same as above, only different direction.

                    if (true == l_Clnt.NeedsToSend())
                    {
                        if (null == l_Clnt.p_UlPeerCntxt ||
                            false == l_Clnt.IsUploadingFileContent())
                        {
                            l_List.Add(l_Clnt);
                        }
                        else
                        {
                            KnownFile l_Known = l_Clnt.p_UlPeerCntxt.f_Known;
                            if (false == l_Files.ContainsKey(l_Known))
                            {
                                l_Files.Add(l_Known, l_Known);
                            }
                        }
                    }
                }
            }

            foreach (KnownFile l_Known in l_Files.Values)
            {
                l_List.Add(l_Known);
            }

            return l_List;
        }

        private static byte[] BuildMuleInfoPacketBody()
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            // eMuleVer.
            l_Buffer.WriteByte(68);
            // ProtVer.
            l_Buffer.WriteByte(byte.MaxValue);

            TagList l_TagList = new TagList();

            // We don't reveal the OS we run on.
            //l_TagList.AddTag(new StringTag(TagCodes.ET_OS_INFO, "Windows"));

            l_TagList.WriteTo(l_Buffer, sizeof(uint), PeerInfo.c_ExpectedUnicodeSupportForOutgoingHello);

            return l_Buffer.ToArray(true);
        }

        public uint p_ProcessingGeneration
        {
            get { return f_EstablishedConnProcessingGen; }
        }

        public uint GetTotalConnectionCount()
        {
            return (uint)(f_PendingConns.Count + f_EstablishedConns.Count);
        }

        public void ContentMngrBanUser(GeneralAPCCookie a_Cookie)
        {
            // Invoked by the content manager. This routine may be
            // called at any time after initialization, so beware.

            object[] l_Args = a_Cookie.GetArgAsObjArray();
            UserNum l_UserNum = (UserNum)l_Args[0];
            int l_RepDelta = (int)l_Args[1];

            UserContext l_UserCntxt = f_Inst.f_UserMngr.GetUserContext(l_UserNum);

            if (null == l_UserCntxt)
            {
                a_Cookie.SetState(CookieStates.Failure);

                return;
            }

            l_UserCntxt.AddToReputation(l_RepDelta);

            if (false == f_ShutdownInProgress)
            {
                BanAddress(l_UserCntxt.f_StreamAddr);
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiBanOrUnbanUsers(GeneralAPCCookie a_Cookie)
        {
            // Invoked by the GUI.

            object[] l_Args = a_Cookie.GetArgAsObjArray();
            UserContext[] l_UserCntxts = (UserContext[])l_Args[0];
            bool l_Ban = (true == (bool)l_Args[1]);

            f_Inst.f_UserMngr.BanOrUnbanUsers(l_UserCntxts, l_Ban);

            a_Cookie.SetState(CookieStates.Success);
        }

        public void GuiAdjustReputationOfUsers(GeneralAPCCookie a_Cookie)
        {
            // Invoked by the GUI.

            object[] l_Args = a_Cookie.GetArgAsObjArray();
            UserContext[] l_UserCntxts = (UserContext[])l_Args[0];
            bool l_NullifyIfNeg = (true == (bool)l_Args[1]);
            int l_Reputation = (false == l_NullifyIfNeg ? (int)l_Args[2] : 0);

            foreach (UserContext l_UserCntxt in l_UserCntxts)
            {
                if (false == l_NullifyIfNeg)
                {
                    l_UserCntxt.SetReputation(l_Reputation);
                }
                else
                {
                    l_UserCntxt.NullifyReputationIfNegative();
                }
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void BanAddress(I2PAddress a_StreamAddr)
        {
            f_Bridge.BanDestination(a_StreamAddr);
        }

        public void UnbanAddress(I2PAddress a_StreamAddr)
        {
            f_Bridge.UnbanDestination(a_StreamAddr);
        }

        private void NotifyBannedAddress(I2PAddress a_BannedAddr)
        {
            UserContext l_UserCntxt;

            if (true == f_Inst.f_UserMngr.TryGetUserContext(a_BannedAddr, out l_UserCntxt))
            {
                // We only kill established connections, since pending
                // connections to a banned destination will not be able
                // to become established and die.
                l_UserCntxt.SetBanned(true);
            }

            f_Inst.f_Kad.NotifyBannedAddress(a_BannedAddr);
        }

        public bool CanConnectTo(I2PAddress a_StreamAddr,
                                 out SourceStates a_ErrSourceState)
        {
            // Even if zero is returned, a_ErrSourceState can be
            // SourceStates.Undefined on return.

            if (BridgeStatus.Connected != f_BridgeStatus)
            {
                a_ErrSourceState = SourceStates.Undefined;

                return false;
            }

            if (f_NetworkPrefs.f_MaxGlobalConns <= GetTotalConnectionCount())
            {
                a_ErrSourceState = SourceStates.OwnGlobalLimitReached;

                return false;
            }

            UserContext l_UserCntxt;

            if (false == f_Inst.f_UserMngr.TryGetUserContext(a_StreamAddr, out l_UserCntxt))
            {
                if (true == f_Inst.f_NetworkMngr.IsAddressBanned(a_StreamAddr))
                {
                    a_ErrSourceState = SourceStates.PeerBanned;

                    return false;
                }
            }
            else
            {
                if (true == l_UserCntxt.IsBanned())
                {
                    a_ErrSourceState = SourceStates.PeerBanned;

                    return false;
                }

                if (false == l_UserCntxt.CanMakeAnotherConnAttempt(f_NetworkPrefs.f_MaxOutgoingConnsPerAddr))
                {
                    a_ErrSourceState = SourceStates.OwnPerUserLimitReached;

                    return false;
                }

                if (true == l_UserCntxt.PreventsA4AFCreatingAnotherDlContext(true))
                {
                    a_ErrSourceState = SourceStates.A4AFLimitation;

                    return false;
                }
            }

            a_ErrSourceState = SourceStates.Undefined;

            return true;
        }

        public bool RequestConnection(I2PAddress a_StreamAddr,
                                      ConnResultContext a_ConnResCntxt)
        {
            // The caller has to check CanConnectTo() before calling.

            I2PConnectCookie l_Cookie;
            ConnectResult l_ConnRes = f_Bridge.BeginConnect(a_StreamAddr, out l_Cookie);

            if (ConnectResult.Pending == l_ConnRes)
            {
                UserContext l_UserCntxt = f_Inst.f_UserMngr.GetOrCreateUserContext(a_StreamAddr);
                PendingConnContext l_PendConn = new PendingConnContext(l_Cookie, l_UserCntxt, a_ConnResCntxt);

                l_UserCntxt.RegisterPendingConn();
                f_PendingConns.Add(l_PendConn, l_PendConn);

                return true;
            }

            return false;
        }

        private void ManageIncomingConns(Time a_Time)
        {
            for (int i = 0; i < c_MaxIncomingConnsPerTurn; i++)
            {
                I2PStream l_Strm = f_Bridge.TryAcceptConnection();

                if (null == l_Strm)
                {
                    break;
                }

                if (GetTotalConnectionCount() >= f_NetworkPrefs.f_MaxGlobalConns)
                {
                    l_Strm.Close();
                }
                else
                {
                    UserContext l_DestCntxt = f_Inst.f_UserMngr.GetOrCreateUserContext(l_Strm.p_RemoteAddr);

                    if (l_DestCntxt.p_IngoingConns >= f_NetworkPrefs.f_MaxIngoingConnsPerAddr)
                    {
                        l_Strm.Close();
                    }
                    else
                    {
                        DonkeyClient l_Clnt = new DonkeyClient(f_Inst, f_NextDebugConnID++, l_DestCntxt, l_Strm, null);

                        f_Inst.f_UserMngr.AddConnection(l_Clnt);
                        f_EstablishedConns.Add(l_Clnt, l_Clnt);
                        f_AcceptedIngoingConns++;
                        f_ActiveIngoingConns++;
                    }
                }
            }
        }

        private void ManagePendingConns(Time a_Time)
        {
            // This routines takes care of out outgoing connection attempts.

            List<PendingConnContext> l_RemLs = null;

            foreach (PendingConnContext l_PendConn in f_PendingConns.Values)
            {
                ConnectResult l_Res;
                I2PStream l_Strm;

                if (BridgeStatus.Connected == f_BridgeStatus)
                {
                    l_Res = f_Bridge.TryFinishConnect(l_PendConn.f_Cookie, out l_Strm);

                    if (ConnectResult.Pending == l_Res)
                    {
                        continue;
                    }
                }
                else
                {
                    l_PendConn.f_Cookie.Close();

                    l_Res = ConnectResult.Shutdown;
                    l_Strm = null;
                }

                l_PendConn.f_UserCntxt.UnregisterPendingConn();

                if (null == l_RemLs)
                {
                    l_RemLs = new List<PendingConnContext>(1);
                }

                l_RemLs.Add(l_PendConn);

                if (ConnectResult.Success == l_Res)
                {
                    if (true == l_PendConn.f_UserCntxt.IsBanned())
                    {
                        l_Res = ConnectResult.UserBanned;

                        l_Strm.Close();
                    }
                    else if (true == f_ShutdownInProgress)
                    {
                        l_Res = ConnectResult.Shutdown;

                        l_Strm.Close();
                    }
                    else
                    {
                        DonkeyClient l_Clnt = new DonkeyClient(f_Inst, f_NextDebugConnID++, l_PendConn.f_UserCntxt, l_Strm, l_PendConn.f_ConnResCntxt);

                        f_Inst.f_UserMngr.AddConnection(l_Clnt);
                        f_EstablishedConns.Add(l_Clnt, l_Clnt);
                        f_SuccessfulOutgoingConns++;

                        continue;
                    }
                }

                // An error occurred.

                l_PendConn.f_ConnResCntxt.Invoke(f_Inst, l_Res);
            }

            if (null != l_RemLs)
            {
                foreach (PendingConnContext l_PendConn in l_RemLs)
                {
                    f_PendingConns.Remove(l_PendConn);
                }
            }
        }

        private void ManageConns(Time a_Time)
        {
            f_EstablishedConnProcessingGen++;

            // Process connections that are associated with a
            // download or upload context.

            foreach (KnownFile l_Known in f_ActiveFiles.Keys)
            {
                l_Known.p_ActCntxt.ProcessPeerContexts(a_Time);
            }

            // Not all connections belong to a file, so there might
            // be some left that we are processing now. At the same time
            // we will check for dead connections. Finally we'll create
            // upload contexts (outside the enumeration of f_ActiveFiles).

            List<DonkeyClient> l_RemLs = null;

            foreach (DonkeyClient l_Clnt in f_EstablishedConns.Values)
            {
                bool l_Keep = l_Clnt.p_IsAlive;

                if (true == l_Keep)
                {
                    if (false == l_Clnt.TryEndCreateUlPeerContext())
                    {
                        l_Keep = false;
                    }
                    else if (false == l_Clnt.WasProcessed(f_EstablishedConnProcessingGen))
                    {
                        if (false == l_Clnt.Process(a_Time))
                        {
                            l_Keep = false;
                        }
                        else
                        {
                            DlPeerContext l_DlPeerCntxt = l_Clnt.p_DlPeerCntxt;
                            UlPeerContext l_UlPeerCntxt = l_Clnt.p_UlPeerCntxt;

                            if (null != l_DlPeerCntxt &&
                                true == l_DlPeerCntxt.ShouldBeRemoved())
                            {
                                l_DlPeerCntxt.f_Known.DestroyDlPeerContext(l_DlPeerCntxt);
                            }

                            if (null != l_UlPeerCntxt &&
                                true == l_UlPeerCntxt.ShouldBeRemoved())
                            {
                                l_UlPeerCntxt.f_Known.DestroyUlPeerContext(l_UlPeerCntxt);
                            }
                        }
                    }

                    if (true == l_Keep &&
                        false == l_Clnt.DoLivelinessCheck(a_Time))
                    {
                        l_Keep = false;
                    }
                }

                if (false == l_Keep)
                {
                    if (null == l_RemLs)
                    {
                        l_RemLs = new List<DonkeyClient>(1);
                    }

                    l_RemLs.Add(l_Clnt);
                }
            }

            if (null != l_RemLs)
            {
                // Deinitialize connections that are about to die.

                foreach (DonkeyClient l_Clnt in l_RemLs)
                {
                    DlPeerContext l_DlPeerCntxt = l_Clnt.p_DlPeerCntxt;
                    UlPeerContext l_UlPeerCntxt = l_Clnt.p_UlPeerCntxt;

                    if (null != l_DlPeerCntxt)
                    {
                        l_DlPeerCntxt.f_Known.DestroyDlPeerContext(l_DlPeerCntxt);
                    }

                    if (null != l_UlPeerCntxt)
                    {
                        l_UlPeerCntxt.f_Known.DestroyUlPeerContext(l_UlPeerCntxt);
                    }

                    f_Inst.f_UserMngr.RemoveConnection(l_Clnt);
                    f_EstablishedConns.Remove(l_Clnt);

                    if (false == l_Clnt.p_Outgoing)
                    {
                        f_ActiveIngoingConns--;
                    }

                    // Call the shutdown code after the connection was
                    // unregistered, so that the handler sees the new
                    // situation.
                    l_Clnt.Shutdown();
                }
            }
        }

        public static bool IsValidRank(uint a_Rank)
        {
            return (c_BestRank <= a_Rank &&
                    c_MaxRank >= a_Rank);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct StatsBuilder
        {
            // Don't use the struct default constructor.

            private readonly StringBuilder f_Bld;
            private readonly TextSet f_TextSet;
            private readonly MainForm f_Form;
            private readonly string f_FormatVerifiedNum;

            private StringBuilder f_TimeBld;

            public StatsBuilder(Instance a_Inst)
            {
                f_Bld = new StringBuilder();
                f_TextSet = a_Inst.f_TextSet;
                f_Form = a_Inst.f_MainForm;
                f_FormatVerifiedNum = f_TextSet[TextKeys.StatsTabFormatVerifiedNum];

                f_TimeBld = null;
            }

            public void AddLineBreak()
            {
                f_Bld.AppendLine();
            }

            public void AddCaption(TextKeys a_Key)
            {
                f_Bld.AppendLine(f_TextSet[a_Key]);
            }

            public void AddTime(TextKeys a_Key,
                                long a_Seconds)
            {
                if (null == f_TimeBld)
                {
                    f_TimeBld = new StringBuilder();
                }
                else
                {
                    f_TimeBld.Length = 0;
                }

                TimeStrHelper.AppendTimeString(f_TimeBld, a_Seconds);
                AddText(a_Key, f_TimeBld.ToString());
            }

            public void AddSize(TextKeys a_Key,
                                ulong a_Value)
            {
                AddText(a_Key, f_Form.BuildDefaultSizeStr(a_Value));
            }

            public void AddText(TextKeys a_Key,
                                TextKeys a_Value)
            {
                AddText(a_Key, f_TextSet[a_Value]);
            }

            public void AddText(TextKeys a_Key,
                                string a_Value)
            {
                f_TextSet.AddText(f_Bld, a_Key, a_Value);
            }

            public void AddTextDirect(string a_Key,
                                      string a_Value)
            {
                f_Bld.AppendFormat(f_TextSet.Format(TextKeys.KeyValuePairFormat, a_Key, a_Value));
                f_Bld.AppendLine();
            }

            public void AddInteger(TextKeys a_Key,
                                   long a_Value)
            {
                f_TextSet.AddInteger(f_Bld, a_Key, a_Value);
            }

            public void AddVerInts(TextKeys a_Key,
                                   long a_Value,
                                   long a_VerifiedValue)
            {
                AddText(a_Key, string.Format(f_FormatVerifiedNum, a_Value, a_VerifiedValue));
            }

            public string GetStatsAsString()
            {
                return f_Bld.ToString();
            }
        }

        public class ShutdownCookie : GeneralAPCCookie
        {
            private readonly MonitorEvent f_ShutdownFinished;

            public ShutdownCookie(NetworkManager a_NetworkMngr) :
                             base(a_NetworkMngr.GuiBeginShutdown)
            {
                f_ShutdownFinished = new MonitorEvent(false);
            }

            public void SetFinished()
            {
                f_ShutdownFinished.Set();
            }

            public void Wait()
            {
                base.WaitForCompletion();
                f_ShutdownFinished.Wait();
            }
        }

        public class ConnResultContext
        {
            public delegate void Handler(Instance a_Inst,
                                         ConnResultContext a_ConnResCntxt);

            private readonly Handler f_Handler;
            public readonly KnownFile f_Known;
            public readonly FileID f_FileID;
            public readonly I2PAddress f_Addr;

            private ConnectResult f_Result;
            private DonkeyClient f_Client;

            public ConnectResult p_Result
            {
                get { return f_Result; }
            }

            public DonkeyClient p_Client
            {
                // p_Client can be null.
                get { return f_Client; }
            }

            public ConnResultContext(Handler a_Handler,
                                     KnownFile a_Known,
                                     FileID a_FileID,
                                     I2PAddress a_Addr)
            {
                f_Handler = a_Handler;
                f_Known = a_Known;
                f_FileID = a_FileID;
                f_Addr = a_Addr;
            }

            public void Invoke(Instance a_Inst,
                               ConnectResult a_Result)
            {
                f_Result = a_Result;

                f_Handler(a_Inst, this);
            }

            public void Invoke(Instance a_Inst,
                               ConnectResult a_Result,
                               DonkeyClient a_Client)
            {
                f_Result = a_Result;
                f_Client = a_Client;

                f_Handler(a_Inst, this);
            }

            public void NotifySourceState(Instance a_Inst,
                                          SourceStates a_SourceState)
            {
                a_Inst.f_NetworkMngr.f_DlQueue.NotifySourceState(f_Known, f_Addr, a_SourceState);
            }
        }
    }

    class PendingConnContext
    {
        public readonly I2PConnectCookie f_Cookie;
        public readonly UserContext f_UserCntxt;
        public readonly NetworkManager.ConnResultContext f_ConnResCntxt;

        public PendingConnContext(I2PConnectCookie a_Cookie,
                                  UserContext a_UserCntxt,
                                  NetworkManager.ConnResultContext a_ConnResCntxt)
        {
            f_Cookie = a_Cookie;
            f_UserCntxt = a_UserCntxt;
            f_ConnResCntxt = a_ConnResCntxt;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TunnelSettings
    {
        public readonly byte f_DatagramQuantity;
        public readonly byte f_StreamQuantity;
        public readonly byte f_MinHops;
        public readonly byte f_RandomExtraHops;

        public TunnelSettings(uint a_DatagramQuantity,
                              uint a_StreamQuantity,
                              uint a_MinHops,
                              uint a_RandomExtraHops)
        {
            f_DatagramQuantity = (byte)a_DatagramQuantity;
            f_StreamQuantity = (byte)a_StreamQuantity;
            f_MinHops = (byte)a_MinHops;
            f_RandomExtraHops = (byte)a_RandomExtraHops;
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      out TunnelSettings a_Dest)
        {
            uint l_DatagramQuantity = a_Buffer.ReadByte();
            uint l_StreamQuantity = a_Buffer.ReadByte();
            uint l_MinHops = a_Buffer.ReadByte();
            uint l_RandomExtraHops = a_Buffer.ReadByte();

            if (AdvancedTunnelSettings.c_MaxTunnels < l_DatagramQuantity ||
                AdvancedTunnelSettings.c_MaxTunnels < l_StreamQuantity ||
                AdvancedTunnelSettings.c_MaxHops < l_MinHops ||
                AdvancedTunnelSettings.c_MaxHops < l_RandomExtraHops)
            {
                a_Dest = new TunnelSettings();

                return false;
            }

            a_Dest = new TunnelSettings(l_DatagramQuantity, l_StreamQuantity, l_MinHops, l_RandomExtraHops);

            return true;
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteByte(f_DatagramQuantity);
            a_Buffer.WriteByte(f_StreamQuantity);
            a_Buffer.WriteByte(f_MinHops);
            a_Buffer.WriteByte(f_RandomExtraHops);
        }

        public static void CreateDefault(out TunnelSettings a_Dest)
        {
            a_Dest = new TunnelSettings(2, 3, 2, 0);
        }

        public bool IsShortTunnel()
        {
            return (0 == f_MinHops ||
                    (1 == f_MinHops &&
                     0 == f_RandomExtraHops));
        }

        public bool ValueEquals(TunnelSettings a_Other)
        {
            return (this.f_DatagramQuantity == a_Other.f_DatagramQuantity &&
                    this.f_StreamQuantity == a_Other.f_StreamQuantity &&
                    this.f_MinHops == a_Other.f_MinHops &&
                    this.f_RandomExtraHops == a_Other.f_RandomExtraHops);
        }

        public void ToString(TextSet a_TextSet,
                             StringBuilder a_Bld)
        {
            string l_RandHopsStr;

            if (0 == f_RandomExtraHops)
            {
                l_RandHopsStr = a_TextSet[TextKeys.OptionsDlgTunnelsNoHops];
            }
            else
            {
                l_RandHopsStr = a_TextSet.Format(TextKeys.OptionsDlgTunnelsPosNumFormat, f_RandomExtraHops);
            }

            a_TextSet.AddInteger(a_Bld, TextKeys.OptionsDlgTunnelsDatagramQuant, f_DatagramQuantity);
            a_TextSet.AddInteger(a_Bld, TextKeys.OptionsDlgTunnelsStreamQuant, f_StreamQuantity);
            a_TextSet.AddInteger(a_Bld, TextKeys.OptionsDlgTunnelsMinHops, f_MinHops);
            a_TextSet.AddText(a_Bld, TextKeys.OptionsDlgTunnelsRandExtraHops, l_RandHopsStr);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct NetworkPrefs
    {
        public const ushort c_MaxConns = (10 * 1000);
        public const byte c_MaxDlFiles = byte.MaxValue;
        // Since the NetworkAvailability class uses a byte array,
        // the following isn't just an arbitrary limit (still this
        // limit has no practical meaning for P2P file sharing,
        // since bandwidth restrictions are the real issue).
        public const byte c_MaxDlOrUlConnsPerFile = (byte.MaxValue / 2);
        public const byte c_MaxGlobalUlSlots = byte.MaxValue;

        public readonly Speed f_DlLimit;
        public readonly Speed f_UlLimit;
        public readonly ushort f_MaxGlobalConns;
        public readonly ushort f_MaxIngoingConnsPerAddr;
        public readonly ushort f_MaxOutgoingConnsPerAddr;
        public readonly ushort f_MaxDlFiles;
        public readonly byte f_MaxDownloadConnsPerFile;
        // f_MaxUploadConnsPerFile is not saved, since the
        // default value is always used.
        public readonly byte f_MaxUploadConnsPerFile;
        public readonly byte f_MaxGlobalUlSlots;

        public readonly string f_UserName;
        public readonly bool f_AllowDataCompression;
        // f_PoisonPeers works only for debug builds.
        public readonly bool f_PoisonPeers;

        public NetworkPrefs(Speed a_DlLimit,
                            Speed a_UlLimit,
                            uint a_MaxConnsGlobal,
                            uint a_MaxIngoingConnsPerAddr,
                            uint a_MaxOutgoingConnsPerAddr,
                            uint a_MaxDlFiles,
                            uint a_MaxDownloadConnsPerFile,
                            uint a_MaxGlobalUlSlots,
                            string a_UserName,
                            bool a_AllowDataCompression,
                            bool a_PoisonPeers)
        {
            if (c_MaxDlOrUlConnsPerFile < a_MaxDownloadConnsPerFile)
            {
                throw (new ApplicationException());
            }

            f_DlLimit = a_DlLimit;
            f_UlLimit = a_UlLimit;
            f_MaxGlobalConns = (ushort)a_MaxConnsGlobal;
            f_MaxIngoingConnsPerAddr = (ushort)a_MaxIngoingConnsPerAddr;
            f_MaxOutgoingConnsPerAddr = (ushort)a_MaxOutgoingConnsPerAddr;
            f_MaxDlFiles = (ushort)a_MaxDlFiles;
            f_MaxDownloadConnsPerFile = (byte)a_MaxDownloadConnsPerFile;
            f_MaxUploadConnsPerFile = c_MaxDlOrUlConnsPerFile;
            f_MaxGlobalUlSlots = (byte)a_MaxGlobalUlSlots;
            f_UserName = a_UserName;
            f_AllowDataCompression = a_AllowDataCompression;
            f_PoisonPeers = a_PoisonPeers;
        }

        public static void CreateDefault(out NetworkPrefs a_Dest)
        {
            a_Dest = new NetworkPrefs(Speed.s_MaxSpeed,
                                      Speed.s_MaxSpeed,
                                      150,
                                      10,
                                      10,
                                      15,
                                      10,
                                      10,
                                      Software.c_AppName + " User",
                                      true,
                                      false);
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      uint a_Revision,
                                      out NetworkPrefs a_Dest)
        {
            Speed l_DlLimit = new Speed(a_Buffer.ReadUInt32LE());
            Speed l_UlLimit = new Speed(a_Buffer.ReadUInt32LE());
            uint l_MaxGlobalConns = a_Buffer.ReadUInt16LE();
            uint l_MaxIngoingConnsPerDest = a_Buffer.ReadUInt16LE();
            uint l_MaxOutgoingConnsPerDest = a_Buffer.ReadUInt16LE();
            uint l_MaxDlFiles = a_Buffer.ReadUInt16LE();
            uint l_MaxDownloadConnsPerFile = a_Buffer.ReadByte();
            uint l_MaxGlobalUlSlots = a_Buffer.ReadByte();
            string l_UserName = a_Buffer.ReadUTF16LEStringEncLen();
            bool l_AllowDataCompression = a_Buffer.ReadBool();
            bool l_PoisonPeers = a_Buffer.ReadBool();

            if (false == l_DlLimit.IsValid() ||
                false == l_UlLimit.IsValid() ||
                c_MaxConns < l_MaxGlobalConns ||
                c_MaxConns < l_MaxIngoingConnsPerDest ||
                c_MaxConns < l_MaxOutgoingConnsPerDest ||
                c_MaxDlFiles < l_MaxDlFiles ||
                c_MaxDlOrUlConnsPerFile < l_MaxDownloadConnsPerFile ||
                c_MaxGlobalUlSlots < l_MaxGlobalUlSlots ||
                eD2k.PeerInfo.c_MaxUserNameChars < l_UserName.Length)
            {
                Util.DebugBreak();

                a_Dest = new NetworkPrefs();

                return false;
            }

            a_Dest = new NetworkPrefs(l_DlLimit, l_UlLimit, l_MaxGlobalConns, l_MaxIngoingConnsPerDest, l_MaxOutgoingConnsPerDest, l_MaxDlFiles, l_MaxDownloadConnsPerFile, l_MaxGlobalUlSlots, l_UserName, l_AllowDataCompression, l_PoisonPeers);

            return true;
        }

        public uint ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt32LE(f_DlLimit.p_BytesPerSec);
            a_Buffer.WriteUInt32LE(f_UlLimit.p_BytesPerSec);
            a_Buffer.WriteUInt16LE(f_MaxGlobalConns);
            a_Buffer.WriteUInt16LE(f_MaxIngoingConnsPerAddr);
            a_Buffer.WriteUInt16LE(f_MaxOutgoingConnsPerAddr);
            a_Buffer.WriteUInt16LE(f_MaxDlFiles);
            a_Buffer.WriteByte(f_MaxDownloadConnsPerFile);
            a_Buffer.WriteByte(f_MaxGlobalUlSlots);
            a_Buffer.WriteUTF16LEStringEncLen(f_UserName);
            a_Buffer.WriteBool(f_AllowDataCompression);
            a_Buffer.WriteBool(f_PoisonPeers);

            // Increase with every change.
            return 1;
        }

        public void DetermineChangeKind(ref NetworkPrefs a_ChangeTo,
                                        ref PrefChangeKinds a_ChangeKind)
        {
            if (PrefChangeKinds.AppRestartNeededToFullyTakeOver == a_ChangeKind)
            {
                return;
            }

            if (PrefChangeKinds.ReconnectNeededToFullyTakeOver == a_ChangeKind ||
                // NOTE: Decreasing the connection limits will not kill
                // connections, so a reconnect might be necessary to
                // enforce the new limit.
                this.f_MaxGlobalConns > a_ChangeTo.f_MaxGlobalConns ||
                this.f_MaxIngoingConnsPerAddr > a_ChangeTo.f_MaxIngoingConnsPerAddr ||
                this.f_MaxOutgoingConnsPerAddr > a_ChangeTo.f_MaxOutgoingConnsPerAddr ||
                this.f_UserName != a_ChangeTo.f_UserName ||
                this.f_AllowDataCompression != a_ChangeTo.f_AllowDataCompression)
            {
                a_ChangeKind = PrefChangeKinds.ReconnectNeededToFullyTakeOver;

                return;
            }

            if (PrefChangeKinds.ApplicableWithoutReconnect == a_ChangeKind ||
                false == this.f_DlLimit.ValueEquals(a_ChangeTo.f_DlLimit) ||
                false == this.f_UlLimit.ValueEquals(a_ChangeTo.f_UlLimit) ||
                this.f_MaxGlobalConns != a_ChangeTo.f_MaxGlobalConns ||
                this.f_MaxIngoingConnsPerAddr != a_ChangeTo.f_MaxIngoingConnsPerAddr ||
                this.f_MaxOutgoingConnsPerAddr != a_ChangeTo.f_MaxOutgoingConnsPerAddr ||
                this.f_MaxDlFiles != a_ChangeTo.f_MaxDlFiles ||
                this.f_MaxDownloadConnsPerFile != a_ChangeTo.f_MaxDownloadConnsPerFile ||
                this.f_MaxUploadConnsPerFile != a_ChangeTo.f_MaxUploadConnsPerFile ||
                this.f_MaxGlobalUlSlots != a_ChangeTo.f_MaxGlobalUlSlots ||
                this.f_PoisonPeers != a_ChangeTo.f_PoisonPeers)
            {
                a_ChangeKind = PrefChangeKinds.ApplicableWithoutReconnect;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TransferVolume
    {
        private ulong f_BytesReceived;
        private ulong f_BytesSent;

        public ulong p_BytesReceived
        {
            get { return f_BytesReceived; }
        }

        public ulong p_BytesSent
        {
            get { return f_BytesSent; }
        }

        public ulong p_Sum
        {
            get { return (f_BytesReceived + f_BytesSent); }
        }

        public void CopyTo(out TransferVolume a_Dest)
        {
            a_Dest = new TransferVolume();
            a_Dest.f_BytesReceived = f_BytesReceived;
            a_Dest.f_BytesSent = f_BytesSent;
        }

        public void CopyToVolatile(out TransferVolume a_Dest)
        {
            // Copies from a volatile source.
            // NOTE: On 32 bit machines this can't be done really
            // thread safe, since the owning thread of a_VolatileSrc
            // might update it while we are reading (without minding
            // thread safety). But since only the GUI relies on
            // this routine, we just neglect that unlikely
            // scenario. In the worst case a garbage value will be
            // displayed for one second.

            CopyTo(out a_Dest);
            Util.MemoryBarrier();
        }

        public float GetUlDlRatio()
        {
            if (0 == f_BytesReceived)
            {
                if (0 == f_BytesSent)
                {
                    return 1f;
                }

                return float.MaxValue;
            }

            if (0 == f_BytesSent)
            {
                return float.MinValue;
            }

            return (f_BytesSent / (float)f_BytesReceived);
        }

        public int ToBoosts()
        {
            if (f_BytesReceived > f_BytesSent)
            {
                return 1;
            }

            if (f_BytesReceived < f_BytesSent)
            {
                return (-1);
            }

            return 0;
        }

        public int AsCompareWeight()
        {
            // Always returns a value in the short range. We will
            // stop accounting for difference at one GiB.

            long l_Diff = (long)(Math.Min(long.MaxValue, f_BytesReceived) - Math.Min(long.MaxValue, f_BytesSent));

            return (int)Math.Max(short.MinValue, Math.Min(short.MaxValue, l_Diff >> 15));
        }

        public void Sent(uint a_Bytes)
        {
            f_BytesSent += a_Bytes;
        }

        public void Received(uint a_Bytes)
        {
            f_BytesReceived += a_Bytes;
        }

        public void Add(TransferVolume a_Other)
        {
            f_BytesReceived += a_Other.f_BytesReceived;
            f_BytesSent += a_Other.f_BytesSent;
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt64LE(f_BytesReceived);
            a_Buffer.WriteUInt64LE(f_BytesSent);
        }

        public static void FromBuffer(BufferIO a_Buffer,
                                      out TransferVolume a_Dest)
        {
            a_Dest.f_BytesReceived = a_Buffer.ReadUInt64LE();
            a_Dest.f_BytesSent = a_Buffer.ReadUInt64LE();
        }
    }

    class OwnUserPersona
    {
        private const ushort c_MaxPubKeyBytesiMule = 80;
        // A 80 bytes long ASN.1 pubkey (in the ASN.1 dialect we
        // support) is 400 bits large. Note that a 400 bit key
        // is broken from the start nowadays.
        public const ushort c_MaxKeyBitsiMule = 400;

        // The limits of the RSACryptoServiceProvider class.
        public const ushort c_MinKeyBits = 384;
        public const ushort c_MaxKeyBits = 16384;

        public const ushort c_DefaultTinyKeyBits = c_MaxKeyBitsiMule;
        public const ushort c_MinLargeKeyBits = (4 * 1024);

        public readonly UserID f_UserID;
        // The tiny key is for iMule compatibility.
        public readonly RsaKeyPair f_TinyKeyPair;
        public readonly RsaKeyPair f_LargeKeyPair;

        private OwnUserPersona(UserID a_UserID,
                               RsaKeyPair a_TinyKeyPair,
                               RsaKeyPair a_LargeKeyPair)
        {
            f_UserID = a_UserID;
            f_TinyKeyPair = a_TinyKeyPair;
            f_LargeKeyPair = a_LargeKeyPair;
        }

        public static OwnUserPersona CreateDummy()
        {
            // Always returns very quick.

            UserID l_UserID = new UserID(UInt128.s_Zero);
            RsaKeyPair l_TinyKeyPair = RsaKeyPair.GenerateKey(c_MinKeyBits);
            RsaKeyPair l_LargeKeyPair = RsaKeyPair.GenerateKey(c_MinKeyBits);

            return new OwnUserPersona(l_UserID, l_TinyKeyPair, l_LargeKeyPair);
        }

        public static OwnUserPersona CreateNew()
        {
            // Key generation may take a while (even longer than
            // one minute).

            // We don't use a constant key size since version 1.4.8
            // to potentially make attacks harder. The key size will
            // be divisible by 32 though, to make it less likely that
            // any library isn't accepting it. Max key size is
            // currently 4,576 bits.
            uint l_EffLargeKeySize = (c_MinLargeKeyBits + ((Util.GetRandomUInt() << 5) % 512));

            UserID l_UserID = UserID.GetRandom();
            RsaKeyPair l_TinyKeyPair = RsaKeyPair.GenerateKey(c_DefaultTinyKeyBits);
            RsaKeyPair l_LargeKeyPair = RsaKeyPair.GenerateKey(l_EffLargeKeySize);

            return new OwnUserPersona(l_UserID, l_TinyKeyPair, l_LargeKeyPair);
        }

        public static OwnUserPersona FromFile(BufferIO a_Buffer)
        {
            // Return can be null if the presented data is garbage.

            UserID l_UserID = new UserID(a_Buffer.ReadUInt128LE());
            RsaKeyPair l_TinyKeyPair = RsaKeyPair.FromBuffer(a_Buffer);

            if (null == l_TinyKeyPair)
            {
                return null;
            }

            RsaKeyPair l_LargeKeyPair = RsaKeyPair.FromBuffer(a_Buffer);

            if (null == l_LargeKeyPair)
            {
                return null;
            }

            return new OwnUserPersona(l_UserID, l_TinyKeyPair, l_LargeKeyPair);
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt128LE(f_UserID.f_Val);
            f_TinyKeyPair.ToBuffer(a_Buffer);
            f_LargeKeyPair.ToBuffer(a_Buffer);
        }

        public bool ValueEquals(OwnUserPersona a_Other)
        {
            return (true == this.f_UserID.f_Val.ValueEquals(a_Other.f_UserID.f_Val) &&
                    true == this.f_TinyKeyPair.ValueEquals(a_Other.f_TinyKeyPair) &&
                    true == this.f_LargeKeyPair.ValueEquals(a_Other.f_LargeKeyPair));
        }

        public bool IsMyUserID(UserID a_UserID)
        {
            return f_UserID.f_Val.Equals(a_UserID.f_Val);
        }

        public RsaKeyPair GetSuitableKeyPair(uint a_SecIdentVer)
        {
            return (2 >= a_SecIdentVer ? f_TinyKeyPair : f_LargeKeyPair);
        }
    }

    class UserContext
    {
        public const ushort c_CurRevision = 2;
        private const ushort c_Magic = 0x17A5;
        // 4096 equals 32 MiB of good data.
        private const ushort c_MayPublishReputationThreshold = (4 * 1024);
        private const bool c_AssumedA4AFStateOfUnknownPeer = true;

        public readonly UserNum f_UserNum;
        public readonly I2PAddress f_StreamAddr;

        // f_DatagramAddr can be null and can be changed whenever the
        // peer likes.
        private I2PAddress f_DatagramAddr;
        // f_UserID and f_RsaPubKey can be uninitialized. After the peer
        // provided this information for the first time, it cannot be
        // changed again.
        private UserID f_UserID;
        private RsaPublicKey f_RsaPubKey;
        // f_ValidSignatureReceived tells us if the user at least once
        // showed us that he/she has the correct private key for the
        // alleged public key.
        private bool f_ValidSignatureReceived;
        // f_UserName can be null.
        private string f_LastUserName;

        // f_LastConnectedTime is only updated whenever the last
        // connection to the peer ends.
        private Time f_LastConnectedTime;
        private TransferVolume f_TransferVol;
        private int f_Reputation;
        private bool f_Friend;
        private bool f_HasA4AFLimitation;

        // f_EstablishedConns and f_UserName can be null.
        private Dictionary<DonkeyClient, DonkeyClient> f_EstablishedConns;

        private bool f_Banned;
        private bool f_PriorityUpToDate;
        private Priority f_Priority;
        private ushort f_IngoingConns;
        private ushort f_OutgoingConns;

        public UploadQueue.UploadQueueFields f_UploadQueueFields;

        private Time f_NextCompareWeightUpdateTime;
        private int f_UlQueueCompareWeight;

        private Time f_LastAcceptedForUlTime;

        public I2PAddress p_DatagramAddr
        {
            get { return f_DatagramAddr; }
        }

        public UserID p_UserID
        {
            get { return f_UserID; }
        }

        public Priority p_Priority
        {
            get
            {
                if (false == f_PriorityUpToDate)
                {
                    UpdatePriority();
                }

                return f_Priority;
            }
        }

        public ushort p_IngoingConns
        {
            get { return f_IngoingConns; }
        }

        public ushort p_OutgoingConns
        {
            get { return f_OutgoingConns; }
        }

        private uint p_TotalConns
        {
            get { return ((uint)f_IngoingConns + f_OutgoingConns); }
        }

        public int p_Reputation
        {
            get { return f_Reputation; }
        }

        public int p_UlQueueCompareWeight
        {
            get
            {
                return f_UlQueueCompareWeight;
            }
        }

        public Time p_LastAcceptedForUlTime
        {
            get { return f_LastAcceptedForUlTime; }
        }

        private UserContext(UserNum a_UserNum,
                            I2PAddress a_StreamAddr)
        {
            f_UserNum = a_UserNum;
            f_StreamAddr = a_StreamAddr;
        }

        public static UserContext CreateNew(UserNum a_UserNum,
                                            I2PAddress a_StreamAddr)
        {
            // Should only be used by the UserManager class.

            UserContext l_UserContext = new UserContext(a_UserNum, a_StreamAddr);
            l_UserContext.f_HasA4AFLimitation = c_AssumedA4AFStateOfUnknownPeer;

            return l_UserContext;
        }

        public static BufferIO.LoadResult FromBuffer(BufferIO a_Buffer,
                                                     bool a_OldAddrFormat,
                                                     out UserContext a_UserCntxt)
        {
            // Unless the return value is BufferIO.LoadResult.Success,
            // a_UserCntxt will be null on return. If only the stream
            // destination is illegal, InvalidDataButBufferPosIsStillGood
            // is returned.

            a_UserCntxt = null;

            if (c_Magic != a_Buffer.ReadUInt16LE())
            {
                return BufferIO.LoadResult.Corruption;
            }

            ulong l_UserNum = a_Buffer.ReadUInt64LE();

            if (UserNum.c_ReservedUserNums > l_UserNum)
            {
                // Invalid user num from generic pool.

                return BufferIO.LoadResult.Corruption;
            }

            I2PAddress l_StreamAddr;
            BufferIO.LoadResult l_Res = I2PAddress.TryCreateFromBufferLegayFormat(a_Buffer, a_OldAddrFormat, out l_StreamAddr);

            if (BufferIO.LoadResult.Success != l_Res)
            {
                return BufferIO.LoadResult.Corruption;
            }

            SaveFlags l_Flags = (SaveFlags)a_Buffer.ReadUInt16LE();
            UserContext l_UserCntxt = new UserContext(new UserNum(l_UserNum), l_StreamAddr);

            if (0 != (SaveFlags.GotDatagramAddr & l_Flags))
            {
                l_Res = I2PAddress.TryCreateFromBufferLegayFormat(a_Buffer, a_OldAddrFormat, out l_UserCntxt.f_DatagramAddr);

                if (BufferIO.LoadResult.Corruption == l_Res)
                {
                    return BufferIO.LoadResult.Corruption;
                }
            }

            if (0 != (SaveFlags.GoteD2kUserID & l_Flags))
            {
                l_UserCntxt.f_UserID = new UserID(a_Buffer.ReadUInt128LE());
            }

            if (0 != (SaveFlags.GotRSAKey & l_Flags))
            {
                if (false == RsaPublicKey.FromBuffer(a_Buffer, out l_UserCntxt.f_RsaPubKey))
                {
                    return BufferIO.LoadResult.Corruption;
                }

                if (0 != (SaveFlags.KeyVerified & l_Flags))
                {
                    l_UserCntxt.f_ValidSignatureReceived = true;
                }
            }

            if (0 != (SaveFlags.LastUserName & l_Flags))
            {
                l_UserCntxt.f_LastUserName = a_Buffer.ReadUTF16LEStringEncLen(true, PeerInfo.c_MaxUserNameChars);
            }

            if (0 != (SaveFlags.IsFriend & l_Flags))
            {
                l_UserCntxt.f_Friend = true;
            }

            if (0 != (SaveFlags.HasA4AFLimitation & l_Flags))
            {
                l_UserCntxt.f_HasA4AFLimitation = true;
            }

            l_UserCntxt.f_LastConnectedTime = a_Buffer.ReadTime();
            TransferVolume.FromBuffer(a_Buffer, out l_UserCntxt.f_TransferVol);
            l_UserCntxt.f_Reputation = (int)a_Buffer.ReadUInt32LE();

            if (null == l_UserCntxt.f_StreamAddr)
            {
                return BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood;
            }

            a_UserCntxt = l_UserCntxt;

            return BufferIO.LoadResult.Success;
        }

        public void ToBuffer(Time a_Time,
                             BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt16LE(c_Magic);
            a_Buffer.WriteUInt64LE(f_UserNum.f_Val);
            f_StreamAddr.ToBufferOwnFormat(a_Buffer);

            SaveFlags l_Flags = SaveFlags.None;

            if (null != f_DatagramAddr)
            {
                l_Flags = SaveFlags.GotDatagramAddr;
            }

            if (null != f_UserID.f_Val)
            {
                l_Flags |= SaveFlags.GoteD2kUserID;
            }

            if (true == f_RsaPubKey.IsValid())
            {
                l_Flags |= SaveFlags.GotRSAKey;

                if (true == f_ValidSignatureReceived)
                {
                    l_Flags |= SaveFlags.KeyVerified;
                }
            }

            if (null != f_LastUserName)
            {
                l_Flags |= SaveFlags.LastUserName;
            }

            if (true == f_Friend)
            {
                l_Flags |= SaveFlags.IsFriend;
            }

            if (true == f_HasA4AFLimitation)
            {
                l_Flags |= SaveFlags.HasA4AFLimitation;
            }

            a_Buffer.WriteUInt16LE((ushort)l_Flags);

            if (null != f_DatagramAddr)
            {
                f_DatagramAddr.ToBufferOwnFormat(a_Buffer);
            }

            if (null != f_UserID.f_Val)
            {
                a_Buffer.WriteUInt128LE(f_UserID.f_Val);
            }

            if (true == f_RsaPubKey.IsValid())
            {
                f_RsaPubKey.ToBuffer(a_Buffer);
            }

            if (null != f_LastUserName)
            {
                a_Buffer.WriteUTF16LEStringEncLen(f_LastUserName);
            }

            a_Buffer.WriteTime(GetLastConnectedTime(a_Time));
            f_TransferVol.ToBuffer(a_Buffer);
            a_Buffer.WriteUInt32LE((uint)f_Reputation);
        }

        public bool HasAnyCurrentSecIdentAuthenticatedConnection(out uint a_AuthKeyStrength)
        {
            // Returns a nonzero value if the peer is considered
            // securely identified regarding any current connection.

            if (null != f_EstablishedConns)
            {
                foreach (DonkeyClient l_Clnt in f_EstablishedConns.Keys)
                {
                    if (true == l_Clnt.SuccessfullyIdentifiedPeerWithSecIdent())
                    {
                        a_AuthKeyStrength = f_RsaPubKey.GetBitSize();

                        return true;
                    }
                }
            }

            a_AuthKeyStrength = 0;

            return false;
        }

        private Time GetLastConnectedTime(Time a_Time)
        {
            if (null != f_EstablishedConns)
            {
                // We are currently connected.

                return a_Time;
            }

            return f_LastConnectedTime;
        }

        public bool IsSaveWorthy()
        {
            if (null != f_EstablishedConns ||
                true == f_Friend)
            {
                return true;
            }

            if (0 != f_Reputation ||
                0 != f_TransferVol.p_BytesReceived ||
                0 != f_TransferVol.p_BytesSent)
            {
                return true;
            }

            return false;
        }

        public void DonateAttributes(UserContext a_DoneeUserCntxt)
        {
            // Only three attributes are donated: f_TransferVol,
            // f_Reputation and f_Friend.

            a_DoneeUserCntxt.f_TransferVol.Add(this.f_TransferVol);
            a_DoneeUserCntxt.AddToReputation(this.f_Reputation);

            if (true == this.f_Friend)
            {
                a_DoneeUserCntxt.f_Friend = true;
            }

            a_DoneeUserCntxt.f_PriorityUpToDate = false;
        }

        public bool MayPublish()
        {
            return (c_MayPublishReputationThreshold <= f_Reputation);
        }

        public void NullifyReputationIfNegative()
        {
            if (0 > f_Reputation)
            {
                f_Reputation = 0;
                f_PriorityUpToDate = false;
            }
        }

        public void SetReputation(int a_Reputation)
        {
            if (a_Reputation != f_Reputation)
            {
                f_Reputation = a_Reputation;
                f_PriorityUpToDate = false;
            }
        }

        public void AddToReputation(int a_ReputationDelta)
        {
            // This routine is overflow safe.

            int l_Sum = (f_Reputation + a_ReputationDelta);
            int l_Rep;

            if (0 > a_ReputationDelta)
            {
                l_Rep = (f_Reputation >= l_Sum ? l_Sum : int.MinValue);
            }
            else
            {
                l_Rep = (f_Reputation <= l_Sum ? l_Sum : int.MaxValue);
            }

            SetReputation(l_Rep);
        }

        private void UpdatePriority()
        {
            int l_Boosts = f_TransferVol.ToBoosts();
            l_Boosts += ReputationToBoosts(f_Reputation);

            if (true == f_Friend)
            {
                l_Boosts++;
            }

            f_PriorityUpToDate = true;
            f_Priority = Priority.FromBoosts(l_Boosts);
        }

        private static int ReputationToBoosts(int a_Reputation)
        {
            if (1024 <= a_Reputation)
            {
                if ((64 * 1024) <= a_Reputation)
                {
                    return 2;
                }

                return 1;
            }

            if (0 > a_Reputation)
            {
                if (-(10 * 1024) >= a_Reputation)
                {
                    return (-2);
                }

                return (-1);
            }

            return 0;
        }

        public void UpdateUlQueueCompareWeight(Time a_Time,
                                               bool a_Force)
        {
            if (false == a_Force &&
                a_Time < f_NextCompareWeightUpdateTime)
            {
                return;
            }

            f_NextCompareWeightUpdateTime = a_Time.AddSeconds(Time.c_SecsPerHalfMinute);
            f_UlQueueCompareWeight = BuildUlQueueCompareWeightCore();
        }

        private int BuildUlQueueCompareWeightCore()
        {
            // Returns a value suitable for sorting.

            long l_Res = f_TransferVol.AsCompareWeight();
            l_Res += (4L * f_Reputation);

            if (true == f_Friend)
            {
                l_Res += (Math.Abs(Math.Max(long.MinValue + 1, l_Res)) / 20);
            }

            if (0 > l_Res)
            {
                return (int)Math.Max(int.MinValue, l_Res);
            }

            if (0 > f_Reputation)
            {
                // Never return positive if the reputation is negative.

                return 0;
            }

            return (int)Math.Min(int.MaxValue, l_Res);
        }

        public uint GetMinExpectedSignatureLength()
        {
            return f_RsaPubKey.GetExpectedSignatureLength();
        }

        public bool HasDatagramAddr()
        {
            return (null != f_DatagramAddr);
        }

        public bool HasSameDatagramAddr(I2PAddress a_DatagramAddr)
        {
            return (null != f_DatagramAddr &&
                    true == f_DatagramAddr.ValueEquals(a_DatagramAddr));
        }

        public void SetDatagramAddr(I2PAddress a_DatagramAddr)
        {
            // May only be used by the UserManager class.

            f_DatagramAddr = a_DatagramAddr;
        }

        public bool HasUserID()
        {
            return (null != f_UserID.f_Val);
        }

        public bool HasSameUserID(UserID a_UserID)
        {
            return (null != f_UserID.f_Val &&
                    true == f_UserID.f_Val.ValueEquals(a_UserID.f_Val));
        }

        public void SetUserID(UserID a_UserID)
        {
            // May only be used by the UserManager class.

            f_UserID = a_UserID;
        }

        public bool HasPublicKey()
        {
            return f_RsaPubKey.IsValid();
        }

        public bool HasPublicKeyWithMinimumSize(uint a_MinBits)
        {
            return (true == f_RsaPubKey.IsValid() &&
                    a_MinBits <= f_RsaPubKey.GetBitSize());
        }

        public bool HasSamePublicKey(UserContext a_Other)
        {
            return (true == a_Other.f_RsaPubKey.IsValid() &&
                    true == this.f_RsaPubKey.ValueEquals(a_Other.f_RsaPubKey));
        }

        public bool HasThisPublicKey(RsaPublicKey a_RsaPubKey)
        {
            return this.f_RsaPubKey.ValueEquals(a_RsaPubKey);
        }

        public RsaPublicKey ExposePublicKey()
        {
            return f_RsaPubKey;
        }

        public bool CalcPublicKeyHash(out uint a_PubKeyHash)
        {
            if (false == f_RsaPubKey.IsValid())
            {
                a_PubKeyHash = 0;

                return false;
            }

            a_PubKeyHash = f_RsaPubKey.CalcHashCode();

            return true;
        }

        public void SetPublicKey(RsaPublicKey a_RsaPubKey)
        {
            // May only be used by the UserManager class.

            f_RsaPubKey = a_RsaPubKey;
        }

        public bool VerifySignature(ref RsaPublicKey.SignatureVerificationContext a_VerCntxt)
        {
            // May only be used by the UserManager class.

            return f_RsaPubKey.VerifySignature(ref a_VerCntxt);
        }

        public void SetValidSignatureReceived()
        {
            f_ValidSignatureReceived = true;
        }

        public bool GetLastUserName(out string a_UserName)
        {
            a_UserName = f_LastUserName;

            return (null != a_UserName);
        }

        public void NotifyUserName(string a_UserName)
        {
            f_LastUserName = (0 == a_UserName.Length ? null : a_UserName);
        }

        public bool IsConnected(bool a_IncludePendingConns)
        {
            return (null != f_EstablishedConns ||
                    (true == a_IncludePendingConns &&
                     0 != f_OutgoingConns));
        }

        public IEnumerable<DonkeyClient> Connections()
        {
            return f_EstablishedConns.Values;
        }

        public void RegisterPendingConn()
        {
            f_OutgoingConns++;
        }

        public void UnregisterPendingConn()
        {
            f_OutgoingConns--;
        }

        public void RegisterEstabConn(DonkeyClient a_Client)
        {
            // May only be used by the UserManager class.

            if (null == f_EstablishedConns)
            {
                f_EstablishedConns = new Dictionary<DonkeyClient, DonkeyClient>(1);
            }

            f_EstablishedConns.Add(a_Client, a_Client);

            if (false == a_Client.p_Outgoing) f_IngoingConns++;
            else f_OutgoingConns++;
        }

        public void UnregisterEstabConn(Time a_Time,
                                        DonkeyClient a_Client)
        {
            // May only be used by the UserManager class.

            bool l_OK = f_EstablishedConns.Remove(a_Client);

            if (0 == f_EstablishedConns.Count)
            {
                f_EstablishedConns = null;
                f_LastConnectedTime = a_Time;
            }

            if (false == a_Client.p_Outgoing) f_IngoingConns--;
            else f_OutgoingConns--;
        }

        public bool CanMakeAnotherConnAttempt(ushort a_MaxOutgoingConns)
        {
            // Do not make more than one outgoing connection attempt
            // at a time, except if there are established conns.

            return (a_MaxOutgoingConns > f_OutgoingConns &&
                    (0 == f_OutgoingConns ||
                     null != f_EstablishedConns));
        }

        public bool IsFriend()
        {
            return f_Friend;
        }

        public void GuiAdjustFriendStatus(bool a_Set)
        {
            // This routine may be used directly by any thread.

            f_Friend = a_Set;
        }

        public bool HasA4AFLimitation()
        {
            return f_HasA4AFLimitation;
        }

        public void AdjustA4AFState(bool a_Set)
        {
            f_HasA4AFLimitation = a_Set;
        }

        public bool PreventsA4AFCreatingAnotherDlContext(bool a_IncludeAttempts)
        {
            if (false == f_HasA4AFLimitation)
            {
                return false;
            }

            if (true == a_IncludeAttempts &&
                0 != f_OutgoingConns)
            {
                // NOTE: This works only if the single reason for
                // connecting to a A4AF limited client is to download.

                return true;
            }

            if (null != f_EstablishedConns)
            {
                foreach (DonkeyClient l_Clnt in f_EstablishedConns.Keys)
                {
                    if (null != l_Clnt.p_DlPeerCntxt)
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public bool IsBanned()
        {
            return f_Banned;
        }

        public void SetBanned(bool a_KillEstablishedConns)
        {
            f_Banned = true;

            if (true == a_KillEstablishedConns &&
                null != f_EstablishedConns)
            {
                foreach (DonkeyClient l_Client in f_EstablishedConns.Values)
                {
                    l_Client.Close();
                }
            }
        }

        public void SetUnbanned()
        {
            f_Banned = false;
        }

        public void CopyTransferVolumeVolatile(out TransferVolume a_Dest)
        {
            f_TransferVol.CopyToVolatile(out a_Dest);
        }

        public void Sent(uint a_Size)
        {
            // Call this routine for data that was sent by us because
            // the peer requested it (so we can charge the user).

            f_TransferVol.Sent(a_Size);
            f_PriorityUpToDate = false;
        }

        public void Rcvd(Packet a_Packet)
        {
            // Call this routine with packets that were sent to us on
            // our own request (so that we get charged).

            f_TransferVol.Received(a_Packet.f_TransferSize);
            f_PriorityUpToDate = false;
        }

        public void GetInfoForGeneralPeersPanelVolatile(out GeneralPeersPanel.PeerInfo a_Dest)
        {
            // Calculate the connection count from volatile data.

            Dictionary<DonkeyClient, DonkeyClient> l_EstabDic = f_EstablishedConns;
            ushort l_IngoingEstabConns = f_IngoingConns;
            Util.MemoryBarrier();

            bool l_Connected = (null != l_EstabDic);
            ushort l_OutgoingEstabConns;

            if (false == l_Connected)
            {
                l_IngoingEstabConns = 0;
                l_OutgoingEstabConns = 0;
            }
            else
            {
                ushort l_EstabCount = (ushort)l_EstabDic.Count;

                if (l_EstabCount <= l_IngoingEstabConns)
                {
                    if (0 == l_IngoingEstabConns)
                    {
                        l_Connected = false;
                    }

                    l_OutgoingEstabConns = 0;
                }
                else
                {
                    l_OutgoingEstabConns = (ushort)(l_EstabCount - l_IngoingEstabConns);
                }
            }

            // Mostly initialize using volatile data right away.
            RsaPublicKey l_RsaPubKey = (false == f_ValidSignatureReceived ? new RsaPublicKey() : f_RsaPubKey);
            a_Dest = new GeneralPeersPanel.PeerInfo(l_RsaPubKey, f_UserID, f_LastUserName, f_Friend, l_Connected, l_IngoingEstabConns, l_OutgoingEstabConns, f_LastConnectedTime, ref f_TransferVol, f_Reputation, f_Banned);
            Util.MemoryBarrier();
        }

        public static int CompareUserContext(ref UserContext a_A,
                                             ref UserContext a_B)
        {
            // Returns a positive value if a_A is more keep-worthy
            // than a_B.

            int l_Res = a_A.f_Friend.CompareTo(a_B.f_Friend);

            if (0 != l_Res)
            {
                return l_Res;
            }

            l_Res = a_A.f_Reputation.CompareTo(a_B.f_Reputation);

            if (0 != l_Res)
            {
                return l_Res;
            }

            l_Res = a_A.f_TransferVol.p_Sum.CompareTo(a_B.f_TransferVol.p_Sum);

            if (0 != l_Res)
            {
                return l_Res;
            }

            return a_A.f_LastConnectedTime.CompareTo(a_B.f_LastConnectedTime);
        }

        public void SetLastAcceptedForUlTime(Time a_Time)
        {
            f_LastAcceptedForUlTime = a_Time;
        }

        public static int CompareUserContextForGui(ref UserContext a_A,
                                                   ref UserContext a_B)
        {
            // Returns a positive value if a_B is more keep-worthy
            // than a_A.

            int l_Res = a_B.p_TotalConns.CompareTo(a_A.p_TotalConns);

            if (0 != l_Res)
            {
                return l_Res;
            }

            return CompareUserContext(ref a_B, ref a_A);
        }

        [Flags]
        private enum SaveFlags : ushort
        {
            None = 0,
            GotDatagramAddr = 0x1,
            GoteD2kUserID = 0x2,
            GotRSAKey = 0x4,
            KeyVerified = 0x8,
            IsFriend = 0x10,
            // LastUserName exists from revision 1 on.
            LastUserName = 0x20,
            // HasA4AFLimitation exists from v1.4.2 on (introduced late
            // in revision 1).
            HasA4AFLimitation = 0x40,
        }
    }

    class UserManager
    {
        private const uint c_UserCntxtsMagic = 0xA91ACB99;
        private const uint c_UserCntxtsHeaderSize = 512;

        public readonly Instance f_Inst;

        // f_UsersByNum and f_UsersByStreamAddr contains all users.
        private readonly Dictionary<UserNum, UserContext> f_UsersByNum;
        private readonly Dictionary<I2PAddress, UserContext> f_UsersByStreamAddr;

        // Collections with special subsets of the user contexts.
        // NOTE: There can be several users with the same alleged
        // datagram dest and public key.
        private readonly MultiMap<I2PAddress, UserContext> f_UsersByDatagramAddr;
        private readonly MultiMap<uint, UserContext> f_UsersByPublicKey;
        private readonly Dictionary<UserContext, UserContext> f_ConnectedUsers;

        private ulong f_NextUserNum;

        public UserContext[] BuildArrayForGui(uint a_SuggestedLimit,
                                              out uint a_ConnectedUsers)
        {
            // The returned array can contain more entries than
            // a_SuggestedLimit. Return can be null, if there are no
            // contexts.

            uint l_Count = (uint)f_UsersByNum.Count;

            if (0 == l_Count)
            {
                goto FAIL;
            }

            uint l_ConnUsers = (uint)f_ConnectedUsers.Count;
            uint l_EffLimit = Math.Max(l_ConnUsers, a_SuggestedLimit);

            if (0 == l_EffLimit)
            {
                goto FAIL;
            }

            FlexArray<UserContext> l_Ls;
            FlexArray<UserContext>.Create((int)l_Count, out l_Ls);
            uint l_EffCount = 0;

            foreach (UserContext l_UserCntxt in f_UsersByNum.Values)
            {
                if (true == l_UserCntxt.IsSaveWorthy())
                {
                    l_Ls.ReserveAtEnd();
                    l_Ls.p_Arr[l_EffCount++] = l_UserCntxt;
                }
            }

            if (l_EffCount > l_EffLimit)
            {
                l_Ls.Sort(UserContext.CompareUserContextForGui);
                l_EffCount = l_EffLimit;
            }

            a_ConnectedUsers = l_ConnUsers;
            UserContext[] l_Arr = l_Ls.ToArrayWithReset((int)l_EffCount);

            return l_Arr;

            FAIL:

            a_ConnectedUsers = 0;

            return null;
        }

        public UserManager(Instance a_Inst)
        {
            f_Inst = a_Inst;

            f_UsersByNum = new Dictionary<UserNum, UserContext>();
            f_UsersByStreamAddr = new Dictionary<I2PAddress, UserContext>();

            MultiMap<I2PAddress, UserContext>.Create(out f_UsersByDatagramAddr);
            MultiMap<uint, UserContext>.Create(out f_UsersByPublicKey);
            f_ConnectedUsers = new Dictionary<UserContext, UserContext>();

            StorageStates l_State = LoadContexts();

            if (StorageStates.Corruption == l_State)
            {
                Debugger.Break();
            }
        }

        private void RemoveAllContexts()
        {
            f_UsersByNum.Clear();
            f_UsersByStreamAddr.Clear();

            f_UsersByDatagramAddr.Clear();
            f_ConnectedUsers.Clear();

            f_NextUserNum = UserNum.c_ReservedUserNums;
        }

        private UserNum GetNextUserNum()
        {
            while (true)
            {
                if (UserNum.c_ReservedUserNums > f_NextUserNum)
                {
                    f_NextUserNum = UserNum.c_ReservedUserNums;
                }

                if (false == f_UsersByNum.ContainsKey(new UserNum(f_NextUserNum)))
                {
                    return new UserNum(f_NextUserNum++);
                }

                f_NextUserNum++;
            }
        }

        public IEnumerable<UserContext> ConnectedUsers()
        {
            // Returns users that have at least one established
            // connection.

            return f_UsersByStreamAddr.Values;
        }

        public UserContext GetUserContext(UserNum a_UserNum)
        {
            // Return can be null.

            UserContext l_DestCntxt;
            if (true == f_UsersByNum.TryGetValue(a_UserNum, out l_DestCntxt))
            {
                return l_DestCntxt;
            }

            return null;
        }

        public bool TryGetUserContext(I2PAddress a_StreamAddr,
                                      out UserContext a_UserCntxt)
        {
            return f_UsersByStreamAddr.TryGetValue(a_StreamAddr, out a_UserCntxt);
        }

        public UserContext GetOrCreateUserContext(I2PAddress a_StreamAddr)
        {
            UserContext l_UserCntxt;
            if (false == f_UsersByStreamAddr.TryGetValue(a_StreamAddr, out l_UserCntxt))
            {
                l_UserCntxt = UserContext.CreateNew(GetNextUserNum(), a_StreamAddr);
                f_UsersByNum.Add(l_UserCntxt.f_UserNum, l_UserCntxt);
                f_UsersByStreamAddr.Add(a_StreamAddr, l_UserCntxt);
            }

            return l_UserCntxt;
        }

        public void AddConnection(DonkeyClient a_Clnt)
        {
            UserContext l_UserCntxt = a_Clnt.p_UserCntxt;

            if (false == l_UserCntxt.IsConnected(false))
            {
                f_ConnectedUsers.Add(l_UserCntxt, l_UserCntxt);
            }

            l_UserCntxt.RegisterEstabConn(a_Clnt);
        }

        public void RemoveConnection(DonkeyClient a_Clnt)
        {
            UserContext l_UserCntxt = a_Clnt.p_UserCntxt;

            l_UserCntxt.UnregisterEstabConn(f_Inst.GetTime(), a_Clnt);

            if (false == l_UserCntxt.IsConnected(false))
            {
                bool l_OK = f_ConnectedUsers.Remove(l_UserCntxt);
            }
        }

        public void ResetBans()
        {
            // This routine may be used even when there's no connection.
            // Does only reset the user context ban (not bridge).

            foreach (UserContext l_UserCntxt in f_UsersByNum.Values)
            {
                if (true == l_UserCntxt.IsBanned())
                {
                    l_UserCntxt.SetUnbanned();
                }
            }
        }

        public void BanOrUnbanUsers(UserContext[] a_Users,
                                    bool a_Ban)
        {
            // This routine may be used even when there's no connection.
            // The items of a_Users are treated as untrusted (it is not
            // relied on them being part of f_UsersByNum still).

            foreach (UserContext l_UserCntxt in a_Users)
            {
                if (l_UserCntxt.IsBanned() == a_Ban)
                {
                    continue;
                }

                if (l_UserCntxt != f_UsersByNum[l_UserCntxt.f_UserNum])
                {
                    // This user context was removed in the meantime.

                    continue;
                }

                if (false == a_Ban)
                {
                    f_Inst.f_NetworkMngr.UnbanAddress(l_UserCntxt.f_StreamAddr);
                    l_UserCntxt.SetUnbanned();
                }
                else
                {
                    f_Inst.f_NetworkMngr.BanAddress(l_UserCntxt.f_StreamAddr);
                }
            }
        }

        public bool TrySetUserID(UserContext a_UserCntxt,
                              UserID a_UserID)
        {
            if (false == a_UserCntxt.HasUserID())
            {
                if (true == f_Inst.f_NetworkMngr.p_OwnPersona.IsMyUserID(a_UserID))
                {
                    // An impostor.

                    return false;
                }

                a_UserCntxt.SetUserID(a_UserID);
            }
            else
            {
                if (false == a_UserCntxt.p_UserID.f_Val.Equals(a_UserID.f_Val))
                {
                    // We don't allow the user id to be changed once
                    // we were told it.

                    return false;
                }
            }

            return true;
        }

        public void SetDatagramDest(UserContext a_UserCntxt,
                                    I2PAddress a_DatagramAddr)
        {
            if (true == a_UserCntxt.HasSameDatagramAddr(a_DatagramAddr))
            {
                return;
            }

            I2PAddress l_OldAddr = a_UserCntxt.p_DatagramAddr;

            if (null != l_OldAddr)
            {
                bool l_Removed = f_UsersByDatagramAddr.Remove(l_OldAddr, a_UserCntxt);
            }

            a_UserCntxt.SetDatagramAddr(a_DatagramAddr);
            bool l_Added = f_UsersByDatagramAddr.Add(a_DatagramAddr, a_UserCntxt);
        }

        public bool SetPublicKey(UserContext a_UserCntxt,
                                 RsaPublicKey a_RsaPubKey)
        {
            if (true == a_UserCntxt.HasPublicKey())
            {
                return a_UserCntxt.HasThisPublicKey(a_RsaPubKey);
            }

            a_UserCntxt.SetPublicKey(a_RsaPubKey);
            bool l_Added = f_UsersByPublicKey.Add(a_RsaPubKey.CalcHashCode(), a_UserCntxt);

            return true;
        }

        public bool VerifySignature(UserContext a_UserCntxt,
                                    ref RsaPublicKey.SignatureVerificationContext a_VerCntxt)
        {
            if (false == a_UserCntxt.VerifySignature(ref a_VerCntxt))
            {
                // If this fails it can be that the remote peer used
                // a different key for signing than we belived. This
                // happens when a known peer changes his/her key but
                // doesn't change the I2P address as well, so that we
                // silently assume we still know the public key, while
                // we in fact don't. ALWAYS change the address when
                // changing the key.

                return false;
            }

            a_UserCntxt.SetValidSignatureReceived();

            if (false == a_UserCntxt.HasUserID())
            {
                // We demand that the user id is present.

                return false;
            }

            UserID l_UserID = a_UserCntxt.p_UserID;
            uint l_PubKeyHash;
            bool l_OK = a_UserCntxt.CalcPublicKeyHash(out l_PubKeyHash);

            MultiMap<uint, UserContext>.Enumerator l_Enum;
            if (true == f_UsersByPublicKey.CreateEnumerator(l_PubKeyHash, out l_Enum))
            {
                UserContext l_OldUserCntxt;
                while (true == l_Enum.GetNext(out l_OldUserCntxt))
                {
                    if (a_UserCntxt != l_OldUserCntxt &&
                        true == a_UserCntxt.HasSamePublicKey(l_OldUserCntxt))
                    {
                        // There is already a user with that exact
                        // public key. So we assume that someone has
                        // changed the stream address, but kept the
                        // user id and public key. We will copy important
                        // values from the old user context into the
                        // new user context.

                        if (false == l_OldUserCntxt.HasSameUserID(l_UserID) ||
                            true == l_OldUserCntxt.IsBanned() ||
                            true == l_OldUserCntxt.IsConnected(false))
                        {
                            // The user id must stay the same and the
                            // old context must not be banned or in use.

                            return false;
                        }

                        if (0 != l_OldUserCntxt.p_OutgoingConns)
                        {
                            // Abort any pending connection attempts
                            // concerning the old dest.

                            f_Inst.f_NetworkMngr.AbortPendingConnectionAttempts(l_OldUserCntxt);
                        }

                        l_OldUserCntxt.DonateAttributes(a_UserCntxt);
                        RemoveUserContext(l_OldUserCntxt);

                        break;
                    }
                }
            }

            return true;
        }

        public bool MayDatagramDestinationPublish(I2PAddress a_DatagramAddr)
        {
            // Since there can be several peers claiming to have that
            // address, just see if any of them has a reputation high
            // enough.

            MultiMap<I2PAddress, UserContext>.Enumerator l_Enum;
            if (true == f_UsersByDatagramAddr.CreateEnumerator(a_DatagramAddr, out l_Enum))
            {
                UserContext l_UserCntxt;
                while (true == l_Enum.GetNext(out l_UserCntxt))
                {
                    if (true == l_UserCntxt.MayPublish())
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public void ContentMngrChangeReputation(GeneralAPCCookie a_Cookie)
        {
            UserRepChange[] l_Users = (UserRepChange[])a_Cookie.GetArg();
            for (int i = (l_Users.Length - 1); 0 <= i; i--)
            {
                UserContext l_Cntxt;
                if (true == f_UsersByNum.TryGetValue(l_Users[i].f_UserNum, out l_Cntxt))
                {
                    l_Cntxt.AddToReputation(l_Users[i].f_ReputationDelta);
                }
            }

            a_Cookie.SetState(CookieStates.Success);
        }

        public void ContentMngrQueryUserInfos(GeneralAPCCookie a_Cookie)
        {
            QueryUserInfosCookie l_Cookie = (QueryUserInfosCookie)a_Cookie;
            UserInfo[] l_Infos = l_Cookie.f_Infos;

            for (int i = (l_Infos.Length - 1); 0 <= i; i--)
            {
                UserContext l_Cntxt;
                if (true == f_UsersByNum.TryGetValue(l_Infos[i].f_UserNum, out l_Cntxt))
                {
                    l_Infos[i].f_KnownUser = true;
                    l_Infos[i].f_Reputation = l_Cntxt.p_Reputation;
                }
            }

            l_Cookie.f_QueryTime = f_Inst.GetTime();
            l_Cookie.SetState(CookieStates.Success);
        }

        public void PrepareSavingOnShutdown()
        {
            // Only the NetworkManager class may call this routine from
            // outside.

            int l_MaxUserCntxts = (10 * 1000);
            int l_Count = f_UsersByNum.Count;

            if (l_MaxUserCntxts >= l_Count)
            {
                return;
            }

            int l_EffCount = 0;
            FlexArray<UserContext> l_Ls;
            FlexArray<UserContext>.Create(l_Count, out l_Ls);

            foreach (UserContext l_UserCntxt in f_UsersByNum.Values)
            {
                if (true == l_UserCntxt.IsSaveWorthy())
                {
                    l_Ls.ReserveAtEnd();
                    l_Ls.p_Arr[l_EffCount++] = l_UserCntxt;
                }
            }

            if (l_MaxUserCntxts >= l_EffCount)
            {
                return;
            }

            l_Ls.Sort(UserContext.CompareUserContext);

            // Remove all (least important) items at the start.

            int l_Excess = (l_EffCount - l_MaxUserCntxts);
            UserContext[] l_Arr = l_Ls.p_Arr;

            for (int i = 0; i < l_Excess; i++)
            {
                RemoveUserContext(l_Arr[i]);
            }
        }

        public StorageStates SaveContexts()
        {
            // Only the NetworkManager class may call this routine from
            // outside.

            return f_Inst.WriteToElement(new ElmKey(Instance.c_UserCntxtsElmKey), SaveContextsCore);
        }

        private StorageStates SaveContextsCore(BufferIO a_Buffer)
        {
            BufferIO.CountWriter l_CountWtr;
            Time l_Time = f_Inst.GetTime();

            a_Buffer.WriteUInt32LE(c_UserCntxtsMagic);
            BufferIO.CountWriter.InitAndMoveBuffer(a_Buffer, sizeof(uint), out l_CountWtr);
            a_Buffer.WriteUInt16LE(UserContext.c_CurRevision);
            a_Buffer.SetPosition(c_UserCntxtsHeaderSize);

            uint l_Count = 0;

            foreach (UserContext l_UserCntxt in f_UsersByNum.Values)
            {
                if (true == l_UserCntxt.IsSaveWorthy())
                {
                    l_UserCntxt.ToBuffer(l_Time, a_Buffer);
                    l_Count++;
                }
            }

            l_CountWtr.Finish(l_Count);

            return StorageStates.Success;
        }

        private StorageStates LoadContexts()
        {
            return f_Inst.ReadFromElement(new ElmKey(Instance.c_UserCntxtsElmKey), LoadContextsCore);
        }

        private StorageStates LoadContextsCore(BufferIO a_Buffer)
        {
            if (c_UserCntxtsHeaderSize > a_Buffer.GetRemaining() ||
                c_UserCntxtsMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            uint l_Count = a_Buffer.ReadUInt32LE();
            ushort l_Revision = a_Buffer.ReadUInt16LE();

            if (UserContext.c_CurRevision < l_Revision)
            {
                return StorageStates.Corruption;
            }

            a_Buffer.SetPosition(c_UserCntxtsHeaderSize);
            bool l_OldAddrFormat = (1 >= l_Revision);

            for (uint i = 0; i < l_Count; i++)
            {
                BufferIO.LoadResult l_Res;
                UserContext l_UserCntxt;

                try
                {
                    l_Res = UserContext.FromBuffer(a_Buffer, l_OldAddrFormat, out l_UserCntxt);
                }
                catch (BufferIO.InterpretBufferException)
                {
                    l_Res = BufferIO.LoadResult.Corruption;
                    l_UserCntxt = null;
                }

                switch (l_Res)
                {
                    case BufferIO.LoadResult.Success:
                        break;

                    case BufferIO.LoadResult.InvalidDataButBufferPosIsStillGood:
                        continue;

                    case BufferIO.LoadResult.Corruption:
                        goto HANDLE_CORRUPTION;

                    default: throw (new ApplicationException());
                }

                try
                {
                    f_UsersByNum.Add(l_UserCntxt.f_UserNum, l_UserCntxt);
                    f_UsersByStreamAddr.Add(l_UserCntxt.f_StreamAddr, l_UserCntxt);
                }
                catch (ArgumentException)
                {
                    // Key already exists.

                    goto HANDLE_CORRUPTION;
                }

                I2PAddress l_DatagramAddr = l_UserCntxt.p_DatagramAddr;
                if (null != l_DatagramAddr)
                {
                    bool l_Added = f_UsersByDatagramAddr.Add(l_DatagramAddr, l_UserCntxt);
                }

                uint l_PubKeyHash;
                if (true == l_UserCntxt.CalcPublicKeyHash(out l_PubKeyHash))
                {
                    bool l_Added = f_UsersByPublicKey.Add(l_PubKeyHash, l_UserCntxt);
                }

                f_NextUserNum = Math.Max(f_NextUserNum, l_UserCntxt.f_UserNum.f_Val);
            }

            if (false == a_Buffer.EndReached())
            {
                goto HANDLE_CORRUPTION;
            }

            if (0 != l_Count)
            {
                f_NextUserNum++;
            }

            return StorageStates.Success;

            HANDLE_CORRUPTION:

            Util.DebugBreak();

            RemoveAllContexts();

            return StorageStates.Corruption;
        }

        private void RemoveUserContext(UserContext a_UserCntxt)
        {
            bool l_Removed = f_UsersByNum.Remove(a_UserCntxt.f_UserNum);

            l_Removed = f_UsersByStreamAddr.Remove(a_UserCntxt.f_StreamAddr);

            I2PAddress l_DatagramAddr = a_UserCntxt.p_DatagramAddr;
            if (null != l_DatagramAddr)
            {
                l_Removed = f_UsersByDatagramAddr.Remove(l_DatagramAddr, a_UserCntxt);
            }

            uint l_PubKeyHash;
            if (true == a_UserCntxt.CalcPublicKeyHash(out l_PubKeyHash))
            {
                l_Removed = f_UsersByPublicKey.Remove(l_PubKeyHash, a_UserCntxt);
            }
        }

        public class QueryUserInfosCookie : GeneralAPCCookie
        {
            public readonly UserInfo[] f_Infos;
            public Time f_QueryTime;

            public QueryUserInfosCookie(UserManager a_UserMngr,
                                        UserInfo[] a_UserInfos) :
                                   base(a_UserMngr.ContentMngrQueryUserInfos)
            {
                f_Infos = a_UserInfos;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct UserInfo
        {
            public readonly UserNum f_UserNum;

            public bool f_KnownUser;
            public int f_Reputation;

            public UserInfo(UserNum a_UserNum)
            {
                f_UserNum = a_UserNum;

                f_KnownUser = false;
                f_Reputation = 0;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct UserRepChange
        {
            public readonly UserNum f_UserNum;
            public int f_ReputationDelta;

            public UserRepChange(UserNum a_UserNum,
                                 int a_ReputationDelta)
            {
                f_UserNum = a_UserNum;
                f_ReputationDelta = a_ReputationDelta;
            }
        }
    }

    class UpDownManager
    {
        public const byte c_TurnsPerSecond = 2;
        private const ushort c_Interval = (1000 / c_TurnsPerSecond);
        private const byte c_SecsPerTimelineElement = 4;

        private readonly Instance f_Inst;
        private readonly SamBridgeManager f_Bridge;

        private SpeedOMeter f_SpeedOMeter;
        private readonly uint f_MaxDatagramInQueueSize;

        private Time f_ConnectionStartingTime;
        private uint f_DatagramsReceived;
        private uint f_DatagramsSent;
        // This does not include SAM or I2P overhead.
        private TransferVolume f_TotalDatagramVolume;
        private TransferVolume f_TotalVolume;

        private Time f_LastTime;
        private Timeline<uint> f_DlSpeedTimeline;
        private Timeline<uint> f_UlSpeedTimeline;

        private Queue<Packet> f_DatagramOutQueue;
        private Queue<Packet> f_DatagramInQueue;

        private uint f_DatagramOutQueueSize;
        private uint f_DatagramInQueueSize;

        private uint f_PreconsumedDatagramOutBw;
        private uint f_PreconsumedDatagramInBw;

        public UpDownManager(Instance a_Inst)
        {
            f_Inst = a_Inst;
            f_Bridge = f_Inst.f_NetworkMngr.ExposeBridge();

            SpeedOMeter.Create(out f_SpeedOMeter);
            f_MaxDatagramInQueueSize = (30 * 1024 * 1024);

            const byte lc_TimelineCapacity = 75;
            Timeline<uint>.Create(c_SecsPerTimelineElement * 10, lc_TimelineCapacity, out f_DlSpeedTimeline);
            Timeline<uint>.Create(c_SecsPerTimelineElement * 10, lc_TimelineCapacity, out f_UlSpeedTimeline);
            f_DatagramOutQueue = new Queue<Packet>();
            f_DatagramInQueue = new Queue<Packet>();
        }

        public void NotifyNewSamConnection(Time a_Time)
        {
            f_ConnectionStartingTime = a_Time;
        }

        public void NotifySamConnectionBroken()
        {
            f_ConnectionStartingTime = new Time();
            f_DatagramsReceived = 0;
            f_DatagramsSent = 0;
            f_TotalDatagramVolume = new TransferVolume();
            f_TotalVolume = new TransferVolume();

            f_DatagramOutQueue.Clear();
            f_DatagramInQueue.Clear();

            f_DatagramOutQueueSize = 0;
            f_DatagramInQueueSize = 0;

            f_PreconsumedDatagramOutBw = 0;
            f_PreconsumedDatagramInBw = 0;
        }

        public void RetrieveStats(Time a_Time,
                                  ref NetworkManager.StatsBuilder a_Stats)
        {
            MainForm l_Form = f_Inst.f_MainForm;

            long l_ConnUpsecs;
            if (false == f_Inst.f_UpDwnMngr.GetConnectionUptime(a_Time, out l_ConnUpsecs))
            {
                a_Stats.AddText(TextKeys.StatsTabI2PConnUptime, TextKeys.NotAvailable);
            }
            else
            {
                a_Stats.AddTime(TextKeys.StatsTabI2PConnUptime, l_ConnUpsecs);
            }

            a_Stats.AddInteger(TextKeys.StatsTabI2PUnrecognizedDgsRcvd, f_Bridge.p_MalformedDatagrams);
            a_Stats.AddInteger(TextKeys.StatsTabI2PRecognizedDgsRcvd, f_DatagramsReceived);
            a_Stats.AddInteger(TextKeys.StatsTabI2PDgsSent, f_DatagramsSent);
            a_Stats.AddSize(TextKeys.StatsTabI2PDgBytesRcvd, f_TotalDatagramVolume.p_BytesReceived);
            a_Stats.AddSize(TextKeys.StatsTabI2PDgBytesSent, f_TotalDatagramVolume.p_BytesSent);
            a_Stats.AddSize(TextKeys.StatsTabI2PTotalBytesRcvd, f_TotalVolume.p_BytesReceived);
            a_Stats.AddSize(TextKeys.StatsTabI2PTotalBytesSent, f_TotalVolume.p_BytesSent);
            a_Stats.AddInteger(TextKeys.StatsTabI2PInternedDests, I2PAddress.GetInternedAddressCount());
        }

        public bool GetConnectionUptime(Time a_Time,
                                        out long a_Secs)
        {
            if (false == f_ConnectionStartingTime.IsSet())
            {
                a_Secs = 0;

                return false;
            }

            a_Secs = a_Time.SecondsSince(f_ConnectionStartingTime);

            return true;
        }

        public Packet DequeueDatagram()
        {
            if (0 < f_DatagramInQueue.Count)
            {
                Packet l_Packet = f_DatagramInQueue.Dequeue();

                f_DatagramInQueueSize -= l_Packet.GetReceivedBodySize();

                return l_Packet;
            }

            return null;
        }

        public bool EnqueueDatagram(Packet a_Packet)
        {
            f_DatagramOutQueue.Enqueue(a_Packet);
            f_DatagramOutQueueSize += a_Packet.f_TransferSize;

            return true;
        }

        private Time f_NextSleepPreventionTime;

        public void Process(Time a_Time,
                            out Time a_NextUlTime)
        {
            if (f_NextSleepPreventionTime <= a_Time)
            {
                f_NextSleepPreventionTime = a_Time.AddSeconds(25);
                bool l_OK = Util.AdjustSystemSleepPrevention(true);
            }

            double l_Factor = 1d;

            if (true == f_LastTime.IsSet())
            {
                // We try to compensate for the small amout of time
                // we are off schedule.

                uint l_MsSince = (uint)a_Time.MillisecondsSince(f_LastTime);

                if ((2 * c_Interval) > l_MsSince)
                {
                    l_Factor = ((double)l_MsSince / c_Interval);
                }
            }

            Speed l_HistMaxDlSpeed = GetMaxSpeedFromTimeline(ref f_DlSpeedTimeline);
            Speed l_HistMaxUlSpeed = GetMaxSpeedFromTimeline(ref f_UlSpeedTimeline);
            uint l_DlBandwidth = GetBandwidthAim(l_HistMaxDlSpeed, f_Inst.f_NetworkMngr.f_NetworkPrefs.f_DlLimit, l_Factor);
            uint l_UlBandwidth = GetBandwidthAim(l_HistMaxUlSpeed, f_Inst.f_NetworkMngr.f_NetworkPrefs.f_UlLimit, l_Factor);

            // We reserve one fourth of our bandwidth for datagrams.
            uint l_ConsumedDlBw = ReceiveDatagrams(l_DlBandwidth / 4);
            uint l_ConsumedUlBw = SendDatagrams(l_UlBandwidth / 4);

            if (l_ConsumedDlBw < l_DlBandwidth)
            {
                List<ITransferable> l_Streams = f_Inst.f_NetworkMngr.GetStreamsForTransfer(false);
                l_ConsumedDlBw += ReceiveStreamPackets(l_DlBandwidth - l_ConsumedDlBw, l_Streams);

                // Whatever is still left might be used for datagrams again.
                if (l_ConsumedDlBw < l_DlBandwidth)
                {
                    l_ConsumedDlBw += ReceiveDatagrams(l_DlBandwidth - l_ConsumedDlBw);
                }
            }

            if (l_ConsumedUlBw < l_UlBandwidth)
            {
                List<ITransferable> l_Streams = f_Inst.f_NetworkMngr.GetStreamsForTransfer(true);
                l_ConsumedUlBw += SendStreamPackets(l_UlBandwidth - l_ConsumedUlBw, l_Streams);

                // Whatever is still left might be used for datagrams again.
                if (l_ConsumedUlBw < l_UlBandwidth)
                {
                    l_ConsumedUlBw += SendDatagrams(l_UlBandwidth - l_ConsumedUlBw);
                }
            }

            f_TotalVolume.Received(l_ConsumedDlBw);
            f_TotalVolume.Sent(l_ConsumedUlBw);
            f_SpeedOMeter.SaveProgress(a_Time, ref f_TotalVolume);

            f_LastTime = a_Time;
            UpdateTimeline(a_Time, l_ConsumedDlBw, ref f_DlSpeedTimeline);
            UpdateTimeline(a_Time, l_ConsumedUlBw, ref f_UlSpeedTimeline);

            a_NextUlTime = a_Time.AddMilliseconds(c_Interval - 50u);

            Speed l_DlSpeed;
            Speed l_UlSpeed;
            f_SpeedOMeter.GetSpeed(a_Time, out l_DlSpeed, out l_UlSpeed);

            MainForm.FormInput l_Input = f_Inst.f_MainForm.f_Input;
            Speed.CopyVolatile(ref l_DlSpeed, out l_Input.f_GlobalDlSpeed);
            Speed.CopyVolatile(ref l_UlSpeed, out l_Input.f_GlobalUlSpeed);
        }

        private static Speed GetMaxSpeedFromTimeline(ref Timeline<uint> a_SpeedTimeline)
        {
            // Returns the maximum speed of the passed timeline. The
            // return value can be Speed.s_UnknownSpeed.

            uint l_Len = a_SpeedTimeline.p_Count;

            if (1 >= l_Len)
            {
                // Not enough data.

                return Speed.s_UnknownSpeed;
            }

            // Do not take the most recent item into account, as it's
            // not ready yet.

            uint l_MaxValue = 0;
            uint l_StartingIdx;
            uint[] l_Arr = a_SpeedTimeline.ExposeArray(out l_StartingIdx);

            for (uint i = l_StartingIdx, k = 0; k != l_Len; k++)
            {
                if (++i >= a_SpeedTimeline.f_Capacity)
                {
                    i = 0;
                }

                l_MaxValue = Math.Max(l_MaxValue, l_Arr[i]);
            }

            return new Speed(l_MaxValue / c_SecsPerTimelineElement);
        }

        private static void UpdateTimeline(Time a_Time,
                                           uint a_ConsumedBandwidth,
                                           ref Timeline<uint> a_SpeedTimeline)
        {
            // Adds a_ConsumedBandwidth to a_SpeedTimeline.

            if (0 == a_ConsumedBandwidth)
            {
                return;
            }

            uint l_SlotIdx;
            Timeline<uint>.SlotResult l_Res = a_SpeedTimeline.TryCreateSlot(a_Time, out l_SlotIdx);

            if (Timeline<uint>.SlotResult.Outdated == l_Res)
            {
                return;
            }

            uint l_EffBandwidth;
            uint[] l_Arr = a_SpeedTimeline.ExposeArray();

            if (Timeline<uint>.SlotResult.ExistsAlready == l_Res)
            {
                l_EffBandwidth = (l_Arr[l_SlotIdx] + a_ConsumedBandwidth);

                if (l_EffBandwidth < a_ConsumedBandwidth)
                {
                    // Overflow.

                    l_EffBandwidth = uint.MaxValue;
                }
            }
            else
            {
                l_EffBandwidth = a_ConsumedBandwidth;
            }

            l_Arr[l_SlotIdx] = l_EffBandwidth;
        }

        private static uint GetBandwidthAim(Speed a_LastSpeed,
                                            Speed a_Limit,
                                            double a_Factor)
        {
            // a_LastSpeed may be Speed.s_UnknownSpeed.

            uint l_SpeedBytesPerSec;

            if (false == a_LastSpeed.IsValid())
            {
                l_SpeedBytesPerSec = (32 * 1024);
            }
            else
            {
                l_SpeedBytesPerSec = Math.Max(24 * 1024, a_LastSpeed.p_BytesPerSec);
            }

            // Try to be five percent faster than our most recent top
            // speed.
            double l_Bandwidth = (l_SpeedBytesPerSec * 1.05d);

            // But never be faster than our limit.
            l_Bandwidth = Math.Min(l_Bandwidth, (double)a_Limit.p_BytesPerSec);

            // Convert from 'per second' to 'per turn'.
            l_Bandwidth /= c_TurnsPerSecond;

            // Apply time correction factor.
            l_Bandwidth *= a_Factor;

            return (uint)l_Bandwidth;
        }

        private uint ReceiveDatagrams(uint a_AimedBandwidth)
        {
            // Returns the consumed bandwidth, which can be higher
            // than a_AimedBandwidth.

            uint l_RemainingBw = a_AimedBandwidth;

            if (0 != f_PreconsumedDatagramInBw)
            {
                if (f_PreconsumedDatagramInBw >= l_RemainingBw)
                {
                    f_PreconsumedDatagramInBw -= l_RemainingBw;

                    return 0;
                }

                l_RemainingBw -= f_PreconsumedDatagramInBw;
                f_PreconsumedDatagramInBw = 0;
            }

            uint l_ConsumedBw = 0;
            uint l_TotalPayload = 0;
            bool l_Go = true;

            while (true == l_Go)
            {
                // l_CurConsumedBw can be much larger than the maximum
                // size of one datagram, since it also includes garbage
                // that was rejected.
                uint l_CurConsumedBw;
                Packet l_Packet = f_Bridge.TryReceiveDatagram(out l_CurConsumedBw);

                l_ConsumedBw += l_CurConsumedBw;

                if (l_ConsumedBw >= l_RemainingBw)
                {
                    f_PreconsumedDatagramInBw = (l_ConsumedBw - l_RemainingBw);

                    l_Go = false;
                }

                if (null == l_Packet)
                {
                    break;
                }

                f_DatagramInQueue.Enqueue(l_Packet);
                f_DatagramInQueueSize += l_Packet.GetReceivedBodySize();

                if (f_DatagramInQueueSize >= f_MaxDatagramInQueueSize)
                {
                    l_Go = false;
                }

                f_DatagramsReceived++;
                l_TotalPayload += l_Packet.f_TransferSize;
            }

            f_TotalDatagramVolume.Received(l_TotalPayload);

            return l_ConsumedBw;
        }

        private uint SendDatagrams(uint a_AimedBandwidth)
        {
            // Returns the consumed bandwidth, which can be higher
            // than a_AimedBandwidth.

            if (true == f_Bridge.IsBusySendingDatagrams())
            {
                return 0;
            }

            uint l_RemainingBw = a_AimedBandwidth;

            if (0 != f_PreconsumedDatagramOutBw)
            {
                if (f_PreconsumedDatagramOutBw >= l_RemainingBw)
                {
                    f_PreconsumedDatagramOutBw -= l_RemainingBw;

                    return 0;
                }

                l_RemainingBw -= f_PreconsumedDatagramOutBw;
                f_PreconsumedDatagramOutBw = 0;
            }

            uint l_ConsumedBw = 0;

            if (0 != l_RemainingBw &&
                0 < f_DatagramOutQueue.Count)
            {
                uint l_TotalPayload = 0;
                List<Packet> l_Packets = new List<Packet>();

                do
                {
                    Packet l_Packet = f_DatagramOutQueue.Dequeue();
                    f_DatagramOutQueueSize -= l_Packet.f_TransferSize;

                    l_Packets.Add(l_Packet);

                    l_TotalPayload += l_Packet.f_TransferSize;

                    if (l_TotalPayload >= l_RemainingBw)
                    {
                        // We might not have enough bandwidth left to
                        // send this packet, but we do it anyway (then
                        // we will pause).

                        f_PreconsumedDatagramOutBw = (l_TotalPayload - l_RemainingBw);

                        break;
                    }
                }
                while (0 < f_DatagramOutQueue.Count);

                if (true == f_Bridge.SendDatagrams(l_Packets.ToArray(), out l_ConsumedBw))
                {
                    foreach (Packet l_Packet in l_Packets)
                    {
                        l_Packet.NotifySendingPacket(l_Packet.f_TransferSize, true);
                    }

                    f_DatagramsSent += (uint)l_Packets.Count;
                    f_TotalDatagramVolume.Sent(l_TotalPayload);
                }
            }

            return l_ConsumedBw;
        }

        private uint ReceiveStreamPackets(uint a_MaxBandwidth,
                                          List<ITransferable> a_Coll)
        {
            // Returns the consumed bandwidth.

            return DistributeWithPrioritizer(a_MaxBandwidth, false, a_Coll, a_Coll.Count);
        }

        private uint SendStreamPackets(uint a_MaxBandwidth,
                                       List<ITransferable> a_Coll)
        {
            // Returns the consumed bandwidth.

            return DistributeWithPrioritizer(a_MaxBandwidth, true, a_Coll, a_Coll.Count);
        }

        private uint DistributeWithPrioritizer(uint a_MaxBandwidth,
                                               bool a_Upload,
                                               IEnumerable<ITransferable> a_Enum,
                                               int a_Count)
        {
            // a_Count must be the number of elements that will be
            // enumerated through a_Enum.

            uint a_NeededBw;
            BandwidthPrioritizer l_Prioritizer;
            BandwidthPrioritizer.Create(a_Upload, a_Enum, a_Count, out a_NeededBw, out l_Prioritizer);

            if (0 == a_NeededBw)
            {
                return 0;
            }

            uint l_Assigned = l_Prioritizer.DistributeBandwidth(a_MaxBandwidth);

            return l_Prioritizer.Transfer();
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct SpeedOMeter
    {
        // In milliseconds.
        public const ushort c_SuggestedMeasureSpeedInterval = 125;

        private FlexArray<Item> f_Items;

        public static void Create(out SpeedOMeter a_SpeedOMeter)
        {
            a_SpeedOMeter = new SpeedOMeter();
            FlexArray<Item>.Create(out a_SpeedOMeter.f_Items);
        }

        public void Clear()
        {
            FlexArray<Item>.Create(out f_Items);
        }

        public void SaveProgress(Time a_Time,
                                 ref TransferVolume a_Volume)
        {
            TidyUp(a_Time);

            int l_Idx = f_Items.ReserveAtEnd();
            f_Items.p_Arr[l_Idx] = new Item(a_Time, a_Volume);
        }

        public void GetSpeed(Time a_Time,
                             out Speed a_DlSpeed,
                             out Speed a_UlSpeed)
        {
            TidyUp(a_Time);

            if (2 > f_Items.p_Count)
            {
                a_DlSpeed = Speed.s_UnknownSpeed;
                a_UlSpeed = Speed.s_UnknownSpeed;

                return;
            }

            Item[] l_Arr = f_Items.p_Arr;

            float l_Factor0 = a_Time.SecondsSinceAsFloat(l_Arr[0].f_Time);
            float l_Factor1 = a_Time.SecondsSinceAsFloat(l_Arr[1].f_Time);
            int l_Comp0 = l_Factor0.CompareTo(1f);
            int l_Comp1 = l_Factor1.CompareTo(1f);

            bool l_CompToFirst;

            if (0 <= l_Comp0 &&
                0 >= l_Comp1)
            {
                // If the first is more than a second ago and the next
                // is less than a second ago, pick the one nearer to one
                // second.

                l_CompToFirst = ((l_Factor0 - 1f) <= (1f - l_Factor1));
            }
            else
            {
                // Take the one nearer to one second.

                l_CompToFirst = (0 >= l_Comp0 &&
                                 0 >= l_Comp1);
            }

            l_Arr[f_Items.p_Count - 1].CalcSpeed(ref l_Arr[false == l_CompToFirst ? 1 : 0], out a_DlSpeed, out a_UlSpeed);
        }

        private void TidyUp(Time a_Time)
        {
            // Deletes all entries that are older than one second,
            // except one (the most recent).

            if (0 == f_Items.p_Count)
            {
                return;
            }

            Time l_Border = a_Time.SubSeconds(1);
            Item[] l_Arr = f_Items.p_Arr;
            for (int i = 0, lim = f_Items.p_Count; i < lim; i++)
            {
                if (false == l_Arr[i].f_Time.IsEarlierThan(l_Border))
                {
                    // Keep this one and the one before, but delete
                    // the rest of the starting.

                    if (2 <= i)
                    {
                        f_Items.DeleteRange(0, i - 1);
                    }

                    return;
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private class Item
        {
            public readonly Time f_Time;
            public readonly TransferVolume f_Volume;

            public Item(Time a_Time,
                        TransferVolume a_Volume)
            {
                f_Time = a_Time;
                f_Volume = a_Volume;
            }

            public bool CalcSpeed(ref Item a_RefItem,
                                  out Speed a_DlSpeed,
                                  out Speed a_UlSpeed)
            {
                float l_Factor = f_Time.SecondsSinceAsFloat(a_RefItem.f_Time);

                if (0.4f <= l_Factor &&
                    1.6f >= l_Factor)
                {
                    ulong l_DlBw = (this.f_Volume.p_BytesReceived - a_RefItem.f_Volume.p_BytesReceived);
                    ulong l_UlBw = (this.f_Volume.p_BytesSent - a_RefItem.f_Volume.p_BytesSent);

                    a_DlSpeed = new Speed(0 == l_DlBw ? 0u : (uint)((float)l_DlBw / l_Factor));
                    a_UlSpeed = new Speed(0 == l_UlBw ? 0u : (uint)((float)l_UlBw / l_Factor));

                    return true;
                }

                a_DlSpeed = Speed.s_UnknownSpeed;
                a_UlSpeed = Speed.s_UnknownSpeed;

                return false;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct Speed : IEquatable<Speed>
    {
        private const uint c_UnknownSpeedValue = uint.MaxValue;
        public const uint c_MaxSpeedValue = (uint.MaxValue - 1);

        private static string s_UnknownSpeedStr;

        public static readonly Speed s_NoSpeed = new Speed(0);
        public static readonly Speed s_MaxSpeed = new Speed(c_MaxSpeedValue);
        public static readonly Speed s_UnknownSpeed = new Speed(c_UnknownSpeedValue);

        // We can't imagine an I2P speed of more than 4 GiB per second
        // yet.
        private uint f_BytesPerSec;

        public uint p_BytesPerSec
        {
            get { return f_BytesPerSec; }
        }

        public Speed(uint a_BytesPerSec)
        {
            f_BytesPerSec = a_BytesPerSec;
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_UnknownSpeedStr = a_TextSet[TextKeys.SpeedUnknown];
        }

        public static void CopyVolatile(ref Speed a_VolatileSrc,
                                        out Speed a_Dest)
        {
            // Only the source is treated as volatile.

            uint l_BytesPerSec = Util.VolatileRead(ref a_VolatileSrc.f_BytesPerSec);
            a_Dest = new Speed(l_BytesPerSec);
        }

        public static Speed CreateSafe(uint a_Value,
                                       Speed a_FallbackValue)
        {
            Speed l_Speed = new Speed(a_Value);

            if (false == l_Speed.IsValid())
            {
                Util.DebugBreak();

                return a_FallbackValue;
            }

            return l_Speed;
        }

        public bool IsValid()
        {
            return (c_UnknownSpeedValue != f_BytesPerSec);
        }

        public bool IsZero()
        {
            return (0 == f_BytesPerSec);
        }

        public bool IsNearZero()
        {
            return (512 > f_BytesPerSec);
        }

        public bool IsMax()
        {
            return (c_MaxSpeedValue == f_BytesPerSec);
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(Speed a_Other)
        {
            return (this.f_BytesPerSec == a_Other.f_BytesPerSec);
        }

        public bool ValueEquals(Speed a_Other)
        {
            return Equals(a_Other);
        }

        public override int GetHashCode()
        {
            return (int)f_BytesPerSec;
        }

        public void Add(Speed a_Other)
        {
            if (c_UnknownSpeedValue == this.f_BytesPerSec ||
                c_UnknownSpeedValue == a_Other.f_BytesPerSec)
            {
                f_BytesPerSec = c_UnknownSpeedValue;
            }
            else
            {
                f_BytesPerSec += a_Other.f_BytesPerSec;
            }
        }

        public uint GetAsLimit()
        {
            return (f_BytesPerSec / UpDownManager.c_TurnsPerSecond);
        }

        public string ToGUIString(bool a_SI)
        {
            // Returns an empty string in case of zero speed.

            if (0 == f_BytesPerSec)
            {
                return string.Empty;
            }

            if (c_UnknownSpeedValue == this.f_BytesPerSec)
            {
                return s_UnknownSpeedStr;
            }

            return ToMenuString(a_SI);
        }

        public string ToMenuString(bool a_SI)
        {
            return SizeFormatter.BuildSizeStr(a_SI, false, true, true, f_BytesPerSec);
        }

        public string ToKiloString(bool a_SI)
        {
            if (c_UnknownSpeedValue == this.f_BytesPerSec)
            {
                return s_UnknownSpeedStr;
            }

            // We don't append the 'per second' part in this case.
            return SizeFormatter.BuildKiloStr(a_SI, f_BytesPerSec);
        }

        public static Speed Min(Speed a_A,
                                Speed a_B)
        {
            // Boths arguments must be valid (not undefined).

            if (true == a_A.IsMax())
            {
                return a_B;
            }

            return new Speed(Math.Min(a_A.f_BytesPerSec, a_B.f_BytesPerSec));
        }

        public static bool operator ==(Speed a_A,
                                       Speed a_B)
        {
            return (a_A.f_BytesPerSec == a_B.f_BytesPerSec);
        }

        public static bool operator !=(Speed a_A,
                                       Speed a_B)
        {
            return (a_A.f_BytesPerSec != a_B.f_BytesPerSec);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TransferDemand
    {
        public readonly Priority f_Priority;
        public readonly uint f_NeededBw;

        public TransferDemand(uint a_NeededBw)
        {
            f_Priority = Priority.s_Default;
            f_NeededBw = a_NeededBw;
        }

        public TransferDemand(Priority a_Priority,
                              uint a_NeededBw)
        {
            f_Priority = a_Priority;
            f_NeededBw = a_NeededBw;
        }

        public TransferDemand(Priority a_Priority,
                              uint a_NeededBw,
                              Speed a_Limit)
        {
            f_Priority = a_Priority;
            f_NeededBw = Math.Min(a_NeededBw, a_Limit.GetAsLimit());
        }
    }

    interface ITransferable
    {
        // Used for BandwidthPrioritizer.

        bool ReinitializeForDl(out TransferDemand a_Demand);
        bool ReinitializeForUl(out TransferDemand a_Demand);

        // The following routines may only be called after a
        // successful call to the corresponding
        // ReinitializeForX() routine.

        // Return the assigned bandwidth.
        uint AssignDlBandwidth(uint a_ExtraBytes);
        uint AssignUlBandwidth(uint a_ExtraBytes);

        // Returns the transferred bytes.
        uint Download(uint a_MaxBytes);
        uint Upload(uint a_MaxBytes);
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct BandwidthPrioritizer
    {
        private FlexArray<Item> f_Items;
        private uint f_UnsolvedItems;
        private uint f_TotalAssignedBw;

        private BandwidthPrioritizer(ref FlexArray<Item> a_Items)
        {
            f_Items = a_Items;
            f_UnsolvedItems = (uint)f_Items.p_Count;
            f_TotalAssignedBw = 0;
        }

        public static void Create<T>(bool a_Upload,
                                     IEnumerable<T> a_Enum,
                                     int a_Count,
                                     out uint a_NeededBytes,
                                     out BandwidthPrioritizer a_Prioritizer)
                            where T : ITransferable
        {
            a_NeededBytes = 0;

            FlexArray<Item> l_Items;
            FlexArray<Item>.Create(a_Count, out l_Items);

            foreach (ITransferable l_Obj in a_Enum)
            {
                bool l_Add;
                TransferDemand l_Demand;

                if (false == a_Upload)
                {
                    l_Add = l_Obj.ReinitializeForDl(out l_Demand);
                }
                else
                {
                    l_Add = l_Obj.ReinitializeForUl(out l_Demand);
                }

                if (true == l_Add)
                {
                    int l_Idx = l_Items.ReserveAtEnd();
                    l_Items.p_Arr[l_Idx] = new Item(a_Upload, l_Obj, ref l_Demand);

                    a_NeededBytes += l_Demand.f_NeededBw;
                }
            }

            a_Prioritizer = new BandwidthPrioritizer(ref l_Items);
        }

        public uint GetAssignedBandwidth()
        {
            return f_TotalAssignedBw;
        }

        public uint DistributeBandwidth(uint a_Bandwidth)
        {
            // Returns the assigned bandwidth.

            uint l_AssignedBw = 0;

            for (int i = 0; i < 2; i++)
            {
                if (0 == f_UnsolvedItems ||
                    l_AssignedBw >= a_Bandwidth)
                {
                    break;
                }

                l_AssignedBw += DistributeExtraBandwidthCore(a_Bandwidth - l_AssignedBw);
            }

            f_TotalAssignedBw += l_AssignedBw;

            return l_AssignedBw;
        }

        private uint DistributeExtraBandwidthCore(uint a_ExtraBw)
        {
            uint l_TotalPriority = 0;
            ulong l_TotalNeededBw = 0;
            int l_LastUnsolvedIdx = 0;
            Item[] l_Arr = f_Items.p_Arr;

            for (int i = (f_Items.p_Count - 1); 0 <= i; i--)
            {
                if (false == l_Arr[i].p_Done)
                {
                    l_TotalPriority += l_Arr[i].f_Demand.f_Priority.f_Value;
                    l_TotalNeededBw += l_Arr[i].GetRemainingNeed();
                    l_LastUnsolvedIdx = i;
                }
            }

            if (0 == l_TotalPriority)
            {
                // Should never happen, but we need to prevent a
                // division by zero.
                l_TotalPriority++;
            }

            uint l_TotalAssignedBw = 0;
            uint l_BwToDistribute = (uint)Math.Min(l_TotalNeededBw, a_ExtraBw);

            for (int i = (f_Items.p_Count - 1); 0 <= i; i--)
            {
                if (false == l_Arr[i].p_Done)
                {
                    uint l_PossibleBw;

                    if (l_LastUnsolvedIdx == i)
                    {
                        // Now we can distribute what's left directly.

                        l_PossibleBw = l_BwToDistribute;
                    }
                    else
                    {
                        // We do this calculation on an ulong to achieve
                        // sufficient precision.
                        l_PossibleBw = (uint)(((ulong)l_BwToDistribute * l_Arr[i].f_Demand.f_Priority.f_Value) /
                                              l_TotalPriority);
                    }

                    uint l_AssignedBw = l_Arr[i].AssignExtraBandwidth(l_PossibleBw);
                    l_TotalAssignedBw += l_AssignedBw;
                    l_BwToDistribute -= l_AssignedBw;

                    if (true == l_Arr[i].p_Done)
                    {
                        f_UnsolvedItems--;
                    }
                }
            }

            return l_TotalAssignedBw;
        }

        public uint Transfer()
        {
            uint l_BytesTransferred = 0;

            Item[] l_Arr = f_Items.p_Arr;
            for (int i = (f_Items.p_Count - 1); 0 <= i; i--)
            {
                l_BytesTransferred += l_Arr[i].Transfer();
            }

            return l_BytesTransferred;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Item
        {
            private readonly bool f_Upload;
            public readonly ITransferable f_Obj;
            public readonly TransferDemand f_Demand;
            public uint f_AssignedBw;

            public bool p_Done
            {
                get { return (f_AssignedBw >= f_Demand.f_NeededBw); }
            }

            public Item(bool a_Upload,
                        ITransferable a_Obj,
                        ref TransferDemand a_Demand)
            {
                f_Upload = a_Upload;
                f_Obj = a_Obj;
                f_Demand = a_Demand;
                f_AssignedBw = 0;
            }

            public uint GetRemainingNeed()
            {
                return (f_Demand.f_NeededBw - f_AssignedBw);
            }

            public uint AssignExtraBandwidth(uint a_MaxExtraBw)
            {
                // Returns the assigned bandwidth.

                uint l_AvailBw = Math.Min(a_MaxExtraBw, GetRemainingNeed());
                uint l_Assigned;

                if (false == f_Upload)
                {
                    l_Assigned = f_Obj.AssignDlBandwidth(l_AvailBw);
                }
                else
                {
                    l_Assigned = f_Obj.AssignUlBandwidth(l_AvailBw);
                }

                f_AssignedBw += l_Assigned;

                return l_Assigned;
            }

            public uint Transfer()
            {
                if (0 == f_AssignedBw)
                {
                    return 0;
                }

                uint f_Transferred;

                if (false == f_Upload)
                {
                    f_Transferred = f_Obj.Download(f_AssignedBw);
                }
                else
                {
                    f_Transferred = f_Obj.Upload(f_AssignedBw);
                }

                return f_Transferred;
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct Priority
    {
        public const byte c_GuiPriorityStages = 5;

        private const byte c_MinPrio = 1;
        private const byte c_MaxPrio = byte.MaxValue;
        private const byte c_VeryLowPrio = 33;
        private const byte c_LowPrio = 67;
        private const byte c_NormalPrio = 100;
        private const byte c_HighPrio = 133;
        private const byte c_VeryHighPrio = 167;

        private const byte c_BoostIncrement = 10;

        public static readonly Priority s_Min = new Priority(c_MinPrio);
        public static readonly Priority s_Max = new Priority(c_MaxPrio);

        public static readonly Priority s_VeryLow = new Priority(c_VeryLowPrio);
        public static readonly Priority s_Low = new Priority(c_LowPrio);
        public static readonly Priority s_Normal = new Priority(c_NormalPrio);
        public static readonly Priority s_High = new Priority(c_HighPrio);
        public static readonly Priority s_VeryHigh = new Priority(c_VeryHighPrio);

        public static readonly Priority s_Default = s_Normal;

        public readonly byte f_Value;

        public Priority(byte a_Value)
        {
            f_Value = a_Value;
        }

        public static Priority CreateSafe(byte a_Value)
        {
            Priority l_Prio = new Priority(a_Value);

            if (false == l_Prio.IsValid())
            {
                Util.DebugBreak();

                return s_Default;
            }

            return l_Prio;
        }

        public bool IsValid()
        {
            return (0 != f_Value);
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public override int GetHashCode()
        {
            return f_Value;
        }

        public int EncodeIntoIndex()
        {
            // Returns minus one if no index is suitable.

            switch (f_Value)
            {
                case c_NormalPrio: return 2;
                case c_HighPrio: return 3;
                case c_VeryHighPrio: return 4;
                case c_LowPrio: return 1;
                case c_VeryLowPrio: return 0;
            }

            return (-1);
        }

        public static Priority DecodeFromIndex(int a_Index)
        {
            switch (a_Index)
            {
                case 2: return s_Normal;
                case 3: return s_High;
                case 4: return s_VeryHigh;
                case 1: return s_Low;
                case 0: return s_VeryLow;
            }

            throw (new ApplicationException());
        }

        public static Priority FromDouble(double a_Value)
        {
            if (c_MinPrio > a_Value)
            {
                return Priority.s_Min;
            }

            if (c_MaxPrio < a_Value)
            {
                return Priority.s_Max;
            }

            return new Priority((byte)a_Value);
        }

        public static Priority FromBoosts(int a_Count)
        {
            int l_Prio = (c_NormalPrio + (a_Count * c_BoostIncrement));

            if (c_MinPrio >= l_Prio)
            {
                return s_Min;
            }

            if (c_MaxPrio <= l_Prio)
            {
                return s_Max;
            }

            return new Priority((byte)l_Prio);
        }

        public static bool operator ==(Priority a_A,
                                       Priority a_B)
        {
            return (a_A.f_Value == a_B.f_Value);
        }

        public static bool operator !=(Priority a_A,
                                       Priority a_B)
        {
            return (a_A.f_Value != a_B.f_Value);
        }
    }
}

// End of file.
