/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\Tags.cs
 * PURPOSE: Tag handling
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *            Some portions of this file are loosely based on the
 *            GPLed code of iMule v1.4.5.
 *            Copyright (C) 2004-2006 iMule Project.
 *            Copyright (C) 2002 Merkur.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using NB.I2P;
using NB.Language;
using Support;

namespace NB.Network.eD2k
{
    enum TagTypes : byte
    {
        TAGTYPE_INVALID = 0x00,
        TAGTYPE_HASH = 0x01,
        TAGTYPE_STRING = 0x02,
        TAGTYPE_UINT32 = 0x03,
        TAGTYPE_FLOAT32 = 0x04,
        TAGTYPE_BOOL = 0x05,
        TAGTYPE_BOOLARRAY = 0x06,
        TAGTYPE_BLOB = 0x07,
        TAGTYPE_UINT16 = 0x08,
        TAGTYPE_UINT8 = 0x09,
        TAGTYPE_BSOB = 0x0A,
        TAGTYPE_STR1 = 0x11,
        // ...
        TAGTYPE_STR16 = 0x20,
        // ...
        TAGTYPE_STR22 = 0x26,
        TAGTYPE_ADDRESS = 0x27,
        TAGTYPE_UINT64UINT64 = 0x28,
        TAGTYPE_UINT64 = 0x29,

        // Own extensions.

        // The following string types indicate a UTF-x encoded
        // string with the byte length (not char length)
        // prefixed. The length is encoded as an integer seven
        // bits at a time. For UTF-16 and UTF-32 the chars
        // itself are in the little endian byte order.
        TAGTYPE_STRING_UTF8 = 0x30,
        TAGTYPE_STRING_UTF16_LE = 0x31,
        TAGTYPE_STRING_UTF32_LE = 0x32,
    }

    enum TagCodes : byte
    {
        NONE = 0,

        // Client info tags.
        CT_NAME = 16,
        CT_VERSION = 17,
        CT_SERVER_FLAGS = 18,
        CT_EMULECOMPAT_OPTIONS = 19,
        CT_EMULE_MISCOPTIONS1 = 20,
        CT_EMULE_VERSION = 21,
        CT_DEST = 22,
        CT_EMULE_MISCOPTIONS2 = 23,

        // Mule tagnames.
        ET_COMPRESSION = 24,
        ET_UDPDEST = 25,
        ET_UDPVER = 26,
        ET_SOURCEEXCHANGE = 27,
        ET_COMMENTS = 28,
        ET_EXTENDEDREQUEST = 29,
        ET_COMPATIBLECLIENT = 30,
        ET_FEATURES = 31,
        ET_MOD_VERSION = 32,
        ET_FEATURESET = 33,
        ET_OS_INFO = 34,

        // File tags.
        TAG_FILENAME = 35,
        TAG_FILESIZE = 36,
        TAG_FILETYPE = 37,
        TAG_FILEFORMAT = 38,
        TAG_LASTSEENCOMPLETE = 39,
        TAG_PART_PATH = 40,
        TAG_PART_HASH = 41,
        TAG_TRANSFERRED = 42,
        TAG_GAP_START = 43,
        TAG_GAP_END = 44,
        TAG_DESCRIPTION = 45,
        TAG_PING = 46,
        TAG_FAIL = 47,
        TAG_PREFERENCE = 48,
        TAG_VERSION = 49,
        TAG_PARTFILENAME = 50,
        TAG_PRIORITY = 51,
        TAG_STATUS = 52,
        TAG_SOURCES = 53,
        TAG_QTIME = 54,
        TAG_PARTS = 55,
        TAG_DLPRIORITY = 56,
        TAG_ULPRIORITY = 57,
        TAG_KADLASTPUBLISHKEY = 58,
        TAG_KADLASTPUBLISHSRC = 59,
        TAG_FLAGS = 60,
        TAG_DL_ACTIVE_TIME = 61,
        TAG_CORRUPTEDPARTS = 62,
        TAG_DL_PREVIEW = 63,
        TAG_KADLASTPUBLISHNOTES = 64,
        TAG_AICH_HASH = 65,
        TAG_COMPLETE_SOURCES = 66,
        TAG_COLLECTION = 67,
        TAG_PERMISSIONS = 68,

        // Stats.
        TAG_ATTRANSFERRED = 69,
        TAG_ATREQUESTED = 70,
        TAG_ATACCEPTED = 71,
        TAG_CATEGORY = 72,
        TAG_ATTRANSFERREDHI = 73,
        TAG_MEDIA_ARTIST = 74,
        TAG_MEDIA_ALBUM = 75,
        TAG_MEDIA_TITLE = 76,
        TAG_MEDIA_LENGTH = 77,
        TAG_MEDIA_BITRATE = 78,
        TAG_MEDIA_CODEC = 79,
        TAG_FILERATING = 80,

        TAG_SERVERDEST = 81,
        TAG_SOURCE_DATAGRAM_DEST = 82,
        TAG_SOURCE_STREAM_DEST = 83,
        TAG_SOURCETYPE = 84,

        // Own extensions.

        TAG_FILENAME_UTF8 = 117,
        TAG_DESCRIPTION_UTF8 = 118,

        TAG_INDEX_MAX_MINUTES = 126,
        TAG_APPLICATION_VERSION = 127,
    }

    [Flags]
    enum TagFlags : uint
    {
        NONE = 0,
        TAG_FLAGS_COMPLETEFILE = 0x1,
    }

    abstract class Tag
    {
        // For iMule this flag is a marker that tells
        // if tag names are ints instead of strings.
        private const byte c_CodeAsNumberFlag = 0x80;

        public readonly TagCodes f_Code;

        public Tag(TagCodes a_Code)
        {
            f_Code = a_Code;
        }

        public bool IsHash()
        {
            return (typeof(HashTag) == this.GetType());
        }

        public bool IsStr()
        {
            return (typeof(StringTag) == this.GetType());
        }

        public bool IsNumber()
        {
            return (typeof(NumberTag) == this.GetType());
        }

        public bool IsFloat()
        {
            return (typeof(FloatTag) == this.GetType());
        }

        public bool IsBool()
        {
            return (typeof(BoolTag) == this.GetType());
        }

        public bool IsBlob()
        {
            return (typeof(BlobTag) == this.GetType());
        }

        public bool IsAddr()
        {
            return (typeof(AddressTag) == this.GetType());
        }

        public virtual UInt128 GetHash()
        {
            throw (new ApplicationException());
        }

        public virtual string GetStr()
        {
            throw (new ApplicationException());
        }

        public virtual ulong GetAsULong()
        {
            throw (new ApplicationException());
        }

        public virtual float GetFloat()
        {
            throw (new ApplicationException());
        }

        public virtual bool GetBool()
        {
            throw (new ApplicationException());
        }

        public virtual byte[] GetBlob()
        {
            throw (new ApplicationException());
        }

        public virtual uint GetBlobSize()
        {
            throw (new ApplicationException());
        }

        public virtual I2PAddress GetAddr()
        {
            throw (new ApplicationException());
        }

        public bool TryGetAsULong(out ulong a_Dest)
        {
            // Does intentionally not work with floats.

            if (false == IsNumber())
            {
                a_Dest = 0;

                return false;
            }

            a_Dest = ((NumberTag)this).GetAsULong();

            return true;
        }

        public static bool AreEqualNumberTags(Tag a_Tag0,
                                              Tag a_Tag1)
        {
            ulong l_Num0;
            ulong l_Num1;
            return (true == a_Tag0.TryGetAsULong(out l_Num0) &&
                    true == a_Tag1.TryGetAsULong(out l_Num1) &&
                    l_Num0 == l_Num1);
        }

        public static Tag ReadTagFromBuffer(BufferIO a_Buffer,
                                            ref ReadTagContext a_ReadCntxt)
        {
            // Return can be null if the tag is not supported,
            // but this is not an error. On error, this routine
            // will throw an exception (InterpretTagListException).

            uint l_TypeNum = a_Buffer.ReadByte();
            TagTypes l_Type;
            TagCodes l_Code;

            if (0 == (l_TypeNum & c_CodeAsNumberFlag))
            {
                l_Type = (TagTypes)l_TypeNum;

                if (1 == a_Buffer.ReadUInt16LE())
                {
                    l_Code = (TagCodes)a_Buffer.ReadByte();
                }
                else
                {
                    l_Code = TagCodes.NONE;
                }
            }
            else
            {
                l_Type = (TagTypes)(l_TypeNum & ~c_CodeAsNumberFlag);
                l_Code = (TagCodes)a_Buffer.ReadByte();
            }

            // Jump over debug data.
            a_Buffer.JumpAhead(sizeof(uint));

            a_ReadCntxt.f_Precedence = false;
            Tag l_NewTag;

            switch (l_Type)
            {
                case TagTypes.TAGTYPE_UINT8:
                    l_NewTag = new NumberTag(l_Code, a_Buffer.ReadByte());
                    break;

                case TagTypes.TAGTYPE_UINT16:
                    l_NewTag = new NumberTag(l_Code, a_Buffer.ReadUInt16LE());
                    break;

                case TagTypes.TAGTYPE_UINT32:
                    l_NewTag = new NumberTag(l_Code, a_Buffer.ReadUInt32LE());
                    break;

                case TagTypes.TAGTYPE_UINT64:
                    l_NewTag = new NumberTag(l_Code, a_Buffer.ReadUInt64LE());
                    break;

                case TagTypes.TAGTYPE_ADDRESS:
                {
                    I2PAddress l_Addr;
                    BufferIO.LoadResult l_Res = I2PAddress.TryCreateFromBufferRaw(a_Buffer, out l_Addr);

                    if (BufferIO.LoadResult.Success != l_Res)
                    {
                        throw (new TagList.InterpretTagListException());
                    }

                    l_NewTag = new AddressTag(l_Code, l_Addr);
                }
                break;

                case TagTypes.TAGTYPE_STRING:
                {
                    string l_Str = a_Buffer.ReadStringWithAutoDetection(a_Buffer.ReadUInt16LE(), a_ReadCntxt.f_NormalStrsAreInUTF8);
                    l_NewTag = new StringTag(l_Code, l_Str);
                }
                break;

                case TagTypes.TAGTYPE_STRING_UTF8:
                    l_NewTag = new StringTag(l_Code, a_Buffer.ReadUTF8StringEncLen());
                    break;

                case TagTypes.TAGTYPE_STRING_UTF16_LE:
                    l_NewTag = new StringTag(l_Code, a_Buffer.ReadUTF16LEStringEncLen());
                    break;

                case TagTypes.TAGTYPE_STRING_UTF32_LE:
                    l_NewTag = new StringTag(l_Code, a_Buffer.ReadUTF32LEStringEncLen());
                    break;

                case TagTypes.TAGTYPE_FLOAT32:
                    l_NewTag = new FloatTag(l_Code, a_Buffer.ReadFloat());
                    break;

                case TagTypes.TAGTYPE_HASH:
                    l_NewTag = new HashTag(l_Code, a_Buffer.ReadUInt128BE());
                    break;

                case TagTypes.TAGTYPE_UINT64UINT64:
                    // ReadNumberTagFromBuffer() will make sure that
                    // the tag read is a number tag, so that there can
                    // be no further recursion.
                    l_NewTag = new NumberPairTag(l_Code, ReadNumberTagFromBuffer(a_Buffer).GetAsULong(), ReadNumberTagFromBuffer(a_Buffer).GetAsULong());
                    break;

                case TagTypes.TAGTYPE_BOOL:
                    l_NewTag = new BoolTag(l_Code, a_Buffer.ReadBool());
                    break;

                case TagTypes.TAGTYPE_BLOB:
                    l_NewTag = ReadBlob(a_Buffer, l_Code, out a_ReadCntxt.f_Precedence);
                    break;

                case TagTypes.TAGTYPE_BOOLARRAY:
                {
                    // Not supported, so jump and return null.

                    uint l_Len = a_Buffer.ReadUInt16LE();
                    a_Buffer.JumpAhead((l_Len + (Util.c_BitsPerByte - 1)) / Util.c_BitsPerByte);

                    l_NewTag = null;
                }
                break;

                default:
                {
                    if (TagTypes.TAGTYPE_STR1 <= l_Type &&
                        TagTypes.TAGTYPE_STR16 >= l_Type)
                    {
                        uint l_Len = (l_Type - TagTypes.TAGTYPE_STR1 + 1u);
                        string l_Str = a_Buffer.ReadStringWithAutoDetection(l_Len, a_ReadCntxt.f_NormalStrsAreInUTF8);
                        l_NewTag = new StringTag(l_Code, l_Str);
                    }
                    else
                    {
                        throw (new TagList.InterpretTagListException());
                    }
                    break;
                }
            }

            return l_NewTag;
        }

        private static NumberTag ReadNumberTagFromBuffer(BufferIO a_Buffer)
        {
            // Throws an exception if the tag is not a number tag.

            // Peek ahead to find out about the type.
            TagTypes l_Type = (TagTypes)(a_Buffer.ReadByte() & ~c_CodeAsNumberFlag);

            if (TagTypes.TAGTYPE_UINT8 != l_Type &&
                TagTypes.TAGTYPE_UINT16 != l_Type &&
                TagTypes.TAGTYPE_UINT32 != l_Type &&
                TagTypes.TAGTYPE_UINT64 != l_Type)
            {
                throw (new TagList.InterpretTagListException());
            }

            a_Buffer.JumpBack(sizeof(byte));
            Tag.ReadTagContext l_ReadTagCntxt = new Tag.ReadTagContext(false);

            return (NumberTag)ReadTagFromBuffer(a_Buffer, ref l_ReadTagCntxt);
        }

        private static Tag ReadBlob(BufferIO a_Buffer,
                                    TagCodes a_Code,
                                    out bool a_Precedence)
        {
            uint l_ByteLen = a_Buffer.ReadUInt32LE();

            TagCodes l_EffCode;

            if (TagCodes.TAG_FILENAME_UTF8 == a_Code)
            {
                l_EffCode = TagCodes.TAG_FILENAME;
            }
            else if (TagCodes.TAG_DESCRIPTION_UTF8 == a_Code)
            {
                l_EffCode = TagCodes.TAG_DESCRIPTION;
            }
            else
            {
                // This is a normal blob.

                a_Precedence = false;

                return new BlobTag(a_Code, a_Buffer.ReadBytes(l_ByteLen));
            }

            // This is a UTF-8 string.

            a_Precedence = true;

            return new StringTag(l_EffCode, a_Buffer.ReadUTF8String(l_ByteLen));
        }

        protected static void WriteTagHeader(BufferIO a_Buffer,
                                             TagCodes a_Code,
                                             TagTypes a_Type)
        {
            a_Buffer.WriteByte((byte)((byte)a_Type | c_CodeAsNumberFlag));
            a_Buffer.WriteByte((byte)a_Code);

            // This is because iMule expects 32 bit that
            // are only used for debugging.
            a_Buffer.ZeroBytes(sizeof(uint));
        }

        public void WriteToBuffer(ref WriteTagContext a_WriteCntxt)
        {
            if (false == IsStr())
            {
                Write(a_WriteCntxt.f_Buffer);
                a_WriteCntxt.f_TagCount++;
            }
            else
            {
                ((StringTag)this).WriteString(ref a_WriteCntxt);
            }
        }

        protected virtual void Write(BufferIO a_Buffer)
        {
            throw (new ApplicationException());
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ReadTagContext
        {
            public readonly bool f_NormalStrsAreInUTF8;
            public bool f_Precedence;

            public ReadTagContext(bool a_NormalStrsAreInUTF8)
            {
                f_NormalStrsAreInUTF8 = a_NormalStrsAreInUTF8;
                f_Precedence = false;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct WriteTagContext
        {
            // Don't use the struct default constructor.

            public readonly BufferIO f_Buffer;
            public readonly UnicodeSupport f_UnicodeSupport;

            // f_TagCount gets incremented by Tag.WriteToBuffer().
            public uint f_TagCount;

            public WriteTagContext(BufferIO a_Buffer,
                                   UnicodeSupport a_UnicodeSupport)
            {
                f_Buffer = a_Buffer;
                f_UnicodeSupport = a_UnicodeSupport;
                f_TagCount = 0;
            }
        }

        [Flags]
        public enum UnicodeSupport : byte
        {
            None = 0,

            NBStyle = 0x1,
            // ImplicitUTF8 means all normal strings are encoded in
            // UTF-8 instead of Latin-1.
            ImplicitUTF8 = 0x2,
        }
    }

    sealed class HashTag : Tag
    {
        private readonly UInt128 f_Hash;

        public HashTag(TagCodes a_Code,
                       UInt128 a_Hash) :
                  base(a_Code)
        {
            f_Hash = a_Hash;
        }

        public override UInt128 GetHash()
        {
            return f_Hash;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            WriteTagHeader(a_Buffer, f_Code, TagTypes.TAGTYPE_HASH);
            a_Buffer.WriteUInt128LE(f_Hash);
        }
    }

    sealed class StringTag : Tag
    {
        private readonly string f_Str;

        private InspectionState f_State;
        private TagCodes f_SpecificCode;

        public StringTag(TagCodes a_Code,
                         string a_Str) :
                    base(a_Code)
        {
            f_Str = Util.UnicodeNormalizeString(a_Str);
        }

        public override string GetStr()
        {
            return f_Str;
        }

        public void WriteString(ref WriteTagContext a_WriteCntxt)
        {
            if (InspectionState.NotAnalyzed == f_State)
            {
                f_State = AnalyzeString(f_Code, f_Str, out f_SpecificCode);
            }

            StaticWriteString(ref a_WriteCntxt, f_Code, f_State, f_SpecificCode, f_Str);
        }

        private static InspectionState AnalyzeString(TagCodes a_Code,
                                                     string a_Str,
                                                     out TagCodes a_SpecificCode)
        {
            if (false == Util.IsLatin1AndUTF8Equivalent(a_Str))
            {
                switch (a_Code)
                {
                    case TagCodes.TAG_FILENAME:
                        a_SpecificCode = TagCodes.TAG_FILENAME_UTF8;
                        break;

                    case TagCodes.TAG_DESCRIPTION:
                        a_SpecificCode = TagCodes.TAG_DESCRIPTION_UTF8;
                        break;

                    default:
                        a_SpecificCode = TagCodes.NONE;
                        break;
                }

                return (false == Util.IsLatin1Compatible(a_Str) ? InspectionState.UnicodeRequired : InspectionState.UnicodeRecommended);
            }

            a_SpecificCode = TagCodes.NONE;

            return InspectionState.UnicodeNotRequired;
        }

        public static void DetermineTypeAndWrite(ref WriteTagContext a_Cntxt,
                                                 TagCodes a_Code,
                                                 string a_Str)
        {
            TagCodes l_SpecificCode;
            InspectionState l_State = AnalyzeString(a_Code, a_Str, out l_SpecificCode);
            StaticWriteString(ref a_Cntxt, a_Code, l_State, l_SpecificCode, a_Str);
        }

        private static void StaticWriteString(ref WriteTagContext a_Cntxt,
                                              TagCodes a_Code,
                                              InspectionState a_State,
                                              TagCodes a_SpecialCode,
                                              string a_Str)
        {
            a_Cntxt.f_TagCount++;

            if (InspectionState.UnicodeNotRequired == a_State)
            {
                StaticWriteAsLatinString(a_Cntxt.f_Buffer, a_Code, a_Str);
            }
            else
            {
                if (0 != (UnicodeSupport.NBStyle & a_Cntxt.f_UnicodeSupport))
                {
                    if (false == Util.IsUTF16ShorterThanUTF8(a_Str, false))
                    {
                        WriteTagHeader(a_Cntxt.f_Buffer, a_Code, TagTypes.TAGTYPE_STRING_UTF8);
                        a_Cntxt.f_Buffer.WriteUTF8StringEncLen(a_Str);
                    }
                    else
                    {
                        WriteTagHeader(a_Cntxt.f_Buffer, a_Code, TagTypes.TAGTYPE_STRING_UTF16_LE);
                        a_Cntxt.f_Buffer.WriteUTF16LEStringEncLen(a_Str);
                    }
                }
                else if (0 != (UnicodeSupport.ImplicitUTF8 & a_Cntxt.f_UnicodeSupport))
                {
                    WriteTagHeader(a_Cntxt.f_Buffer, a_Code, TagTypes.TAGTYPE_STRING);
                    a_Cntxt.f_Buffer.WriteUTF8StringUInt16Len(a_Str);
                }
                else
                {
                    // When we write a latin-1 and an UTF-8 string,
                    // we need to make sure the unicode string is
                    // written last.

                    StaticWriteAsLatinString(a_Cntxt.f_Buffer, a_Code, a_Str);

                    if (InspectionState.UnicodeRequired == a_State &&
                        TagCodes.NONE != a_SpecialCode)
                    {
                        // The receiver may not have UTF-8 support, but
                        // we will still write this out as UTF-8 blob
                        // using a special code.

                        WriteTagHeader(a_Cntxt.f_Buffer, a_SpecialCode, TagTypes.TAGTYPE_BLOB);
                        a_Cntxt.f_Buffer.WriteUTF8StringUInt32Len(a_Str);
                        a_Cntxt.f_TagCount++;
                    }
                }
            }
        }

        private static void StaticWriteAsLatinString(BufferIO a_Buffer,
                                                     TagCodes a_Code,
                                                     string a_Str)
        {
            byte[] l_LatinBytes = Util.s_Latin1BestFit.GetBytes(a_Str);
            int l_LatinByteLen = l_LatinBytes.Length;

            if (16 >= l_LatinByteLen &&
                1 <= l_LatinByteLen)
            {
                TagTypes l_TagType = (TagTypes)((int)TagTypes.TAGTYPE_STR1 + l_LatinByteLen - 1);

                WriteTagHeader(a_Buffer, a_Code, l_TagType);
            }
            else
            {
                if (ushort.MaxValue < l_LatinByteLen)
                {
                    throw (new ApplicationException());
                }

                WriteTagHeader(a_Buffer, a_Code, TagTypes.TAGTYPE_STRING);
                a_Buffer.WriteUInt16LE((ushort)l_LatinByteLen);
            }

            a_Buffer.WriteBytes(l_LatinBytes);
        }

        private enum InspectionState : byte
        {
            NotAnalyzed = 0,
            UnicodeNotRequired = 1,
            UnicodeRecommended = 2,
            UnicodeRequired = 3,
        }
    }

    sealed class NumberTag : Tag
    {
        private readonly ulong f_Num;

        private TagTypes f_SpecificType;

        public NumberTag(TagCodes a_Code,
                         ulong a_Num) :
                    base(a_Code)
        {
            f_Num = a_Num;
        }

        public override ulong GetAsULong()
        {
            return f_Num;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            if (TagTypes.TAGTYPE_INVALID == f_SpecificType)
            {
                f_SpecificType = GetTagType(f_Num);
            }

            StaticWrite(a_Buffer, f_Code, f_SpecificType, f_Num);
        }

        private static TagTypes GetTagType(ulong a_Num)
        {
            if (byte.MaxValue >= a_Num)
            {
                return TagTypes.TAGTYPE_UINT8;
            }

            if (ushort.MaxValue >= a_Num)
            {
                return TagTypes.TAGTYPE_UINT16;
            }

            if (uint.MaxValue >= a_Num)
            {
                return TagTypes.TAGTYPE_UINT32;
            }

            return TagTypes.TAGTYPE_UINT64;
        }

        public static void DetermineTypeAndWrite(BufferIO a_Buffer,
                                                 TagCodes a_Code,
                                                 ulong a_Num)
        {
            TagTypes l_SpecificType = GetTagType(a_Num);
            StaticWrite(a_Buffer, a_Code, l_SpecificType, a_Num);
        }

        public static void StaticWrite(BufferIO a_Buffer,
                                       TagCodes a_Code,
                                       TagTypes a_SpecificType,
                                       ulong a_Num)
        {
            WriteTagHeader(a_Buffer, a_Code, a_SpecificType);

            switch (a_SpecificType)
            {
                case TagTypes.TAGTYPE_UINT8:
                    a_Buffer.WriteByte((byte)a_Num);
                    break;

                case TagTypes.TAGTYPE_UINT16:
                    a_Buffer.WriteUInt16LE((ushort)a_Num);
                    break;

                case TagTypes.TAGTYPE_UINT32:
                    a_Buffer.WriteUInt32LE((uint)a_Num);
                    break;

                case TagTypes.TAGTYPE_UINT64:
                    a_Buffer.WriteUInt64LE(a_Num);
                    break;

                default: throw (new ApplicationException());
            }
        }
    }

    sealed class NumberPairTag : Tag
    {
        public readonly ulong f_First;
        public readonly ulong f_Second;

        public NumberPairTag(TagCodes a_Code,
                             ulong a_First,
                             ulong a_Second) :
                        base(a_Code)
        {
            f_First = a_First;
            f_Second = a_Second;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            WriteTagHeader(a_Buffer, f_Code, TagTypes.TAGTYPE_UINT64UINT64);
            NumberTag.DetermineTypeAndWrite(a_Buffer, f_Code, f_First);
            NumberTag.DetermineTypeAndWrite(a_Buffer, f_Code, f_Second);
        }
    }

    sealed class FloatTag : Tag
    {
        private readonly float f_Float;

        public FloatTag(TagCodes a_Code,
                        float a_Float) :
                   base(a_Code)
        {
            f_Float = a_Float;
        }

        public override float GetFloat()
        {
            return f_Float;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            WriteTagHeader(a_Buffer, f_Code, TagTypes.TAGTYPE_FLOAT32);
            a_Buffer.WriteFloat(f_Float);
        }
    }

    sealed class BoolTag : Tag
    {
        private readonly bool f_Bool;

        public BoolTag(TagCodes a_Code,
                       bool a_Bool) :
                  base(a_Code)
        {
            f_Bool = a_Bool;
        }

        public override bool GetBool()
        {
            return f_Bool;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            WriteTagHeader(a_Buffer, f_Code, TagTypes.TAGTYPE_BOOL);
            a_Buffer.WriteBool(f_Bool);
        }
    }

    sealed class BlobTag : Tag
    {
        private readonly byte[] f_Data;

        public BlobTag(TagCodes a_Code,
                       byte[] a_Data) :
                  base(a_Code)
        {
            f_Data = a_Data;
        }

        public override byte[] GetBlob()
        {
            return f_Data;
        }

        public override uint GetBlobSize()
        {
            return (uint)f_Data.Length;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            WriteTagHeader(a_Buffer, f_Code, TagTypes.TAGTYPE_BLOB);
            a_Buffer.WriteUInt32LE((uint)f_Data.Length);
            a_Buffer.WriteBytes(f_Data);
        }
    }

    sealed class AddressTag : Tag
    {
        private readonly I2PAddress f_Addr;

        public AddressTag(TagCodes a_Code,
                          I2PAddress a_Addr) :
                     base(a_Code)
        {
            f_Addr = a_Addr;
        }

        public override I2PAddress GetAddr()
        {
            return f_Addr;
        }

        protected override void Write(BufferIO a_Buffer)
        {
            StaticWrite(a_Buffer, f_Code, f_Addr);
        }

        public static void StaticWrite(BufferIO a_Buffer,
                                       TagCodes a_Code,
                                       I2PAddress a_Addr)
        {
            WriteTagHeader(a_Buffer, a_Code, TagTypes.TAGTYPE_ADDRESS);
            a_Addr.ToBufferRaw(a_Buffer);
        }
    }

    class TagList
    {
        public const byte c_MaxEntries = byte.MaxValue;

        private readonly Dictionary<TagCodes, Tag> f_Dic;

        public uint p_Count
        {
            get { return (uint)f_Dic.Count; }
        }

        private TagList(Dictionary<TagCodes, Tag> a_Dic)
        {
            f_Dic = a_Dic;
        }

        public TagList() :
                  this(new Dictionary<TagCodes, Tag>())
        {
        }

        public static TagList ReadListUntrusted(BufferIO a_Buffer,
                                                uint a_Count,
                                                bool a_NormalStrsAreInUTF8)
        {
            // Return can be null. Does not catch any
            // FileAccessExceptions.

            try
            {
                return ReadListTrusted(a_Buffer, a_Count, a_NormalStrsAreInUTF8);
            }
            catch (BufferIO.InterpretBufferException)
            {
                return null;
            }
        }

        public static TagList ReadListTrusted(BufferIO a_Buffer,
                                              bool a_NormalStrsAreInUTF8)
        {
            // Return can be null.

            return ReadListTrusted(a_Buffer, a_Buffer.ReadByte(), a_NormalStrsAreInUTF8);
        }

        public static TagList ReadListTrusted(BufferIO a_Buffer,
                                              uint a_Count,
                                              bool a_NormalStrsAreInUTF8)
        {
            // Return can be null.

            FlexArray<FileRange> l_GapRanges = new FlexArray<FileRange>();
            return ReadListTrustedCore(a_Buffer, a_Count, a_NormalStrsAreInUTF8, false, ref l_GapRanges);
        }

        public static TagList ReadListTrustedWithFileRanges(BufferIO a_Buffer,
                                                            uint a_Count,
                                                            bool a_NormalStrsAreInUTF8,
                                                            out FlexArray<FileRange> a_GapRanges)
        {
            // Return can be null.

            FlexArray<FileRange>.Create(out a_GapRanges);
            return ReadListTrustedCore(a_Buffer, a_Count, a_NormalStrsAreInUTF8, true, ref a_GapRanges);
        }

        private static TagList ReadListTrustedCore(BufferIO a_Buffer,
                                                   uint a_Count,
                                                   bool a_NormalStrsAreInUTF8,
                                                   bool a_ExpectRanges,
                                                   ref FlexArray<FileRange> a_GapRanges)
        {
            // Return can be null if the input is corrupt. Exceptions
            // are possible (InterpretTagListException).

            Tag.ReadTagContext l_ReadTagCntxt = new Tag.ReadTagContext(a_NormalStrsAreInUTF8);
            Dictionary<TagCodes, Tag> l_Dic = new Dictionary<TagCodes, Tag>((int)a_Count);
            NumberPairTag l_GapStart = null;
            int l_GapCount = 0;

            for (int i = 0; i < a_Count; i++)
            {
                Tag l_Tag = Tag.ReadTagFromBuffer(a_Buffer, ref l_ReadTagCntxt);

                if (null == l_Tag)
                {
                    // This tag is not supported and will be skipped.

                    continue;
                }

                if (true == a_ExpectRanges &&
                    typeof(NumberPairTag) == l_Tag.GetType())
                {
                    // Since we cannot add multiple tags with the same
                    // code, we will filter those tags.

                    if (TagCodes.TAG_GAP_START == l_Tag.f_Code)
                    {
                        if (null != l_GapStart)
                        {
                            return null;
                        }

                        l_GapStart = (NumberPairTag)l_Tag;

                        if (l_GapStart.f_First != (uint)l_GapCount)
                        {
                            return null;
                        }

                        continue;
                    }
                    else if (TagCodes.TAG_GAP_END == l_Tag.f_Code)
                    {
                        if (null == l_GapStart)
                        {
                            return null;
                        }

                        NumberPairTag l_GapEnd = (NumberPairTag)l_Tag;

                        if (l_GapStart.f_First != l_GapEnd.f_First)
                        {
                            return null;
                        }

                        a_GapRanges.ReserveAtEnd();

                        if (false == FileRange.CreateFromUnsafeBounds(l_GapStart.f_Second, l_GapEnd.f_Second, out a_GapRanges.p_Arr[l_GapCount]))
                        {
                            if (l_GapStart.f_Second == l_GapEnd.f_Second)
                            {
                                // Allow zero ranges, but don't add
                                // them.

                                a_GapRanges.DeleteAt(l_GapCount);
                                l_GapStart = null;

                                continue;
                            }

                            return null;
                        }

                        l_GapStart = null;
                        l_GapCount++;

                        continue;
                    }
                }

                if (false == l_ReadTagCntxt.f_Precedence)
                {
                    try
                    {
                        l_Dic.Add(l_Tag.f_Code, l_Tag);
                    }
                    catch (ArgumentException)
                    {
                        // If this is not a string tag, a collision
                        // is an error.

                        if (false == l_Tag.IsStr())
                        {
                            // IMULETWEAK: iMule 2.3.0 implements the
                            // same number tag with the same value
                            // several times in some cases (e.g. two
                            // file sizes).
                            if (true == Tag.AreEqualNumberTags(l_Tag, l_Dic[l_Tag.f_Code]))
                            {
                                continue;
                            }

                            return null;
                        }
                    }
                }
                else
                {
                    // In this case it's okay if the code is already
                    // there.

                    l_Dic[l_Tag.f_Code] = l_Tag;
                }
            }

            if (true == a_ExpectRanges)
            {
                if (null != l_GapStart)
                {
                    // Range not closed.

                    return null;
                }

                // The ranges are not necessarily sorted when loaded.
                a_GapRanges.Sort(FileRange.CompareByStartingOffset);
            }

            return new TagList(l_Dic);
        }

        public void Clear()
        {
            f_Dic.Clear();
        }

        // The following routines help us to write tag lists for
        // publish requests without having to instance several
        // reference types (tag list + tags) each time.

        public static void WriteFileEntryForPublishing(BufferIO a_Buffer,
                                                       FileEntry a_FileEntry)
        {
            bool l_HasFileSize = (0 != a_FileEntry.f_FileSize);
            a_Buffer.WriteByte(false == l_HasFileSize ? (byte)3 : (byte)4);

            NumberTag.StaticWrite(a_Buffer, TagCodes.TAG_SOURCETYPE, TagTypes.TAGTYPE_UINT8, 1);

            AddressTag.StaticWrite(a_Buffer, TagCodes.TAG_SOURCE_STREAM_DEST, a_FileEntry.f_StreamAddr);

            TagFlags l_Flags = (false == a_FileEntry.f_Complete ? TagFlags.NONE : TagFlags.TAG_FLAGS_COMPLETEFILE);
            NumberTag.StaticWrite(a_Buffer, TagCodes.TAG_FLAGS, TagTypes.TAGTYPE_UINT8, (byte)l_Flags);

            if (true == l_HasFileSize)
            {
                NumberTag.DetermineTypeAndWrite(a_Buffer, TagCodes.TAG_FILESIZE, a_FileEntry.f_FileSize);
            }
        }

        public static void WriteKeywordEntryForPublishing(ref Tag.WriteTagContext a_Cntxt,
                                                          KeywordEntry a_KeywordEntry)
        {
            // a_Cntxt.f_TagCount is not maintained by this routine.

            // We know that there will be at least three number tags.
            a_Cntxt.f_TagCount = 3;

            BufferIO.CountWriter l_CountWtr;
            BufferIO.CountWriter.InitAndMoveBuffer(a_Cntxt.f_Buffer, sizeof(byte), out l_CountWtr);

            StringTag.DetermineTypeAndWrite(ref a_Cntxt, TagCodes.TAG_FILENAME, a_KeywordEntry.f_FileName);

            NumberTag.DetermineTypeAndWrite(a_Cntxt.f_Buffer, TagCodes.TAG_FILESIZE, a_KeywordEntry.f_FileSize);

            NumberTag.DetermineTypeAndWrite(a_Cntxt.f_Buffer, TagCodes.TAG_COMPLETE_SOURCES, a_KeywordEntry.f_CompleteSources);

            NumberTag.DetermineTypeAndWrite(a_Cntxt.f_Buffer, TagCodes.TAG_SOURCES, a_KeywordEntry.f_Sources);

            l_CountWtr.Finish(a_Cntxt.f_TagCount);
        }

        public static void WriteNoteEntryForPublishing(ref Tag.WriteTagContext a_Cntxt,
                                                       NoteEntry a_NoteEntry)
        {
            // a_Cntxt.f_TagCount is not maintained by this routine.

            a_Cntxt.f_TagCount = 0;

            BufferIO.CountWriter l_CountWtr;
            BufferIO.CountWriter.InitAndMoveBuffer(a_Cntxt.f_Buffer, sizeof(byte), out l_CountWtr);

            StringTag.DetermineTypeAndWrite(ref a_Cntxt, TagCodes.TAG_FILENAME, a_NoteEntry.f_FileName);

            if (0 != a_NoteEntry.f_FileSize)
            {
                NumberTag.DetermineTypeAndWrite(a_Cntxt.f_Buffer, TagCodes.TAG_FILESIZE, a_NoteEntry.f_FileSize);
                a_Cntxt.f_TagCount++;
            }

            if (FileRatings.NotRated != a_NoteEntry.f_Rating)
            {
                NumberTag.DetermineTypeAndWrite(a_Cntxt.f_Buffer, TagCodes.TAG_FILERATING, (byte)a_NoteEntry.f_Rating);
                a_Cntxt.f_TagCount++;
            }

            if (null != a_NoteEntry.f_Description)
            {
                StringTag.DetermineTypeAndWrite(ref a_Cntxt, TagCodes.TAG_DESCRIPTION, a_NoteEntry.f_Description);
            }

            l_CountWtr.Finish(a_Cntxt.f_TagCount);
        }

        public void WriteTo(BufferIO a_Buffer,
                            uint a_CountByteSize,
                            Tag.UnicodeSupport a_UnicodeSupport)
        {
            BufferIO.CountWriter l_CountWtr;

            BufferIO.CountWriter.InitAndMoveBuffer(a_Buffer, a_CountByteSize, out l_CountWtr);
            Tag.WriteTagContext l_WriteTagCntxt = new Tag.WriteTagContext(a_Buffer, a_UnicodeSupport);

            foreach (Tag CurTag in f_Dic.Values)
            {
                CurTag.WriteToBuffer(ref l_WriteTagCntxt);
            }

            if (sizeof(byte) == a_CountByteSize &&
                byte.MaxValue < l_WriteTagCntxt.f_TagCount)
            {
                throw (new ApplicationException());
            }

            l_CountWtr.Finish(l_WriteTagCntxt.f_TagCount);
        }

        public byte[] TagListToBytes(Tag.UnicodeSupport a_UnicodeSupport)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            WriteTo(l_Buffer, sizeof(byte), a_UnicodeSupport);

            return l_Buffer.ToArray(true);
        }

        public void AddTag(Tag a_Tag)
        {
            if (f_Dic.Count >= c_MaxEntries)
            {
                throw (new ApplicationException());
            }

            f_Dic.Add(a_Tag.f_Code, a_Tag);
        }

        public void RemoveTag(TagCodes a_Code)
        {
            f_Dic.Remove(a_Code);
        }

        public ushort GetAndDeleteMaxLifetime()
        {
            ushort l_Mins = ushort.MaxValue;

            ulong l_Num;
            if (true == TryGetTagAsULong(TagCodes.TAG_INDEX_MAX_MINUTES, out l_Num) &&
                ushort.MaxValue >= l_Num)
            {
                l_Mins = (ushort)l_Num;
            }

            RemoveTag(TagCodes.TAG_INDEX_MAX_MINUTES);

            return l_Mins;
        }

        public Tag GetTag(TagCodes a_Code)
        {
            Tag l_Tag;
            if (true == f_Dic.TryGetValue(a_Code, out l_Tag))
            {
                return l_Tag;
            }

            return null;
        }

        public bool TryGetTagAsUInt(TagCodes a_Code,
                                    out uint a_Val)
        {
            ulong l_Num;
            if (true == TryGetTagAsULong(a_Code, out l_Num) &&
                uint.MaxValue >= l_Num)
            {
                a_Val = (uint)l_Num;

                return true;
            }

            a_Val = 0;

            return false;
        }

        public bool TryGetTagAsULong(TagCodes a_Code,
                                     out ulong a_Dest)
        {
            Tag l_Tag = GetTag(a_Code);

            if (null != l_Tag)
            {
                return l_Tag.TryGetAsULong(out a_Dest);
            }

            a_Dest = 0;

            return false;
        }

        public bool TryGetTagAsString(TagCodes a_Code,
                                      out string a_Str)
        {
            // Return can be null.

            Tag l_Tag = GetTag(a_Code);

            if (null != l_Tag &&
                true == l_Tag.IsStr())
            {
                a_Str = l_Tag.GetStr();

                return true;
            }

            a_Str = null;

            return false;
        }

        public bool TryGetTagAsAddress(TagCodes a_Code,
                                       out I2PAddress a_Addr)
        {
            Tag l_Tag = GetTag(a_Code);

            if (null != l_Tag &&
                true == l_Tag.IsAddr())
            {
                a_Addr = l_Tag.GetAddr();

                return true;
            }

            a_Addr = null;

            return false;
        }

        public class InterpretTagListException : BufferIO.InterpretBufferException
        {
        }
    }

    enum FileType : byte
    {
        FT_ANY = 0,
        FT_AUDIO = 1,
        FT_VIDEO = 2,
        FT_IMAGE = 3,
        FT_PROGRAM = 4,
        FT_DOCUMENT = 5,
        FT_ARCHIVE = 6,
        FT_CDIMAGE = 7,
        FT_COLLECTION = 8,
    }

    static class FileTypeDetector
    {
        private const byte c_FileTypeCount = 9;

        // s_FileTypesAsNetworkStr contains strings the protocol uses.
        private static readonly string[] s_FileTypesAsNetworkStr;
        private static readonly Dictionary<string, FileType> s_Extensions;

        // s_FileTypesAsPrettyStr contains strings for the GUI only.
        private static string[] s_FileTypesAsPrettyStr;

        static FileTypeDetector()
        {
            s_FileTypesAsNetworkStr = new string[c_FileTypeCount]
            {
                null,
                "Audio",
                "Video",
                "Image",
                "Pro",
                "Doc",
                "Arc",
                "Iso",
                "Collect"
            };

            const int l_FileTypeCount = 228;
            s_Extensions = new Dictionary<string, FileType>(l_FileTypeCount);

            AddFileExtension("aac", FileType.FT_AUDIO);
            AddFileExtension("ac3", FileType.FT_AUDIO);
            AddFileExtension("aif", FileType.FT_AUDIO);
            AddFileExtension("aifc", FileType.FT_AUDIO);
            AddFileExtension("aiff", FileType.FT_AUDIO);
            AddFileExtension("amr", FileType.FT_AUDIO);
            AddFileExtension("ape", FileType.FT_AUDIO);
            AddFileExtension("au", FileType.FT_AUDIO);
            AddFileExtension("aud", FileType.FT_AUDIO);
            AddFileExtension("audio", FileType.FT_AUDIO);
            AddFileExtension("cda", FileType.FT_AUDIO);
            AddFileExtension("dmf", FileType.FT_AUDIO);
            AddFileExtension("dsm", FileType.FT_AUDIO);
            AddFileExtension("dts", FileType.FT_AUDIO);
            AddFileExtension("far", FileType.FT_AUDIO);
            AddFileExtension("flac", FileType.FT_AUDIO);
            AddFileExtension("it", FileType.FT_AUDIO);
            AddFileExtension("m1a", FileType.FT_AUDIO);
            AddFileExtension("m2a", FileType.FT_AUDIO);
            AddFileExtension("m4a", FileType.FT_AUDIO);
            AddFileExtension("mdl", FileType.FT_AUDIO);
            AddFileExtension("med", FileType.FT_AUDIO);
            AddFileExtension("mid", FileType.FT_AUDIO);
            AddFileExtension("midi", FileType.FT_AUDIO);
            AddFileExtension("mka", FileType.FT_AUDIO);
            AddFileExtension("mod", FileType.FT_AUDIO);
            AddFileExtension("mp1", FileType.FT_AUDIO);
            AddFileExtension("mp2", FileType.FT_AUDIO);
            AddFileExtension("mp3", FileType.FT_AUDIO);
            AddFileExtension("mpa", FileType.FT_AUDIO);
            AddFileExtension("mpc", FileType.FT_AUDIO);
            AddFileExtension("mtm", FileType.FT_AUDIO);
            AddFileExtension("ogg", FileType.FT_AUDIO);
            AddFileExtension("psm", FileType.FT_AUDIO);
            AddFileExtension("ptm", FileType.FT_AUDIO);
            AddFileExtension("ra", FileType.FT_AUDIO);
            AddFileExtension("rmi", FileType.FT_AUDIO);
            AddFileExtension("s3m", FileType.FT_AUDIO);
            AddFileExtension("snd", FileType.FT_AUDIO);
            AddFileExtension("stm", FileType.FT_AUDIO);
            AddFileExtension("umx", FileType.FT_AUDIO);
            AddFileExtension("wav", FileType.FT_AUDIO);
            AddFileExtension("wma", FileType.FT_AUDIO);
            AddFileExtension("xm", FileType.FT_AUDIO);

            AddFileExtension("3g2", FileType.FT_VIDEO);
            AddFileExtension("3gp", FileType.FT_VIDEO);
            AddFileExtension("3gp2", FileType.FT_VIDEO);
            AddFileExtension("3gpp", FileType.FT_VIDEO);
            AddFileExtension("amv", FileType.FT_VIDEO);
            AddFileExtension("asf", FileType.FT_VIDEO);
            AddFileExtension("avi", FileType.FT_VIDEO);
            AddFileExtension("bik", FileType.FT_VIDEO);
            AddFileExtension("divx", FileType.FT_VIDEO);
            AddFileExtension("dvr-ms", FileType.FT_VIDEO);
            AddFileExtension("flc", FileType.FT_VIDEO);
            AddFileExtension("fli", FileType.FT_VIDEO);
            AddFileExtension("flic", FileType.FT_VIDEO);
            AddFileExtension("flv", FileType.FT_VIDEO);
            AddFileExtension("hdmov", FileType.FT_VIDEO);
            AddFileExtension("ifo", FileType.FT_VIDEO);
            AddFileExtension("m1v", FileType.FT_VIDEO);
            AddFileExtension("m2t", FileType.FT_VIDEO);
            AddFileExtension("m2ts", FileType.FT_VIDEO);
            AddFileExtension("m2v", FileType.FT_VIDEO);
            AddFileExtension("m4b", FileType.FT_VIDEO);
            AddFileExtension("m4v", FileType.FT_VIDEO);
            AddFileExtension("mkv", FileType.FT_VIDEO);
            AddFileExtension("mov", FileType.FT_VIDEO);
            AddFileExtension("movie", FileType.FT_VIDEO);
            AddFileExtension("mp1v", FileType.FT_VIDEO);
            AddFileExtension("mp2v", FileType.FT_VIDEO);
            AddFileExtension("mp4", FileType.FT_VIDEO);
            AddFileExtension("mpe", FileType.FT_VIDEO);
            AddFileExtension("mpeg", FileType.FT_VIDEO);
            AddFileExtension("mpg", FileType.FT_VIDEO);
            AddFileExtension("mpv", FileType.FT_VIDEO);
            AddFileExtension("mpv1", FileType.FT_VIDEO);
            AddFileExtension("mpv2", FileType.FT_VIDEO);
            AddFileExtension("ogm", FileType.FT_VIDEO);
            AddFileExtension("pva", FileType.FT_VIDEO);
            AddFileExtension("qt", FileType.FT_VIDEO);
            AddFileExtension("ram", FileType.FT_VIDEO);
            AddFileExtension("ratdvd", FileType.FT_VIDEO);
            AddFileExtension("rm", FileType.FT_VIDEO);
            AddFileExtension("rmm", FileType.FT_VIDEO);
            AddFileExtension("rmvb", FileType.FT_VIDEO);
            AddFileExtension("rv", FileType.FT_VIDEO);
            AddFileExtension("smil", FileType.FT_VIDEO);
            AddFileExtension("smk", FileType.FT_VIDEO);
            AddFileExtension("swf", FileType.FT_VIDEO);
            AddFileExtension("tp", FileType.FT_VIDEO);
            AddFileExtension("ts", FileType.FT_VIDEO);
            AddFileExtension("vid", FileType.FT_VIDEO);
            AddFileExtension("video", FileType.FT_VIDEO);
            AddFileExtension("vob", FileType.FT_VIDEO);
            AddFileExtension("vp6", FileType.FT_VIDEO);
            AddFileExtension("wm", FileType.FT_VIDEO);
            AddFileExtension("wmv", FileType.FT_VIDEO);
            AddFileExtension("xvid", FileType.FT_VIDEO);

            AddFileExtension("bmp", FileType.FT_IMAGE);
            AddFileExtension("emf", FileType.FT_IMAGE);
            AddFileExtension("gif", FileType.FT_IMAGE);
            AddFileExtension("ico", FileType.FT_IMAGE);
            AddFileExtension("jfif", FileType.FT_IMAGE);
            AddFileExtension("jpe", FileType.FT_IMAGE);
            AddFileExtension("jpeg", FileType.FT_IMAGE);
            AddFileExtension("jpg", FileType.FT_IMAGE);
            AddFileExtension("pct", FileType.FT_IMAGE);
            AddFileExtension("pcx", FileType.FT_IMAGE);
            AddFileExtension("pic", FileType.FT_IMAGE);
            AddFileExtension("pict", FileType.FT_IMAGE);
            AddFileExtension("png", FileType.FT_IMAGE);
            AddFileExtension("psd", FileType.FT_IMAGE);
            AddFileExtension("psp", FileType.FT_IMAGE);
            AddFileExtension("tga", FileType.FT_IMAGE);
            AddFileExtension("tif", FileType.FT_IMAGE);
            AddFileExtension("tiff", FileType.FT_IMAGE);
            AddFileExtension("wmf", FileType.FT_IMAGE);
            AddFileExtension("wmp", FileType.FT_IMAGE);
            AddFileExtension("xif", FileType.FT_IMAGE);

            AddFileExtension("7z", FileType.FT_ARCHIVE);
            AddFileExtension("ace", FileType.FT_ARCHIVE);
            AddFileExtension("alz", FileType.FT_ARCHIVE);
            AddFileExtension("arc", FileType.FT_ARCHIVE);
            AddFileExtension("arj", FileType.FT_ARCHIVE);
            AddFileExtension("bz2", FileType.FT_ARCHIVE);
            AddFileExtension("cab", FileType.FT_ARCHIVE);
            AddFileExtension("cbr", FileType.FT_ARCHIVE);
            AddFileExtension("cbz", FileType.FT_ARCHIVE);
            AddFileExtension("gz", FileType.FT_ARCHIVE);
            AddFileExtension("hqx", FileType.FT_ARCHIVE);
            AddFileExtension("lha", FileType.FT_ARCHIVE);
            AddFileExtension("lzh", FileType.FT_ARCHIVE);
            AddFileExtension("pak", FileType.FT_ARCHIVE);
            AddFileExtension("par", FileType.FT_ARCHIVE);
            AddFileExtension("par2", FileType.FT_ARCHIVE);
            AddFileExtension("rar", FileType.FT_ARCHIVE);
            AddFileExtension("sit", FileType.FT_ARCHIVE);
            AddFileExtension("sitx", FileType.FT_ARCHIVE);
            AddFileExtension("tar", FileType.FT_ARCHIVE);
            AddFileExtension("tbz2", FileType.FT_ARCHIVE);
            AddFileExtension("tgz", FileType.FT_ARCHIVE);
            AddFileExtension("z", FileType.FT_ARCHIVE);
            AddFileExtension("zip", FileType.FT_ARCHIVE);

            AddFileExtension("air", FileType.FT_PROGRAM);
            AddFileExtension("app", FileType.FT_PROGRAM);
            AddFileExtension("application", FileType.FT_PROGRAM);
            AddFileExtension("appx", FileType.FT_PROGRAM);
            AddFileExtension("azw2", FileType.FT_PROGRAM);
            AddFileExtension("bat", FileType.FT_PROGRAM);
            AddFileExtension("bin", FileType.FT_PROGRAM);
            AddFileExtension("cmd", FileType.FT_PROGRAM);
            AddFileExtension("com", FileType.FT_PROGRAM);
            AddFileExtension("deb", FileType.FT_PROGRAM);
            AddFileExtension("dll", FileType.FT_PROGRAM);
            AddFileExtension("elf", FileType.FT_PROGRAM);
            AddFileExtension("exe", FileType.FT_PROGRAM);
            AddFileExtension("fox", FileType.FT_PROGRAM);
            AddFileExtension("fpx", FileType.FT_PROGRAM);
            AddFileExtension("fxp", FileType.FT_PROGRAM);
            AddFileExtension("hta", FileType.FT_PROGRAM);
            AddFileExtension("jar", FileType.FT_PROGRAM);
            AddFileExtension("jax", FileType.FT_PROGRAM);
            AddFileExtension("js", FileType.FT_PROGRAM);
            AddFileExtension("jse", FileType.FT_PROGRAM);
            AddFileExtension("lnk", FileType.FT_PROGRAM);
            AddFileExtension("msc", FileType.FT_PROGRAM);
            AddFileExtension("msi", FileType.FT_PROGRAM);
            AddFileExtension("nt", FileType.FT_PROGRAM);
            AddFileExtension("pe", FileType.FT_PROGRAM);
            AddFileExtension("prx", FileType.FT_PROGRAM);
            AddFileExtension("pwz", FileType.FT_PROGRAM);
            AddFileExtension("rpm", FileType.FT_PROGRAM);
            AddFileExtension("scr", FileType.FT_PROGRAM);
            AddFileExtension("shs", FileType.FT_PROGRAM);
            AddFileExtension("sys", FileType.FT_PROGRAM);
            AddFileExtension("vb", FileType.FT_PROGRAM);
            AddFileExtension("vbe", FileType.FT_PROGRAM);
            AddFileExtension("vbs", FileType.FT_PROGRAM);
            AddFileExtension("wgt", FileType.FT_PROGRAM);
            AddFileExtension("wince", FileType.FT_PROGRAM);
            AddFileExtension("wiz", FileType.FT_PROGRAM);
            AddFileExtension("wsf", FileType.FT_PROGRAM);
            AddFileExtension("wsh", FileType.FT_PROGRAM);
            AddFileExtension("xap", FileType.FT_PROGRAM);
            AddFileExtension("xlnk", FileType.FT_PROGRAM);
            AddFileExtension("xpi", FileType.FT_PROGRAM);

            AddFileExtension("bwa", FileType.FT_CDIMAGE);
            AddFileExtension("bwi", FileType.FT_CDIMAGE);
            AddFileExtension("bws", FileType.FT_CDIMAGE);
            AddFileExtension("bwt", FileType.FT_CDIMAGE);
            AddFileExtension("ccd", FileType.FT_CDIMAGE);
            AddFileExtension("cue", FileType.FT_CDIMAGE);
            AddFileExtension("dmg", FileType.FT_CDIMAGE);
            AddFileExtension("img", FileType.FT_CDIMAGE);
            AddFileExtension("iso", FileType.FT_CDIMAGE);
            AddFileExtension("mdf", FileType.FT_CDIMAGE);
            AddFileExtension("mds", FileType.FT_CDIMAGE);
            AddFileExtension("nrg", FileType.FT_CDIMAGE);
            AddFileExtension("sub", FileType.FT_CDIMAGE);
            AddFileExtension("toast", FileType.FT_CDIMAGE);

            AddFileExtension("chm", FileType.FT_DOCUMENT);
            AddFileExtension("css", FileType.FT_DOCUMENT);
            AddFileExtension("diz", FileType.FT_DOCUMENT);
            AddFileExtension("doc", FileType.FT_DOCUMENT);
            AddFileExtension("dot", FileType.FT_DOCUMENT);
            AddFileExtension("hlp", FileType.FT_DOCUMENT);
            AddFileExtension("htm", FileType.FT_DOCUMENT);
            AddFileExtension("html", FileType.FT_DOCUMENT);
            AddFileExtension("nfo", FileType.FT_DOCUMENT);
            AddFileExtension("nfr", FileType.FT_DOCUMENT);
            AddFileExtension("pdf", FileType.FT_DOCUMENT);
            AddFileExtension("pps", FileType.FT_DOCUMENT);
            AddFileExtension("ppt", FileType.FT_DOCUMENT);
            AddFileExtension("ps", FileType.FT_DOCUMENT);
            AddFileExtension("rtf", FileType.FT_DOCUMENT);
            AddFileExtension("text", FileType.FT_DOCUMENT);
            AddFileExtension("txt", FileType.FT_DOCUMENT);
            AddFileExtension("wri", FileType.FT_DOCUMENT);
            AddFileExtension("xls", FileType.FT_DOCUMENT);
            AddFileExtension("xml", FileType.FT_DOCUMENT);

            AddFileExtension("emulecollection", FileType.FT_COLLECTION);
            AddFileExtension("nbc", FileType.FT_COLLECTION);
            AddFileExtension("torrent", FileType.FT_COLLECTION);

            AddFileExtension("dor", FileType.FT_AUDIO);
            AddFileExtension("apk", FileType.FT_PROGRAM);
            AddFileExtension("webm", FileType.FT_VIDEO);
            AddFileExtension("dsk", FileType.FT_CDIMAGE);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_FileTypesAsPrettyStr = a_TextSet.ReadArray(TextKeys.FileTypeAny, c_FileTypeCount);
        }

        private static void AddFileExtension(string a_Ext,
                                             FileType a_Type)
        {
            s_Extensions.Add(a_Ext, a_Type);
        }

        public static FileType GuessFileType(string a_FileName)
        {
            string l_Extension = LongPath.GetExtension(a_FileName);

            return (null == l_Extension ? FileType.FT_ANY :
                    GuessFileTypeFromExtension(l_Extension));
        }

        public static FileType GuessFileTypeFromExtension(string a_Extension)
        {
            string l_LowercaseExt = a_Extension.ToLowerInvariant();

            FileType l_FileType;
            if (true == s_Extensions.TryGetValue(l_LowercaseExt, out l_FileType))
            {
                return l_FileType;
            }

            return FileType.FT_ANY;
        }

        public static bool TryToRecognizeStringFileType(string a_TypeStr,
                                                        out FileType a_TypeEnum)
        {
            for (int i = 1; i < s_FileTypesAsNetworkStr.Length; i++)
            {
                if (true == Util.StringsEqualIgnoreCase(s_FileTypesAsNetworkStr[i], a_TypeStr))
                {
                    a_TypeEnum = (FileType)i;

                    return true;
                }
            }

            a_TypeEnum = FileType.FT_ANY;

            return false;
        }

        public static string FileTypeToNetworkString(FileType a_TypeEnum)
        {
            // For ED2KFT_ANY this routine will return null.

            return s_FileTypesAsNetworkStr[(int)a_TypeEnum];
        }

        public static string FileTypeToPrettyString(FileType a_TypeEnum)
        {
            return s_FileTypesAsPrettyStr[(int)a_TypeEnum];
        }

        public static string[] GetFileTypesStrings()
        {
            // Used by the GUI.

            return s_FileTypesAsPrettyStr;
        }
    }
}

// End of file.
