/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\Types.cs
 * PURPOSE: Wrapper structs for type security
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Support;

namespace NB.Network.eD2k
{
    interface IUInt128
    {
        UInt128 GetUInt128();
    }

    struct UserID : IEquatable<UserID>, IComparable, IComparable<UserID>, IUInt128
    {
        public readonly UInt128 f_Val;

        public UserID(UInt128 a_Val)
        {
            f_Val = a_Val;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(UserID a_Other)
        {
            return this.f_Val.Equals(a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            return f_Val.GetHashCode();
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((UserID)a_Other);
        }

        public int CompareTo(UserID a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public UInt128 GetUInt128()
        {
            return f_Val;
        }

        public string AsStringNullSafe(string a_FailStr)
        {
            if (null == f_Val)
            {
                return a_FailStr;
            }

            return f_Val.ToHexString();
        }

        public bool IsValid()
        {
            // Checks if the magic bytes (see GetRandom()) are there.

            return (0x0E0000 == (f_Val.Get32BitChunk(2) & 0xFF0000) &&
                    0x6F00 == (f_Val.Get32BitChunk(0) & 0xFF00));
        }

        public static UserID GetRandom()
        {
            // A valid user id consists of random data except two
            // special bytes (in big endian notation at byte offsets
            // 5 and 14, e.g. 5931A9D2600E1411241670AA39396F9A).

            byte[] l_Bytes = Util.GetRandomBytes(UInt128.c_ByteSize);
            // Our input is in little endian, so flip offsets.
            l_Bytes[UInt128.c_ByteSize - 5 - 1] = 0x0E;
            l_Bytes[UInt128.c_ByteSize - 14 - 1] = 0x6F;

            UserID l_NewID = new UserID(new UInt128(l_Bytes));

            return l_NewID;
        }
    }

    struct ClientID : IEquatable<ClientID>, IComparable, IComparable<ClientID>, IUInt128
    {
        public readonly UInt128 f_Val;

        public ClientID(UInt128 a_Val)
        {
            f_Val = a_Val;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(ClientID a_Other)
        {
            return this.f_Val.Equals(a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            return f_Val.GetHashCode();
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((ClientID)a_Other);
        }

        public int CompareTo(ClientID a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public UInt128 GetUInt128()
        {
            return f_Val;
        }
    }

    struct FileID : IEquatable<FileID>, IComparable<FileID>, IUInt128
    {
        public readonly UInt128 f_Val;

        public FileID(UInt128 a_Val)
        {
            f_Val = a_Val;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(FileID a_Other)
        {
            return this.f_Val.Equals(a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            return f_Val.GetHashCode();
        }

        public int CompareTo(FileID a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public UInt128 GetUInt128()
        {
            return f_Val;
        }
    }

    struct KeywordID : IEquatable<KeywordID>, IComparable<KeywordID>, IUInt128
    {
        public readonly UInt128 f_Val;

        public KeywordID(UInt128 a_Val)
        {
            f_Val = a_Val;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(KeywordID a_Other)
        {
            return this.f_Val.Equals(a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            return f_Val.GetHashCode();
        }

        public int CompareTo(KeywordID a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public UInt128 GetUInt128()
        {
            return f_Val;
        }
    }

    struct Distance : IComparable, IComparable<Distance>, IUInt128
    {
        public readonly UInt128 f_Val;

        public Distance(UInt128 a_Val)
        {
            f_Val = a_Val;
        }

        public bool HasMinFitness(uint a_Bits)
        {
            return f_Val.AreHighBitsClear(a_Bits);
        }

        public uint GetFitness()
        {
            return f_Val.GetClearHighBitCount();
        }

        public bool BelongsToUpperHalf()
        {
            return f_Val.BelongsToUpperHalf();
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public override int GetHashCode()
        {
            return f_Val.GetHashCode();
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((Distance)a_Other);
        }

        public int CompareTo(Distance a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public UInt128 GetUInt128()
        {
            return f_Val;
        }
    }
}

namespace NB.Network
{
    struct UserNum : IEquatable<UserNum>, IComparable<UserNum>
    {
        public const ulong c_ReservedUserNums = 1024;
        private const ulong c_SelfUserNum = 0;

        public static readonly UserNum s_Self = new UserNum(c_SelfUserNum);
        // Used to identify the origin of data when a completed chunk
        // turns into an incomplete chunk again.
        public static readonly UserNum s_Anonymous = new UserNum(1);
        // Used by the GUI to extract preview data.
        public static readonly UserNum s_PreviewExtraction = new UserNum(2);
        // Used by the content manager to create complete files.
        public static readonly UserNum s_ComplFileExtraction = new UserNum(3);
        // Used to query sources.
        public static readonly UserNum s_QueryUsers = new UserNum(4);
        // Used when importing partial downloads from another client.
        public static readonly UserNum s_Import = new UserNum(5);

        public readonly ulong f_Val;

        public UserNum(ulong a_Val)
        {
            f_Val = a_Val;
        }

        public bool IsSelf()
        {
            return (0 == f_Val);
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(UserNum a_Other)
        {
            return (this.f_Val == a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            // We don't include the upper half in the hash code,
            // since conflicts are very unlikely already with
            // the lower half.
            return (int)f_Val;
        }

        public int CompareTo(UserNum a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public bool IsGeneric()
        {
            return (c_ReservedUserNums > f_Val);
        }

        public bool IsNonGeneric()
        {
            return (false == IsGeneric());
        }

        public static bool operator ==(UserNum a_A,
                                       UserNum a_B)
        {
            return (a_A.f_Val == a_B.f_Val);
        }

        public static bool operator !=(UserNum a_A,
                                       UserNum a_B)
        {
            return (a_A.f_Val != a_B.f_Val);
        }
    }
}

namespace NB.Content
{
    struct FileNum : IEquatable<FileNum>, IComparable<FileNum>
    {
        public readonly ulong f_Val;

        public FileNum(ulong a_Val)
        {
            f_Val = a_Val;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(FileNum a_Other)
        {
            return (this.f_Val == a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            // We don't include the upper half in the hash code.
            return (int)f_Val;
        }

        public int CompareTo(FileNum a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public static bool operator ==(FileNum a_A,
                                       FileNum a_B)
        {
            return (a_A.f_Val == a_B.f_Val);
        }

        public static bool operator !=(FileNum a_A,
                                       FileNum a_B)
        {
            return (a_A.f_Val != a_B.f_Val);
        }
    }

    struct ElmKey : IEquatable<ElmKey>, IComparable<ElmKey>
    {
        public readonly ulong f_Val;

        public ElmKey(ulong a_Val)
        {
            f_Val = a_Val;
        }

        public bool IsZero()
        {
            return (0 == f_Val);
        }

        public bool IsNonzero()
        {
            return (0 != f_Val);
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(ElmKey a_Other)
        {
            return (this.f_Val == a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            // We don't include the upper half in the hash code.
            return (int)f_Val;
        }

        public int CompareTo(ElmKey a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public static bool operator ==(ElmKey a_A,
                                       ElmKey a_B)
        {
            return (a_A.f_Val == a_B.f_Val);
        }

        public static bool operator !=(ElmKey a_A,
                                       ElmKey a_B)
        {
            return (a_A.f_Val != a_B.f_Val);
        }
    }

    struct Category : IEquatable<Category>, IComparable<Category>
    {
        public const ushort c_GenericCount = 7;
        public const ushort c_ReservedGenericNums = 128;

        public static readonly Category s_All = new Category(0);
        public static readonly Category s_Incomplete = new Category(1);
        public static readonly Category s_Missing = new Category(2);
        public static readonly Category s_Complete = new Category(3);
        public static readonly Category s_Active = new Category(4);
        public static readonly Category s_Downloads = new Category(5);
        public static readonly Category s_Uploads = new Category(6);
        public static readonly Category s_Unassigned = new Category(c_ReservedGenericNums - 1);

        public readonly ushort f_Val;

        public Category(uint a_Val)
        {
            f_Val = (ushort)a_Val;
        }

        public bool IsGeneric()
        {
            return (c_ReservedGenericNums > f_Val);
        }

        public bool IsNonGeneric()
        {
            return !IsGeneric();
        }

        public bool IsUnassignedOrNonGeneric()
        {
            return ((c_ReservedGenericNums - 1) <= f_Val);
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool Equals(Category a_Other)
        {
            return (this.f_Val == a_Other.f_Val);
        }

        public override int GetHashCode()
        {
            return f_Val;
        }

        public int CompareTo(Category a_Other)
        {
            return this.f_Val.CompareTo(a_Other.f_Val);
        }

        public bool FilterRoutine(KnownFile a_Known)
        {
            // Returns true if the file should be filtered away.

            return (a_Known.p_Category.f_Val != f_Val);
        }

        public static bool operator ==(Category a_A,
                                       Category a_B)
        {
            return (a_A.f_Val == a_B.f_Val);
        }

        public static bool operator !=(Category a_A,
                                       Category a_B)
        {
            return (a_A.f_Val != a_B.f_Val);
        }
    }
}

namespace NB
{
#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct FileRange : IEquatable<FileRange>
    {
        public readonly ulong f_StartingOffset;
        public readonly uint f_Length;

        public ulong p_FinalOffset
        {
            get { return (f_StartingOffset + f_Length); }
        }

        private FileRange(ulong a_StartingOffset,
                          uint a_Length)
        {
            f_StartingOffset = a_StartingOffset;
            f_Length = a_Length;
        }

        public bool IsValid()
        {
            return (0 != f_Length);
        }

        public static bool CreateFromUnsafeBounds(ulong a_StartingOffset,
                                                  ulong a_FinalOffset,
                                                  out FileRange a_FileRange)
        {
            // We won't allow zero ranges.

            if (a_StartingOffset < a_FinalOffset)
            {
                ulong l_Length = (a_FinalOffset - a_StartingOffset);

                if (uint.MaxValue >= l_Length)
                {
                    a_FileRange = new FileRange(a_StartingOffset, (uint)l_Length);

                    return true;
                }
            }

            a_FileRange = new FileRange();

            return false;
        }

        public static void CreateFromChunkRange(ulong a_ChunkBaseOffset,
                                                ref Content.Internal.ChunkRange a_ChunkRange,
                                                out FileRange a_FileRange)
        {
            a_FileRange = new FileRange(a_ChunkBaseOffset + a_ChunkRange.f_Offset, a_ChunkRange.f_Size);
        }

        public bool Equals(FileRange a_Range)
        {
            return (this.f_StartingOffset == a_Range.f_StartingOffset &&
                    this.f_Length == a_Range.f_Length);
        }

        public bool DoesOverlap(FileRange a_Other)
        {
            return (this.f_StartingOffset < a_Other.p_FinalOffset &&
                    this.p_FinalOffset > a_Other.f_StartingOffset);
        }

        public static int CompareByStartingOffset(ref FileRange a_A,
                                                  ref FileRange a_B)
        {
            return a_A.f_StartingOffset.CompareTo(a_B.f_StartingOffset);
        }
    }
}

// End of file.
