/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Native\MD4.c
 * PURPOSE: MD4 implementation
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

// See: RFC 1320 MD4 Message-Digest Algorithm April 1992.

#include "general.h"

// a = (a + F(b,c,d) + X[k]) <<< s;
// F(X,Y,Z) = XY v not(X) Z.
#define Round1Op(a, b, c, d, k, s) (a) += ((((b) & (c)) | ((~(b)) & (d))) + l_X[(k)]); (a) = ROT_L((a), (s));

// a = (a + G(b,c,d) + X[k] + 5A827999) <<< s;
// G(X,Y,Z) = XY v XZ v YZ.
#define Round2Op(a, b, c, d, k, s) (a) += ((((b) & (c)) | ((b) & (d)) | ((c) & (d))) + l_X[(k)] + 0x5A827999UL); (a) = ROT_L((a), (s));

// a = (a + H(b,c,d) + X[k] + 6ED9EBA1) <<< s;
// H(X,Y,Z) = X xor Y xor Z.
#define Round3Op(a, b, c, d, k, s) (a) += (((b) ^ (c) ^ (d)) + l_X[(k)] + 0x6ED9EBA1UL); (a) = ROT_L((a), (s));

// Exports.

EXPORT void CALLCONV
MD4MultiTransform_EXPORT(IN OUT UINT32 * const a_Registers /* 4 UINT32s */,
                         IN UINT32 const * const a_InputBlockBuf /* 16 UINT32s per block */,
                         IN UINT32 const a_InputBlockCount)
{
    // a_Registers and a_InputBlockBuf may reference the same memory
    // location.

    UINT32 const * l_X;
    UINT32 l_ToGo;
    UINT32 A, B, C, D;

    for (l_X = a_InputBlockBuf, l_ToGo = a_InputBlockCount; 0 != l_ToGo; l_X += 16, l_ToGo--)
    {
        A = a_Registers[0];
        B = a_Registers[1];
        C = a_Registers[2];
        D = a_Registers[3];

        Round1Op(A, B, C, D, 0, 3);
        Round1Op(D, A, B, C, 1, 7);
        Round1Op(C, D, A, B, 2, 11);
        Round1Op(B, C, D, A, 3, 19);
        Round1Op(A, B, C, D, 4, 3);
        Round1Op(D, A, B, C, 5, 7);
        Round1Op(C, D, A, B, 6, 11);
        Round1Op(B, C, D, A, 7, 19);
        Round1Op(A, B, C, D, 8, 3);
        Round1Op(D, A, B, C, 9, 7);
        Round1Op(C, D, A, B, 10, 11);
        Round1Op(B, C, D, A, 11, 19);
        Round1Op(A, B, C, D, 12, 3);
        Round1Op(D, A, B, C, 13, 7);
        Round1Op(C, D, A, B, 14, 11);
        Round1Op(B, C, D, A, 15, 19);

        Round2Op(A, B, C, D, 0, 3);
        Round2Op(D, A, B, C, 4, 5);
        Round2Op(C, D, A, B, 8, 9);
        Round2Op(B, C, D, A, 12, 13);
        Round2Op(A, B, C, D, 1, 3);
        Round2Op(D, A, B, C, 5, 5);
        Round2Op(C, D, A, B, 9, 9);
        Round2Op(B, C, D, A, 13, 13);
        Round2Op(A, B, C, D, 2, 3);
        Round2Op(D, A, B, C, 6, 5);
        Round2Op(C, D, A, B, 10, 9);
        Round2Op(B, C, D, A, 14, 13);
        Round2Op(A, B, C, D, 3, 3);
        Round2Op(D, A, B, C, 7, 5);
        Round2Op(C, D, A, B, 11, 9);
        Round2Op(B, C, D, A, 15, 13);

        Round3Op(A, B, C, D, 0, 3);
        Round3Op(D, A, B, C, 8, 9);
        Round3Op(C, D, A, B, 4, 11);
        Round3Op(B, C, D, A, 12, 15);
        Round3Op(A, B, C, D, 2, 3);
        Round3Op(D, A, B, C, 10, 9);
        Round3Op(C, D, A, B, 6, 11);
        Round3Op(B, C, D, A, 14, 15);
        Round3Op(A, B, C, D, 1, 3);
        Round3Op(D, A, B, C, 9, 9);
        Round3Op(C, D, A, B, 5, 11);
        Round3Op(B, C, D, A, 13, 15);
        Round3Op(A, B, C, D, 3, 3);
        Round3Op(D, A, B, C, 11, 9);
        Round3Op(C, D, A, B, 7, 11);
        Round3Op(B, C, D, A, 15, 15);

        a_Registers[0] += A;
        a_Registers[1] += B;
        a_Registers[2] += C;
        a_Registers[3] += D;
    }
}

// For backward compatibility:

EXPORT void CALLCONV
MD4Transform_EXPORT(IN OUT UINT32 * const a_Registers,
                    IN UINT32 const * const a_InputBlockBuf)
{
    MD4MultiTransform_EXPORT(a_Registers, a_InputBlockBuf, 1);
}

// End of file.
