/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Native\Timer.c
 * PURPOSE: High-resolution timer interface
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

#define INCLUDE_WINDOWS
#include "general.h"

// This interface basically exposes GetTickCount64() with the
// granularity of FILETIME, using the timer with the highest
// possible precision.

// One tick represents 100 nanoseconds.
#define TICKS_PER_SECOND            (10ULL * 1000 * 1000)

static BOOLEAN g_Initialized = FALSE;
static UINT64 g_Frequency = 0;

BOOLEAN
HiResTimerInitOrShutdownInternal(IN BOOLEAN const a_Init)
{
    // This routine is used for initialization and shutdown, regarding
    // global data. Returns a nonzero value if the desired final state
    // was reached.

    if (a_Init == g_Initialized)
    {
        return TRUE;
    }

    if (FALSE != a_Init)
    {
        // We expect a precision that is clearly better than one
        // millisecond (so require a precision of 100 microseconds).
        const UINT64 l_MinUnitsPerSecond = (10ULL * 1000);
        // NOTE: (2^63 - 1) is the largest value we expect the counter
        // to ever assume. The following limit makes sure that the
        // remainder (counter % frequency) can't become too large.
        // This limit allows for a precision of up to about 1.0842
        // picoseconds.
        const UINT64 l_MaxUnitsPerSecond = (((1ULL << 63) - 1) / TICKS_PER_SECOND);
        LARGE_INTEGER l_Frequency;

        if (FALSE != QueryPerformanceFrequency(&l_Frequency) &&
            l_MinUnitsPerSecond <= (UINT64)l_Frequency.QuadPart &&
            l_MaxUnitsPerSecond >= (UINT64)l_Frequency.QuadPart)
        {
            g_Initialized = TRUE;
            g_Frequency = (UINT64)l_Frequency.QuadPart;

            return TRUE;
        }
    }

    g_Initialized = FALSE;
    g_Frequency = 0;

    return !a_Init;
}

EXPORT BOOL CALLCONV
HiResTimerQuery_EXPORT(METAIN OUT UINT64 * a_Ticks)
{
    // On successful return, a_Ticks will be a number that increases
    // by one about every 100 nanoseconds. This routine does not check
    // for/workaround hardware or OS defects on multiprocessor systems.
    // There is no mechanism here to prevent that a_Ticks will somehow
    // become smaller than before as time passes (the callers should
    // employ such a mechanism). This routine is thread-safe.

    // NOTE: If the system is suspended (sleep/hibernation), the counter
    // behaves as if it had kept running during suspension. GetTickCount()
    // works the same way. QueryUnbiasedInterruptTime() works different
    // and returns a count that only reflects the time the system spent
    // in working state.

    LARGE_INTEGER l_Counter;
    UINT64 l_SecTicks;
    UINT64 l_RestTicks;

    if (FALSE == g_Initialized ||
        FALSE == QueryPerformanceCounter(&l_Counter))
    {
        goto FAIL;
    }

    // Calc the number of whole seconds.
    l_SecTicks = ((UINT64)l_Counter.QuadPart / g_Frequency);

    if ((MAX_UINT64 / TICKS_PER_SECOND) < l_SecTicks)
    {
        // There are so many whole seconds, that we aren't able to
        // represent them.

        goto FAIL;
    }

    // Shift the number of seconds seven decimal places to the left.
    l_SecTicks *= TICKS_PER_SECOND;

    // Calc the number of 100-ms from the fractal seconds. This can't
    // overflow, since we've already made sure it won't when
    // initializing g_Frequency.
    l_RestTicks = ((((UINT64)l_Counter.QuadPart % g_Frequency) * TICKS_PER_SECOND) / g_Frequency);

    *a_Ticks = (l_SecTicks + l_RestTicks);

    return TRUE;

    FAIL:

    *a_Ticks = 0;

    return FALSE;
}

// End of file.
