/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Support\Threading.cs
 * PURPOSE: Threading related code
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Threading;

namespace Support
{
    public enum ResourceAccessType : byte
    {
        None = 0,
        Shared = 1,
        Exclusive = 2,
    }

    public abstract class Resource
    {
        // This class will refresh/flush the thread cache whenever
        // a lock is acquired or released, as the Monitor class does.

        public bool NoLockHeld()
        {
            return (ResourceAccessType.None == GetAccessType());
        }

        public bool HasSharedAccess()
        {
            return (ResourceAccessType.Shared == GetAccessType());
        }

        public bool HasExclusiveAccess()
        {
            return (ResourceAccessType.Exclusive == GetAccessType());
        }

        public bool HasSharedOrExclusiveAccess()
        {
            return (ResourceAccessType.Shared <= GetAccessType());
        }

        public abstract ResourceAccessType GetAccessType();

        public void AcquireShared()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.AcquireShrd);
        }

        public void AcquireExclusive()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.AcquireExcl);
        }

        public void ReleaseShared()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.ReleaseShrd);
        }

        public void ReleaseExclusive()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.ReleaseExcl);
        }

        public void Upgrade()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.UpgradeToExclWithPossibleIntermediateRelease);
        }

        public void Downgrade()
        {
            ResourceAccessType l_Type = HandleRequestBase(Action.DowngradeToShrd);
        }

        public ResourceAccessType SwitchToShared()
        {
            // Returns the old access state.

            return HandleRequestBase(Action.SwitchToShrd);
        }

        public ResourceAccessType SwitchToExclusive()
        {
            // Returns the old access state.

            return HandleRequestBase(Action.SwitchToExcl);
        }

        public ResourceAccessType SwitchToSharedOrBetter()
        {
            // Returns the old access state.

            return HandleRequestBase(Action.SwitchToShrdOrBetter);
        }

        public ResourceAccessType ReleaseAny()
        {
            // Returns the old access state.

            return HandleRequestBase(Action.SwitchToNone);
        }

        public void SwitchToIfNecessary(ResourceAccessType a_CurrentAccessType,
                                        ResourceAccessType a_AimedAccessType)
        {
            if (a_CurrentAccessType == a_AimedAccessType)
            {
                // Nothing to do.
            }
            else
            {
                SwitchTo(a_AimedAccessType);
            }
        }

        public void SwitchToIfNecessaryAdv(ref ResourceAccessType a_CurrentAccessType,
                                           ResourceAccessType a_AimedAccessType)
        {
            // Same as SwitchToIfNecessary(), but updates the first
            // argument.

            SwitchToIfNecessary(a_CurrentAccessType, a_AimedAccessType);
            a_CurrentAccessType = a_AimedAccessType;
        }

        public ResourceAccessType SwitchTo(ResourceAccessType a_AccessType)
        {
            // Returns the old access state.

            Action l_Action;

            switch (a_AccessType)
            {
                case ResourceAccessType.None:
                    l_Action = Action.SwitchToNone;
                    break;

                case ResourceAccessType.Shared:
                    l_Action = Action.SwitchToShrd;
                    break;

                case ResourceAccessType.Exclusive:
                    l_Action = Action.SwitchToExcl;
                    break;

                default: throw (new ApplicationException());
            }

            return HandleRequestBase(l_Action);
        }

        private ResourceAccessType HandleRequestBase(Action a_InAction)
        {
            // We cannot be sure that the result will be true for
            // Action.UpgradeDirectlyToExcl.

            ResourceAccessType l_OldAccessType;
            bool l_OK = HandleRequest(a_InAction, true, out l_OldAccessType);

            return l_OldAccessType;
        }

        protected abstract bool HandleRequest(Action a_InAction,
                                              bool a_Wait,
                                              out ResourceAccessType a_OldAccessType);

        protected static void AssertSameAccessType(ResourceAccessType a_RealAccessType,
                                                   ResourceAccessType a_ExpectedAccessType)
        {
            if (a_RealAccessType != a_ExpectedAccessType)
            {
                throw (new ApplicationException("Object synchronization conflict. Actual type: " + a_RealAccessType + "; expected type: " + a_ExpectedAccessType));
            }
        }

        public static Resource CreateSMP()
        {
            return new ResourceSMP();
        }

        public static Resource Create(int a_ThreadCount)
        {
            if (1 >= a_ThreadCount)
            {
                return new ResourceDummy();
            }

            ResourceSMP l_Res = new ResourceSMP();

            return l_Res;
        }

        protected static bool IsSwitchAction(Action a_Action)
        {
            return (Action.SWITCH_BEGIN <= a_Action &&
                    Action.SWITCH_END >= a_Action);
        }

        protected static Action TranslateSwitchAction(ResourceAccessType a_AccessType,
                                                      Action a_Action)
        {
            switch (a_Action)
            {
                case Action.SwitchToNone:
                    switch (a_AccessType)
                    {
                        case ResourceAccessType.None:
                            return Action.DoNothing;

                        case ResourceAccessType.Shared:
                            return Action.ReleaseShrd;

                        case ResourceAccessType.Exclusive:
                            return Action.ReleaseExcl;
                    }
                    break;

                case Action.SwitchToShrd:
                    switch (a_AccessType)
                    {
                        case ResourceAccessType.None:
                            return Action.AcquireShrd;

                        case ResourceAccessType.Shared:
                            return Action.DoNothing;

                        case ResourceAccessType.Exclusive:
                            return Action.DowngradeToShrd;
                    }
                    break;

                case Action.SwitchToExcl:
                    switch (a_AccessType)
                    {
                        case ResourceAccessType.None:
                            return Action.AcquireExcl;

                        case ResourceAccessType.Shared:
                            return Action.UpgradeToExclWithPossibleIntermediateRelease;

                        case ResourceAccessType.Exclusive:
                            return Action.DoNothing;
                    }
                    break;

                case Action.SwitchToShrdOrBetter:
                    switch (a_AccessType)
                    {
                        case ResourceAccessType.None:
                            return Action.AcquireShrd;

                        case ResourceAccessType.Shared:
                        case ResourceAccessType.Exclusive:
                            return Action.DoNothing;
                    }
                    break;
            }

            throw (new ApplicationException());
        }

        protected enum Action : byte
        {
            DoNothing = 0,

            AcquireShrd = 1,
            AcquireExcl = 2,

            ReleaseShrd = 3,
            ReleaseExcl = 4,

            // The UpgradeDirectlyToExcl operation may fail even
            // if the thread is allowed to wait.
            UpgradeDirectlyToExcl = 5,
            UpgradeToExclWithPossibleIntermediateRelease = 6,
            DowngradeToShrd = 7,

            SwitchToShrd = 8,
            SwitchToExcl = 9,
            SwitchToShrdOrBetter = 10,
            SwitchToNone = 11,

            SWITCH_BEGIN = SwitchToShrd,
            SWITCH_END = SwitchToNone,
        }
    }

    class ResourceSMP : Resource
    {
        // A single-writer/multiple-reader synchronization object
        // implementation optimized for a small number of consumers.

        private readonly object f_Lock;

        private uint f_OwnedShrdCount;
        private bool f_OwnedExcl;
        private FlexArray<ThreadAccessContext> f_Consumers;

        private uint f_SharedWaiting;
        private uint f_ExclWaiting;
        private bool f_TryingToUpgrade;

        public ResourceSMP()
        {
            f_Lock = new object();
            FlexArray<ThreadAccessContext>.Create(out f_Consumers);
        }

        public override ResourceAccessType GetAccessType()
        {
            ResourceAccessType l_AccessType;

            Enter();
            {
                int l_CntxtIdx = FindContextIndex();
                l_AccessType = GetThreadAccessType(l_CntxtIdx);
            }
            Leave();

            return l_AccessType;
        }

        protected override bool HandleRequest(Action a_InAction,
                                              bool a_Wait,
                                              out ResourceAccessType a_OldAccessType)
        {
            bool l_Succ = true;

            Enter();
            {
                int l_CntxtIdx = FindContextIndex();
                a_OldAccessType = GetThreadAccessType(l_CntxtIdx);
                Action l_Action = a_InAction;

                if (true == IsSwitchAction(l_Action))
                {
                    l_Action = TranslateSwitchAction(a_OldAccessType, l_Action);

                    if (Action.DoNothing == l_Action)
                    {
                        Leave();

                        return true;
                    }
                }

                switch (l_Action)
                {
                    case Action.AcquireShrd:
                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.None);

                        l_Succ = AcquireShrdCore(ref l_CntxtIdx, a_Wait);

                        if (true == l_Succ)
                        {
                            SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.Shared);
                            f_OwnedShrdCount++;
                        }
                        break;

                    case Action.AcquireExcl:
                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.None);

                        l_Succ = AcquireExclCore(ref l_CntxtIdx, a_Wait);

                        if (true == l_Succ)
                        {
                            SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.Exclusive);
                            f_OwnedExcl = true;
                        }
                        break;

                    case Action.ReleaseShrd:
                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.Shared);

                        ReleaseShrdCore();

                        SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.None);
                        f_OwnedShrdCount--;
                        break;

                    case Action.ReleaseExcl:
                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.Exclusive);

                        ReleaseExclCore();

                        SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.None);
                        f_OwnedExcl = false;
                        break;

                    case Action.UpgradeToExclWithPossibleIntermediateRelease:
                        // Action.UpgradeDirectlyToExcl also lands here.

                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.Shared);

                        l_Succ = UpgradeToExclCore(ref l_CntxtIdx, a_Wait);

                        if (false == l_Succ)
                        {
                            // If a_Wait is true, we would have to
                            // release first (since someone else is
                            // upgrading too currently).

                            if (true == a_Wait &&
                                Action.UpgradeToExclWithPossibleIntermediateRelease == l_Action)
                            {
                                // Same as Action.ReleaseShrd and
                                // followed by Action.AcquireExcl:

                                ReleaseShrdCore();

                                f_OwnedShrdCount--;

                                l_Succ = AcquireExclCore(ref l_CntxtIdx, true);

                                if (false == l_Succ)
                                {
                                    // Should never happen.

                                    throw (new ApplicationException());
                                }
                                else
                                {
                                    SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.Exclusive);
                                    f_OwnedExcl = true;
                                }
                            }
                        }
                        else
                        {
                            SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.Exclusive);
                            f_OwnedShrdCount--;
                            f_OwnedExcl = true;
                        }
                        break;

                    case Action.DowngradeToShrd:
                        AssertSameAccessType(a_OldAccessType, ResourceAccessType.Exclusive);

                        DowngradeToShrdCore();

                        SetThreadAccessType(ref l_CntxtIdx, ResourceAccessType.Shared);
                        f_OwnedShrdCount++;
                        f_OwnedExcl = false;
                        break;

                    case Action.DoNothing:
                        break;

                    case Action.UpgradeDirectlyToExcl:
                        goto case Action.UpgradeToExclWithPossibleIntermediateRelease;

                    default: throw (new ApplicationException());
                }
            }
            Leave();

            return l_Succ;
        }

        private bool AcquireShrdCore(ref int a_CntxtIdx,
                                     bool a_Wait)
        {
            // The caller must hold f_Lock.

            if (true == CanEnterShrd())
            {
                return true;
            }

            if (false == a_Wait)
            {
                return false;
            }

            f_SharedWaiting++;
            {
                do
                {
                    Wait();
                }
                while (false == CanEnterShrd());

                a_CntxtIdx = FindContextIndex();
            }
            f_SharedWaiting--;

            return true;
        }

        private bool AcquireExclCore(ref int a_CntxtIdx,
                                     bool a_Wait)
        {
            // The caller must hold f_Lock.

            if (true == CanEnterExcl())
            {
                return true;
            }

            if (false == a_Wait)
            {
                return false;
            }

            f_ExclWaiting++;
            {
                do
                {
                    Wait();
                }
                while (false == CanEnterExcl());

                a_CntxtIdx = FindContextIndex();
            }
            f_ExclWaiting--;

            return true;
        }

        private void ReleaseShrdCore()
        {
            // The caller must hold f_Lock.

            bool l_Pulse;

            if (0 == f_ExclWaiting)
            {
                l_Pulse = (0 != f_SharedWaiting);
            }
            else
            {
                l_Pulse = ((false == f_TryingToUpgrade ? 1 : 2) == f_OwnedShrdCount);
            }

            if (true == l_Pulse)
            {
                Pulse();
            }
        }

        private void ReleaseExclCore()
        {
            // The caller must hold f_Lock.

            if (0 != f_SharedWaiting ||
                0 != f_ExclWaiting)
            {
                Pulse();
            }
        }

        private bool UpgradeToExclCore(ref int a_CntxtIdx,
                                       bool a_Wait)
        {
            // The caller must hold f_Lock.

            if (true == CanUpgradeToExcl())
            {
                return true;
            }

            if (false == a_Wait ||
                true == f_TryingToUpgrade)
            {
                // We may not wait or someone else already tries to
                // upgrade.

                return false;
            }

            f_ExclWaiting++;
            f_TryingToUpgrade = true;
            {
                do
                {
                    Wait();
                }
                while (false == CanUpgradeToExcl());

                a_CntxtIdx = FindContextIndex();
            }
            f_ExclWaiting--;
            f_TryingToUpgrade = false;

            return true;
        }

        private void DowngradeToShrdCore()
        {
            // The caller must hold f_Lock.

            if (0 != f_SharedWaiting &&
                0 == f_ExclWaiting)
            {
                Pulse();
            }
        }

        private bool CanEnterShrd()
        {
            return (false == f_OwnedExcl &&
                    0 == f_ExclWaiting);
        }

        private bool CanEnterExcl()
        {
            return (0 == f_OwnedShrdCount &&
                    false == f_OwnedExcl);
        }

        private bool CanUpgradeToExcl()
        {
            return (1 == f_OwnedShrdCount &&
                    false == f_OwnedExcl);
        }

        private void Enter()
        {
            Util.MonitorEnter(f_Lock);
        }

        private void Leave()
        {
            Util.MonitorLeave(f_Lock);
        }

        private void Wait()
        {
            // The caller must update any stored index after a series
            // of waits.

            Monitor.Wait(f_Lock);
        }

        private void Pulse()
        {
            Monitor.PulseAll(f_Lock);
        }

        private int FindContextIndex()
        {
            // The caller must hold f_Lock.

            int l_ThreadId = GetCurrentThreadId();

            ThreadAccessContext[] l_Arr = f_Consumers.p_Arr;
            for (int i = (f_Consumers.p_Count - 1); 0 <= i; i--)
            {
                if (l_Arr[i].f_ThreadId == l_ThreadId)
                {
                    return i;
                }
            }

            return (-1);
        }

        private ResourceAccessType GetThreadAccessType(int a_CntxtIdx)
        {
            // The caller must hold f_Lock. a_ContextIdx may be
            // negative if the thread has no context yet.

            return (0 > a_CntxtIdx ? ResourceAccessType.None :
                    f_Consumers.p_Arr[a_CntxtIdx].f_AccessType);
        }

        private void SetThreadAccessType(ref int a_CntxtIdx,
                                         ResourceAccessType a_AccessType)
        {
            // The caller must hold f_Lock. a_ContextIdx may be
            // negative if the thread has no context yet.

            if (0 > a_CntxtIdx)
            {
                if (ResourceAccessType.None != a_AccessType)
                {
                    // Create.

                    a_CntxtIdx = f_Consumers.ReserveAtEnd();
                    f_Consumers.p_Arr[a_CntxtIdx] = new ThreadAccessContext(GetCurrentThreadId(), a_AccessType);
                }
            }
            else
            {
                if (ResourceAccessType.None == a_AccessType)
                {
                    // Delete.

                    f_Consumers.DeleteAt(a_CntxtIdx);
                    a_CntxtIdx = (-1);
                }
                else
                {
                    // Update.

                    f_Consumers.p_Arr[a_CntxtIdx].f_AccessType = a_AccessType;
                }
            }
        }

        private static int GetCurrentThreadId()
        {
            return Thread.CurrentThread.ManagedThreadId;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ThreadAccessContext
        {
            public readonly int f_ThreadId;
            public ResourceAccessType f_AccessType;

            public ThreadAccessContext(int a_ThreadId,
                                       ResourceAccessType a_AccessType)
            {
                f_ThreadId = a_ThreadId;
                f_AccessType = a_AccessType;
            }
        }
    }

    class ResourceDummy : Resource
    {
        private ResourceAccessType f_AccessType;

        public override ResourceAccessType GetAccessType()
        {
            return f_AccessType;
        }

        protected override bool HandleRequest(Action a_InAction,
                                              bool a_Wait,
                                              out ResourceAccessType a_OldAccessType)
        {
            a_OldAccessType = f_AccessType;

            Action l_Action = a_InAction;

            if (true == IsSwitchAction(l_Action))
            {
                l_Action = TranslateSwitchAction(f_AccessType, l_Action);

                if (Action.DoNothing == l_Action)
                {
                    return true;
                }
            }

            switch (l_Action)
            {
                case Action.AcquireShrd:
                    ChangeTypeWithCheck(ResourceAccessType.None, ResourceAccessType.Shared);
                    break;

                case Action.AcquireExcl:
                    ChangeTypeWithCheck(ResourceAccessType.None, ResourceAccessType.Exclusive);
                    break;

                case Action.ReleaseShrd:
                    ChangeTypeWithCheck(ResourceAccessType.Shared, ResourceAccessType.None);
                    break;

                case Action.ReleaseExcl:
                    ChangeTypeWithCheck(ResourceAccessType.Exclusive, ResourceAccessType.None);
                    break;

                case Action.UpgradeDirectlyToExcl:
                case Action.UpgradeToExclWithPossibleIntermediateRelease:
                    ChangeTypeWithCheck(ResourceAccessType.Shared, ResourceAccessType.Exclusive);
                    break;

                case Action.DowngradeToShrd:
                    ChangeTypeWithCheck(ResourceAccessType.Exclusive, ResourceAccessType.Shared);
                    break;

                case Action.DoNothing:
                    break;

                default: throw (new ApplicationException());
            }

            return true;
        }

        private void ChangeTypeWithCheck(ResourceAccessType a_ExpectedAccessType,
                                         ResourceAccessType a_NewAccessType)
        {
            AssertSameAccessType(f_AccessType, a_ExpectedAccessType);
            f_AccessType = a_NewAccessType;
        }
    }

    public class MonitorEvent
    {
        // Our way of implementing EventWaitHandle (without OS, using
        // monitors instead). This is faster and there's no need to
        // close or dispose anything.

        private readonly EventResetMode f_ResetMode;
        private bool f_State;

        public MonitorEvent(bool a_InitialState) :
                       this(a_InitialState, EventResetMode.ManualReset)
        {
        }

        public MonitorEvent(bool a_InitialState,
                            EventResetMode a_ResetMode)
        {
            f_ResetMode = a_ResetMode;
            f_State = a_InitialState;
        }

        public static void CreateManualNotSet(out MonitorEvent a_Dest)
        {
            a_Dest = new MonitorEvent(false, EventResetMode.ManualReset);
        }

        public static void CreateAutoNotSet(out MonitorEvent a_Dest)
        {
            a_Dest = new MonitorEvent(false, EventResetMode.AutoReset);
        }

        public bool IsSet()
        {
            Util.MemoryBarrier();

            return f_State;
        }

        public void Reset()
        {
            Enter();
            {
                f_State = false;
            }
            Leave();
        }

        public void Set()
        {
            Enter();
            {
                f_State = true;

                if (EventResetMode.AutoReset == f_ResetMode)
                {
                    Monitor.Pulse(this);
                }
                else
                {
                    Monitor.PulseAll(this);
                }
            }
            Leave();
        }

        public void Wait()
        {
            Enter();
            {
                while (false == f_State)
                {
                    Monitor.Wait(this);
                }

                if (EventResetMode.AutoReset == f_ResetMode)
                {
                    f_State = false;
                }
            }
            Leave();
        }

        public static void WaitAll(MonitorEvent[] a_Events)
        {
            foreach (MonitorEvent l_Event in a_Events)
            {
                l_Event.Wait();
            }
        }

        private void Enter()
        {
            Util.MonitorEnter(this);
        }

        private void Leave()
        {
            Util.MonitorLeave(this);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct ProcessCPUInfo : IDisposable
    {
        // Don't use the struct default constructor.

        private readonly Process f_Process;
        private readonly ulong f_MaxCPUTicksPerSec;

        private Time f_LastPerfTime;
        private ulong f_LastPerfTicks;

        private ProcessCPUInfo(Process a_Process)
        {
            f_Process = a_Process;
            f_MaxCPUTicksPerSec = ((ulong)TimeSpan.TicksPerSecond * (uint)Environment.ProcessorCount);

            f_LastPerfTime = new Time();
            f_LastPerfTicks = 0;
        }

        public static void Create(out ProcessCPUInfo a_Dest)
        {
            a_Dest = new ProcessCPUInfo(Process.GetCurrentProcess());
        }

        public void Dispose()
        {
            f_Process.Dispose();
        }

        public bool GetInfo(Time a_Time,
                            out float a_CPUUsage,
                            out ulong a_CPUTimeSeconds)
        {
            // On return, a_CPUTimeSeconds will be valid in any case.

            ulong l_CurPerfTicks = (ulong)f_Process.TotalProcessorTime.Ticks;
            a_CPUTimeSeconds = (l_CurPerfTicks / (ulong)TimeSpan.TicksPerSecond);

            float l_Since = a_Time.SecondsSinceAsFloat(f_LastPerfTime);
            bool l_Valid = (2f >= l_Since &&
                            0.1f <= l_Since);

            if (false == l_Valid)
            {
                a_CPUUsage = 0;
            }
            else
            {
                ulong l_TicksPassed = (l_CurPerfTicks - f_LastPerfTicks);
                a_CPUUsage = Math.Min(1f, ((float)l_TicksPassed / f_MaxCPUTicksPerSec) / l_Since);
            }

            f_LastPerfTime = a_Time;
            f_LastPerfTicks = l_CurPerfTicks;

            return l_Valid;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct ThreadSafeCounter32
    {
        // The default constructor may be used.

        private int f_Count;

        public uint p_Count
        {
            get { return (uint)f_Count; }
        }

        public void Inc()
        {
            Interlocked.Increment(ref f_Count);
        }

        public void Dec()
        {
            Interlocked.Decrement(ref f_Count);
        }
    }
}

// End of file.
