/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Support\Util.cs
 * PURPOSE: General helpers
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Support.Encodings;
using Support.HashFunctions;

namespace Support
{
    public static class SupportLibrary
    {
        public static bool Init(uint a_MainModuleReleaseNumber,
                                string a_Folder,
                                out string a_ErrMsg)
        {
            if (33 != a_MainModuleReleaseNumber)
            {
                // The support library isn't aiming at backward/
                // forward compatibility.

                a_ErrMsg = "Version mismatch (Support.dll).";

                return false;
            }

            bool l_VendorConflict = (null != Type.GetType("Mono.Runtime"));
#if MONO
            l_VendorConflict = !l_VendorConflict;
#endif

            if (true == l_VendorConflict)
            {
                a_ErrMsg = "Framework and runtime not supported (Support.dll).";

                return false;
            }

            a_ErrMsg = "Failed to load internal resource.";

            Stream l_Strm = OpenResource(a_Folder, "tiger.boxes");
            {
                if (false == Tiger.LoadBoxes(l_Strm))
                {
                    return false;
                }
            }
            l_Strm.Dispose();

#if FAST
            l_Strm = OpenResource(a_Folder, "sha256.boxes");
            {
                if (false == SHA_256.LoadBoxesForFastCode(l_Strm))
                {
                    return false;
                }
            }
            l_Strm.Dispose();
#endif

            Util.MemoryBarrier();

            if (false == SelfTest())
            {
                a_ErrMsg = "Self test failed.";

                return false;
            }

            a_ErrMsg = null;

            return true;
        }

        private static Stream OpenResource(string a_Folder,
                                           string a_ResourceName)
        {
            Assembly l_Assembly = Assembly.GetExecutingAssembly();
            Stream l_Strm = l_Assembly.GetManifestResourceStream(a_ResourceName);

            if (null == l_Strm)
            {
                // We may land here if we were built through Visual
                // Studio.

                System.Resources.ResourceManager l_ResMngr = new System.Resources.ResourceManager("Support.Resources", l_Assembly);
                object l_ResObj = l_ResMngr.GetObject(a_ResourceName);

                l_Strm = new MemoryStream((byte[])l_ResObj, false);
            }

            return l_Strm;
        }

        private static bool SelfTest()
        {
            bool l_Hashes = (true == HMACWrapper.SelfTest() &&
                             true == MD_4.SelfTest() &&
                             true == MD_5.SelfTest() &&
                             true == RIPEMD_160.SelfTest() &&
                             true == SHA_1.SelfTest() &&
                             true == SHA_256.SelfTest() &&
                             true == SHA_384.SelfTest() &&
                             true == SHA_512.SelfTest() &&
                             true == Keccak.SelfTest() &&
                             true == Adler32.SelfTest() &&
                             true == CRC32.SelfTest() &&
                             true == Lookup3.SelfTest() &&
                             true == SipHash.SelfTest() &&
                             true == Tiger.SelfTest());

            return l_Hashes;
        }
    }

    public static class Clock
    {
        // Implements a clock with high resolution that is prone to
        // system time changes.

        private static readonly TimeSources s_TimeSource;
        private static readonly ulong s_BaseTicks;

        private static ulong s_LastValue;
        // s_LastTickCount32 is used by QueryThroughFramework() only.
        private static uint s_LastTickCount32;

        static Clock()
        {
#if FAST
            try
            {
                ulong l_Junk = QueryHighResolutionTimer();
                s_TimeSource = TimeSources.NativeHighResolutionTimer;
            }
            catch (BadImageFormatException)
            {
            }
            catch (TypeLoadException)
            {
            }

            if (TimeSources.Undefined == s_TimeSource)
            {
                try
                {
                    ulong l_Junk = GetTickCount64();
                    s_TimeSource = TimeSources.NativeTickCount64;
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
            }
#endif

            if (TimeSources.Undefined == s_TimeSource)
            {
                s_TimeSource = TimeSources.FrameworkTickCount;
            }

            // NOTE: Run the following code at least twice, since there
            // can be costly initializations (type constructors) when run
            // the first time, so that the results possibly aren't good
            // enough at first.

            const byte l_RoundsRequired = 2;
            uint l_Round = 0;
            ulong l_RawTicks;

            do
            {
                // We need a fairly huge base in any case that allows
                // us to subtract without care.
                s_BaseTicks = Math.Max((ulong)DateTime.UtcNow.Ticks, 366UL * TimeSpan.TicksPerDay);
                l_RawTicks = QueryTicksCore();

                if (l_RawTicks >= s_BaseTicks)
                {
                    s_BaseTicks = 0;
                }
                else
                {
                    s_BaseTicks -= l_RawTicks;
                }
            }
            while (l_RoundsRequired != ++l_Round);
        }

        public static ulong QueryTicks()
        {
            // Returns the time in 100 nanoseconds. The returned value
            // represents the UTC system time of the moment the
            // application was launched plus all time that passed
            // since that moment. System time changes after launch are
            // irrelevant. The UTC system time and the return value of
            // this routine are quite the same value, unless a (dramatic)
            // system time change happens during the lifetime of the
            // process.

            return (s_BaseTicks + QueryTicksCore());
        }

        private static ulong QueryTicksCore()
        {
            // Returns the time as a number that gets incremented by
            // one every 100 nanoseconds. The actual precision is
            // usually much worse. Normally the number represents the
            // number of ticks since the system started. This routine
            // will never return a smaller number than before when
            // invoked subsequently. This routine most likely performs
            // a syscall.

            switch (s_TimeSource)
            {
#if FAST
                case TimeSources.NativeHighResolutionTimer:
                    return QueryHighResolutionTimer();
#endif

                case TimeSources.NativeTickCount64:
                    return (GetTickCount64() * (uint)TimeSpan.TicksPerMillisecond);

                case TimeSources.FrameworkTickCount:
                    return QueryThroughFramework();

                default: throw (new UnexpectedStateException());
            }
        }

#if FAST
        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern bool HiResTimerQuery_x86([Out] out ulong a_Ticks);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.StdCall)]
        private static extern bool HiResTimerQuery_x64([Out] out ulong a_Ticks);

        private static ulong QueryHighResolutionTimer()
        {
            bool l_Succ;
            ulong l_Ticks;

            if (sizeof(uint) == IntPtr.Size)
            {
                l_Succ = HiResTimerQuery_x86(out l_Ticks);
            }
            else
            {
                l_Succ = HiResTimerQuery_x64(out l_Ticks);
            }

            if (false == l_Succ)
            {
                // Should never happen.

                throw (new UnexpectedStateException());
            }

            // Make sure that the value doesn't get smaller than it
            // previously was.

            Util.MonitorEnter(typeof(Clock));
            {
                if (l_Ticks <= s_LastValue)
                {
                    l_Ticks = s_LastValue;
                }
                else
                {
                    s_LastValue = l_Ticks;
                }
            }
            Util.MonitorLeave(typeof(Clock));

            return l_Ticks;
        }
#endif

        // GetTickCount64() is an optional way of retrieving a low
        // resolution timer. It requires NT6.
        [DllImport("Kernel32.dll", CallingConvention = CallingConvention.Winapi)]
        private static extern ulong GetTickCount64();

        private static ulong QueryThroughFramework()
        {
            // This approach works only if this routine is at least
            // called once every few days while the application is
            // running (we assume that this is always so). Otherwise
            // we can't determine how many times the tick value wrapped
            // around in between two calls.

            // TickCount is a uint in fact.
            uint l_TickCount32 = (uint)Environment.TickCount;
            ulong l_Ticks;

            Util.MonitorEnter(typeof(Clock));
            {
                ulong l_Milliseconds;

                if (l_TickCount32 > s_LastTickCount32)
                {
                    l_Milliseconds = (l_TickCount32 - s_LastTickCount32);
                }
                else if (l_TickCount32 == s_LastTickCount32)
                {
                    l_Milliseconds = 0;
                }
                else
                {
                    const uint l_CreditableOverflowThreshold = (1u << 24);

                    if (l_CreditableOverflowThreshold > (s_LastTickCount32 - l_TickCount32))
                    {
                        // The value didn't wrap around, but a threading
                        // issue occurred (so return the old count).

                        l_Milliseconds = 0;
                    }
                    else
                    {
                        // An overflow really happened.

                        l_Milliseconds = (uint.MaxValue - s_LastTickCount32);
                        l_Milliseconds += (1 + l_TickCount32);
                    }
                }

                if (0 != l_Milliseconds)
                {
                    s_LastValue += (l_Milliseconds * Time.c_TicksPerMillisecond);
                    s_LastTickCount32 = l_TickCount32;
                }

                l_Ticks = s_LastValue;
            }
            Util.MonitorLeave(typeof(Clock));

            return l_Ticks;
        }

        private enum TimeSources : byte
        {
            Undefined = 0,

            NativeHighResolutionTimer = 1,
            NativeTickCount64 = 2,
            FrameworkTickCount = 3,
        }
    }

    public struct Time : IEquatable<Time>, IComparable, IComparable<Time>
    {
        public const uint c_SecsPerWeek = (7 * c_SecsPerDay);
        public const uint c_SecsPerDay = (24 * c_SecsPerHour);
        public const uint c_SecsPerHour = (60 * c_SecsPerMinute);
        public const uint c_SecsPerHalfHour = (30 * c_SecsPerMinute);
        public const uint c_SecsPerQuarterHour = (15 * c_SecsPerMinute);
        public const uint c_SecsPerMinute = (60 * c_Second);
        public const uint c_SecsPerHalfMinute = (30 * c_Second);
        public const uint c_Second = 1;

        public const uint c_TicksPerSecond = (uint)TimeSpan.TicksPerSecond;
        public const uint c_TicksPerMillisecond = (uint)TimeSpan.TicksPerMillisecond;

        private const ulong c_MaxDateTimeTicks = 0x2BCA2875F4373FFF;
        // 1-Jan-1601 00:00:00 UTC
        private const ulong c_WinFileTimeBaseTicks = 0x701CE1722770000;

        // f_Ticks is in 100 nanoseconds.
        private readonly ulong f_Ticks;

        private Time(ulong a_Ticks)
        {
            f_Ticks = a_Ticks;
        }

        public override bool Equals(object a_Other)
        {
            throw (new ApplicationException());
        }

        public bool Equals(Time a_Other)
        {
            return (this.f_Ticks == a_Other.f_Ticks);
        }

        public override int GetHashCode()
        {
            return f_Ticks.GetHashCode();
        }

        public bool IsSet()
        {
            return (0 != f_Ticks);
        }

        public Time AddTicks(ulong a_Ticks)
        {
            return new Time(f_Ticks + a_Ticks);
        }

        public Time AddMilliseconds(uint a_Millisecs)
        {
            return new Time(f_Ticks + ((ulong)a_Millisecs * TimeSpan.TicksPerMillisecond));
        }

        public long MillisecondsSince(Time a_Other)
        {
            return (TicksSince(a_Other) / TimeSpan.TicksPerMillisecond);
        }

        public Time AddSeconds(uint a_Secs)
        {
            return new Time(f_Ticks + ((ulong)a_Secs * TimeSpan.TicksPerSecond));
        }

        public Time SubSeconds(uint a_Secs)
        {
            return new Time(f_Ticks - ((ulong)a_Secs * TimeSpan.TicksPerSecond));
        }

        public long SecondsSince(Time a_Other)
        {
            return (TicksSince(a_Other) / TimeSpan.TicksPerSecond);
        }

        public long FullDaysSince(Time a_Other)
        {
            return (TicksSince(a_Other) / TimeSpan.TicksPerDay);
        }

        private long TicksSince(Time a_Other)
        {
            return ((long)this.f_Ticks - (long)a_Other.f_Ticks);
        }

        public float SecondsSinceAsFloat(Time a_Other)
        {
            long l_Diff = ((long)this.f_Ticks - (long)a_Other.f_Ticks);

            return ((float)l_Diff / TimeSpan.TicksPerSecond);
        }

        public ulong AsSeconds()
        {
            return (f_Ticks / TimeSpan.TicksPerSecond);
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((Time)a_Other);
        }

        public int CompareTo(Time a_Other)
        {
            return this.f_Ticks.CompareTo(a_Other.f_Ticks);
        }

        public ulong ToBinary()
        {
            return f_Ticks;
        }

        public static Time FromBinary(ulong a_Num)
        {
            return new Time(a_Num);
        }

        public DateTime ToDateTime()
        {
            return new DateTime((long)Math.Min(f_Ticks, c_MaxDateTimeTicks), DateTimeKind.Utc);
        }

        public static Time FromDateTime(DateTime a_DateTime)
        {
            return new Time((ulong)a_DateTime.Ticks);
        }

        public static Time FromWinFileTime(System.Runtime.InteropServices.ComTypes.FILETIME a_FileTime)
        {
            ulong l_TicksSinceBase = (((ulong)a_FileTime.dwHighDateTime << Util.c_BitsPerUInt) | (uint)a_FileTime.dwLowDateTime);
            ulong l_Ticks = (c_WinFileTimeBaseTicks + l_TicksSinceBase);

            return new Time(c_WinFileTimeBaseTicks > l_Ticks ? ulong.MaxValue : l_Ticks);
        }

        public static Time Min(Time a_A,
                               Time a_B)
        {
            return (a_A < a_B ? a_A : a_B);
        }

        public static Time Max(Time a_A,
                               Time a_B)
        {
            return (a_A > a_B ? a_A : a_B);
        }

        public static Time Now()
        {
            // The return value is not prone to system time changes.

            ulong l_Ticks = Clock.QueryTicks();

            return new Time(l_Ticks);
        }

        public static Time SysTimeNow()
        {
            return new Time((ulong)DateTime.UtcNow.Ticks);
        }

        public static Time SysTimeLastMidnight()
        {
            return new Time((ulong)DateTime.UtcNow.Date.Ticks);
        }

        public bool IsEarlierThan(Time a_Other)
        {
            return (this.f_Ticks < a_Other.f_Ticks);
        }

        public static bool operator ==(Time a_A,
                                       Time a_B)
        {
            return (a_A.f_Ticks == a_B.f_Ticks);
        }

        public static bool operator !=(Time a_A,
                                       Time a_B)
        {
            return (a_A.f_Ticks != a_B.f_Ticks);
        }

        public static bool operator <(Time a_A,
                                      Time a_B)
        {
            return (a_A.f_Ticks < a_B.f_Ticks);
        }

        public static bool operator >(Time a_A,
                                      Time a_B)
        {
            return (a_A.f_Ticks > a_B.f_Ticks);
        }

        public static bool operator <=(Time a_A,
                                       Time a_B)
        {
            return (a_A.f_Ticks <= a_B.f_Ticks);
        }

        public static bool operator >=(Time a_A,
                                       Time a_B)
        {
            return (a_A.f_Ticks >= a_B.f_Ticks);
        }
    }

    public struct StorableTime : IEquatable<StorableTime>, IComparable, IComparable<StorableTime>
    {
        // This struct represents the number of 2000-ms units since
        // January 1, 1970 (UTC).

        public const byte c_SecondsPerUnit = 2;
        private const uint c_TicksPerUnit = (c_SecondsPerUnit * Time.c_TicksPerSecond);

        private readonly uint f_Units;

        private StorableTime(uint a_Units)
        {
            f_Units = a_Units;
        }

        public override bool Equals(object a_Other)
        {
            throw (new ApplicationException());
        }

        public bool Equals(StorableTime a_Other)
        {
            return (this.f_Units == a_Other.f_Units);
        }

        public override int GetHashCode()
        {
            return f_Units.GetHashCode();
        }

        public bool IsSet()
        {
            return (0 != f_Units);
        }

        public Time ToTime()
        {
            if (0 == f_Units)
            {
                return new Time();
            }

            ulong l_Ticks = (0x89F7FF5F7B58000UL + Util.Mul64(f_Units - 1, c_TicksPerUnit));

            return Time.FromBinary(l_Ticks);
        }

        public static StorableTime FromTime(Time a_Time)
        {
            bool l_Junk;

            return FromTime(a_Time, false, out l_Junk);
        }

        public static StorableTime FromTime(Time a_Time,
                                            bool a_RoundUp,
                                            out bool a_Clamped)
        {
            // See comment in SlimTime.FromTime() about the meaning of
            // the arguments.

            // NOTE: Zero represents a undefined value. One represents
            // the lower limit and uint.MaxValue represents the upper
            // limit.

            const byte lc_BaseUnits = 1;
            // 1-Jan-1970 00:00:00 UTC.
            const ulong lc_LowerLimit = 0x89F7FF5F7B58000UL;
            // 16-March-2242 12:56:28 UTC.
            const ulong lc_UpperLimit = (lc_LowerLimit + ((ulong)(uint.MaxValue - lc_BaseUnits) * c_TicksPerUnit));

            ulong l_Ticks = a_Time.ToBinary();

            if (lc_LowerLimit > l_Ticks)
            {
                // a_Time is too low (potentially).

                a_Clamped = (false == a_RoundUp ||
                             (lc_LowerLimit - c_TicksPerUnit) >= l_Ticks);

                return new StorableTime(lc_BaseUnits);
            }

            if (lc_UpperLimit < l_Ticks)
            {
                // a_Time is too high (potentially).

                a_Clamped = (true == a_RoundUp ||
                             (lc_UpperLimit + c_TicksPerUnit) <= l_Ticks);

                return new StorableTime(uint.MaxValue);
            }

            uint l_Units = (uint)((l_Ticks - lc_LowerLimit) / c_TicksPerUnit);

            if (true == a_RoundUp &&
                0 != (l_Ticks % c_TicksPerUnit))
            {
                l_Units++;
            }

            a_Clamped = false;

            return new StorableTime(lc_BaseUnits + l_Units);
        }

        public int CompareTo(object a_Other)
        {
            return CompareTo((StorableTime)a_Other);
        }

        public int CompareTo(StorableTime a_Other)
        {
            return this.f_Units.CompareTo(a_Other.f_Units);
        }

        public uint GetAsUInt32()
        {
            return f_Units;
        }

        public static StorableTime FromUInt32(uint a_Units)
        {
            return new StorableTime(a_Units);
        }

        public static StorableTime Min(StorableTime a_A,
                                       StorableTime a_B)
        {
            return (a_A < a_B ? a_A : a_B);
        }

        public static StorableTime Max(StorableTime a_A,
                                       StorableTime a_B)
        {
            return (a_A > a_B ? a_A : a_B);
        }

        public static bool operator ==(StorableTime a_A,
                                       StorableTime a_B)
        {
            return (a_A.f_Units == a_B.f_Units);
        }

        public static bool operator !=(StorableTime a_A,
                                       StorableTime a_B)
        {
            return !(a_A == a_B);
        }

        public static bool operator <(StorableTime a_A,
                                      StorableTime a_B)
        {
            return (a_A.f_Units < a_B.f_Units);
        }

        public static bool operator >(StorableTime a_A,
                                      StorableTime a_B)
        {
            return (a_A.f_Units > a_B.f_Units);
        }

        public static bool operator <=(StorableTime a_A,
                                       StorableTime a_B)
        {
            return (a_A.f_Units <= a_B.f_Units);
        }

        public static bool operator >=(StorableTime a_A,
                                       StorableTime a_B)
        {
            return (a_A.f_Units >= a_B.f_Units);
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct PermutationGenerator
    {
        // Based on code placed under the public domain by Michael
        // Gilleland. See: <http://www.merriampark.com/perm.htm>

        public const byte c_MinElements = 1;
        public const byte c_MaxElements = 12;

        public readonly int f_ElementCount;
        public readonly uint f_PermutationCount;
        private readonly int[] f_Arr;

        private uint f_Remaining;

        public PermutationGenerator(int a_Count)
        {
            f_ElementCount = a_Count;
            f_PermutationCount = GetFactorial(f_ElementCount);
            f_Arr = new int[f_ElementCount];

            f_Remaining = 0;

            BuildFirst();
        }

        public bool IsValid()
        {
            return (null != f_Arr);
        }

        public void BuildFirst()
        {
            for (int i = 0; i < f_ElementCount; i++)
            {
                f_Arr[i] = i;
            }

            f_Remaining = (f_PermutationCount - 1);
        }

        public bool HasNext()
        {
            return (0 != f_Remaining);
        }

        public void BuildNext()
        {
            // Algorithm from Kenneth H. Rosen, Discrete Mathematics and
            // Its Applications, 2nd edition (NY: McGraw-Hill, 1991),
            // pp. 282-284.

            // Find largest index j with a[j] < a[j+1]

            int j = f_ElementCount - 2;
            while (f_Arr[j] > f_Arr[j + 1])
            {
                j--;
            }

            // Find index k such that a[k] is smallest integer
            // greater than a[j] to the right of a[j]

            int k = f_ElementCount - 1;
            while (f_Arr[j] > f_Arr[k])
            {
                k--;
            }

            // Interchange a[j] and a[k]

            int temp = f_Arr[k];
            f_Arr[k] = f_Arr[j];
            f_Arr[j] = temp;

            // Put tail end of permutation after jth position in increasing order

            int r = f_ElementCount - 1;
            int s = j + 1;

            while (r > s)
            {
                temp = f_Arr[s];
                f_Arr[s] = f_Arr[r];
                f_Arr[r] = temp;
                r--;
                s++;
            }

            f_Remaining--;
        }

        public int[] ExposeArray()
        {
            return f_Arr;
        }

        private static uint GetFactorial(int a_Count)
        {
            // If l_Count is more than 12, our 32 bit integer
            // will overflow.

            uint l_Res = 1;

            for (uint i = 1; i <= a_Count; i++)
            {
                l_Res *= i;
            }

            return l_Res;
        }
    }

    public static partial class Util
    {
        public const uint c_NoIndex = uint.MaxValue;

        public const byte c_BitsPerNibble = (c_BitsPerByte / 2);
        public const byte c_BitsPerByte = 8;
        public const byte c_BitsPerUShort = (sizeof(ushort) * c_BitsPerByte);
        public const byte c_BitsPerUInt = (sizeof(uint) * c_BitsPerByte);
        public const byte c_BitsPerULong = (sizeof(ulong) * c_BitsPerByte);

        public const byte c_HexCharsPerByte = 2;
        public const byte c_HexCharsPerUInt = (sizeof(uint) * c_HexCharsPerByte);

        public const char c_Space = '\x0020';
        public const char c_WildcardUndefinedChar = '?';
        public const char c_WildcardAnySeqChar = '*';

        public const byte c_Max7BitEncUInt32ByteLen = 5;
        private const string c_LatinReplacementForUnsupportedUnicodeChars = "~";

        // In milliseconds.
        public const byte c_MiniSleepTimeout = 10;

        // This is the codepage of ISO-8859-1, not that of
        // Windows-1252.
        private const int c_Latin1Codepage = 28591;

        public static readonly System.Globalization.CultureInfo s_InvariantCulture;
        public static readonly byte[] s_Empty;
        public static readonly bool s_IsWindows;

        public static readonly Encoding s_Latin1BestFit;
        public static readonly Encoding s_UTF8Picky;

        public static readonly bool s_IsRunningOnWine;

        private static bool s_SocketCheckTried;
        private static bool s_SocketCheckPerformed;

        public static bool sp_NeedToSendToTestConnLife
        {
            get { return !s_SocketCheckPerformed; }
        }

        static Util()
        {
            s_InvariantCulture = System.Globalization.CultureInfo.InvariantCulture;
            s_Empty = new byte[0];
            s_IsWindows = IsWindows();

            // We use the ascii character tilde as a replacement char,
            // since it can be included in file names (contrary to
            // question mark) and it is not a character that leads
            // to keyword separation (like underscore).
            s_Latin1BestFit = Encoding.GetEncoding(c_Latin1Codepage,
                                                   new EncoderReplacementFallback(c_LatinReplacementForUnsupportedUnicodeChars),
                                                   new DecoderReplacementFallback(c_LatinReplacementForUnsupportedUnicodeChars));
            const int l_UTF8Codepage = 65001;
            s_UTF8Picky = Encoding.GetEncoding(l_UTF8Codepage, new EncoderExceptionFallback(), new DecoderExceptionFallback());
            s_IsRunningOnWine = IsRunningOnWine();
        }

        public static void MonitorEnter(object a_Obj)
        {
            Monitor.Enter(a_Obj);
        }

        public static void MonitorLeave(object a_Obj)
        {
            Monitor.Exit(a_Obj);
        }

        private static bool IsRunningOnWine()
        {
            try
            {
                IntPtr l_StrPtr = wine_get_version();

                return true;
            }
            catch (BadImageFormatException)
            {
            }
            catch (TypeLoadException)
            {
            }

            return false;
        }

        [DllImport("ntdll.dll", CallingConvention = CallingConvention.StdCall)]
        private static extern IntPtr wine_get_version();

        public static bool EnsureDirectoryExists(string a_Dir)
        {
            // Returns true if the directory seems to exist.

            if (false == Directory.Exists(a_Dir))
            {
                try
                {
                    return FHandle.TryCreateDirectory(a_Dir);
                }
                catch (ArgumentNullException l_Exc)
                {
                    throw (l_Exc);
                }
                catch
                {
                }

                return false;
            }

            return true;
        }

        public static bool ShellOpenOrExecuteFile(string a_FullFileName)
        {
            if (false == File.Exists(a_FullFileName))
            {
                return false;
            }

            return ShellExecuteCore(a_FullFileName);
        }

        public static bool ShellOpenDirectory(string a_DirPath)
        {
            if (false == Directory.Exists(a_DirPath))
            {
                return false;
            }

            return ShellExecuteCore(a_DirPath);
        }

        private static bool ShellExecuteCore(string a_FileName)
        {
            ProcessStartInfo l_StartInfo = new ProcessStartInfo();
            l_StartInfo.FileName = a_FileName;
            l_StartInfo.Verb = "open";

            Process l_Process;

            try
            {
                l_Process = Process.Start(l_StartInfo);
            }
            catch (ArgumentException l_Exc)
            {
                throw (l_Exc);
            }
            catch
            {
                return false;
            }

            // NOTE: l_Process can be null even on success, e.g. if
            // the Explorer handled the command.

            if (null != l_Process)
            {
                // Simply close the handle to this process, but let it
                // run as long as it likes.
                l_Process.Dispose();
            }

            return true;
        }

        public static void SetClipboard(string a_Str)
        {
            // This must be done by a STA thread.

            if (0 == a_Str.Length)
            {
                Clipboard.Clear();
            }
            else
            {
                // Trying to set empty text throws an exception.

                Clipboard.SetText(a_Str);
            }
        }

        public static string UnicodeNormalizeString(string a_String)
        {
            // This will (tested) return the input string if the
            // input string is already normalized in form C.

            return a_String.Normalize(NormalizationForm.FormC);
        }

        public static bool IsStringUnicodeNormalized(string a_String)
        {
            return a_String.IsNormalized(NormalizationForm.FormC);
        }

        public static bool IsLatin1AndUTF8Equivalent(string a_Str)
        {
            // Returns true if a_Str can be expressed in a byte
            // sequence that is interpreted as the same text by
            // the latin1 and the UTF-8 encoding.

            foreach (char l_Char in a_Str)
            {
                if (sbyte.MaxValue < l_Char)
                {
                    return false;
                }
            }

            return true;
        }

        public static bool IsLatin1Compatible(string a_Str)
        {
            // Returns a nonzero value if the UTF-16 string a_Str can
            // be expressed as latin1 string.

            foreach (char l_Char in a_Str)
            {
                if (byte.MaxValue < l_Char)
                {
                    return false;
                }
            }

            return true;
        }

        public static bool GetUTF8ByteLen(string a_Str,
                                          uint a_StrIdx,
                                          uint a_StrLen,
                                          out uint a_ByteLen)
        {
            // Returns zero if the length could not be determined. This
            // routine does not throw an exception on any encoding errors.

            Encoding l_UTF8 = Encoding.UTF8;

            if (0 == a_StrIdx &&
                (uint)a_Str.Length == a_StrLen)
            {
                a_ByteLen = (uint)l_UTF8.GetByteCount(a_Str);

                return true;
            }

#if FAST
            Util.VerifyRangeOrFail((int)a_StrIdx, (int)a_StrLen, a_Str.Length);

            unsafe
            {
                fixed (char* l_Ptr = a_Str)
                {
                    a_ByteLen = (uint)l_UTF8.GetByteCount(&l_Ptr[a_StrIdx], (int)a_StrLen);

                    return true;
                }
            }
#else
            char[] l_CharArr = a_Str.ToCharArray((int)a_StrIdx, (int)a_StrLen);
            a_ByteLen = (uint)l_UTF8.GetByteCount(l_CharArr);

            return true;
#endif
        }

        public static bool IsUTF16ShorterThanUTF8(string a_Str,
                                                  bool a_PreferUTF16)
        {
            return IsUTF16ShorterThanUTF8(a_Str, 0, (uint)a_Str.Length, a_PreferUTF16);
        }

        public static bool IsUTF16ShorterThanUTF8(string a_Str,
                                                  uint a_Idx,
                                                  uint a_Len,
                                                  bool a_PreferUTF16)
        {
            // a_PreferUTF16 is the return value if no clear answer
            // can be given.

            uint l_UTF8ByteLen;

            if (0 != a_Len &&
                true == GetUTF8ByteLen(a_Str, a_Idx, a_Len, out l_UTF8ByteLen))
            {
                uint l_UTF16ByteLen = (a_Len * sizeof(char));

                if (l_UTF8ByteLen != l_UTF16ByteLen)
                {
                    return (l_UTF8ByteLen > l_UTF16ByteLen);
                }
            }

            return a_PreferUTF16;
        }

        public static bool IsCharInArray(char a_Char,
                                         char[] a_Arr)
        {
            foreach (char l_Char in a_Arr)
            {
                if (l_Char == a_Char)
                {
                    return true;
                }
            }

            return false;
        }

        public static bool IsStringLowercase(string a_RawStr,
                                             out string a_LowercaseStr)
        {
            string l_Str = a_RawStr.ToLowerInvariant();

            if (true == StringsEqual(a_RawStr, l_Str))
            {
                a_LowercaseStr = a_RawStr;

                return true;
            }

            a_LowercaseStr = l_Str;

            return false;
        }

        public static string[] Split(string a_Str,
                                     int a_Idx,
                                     char a_Delimiter)
        {
            return Split(a_Str, a_Idx, a_Str.Length - a_Idx, a_Delimiter);
        }

        public static string[] Split(string a_Str,
                                     int a_Idx,
                                     int a_Len,
                                     char a_Delimiter)
        {
            string l_Core = a_Str.Substring(a_Idx, a_Len);

            return l_Core.Split(new char[] { a_Delimiter }, StringSplitOptions.None);
        }

        public static string Join(char a_SeparationChar,
                                  char[] a_Elements)
        {
            int l_Count = a_Elements.Length;
            StringBuilder l_Bld = new StringBuilder(l_Count * 2);

            for (int i = 0, lim = (l_Count - 1); i <= lim; i++)
            {
                l_Bld.Append(a_Elements[i]);

                if (i < lim)
                {
                    l_Bld.Append(a_SeparationChar);
                }
            }

            return l_Bld.ToString();
        }

        public static bool StartsWith(string a_LongStr,
                                      string a_PossibleStarting)
        {
            return a_LongStr.StartsWith(a_PossibleStarting, StringComparison.InvariantCulture);
        }

        public static bool EndsWith(string a_LongStr,
                                    string a_PossibleEnding)
        {
            return a_LongStr.EndsWith(a_PossibleEnding, StringComparison.InvariantCulture);
        }

        public static bool StringsEqual(string a_A,
                                        string a_B)
        {
            return string.Equals(a_A, a_B, StringComparison.InvariantCulture);
        }

        public static bool StringsEqualIgnoreCase(string a_A,
                                                  string a_B)
        {
            return string.Equals(a_A, a_B, StringComparison.InvariantCultureIgnoreCase);
        }

        public static bool StringsEqual(string a_A,
                                        string a_B,
                                        bool a_CaseSensitive)
        {
            return string.Equals(a_A, a_B, false == a_CaseSensitive ? StringComparison.InvariantCultureIgnoreCase : StringComparison.InvariantCulture);
        }

        public static bool StringsEqual(string a_A,
                                        uint a_AIdx,
                                        uint a_ALen,
                                        string a_B,
                                        uint a_BIdx,
                                        uint a_BLen)
        {
            return StringsEqual(a_A, a_AIdx, a_ALen, a_B, a_BIdx, a_BLen, true);
        }

        public static bool StringsEqual(string a_A,
                                        uint a_AIdx,
                                        uint a_ALen,
                                        string a_B,
                                        uint a_BIdx,
                                        uint a_BLen,
                                        bool a_CaseSensitive)
        {
            return (a_ALen == a_BLen &&
                    0 == string.Compare(a_A, (int)a_AIdx, a_B, (int)a_BIdx, (int)a_ALen, false == a_CaseSensitive ? StringComparison.InvariantCultureIgnoreCase : StringComparison.InvariantCulture));
        }

        public static int CompareStringsOrdinal(string a_A,
                                                string a_B)
        {
            return CompareStringsOrdinal(a_A, 0, GetStringLength(a_A), a_B, 0, GetStringLength(a_B));
        }

        public static int CompareStringsOrdinal(string a_A,
                                                uint a_IdxA,
                                                uint a_LenA,
                                                string a_B,
                                                uint a_IdxB,
                                                uint a_LenB)
        {
            if (0 == a_LenA)
            {
                return (0 == a_LenB ? 0 : (-1));
            }

            if (0 == a_LenB)
            {
                return 1;
            }

            UTF16Reader l_RdrA = new UTF16Reader(a_A, a_IdxA, a_LenA);
            UTF16Reader l_RdrB = new UTF16Reader(a_B, a_IdxB, a_LenB);

            while (true)
            {
                uint l_CodePointA;
                bool l_GotA = l_RdrA.TryReadNextCodePoint(out l_CodePointA);

                uint l_CodePointB;
                bool l_GotB = l_RdrB.TryReadNextCodePoint(out l_CodePointB);

                if (false == l_GotA)
                {
                    return (false == l_GotB ? 0 : (-1));
                }

                if (false == l_GotB)
                {
                    return 1;
                }

                if (l_CodePointA != l_CodePointB)
                {
                    return (l_CodePointA < l_CodePointB ? (-1) : 1);
                }
            }
        }

        public static bool FindString(string a_A,
                                      uint a_AIdx,
                                      uint a_ALen,
                                      string a_B,
                                      uint a_BIdx,
                                      uint a_BLen,
                                      bool a_CaseSensitive,
                                      out uint a_TrgtIdx)
        {
            // Tries to find the segment of a_B in the segment of a_A.
            // Expects both strings to be normalized.

            if (0 == a_BLen ||
                a_ALen < a_BLen)
            {
                goto FAIL;
            }

            uint l_AFinalIdx = (a_AIdx + a_ALen - a_BLen + 1);
            char l_FirstChar = a_B[(int)a_BIdx];

            if (false == a_CaseSensitive)
            {
                l_FirstChar = char.ToLowerInvariant(l_FirstChar);
            }

            for (uint i = a_AIdx; i != l_AFinalIdx; i++)
            {
                char l_CurChar = a_A[(int)i];

                if (false == a_CaseSensitive)
                {
                    l_CurChar = char.ToLowerInvariant(l_CurChar);
                }

                if (l_CurChar != l_FirstChar)
                {
                    continue;
                }

                // The first char matches.

                uint l_RestLen = (a_BLen - 1);

                if (0 == l_RestLen ||
                    true == StringsEqual(a_A, i + 1, l_RestLen, a_B, a_BIdx + 1, l_RestLen, a_CaseSensitive))
                {
                    a_TrgtIdx = i;

                    return true;
                }
            }

            FAIL:

            a_TrgtIdx = Util.c_NoIndex;

            return false;
        }

        public static bool IsLatinLetter(char a_Char)
        {
            return (('A' <= (a_Char) &&
                     'Z' >= (a_Char)) ||
                    ('a' <= (a_Char) &&
                     'z' >= (a_Char)));
        }

        public static bool IsSameChar(char a_A,
                                      char a_B,
                                      bool a_CaseSensitive)
        {
            if (true == a_CaseSensitive)
            {
                return (a_A == a_B);
            }

            return (char.ToLowerInvariant(a_A) == char.ToLowerInvariant(a_B));
        }

        public static int IndexOfIgnoreCase(string a_LongStr,
                                            string a_PossibleSubstr)
        {
            // Returns a negative value if a_LongStr does not contain
            // a_PossibleSubstr.

            return a_LongStr.IndexOf(a_PossibleSubstr, StringComparison.InvariantCultureIgnoreCase);
        }

        public static int CountOccurrencesOfChar(string a_Str,
                                                 char a_Char)
        {
            int l_Count = 0;

            foreach (char l_Char in a_Str)
            {
                if (l_Char == a_Char)
                {
                    l_Count++;
                }
            }

            return l_Count;
        }

        public static
#if FAST
        unsafe
#endif
        int FindCharIndex(string a_Str,
                          char a_Char,
                          int a_OccurrenceCount)
        {
            // Finds the n-th occurrence of the char in the string,
            // couting from the start, or a negative value if the char
            // does not occur that often.

            int l_Count = 0;

#if FAST
            fixed (char* l_TempStr = a_Str)
            {
#else
            string l_TempStr = a_Str;
#endif
            for (int i = 0, lim = a_Str.Length; i < lim; i++)
            {
                if (a_Char == l_TempStr[i])
                {
                    if (++l_Count >= a_OccurrenceCount)
                    {
                        return i;
                    }
                }
            }
#if FAST
            }
#endif

            return (-1);
        }

        public static void TransformNewLineIntoPlatformLineBreak(StringBuilder a_Bld,
                                                                 int a_Idx,
                                                                 int a_Len)
        {
            if (1 == Environment.NewLine.Length)
            {
                return;
            }

            // Assume the new line string is "\r\n".

            int l_FinalIdx = (a_Idx + a_Len);
            for (int i = a_Idx; i < l_FinalIdx; i++)
            {
                if ('\n' == a_Bld[i])
                {
                    if (i == a_Idx ||
                        '\r' != a_Bld[i - 1])
                    {
                        a_Bld.Insert(i, '\r');
                        l_FinalIdx++;
                        i++;
                    }
                }
            }
        }

        public static string FixUTF8Latin1ConversionErrors(string a_RawStr)
        {
            // This routine is able to fix a number of conversion
            // errors that happen when a UTF-8 string is interpreted
            // as if ISO-8859-1 encoded.

            const byte l_RangeBegin = 192;
            const byte l_RangeEnd = 255;
            const byte l_Except = 215;

            StringBuilder l_StrBld = null;
            string l_Str = a_RawStr;

            for (uint i = l_RangeBegin; i <= l_RangeEnd; i++)
            {
                if (l_Except == i ||
                    (l_Except + 32) == i)
                {
                    continue;
                }

                l_Str = FixUTF8Latin1ConversionErrorsCore(l_StrBld, l_Str, (char)(i - 64), (char)i);
            }

            return l_Str;
        }

        private static string FixUTF8Latin1ConversionErrorsCore(StringBuilder a_StrBld,
                                                                string a_Str,
                                                                char a_CharBefore,
                                                                char a_CharAfter)
        {
            // a_StrBld may be null or contain garbage on entry.

            const char l_ConvChar0 = (char)194;
            const char l_ConvChar1 = (char)195;
            const char l_RemChar0 = (char)130;
            const char l_RemChar1 = (char)131;

            for (int l_Idx = 0, l_DoneCount = 0; true; l_Idx++)
            {
                // l_Idx can and may be equal to a_Str.Length.
                l_Idx = a_Str.IndexOf(a_CharBefore, l_Idx);

                if (0 > l_Idx)
                {
                    if (0 == l_DoneCount)
                    {
                        return a_Str;
                    }

                    a_StrBld.Append(a_Str, l_DoneCount, a_Str.Length - l_DoneCount);

                    return a_StrBld.ToString();
                }

                if (l_DoneCount == l_Idx)
                {
                    continue;
                }

                int l_StopIdx = (l_Idx - 1);
                char l_FirstChar = a_Str[l_StopIdx];

                if (l_ConvChar0 == l_FirstChar ||
                    l_ConvChar1 == l_FirstChar)
                {
                    for (; l_DoneCount < l_StopIdx; l_StopIdx--)
                    {
                        char l_Char = a_Str[l_StopIdx - 1];

                        bool l_Remove = (l_ConvChar0 == l_Char ||
                                         l_ConvChar1 == l_Char ||
                                         l_RemChar0 == l_Char ||
                                         l_RemChar1 == l_Char);

                        if (false == l_Remove)
                        {
                            break;
                        }
                    }
                }
                else
                {
                    continue;
                }

                if (null == a_StrBld)
                {
                    a_StrBld = new StringBuilder(a_Str.Length);
                }
                else if (0 == l_DoneCount)
                {
                    // Reset the string builder.

                    a_StrBld.Length = 0;
                }

                if (l_DoneCount < l_StopIdx)
                {
                    a_StrBld.Append(a_Str, l_DoneCount, l_StopIdx - l_DoneCount);
                }

                a_StrBld.Append(a_CharAfter);
                l_DoneCount = (l_Idx + 1);
            }
        }

#region WildcardHandling

        public static bool IsWildcardPatternOrderly(string a_Pattern,
                                                    bool a_Strict,
                                                    StringBuilder a_Bld)
        {
            return IsWildcardPatternOrderly(a_Pattern, 0, GetStringLength(a_Pattern), a_Strict, a_Bld);
        }

        public static bool IsWildcardPatternOrderly(string a_Pattern,
                                                    uint a_PatternIdx,
                                                    uint a_PatternLen,
                                                    bool a_Strict,
                                                    StringBuilder a_Bld)
        {
            // Returns a nonzero value if the patern is orderly
            // already (requirements depending on a_Strict).
            // If zero is returned, the orderly pattern was
            // written to a_Bld if it's not null. a_Bld may be
            // null.

            bool l_PatternIsFine = true;
            uint l_PatternFinalIdx = (a_PatternIdx + a_PatternLen);
            uint l_SeqLen;

            for (uint i = a_PatternIdx; i != l_PatternFinalIdx; i += l_SeqLen)
            {
                bool l_AnyLen;
                uint l_MinChars;
                bool l_Wildcard = TryInterpretWildcardSequence(a_Pattern, i, l_PatternFinalIdx, out l_AnyLen, out l_MinChars, out l_SeqLen);

                if (true == l_Wildcard)
                {
                    bool l_MinimalLen = ((l_MinChars + BoolToByte(l_AnyLen)) == l_SeqLen);

                    if (false == l_MinimalLen ||
                        (true == a_Strict &&
                         c_WildcardAnySeqChar != a_Pattern[(int)i]))
                    {
                        if (true == l_PatternIsFine)
                        {
                            l_PatternIsFine = false;

                            if (null == a_Bld)
                            {
                                break;
                            }

                            a_Bld.Append(a_Pattern, (int)a_PatternIdx, (int)(i - a_PatternIdx));
                        }

                        if (true == l_AnyLen)
                        {
                            a_Bld.Append(c_WildcardAnySeqChar);
                        }

                        if (0 != l_MinChars)
                        {
                            a_Bld.Append(c_WildcardUndefinedChar, (int)l_MinChars);
                        }

                        continue;
                    }
                }

                if (false == l_PatternIsFine)
                {
                    a_Bld.Append(a_Pattern, (int)i, (int)l_SeqLen);
                }
            }

            return l_PatternIsFine;
        }

        public static bool DoesAnyStringMatchPattern(string[] a_TestStrs,
                                                     string a_Pattern,
                                                     uint a_PatternIdx,
                                                     uint a_PatternLen,
                                                     bool a_CaseSensitive)
        {
            // a_TestStrs must not be an empty array.

            uint l_StrCount = (uint)a_TestStrs.Length;

            if (1 == l_StrCount)
            {
                return DoesStringMatchPattern(a_TestStrs[0], a_Pattern, a_PatternIdx, a_PatternLen, a_CaseSensitive);
            }

            bool l_ExactLen;
            uint l_MinLen;

            if (false == AnalyzeWildcardPattern(a_Pattern, a_PatternIdx, a_PatternLen, out l_ExactLen, out l_MinLen))
            {
                foreach (string l_TestStr in a_TestStrs)
                {
                    uint l_TestStrLen = (uint)l_TestStr.Length;

                    if (l_TestStrLen == l_MinLen &&
                        true == StringsEqual(l_TestStr, 0, l_TestStrLen, a_Pattern, a_PatternIdx, a_PatternLen, a_CaseSensitive))
                    {
                        return true;
                    }
                }

                return false;
            }

            char l_FirstPatternChar = a_Pattern[(int)a_PatternIdx];

            if (true == IsWildcardChar(l_FirstPatternChar))
            {
                l_FirstPatternChar = '\0';
            }

            char l_LastPatternChar;

            if (1 >= a_PatternLen)
            {
                l_LastPatternChar = '\0';
            }
            else
            {
                l_LastPatternChar = a_Pattern[(int)(a_PatternIdx + a_PatternLen - 1)];

                if (true == IsWildcardChar(l_LastPatternChar))
                {
                    l_LastPatternChar = '\0';
                }
            }

            foreach (string l_TestStr in a_TestStrs)
            {
                uint l_TestStrLen = (uint)l_TestStr.Length;

                if (false == l_ExactLen ? (l_TestStrLen < l_MinLen) : (l_TestStrLen != l_MinLen))
                {
                    continue;
                }

                if (1 == a_PatternLen)
                {
                    return true;
                }

                if ('\0' != l_FirstPatternChar &&
                    false == IsSameChar(l_FirstPatternChar, l_TestStr[0], a_CaseSensitive))
                {
                    continue;
                }

                if ('\0' != l_LastPatternChar &&
                    false == IsSameChar(l_LastPatternChar, l_TestStr[(int)(l_TestStrLen - 1)], a_CaseSensitive))
                {
                    continue;
                }

                if (true == DoesStringMatchPattern(l_TestStr, 0, l_TestStrLen, a_Pattern, a_PatternIdx, a_PatternLen, a_CaseSensitive, false))
                {
                    return true;
                }
            }

            return false;
        }

        public static bool DoesStringMatchPattern(string a_TestStr,
                                                  string a_Pattern,
                                                  uint a_PatternIdx,
                                                  uint a_PatternLen,
                                                  bool a_CaseSensitive)
        {
            return DoesStringMatchPattern(a_TestStr, 0, (uint)a_TestStr.Length, a_Pattern, a_PatternIdx, a_PatternLen, a_CaseSensitive, true);
        }

        public static bool DoesStringMatchPattern(string a_TestStr,
                                                  uint a_TestIdx,
                                                  uint a_TestLen,
                                                  string a_Pattern,
                                                  uint a_PatternIdx,
                                                  uint a_PatternLen,
                                                  bool a_CaseSensitive,
                                                  bool a_DoPreAnalysis)
        {
            // a_Pattern may contain wildcars, such as '?' and '*'.

            uint l_TestFinalIdx = (a_TestIdx + a_TestLen);
            uint l_PatternFinalIdx = (a_PatternIdx + a_PatternLen);

            if (true == a_DoPreAnalysis)
            {
                bool l_ExactLen;
                uint l_MinLen;

                if (false == AnalyzeWildcardPattern(a_Pattern, a_PatternIdx, a_PatternLen, out l_ExactLen, out l_MinLen))
                {
                    // No wildcards involved.

                    return StringsEqual(a_TestStr, a_TestIdx, a_TestLen, a_Pattern, a_PatternIdx, a_PatternLen, a_CaseSensitive);
                }
                else if (1 >= a_PatternLen)
                {
                    if (true == l_ExactLen)
                    {
                        return (l_MinLen == a_TestLen);
                    }

                    return true;
                }
                else
                {
                    if (false == l_ExactLen)
                    {
                        if (l_MinLen > a_TestLen)
                        {
                            return false;
                        }
                    }
                    else
                    {
                        if (l_MinLen != a_TestLen)
                        {
                            return false;
                        }
                    }

                    char l_FirstPatternChar = a_Pattern[(int)a_PatternIdx];

                    if (false == IsWildcardChar(l_FirstPatternChar) &&
                        false == IsSameChar(l_FirstPatternChar, a_TestStr[(int)a_TestIdx], a_CaseSensitive))
                    {
                        return false;
                    }

                    char l_LastPatternChar = a_Pattern[(int)(l_PatternFinalIdx - 1)];

                    if (false == IsWildcardChar(l_LastPatternChar) &&
                        false == IsSameChar(l_LastPatternChar, a_TestStr[(int)(l_TestFinalIdx - 1)], a_CaseSensitive))
                    {
                        return false;
                    }
                }
            }

            Stack<ulong> l_Stack = null;
            uint l_TestIdx = a_TestIdx;
            uint l_PatternIdx = a_PatternIdx;
            bool l_Asterisk = false;

            while (true)
            {
                for (uint i = l_PatternIdx; i != l_PatternFinalIdx; )
                {
                    bool l_AnyLen;
                    uint l_MinChars;
                    uint l_SeqLen;
                    bool l_Wildcard = TryInterpretWildcardSequence(a_Pattern, i, l_PatternFinalIdx, out l_AnyLen, out l_MinChars, out l_SeqLen);

                    if (true == l_Asterisk)
                    {
                        // We land here after a wildcard sequence
                        // involving '*' was found. Now scan a_Test
                        // for the sequence after the asterisk in the
                        // pattern.

                        uint l_RemTestLen = (l_TestFinalIdx - l_TestIdx);
                        uint l_NextTestIdx;

                        if (l_RemTestLen < l_SeqLen ||
                            false == FindString(a_TestStr, l_TestIdx, l_RemTestLen, a_Pattern, i, l_SeqLen, a_CaseSensitive, out l_NextTestIdx))
                        {
                            goto SKIP;
                        }

                        l_TestIdx = (l_NextTestIdx + l_SeqLen);

                        if (l_TestIdx < l_TestFinalIdx)
                        {
                            // There are possibly other valid paths
                            // ahead, so place the corresponding state
                            // on the stack.

                            if (null == l_Stack)
                            {
                                const byte l_DefaultStackLen = 4;
                                l_Stack = new Stack<ulong>(l_DefaultStackLen);
                            }

                            ulong l_NewVal = (((ulong)(l_NextTestIdx + 1) << Util.c_BitsPerUInt) |
                                              i);
                            l_Stack.Push(l_NewVal);
                        }

                        i += l_SeqLen;
                        l_Asterisk = false;
                    }
                    else if (false == l_Wildcard)
                    {
                        // Expect the same non-wildcard sequence in
                        // both a_TestStr and a_Pattern.

                        uint l_NextTestIdx = (l_TestIdx + l_SeqLen);

                        if (l_NextTestIdx > l_TestFinalIdx ||
                            false == StringsEqual(a_TestStr, l_TestIdx, l_SeqLen, a_Pattern, i, l_SeqLen, a_CaseSensitive))
                        {
                            goto SKIP;
                        }

                        l_TestIdx = l_NextTestIdx;
                        i += l_SeqLen;
                    }
                    else
                    {
                        // A wildcard sequence was found.

                        l_TestIdx += l_MinChars;

                        if (l_TestIdx > l_TestFinalIdx)
                        {
                            goto SKIP;
                        }

                        i += l_SeqLen;

                        if (true == l_AnyLen)
                        {
                            // There is a '*' char involved.

                            if (i == l_PatternFinalIdx)
                            {
                                // End of pattern. Anything goes.

                                return true;
                            }

                            l_Asterisk = true;

                            continue;
                        }
                    }

                    if (i == l_PatternFinalIdx)
                    {
                        if (l_TestIdx != l_TestFinalIdx)
                        {
                            goto SKIP;
                        }

                        // Match.

                        return true;
                    }
                }

                SKIP:

                if (null == l_Stack ||
                    0 >= l_Stack.Count)
                {
                    return false;
                }

                // Unwind stack by one position.

                ulong l_OldVal = l_Stack.Pop();
                l_TestIdx = (uint)(l_OldVal >> Util.c_BitsPerUInt);
                l_PatternIdx = (uint)l_OldVal;
                l_Asterisk = true;
            }
        }

        private static bool TryInterpretWildcardSequence(string a_Pattern,
                                                         uint a_PatternIdx,
                                                         uint a_PatternFinalIdx,
                                                         out bool a_AnyLen,
                                                         out uint a_MinChars,
                                                         out uint a_SeqLen)
        {
            // Helper routine of DoesStringMatchPattern().

            a_AnyLen = false;
            a_MinChars = 0;
            bool l_Wildcard = false;
            uint i;

            for (i = a_PatternIdx; i != a_PatternFinalIdx; i++)
            {
                char l_Char = a_Pattern[(int)i];

                if (false == IsWildcardChar(l_Char))
                {
                    if (true == l_Wildcard)
                    {
                        // This is the end of a wildcard sequence.

                        break;
                    }
                }
                else
                {
                    if (false == l_Wildcard)
                    {
                        if (i != a_PatternIdx)
                        {
                            // We already encountered a non-wildcard
                            // char.

                            break;
                        }

                        l_Wildcard = true;
                    }

                    if (c_WildcardUndefinedChar == l_Char)
                    {
                        a_MinChars++;
                    }
                    else
                    {
                        a_AnyLen = true;
                    }
                }
            }

            a_SeqLen = (i - a_PatternIdx);

            return l_Wildcard;
        }

        public static bool AnalyzeWildcardPattern(string a_Pattern,
                                                  uint a_Idx,
                                                  uint a_Len,
                                                  out bool a_ExactLen,
                                                  out uint a_MinLen)
        {
            a_ExactLen = true;
            a_MinLen = 0;
            bool l_Wildcard = false;

            for (uint i = (a_Idx + a_Len); i-- != a_Idx; )
            {
                char l_Char = a_Pattern[(int)i];

                if (c_WildcardAnySeqChar == l_Char)
                {
                    a_ExactLen = false;
                    l_Wildcard = true;

                    continue;
                }

                if (c_WildcardUndefinedChar == l_Char)
                {
                    l_Wildcard = true;
                }

                a_MinLen++;
            }

            return l_Wildcard;
        }

        public static bool IsWildcardChar(char a_Char)
        {
            return (c_WildcardUndefinedChar == a_Char ||
                    c_WildcardAnySeqChar == a_Char);
        }

#endregion // WildcardHandling

        public static bool IsWindows()
        {
            // Xbox is not known to .NET 2.0.
            const PlatformID lc_XboxPlatform = (PlatformID)5;

            PlatformID l_Platf = Environment.OSVersion.Platform;

            return (PlatformID.Win32NT == l_Platf ||
                    PlatformID.Win32Windows == l_Platf ||
                    PlatformID.Win32S == l_Platf ||
                    lc_XboxPlatform == l_Platf);
        }

        public static bool CheckForByteSequence(ref ByteArray a_Data,
                                                ref ByteArray a_Sequence)
        {
            // Scans for a_Sequence in a_Data and returns a nonzero
            // value if the sequence was found.

            if (0 >= a_Sequence.f_Len ||
                a_Data.f_Len < a_Sequence.f_Len)
            {
                return false;
            }

            byte l_FirstSequChar = a_Sequence.f_Arr[a_Sequence.f_Idx];

            if (1 == a_Sequence.f_Len)
            {
                for (int i = a_Data.f_Idx, lim = (a_Data.f_Idx + a_Data.f_Len); i != lim; i++)
                {
                    if (l_FirstSequChar == a_Data.f_Arr[i])
                    {
                        return true;
                    }
                }

                return false;
            }

            int l_SecondSequIndex = (a_Sequence.f_Idx + 1);
            int l_FinalSequIndex = (a_Sequence.f_Idx + a_Sequence.f_Len);
            byte[] l_Arr = a_Data.f_Arr;

            for (int i = a_Data.f_Idx, lim = (a_Data.f_Idx + a_Data.f_Len - a_Sequence.f_Len + 1); i != lim; i++)
            {
                if (l_FirstSequChar != l_Arr[i])
                {
                    continue;
                }

                bool l_Match = true;

                for (int k = l_SecondSequIndex, g = i; k != l_FinalSequIndex; k++)
                {
                    if (a_Sequence.f_Arr[k] != l_Arr[++g])
                    {
                        l_Match = false;

                        break;
                    }
                }

                if (true == l_Match)
                {
                    return true;
                }
            }

            return false;
        }

        [Conditional("DEBUG")]
        public static void DebugBreak()
        {
        }

        public static byte BoolToByte(bool a_Bool)
        {
            return (byte)(false == a_Bool ? 0 : 1);
        }

        public static string TruncateStr(string a_Str,
                                         bool a_ReturnNullIfEmpty,
                                         uint a_TruncationCharLimit)
        {
            // Reads a UTF-16 string and returns no more than
            // a_TruncationCharLimit chars of it.

            uint l_Len = (uint)a_Str.Length;

            if (true == a_ReturnNullIfEmpty &&
                0 == l_Len)
            {
                return null;
            }

            if (l_Len > a_TruncationCharLimit)
            {
                return a_Str.Substring(0, (int)a_TruncationCharLimit);
            }

            return a_Str;
        }

        public static uint CalcStringWorstCaseByteSize(uint a_StringLength)
        {
            return (a_StringLength * 4);
        }

        public static uint GetArrayLength<T>(T[] a_Arr)
        {
            return (uint)a_Arr.Length;
        }

        public static uint GetStringLength(string a_Str)
        {
            return (uint)a_Str.Length;
        }

        public static uint GetStringLengthNullSafe(string a_Str)
        {
            return (null == a_Str ? 0 : (uint)a_Str.Length);
        }

        public static int MinMaxSigned(int a_Minimum,
                                       int a_Maximum,
                                       int a_Value)
        {
            return Math.Max(a_Minimum, Math.Min(a_Maximum, a_Value));
        }

        public static uint MinMax(uint a_Minimum,
                                  uint a_Maximum,
                                  uint a_Value)
        {
            return Math.Max(a_Minimum, Math.Min(a_Maximum, a_Value));
        }

        public static void DisposeArray<T>(T[] a_Arr)
                                  where T : class, IDisposable
        {
            // Each element of the array may be null.

            foreach (T l_Obj in a_Arr)
            {
                if (null != l_Obj)
                {
                    l_Obj.Dispose();
                }
            }
        }

        public static void DisposeStructArray<T>(T[] a_Arr)
                                        where T : struct, IDisposable
        {
            for (int i = a_Arr.Length; 0 != i--;)
            {
                a_Arr[i].Dispose();
            }
        }

        public static void DisposeAndZero<T>(ref T a_Obj)
                                    where T : class, IDisposable
        {
            // a_Obj may be null.

            if (null != a_Obj)
            {
                a_Obj.Dispose();
                a_Obj = null;
            }
        }

        public static void Beep()
        {
            Beep(650, 250);
        }

        public static void Beep(uint a_Frequency,
                                uint a_Duration)
        {
            try
            {
                Console.Beep((int)a_Frequency, (int)a_Duration);
            }
            catch (System.Security.HostProtectionException)
            {
                System.Media.SystemSounds.Beep.Play();
            }
        }

        public static void Swap<T>(ref T a_A,
                                   ref T a_B)
        {
            T l_Temp = a_A;
            a_A = a_B;
            a_B = l_Temp;
        }

        public static bool AddSorted<T>(IList<T> a_Ls,
                                        T a_HashIndex)
                               where T : IComparable
        {
            int l_TrgtIdx = 0;

            for (int l_Count = a_Ls.Count; l_TrgtIdx < l_Count; l_TrgtIdx++)
            {
                int l_Res = a_HashIndex.CompareTo(a_Ls[l_TrgtIdx]);

                if (0 > l_Res)
                {
                    break;
                }
                else if (0 == l_Res)
                {
                    return false;
                }
            }

            a_Ls.Insert(l_TrgtIdx, a_HashIndex);

            return true;
        }

        public static string BytesToString(byte[] a_Bytes)
        {
            return HexEncoding.ToHex(a_Bytes, 0, a_Bytes.Length);
        }

        public static uint GetClearLowBits(uint a_Num)
        {
            uint l_Bits = 0;
            uint l_Temp = a_Num;

            if (0 == (l_Temp & 0x0000FFFF))
            {
                // The lowest 16 bits are clear.

                l_Bits += c_BitsPerUShort;
                l_Temp >>= c_BitsPerUShort;
            }

            if (0 == (l_Temp & 0x00FF))
            {
                l_Bits += c_BitsPerByte;
                l_Temp >>= c_BitsPerByte;
            }

            if (0 == (l_Temp & 0x0F))
            {
                l_Bits += c_BitsPerNibble;
                l_Temp >>= c_BitsPerNibble;
            }

            for (uint l_Mask = 1; 0xF >= l_Mask; l_Mask <<= 1)
            {
                if (0 != (l_Temp & l_Mask))
                {
                    break;
                }

                l_Bits++;
            }

            return l_Bits;
        }

        public static uint GetClearHighBits(uint a_Num)
        {
            uint l_Bits = 0;
            uint l_Temp = a_Num;

            if (0 == (l_Temp & 0xFFFF0000))
            {
                // The highest 16 bits are clear.

                l_Bits += c_BitsPerUShort;
            }
            else
            {
                l_Temp >>= c_BitsPerUShort;
            }

            if (0 == (l_Temp & 0xFF00))
            {
                l_Bits += c_BitsPerByte;
            }
            else
            {
                l_Temp >>= c_BitsPerByte;
            }

            if (0 == (l_Temp & 0xF0))
            {
                l_Bits += c_BitsPerNibble;
            }
            else
            {
                l_Temp >>= c_BitsPerNibble;
            }

            for (uint l_Mask = 0x8; 0 != l_Mask; l_Mask >>= 1)
            {
                if (0 != (l_Temp & l_Mask))
                {
                    break;
                }

                l_Bits++;
            }

            return l_Bits;
        }

        public static uint GetClearLowBits64(ulong a_Num)
        {
            uint l_Bits = GetClearLowBits((uint)a_Num);

            if (32 == l_Bits)
            {
                l_Bits += GetClearLowBits((uint)(a_Num >> c_BitsPerUInt));
            }

            return l_Bits;
        }

        public static uint GetClearHighBits64(ulong a_Num)
        {
            uint l_Bits = GetClearHighBits((uint)(a_Num >> c_BitsPerUInt));

            if (32 == l_Bits)
            {
                l_Bits += GetClearHighBits((uint)a_Num);
            }

            return l_Bits;
        }

        public static byte[] CopyBytes(byte[] a_In,
                                       int a_Idx)
        {
            return CopyBytes(a_In, a_Idx, a_In.Length - a_Idx);
        }

        public static byte[] CopyBytes(byte[] a_In,
                                       int a_Idx,
                                       int a_Len)
        {
            byte[] l_Out = new byte[a_Len];

            Buffer.BlockCopy(a_In, a_Idx, l_Out, 0, a_Len);

            return l_Out;
        }

#region 7BitEncodedUInt32

        public static Read7BitEncUInt32Status TryRead7BitEncUInt32(byte[] a_Src,
                                                                   int a_Idx,
                                                                   int a_MaxLength,
                                                                   out uint a_Size,
                                                                   out uint a_Dest)
        {
            a_Size = 0;
            a_Dest = 0;

            for (int l_Shift = 0; a_Size < a_MaxLength; a_Size++)
            {
                uint l_Byte = a_Src[a_Idx + a_Size];

                if (4 <= a_Size)
                {
                    // Check for too large input.

                    if (4 < a_Size)
                    {
                        // The fifth byte must be the last.

                        return Read7BitEncUInt32Status.FormatError;
                    }

                    if (0 != (0xF0 & l_Byte))
                    {
                        // The high nibble of the fourth byte must
                        // be clear.

                        return Read7BitEncUInt32Status.FormatError;
                    }
                }

                a_Dest |= ((l_Byte & 0x7F) << l_Shift);
                l_Shift += 7;

                if (0 == (l_Byte & 0x80))
                {
                    a_Size++;

                    return Read7BitEncUInt32Status.Success;
                }
            }

            return Read7BitEncUInt32Status.TooFewSrcBytes;
        }

        public static uint Get7BitEncUInt32Size(uint a_UInt)
        {
            // Returns the same value that Write7BitEncUInt32() will
            // return when writing.

            if (0x80 > a_UInt) return 1;
            if (0x4000 > a_UInt) return 2;
            if (0x200000 > a_UInt) return 3;
            if (0x10000000 > a_UInt) return 4;

            return c_Max7BitEncUInt32ByteLen;
        }

        public static uint Write7BitEncUInt32(uint a_Src,
                                              byte[] a_Dest,
                                              int a_Idx)
        {
            // Returns the number of bytes written. Five at most.

            int l_Size = 0;
            uint l_Rest = a_Src;

            do
            {
                uint l_Byte = (l_Rest & 0x7F);
                l_Rest >>= 7;

                if (0 != l_Rest)
                {
                    l_Byte |= 0x80;
                }

                a_Dest[a_Idx + l_Size] = (byte)l_Byte;
                l_Size++;
            }
            while (0 != l_Rest);

            return (uint)l_Size;
        }

#endregion

        public static bool TryBuildDateTime(uint a_Year,
                                            uint a_Month,
                                            uint a_Day,
                                            DateTimeKind a_Kind,
                                            out DateTime a_DateTime)
        {
            try
            {
                a_DateTime = new DateTime((int)a_Year, (int)a_Month, (int)a_Day, 0, 0, 0, 0, Util.s_InvariantCulture.Calendar, a_Kind);
            }
            catch (ArgumentException)
            {
                a_DateTime = new DateTime();

                return false;
            }

            return true;
        }

        public static ushort ByteSwapUInt16(ushort a_Num)
        {
            // Don't use IPAddress.HostToNetworkOrder() for this
            // (since it works only on machines with a certain
            // byte order and does nothing on others).

            return (ushort)((a_Num << c_BitsPerByte) |
                            (a_Num >> c_BitsPerByte));
        }

        public static uint ByteSwapUInt32(uint a_Num)
        {
            // Don't use IPAddress.HostToNetworkOrder() for this.

            return ((((uint)ByteSwapUInt16((ushort)a_Num)) << c_BitsPerUShort) |
                    (uint)ByteSwapUInt16((ushort)(a_Num >> c_BitsPerUShort)));
        }

        public static ulong ByteSwapUInt64(ulong a_Num)
        {
            // Don't use IPAddress.HostToNetworkOrder() for this.

            return ((((ulong)ByteSwapUInt32((uint)a_Num)) << c_BitsPerUInt) |
                    ByteSwapUInt32((uint)(a_Num >> c_BitsPerUInt)));
        }

        public static void WriteUInt16LE(ushort a_Src,
                                         byte[] a_Dest,
                                         int a_Idx)
        {
            a_Dest[a_Idx] = (byte)(a_Src & byte.MaxValue);
            a_Dest[a_Idx + 1] = (byte)((a_Src >> c_BitsPerByte) & byte.MaxValue);
        }

        public static void WriteUInt32LE(uint a_Src,
                                         byte[] a_Dest,
                                         int a_Idx)
        {
            a_Dest[a_Idx] = (byte)(a_Src & byte.MaxValue);
            a_Dest[a_Idx + 1] = (byte)((a_Src >> c_BitsPerByte) & byte.MaxValue);
            a_Dest[a_Idx + 2] = (byte)((a_Src >> c_BitsPerUShort) & byte.MaxValue);
            a_Dest[a_Idx + 3] = (byte)((a_Src >> 24) & byte.MaxValue);
        }

        public static void WriteUInt32BE(uint a_Src,
                                         byte[] a_Dest,
                                         int a_Idx)
        {
            a_Dest[a_Idx] = (byte)((a_Src >> 24) & byte.MaxValue);
            a_Dest[a_Idx + 1] = (byte)((a_Src >> c_BitsPerUShort) & byte.MaxValue);
            a_Dest[a_Idx + 2] = (byte)((a_Src >> c_BitsPerByte) & byte.MaxValue);
            a_Dest[a_Idx + 3] = (byte)(a_Src & byte.MaxValue);
        }

        public static void WriteUInt64LE(ulong a_Src,
                                         byte[] a_Dest,
                                         int a_Idx)
        {
            a_Dest[a_Idx] = (byte)(a_Src & byte.MaxValue);
            a_Dest[a_Idx + 1] = (byte)((a_Src >> c_BitsPerByte) & byte.MaxValue);
            a_Dest[a_Idx + 2] = (byte)((a_Src >> c_BitsPerUShort) & byte.MaxValue);
            a_Dest[a_Idx + 3] = (byte)((a_Src >> 24) & byte.MaxValue);
            a_Dest[a_Idx + 4] = (byte)((a_Src >> c_BitsPerUInt) & byte.MaxValue);
            a_Dest[a_Idx + 5] = (byte)((a_Src >> 40) & byte.MaxValue);
            a_Dest[a_Idx + 6] = (byte)((a_Src >> 48) & byte.MaxValue);
            a_Dest[a_Idx + 7] = (byte)((a_Src >> 56) & byte.MaxValue);
        }

        public static void WriteUInt64BE(ulong a_Src,
                                         byte[] a_Dest,
                                         int a_Idx)
        {
            WriteUInt32BE((uint)(a_Src >> c_BitsPerUInt) & uint.MaxValue, a_Dest, a_Idx);
            WriteUInt32BE((uint)(a_Src & uint.MaxValue), a_Dest, a_Idx + sizeof(uint));
        }

        public static ushort ReadUInt16LE(byte[] a_Src,
                                          int a_Idx)
        {
            // Don't use BitConverter.ToUInt16() for this.

            return (ushort)(a_Src[a_Idx] |
                            (a_Src[a_Idx + sizeof(byte)] << c_BitsPerByte));
        }

        public static uint ReadUInt32LE(byte[] a_Src,
                                        int a_Idx)
        {
            // Don't use BitConverter.ToUInt32() for this.

            return (uint)(a_Src[a_Idx] |
                          (a_Src[a_Idx + 1] << c_BitsPerByte) |
                          (a_Src[a_Idx + 2] << c_BitsPerUShort) |
                          (a_Src[a_Idx + 3] << 24));
        }

        public static ulong ReadUInt64LE(byte[] a_Src,
                                         int a_Idx)
        {
            // Don't use BitConverter.ToUInt64() for this.

            return (ReadUInt32LE(a_Src, a_Idx) |
                    ((ulong)ReadUInt32LE(a_Src, a_Idx + sizeof(uint)) << c_BitsPerUInt));
        }

        public static void WriteFloat(float a_Src,
                                      byte[] a_Dest,
                                      int a_Idx)
        {
            FloatUnion l_Union = new FloatUnion();
            l_Union.f_Float = a_Src;

            WriteUInt32LE(l_Union.f_UInt32, a_Dest, a_Idx);
        }

        public static float ReadFloat(byte[] a_Src,
                                      int a_Idx)
        {
            FloatUnion l_Union = new FloatUnion();
            l_Union.f_UInt32 = ReadUInt32LE(a_Src, a_Idx);

            return l_Union.f_Float;
        }

        public static int RoundDoubleToInt(double a_Double)
        {
            // Can be used with floats too of course.

            return (int)Math.Round(a_Double, MidpointRounding.ToEven);
        }

        public static bool TryParseUInt16Inv(string a_Text,
                                             int a_Idx,
                                             int a_Len,
                                             out ushort a_Num)
        {
            ulong l_Num;
            if (true == TryParseUInt64Inv(a_Text, a_Idx, a_Len, out l_Num) &&
                ushort.MaxValue >= l_Num)
            {
                a_Num = (ushort)l_Num;

                return true;
            }

            a_Num = 0;

            return false;
        }

        public static bool TryParseUInt32Inv(string a_Text,
                                             int a_Idx,
                                             int a_Len,
                                             out uint a_Num)
        {
            ulong l_Num;
            if (true == TryParseUInt64Inv(a_Text, a_Idx, a_Len, out l_Num) &&
                uint.MaxValue >= l_Num)
            {
                a_Num = (uint)l_Num;

                return true;
            }

            a_Num = 0;

            return false;
        }

        public static bool TryParseUInt64Inv(string a_Text,
                                             out ulong a_Num)
        {
            return TryParseUInt64Inv(a_Text, 0, a_Text.Length, out a_Num);
        }

        public static bool TryParseUInt64Inv(string a_Text,
                                             int a_Idx,
                                             int a_Len,
                                             out ulong a_Num)
        {
            // This routine parses decimal integers without the need
            // of creating a new string for partial processing.

            a_Num = 0;

            bool l_CoreReached = false;
            bool l_CoreLeft = false;
            bool l_HadPosSign = false;
            bool l_HadDigit = false;
            for (int i = a_Idx, lim = (a_Idx + a_Len); i < lim; i++)
            {
                char l_Char = a_Text[i];

                if (false == char.IsWhiteSpace(l_Char))
                {
                    if (true == l_CoreLeft)
                    {
                        return false;
                    }

                    if (false == l_CoreReached)
                    {
                        l_CoreReached = true;

                        if ('+' == l_Char)
                        {
                            if (true == l_HadPosSign)
                            {
                                return false;
                            }

                            l_HadPosSign = true;

                            continue;
                        }
                    }

                    if (false == IsDigitInvariant(l_Char))
                    {
                        return false;
                    }

                    try
                    {
                        if (false == l_HadDigit)
                        {
                            l_HadDigit = true;
                        }
                        else
                        {
                            a_Num = checked(a_Num * 10);
                        }

                        uint l_Val = (uint)(l_Char - '0');
                        a_Num = checked(a_Num + l_Val);
                    }
                    catch (System.OverflowException)
                    {
                        return false;
                    }
                }
                else
                {
                    if (false == l_CoreReached ||
                        true == l_CoreLeft)
                    {
                        continue;
                    }

                    if (true == l_HadDigit)
                    {
                        l_CoreLeft = true;

                        continue;
                    }

                    return false;
                }
            }

            return l_HadDigit;
        }

        public static bool IsDigitInvariant(char a_Char)
        {
            return ('0' <= a_Char &&
                    '9' >= a_Char);
        }

        public static string NumberToInvString(int a_Num)
        {
            return a_Num.ToString(s_InvariantCulture);
        }

        public static string ULongToInvString(ulong a_Num)
        {
            return a_Num.ToString(s_InvariantCulture);
        }

        public static void ReseedRandomProvider()
        {
            RandomProvider.ReseedSecure();
        }

        public static byte[] GetRandomBytes(int a_Count)
        {
            ByteArray l_Data = new ByteArray(a_Count);
            RandomProvider.FillBufferSecure(ref l_Data);

            return l_Data.f_Arr;
        }

        public static bool GetRandomBool()
        {
            return (sbyte.MaxValue >= GetRandomByte());
        }

        public static byte GetRandomByte()
        {
            return (byte)GetRandomUInt();
        }

        public static uint GetRandomUInt()
        {
            ulong l_Junk0, l_Junk1;

            return RandomProvider.GetSecureUInts(out l_Junk0, out l_Junk1);
        }

        public static ulong GetRandomULong()
        {
            ulong l_Res, l_Junk0;
            uint l_Junk1 = RandomProvider.GetSecureUInts(out l_Res, out l_Junk0);

            return l_Res;
        }

        public static UInt128 GetRandomUInt128()
        {
            ulong l_Rand0;
            ulong l_Rand1;
            uint l_Rand2 = RandomProvider.GetSecureUInts(out l_Rand0, out l_Rand1);

            return new UInt128((uint)((l_Rand0 >> Util.c_BitsPerUInt) & uint.MaxValue), (uint)l_Rand0, (uint)l_Rand1, l_Rand2);
        }

        public static uint GetRandomUIntFast()
        {
            return RandomProvider.GetUInt32Quick();
        }

        public static int GetRandomIndexFast(int a_Count)
        {
            return ((int)(GetRandomUIntFast() & int.MaxValue) % a_Count);
        }

        public static float GetRandomFactorFast()
        {
            // Returns a float between zero and one.

            return ((float)GetRandomUIntFast() / uint.MaxValue);
        }

        public static TValue[] DictionaryToArray<TKey, TValue>(Dictionary<TKey, TValue> a_Dic)
        {
            // If a_Dic is null or empty, the output will be null.

            if (null == a_Dic ||
                0 == a_Dic.Count)
            {
                return null;
            }

            TValue[] l_Arr = new TValue[a_Dic.Count];
            int l_Idx = 0;

            foreach (TValue l_Item in a_Dic.Values)
            {
                l_Arr[l_Idx++] = l_Item;
            }

            return l_Arr;
        }

        public static void ReverseArray<T>(T[] a_Arr)
        {
            ReverseArray<T>(a_Arr, 0, a_Arr.Length);
        }

        public static void ReverseArray<T>(T[] a_Arr,
                                           int a_Index,
                                           int a_Count)
        {
            Array.Reverse(a_Arr, a_Index, a_Count);
        }

        public static void RandomizeArrayOrder<T>(T[] a_Arr,
                                                  int a_Index,
                                                  int a_Count)
        {
            for (int i = 0; i < a_Count; i++)
            {
                int l_RandIdx1 = (a_Index + (int)(Util.GetRandomUIntFast() % (uint)a_Count));
                int l_RandIdx2 = (a_Index + (int)(Util.GetRandomUIntFast() % (uint)a_Count));

                T l_Temp = a_Arr[l_RandIdx1];
                a_Arr[l_RandIdx1] = a_Arr[l_RandIdx2];
                a_Arr[l_RandIdx2] = l_Temp;
            }
        }

        public static void SortArrayIfNecessary<T>(T[] a_Arr,
                                                   Comparison<T> a_Comparison)
        {
            int l_Count = a_Arr.Length;

            if (1 >= l_Count)
            {
                return;
            }

            for (int i = 0, lim = (l_Count - 1); i < lim; i++)
            {
                int l_Res = a_Comparison(a_Arr[i], a_Arr[i + 1]);

                if (0 < l_Res)
                {
                    // The array is not sorted.

                    SortArray(a_Arr, a_Comparison);

                    return;
                }
            }
        }

        public static void SortArray<T>(T[] a_Arr,
                                        Comparison<T> a_Comparison)
        {
            List<T> l_Ls = new List<T>(a_Arr);
            l_Ls.Sort(a_Comparison);
            l_Ls.CopyTo(a_Arr);
        }

        public static T[] CloneArray<T>(T[] a_Arr)
        {
            return (T[])a_Arr.Clone();
        }

        public static bool GetAny<T>(ICollection<T> a_Coll,
                                     out T a_Res)
        {
            if (0 < a_Coll.Count)
            {
                foreach (T l_T in a_Coll)
                {
                    a_Res = l_T;

                    return true;
                }
            }

            a_Res = default(T);

            return false;
        }

        public static ulong Mul64(ulong a_Val,
                                  ulong a_Multiplier)
        {
            return (a_Val * a_Multiplier);
        }

        public static bool FindFrameworkVersion(Assembly a_From,
                                                out Version a_Version)
        {
            AssemblyName[] l_Names = a_From.GetReferencedAssemblies();

            foreach (AssemblyName l_Name in l_Names)
            {
                if ("mscorlib" == l_Name.Name)
                {
                    a_Version = l_Name.Version;

                    return true;
                }
            }

            a_Version = null;

            return false;
        }

        public static string ToShortString(string a_Str,
                                           int a_MaxPartLen)
        {
            if ((2 * a_MaxPartLen) >= a_Str.Length)
            {
                return a_Str;
            }

            return (a_Str.Substring(0, a_MaxPartLen) + "..." +
                    a_Str.Substring(a_Str.Length - a_MaxPartLen, a_MaxPartLen));
        }

        private static void InitializeThread(Thread a_Thread,
                                             string a_Name,
                                             ThreadPriority a_Priority)
        {
            a_Thread.Name = a_Name;
            a_Thread.Priority = a_Priority;
            a_Thread.SetApartmentState(ApartmentState.MTA);
            a_Thread.CurrentCulture = s_InvariantCulture;
            a_Thread.CurrentUICulture = s_InvariantCulture;
        }

        public static Thread StartThread(string a_Name,
                                         ThreadStart a_Entry)
        {
            return StartThread(a_Name, a_Entry, ThreadPriority.Normal);
        }

        public static Thread StartThread(string a_Name,
                                         ThreadStart a_Entry,
                                         ThreadPriority a_Priority)
        {
            Thread l_Thread = new Thread(a_Entry);
            InitializeThread(l_Thread, a_Name, a_Priority);
            l_Thread.Start();

            return l_Thread;
        }

        public static Thread StartThreadWithParam(string a_Name,
                                                  ParameterizedThreadStart a_Entry,
                                                  ThreadPriority a_Priority,
                                                  object a_Param)
        {
            Thread l_Thread = new Thread(a_Entry);
            InitializeThread(l_Thread, a_Name, a_Priority);
            l_Thread.Start(a_Param);

            return l_Thread;
        }

        public static void MiniSleep()
        {
            System.Threading.Thread.Sleep(c_MiniSleepTimeout);
        }

        public static bool WaitForChangeWithTimeout<T>(ref T a_Target,
                                                       T a_WaitValue,
                                                       uint a_MaxMillisecs)
                                              where T : IEquatable<T>
        {
            Time l_Start = Time.Now();

            while (true == Util.VolatileRead(ref a_Target).Equals(a_WaitValue))
            {
                if (Time.Now().MillisecondsSince(l_Start) >= a_MaxMillisecs)
                {
                    return false;
                }

                Util.MiniSleep();
            }

            return true;
        }

        public static uint RotateLeft(uint x,
                                      int n)
        {
            return ((x << n) | (x >> (32 - n)));
        }

        public static uint RotateRight(uint x,
                                       int n)
        {
            return ((x >> n) | (x << (32 - n)));
        }

        public static void MemoryBarrier()
        {
            Thread.MemoryBarrier();
        }

        public static T VolatileRead<T>(ref T a_Ref)
        {
            // Do not use this routine to read value types that are
            // larger than four bytes, since we can't reliably do this
            // thread safe.

            T l_Temp = a_Ref;

            Thread.MemoryBarrier();

            return l_Temp;
        }

        public static void VolatileWrite<T>(out T a_Dest,
                                            T a_Src)
        {
            Thread.MemoryBarrier();

            a_Dest = a_Src;
        }

        public static bool CheckRangeUnsigned(uint a_Idx,
                                              uint a_Len,
                                              uint a_MaxLen)
        {
            uint l_Final = (a_Idx + a_Len);

            return (l_Final <= a_MaxLen &&
                    // Check for overflow.
                    l_Final >= a_Idx);
        }

        public static bool CheckRange(int a_Idx,
                                      int a_Len,
                                      int a_MaxLen)
        {
            // a_MaxLen should never be negative.

            return (0 <= a_MaxLen &&
                    true == CheckRangeUnsigned((uint)a_Idx, (uint)a_Len, (uint)a_MaxLen));
        }

        public static bool CheckRangeMixed(uint a_Idx,
                                           int a_Len,
                                           uint a_MaxLen)
        {
            // a_Len is a signed integer.

            return (0 <= a_Len &&
                    true == CheckRangeUnsigned(a_Idx, (uint)a_Len, a_MaxLen));
        }

        public static void VerifyRangeOrFail(int a_Idx,
                                             int a_Len,
                                             int a_MaxLen)
        {
            // a_MaxLen is not treated as unsafe.

            if (false == CheckRange(a_Idx, a_Len, a_MaxLen))
            {
                throw (new ArgumentOutOfRangeException());
            }
        }

#if FAST
        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern uint SocketCheckConnectionStatus_x86([In] IntPtr a_Socket);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.StdCall)]
        private static extern uint SocketCheckConnectionStatus_x64([In] IntPtr a_Socket);
#endif

        public static bool CheckConnectionStatus(IntPtr a_Socket,
                                                 out uint l_ConnStatus)
        {
            // Returns a nonzero value if the native library processed
            // the request.

#if FAST
            if (false == s_SocketCheckTried ||
                true == s_SocketCheckPerformed)
            {
                s_SocketCheckTried = true;

                try
                {
                    if (sizeof(uint) == IntPtr.Size)
                    {
                        l_ConnStatus = SocketCheckConnectionStatus_x86(a_Socket);
                    }
                    else
                    {
                        l_ConnStatus = SocketCheckConnectionStatus_x64(a_Socket);
                    }

                    s_SocketCheckPerformed = true;

                    return true;
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }

                s_SocketCheckPerformed = false;
            }
#endif

            l_ConnStatus = 0;

            return false;
        }

#if FAST
        [DllImport("kernel32.dll", CallingConvention = CallingConvention.StdCall)]
        private static extern uint SetThreadExecutionState([In] uint esFlags);
#endif

        public static bool AdjustSystemSleepPrevention(bool a_Prevent)
        {
            // This routine must be called periodically in order to
            // prevent the system from going to sleep.

#if FAST
            const uint ES_CONTINUOUS = 0x80000000;
            const uint ES_SYSTEM_REQUIRED = 0x00000001;

            uint l_Flags = (ES_CONTINUOUS | (false == a_Prevent ? 0 : ES_SYSTEM_REQUIRED));
            uint l_Res;

            try
            {
                l_Res = SetThreadExecutionState(l_Flags);
            }
            catch
            {
                return false;
            }

            return (0 != l_Res);
#else
            return false;
#endif
        }

        public static int ReadFromStream(Stream a_Strm,
                                         ref ByteArray a_Dest)
        {
            // Returns the number of bytes read. Exceptions can occur.

            int l_Done = a_Strm.Read(a_Dest.f_Arr, a_Dest.f_Idx, a_Dest.f_Len);

            if (a_Dest.f_Len != l_Done)
            {
                return l_Done;
            }

            // The Read() routine may give us only a portion of the
            // range we requested, so we need to loop and read again.

            int l_Total = l_Done;

            while (l_Total < a_Dest.f_Len)
            {
                if (0 == l_Done)
                {
                    // If Read() returns zero, we should abort.

                    break;
                }

                l_Done = a_Strm.Read(a_Dest.f_Arr, a_Dest.f_Idx + l_Total, a_Dest.f_Len - l_Total);
                l_Total += l_Done;
            }

            return l_Total;
        }

        public static StreamReader OpenTextFile(string a_FileName,
                                                out ulong a_Size)
        {
            // Return can be null on error. NOTE: We don't support
            // very long paths here.

            FileStream l_Strm = null;

            try
            {
                l_Strm = File.OpenRead(a_FileName);
                a_Size = (ulong)l_Strm.Length;

                return new StreamReader(l_Strm, true);
            }
            catch (ArgumentNullException l_Exc)
            {
                throw (l_Exc);
            }
            catch
            {
            }

            if (null != l_Strm)
            {
                l_Strm.Close();
            }

            a_Size = 0;

            return null;
        }

        public static MemoryStream InflateStream(Stream a_DeflatedStrm)
        {
            // The returned stream must be disposed by the caller.

            MemoryStream l_Mem = new MemoryStream();

            System.IO.Compression.DeflateStream l_InflateStrm = new System.IO.Compression.DeflateStream(a_DeflatedStrm, System.IO.Compression.CompressionMode.Decompress, true);
            {
                byte[] l_Buffer = new byte[64 * 1024];

                while (true)
                {
                    int l_Done = l_InflateStrm.Read(l_Buffer, 0, l_Buffer.Length);

                    if (0 == l_Done)
                    {
                        break;
                    }

                    l_Mem.Write(l_Buffer, 0, l_Done);
                }

                l_InflateStrm.Flush();
            }
            l_InflateStrm.Dispose();

            l_Mem.Position = 0;

            return l_Mem;
        }

#if REFERENCIAL_SOURCE_CODE

        public static byte[] DeflateBytes(ref ByteArray a_Src)
        {
            // This routine can be used to compress data that is
            // embedded as resource into the release executables.
            // NOTE: This code is never used in release builds and
            // included only for referencial purposes.

            using (MemoryStream l_MemStrm = new MemoryStream())
            {
                using (Stream l_ComprStrm = new System.IO.Compression.DeflateStream(l_MemStrm, System.IO.Compression.CompressionMode.Compress, true))
                {
                    l_ComprStrm.Write(a_Src.f_Arr, a_Src.f_Idx, a_Src.f_Len);
                    l_ComprStrm.Flush();

                    return l_MemStrm.ToArray();
                }
            }
        }

#endif

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        internal struct UTF16Reader
        {
            // Don't use the struct default constructor.

            public readonly string f_Str;
            public readonly uint f_Idx;
            public readonly uint f_FinalIdx;

            private uint f_Pos;

            public UTF16Reader(string a_Str,
                               uint a_Idx,
                               uint a_Len)
            {
                f_Str = a_Str;
                f_Idx = a_Idx;
                f_FinalIdx = (a_Idx + a_Len);

                // Do what Reset() does.
                f_Pos = a_Idx;
            }

            public void Reset()
            {
                f_Pos = f_Idx;
            }

            public bool TryReadNextCodePoint(out uint a_CodePoint)
            {
                if (f_Pos >= f_FinalIdx)
                {
                    a_CodePoint = 0;

                    return false;
                }

                char l_Char = f_Str[(int)f_Pos];
                f_Pos++;

                if (true == char.IsHighSurrogate(l_Char))
                {
                    if (f_Pos < f_FinalIdx)
                    {
                        char l_NextChar = f_Str[(int)f_Pos];

                        if (true == char.IsLowSurrogate(l_NextChar))
                        {
                            f_Pos++;
                            a_CodePoint = (((uint)l_Char * 1024) + (uint)l_NextChar - 0x35FDC00);

                            return true;
                        }
                    }

                    // If we land here, there is an encoding error as
                    // this is not a legal surrogate pair. Treat it as
                    // a normal single code unit.
                }

                a_CodePoint = (uint)l_Char;

                return true;
            }
        }

        [StructLayout(LayoutKind.Explicit, Size=sizeof(uint))]
        private struct FloatUnion
        {
            // This simple union even works on Mono (more advanced
            // layouts can be problematic for Mono).

            [FieldOffset(0)]
            public uint f_UInt32;
            [FieldOffset(0)]
            public float f_Float;
        }

        public enum Read7BitEncUInt32Status : byte
        {
            Success = 0,
            TooFewSrcBytes = 1,
            FormatError = 2,
        }

        private static class RandomProvider
        {
            // This is the block size of our native RNG.
            private const byte c_RandBlockByteSize = 20;
            // Allocation should happen in ulongs, since we want
            // buffers passed to the RNG to be 8 byte aligned.
            private const byte c_ULongsPerBlock = ((c_RandBlockByteSize + (sizeof(ulong) - 1)) / sizeof(ulong));

            // The following two references can be null.
            private static RNGCryptoServiceProvider s_RNG;
            private static byte[] s_RNGBuffer;

            // The following two fields can be read/written to by multiple
            // threads without synchronization (that's by design).
            private static uint s_Randz;
            private static uint s_Randw;

            private static bool sp_RelyOnSelf
            {
                get { return (null == s_RNG); }
            }

            static RandomProvider()
            {
#if FAST
                if (true == TryInitializeFAST())
                {
                    return;
                }
#endif

                CreateFrameworkProvider();
                ReseedSecure();
            }

            private static void CreateFrameworkProvider()
            {
                Util.MonitorEnter(typeof(RandomProvider));
                {
                    if (null == s_RNG)
                    {
                        s_RNG = new RNGCryptoServiceProvider();
                        s_RNGBuffer = new byte[c_RandBlockByteSize];
                    }
                }
                Util.MonitorLeave(typeof(RandomProvider));
            }

            public static void ReseedSecure()
            {
#if FAST
                if (true == sp_RelyOnSelf)
                {
                    bool l_Succ = ReseedSecureFAST();
                }
#endif

                SeedQuickRandomGenerator(Util.GetRandomULong());
            }

            public static void FillBufferSecure(ref ByteArray a_Dest)
            {
                FillBufferSecure(a_Dest.f_Arr, (uint)a_Dest.f_Idx, (uint)a_Dest.f_Len);
            }

            public static void FillBufferSecure(byte[] a_Buf,
                                                uint a_Idx,
                                                uint a_Len)
            {
                FillBufferSecure(a_Buf, a_Idx, a_Len, false);
            }

            public static void FillBufferSecure(byte[] a_Buf,
                                                uint a_Idx,
                                                uint a_Len,
                                                bool a_DisallowZeros)
            {
                if (0 == a_Len)
                {
                    return;
                }

#if FAST
                if (true == sp_RelyOnSelf)
                {
                    if (false == Util.CheckRangeUnsigned(a_Idx, a_Len, (uint)a_Buf.Length))
                    {
                        throw (new ArgumentOutOfRangeException());
                    }

                    if (true == FillBufferSecureFAST(a_Buf, a_Idx, a_Len, a_DisallowZeros))
                    {
                        return;
                    }

                    CreateFrameworkProvider();
                }
#endif

                byte[] l_EffBuf;

                if (0 == a_Idx &&
                    (uint)a_Buf.Length == a_Len)
                {
                    // Fill entire buffer.

                    l_EffBuf = a_Buf;
                }
                else
                {
                    // Sadly the interface doesn't allow us to
                    // partially fill a buffer, so we need a temp
                    // buffer.

                    l_EffBuf = new byte[a_Len];
                }

                EnterFrameworkRNG();
                {
                    if (false == a_DisallowZeros)
                    {
                        s_RNG.GetBytes(l_EffBuf);
                    }
                    else
                    {
                        s_RNG.GetNonZeroBytes(l_EffBuf);
                    }
                }
                LeaveFrameworkRNG();

                if (a_Buf != l_EffBuf)
                {
                    Array.Copy(l_EffBuf, 0, a_Buf, (int)a_Idx, (int)a_Len);
                }
            }

            public static uint GetSecureUInts(out ulong a_Rand0,
                                              out ulong a_Rand1)
            {
                uint l_ResRand;

#if FAST
                if (true == sp_RelyOnSelf)
                {
                    if (true == GetSecureUIntsFAST(out a_Rand0, out a_Rand1, out l_ResRand))
                    {
                        return l_ResRand;
                    }

                    CreateFrameworkProvider();
                }
#endif

                EnterFrameworkRNG();
                {
                    s_RNG.GetBytes(s_RNGBuffer);

                    a_Rand0 = Util.ReadUInt64LE(s_RNGBuffer, 0);
                    a_Rand1 = Util.ReadUInt64LE(s_RNGBuffer, sizeof(ulong));
                    l_ResRand = Util.ReadUInt32LE(s_RNGBuffer, 2 * sizeof(ulong));
                }
                LeaveFrameworkRNG();

                return l_ResRand;
            }

            private static void EnterFrameworkRNG()
            {
                MonitorEnter(s_RNG);
            }

            private static void LeaveFrameworkRNG()
            {
                MonitorLeave(s_RNG);
            }

#if FAST

            // The following native library routines are thread-safe.

            [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void RandomReseed_x86();

            [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void RandomReseed_x64();

            [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void RandomQuery_x86([Out] IntPtr a_Dest,
                                                       [In] uint a_Length,
                                                       [In] bool a_DisallowZeros);

            [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
            private static extern void RandomQuery_x64([Out] IntPtr a_Dest,
                                                       [In] ulong a_Length,
                                                       [In] bool a_DisallowZeros);

            private static unsafe bool TryInitializeFAST()
            {
                ulong* l_Buf = stackalloc ulong[c_ULongsPerBlock];

                if (true == ReseedSecureFAST() &&
                    true == FillBufferSecureCoreFAST(l_Buf, c_RandBlockByteSize, false))
                {
                    SeedQuickRandomGenerator(l_Buf[0]);

                    return true;
                }

                return false;
            }

            private static bool ReseedSecureFAST()
            {
                // Returns zero if the attempt failed.

                try
                {
                    if (sizeof(uint) == IntPtr.Size)
                    {
                        RandomReseed_x86();
                    }
                    else
                    {
                        RandomReseed_x64();
                    }

                    return true;
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }

                return false;
            }

            private static unsafe bool FillBufferSecureCoreFAST(void* a_Ptr,
                                                                uint a_Len,
                                                                bool a_DisallowZeros)
            {
                // Returns zero if the attempt failed.

                try
                {
                    if (sizeof(uint) == IntPtr.Size)
                    {
                        RandomQuery_x86((IntPtr)a_Ptr, a_Len, a_DisallowZeros);
                    }
                    else
                    {
                        RandomQuery_x64((IntPtr)a_Ptr, a_Len, a_DisallowZeros);
                    }

                    return true;
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }

                return false;
            }

            private static unsafe bool FillBufferSecureFAST(byte[] a_Buf,
                                                            uint a_Idx,
                                                            uint a_Len,
                                                            bool a_DisallowZeros)
            {
                // The passed range must be valid.

                fixed (byte* l_Buf = a_Buf)
                {
                    bool l_Succ = FillBufferSecureCoreFAST(&l_Buf[a_Idx], a_Len, a_DisallowZeros);

                    return l_Succ;
                }
            }

            private static unsafe bool GetSecureUIntsFAST(out ulong a_Rand0,
                                                          out ulong a_Rand1,
                                                          out uint a_Rand2)
            {
                ulong* l_Buf = stackalloc ulong[c_ULongsPerBlock];

                if (false == FillBufferSecureCoreFAST(l_Buf, c_RandBlockByteSize, false))
                {
                    a_Rand1 = a_Rand0 = 0;
                    a_Rand2 = 0;

                    return false;
                }

                a_Rand0 = l_Buf[0];
                a_Rand1 = l_Buf[1];
                a_Rand2 = *(uint*)(&l_Buf[2]);

                return true;
            }

#endif

            private static void SeedQuickRandomGenerator(ulong a_Value)
            {
                SeedQuickRandomGenerator((uint)(a_Value >> Util.c_BitsPerUInt), (uint)a_Value);
            }

            private static void SeedQuickRandomGenerator(uint a_Value0,
                                                         uint a_Value1)
            {
                // Certain values should not be used.

                s_Randw = (0 == a_Value0 ||
                           0x464FFFFF == a_Value0 ? (~a_Value0) : a_Value0);
                s_Randz = (0 == a_Value1 ||
                           0x9068FFFF == a_Value1 ? (~a_Value1) : a_Value1);
            }

            public static uint GetUInt32Quick()
            {
                // This routine produces cryptographically unsafe, but
                // well enough randomness without the locking overhead of
                // the above random routines. This implementation is
                // intentionally not thread-safe, to make it less
                // predictable.

                // Multiply-with-carry algorithm by George Marsaglia.
                s_Randw = ((18000 * (s_Randw & ushort.MaxValue)) + (s_Randw >> c_BitsPerUShort));
                s_Randz = ((36969 * (s_Randz & ushort.MaxValue)) + (s_Randz >> c_BitsPerUShort));

                return ((s_Randz << c_BitsPerUShort) + s_Randw);
            }

            public static ulong GetUInt64Quick()
            {
                return (((ulong)GetUInt32Quick() << Util.c_BitsPerUInt) | GetUInt32Quick());
            }
        }
    }

    public static class Buffer2
    {
#if FAST
        private static bool s_External;
        private static bool s_64;

        static Buffer2()
        {
            if (sizeof(uint) == IntPtr.Size)
            {
                try
                {
                    memcmp_x86(IntPtr.Zero, IntPtr.Zero, IntPtr.Zero);
                    memset_x86(IntPtr.Zero, 0, IntPtr.Zero);

                    s_External = true;
                }
                catch
                {
                }
            }
            else if (sizeof(ulong) == IntPtr.Size)
            {
                try
                {
                    memcmp_x64(IntPtr.Zero, IntPtr.Zero, IntPtr.Zero);
                    memset_x64(IntPtr.Zero, 0, IntPtr.Zero);

                    s_External = true;
                    s_64 = true;
                }
                catch
                {
                }
            }
        }

        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern int memcmp_x86([In] IntPtr a_Ptr1,
                                             [In] IntPtr a_Ptr2,
                                             [In] IntPtr a_Len);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern int memcmp_x64([In] IntPtr a_Ptr1,
                                             [In] IntPtr a_Ptr2,
                                             [In] IntPtr a_Len);

        [DllImport("Nativex86.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void memset_x86([In] IntPtr a_Ptr,
                                              [In] int a_Val,
                                              [In] IntPtr a_Len);

        [DllImport("Nativex86-64.dll", CallingConvention = CallingConvention.Cdecl)]
        private static extern void memset_x64([In] IntPtr a_Ptr,
                                              [In] int a_Val,
                                              [In] IntPtr a_Len);

#endif

        public static void BlockZero(byte[] a_Dest,
                                     int a_DestOffset,
                                     int a_Count)
        {
            BlockSet(a_Dest, a_DestOffset, a_Count, 0);
        }

        public static void BlockZero(ref ByteArray a_Data)
        {
            BlockSet(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len, 0);
        }

        public static void BlockZero(byte[] a_Bytes)
        {
            BlockSet(a_Bytes, 0, a_Bytes.Length, 0);
        }

        public static void BlockZero(uint[] a_Dest,
                                     int a_ElementIndex,
                                     int a_ElementCount)
        {
            SetUIntArray(a_Dest, a_ElementIndex, a_ElementCount, 0);
        }

        public static
#if FAST
        unsafe
#endif
        void BlockSet(byte[] a_Dest,
                      int a_DestOffset,
                      int a_Count,
                      byte a_Value)
        {
#if FAST
            if (true == s_External)
            {
                Util.VerifyRangeOrFail(a_DestOffset, a_Count, a_Dest.Length);

                fixed (byte* l_Ptr = a_Dest)
                {
                    BlockSetCore(l_Ptr + a_DestOffset, a_Count, a_Value);
                }

                return;
            }
#endif

            if (0 == a_Value)
            {
                // In this case we are better off using the framework.

                Array.Clear(a_Dest, a_DestOffset, a_Count);

                return;
            }

            // Check for overflow.

            for (int i = a_DestOffset, lim = (a_DestOffset + a_Count); i < lim; i++)
            {
                a_Dest[i] = (byte)a_Value;
            }
        }

        public static
#if FAST
        unsafe
#endif
        void SetUIntArray(uint[] a_UIntArr,
                          int a_ElementIndex,
                          int a_ElementCount,
                          uint a_Value2)
        {
            byte a_Value = (byte)a_Value2;

            // a_Value will be assigned to each byte affected.

#if FAST
            if (true == s_External)
            {
                Util.VerifyRangeOrFail(a_ElementIndex, a_ElementCount, a_UIntArr.Length);

                fixed (uint* l_Ptr = a_UIntArr)
                {
                    BlockSetCore((byte*)l_Ptr + (a_ElementIndex * sizeof(uint)), a_ElementCount * sizeof(uint), a_Value);
                }

                return;
            }
#endif

            if (0 == a_Value)
            {
                // In this case we are better off using the framework.

                Array.Clear(a_UIntArr, a_ElementIndex, a_ElementCount);

                return;
            }

            // Check for overflow.

            uint l_Val = (uint)(a_Value |
                                (a_Value << Util.c_BitsPerByte) |
                                (a_Value << Util.c_BitsPerUShort) |
                                (a_Value << 24));

            for (int i = a_ElementIndex, lim = (a_ElementIndex + a_ElementCount); i < lim; i++)
            {
                a_UIntArr[i] = l_Val;
            }
        }

        public static bool ValueEquals(byte[] a_Arr1,
                                       byte[] a_Arr2)
        {
            return (a_Arr1.Length == a_Arr2.Length &&
                    0 == BlockCompare(a_Arr1, 0, a_Arr2, 0, a_Arr1.Length));
        }

        public static bool ValueEquals(byte[] a_Arr1,
                                       int a_Offset1,
                                       byte[] a_Arr2,
                                       int a_Offset2,
                                       int a_Count)
        {
            return (0 == BlockCompare(a_Arr1, a_Offset1, a_Arr2, a_Offset2, a_Count));
        }

        public static int BlockCompare(byte[] a_Arr1,
                                       byte[] a_Arr2)
        {
            return BlockCompare(a_Arr1, 0, a_Arr2, 0, a_Arr1.Length);
        }

        public static
#if FAST
        unsafe
#endif
        int BlockCompare(byte[] a_Arr1,
                         int a_Offset1,
                         byte[] a_Arr2,
                         int a_Offset2,
                         int a_Count)
        {
#if FAST
            if (true == s_External)
            {
                Util.VerifyRangeOrFail(a_Offset1, a_Count, a_Arr1.Length);
                Util.VerifyRangeOrFail(a_Offset2, a_Count, a_Arr2.Length);

                if (0 == a_Count)
                {
                    return 0;
                }

                fixed (byte* l_Ptr1 = a_Arr1, l_Ptr2 = a_Arr2)
                {
                    // Do a quick check if calling native code makes
                    // sense or if we can settle things quickly.
                    if (a_Arr1[a_Offset1] != a_Arr2[a_Offset2])
                    {
                        return (a_Arr1[a_Offset1] < a_Arr2[a_Offset2] ? (-1) : 1);
                    }

                    return BlockCompareCore(l_Ptr1 + a_Offset1, l_Ptr2 + a_Offset2, a_Count);
                }
            }
#endif

            // Check for overflow.

            for (int i = a_Offset1, k = a_Offset2, lim = (a_Offset1 + a_Count); i < lim; i++, k++)
            {
                if (a_Arr1[i] != a_Arr2[k])
                {
                    // This is what memcmp() would do.

                    return (a_Arr1[i] < a_Arr2[k] ? (-1) : 1);
                }
            }

            return 0;
        }

#if FAST
        private static unsafe void BlockSetCore(void* a_Ptr,
                                                int a_Count,
                                                int a_Value)
        {
            if (false == s_64)
            {
                memset_x86((IntPtr)a_Ptr, a_Value, (IntPtr)a_Count);
            }
            else
            {
                memset_x64((IntPtr)a_Ptr, a_Value, (IntPtr)a_Count);
            }
        }

        private static unsafe int BlockCompareCore(void* a_Ptr1,
                                                   void* a_Ptr2,
                                                   int a_Count)
        {
            // The caller should not rely on the possibility that the
            // return value can be safely casted to a smaller type
            // than int.

            if (false == s_64)
            {
                return memcmp_x86((IntPtr)a_Ptr1, (IntPtr)a_Ptr2, (IntPtr)a_Count);
            }
            else
            {
                return memcmp_x64((IntPtr)a_Ptr1, (IntPtr)a_Ptr2, (IntPtr)a_Count);
            }
        }
#endif
    }

    public class UnexpectedStateException : System.ApplicationException
    {
    }
}

// End of file.
