/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Support\Util2.cs
 * PURPOSE: Complex helper types
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Threading;
using Microsoft.VisualBasic.FileIO;
using Support.Encodings;
using Support.HashFunctions;

namespace Support
{
    public class FHandle
    {
        // Open file streams only through this class to avoid handle
        // leaks.

        public const ulong c_MaxFileSize = long.MaxValue;
        public const uint c_SuggestedBufferSize = (512 * 1024);

        private const uint c_GenericRead = 0x80000000;
        private const uint c_GenericWrite = 0x40000000;
        private const uint c_FileAttrNormal = 0x80;

        private const bool c_TreatReadBeyondEndAsArgExcByDefault = true;

        private static readonly bool s_UseWinAPI;
#if !MONO
        // Only used with the original .NET framework.
        private static readonly System.Reflection.FieldInfo s_FieldInfo;
#endif

        private readonly FileStream f_Strm;
        private readonly bool f_ThreadSafe;

        static FHandle()
        {
#if FAST
            if (true == Util.s_IsWindows)
            {
                // Test if the WinAPI is reachable.

                string l_AssemblyWinPath = BuildWinAPIUnicodePath(System.Reflection.Assembly.GetExecutingAssembly().Location);

                try
                {
                    uint l_Res = GetFileAttributesW(l_AssemblyWinPath);

                    s_UseWinAPI = true;
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
            }
#endif

#if !MONO
            // Mono uses a field named "name" for this, but contrary
            // to the original framework the Mono framework depends
            // on the name to be there. So in case of Mono do nothing.

            System.Reflection.BindingFlags l_BindFlags = (System.Reflection.BindingFlags.NonPublic |
                                                          System.Reflection.BindingFlags.Instance);

            s_FieldInfo = typeof(FileStream).GetField("_fileName", l_BindFlags);
#endif
        }

        private FHandle(FileStream a_Strm,
                        bool a_ThreadSafe)
        {
            f_Strm = a_Strm;
            f_ThreadSafe = a_ThreadSafe;
        }

        private static WinAPIWorthiness ShouldTryToUseWinAPI(string a_Path)
        {
#if FAST
            if (true == s_UseWinAPI)
            {
                if (false == LongPath.HasWindowsExtendedLengthPrefix(a_Path) &&
                    false == LongPath.HasWindowsDevicePrefix(a_Path))
                {
                    return WinAPIWorthiness.SuitableForWinAPI;
                }
            }
#endif

            return WinAPIWorthiness.MustNotUseWinAPI;
        }

        public static FHandle OpenR(string a_Path,
                                    bool a_CanBeMovedOrDeleted,
                                    FileOptions a_Options,
                                    uint a_BufferSize,
                                    ThreadSafetyMode a_ThreadSafetyMode,
                                    out Exception a_Exc)
        {
            FileShare l_Share = FileShare.Read;

            if (true == a_CanBeMovedOrDeleted)
            {
                l_Share |= FileShare.Delete;
            }

            return OpenFHandle(a_Path, FileMode.Open, FileAccess.Read, l_Share, a_Options, a_BufferSize, a_ThreadSafetyMode, out a_Exc);
        }

        public static FHandle OpenRWRITE(Type a_CallerType,
                                         string a_Path,
                                         FileMode a_Mode,
                                         FileOptions a_Options,
                                         uint a_BufferSize,
                                         ThreadSafetyMode a_ThreadSafetyMode,
                                         out Exception a_Exc)
        {
            return OpenFHandle(a_Path, a_Mode, FileAccess.ReadWrite, FileShare.None, a_Options, a_BufferSize, a_ThreadSafetyMode, out a_Exc);
        }

        [DllImport("kernel32.dll", CharSet = CharSet.Unicode, CallingConvention = CallingConvention.StdCall, SetLastError = true)]
        private static extern IntPtr CreateFileW([In] string lpFileName,
                                                 [In] uint dwDesiredAccess,
                                                 [In] uint dwShareMode,
                                                 [In] IntPtr SecurityAttributes,
                                                 [In] uint dwCreationDisposition,
                                                 [In] uint dwFlagsAndAttributes,
                                                 [In] IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CallingConvention = CallingConvention.Winapi, CharSet = CharSet.Unicode, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool CreateDirectoryW([In] string lpPathName,
                                                    [In] IntPtr lpSecurityAttributes);

        internal static bool TryCreateDirectory(string a_Path)
        {
            // This routine is able to create several (nested)
            // directories at once if necessary.

            if (WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_Path))
            {
                string l_WinPath = BuildWinAPIUnicodePath(a_Path);

                try
                {
                    return CreateDirectoryW(l_WinPath, IntPtr.Zero);
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
            }
            else
            {
                try
                {
                    Directory.CreateDirectory(a_Path);

                    return true;
                }
                catch (ArgumentNullException)
                {
                    throw;
                }
                catch (ArgumentException)
                {
                }
                catch (IOException)
                {
                }
                catch (UnauthorizedAccessException)
                {
                }
                catch (NotSupportedException)
                {
                }
            }

            return false;
        }

        private static string BuildWinAPIUnicodePath(string a_RawPath)
        {
            return LongPath.AddWindowsExtendedLengthPrefix(a_RawPath);
        }

        private static FHandle OpenFHandle(string a_Path,
                                           FileMode a_Mode,
                                           FileAccess a_Access,
                                           FileShare a_Share,
                                           FileOptions a_Options,
                                           uint a_BufferSize,
                                           ThreadSafetyMode a_ThreadSafetyMode,
                                           out Exception a_Exc)
        {
            // This routine can open files that have very long paths
            // on Windows. We can't do this with the framework.

            // We shouldn't use modes of which we don't know that they
            // can be translated to native API correctly.

            if (0 == a_BufferSize)
            {
                // The framwork doesn't allow zero buffers.

                a_BufferSize = 512;
            }

#if MONO
            // Async io on Mono (v2.10.8) lead to problems, so
            // deactivate.
            if (0 != (FileOptions.Asynchronous & a_Options))
            {
                a_Options &= ~FileOptions.Asynchronous;
            }
#endif

            FileStream l_Strm = null;

            if (WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_Path))
            {
                // Convert .NET access value into native format.
                uint l_Acc = 0;
                if (0 != (a_Access & FileAccess.Read)) l_Acc |= c_GenericRead;
                if (0 != (a_Access & FileAccess.Write)) l_Acc |= c_GenericWrite;

                // The mode and the share value are alrady in the
                // native format.

                bool l_HadExc = false;
                IntPtr l_Handle = new IntPtr();
                string l_ApiPath = BuildWinAPIUnicodePath(a_Path);

                try
                {
                    l_Handle = CreateFileW(l_ApiPath, l_Acc, (uint)a_Share, IntPtr.Zero, (uint)a_Mode, c_FileAttrNormal | (uint)a_Options, IntPtr.Zero);
                }
                catch (BadImageFormatException)
                {
                    l_HadExc = true;
                }
                catch (TypeLoadException)
                {
                    l_HadExc = true;
                }

                if (false == l_HadExc)
                {
                    Microsoft.Win32.SafeHandles.SafeFileHandle l_SafeHandle = new Microsoft.Win32.SafeHandles.SafeFileHandle(l_Handle, true);

                    if (false == l_SafeHandle.IsInvalid)
                    {
                        bool l_Async = (0 != (a_Options & FileOptions.Asynchronous));

                        try
                        {
                            l_Strm = new FileStream(l_SafeHandle, a_Access, (int)a_BufferSize, l_Async);
                        }
                        catch (ArgumentOutOfRangeException)
                        {
                            throw;
                        }
                        catch (ArgumentException)
                        {
                        }
                        catch (IOException)
                        {
                        }
                        catch (UnauthorizedAccessException)
                        {
                        }
                        catch (System.Security.SecurityException)
                        {
                        }

                        if (null == l_Strm)
                        {
                            l_SafeHandle.Close();
                        }
                    }
                    else
                    {
                        l_SafeHandle.Dispose();
                    }
                }
            }

            if (null == l_Strm)
            {
                try
                {
                    l_Strm = new FileStream(a_Path, a_Mode, a_Access, a_Share, (int)a_BufferSize, a_Options);
                }
                catch (ArgumentNullException)
                {
                    throw;
                }
                catch (ArgumentException l_Exc)
                {
                    a_Exc = l_Exc;

                    goto FAIL;
                }
                catch (IOException l_Exc)
                {
                    a_Exc = l_Exc;

                    goto FAIL;
                }
                catch (UnauthorizedAccessException l_Exc)
                {
                    a_Exc = l_Exc;

                    goto FAIL;
                }
                catch (System.Security.SecurityException l_Exc)
                {
                    a_Exc = l_Exc;

                    goto FAIL;
                }
                catch (NotSupportedException l_Exc)
                {
                    a_Exc = l_Exc;

                    goto FAIL;
                }
            }

#if !MONO
            SetFrameworkFileNameField(l_Strm, string.Empty);
#endif

            a_Exc = null;

            return new FHandle(l_Strm, ThreadSafetyMode.NotThreadSafe != a_ThreadSafetyMode);

        FAIL:

            return null;
        }

        ~FHandle()
        {
            f_Strm.Dispose();
        }

        private void Enter()
        {
            if (true == f_ThreadSafe)
            {
                Util.MonitorEnter(f_Strm);
            }
        }

        private void Leave()
        {
            // This routine must be called once for every time Enter()
            // was called.

            if (true == f_ThreadSafe)
            {
                Util.MonitorLeave(f_Strm);
            }
        }

        public void Close()
        {
            Close(false);
        }

        public void Close(bool a_Flush)
        {
            if (true == a_Flush)
            {
                bool l_Flushed = Flush();
            }

            Enter();
            {
                f_Strm.Dispose();
            }
            Leave();

            GC.SuppressFinalize(this);
        }

        public bool Flush()
        {
            // NOTE: It's not an error to flush a file that was opened
            // with read-only access.

            Enter();

            try
            {
                f_Strm.Flush();
            }
            catch (ObjectDisposedException)
            {
                return false;
            }
            catch (IOException)
            {
                return false;
            }
            finally
            {
                Leave();
            }

            return true;
        }

        public bool GetSize(out ulong a_Size)
        {
            Enter();

            try
            {
                // This usually works even if the file was deleted
                // in the meantime, as long as it wasn't closed.

                a_Size = (ulong)f_Strm.Length;

                return true;
            }
            catch (ObjectDisposedException)
            {
            }
            catch (IOException)
            {
            }
            finally
            {
                Leave();
            }

            a_Size = 0;

            return false;
        }

        public bool SetSize(ulong a_Size)
        {
            Enter();

            try
            {
                f_Strm.SetLength((long)a_Size);

                return true;
            }
            catch (ObjectDisposedException)
            {
            }
            catch (IOException)
            {
            }
            finally
            {
                Leave();
            }

            return false;
        }

        public bool Read(ulong a_FilePos,
                         ref ByteArray a_Data)
        {
            return Read(a_FilePos, ref a_Data, c_TreatReadBeyondEndAsArgExcByDefault);
        }

        public bool Read(ulong a_FilePos,
                         ref ByteArray a_Data,
                         bool a_TreatReadBeyondEndAsArgExc)
        {
            // a_TreatReadBeyondEndAsArgExc should not be used with
            // files that have a volatile file size (handles that
            // share write access). NOTE: Zero byte reads are allowed
            // beyond the file's end.

            bool l_ThrowReadPastEndException = false;

            Enter();

            try
            {
                f_Strm.Position = (long)a_FilePos;
                int l_BytesRead = Util.ReadFromStream(f_Strm, ref a_Data);

                if (a_Data.f_Len == l_BytesRead)
                {
                    return true;
                }

                if (true == a_TreatReadBeyondEndAsArgExc &&
                    true == HasTriedToReadBeyondEnd((ulong)f_Strm.Length, a_FilePos, (uint)a_Data.f_Len))
                {
                    l_ThrowReadPastEndException = true;
                }
            }
            catch (IOException)
            {
            }
            catch (NotSupportedException)
            {
            }
            catch (ObjectDisposedException)
            {
            }
            finally
            {
                Leave();
            }

            if (true == l_ThrowReadPastEndException)
            {
                // Attempted to read past the end of the stream.
                throw (new ArgumentException());
            }

            return false;
        }

        public bool Write(ulong a_FilePos,
                          ref ByteArray a_Data)
        {
            Enter();

            try
            {
                f_Strm.Position = (long)a_FilePos;
                f_Strm.Write(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len);

                return true;
            }
            catch (IOException)
            {
            }
            catch (NotSupportedException)
            {
            }
            catch (ObjectDisposedException)
            {
            }
            finally
            {
                Leave();
            }

            return false;
        }

        public bool Zero(ulong a_FilePos,
                         uint a_Len)
        {
            byte[] l_Buf = new byte[(int)Math.Min(4 * 1024 * 1024, a_Len)];

            Enter();

            try
            {
                f_Strm.Position = (long)a_FilePos;

                for (uint l_ToGo = a_Len; 0 != l_ToGo; )
                {
                    int l_Cur = (int)Math.Min(l_Buf.Length, l_ToGo);

                    f_Strm.Write(l_Buf, 0, l_Cur);

                    l_ToGo -= (uint)l_Cur;
                }

                return true;
            }
            catch (IOException)
            {
            }
            catch (NotSupportedException)
            {
            }
            catch (ObjectDisposedException)
            {
            }
            finally
            {
                Leave();
            }

            return false;
        }

        public AsyncOperationContext ReadAsync(ulong a_FilePos,
                                               ref ByteArray a_Data)
        {
            return ReadAsync(a_FilePos, ref a_Data, c_TreatReadBeyondEndAsArgExcByDefault);
        }

        public AsyncOperationContext ReadAsync(ulong a_FilePos,
                                               ref ByteArray a_Data,
                                               bool a_TreatReadBeyondEndAsArgExc)
        {
            // Returns null on error.
            // a_TreatReadBeyondEndAsArgExc should not be used with
            // files that have a volatile file size (handles that
            // share write access). NOTE: Zero byte reads are allowed
            // beyond the file's end.

            AsyncOperationContext l_Cntxt = new AsyncOperationContext(this, FileAccess.Read, a_FilePos, ref a_Data, a_TreatReadBeyondEndAsArgExc);

#if MONO
            // Use sync io.
            bool l_Succ = Read(a_FilePos, ref a_Data, a_TreatReadBeyondEndAsArgExc);
            l_Cntxt.SetCompleted(l_Succ);
#else
            if (false == BeginAsyncOperation(l_Cntxt, ref a_Data))
            {
                return null;
            }
#endif

            return l_Cntxt;
        }

        public AsyncOperationContext WriteAsync(ulong a_FilePos,
                                                ref ByteArray a_Data)
        {
            // Returns null on error.

            AsyncOperationContext l_Cntxt = new AsyncOperationContext(this, FileAccess.Write, a_FilePos, ref a_Data, false);

#if MONO
            // Use sync io.
            bool l_Succ = Write(a_FilePos, ref a_Data);
            l_Cntxt.SetCompleted(l_Succ);
#else
            if (false == BeginAsyncOperation(l_Cntxt, ref a_Data))
            {
                return null;
            }
#endif

            return l_Cntxt;
        }

        private bool BeginAsyncOperation(AsyncOperationContext a_Cntxt,
                                         ref ByteArray a_Data)
        {
            // This routine only makes sense on handles opened using
            // FileOptions.Asynchronous (otherwise it will block).

            IAsyncResult l_AsyncRes;

            Enter();

            try
            {
                f_Strm.Position = (long)a_Cntxt.f_FilePos;

                if (FileAccess.Read == a_Cntxt.f_Access)
                {
                    l_AsyncRes = f_Strm.BeginRead(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len, HandleAsyncCompletion, a_Cntxt);
                }
                else
                {
                    l_AsyncRes = f_Strm.BeginWrite(a_Data.f_Arr, a_Data.f_Idx, a_Data.f_Len, HandleAsyncCompletion, a_Cntxt);
                }
            }
            catch (NotSupportedException)
            {
                return false;
            }
            catch (IOException)
            {
                return false;
            }
            catch (ObjectDisposedException)
            {
                return false;
            }
            finally
            {
                Leave();
            }

            if (true == l_AsyncRes.CompletedSynchronously)
            {
                HandleAsyncCompletion(l_AsyncRes);
            }

            return true;
        }

        private void HandleAsyncCompletion(IAsyncResult a_Res)
        {
            AsyncOperationContext l_Cntxt = (AsyncOperationContext)a_Res.AsyncState;

            bool l_ThrowReadPastEndException = false;
            bool l_Succ = false;

            Enter();

            try
            {
                if (FileAccess.Read == l_Cntxt.f_Access)
                {
                    // Mono has an internal exception when EndRead()
                    // is invoked. That's why we don't use async io
                    // on Mono.
                    if (f_Strm.EndRead(a_Res) == l_Cntxt.f_Data.f_Len)
                    {
                        l_Succ = true;
                    }
                    else if (true == l_Cntxt.f_TreatReadBeyondEndAsArgExc &&
                             true == HasTriedToReadBeyondEnd((ulong)f_Strm.Length, l_Cntxt.f_FilePos, (uint)l_Cntxt.f_Data.f_Len))
                    {
                        l_ThrowReadPastEndException = true;
                    }
                }
                else
                {
                    f_Strm.EndWrite(a_Res);
                    l_Succ = true;
                }
            }
            catch (ArgumentException)
            {
                throw;
            }
            catch (IOException)
            {
            }
            catch (NotSupportedException)
            {
            }
            catch (ObjectDisposedException)
            {
            }
            finally
            {
                Leave();

                // We must complete in any case.
                l_Cntxt.SetCompleted(l_Succ);
            }

            if (true == l_ThrowReadPastEndException)
            {
                // Attempted to read past the end of the stream.
                throw (new ArgumentException());
            }
        }

        private static bool HasTriedToReadBeyondEnd(ulong a_FileSize,
                                                    ulong a_FilePos,
                                                    uint a_Count)
        {
            ulong l_FinalFilePos = (a_FilePos + a_Count);

            return (l_FinalFilePos < a_FilePos ||
                    a_FileSize < l_FinalFilePos);
        }

        [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool DeleteFileW([In] string lpFileName);

        public static bool Delete(string a_Path)
        {
            return Delete(a_Path, false);
        }

        public static bool Delete(string a_Path,
                                  bool a_JustMoveToRecycleBin)
        {
            // If a_JustMoveToRecycleBin is nonzero, this routine may
            // block.

            if (false == a_JustMoveToRecycleBin &&
                WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_Path))
            {
                // Try using the WinAPI, which allows us to use long
                // file names.

                string l_WinPath = BuildWinAPIUnicodePath(a_Path);

                try
                {
                    return DeleteFileW(l_WinPath);
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
            }

            try
            {
                if (false == a_JustMoveToRecycleBin)
                {
                    // If anyone holds a handle to this file, the
                    // operating system will just flag the file for
                    // later deletion (committed when the last handle
                    // is closed).

                    File.Delete(a_Path);
                }
                else
                {
                    // This operation can block (if the a dialog is
                    // shown).

                    FileSystem.DeleteFile(a_Path, UIOption.OnlyErrorDialogs, RecycleOption.SendToRecycleBin);
                }
            }
            catch (ArgumentNullException)
            {
                throw;
            }
            catch (ArgumentException)
            {
            }
            catch (IOException)
            {
                return false;
            }
            catch (UnauthorizedAccessException)
            {
                return false;
            }
            catch (NotSupportedException)
            {
                return false;
            }

            return true;
        }

        [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool MoveFileExW([In] string lpExistingFileName,
                                               [In] string lpNewFileName,
                                               [In] uint dwFlags);

        public static bool Move(string a_OldFullFileName,
                                string a_NewFullFileName)
        {
            if (WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_OldFullFileName) &&
                WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_NewFullFileName))
            {
                // Try using the WinAPI, which allows us to use long
                // file names.

                string l_OldPath = BuildWinAPIUnicodePath(a_OldFullFileName);
                string l_NewPath = BuildWinAPIUnicodePath(a_NewFullFileName);

                try
                {
                    return MoveFileExW(l_OldPath, l_NewPath, 0);
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
            }

            try
            {
                File.Move(a_OldFullFileName, a_NewFullFileName);

                return true;
            }
            catch (ArgumentNullException)
            {
                throw;
            }
            catch (ArgumentException)
            {
            }
            catch (IOException)
            {
            }
            catch (UnauthorizedAccessException)
            {
            }
            catch (NotSupportedException)
            {
            }

            return false;
        }

        public static AsyncCopyResult CopyAsync(FHandle a_Src,
                                                 FHandle a_Dest,
                                                 ref bool a_Run,
                                                 ref ulong a_Done)
        {
            // a_Done gets incremented (the input value is preserved).

            ulong l_Size;
            if (false == a_Src.GetSize(out l_Size))
            {
                return AsyncCopyResult.FailedToGetSrcSize;
            }

            if (false == a_Dest.SetSize(l_Size))
            {
                return AsyncCopyResult.FailedToSetDestSize;
            }

            if (0 == l_Size)
            {
                // The below loop doesn't terminate on zero input.

                return AsyncCopyResult.Success;
            }

            AsyncBuffers l_Bufs = new AsyncBuffers(l_Size);
            AsyncOperationContext l_ReadCntxt = null;
            AsyncOperationContext l_WriteCntxt = null;
            bool l_LastRead = false;
            bool l_NeedToWrite = false;
            ulong l_WritePos = 0;
            ByteArray l_DataForWriting = new ByteArray();

            for (ulong l_ReadPos = 0; true; )
            {
                if (false == Util.VolatileRead(ref a_Run))
                {
                    return AsyncCopyResult.AbortedByCaller;
                }

                if (null != l_ReadCntxt &&
                    true == l_ReadCntxt.p_Completed)
                {
                    // A read completed.

                    if (false == l_ReadCntxt.p_Success)
                    {
                        return AsyncCopyResult.FailedToReadFromSrc;
                    }

                    l_NeedToWrite = true;
                    l_WritePos = l_ReadCntxt.f_FilePos;
                    l_DataForWriting = l_ReadCntxt.f_Data;

                    l_ReadCntxt = null;
                }

                if (null != l_WriteCntxt &&
                    true == l_WriteCntxt.p_Completed)
                {
                    // A write completed.

                    if (false == l_WriteCntxt.p_Success)
                    {
                        return AsyncCopyResult.FailedToWriteToDest;
                    }

                    // NOTE: This add is not thread safe.
                    a_Done += (uint)l_WriteCntxt.f_Data.f_Len;

                    if (true == l_LastRead &&
                        (l_WriteCntxt.f_FilePos + (uint)l_WriteCntxt.f_Data.f_Len) == l_Size)
                    {
                        // We're done.

                        if (false == a_Dest.Flush())
                        {
                            return AsyncCopyResult.FailedToFlushDest;
                        }

                        return AsyncCopyResult.Success;
                    }

                    l_Bufs.ReleaseBuffer(l_WriteCntxt.f_Data.f_Arr);

                    l_WriteCntxt = null;
                }

                bool l_Sleep = true;

                if (null == l_ReadCntxt &&
                    false == l_LastRead)
                {
                    byte[] l_ReadBuf = l_Bufs.AcquireBuffer();

                    if (null != l_ReadBuf)
                    {
                        // Read on.

                        uint l_Cur = (uint)Math.Min((uint)l_ReadBuf.Length, l_Size - l_ReadPos);
                        ByteArray l_Data = new ByteArray(l_ReadBuf, 0, (int)l_Cur);

                        l_ReadCntxt = a_Src.ReadAsync(l_ReadPos, ref l_Data);

                        if (null == l_ReadCntxt)
                        {
                            return AsyncCopyResult.FailedToReadFromSrc;
                        }

                        l_ReadPos += l_Cur;
                        l_Sleep = false;

                        if (l_ReadPos == l_Size)
                        {
                            l_LastRead = true;
                        }
                    }
                }

                if (null == l_WriteCntxt &&
                    true == l_NeedToWrite)
                {
                    // Write on.

                    l_WriteCntxt = a_Dest.WriteAsync(l_WritePos, ref l_DataForWriting);

                    if (null == l_WriteCntxt)
                    {
                        return AsyncCopyResult.FailedToWriteToDest;
                    }

                    l_NeedToWrite = false;
                    l_Sleep = false;
                }

                if (true == l_Sleep)
                {
                    Util.MiniSleep();
                }
            }
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool GetFileTime([In] IntPtr hFile,
                                               [Out] out System.Runtime.InteropServices.ComTypes.FILETIME lpCreationTime,
                                               [Out] out System.Runtime.InteropServices.ComTypes.FILETIME lpLastAccessTime,
                                               [Out] out System.Runtime.InteropServices.ComTypes.FILETIME lpLastWriteTime);

        public bool GetLastWriteTime(string a_FallbackFileName,
                                     out Time a_Time,
                                     out Exception a_Exc)
        {
            // a_FallbackFileName is used if we are unable to retrieve
            // the file name using the handle. NOTE: If the file was
            // opened with write access, the result of this routine
            // is not reliable. In this case, close the file first
            // and call GetLastWriteTimeByFileName().

            if (true == s_UseWinAPI)
            {
                // Try using the WinAPI, which allows us to use long
                // file names. In this case we use the WinAPI also
                // because it allows us to use a handle instead of a
                // file name (the handle is more reliable).

                bool l_Succ = false;
                System.Runtime.InteropServices.ComTypes.FILETIME l_LastWriteFileTime = new System.Runtime.InteropServices.ComTypes.FILETIME();

                bool l_HandleReferenced = false;
                Microsoft.Win32.SafeHandles.SafeFileHandle l_Handle = null;

                Enter();

                try
                {
                    // Retrieving the safe handle resets the file pos
                    // (the manuals claims that).
                    l_Handle = this.f_Strm.SafeFileHandle;

                    if (false == l_Handle.IsClosed)
                    {
                        l_Handle.DangerousAddRef(ref l_HandleReferenced);

                        System.Runtime.InteropServices.ComTypes.FILETIME l_Junk0;
                        System.Runtime.InteropServices.ComTypes.FILETIME l_Junk1;

                        if (true == l_HandleReferenced &&
                            true == GetFileTime(l_Handle.DangerousGetHandle(), out l_Junk0, out l_Junk1, out l_LastWriteFileTime))
                        {
                            l_Succ = true;
                        }
                    }
                }
                catch (BadImageFormatException)
                {
                }
                catch (TypeLoadException)
                {
                }
                finally
                {
                    // Do not dispose l_Handle.

                    if (true == l_HandleReferenced)
                    {
                        l_Handle.DangerousRelease();
                    }

                    Leave();
                }

                if (true == l_Succ)
                {
                    a_Time = Time.FromWinFileTime(l_LastWriteFileTime);
                    a_Exc = null;

                    return true;
                }
            }

            return GetLastWriteTimeByFileName(a_FallbackFileName, out a_Time, out a_Exc);
        }

        public static bool GetLastWriteTimeByFileName(string a_FileName,
                                                      out Time a_Time,
                                                      out Exception a_Exc)
        {
            // NOTE: Changing the file size is also writing in the sense
            // of changing the last write time.

            DateTime l_DateTime;

            try
            {
                l_DateTime = File.GetLastWriteTimeUtc(a_FileName);
            }
            catch (ArgumentNullException)
            {
                throw;
            }
            catch (ArgumentException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (IOException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (UnauthorizedAccessException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (System.Security.SecurityException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (NotSupportedException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }

            a_Time = Time.FromDateTime(l_DateTime);
            a_Exc = null;

            return true;

        FAIL:

            a_Time = new Time();

            return false;
        }

        [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        private static extern uint GetFileAttributesW([In] string lpFileName);

        public static bool GetAttributes(string a_FileName,
                                         out FileAttributes a_Attrs,
                                         out Exception a_Exc)
        {
            // We can't get the file attributes without the file name,
            // even if we have a handle to the file.

            if (WinAPIWorthiness.WorthinessThreshold <= ShouldTryToUseWinAPI(a_FileName))
            {
                // Try using the WinAPI, which allows us to use long
                // file names.

                const uint INVALID_FILE_ATTRIBUTES = uint.MaxValue;

                string l_Path = BuildWinAPIUnicodePath(a_FileName);
                uint l_RetVal;

                try
                {
                    l_RetVal = GetFileAttributesW(l_Path);
                }
                catch (BadImageFormatException)
                {
                    l_RetVal = INVALID_FILE_ATTRIBUTES;
                }
                catch (TypeLoadException)
                {
                    l_RetVal = INVALID_FILE_ATTRIBUTES;
                }

                if (INVALID_FILE_ATTRIBUTES != l_RetVal)
                {
                    a_Attrs = (FileAttributes)l_RetVal;
                    a_Exc = null;

                    return true;
                }
            }

            try
            {
                a_Attrs = File.GetAttributes(a_FileName);
            }
            catch (ArgumentNullException)
            {
                throw;
            }
            catch (ArgumentException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (IOException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (UnauthorizedAccessException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }
            catch (NotSupportedException l_Exc)
            {
                a_Exc = l_Exc;

                goto FAIL;
            }

            a_Exc = null;

            return true;

        FAIL:

            a_Attrs = (FileAttributes)0;

            return false;
        }

#if DEBUG
        [Conditional("DEBUG")]
        public static void DebugCreateHashTestFile(string a_FileName,
                                                   ulong a_FileSize)
        {
            using (FileStream l_Strm = new FileStream(a_FileName, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                l_Strm.SetLength((long)a_FileSize);

                byte[] l_Buf = new byte[sizeof(ulong)];
                ulong l_Left = a_FileSize;
                ulong l_Num = 0;

                while (0 < l_Left)
                {
                    int l_Chunk = (int)Math.Min(l_Left, 1024);
                    int l_NumSize = Math.Min(l_Chunk, l_Buf.Length);

                    Util.WriteUInt64LE(++l_Num, l_Buf, 0);
                    l_Strm.Write(l_Buf, 0, l_NumSize);

                    l_Strm.Position += (l_Chunk - l_NumSize);
                    l_Left -= (uint)l_Chunk;
                }
            }
        }
#endif

#if !MONO
        public static bool SetFrameworkFileNameField(FileStream a_Strm,
                                                     string a_NewName)
        {
            // Use this routine to modify the frameworks internal file
            // name buffer, since it's useless (won't update if file
            // is moved) and wastes tons of memory when thousands of
            // files are open. In case of Mono the name must not be
            // altered.

            if (null != s_FieldInfo)
            {
                try
                {
                    s_FieldInfo.SetValue(a_Strm, a_NewName);

                    return true;
                }
                catch (FieldAccessException)
                {
                }
                catch (System.Reflection.TargetException)
                {
                }
            }

            return false;
        }
#endif

        public static bool CopyFileToFile(FHandle a_SrcFile,
                                          ulong a_SrcOffset,
                                          FHandle a_DestFile,
                                          ulong a_DestOffset,
                                          uint a_Size,
                                          bool a_Flush,
                                          byte[] a_TempBuffer)
        {
            // This routine is not capable of overlap handling.
            // a_TempBuffer may be null.

            if (0 == a_Size)
            {
                return true;
            }

            byte[] l_Buf = (null == a_TempBuffer ? new byte[256 * 1024] : a_TempBuffer);

            ulong l_SrcPos = a_SrcOffset;
            ulong l_DestPos = a_DestOffset;
            uint l_BytesLeft = a_Size;

            while (0 != l_BytesLeft)
            {
                uint l_Count = Math.Min((uint)l_Buf.Length, l_BytesLeft);
                ByteArray l_Data = new ByteArray(l_Buf, 0, (int)l_Count);

                if (false == a_SrcFile.Read(l_SrcPos, ref l_Data) ||
                    false == a_DestFile.Write(l_DestPos, ref l_Data))
                {
                    return false;
                }

                l_DestPos += l_Count;
                l_SrcPos += l_Count;
                l_BytesLeft -= l_Count;
            }

            if (true == a_Flush)
            {
                return a_DestFile.Flush();
            }

            return true;
        }

        public static bool GetAvailableVolumeSpace(string a_Path,
                                                   out ulong a_AvailBytes)
        {
            string l_RootPath = LongPath.GetPathRoot(a_Path);

            try
            {
                DriveInfo l_Info = new DriveInfo(l_RootPath);

                a_AvailBytes = (ulong)l_Info.AvailableFreeSpace;

                return true;
            }
            catch (ArgumentNullException)
            {
                throw;
            }
            catch (ArgumentException)
            {
            }
            catch (IOException)
            {
            }

            a_AvailBytes = 0;

            return false;
        }

        public enum ThreadSafetyMode : byte
        {
            NotThreadSafe = 0,
            NeedSynchronization = 1,
        }

        private enum WinAPIWorthiness : byte
        {
            MustNotUseWinAPI = 0,
            SuitableForWinAPI = 1,
            ShouldUseWinAPI = 2,

            WorthinessThreshold = SuitableForWinAPI,
        }

        public class AsyncOperationContext
        {
            public readonly FHandle f_File;
            public readonly FileAccess f_Access;
            public readonly ulong f_FilePos;
            public readonly ByteArray f_Data;
            public readonly bool f_TreatReadBeyondEndAsArgExc;

            private bool f_Completed;
            private bool f_Success;

            public bool p_Completed
            {
                get { return f_Completed; }
            }

            public bool p_Success
            {
                get { return f_Success; }
            }

            public AsyncOperationContext(FHandle a_File,
                                         FileAccess a_Access,
                                         ulong a_FilePos,
                                         ref ByteArray a_Data,
                                         bool a_TreatReadBeyondEndAsArgExc)
            {
                f_File = a_File;
                f_Access = a_Access;
                f_FilePos = a_FilePos;
                f_Data = a_Data;
                f_TreatReadBeyondEndAsArgExc = a_TreatReadBeyondEndAsArgExc;
            }

            public void SetCompleted(bool a_Success)
            {
                f_Success = a_Success;
                Util.VolatileWrite(out f_Completed, true);
            }
        }

        public enum AsyncCopyResult : byte
        {
            Success = 0,
            AbortedByCaller = 1,
            FailedToGetSrcSize = 2,
            FailedToSetDestSize = 3,
            FailedToReadFromSrc = 4,
            FailedToWriteToDest = 5,
            FailedToFlushDest = 6,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct AsyncBuffers
        {
            // Used by CopyAsync().

            private const uint c_MaxBufferSize = (1024 * 1024);
            private const uint c_MinBufferSize = (64 * 1024);

            public byte[] f_Buf1;
            public byte[] f_Buf2;

            public bool f_Buf1Busy;
            public bool f_Buf2Busy;

            public AsyncBuffers(ulong a_FileSize)
            {
                uint l_BufSize;

                if ((2 * c_MaxBufferSize) < a_FileSize)
                {
                    l_BufSize = c_MaxBufferSize;
                }
                else
                {
                    if (c_MinBufferSize >= a_FileSize)
                    {
                        // In this case make the buffers large enough
                        // to complete in one operation.

                        l_BufSize = c_MinBufferSize;
                    }
                    else
                    {
                        // Complete in two operations.

                        l_BufSize = (((uint)a_FileSize / 2) + 1);
                    }
                }

                f_Buf1 = new byte[(int)l_BufSize];
                f_Buf2 = new byte[(int)l_BufSize];

                f_Buf1Busy = false;
                f_Buf2Busy = false;
            }

            public byte[] AcquireBuffer()
            {
                if (false == f_Buf1Busy)
                {
                    f_Buf1Busy = true;

                    return f_Buf1;
                }

                if (false == f_Buf2Busy)
                {
                    f_Buf2Busy = true;

                    return f_Buf2;
                }

                return null;
            }

            public void ReleaseBuffer(byte[] a_Buf)
            {
                if (f_Buf1 == a_Buf)
                {
                    f_Buf1Busy = false;

                    return;
                }

                if (f_Buf2 == a_Buf)
                {
                    f_Buf2Busy = false;

                    return;
                }

                throw (new ApplicationException());
            }
        }
    }

    public abstract class BufferIO
    {
        // This limit allows for a string with ushort.MaxValue
        // characters in UTF-8, 16 and 32.
        public const uint c_DefaultMaxReadStringByteLength = (4 * ushort.MaxValue);
        private const uint c_AllocGranularity = 1024;

        private const ulong c_DateTimeKindBits = (3ul << 62);

        protected byte[] f_Buffer;
        protected uint f_Position;
        protected uint f_Length;

        public uint p_Position
        {
            get { return f_Position; }
        }

        public uint p_Length
        {
            get { return f_Length; }
        }

        protected BufferIO()
        {
        }

        protected uint GetBufferSize()
        {
            return (null == f_Buffer ? 0 : (uint)f_Buffer.Length);
        }

        public uint GetRemaining()
        {
            return (f_Length - f_Position);
        }

        public bool EndReached()
        {
            return (f_Length == f_Position);
        }

        public virtual void Flush()
        {
        }

        public virtual void Close()
        {
        }

        public abstract void SetLength(uint a_NewLength);

        public abstract void SetPosition(uint a_NewPos);

        public void JumpToStart()
        {
            SetPosition(0);
        }

        public void JumpToEnd()
        {
            SetPosition(f_Length);
        }

        public void JumpBack(uint a_Bytes)
        {
            uint l_NewPos;

            if (false == SubWithUnderflowCheck(f_Position, a_Bytes, out l_NewPos))
            {
                throw (new ArgumentException());
            }

            SetPosition(l_NewPos);
        }

        public void JumpAhead(uint a_Bytes)
        {
            uint l_NewPos;

            if (false == AddWithOverflowCheck(f_Position, a_Bytes, out l_NewPos))
            {
                throw (new ArgumentException());
            }

            SetPosition(l_NewPos);
        }

        public void Align(uint a_Alignment,
                          bool a_Zero)
        {
            uint l_Count = (f_Position % a_Alignment);

            if (0 != l_Count)
            {
                l_Count = (a_Alignment - l_Count);
            }

            if (false == a_Zero)
            {
                JumpAhead(l_Count);
            }
            else
            {
                ZeroBytes(l_Count);
            }
        }

        protected abstract int PrepareRead(uint a_Size);

        protected abstract int PrepareWrite(uint a_Size);

        public virtual uint GetCapacity()
        {
            return 0;
        }

        public virtual void SetCapacity(uint a_NewCapacity)
        {
        }

        public void EnsureCapacity(uint a_MinCapacity)
        {
            if (GetCapacity() < a_MinCapacity)
            {
                SetCapacity(a_MinCapacity);
            }
        }

        public abstract byte[] CopyRange(uint a_Index,
                                         uint a_Count);

        public abstract void ExposeRange(uint a_Index,
                                         uint a_Count,
                                         out ByteArray a_Data);

        public bool ExposeRangeChecked(uint a_Index,
                                       uint a_Count,
                                       out ByteArray a_Data)
        {
            if (false == Util.CheckRangeUnsigned(a_Index, a_Count, f_Length))
            {
                a_Data = new ByteArray();

                return false;
            }

            ExposeRange(a_Index, a_Count, out a_Data);

            return true;
        }

        public abstract void HashRange(IHashFunction a_HashAlgo,
                                       uint a_Index,
                                       uint a_Count);

#region ReadCode

        public byte[] ReadBytes(uint a_Count)
        {
            int l_Idx = PrepareRead(a_Count);
            return Util.CopyBytes(f_Buffer, l_Idx, (int)a_Count);
        }

        public uint[] ReadUIntsLE(uint a_Count)
        {
            return ReadBaseTypeArray<uint>(a_Count, sizeof(uint));
        }

        public T[] ReadBaseTypeArray<T>(uint a_Elements,
                                        byte a_SizeOf)
        {
            int l_TotSize = checked((int)(a_Elements * a_SizeOf));
            int l_Idx = PrepareRead((uint)l_TotSize);

            T[] l_Dest = new T[a_Elements];
            Buffer.BlockCopy(f_Buffer, l_Idx, l_Dest, 0, l_TotSize);

            return l_Dest;
        }

        public bool ReadBool()
        {
            return (0 != ReadByte());
        }

        public byte ReadByte()
        {
            int l_Idx = PrepareRead(sizeof(byte));
            return f_Buffer[l_Idx];
        }

        public ushort ReadUInt16LE()
        {
            int l_Idx = PrepareRead(sizeof(ushort));
            return Util.ReadUInt16LE(f_Buffer, l_Idx);
        }

        public uint ReadUInt32LE()
        {
            int l_Idx = PrepareRead(sizeof(uint));
            return Util.ReadUInt32LE(f_Buffer, l_Idx);
        }

        public ulong ReadUInt64LE()
        {
            int l_Idx = PrepareRead(sizeof(ulong));
            return Util.ReadUInt64LE(f_Buffer, l_Idx);
        }

        public ulong ReadUIntVarSizeLE(uint a_ByteSize)
        {
            int l_Idx = PrepareRead(a_ByteSize);

            switch (a_ByteSize)
            {
                case sizeof(byte):
                    return f_Buffer[l_Idx];

                case sizeof(ushort):
                    return Util.ReadUInt16LE(f_Buffer, l_Idx);

                case sizeof(uint):
                    return Util.ReadUInt32LE(f_Buffer, l_Idx);

                case sizeof(ulong):
                    return Util.ReadUInt64LE(f_Buffer, l_Idx);

                default: throw (new UnexpectedStateException());
            }
        }

        public float ReadFloat()
        {
            int l_Idx = PrepareRead(sizeof(float));
            return Util.ReadFloat(f_Buffer, l_Idx);
        }

        public UInt128 ReadUInt128LE()
        {
            int l_Idx = PrepareRead(UInt128.c_ByteSize);
            return new UInt128(f_Buffer, l_Idx);
        }

        public UInt128 ReadUInt128BE()
        {
            int l_Idx = PrepareRead(UInt128.c_ByteSize);
            return UInt128.FromByteArrayBE(f_Buffer, l_Idx);
        }

        public UInt128 ReadUInt128ME()
        {
            int l_Idx = PrepareRead(UInt128.c_ByteSize);
            return UInt128.FromByteArrayME(f_Buffer, l_Idx);
        }

        public Time ReadTime()
        {
            ulong l_Ticks = ReadUInt64LE();

            return Time.FromBinary(l_Ticks);
        }

        public Time ReadFormerDateTime()
        {
            // Reads a time that once (v1.0.0) was a data time and
            // therefore might need to have the DateTimeKind deleted.
            // Can be seen as a more expensive ReadTime().

            return Time.FromBinary(ReadUInt64LE() & (~c_DateTimeKindBits));
        }

        public string ReadStringWithAutoDetection(uint a_ByteCount,
                                                  bool a_UTF8)
        {
            // This behaviour is similar to what iMule 2.3.2.0 does.

            const byte l_UTF8BOMByteLen = 3;
            const byte l_UTF16BOMByteLen = 2;
            int l_Idx = PrepareRead(a_ByteCount);

            if (l_UTF16BOMByteLen <= a_ByteCount)
            {
                // If a BOM is detected, a_UTF8 will be ignored.

                byte l_FirstByte = f_Buffer[l_Idx];
                Encoding l_Enc = null;
                int l_BOMLen = 0;

                if (l_UTF8BOMByteLen <= a_ByteCount &&
                    0xEF == l_FirstByte &&
                    0xBB == f_Buffer[l_Idx + 1] &&
                    0xBF == f_Buffer[l_Idx + 2])
                {
                    // UTF-8.

                    l_Enc = Encoding.UTF8;
                    l_BOMLen = l_UTF8BOMByteLen;
                }
                else if (0xFE == l_FirstByte &&
                         0xFF == f_Buffer[l_Idx + 1])
                {
                    // UTF-16 big endian.

                    l_Enc = Encoding.BigEndianUnicode;
                    l_BOMLen = l_UTF16BOMByteLen;
                }
                else if (0xFF == l_FirstByte &&
                         0xFE == f_Buffer[l_Idx + 1])
                {
                    // UTF-16 little endian.

                    l_Enc = Encoding.Unicode;
                    l_BOMLen = l_UTF16BOMByteLen;
                }

                if (null != l_Enc)
                {
                    return l_Enc.GetString(f_Buffer, l_Idx + l_BOMLen, (int)a_ByteCount - l_BOMLen);
                }
            }

            if (true == a_UTF8)
            {
                try
                {
                    return Util.s_UTF8Picky.GetString(f_Buffer, l_Idx, (int)a_ByteCount);
                }
                catch (System.Text.DecoderFallbackException)
                {
                    // Reading UTF-8 failed, so assume the ecoding is
                    // latin1.
                }
            }

            return Util.s_Latin1BestFit.GetString(f_Buffer, l_Idx, (int)a_ByteCount);
        }

        public string ReadSBString(uint a_ByteCount)
        {
            return ReadStringCore(Util.s_Latin1BestFit, a_ByteCount);
        }

        public string ReadSBStringUInt16Len()
        {
            return ReadStringCore(Util.s_Latin1BestFit, ReadUInt16LE());
        }

        public string ReadUTF8String(uint a_ByteCount)
        {
            return ReadStringCore(Encoding.UTF8, a_ByteCount);
        }

        public string ReadUTF8StringUInt16Len()
        {
            return ReadStringCore(Encoding.UTF8, ReadUInt16LE());
        }

        public string ReadUTF8StringEncLen()
        {
            return ReadStringCore(Encoding.UTF8, Read7BitEncUInt32());
        }

        public string ReadUTF16LEStringEncLen()
        {
            return ReadStringCore(Encoding.Unicode, Read7BitEncUInt32());
        }

        public string ReadUTF16LEStringEncLen(bool a_ReturnNullIfEmpty,
                                              int a_MaxLength)
        {
            return Util.TruncateStr(ReadUTF16LEStringEncLen(), a_ReturnNullIfEmpty, (uint)a_MaxLength);
        }

        public string ReadUTF32LEStringEncLen()
        {
            return ReadStringCore(Encoding.UTF32, Read7BitEncUInt32());
        }

        public string ReadStringFlexibelEncLen()
        {
            return ReadStringFlexibelEncLen(c_DefaultMaxReadStringByteLength);
        }

        public string ReadStringFlexibelEncLen(uint a_MaxByteCount)
        {
            FlexibleStringMark l_Mark = (FlexibleStringMark)ReadByte();

            if (FlexibleStringMark.HighNibbleMagic != (l_Mark & FlexibleStringMark.HighNibbleMagic))
            {
                // Unexpected value in high nibble.

                throw (new InterpretBufferException());
            }

            Encoding l_Encoding;

            switch (l_Mark & (FlexibleStringMark)0xF)
            {
                case FlexibleStringMark.Latin1:
                    l_Encoding = Util.s_Latin1BestFit;
                    break;

                case FlexibleStringMark.UTF16LE:
                    l_Encoding = Encoding.Unicode;
                    break;

                case FlexibleStringMark.UTF8:
                    l_Encoding = Encoding.UTF8;
                    break;

                default: throw (new InterpretBufferException());
            }

            return ReadStringCore(l_Encoding, Read7BitEncUInt32(), a_MaxByteCount);
        }

        private string ReadStringCore(Encoding a_Encoding,
                                      uint a_ByteCount)
        {
            return ReadStringCore(a_Encoding, a_ByteCount, c_DefaultMaxReadStringByteLength);
        }

        private string ReadStringCore(Encoding a_Encoding,
                                      uint a_ByteCount,
                                      uint a_MaxByteCount)
        {
            // Will never return null.

            if (0 == a_ByteCount)
            {
                return string.Empty;
            }

            if (a_MaxByteCount < a_ByteCount)
            {
                throw (new StringTooLargeException());
            }

            int l_Idx = PrepareRead(a_ByteCount);
            return a_Encoding.GetString(f_Buffer, l_Idx, (int)a_ByteCount);
        }

        public uint Read7BitEncUInt32()
        {
            uint l_ReadAheadSize = Math.Min(GetRemaining(), Util.c_Max7BitEncUInt32ByteLen);
            int l_Idx = PrepareRead(l_ReadAheadSize);

            uint l_Occupied;
            uint l_Num;
            Util.Read7BitEncUInt32Status l_Status = Util.TryRead7BitEncUInt32(f_Buffer, l_Idx, (int)l_ReadAheadSize, out l_Occupied, out l_Num);

            if (Util.Read7BitEncUInt32Status.Success == l_Status)
            {
                if (l_ReadAheadSize > l_Occupied)
                {
                    JumpBack(l_ReadAheadSize - l_Occupied);
                }

                return l_Num;
            }

            if (Util.Read7BitEncUInt32Status.TooFewSrcBytes == l_Status)
            {
                throw (new TryToReadBeyondEndException());
            }

            throw (new EncodedUIntFormatException());
        }

        public bool ReadIPEndPoint(out System.Net.IPEndPoint a_EndPoint)
        {
            int l_AddrLen = this.ReadByte();

            if (4 != l_AddrLen &&
                16 != l_AddrLen)
            {
                a_EndPoint = null;

                return false;
            }

            int l_Idx = PrepareRead((uint)l_AddrLen + sizeof(ushort));
            byte[] l_AddrBytes = Util.CopyBytes(f_Buffer, l_Idx, l_AddrLen);
            ushort l_Port = Util.ReadUInt16LE(f_Buffer, l_Idx + l_AddrLen);
            a_EndPoint = new System.Net.IPEndPoint(new System.Net.IPAddress(l_AddrBytes), l_Port);

            return true;
        }

#endregion

#region WriteCode

        public void WriteBytes(byte[] a_Src)
        {
            WriteBytes(a_Src, 0, a_Src.Length);
        }

        public void WriteBytes(byte[] a_Src,
                               int a_SrcIdx,
                               int a_SrcLen)
        {
            int l_DestIdx = PrepareWrite(checked((uint)a_SrcLen));
            Buffer.BlockCopy(a_Src, a_SrcIdx, f_Buffer, l_DestIdx, a_SrcLen);
        }

        public void WriteBytes(ref ByteArray a_Src)
        {
            WriteBytes(a_Src.f_Arr, a_Src.f_Idx, a_Src.f_Len);
        }

        public void WriteUIntsLE(uint[] a_Src)
        {
            int l_Size = checked(a_Src.Length * sizeof(uint));
            int l_Idx = PrepareWrite((uint)l_Size);
            Buffer.BlockCopy(a_Src, 0, f_Buffer, l_Idx, l_Size);
        }

        public void ZeroBytes(uint a_Count)
        {
            int l_Idx = PrepareWrite(a_Count);
            Buffer2.BlockZero(f_Buffer, l_Idx, (int)a_Count);
        }

        public void WriteBool(bool a_Bool)
        {
            WriteByte((byte)(false == a_Bool ? 0 : 1));
        }

        public void WriteByte(byte a_Src)
        {
            int l_Idx = PrepareWrite(sizeof(byte));
            f_Buffer[l_Idx] = a_Src;
        }

        public void WriteUInt16LE(ushort a_Src)
        {
            int l_Idx = PrepareWrite(sizeof(ushort));
            Util.WriteUInt16LE(a_Src, f_Buffer, l_Idx);
        }

        public void WriteUInt32LE(uint a_Src)
        {
            int l_Idx = PrepareWrite(sizeof(uint));
            Util.WriteUInt32LE(a_Src, f_Buffer, l_Idx);
        }

        public void WriteUInt64LE(ulong a_Src)
        {
            int l_Idx = PrepareWrite(sizeof(ulong));
            Util.WriteUInt64LE(a_Src, f_Buffer, l_Idx);
        }

        public void WriteUIntVarSizeLE(uint a_ByteSize,
                                       ulong a_Src)
        {
            int l_Idx = PrepareWrite(a_ByteSize);

            switch (a_ByteSize)
            {
                case sizeof(byte):
                    f_Buffer[l_Idx] = (byte)a_Src;
                    break;

                case sizeof(ushort):
                    Util.WriteUInt16LE((ushort)a_Src, f_Buffer, l_Idx);
                    break;

                case sizeof(uint):
                    Util.WriteUInt32LE((uint)a_Src, f_Buffer, l_Idx);
                    break;

                case sizeof(ulong):
                    Util.WriteUInt64LE(a_Src, f_Buffer, l_Idx);
                    break;

                default: throw (new UnexpectedStateException());
            }
        }

        public void WriteFloat(float a_Src)
        {
            int l_Idx = PrepareWrite(sizeof(float));
            Util.WriteFloat(a_Src, f_Buffer, l_Idx);
        }

        public void WriteUInt128LE(UInt128 a_Src)
        {
            int l_Idx = PrepareWrite(UInt128.c_ByteSize);
            a_Src.ToByteArrayLE(f_Buffer, l_Idx);
        }

        public void WriteUInt128BE(UInt128 a_Src)
        {
            int l_Idx = PrepareWrite(UInt128.c_ByteSize);
            a_Src.ToByteArrayBE(f_Buffer, l_Idx);
        }

        public void WriteUInt128ME(UInt128 a_Src)
        {
            int l_Idx = PrepareWrite(UInt128.c_ByteSize);
            a_Src.ToByteArrayME(f_Buffer, l_Idx);
        }

        public void WriteTime(Time a_Time)
        {
            WriteUInt64LE(a_Time.ToBinary());
        }

        public void Write7BitEncUInt32(uint a_Value)
        {
            uint l_Size = Util.Get7BitEncUInt32Size(a_Value);
            int l_Idx = PrepareWrite(l_Size);
            Util.Write7BitEncUInt32(a_Value, f_Buffer, l_Idx);
        }

        public void WriteSBStringUInt16Len(string a_Str)
        {
            WriteStringCore(a_Str, Util.s_Latin1BestFit, StringLengthPrefix.BytesAsUInt16);
        }

        public void WriteSBStringUInt32Len(string a_Str)
        {
            WriteStringCore(a_Str, Util.s_Latin1BestFit, StringLengthPrefix.BytesAsUInt32);
        }

        public void WriteUTF8StringUInt16Len(string a_Str)
        {
            WriteStringCore(a_Str, Encoding.UTF8, StringLengthPrefix.BytesAsUInt16);
        }

        public void WriteUTF8StringUInt32Len(string a_Str)
        {
            WriteStringCore(a_Str, Encoding.UTF8, StringLengthPrefix.BytesAsUInt32);
        }

        public void WriteUTF8StringEncLen(string a_Str)
        {
            WriteStringCore(a_Str, Encoding.UTF8, StringLengthPrefix.BytesAs7BitEncodedUInt);
        }

        public void WriteUTF16LEStringEncLen(string a_Str)
        {
            WriteStringCore(a_Str, Encoding.Unicode, StringLengthPrefix.BytesAs7BitEncodedUInt);
        }

        public void WriteUTF32LEStringEncLen(string a_Str)
        {
            WriteStringCore(a_Str, Encoding.UTF32, StringLengthPrefix.BytesAs7BitEncodedUInt);
        }

        public void WriteStringFlexibelEncLen(string a_Str)
        {
            // This routine automatically picks the best suitable
            // encoding (producing the fewest output length), whether
            // it is latin1, UTF-8 oder UTF-16 little endian.

            Encoding l_Encoding;
            FlexibleStringMark l_Mark;

            if (true == Util.IsLatin1Compatible(a_Str))
            {
                l_Encoding = Util.s_Latin1BestFit;
                l_Mark = FlexibleStringMark.Latin1;
            }
            else if (false == Util.IsUTF16ShorterThanUTF8(a_Str, true))
            {
                l_Encoding = Encoding.UTF8;
                l_Mark = FlexibleStringMark.UTF8;
            }
            else
            {
                l_Encoding = Encoding.Unicode;
                l_Mark = FlexibleStringMark.UTF16LE;
            }

            byte l_PrefixByte = (byte)(l_Mark | FlexibleStringMark.HighNibbleMagic);
            WriteStringCore(a_Str, l_Encoding, StringLengthPrefix.BytesAs7BitEncodedUInt, l_PrefixByte);
        }

        private void WriteStringCore(string a_Str,
                                     Encoding a_Encoding,
                                     StringLengthPrefix a_LenPrefixType)
        {
            WriteStringCore(a_Str, a_Encoding, a_LenPrefixType, 0);
        }

        private void WriteStringCore(string a_Str,
                                     Encoding a_Encoding,
                                     StringLengthPrefix a_LenPrefixType,
                                     byte a_PrefixByte)
        {
            // a_Str may be null. a_PrefixByte will be ignored if it's
            // zero. Otherwise it will be the first byte written.

            byte[] l_EncData;
            uint l_EncDataLen;

            if (null == a_Str ||
                0 >= a_Str.Length)
            {
                l_EncData = null;
                l_EncDataLen = 0;
            }
            else
            {
                l_EncData = a_Encoding.GetBytes(a_Str);
                l_EncDataLen = (uint)l_EncData.Length;
            }

            uint l_CountSize;

            switch (a_LenPrefixType)
            {
                case StringLengthPrefix.BytesAsUInt16:
                {
                    if (ushort.MaxValue < l_EncDataLen)
                    {
                        throw (new ApplicationException());
                    }

                    l_CountSize = sizeof(ushort);
                    break;
                }

                case StringLengthPrefix.BytesAsUInt32:
                    l_CountSize = sizeof(uint);
                    break;

                case StringLengthPrefix.BytesAs7BitEncodedUInt:
                    l_CountSize = Util.Get7BitEncUInt32Size(l_EncDataLen);
                    break;

                default: throw (new UnexpectedStateException());
            }

            // This addition will never overflow.
            int l_Idx = PrepareWrite(Util.BoolToByte(0 != a_PrefixByte) + l_CountSize + l_EncDataLen);

            if (0 != a_PrefixByte)
            {
                f_Buffer[l_Idx++] = a_PrefixByte;
            }

            switch (a_LenPrefixType)
            {
                case StringLengthPrefix.BytesAsUInt16:
                    Util.WriteUInt16LE((ushort)l_EncDataLen, f_Buffer, l_Idx);
                    break;

                case StringLengthPrefix.BytesAsUInt32:
                    Util.WriteUInt32LE(l_EncDataLen, f_Buffer, l_Idx);
                    break;

                case StringLengthPrefix.BytesAs7BitEncodedUInt:
                {
                    uint l_WrittenSize = Util.Write7BitEncUInt32(l_EncDataLen, f_Buffer, l_Idx);
                    break;
                }
            }

            if (null != l_EncData)
            {
                Buffer.BlockCopy(l_EncData, 0, f_Buffer, l_Idx + (int)l_CountSize, (int)l_EncDataLen);
            }
        }

        public void WriteIPEndPoint(System.Net.IPEndPoint a_EndPoint)
        {
            byte[] l_AddrBytes = a_EndPoint.Address.GetAddressBytes();
            int l_AddrLen = l_AddrBytes.Length;

            int l_Idx = PrepareWrite(sizeof(byte) + (uint)l_AddrLen + sizeof(ushort));
            f_Buffer[l_Idx] = (byte)l_AddrLen;
            Buffer.BlockCopy(l_AddrBytes, 0, f_Buffer, l_Idx + sizeof(byte), l_AddrLen);
            Util.WriteUInt16LE((ushort)a_EndPoint.Port, f_Buffer, l_Idx + sizeof(byte) + l_AddrLen);
        }

#endregion

        protected static uint SuggestBufferSize(uint a_AimedSize)
        {
            uint l_Remainder = (a_AimedSize % c_AllocGranularity);
            uint l_Res = (a_AimedSize + (0 == l_Remainder ? 0 : (c_AllocGranularity - l_Remainder)));

            return l_Res;
        }

        public static bool AddWithOverflowCheck(uint a_A,
                                                uint a_B,
                                                out uint a_Sum)
        {
            a_Sum = (a_A + a_B);

            return (a_A <= a_Sum);
        }

        public static bool SubWithUnderflowCheck(uint a_A,
                                                 uint a_B,
                                                 out uint a_Res)
        {
            if (a_A < a_B)
            {
                a_Res = 0;

                return false;
            }

            a_Res = (a_A - a_B);

            return true;
        }

        public enum LoadResult : byte
        {
            // Corruption also means the buffer pos is undefined.
            Corruption = 0,
            InvalidDataButBufferPosIsStillGood = 1,
            Success = 2,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct CountWriter
        {
            private readonly BufferIO f_Buffer;
            private readonly byte f_CountByteSize;
            private readonly uint f_CountPos;

            private CountWriter(BufferIO a_Buffer,
                                uint a_CountByteSize)
            {
                // a_Buffer will be moved a_CountByteSize bytes.

                f_Buffer = a_Buffer;
                f_CountByteSize = (byte)a_CountByteSize;
                f_CountPos = f_Buffer.p_Position;

                f_Buffer.JumpAhead(f_CountByteSize);
            }

            public static void InitAndMoveBuffer(BufferIO a_Buffer,
                                                 uint a_CountByteSize,
                                                 out CountWriter a_Dest)
            {
                a_Dest = new CountWriter(a_Buffer, a_CountByteSize);
            }

            public void Finish(uint a_Count)
            {
                uint l_StoredPos = f_Buffer.p_Position;
                f_Buffer.SetPosition(f_CountPos);

                switch (f_CountByteSize)
                {
                    case 1:
                        f_Buffer.WriteByte((byte)a_Count);
                        break;

                    case 2:
                        f_Buffer.WriteUInt16LE((ushort)a_Count);
                        break;

                    case 4:
                        f_Buffer.WriteUInt32LE(a_Count);
                        break;

                    default:
                        throw (new ApplicationException());
                }

                f_Buffer.SetPosition(l_StoredPos);
            }
        }

        public class BufferIOException : Exception
        {
        }

        public class FileAccessException : BufferIOException
        {
        }

        public class InterpretBufferException : BufferIOException
        {
            // This base class exception reflects the subset of errors
            // that are not due to the caller passing invalid args or
            // calling invalid routines, but due to the interpretation
            // process itself.
        }

        public class TryToReadBeyondEndException : InterpretBufferException
        {
        }

        public class EncodedUIntFormatException : InterpretBufferException
        {
        }

        public class StringTooLargeException : InterpretBufferException
        {
        }

        private enum StringLengthPrefix : byte
        {
            BytesAsUInt16 = 0,
            BytesAsUInt32 = 1,
            BytesAs7BitEncodedUInt = 2,
        }

        private enum FlexibleStringMark : byte
        {
            Latin1 = 1,
            UTF8 = 2,
            UTF16LE = 3,

            HighNibbleMagic = (0xB << Util.c_BitsPerNibble),
        }
    }

    sealed public class MemoryIO : BufferIO
    {
        public readonly bool f_IsWritable;
        // If f_IsResizeable is true, f_IsWritable must be true as
        // well.
        public readonly bool f_IsResizeable;

        private MemoryIO(bool a_IsWritable,
                         bool a_IsExpandable,
                         byte[] a_Buffer,
                         uint a_Length)
        {
            f_IsWritable = a_IsWritable;
            f_IsResizeable = a_IsExpandable;
            f_Buffer = a_Buffer;
            f_Length = a_Length;
        }

        public static MemoryIO CreateWritable()
        {
            // Writable memory with unlimited size.

            return new MemoryIO(true, true, null, 0);
        }

        public static MemoryIO CreateWritable(uint a_FixedSize)
        {
            // Writable memory with fixed size.

            return new MemoryIO(true, false, new byte[a_FixedSize], a_FixedSize);
        }

        public static MemoryIO CreateWritableAndGrowable(uint a_InitialCapacity)
        {
            // Writable memory with flexible size.

            return new MemoryIO(true, true, new byte[a_InitialCapacity], 0);
        }

        public static MemoryIO CreateReadable(byte[] a_Data)
        {
            // Read-only memory with fixed size.

            return new MemoryIO(false, false, a_Data, Util.GetArrayLength(a_Data));
        }

        public static MemoryIO CreateReadable(ref ByteArray a_Data)
        {
            // Read-only memory with fixed size.

            MemoryIO l_Buffer = new MemoryIO(false, false, a_Data.f_Arr, (uint)a_Data.f_Len);
            l_Buffer.f_Position = (uint)a_Data.f_Idx;
            l_Buffer.f_Length = (l_Buffer.f_Position + (uint)a_Data.f_Len);

            return l_Buffer;
        }

        public override void SetLength(uint a_NewLength)
        {
            if (a_NewLength <= f_Length)
            {
                if (false == f_IsResizeable)
                {
                    throw (new BufferIOException());
                }

                f_Position = Math.Min(f_Position, a_NewLength);
            }
            else
            {
                SetCapacityPlusExtra(a_NewLength);
            }

            f_Length = a_NewLength;
        }

        public override void SetPosition(uint a_NewPos)
        {
            if (a_NewPos > f_Length)
            {
                SetCapacityPlusExtra(a_NewPos);
                f_Length = a_NewPos;
            }

            f_Position = a_NewPos;
        }

        public void JettisonRange(uint a_Index,
                                  uint a_Count)
        {
            // Removes the given range from the buffer.

            if (false == f_IsResizeable)
            {
                throw (new BufferIOException());
            }

            uint l_FinalIdx = (a_Index + a_Count);

            if (l_FinalIdx < a_Index ||
                l_FinalIdx > f_Length)
            {
                throw (new ArgumentException());
            }

            if (0 == a_Count)
            {
                return;
            }

            Buffer.BlockCopy(f_Buffer, (int)l_FinalIdx, f_Buffer, (int)a_Index, (int)(f_Length - l_FinalIdx));

            f_Length -= a_Count;

            if (a_Index < f_Position)
            {
                if (l_FinalIdx >= f_Position)
                {
                    f_Position = a_Index;
                }
                else
                {
                    f_Position -= a_Count;
                }
            }
        }

        protected override int PrepareRead(uint a_Size)
        {
            uint l_NewPos = (f_Position + a_Size);

            if (l_NewPos < f_Position ||
                l_NewPos > f_Length)
            {
                throw (new TryToReadBeyondEndException());
            }

            int l_Idx = (int)f_Position;
            f_Position = l_NewPos;

            return l_Idx;
        }

        protected override int PrepareWrite(uint a_Size)
        {
            // f_Buffer can be modified (enlarged) by this routine

            uint l_NewPos = (f_Position + a_Size);

            if (false == f_IsWritable ||
                l_NewPos < f_Position)
            {
                throw (new BufferIOException());
            }

            if (l_NewPos > f_Length)
            {
                SetCapacityPlusExtra(l_NewPos);
                f_Length = l_NewPos;
            }

            int l_Idx = (int)f_Position;
            f_Position = l_NewPos;

            return l_Idx;
        }

        public override uint GetCapacity()
        {
            return GetBufferSize();
        }

        public override void SetCapacity(uint a_NewCapacity)
        {
            if (f_Length > a_NewCapacity)
            {
                throw (new ArgumentException());
            }

            if (false == f_IsResizeable)
            {
                throw (new BufferIOException());
            }

            Array.Resize(ref f_Buffer, (int)a_NewCapacity);
        }

        private void SetCapacityPlusExtra(uint a_AimedCapacity)
        {
            if (false == f_IsResizeable)
            {
                throw (new BufferIOException());
            }

            if (GetBufferSize() < a_AimedCapacity)
            {
                Array.Resize(ref f_Buffer, (int)SuggestBufferSize(a_AimedCapacity));
            }
        }

        private void TrimExcess()
        {
            Array.Resize(ref f_Buffer, (int)f_Length);
        }

        public byte[] ToArray(bool a_Cannibalize)
        {
            // Copies all data from the buffer (regardless of position).

            if (false == a_Cannibalize)
            {
                return Util.CopyBytes(f_Buffer, 0, (int)f_Length);
            }

            // Try to return the internal buffer directly, in any case
            // reset.

            // Make sure we don't return a buffer with wrong length.
            TrimExcess();

            byte[] l_Buffer = f_Buffer;

            f_Buffer = null;
            f_Position = 0;
            f_Length = 0;

            return l_Buffer;
        }

        public override byte[] CopyRange(uint a_Index,
                                         uint a_Count)
        {
            ByteArray l_Data = new ByteArray(f_Buffer, (int)a_Index, (int)a_Count);

            return l_Data.ToArray(true);
        }

        public override void ExposeRange(uint a_Index,
                                         uint a_Count,
                                         out ByteArray a_Data)
        {
            a_Data = new ByteArray(f_Buffer, (int)a_Index, (int)a_Count);
        }

        public override void HashRange(IHashFunction a_HashAlgo,
                                       uint a_Index,
                                       uint a_Count)
        {
            a_HashAlgo.Update(f_Buffer, (int)a_Index, (int)a_Count);
        }
    }

    sealed public class FileIO : BufferIO
    {
        public const uint c_FileBufferSize =
            (256 * 1024);

        private readonly FHandle f_File;
        private readonly bool f_IsWritable;

        private uint f_ValidFileSize;
        private uint f_LatestFileSize;

        private uint f_FileBufferPos;
        private uint f_FinalFileBufferPos;
        private uint f_FileBufferDirtyPos;
        private bool f_NeedFlush;

        public FileIO(FHandle a_File,
                      bool a_IsWritable)
        {
            f_File = a_File;
            f_IsWritable = a_IsWritable;

            f_Length = GetFileSize();
            f_ValidFileSize = f_Length;
            f_LatestFileSize = f_Length;
        }

        public static FileIO CreateWritable(FHandle a_File)
        {
            // Can trigger an FileAccessException.

            return new FileIO(a_File, true);
        }

        public static FileIO CreateReadable(FHandle a_File)
        {
            // Can trigger an FileAccessException.

            return new FileIO(a_File, false);
        }

        private uint GetFileSize()
        {
            ulong l_FileSize;

            if (false == f_File.GetSize(out l_FileSize))
            {
                throw (new FileAccessException());
            }

            return (uint.MaxValue >= l_FileSize ? (uint)l_FileSize : uint.MaxValue);
        }

        public override void Flush()
        {
            Flush(true);
        }

        public void Flush(bool a_ToDisk)
        {
            if (false == f_NeedFlush)
            {
                return;
            }

            if (f_Length != f_LatestFileSize)
            {
                if (false == f_File.SetSize(f_Length))
                {
                    throw (new FileAccessException());
                }

                f_ValidFileSize = Math.Min(f_ValidFileSize, f_Length);
                f_LatestFileSize = f_Length;
            }

            if (f_FileBufferDirtyPos != f_FileBufferPos)
            {
                ByteArray l_Data = new ByteArray(f_Buffer, 0, (int)(f_FileBufferDirtyPos - f_FileBufferPos));

                if (false == f_File.Write(f_FileBufferPos, ref l_Data) &&
                    (false == a_ToDisk ||
                     false == f_File.Flush()))
                {
                    throw (new FileAccessException());
                }

                f_ValidFileSize = Math.Max(f_ValidFileSize, f_FileBufferDirtyPos);
                f_FileBufferDirtyPos = f_FileBufferPos;
            }

            f_NeedFlush = false;
        }

        public override void Close()
        {
            Flush(true);

            f_File.Close(false);
        }

        private void PrepareFileBuffer(uint a_Index,
                                       uint a_Count,
                                       out ByteArray a_Data)
        {
            uint l_FinalPos = (a_Index + a_Count);

            if (null != f_Buffer &&
                a_Index >= f_FileBufferPos &&
                l_FinalPos <= f_FinalFileBufferPos)
            {
                a_Data = new ByteArray(f_Buffer, (int)(a_Index - f_FileBufferPos), (int)a_Count);

                return;
            }

            Flush(false);

            uint l_CoreSize = (a_Index >= f_ValidFileSize ? 0 : (f_ValidFileSize - a_Index));
            uint l_SuggSize = Math.Max(SuggestBufferSize(a_Count), false == f_IsWritable ? Math.Min(f_Length, c_FileBufferSize) : c_FileBufferSize);
            l_CoreSize = Math.Min(l_CoreSize, l_SuggSize);

            if (null == f_Buffer ||
                (uint)f_Buffer.Length != l_SuggSize)
            {
                f_Buffer = new byte[(int)l_SuggSize];
            }
            else if (l_CoreSize != l_SuggSize)
            {
                Buffer2.BlockZero(f_Buffer, (int)l_CoreSize, (int)(l_SuggSize - l_CoreSize));
            }

            if (0 != l_CoreSize)
            {
                ByteArray l_Data = new ByteArray(f_Buffer, 0, (int)l_CoreSize);

                if (false == f_File.Read(a_Index, ref l_Data))
                {
                    throw (new FileAccessException());
                }
            }

            f_FileBufferPos = a_Index;
            f_FinalFileBufferPos = (f_FileBufferPos + l_SuggSize);
            f_FileBufferDirtyPos = f_FileBufferPos;
            a_Data = new ByteArray(f_Buffer, 0, (int)a_Count);
        }

        public override void SetLength(uint a_NewLength)
        {
            if (false == f_IsWritable)
            {
                throw (new BufferIOException());
            }

            if (f_Length == a_NewLength)
            {
                return;
            }

            if (f_Length > a_NewLength)
            {
                f_Position = Math.Min(f_Position, a_NewLength);

                if (null != f_Buffer &&
                    a_NewLength < f_FinalFileBufferPos)
                {
                    int l_ValidBufLen = (int)(a_NewLength - f_FileBufferPos);

                    if (false == f_IsWritable)
                    {
                        Array.Resize(ref f_Buffer, l_ValidBufLen);
                        f_FinalFileBufferPos = a_NewLength;
                    }
                    else
                    {
                        Buffer2.BlockZero(f_Buffer, l_ValidBufLen, f_Buffer.Length - l_ValidBufLen);
                    }

                    f_FileBufferDirtyPos = Math.Max(f_FileBufferPos, a_NewLength);
                }
            }

            f_Length = a_NewLength;
            f_NeedFlush = true;
        }

        public override void SetPosition(uint a_NewPos)
        {
            if (a_NewPos > f_Length)
            {
                SetLength(a_NewPos);
            }

            f_Position = a_NewPos;
        }

        protected override int PrepareRead(uint a_Size)
        {
            uint l_NewPos;

            if (false == AddWithOverflowCheck(f_Position, a_Size, out l_NewPos) ||
                f_Length < l_NewPos)
            {
                throw (new TryToReadBeyondEndException());
            }

            ByteArray l_Data;
            PrepareFileBuffer(f_Position, a_Size, out l_Data);
            f_Position = l_NewPos;

            return l_Data.f_Idx;
        }

        protected override int PrepareWrite(uint a_Size)
        {
            uint l_NewPos;

            if (false == f_IsWritable ||
                false == AddWithOverflowCheck(f_Position, a_Size, out l_NewPos))
            {
                throw (new BufferIOException());
            }

            if (f_Length < l_NewPos)
            {
                f_Length = l_NewPos;
            }

            ByteArray l_Data;
            PrepareFileBuffer(f_Position, a_Size, out l_Data);
            f_Position = l_NewPos;
            f_FileBufferDirtyPos = Math.Max(f_FileBufferDirtyPos, l_NewPos);
            f_NeedFlush = true;

            return l_Data.f_Idx;
        }

        public override byte[] CopyRange(uint a_Index,
                                         uint a_Count)
        {
            ByteArray l_Data;
            PrepareFileBuffer(a_Index, a_Count, out l_Data);

            return l_Data.ToArray(true);
        }

        public override void ExposeRange(uint a_Index,
                                         uint a_Count,
                                         out ByteArray a_Data)
        {
            PrepareFileBuffer(a_Index, a_Count, out a_Data);
        }

        public override void HashRange(IHashFunction a_HashAlgo,
                                       uint a_Index,
                                       uint a_Count)
        {
            ByteArray l_Data;
            PrepareFileBuffer(a_Index, a_Count, out l_Data);
            a_HashAlgo.Update(ref l_Data);
        }
    }

    public class SensitiveBuffer : System.Runtime.ConstrainedExecution.CriticalFinalizerObject
    {
        private byte[] f_Buf;
        private byte[] f_Pad;

        public uint p_Length
        {
            get { return (uint)f_Buf.Length; }
        }

        private SensitiveBuffer(byte[] a_Buf,
                                byte[] a_Pad)
        {
            f_Buf = a_Buf;
            f_Pad = a_Pad;
        }

        ~SensitiveBuffer()
        {
            // This is a critical finalizer.

            ReleaseInternal(true);
        }

        private void ReleaseInternal(bool a_Critical)
        {
            byte[] l_Buf = f_Buf;
            byte[] l_Pad = f_Pad;

            if (true == a_Critical)
            {
                Util.MemoryBarrier();
            }

            if (null != l_Buf)
            {
                Array.Clear(l_Buf, 0, l_Buf.Length);
                f_Buf = null;
            }

            if (null != l_Pad)
            {
                Array.Clear(l_Pad, 0, l_Pad.Length);
                f_Pad = null;
            }
        }

        public void Release()
        {
            // Do not use the instance after Release() got called.

            if (true == IsValid())
            {
                ReleaseInternal(false);
                GC.SuppressFinalize(this);
            }
        }

        public bool IsEmpty()
        {
            return (0 == p_Length);
        }

        public bool IsValid()
        {
            return (null != f_Buf);
        }

        public bool IsValid(uint a_ExpectedSize)
        {
            return (true == IsValid() &&
                    p_Length == a_ExpectedSize);
        }

        public bool ValueEquals(SensitiveBuffer a_Other)
        {
            if (this.p_Length != a_Other.p_Length)
            {
                return false;
            }

            // We don't have to expose plain values to do this.

            byte[] l_TempBuf = Util.CloneArray(this.f_Buf);

            ScrambleStatic(l_TempBuf, a_Other.f_Pad);
            ScrambleStatic(l_TempBuf, this.f_Pad);

            uint l_XorSum = 0;
            byte[] l_OtherArr = a_Other.f_Buf;

            for (int i = 0, len = l_OtherArr.Length; i != len; i++)
            {
                l_XorSum |= ((uint)l_TempBuf[i] ^ l_OtherArr[i]);
            }

            Array.Clear(l_TempBuf, 0, l_TempBuf.Length);
            bool l_Same = (0 == l_XorSum);

            return l_Same;
        }

        public void ChangeKey()
        {
            byte[] l_NewPad = GenerateNewPad();

            Unscramble();
            {
                f_Pad = l_NewPad;
            }
            Scramble();
        }

        private byte[] GenerateNewPad()
        {
            return GenerateNewPadStatic(p_Length);
        }

        private static byte[] GenerateNewPadStatic(uint a_BufLen)
        {
            // 2,560.
            const ushort l_MaxPadLen = (128 * 20);
            uint l_EffLen;

            if (l_MaxPadLen <= a_BufLen)
            {
                l_EffLen = l_MaxPadLen;
            }
            else
            {
                uint l_Rem = (a_BufLen % 20);
                l_EffLen = (a_BufLen + (0 == l_Rem ? 0 : (20 - l_Rem)));
            }

            byte[] l_Pad = Util.GetRandomBytes((int)l_EffLen);

            return l_Pad;
        }

        private void Scramble()
        {
            ScrambleStatic(f_Buf, f_Pad);
        }

        private void Unscramble()
        {
            Scramble();
        }

        public SensitiveBuffer Clone()
        {
            if (true == IsEmpty())
            {
                return CreateEmpty();
            }

            // We don't have to expose plain values to do this.

            byte[] l_NewBuf = Util.CloneArray(f_Buf);
            byte[] l_NewPad = GenerateNewPad();

            ScrambleStatic(l_NewBuf, l_NewPad);
            ScrambleStatic(l_NewBuf, f_Pad);

            return new SensitiveBuffer(l_NewBuf, l_NewPad);
        }

        public static SensitiveBuffer CreateEmpty()
        {
            return new SensitiveBuffer(Util.s_Empty, Util.s_Empty);
        }

        public static SensitiveBuffer FromByteArray(byte[] a_PlainBytes)
        {
            // The new instance takes ownership of a_PlainBytes.

            byte[] l_NewPad = GenerateNewPadStatic((uint)a_PlainBytes.Length);

            ScrambleStatic(a_PlainBytes, l_NewPad);

            return new SensitiveBuffer(a_PlainBytes, l_NewPad);
        }

        public static SensitiveBuffer FromBuffer(BufferIO a_Buffer)
        {
            ulong l_UInt64Pad = a_Buffer.ReadUInt64LE();
            uint l_Length = a_Buffer.ReadUInt32LE();

            if (0 == l_Length)
            {
                return CreateEmpty();
            }

            byte[] l_Bytes = a_Buffer.ReadBytes(l_Length);

            SensitiveBuffer l_NewSensitiveBuffer = FromByteArray(l_Bytes);
            ScrambleMiniPadStatic(l_NewSensitiveBuffer.f_Buf, l_UInt64Pad);

            return l_NewSensitiveBuffer;
        }

        public static SensitiveBuffer FromRSAParameters(ref RSAParameters a_Params)
        {
            byte[][] l_Arr = RsaKeyPair.RSAParamAccessor.Get(ref a_Params);
            int l_TotalLen = 0;

            for (int i = 0; RsaKeyPair.RSAParamAccessor.c_ItemCount > i; i++)
            {
                int l_CurLen = l_Arr[i].Length;
                l_TotalLen += ((int)Util.Get7BitEncUInt32Size((uint)l_CurLen) + l_CurLen);
            }

            byte[] l_Bytes = new byte[l_TotalLen];
            int l_Pos = 0;

            for (int i = 0; RsaKeyPair.RSAParamAccessor.c_ItemCount > i; i++)
            {
                int l_CurLen = l_Arr[i].Length;
                l_Pos += (int)Util.Write7BitEncUInt32((uint)l_CurLen, l_Bytes, l_Pos);
                Buffer.BlockCopy(l_Arr[i], 0, l_Bytes, l_Pos, l_CurLen);
                l_Pos += l_CurLen;
            }

            return FromByteArray(l_Bytes);
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            uint l_BufLen = p_Length;
            ulong l_UInt64Pad = (0 == l_BufLen ? 0 : Util.GetRandomULong());
            a_Buffer.WriteUInt64LE(l_UInt64Pad);
            a_Buffer.WriteUInt32LE(l_BufLen);

            ScrambleMiniPadStatic(f_Buf, l_UInt64Pad);
            Unscramble();
            {
                a_Buffer.WriteBytes(f_Buf);
            }
            Scramble();
            ScrambleMiniPadStatic(f_Buf, l_UInt64Pad);
        }

        public bool ToRSAParameters(out RSAParameters a_Dest)
        {
            a_Dest = new RSAParameters();

            Unscramble();
            {
                int l_Count = f_Buf.Length;
                int l_Pos = 0;
                for (int i = 0; RsaKeyPair.RSAParamAccessor.c_ItemCount > i; i++)
                {
                    uint l_Occupied;
                    uint l_CurByteSize;
                    Util.Read7BitEncUInt32Status l_Status = Util.TryRead7BitEncUInt32(f_Buf, l_Pos, l_Count - l_Pos, out l_Occupied, out l_CurByteSize);

                    if (Util.Read7BitEncUInt32Status.Success != l_Status ||
                        ushort.MaxValue < l_CurByteSize)
                    {
                        goto FAIL_EXIT;
                    }

                    l_Pos += (int)l_Occupied;

                    if ((uint)(l_Count - l_Pos) < l_CurByteSize)
                    {
                        goto FAIL_EXIT;
                    }

                    RsaKeyPair.RSAParamAccessor.Set(Util.CopyBytes(f_Buf, l_Pos, (int)l_CurByteSize), ref a_Dest, i);
                    l_Pos += (int)l_CurByteSize;
                }

                if (l_Count == l_Pos)
                {
                    Scramble();

                    return true;
                }
            }

            FAIL_EXIT:

            Scramble();

            RsaKeyPair.RSAParamAccessor.ZeroParameters(ref a_Dest);

            return false;
        }

        public byte[] BeginExpose()
        {
            // The returned buffer may get modified. No other instance
            // members of this class must be used before EndExpose()
            // gets called. The returned buffer must not be used in
            // conjunction with the Monitor class.

            Unscramble();

            return f_Buf;
        }

        public void EndExpose()
        {
            Scramble();
        }

        private static void ScrambleStatic(byte[] a_Buf,
                                           byte[] a_Pad)
        {
            // A second call undoes a first.

            uint l_BufLen = (uint)a_Buf.Length;

            if (0 == l_BufLen)
            {
                return;
            }

            uint l_PadLen = (uint)a_Pad.Length;

            for (uint i = 0, k = 0; i != l_BufLen; i++)
            {
                a_Buf[i] ^= a_Pad[k];

                if (l_PadLen == ++k)
                {
                    k = 0;
                }
            }
        }

        private static void ScrambleMiniPadStatic(byte[] a_Buf,
                                                  ulong a_UInt64Pad)
        {
            // A second call undoes a first.

            if (0 == a_UInt64Pad)
            {
                return;
            }

            uint l_BufLen = (uint)a_Buf.Length;

            if (0 == l_BufLen)
            {
                return;
            }

            byte[] l_PadBytes = new byte[sizeof(ulong)];
            Util.WriteUInt64LE(a_UInt64Pad, l_PadBytes, 0);

            for (uint i = 0, k = 0; i != l_BufLen; i++)
            {
                a_Buf[i] ^= l_PadBytes[k];

                if (sizeof(ulong) == ++k)
                {
                    k = 0;
                }
            }
        }
    }

    public static class LongPath
    {
        // Paths with up to that many characters never cause a
        // PathTooLongException in the .NET framework.
        public const byte c_SafePathChars = 247;

        private static readonly char[] s_SepChars;

        static LongPath()
        {
            s_SepChars = new char[2] { System.IO.Path.DirectorySeparatorChar,
                                       System.IO.Path.AltDirectorySeparatorChar };
        }

        public static bool HasWindowsDevicePrefix(string a_Path)
        {
            return HasWindowsDevicePrefix(a_Path, 0, Util.GetStringLength(a_Path));
        }

        public static bool HasWindowsDevicePrefix(string a_Path,
                                                  uint a_Idx,
                                                  uint a_Len)
        {
            const byte lc_PrefixLen = 4;

            return (lc_PrefixLen <= a_Len &&
                    '\\' == a_Path[(int)a_Idx] &&
                    '\\' == a_Path[(int)a_Idx + 1] &&
                    '.' == a_Path[(int)a_Idx + 2] &&
                    '\\' == a_Path[(int)a_Idx + 3]);
        }

        public static bool HasWindowsExtendedLengthPrefix(string a_Path)
        {
            uint l_Junk;

            return HasWindowsExtendedLengthPrefix(a_Path, out l_Junk);
        }

        public static bool HasWindowsExtendedLengthPrefix(string a_Path,
                                                          out uint a_PrefixLen)
        {
            return HasWindowsExtendedLengthPrefix(a_Path, 0, Util.GetStringLength(a_Path), out a_PrefixLen);
        }

        public static bool HasWindowsExtendedLengthPrefix(string a_Path,
                                                          uint a_Idx,
                                                          uint a_Len,
                                                          out uint a_PrefixLen)
        {
            const byte lc_MinLen = 4;

            bool l_HasPrefix = (lc_MinLen <= a_Len &&
                                '\\' == a_Path[(int)a_Idx] &&
                                '\\' == a_Path[(int)a_Idx + 1] &&
                                '?' == a_Path[(int)a_Idx + 2] &&
                                '\\' == a_Path[(int)a_Idx + 3]);

            if (false == l_HasPrefix)
            {
                a_PrefixLen = 0;

                return false;
            }

            const byte lc_LenWithUNC = 8;
            bool l_WithUNC = (lc_LenWithUNC <= a_Len &&
                              'U' == a_Path[(int)a_Idx + 4] &&
                              'N' == a_Path[(int)a_Idx + 5] &&
                              'C' == a_Path[(int)a_Idx + 6] &&
                              '\\' == a_Path[(int)a_Idx + 7]);
            a_PrefixLen = (false == l_WithUNC ? lc_MinLen : lc_LenWithUNC);

            return true;
        }

        public static string RemoveAnyWindowsExtendedLengthPrefix(string a_Path)
        {
            // If no extended-length prefix is present, the input
            // string is returned  unchanged.

            uint l_PrefixLen;

            if (true == HasWindowsExtendedLengthPrefix(a_Path, out l_PrefixLen))
            {
                return a_Path.Substring((int)l_PrefixLen);
            }

            return a_Path;
        }

        public static string AddWindowsExtendedLengthPrefix(string a_Path)
        {
            // If an extended-length prefix is already present, the
            // input string is returned unchanged.

            uint l_Junk;

            if (false == HasWindowsExtendedLengthPrefix(a_Path, out l_Junk))
            {
                const string lc_WinAPILongPathPrefix = "\\\\?\\";

                return (lc_WinAPILongPathPrefix + a_Path);
            }

            return a_Path;
        }

        private static StartingKind AnalyzeStartingUnix(string a_Path,
                                                        uint a_Idx,
                                                        uint a_Len,
                                                        out uint a_RootLen)
        {
            // a_Len must not be zero.

            char l_FirstChar = a_Path[(int)a_Idx];

            if ('/' == l_FirstChar)
            {
                a_RootLen = 1;

                return StartingKind.AbsolutePath;
            }

            a_RootLen = 0;

            if (2 <= a_Len &&
                ('~' == l_FirstChar ||
                 '.' == l_FirstChar))
            {
                char l_SecondChar = a_Path[(int)a_Idx + 1];

                if ('/' == l_SecondChar)
                {
                    return StartingKind.RelativePath;
                }

                if ('.' == l_FirstChar &&
                    '.' == l_SecondChar &&
                    '/' == a_Path[(int)a_Idx + 2])
                {
                    return StartingKind.RelativePath;
                }
            }

            // Unknown path syntax.

            return StartingKind.Invalid;
        }

        private static StartingKind AnalyzeStartingWindows(string a_Path,
                                                           uint a_Idx,
                                                           uint a_Len,
                                                           out uint a_RootLen)
        {
            // a_Len must not be zero.

            char l_FirstChar = a_Path[(int)a_Idx];

            if ('\\' == l_FirstChar)
            {
                if (1 == a_Len)
                {
                    goto RELATIVE_PATH;
                }

                if ('\\' == a_Path[(int)a_Idx + 1])
                {
                    uint l_SkipLen;

                    if (false == HasWindowsExtendedLengthPrefix(a_Path, a_Idx, a_Len, out l_SkipLen))
                    {
                        l_SkipLen = 2;
                    }
                    else
                    {
                        if (4 == l_SkipLen)
                        {
                            if (7 <= a_Len &&
                                ':' == a_Path[(int)a_Idx + 5] &&
                                '\\' == a_Path[(int)a_Idx + 6] &&
                                Util.IsLatinLetter(a_Path[(int)a_Idx + 4]))
                            {
                                a_RootLen = 7;

                                return StartingKind.AbsolutePath;
                            }
                        }
                    }

                    if (4 <= (a_Len - l_SkipLen))
                    {
                        uint l_FinalIdx = (a_Idx + a_Len);
                        uint l_BeginningIdx = (a_Idx + l_SkipLen);
                        uint l_ShareNameIdx = (uint)a_Path.IndexOf('\\', (int)l_BeginningIdx, (int)(l_FinalIdx - l_BeginningIdx));

                        if (Util.c_NoIndex != l_ShareNameIdx &&
                            true == IsPotentialUNCServerName(a_Path, l_BeginningIdx, l_ShareNameIdx - l_BeginningIdx))
                        {
                            l_ShareNameIdx++;
                            uint l_LastBackslashIdx = (uint)a_Path.IndexOf('\\', (int)l_ShareNameIdx, (int)(l_FinalIdx - l_ShareNameIdx));

                            if (Util.c_NoIndex != l_LastBackslashIdx &&
                                true == IsPotentialUNCShareName(a_Path, l_ShareNameIdx, l_LastBackslashIdx - l_ShareNameIdx))
                            {
                                a_RootLen = (l_LastBackslashIdx + 1 - a_Idx);

                                return StartingKind.AbsolutePath;
                            }
                        }
                    }
                }
                else
                {
                    goto RELATIVE_PATH;
                }
            }
            else if (2 <= a_Len)
            {
                char l_SecondChar = a_Path[(int)a_Idx + 1];

                if (':' == l_SecondChar)
                {
                    if (true == Util.IsLatinLetter(l_FirstChar))
                    {
                        if (2 == a_Len)
                        {
                            goto RELATIVE_PATH;
                        }

                        char l_ThirdChar = a_Path[(int)a_Idx + 2];

                        if ('\\' == l_ThirdChar)
                        {
                            a_RootLen = 3;

                            return StartingKind.AbsolutePath;
                        }

                        if (Util.c_Space <= l_ThirdChar &&
                            '/' != l_ThirdChar &&
                            ':' != l_ThirdChar &&
                            '"' != l_ThirdChar &&
                            '<' != l_ThirdChar &&
                            '>' != l_ThirdChar &&
                            '|' != l_ThirdChar)
                        {
                            goto RELATIVE_PATH;
                        }
                    }
                }
                else if ('.' == l_FirstChar)
                {
                    if ('\\' == l_SecondChar)
                    {
                        goto RELATIVE_PATH;
                    }

                    if (3 <= a_Len &&
                        '.' == l_SecondChar &&
                        '\\' == a_Path[(int)a_Idx + 2])
                    {
                        goto RELATIVE_PATH;
                    }
                }
            }

            a_RootLen = 0;

            return StartingKind.Invalid;

        RELATIVE_PATH:

            a_RootLen = 0;

            return StartingKind.RelativePath;
        }

        private static bool IsPotentialUNCServerName(string a_Name,
                                                     uint a_Idx,
                                                     uint a_Len)
        {
            return (0 != a_Len &&
                    '.' != a_Name[(int)a_Idx]);
        }

        private static bool IsPotentialUNCShareName(string a_Name,
                                                    uint a_Idx,
                                                    uint a_Len)
        {
            return IsPotentialUNCServerName(a_Name, a_Idx, a_Len);
        }

        private static StartingKind AnalyzeStarting(string a_Path,
                                                    uint a_Idx,
                                                    uint a_Len,
                                                    out uint a_RootLen)
        {
            // a_RootLen will only be valid on return if the path
            // is absolute.

            if (0 == a_Len)
            {
                a_RootLen = 0;

                return StartingKind.Invalid;
            }

            StartingKind l_Kind;

            if (false == Util.s_IsWindows)
            {
                l_Kind = AnalyzeStartingUnix(a_Path, a_Idx, a_Len, out a_RootLen);
            }
            else
            {
                l_Kind = AnalyzeStartingWindows(a_Path, a_Idx, a_Len, out a_RootLen);
            }

            if (StartingKind.Invalid == l_Kind &&
                0 != a_Len)
            {
                l_Kind = StartingKind.RelativePath;
            }

            return l_Kind;
        }

        public static bool IsDirSeparator(char a_Char)
        {
            return (Path.DirectorySeparatorChar == a_Char ||
                    Path.AltDirectorySeparatorChar == a_Char);
        }

        public static string Combine(string a_StartingPath,
                                     string a_EndingPath)
        {
            // If one of the paths is an empty string, the other path
            // is returned. a_EndingPath is not supposed to be rooted.

            int l_StartLen = a_StartingPath.Length;

            if (0 == l_StartLen)
            {
                return a_EndingPath;
            }

            int l_EndingLen = a_EndingPath.Length;

            if (0 == l_EndingLen)
            {
                return a_StartingPath;
            }

            StringBuilder l_Bld = new StringBuilder(a_StartingPath, l_StartLen + 1 + l_EndingLen);

            char l_LastChar = a_StartingPath[l_StartLen - 1];
            char l_FirstChar = a_EndingPath[l_EndingLen - 1];
            bool l_SkipFirstChar = false;

            if (false == IsDirSeparator(l_LastChar))
            {
                if (Path.DirectorySeparatorChar != l_FirstChar)
                {
                    if (Path.AltDirectorySeparatorChar == l_FirstChar)
                    {
                        l_SkipFirstChar = true;
                    }

                    l_Bld.Append(Path.DirectorySeparatorChar);
                }
            }
            else if (true == IsDirSeparator(l_FirstChar))
            {
                l_SkipFirstChar = true;
            }

            l_Bld.Append(a_EndingPath, Util.BoolToByte(l_SkipFirstChar), l_EndingLen - Util.BoolToByte(l_SkipFirstChar));

            return l_Bld.ToString();
        }

        public static bool IsPathRooted(string a_Path,
                                        out bool a_IsValid)
        {
            // Returns zero if the path is invalid or not rooted. Otherwise
            // it's valid and rooted.

            uint l_RootLen;
            StartingKind l_Kind = AnalyzeStarting(a_Path, 0, Util.GetStringLength(a_Path), out l_RootLen);

            if (StartingKind.AbsolutePath == l_Kind)
            {
                a_IsValid = true;

                return true;
            }

            a_IsValid = (StartingKind.RelativePath == l_Kind);

            return false;
        }

        public static string GetDirectoryName(string a_RawPath)
        {
            // Return can be null on illegal input.

            string l_DirName;
            string l_Junk = SplitFullPath(a_RawPath, out l_DirName);

            return l_DirName;
        }

        public static string SplitFullPath(string a_Path,
                                           out string a_DirName)
        {
            // This routine has no path length restriction like
            // Path.GetDirectoryName(). Its return is similar to what
            // Path.GetFileName() would return. a_DirName will be null
            // on return if a directory name couldn't be found.

            int l_LastSep = a_Path.LastIndexOfAny(s_SepChars);

            if (0 > l_LastSep)
            {
                a_DirName = null;

                return a_Path;
            }

            uint l_RootLen;

            if (StartingKind.AbsolutePath == AnalyzeStarting(a_Path, 0, Util.GetStringLength(a_Path), out l_RootLen))
            {
                if (Util.GetStringLength(a_Path) == l_RootLen)
                {
                    // Input is just as long as the root.

                    a_DirName = null;

                    return string.Empty;
                }
            }
            else
            {
                l_RootLen = 0;
            }

            if ((int)l_RootLen > l_LastSep)
            {
                a_DirName = a_Path.Substring(0, (int)l_RootLen);
                l_LastSep = (int)l_RootLen;
            }
            else
            {
                a_DirName = a_Path.Substring(0, l_LastSep);
                l_LastSep++;
            }

            return a_Path.Substring(l_LastSep);
        }

        public static bool IsRoot(string a_Path)
        {
            uint l_RootLen;

            return (StartingKind.AbsolutePath == AnalyzeStarting(a_Path, 0, Util.GetStringLength(a_Path), out l_RootLen) &&
                    Util.GetStringLength(a_Path) == l_RootLen);
        }

        public static string GetPathRoot(string a_Path)
        {
            // This routine has no path length restriction like
            // Path.GetPathRoot() (which has one, but it's not
            // documented). Return can be null.

            uint l_RootLen;

            if (StartingKind.AbsolutePath == AnalyzeStarting(a_Path, 0, Util.GetStringLength(a_Path), out l_RootLen))
            {
                return a_Path.Substring(0, (int)l_RootLen);
            }

            return null;
        }

        public static string GetExtension(string a_FileName)
        {
            // Returns the extension without the period, or null if
            // there is no extension.

            int l_Idx = a_FileName.LastIndexOf('.');

            if (0 > l_Idx ||
                ++l_Idx == a_FileName.Length)
            {
                return null;
            }

            return a_FileName.Substring(l_Idx);
        }

        private enum StartingKind : byte
        {
            Invalid = 0,

            RelativePath = 1,
            AbsolutePath = 2,
        }
    }

    public class RsaKeyPair
    {
        // We don't necessarily use SHA-1, but we have to pretent we
        // do to make the RSACryptoServiceProvider class work.
#if MONO
        // Mono expects an OID, not a name.
        public const string c_AllegedHashAlgoName = "1.3.14.3.2.26";
#else
        public const string c_AllegedHashAlgoName = "SHA1";
#endif

        internal const ushort c_SignatureWithSaltHeaderMagic = 0x2B87;

        private readonly SensitiveBuffer f_Keys;
        private readonly RSACryptoServiceProvider f_RSA;
        public readonly RsaPublicKey f_PublicKey;

        private RsaKeyPair(SensitiveBuffer a_Keys,
                           RSACryptoServiceProvider a_RSA,
                           RsaPublicKey a_PublicKey)
        {
            f_Keys = a_Keys;
            f_RSA = a_RSA;
            f_PublicKey = a_PublicKey;
        }

        public static RsaKeyPair GenerateKey(uint a_BitSize)
        {
            // NOTE: The RSACryptoServiceProvider class will generate
            // the key as late as possible (when it is first used).
            // Since we export it right after creating the instance,
            // this is not a problem.

            RSACryptoServiceProvider l_RSA = new RSACryptoServiceProvider((int)a_BitSize);
            RSAParameters l_Params = l_RSA.ExportParameters(true);
            SensitiveBuffer l_Keys = SensitiveBuffer.FromRSAParameters(ref l_Params);
            RsaPublicKey l_PubKey;
            RsaPublicKey.FromRsaParams(true, l_Params.Modulus, l_Params.Exponent, out l_PubKey);

            // Overwrite sensitive stack.
            l_Params = new RSAParameters();

            return new RsaKeyPair(l_Keys, l_RSA, l_PubKey);
        }

        public static RsaKeyPair FromBuffer(BufferIO a_Buffer)
        {
            // Return can be null on error.

            SensitiveBuffer l_Keys = SensitiveBuffer.FromBuffer(a_Buffer);
            RSAParameters l_Params;

            if (false == l_Keys.ToRSAParameters(out l_Params))
            {
                return null;
            }

            // ZeroParams() must be called before we leave.

            RsaKeyPair l_KeyPair = null;

            {
                RSACryptoServiceProvider l_Rsa = new RSACryptoServiceProvider();

                try
                {
                    l_Rsa.ImportParameters(l_Params);
                }
                catch (CryptographicException)
                {
                    l_Rsa = null;
                }

                if (null != l_Rsa &&
                    false == l_Rsa.PublicOnly)
                {
                    RsaPublicKey l_PubKey;
                    RsaPublicKey.FromRsaParams(true, l_Params.Modulus, l_Params.Exponent, out l_PubKey);
                    l_KeyPair = new RsaKeyPair(l_Keys, l_Rsa, l_PubKey);
                }
            }

            RSAParamAccessor.ZeroParameters(ref l_Params);

            return l_KeyPair;
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            f_Keys.ToBuffer(a_Buffer);
        }

        public bool ValueEquals(RsaKeyPair a_Other)
        {
            return this.f_PublicKey.ValueEquals(a_Other.f_PublicKey);
        }

        public void Release()
        {
            f_Keys.Release();
        }

        public byte[] CreateSignature(byte[] a_Hash,
                                      bool a_WithSalt,
                                      IHashFunction a_SaltHashAlgo)
        {
            uint l_SignatureLen = f_PublicKey.GetExpectedSignatureLength();

            if (false == a_WithSalt)
            {
                return f_RSA.SignHash(a_Hash, c_AllegedHashAlgoName);
            }

            const byte lc_SaltIdx = (2 * sizeof(ushort));
            const byte lc_SaltLen = RsaPublicKey.c_SignableHashByteLen;
            const byte lc_SigIdx = (lc_SaltIdx + lc_SaltLen);

            // Create the salt and copy it to the newly created output
            // buffer.

            byte[] l_SigBuf = new byte[lc_SigIdx + l_SignatureLen];
            byte[] l_SaltBuf = Util.GetRandomBytes(lc_SaltLen);
            Buffer.BlockCopy(l_SaltBuf, 0, l_SigBuf, lc_SaltIdx, lc_SaltLen);

            // Hash the input hash with salt.

            HMACWrapper l_HMAC = new HMACWrapper(a_SaltHashAlgo, l_SaltBuf);
            l_HMAC.Update(a_Hash);

            uint l_RawHashLen = a_SaltHashAlgo.p_HashByteSize;
            byte[] l_SaltedHashBuf = new byte[l_RawHashLen];
            l_HMAC.GetDigest(l_SaltedHashBuf, 0);

            if (lc_SaltLen < l_RawHashLen)
            {
                Array.Resize(ref l_SaltedHashBuf, lc_SaltLen);
            }

            // Sign the salted hash and write the signature to the
            // output buffer.

            byte[] l_SigPortionBuf = f_RSA.SignHash(l_SaltedHashBuf, c_AllegedHashAlgoName);
            Buffer.BlockCopy(l_SigPortionBuf, 0, l_SigBuf, lc_SigIdx, (int)l_SignatureLen);

            // Write the header to the output buffer.

            Util.WriteUInt16LE(c_SignatureWithSaltHeaderMagic, l_SigBuf, 0);
            Util.WriteUInt16LE((ushort)l_SignatureLen, l_SigBuf, sizeof(ushort));

            return l_SigBuf;
        }

        public static class RSAParamAccessor
        {
            public const int c_ItemCount = 8;

            public static void Set(byte[] a_Src,
                                   ref RSAParameters a_Dest,
                                   int a_DestIndex)
            {
                // a_Src must contain BE data.

                switch (a_DestIndex)
                {
                    case 0:
                        a_Dest.D = a_Src;
                        break;

                    case 1:
                        a_Dest.DP = a_Src;
                        break;

                    case 2:
                        a_Dest.DQ = a_Src;
                        break;

                    case 3:
                        a_Dest.Exponent = a_Src;
                        break;

                    case 4:
                        a_Dest.InverseQ = a_Src;
                        break;

                    case 5:
                        a_Dest.Modulus = a_Src;
                        break;

                    case 6:
                        a_Dest.P = a_Src;
                        break;

                    case 7:
                        a_Dest.Q = a_Src;
                        break;

                    default: throw (new ApplicationException());
                }
            }

            public static byte[][] Get(ref RSAParameters a_Src)
            {
                // All returned arrays will contain BE data.

                byte[][] l_Arr = new byte[c_ItemCount][];

                l_Arr[0] = a_Src.D;
                l_Arr[1] = a_Src.DP;
                l_Arr[2] = a_Src.DQ;
                l_Arr[3] = a_Src.Exponent;
                l_Arr[4] = a_Src.InverseQ;
                l_Arr[5] = a_Src.Modulus;
                l_Arr[6] = a_Src.P;
                l_Arr[7] = a_Src.Q;

                return l_Arr;
            }

            public static void ZeroParameters(ref RSAParameters a_Params)
            {
                // This routine zeros all private key arrays.

                ZeroByteArray(a_Params.D);
                ZeroByteArray(a_Params.DP);
                ZeroByteArray(a_Params.DQ);
                ZeroByteArray(a_Params.InverseQ);
                ZeroByteArray(a_Params.P);
                ZeroByteArray(a_Params.Q);

                a_Params = new RSAParameters();
            }

            private static void ZeroByteArray(byte[] a_Bytes)
            {
                if (null != a_Bytes)
                {
                    Buffer2.BlockZero(a_Bytes);
                }
            }
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct RsaPublicKey
    {
        // NOTE: The following document says that the RSAParameters
        // struct is in big endian notation (the .NET docs don't
        // mention this).
        // <http://msdn.microsoft.com/en-us/library/ms867080.aspx>

        private const byte c_PortableRsaPublicKeyFormatMinSize = ((3 * sizeof(ushort)) + 2);
        private const ushort c_PortableRsaPublicKeyFormatMagic = 0x1E4C;

        public const byte c_SignableHashByteLen = 20;

        private static SipHash.Key s_SipHashKey;

        private readonly byte[] f_BinaryPubKey;

        static RsaPublicKey()
        {
            SipHash.Key.CreateRandom(out s_SipHashKey);
        }

        private RsaPublicKey(byte[] a_BinaryPubKey)
        {
            f_BinaryPubKey = a_BinaryPubKey;
        }

        public bool IsValid()
        {
            // This routine doesn't check if the key is really usable,
            // instead it checks if there is a potential key at all.

            return (null != f_BinaryPubKey);
        }

        public bool IsValid(RSACryptoServiceProvider a_RSA)
        {
            RSAParameters l_Params;
            ToDotNetRsaParams(out l_Params);

            try
            {
                a_RSA.ImportParameters(l_Params);
            }
            catch (CryptographicException)
            {
                return false;
            }

            return true;
        }

        public override bool Equals(object obj)
        {
            // Warn us if someone operates non-generic.
            throw (new ApplicationException());
        }

        public bool ValueEquals(RsaPublicKey a_Other)
        {
            return Buffer2.ValueEquals(this.f_BinaryPubKey, a_Other.f_BinaryPubKey);
        }

        public override int GetHashCode()
        {
            return (int)CalcHashCode();
        }

        public uint CalcHashCode()
        {
            return (uint)SipHash.CalculateHashCode(ref s_SipHashKey, f_BinaryPubKey, 0, f_BinaryPubKey.Length);
        }

        private uint ReadModulusByteSize()
        {
            return Util.ReadUInt16LE(f_BinaryPubKey, sizeof(ushort));
        }

        public uint GetBitSize()
        {
            return (ReadModulusByteSize() * Util.c_BitsPerByte);
        }

        public uint GetExpectedSignatureLength()
        {
            return ReadModulusByteSize();
        }

        public void GetModAndExp(bool a_BE,
                                 out byte[] a_Modulus,
                                 out byte[] a_Exponent)
        {
            // If a_BE is false, the output will be in LE.

            ByteArray l_Data = new ByteArray(f_BinaryPubKey);
            int l_ModLen;
            int l_ExpLen;
            bool l_Succ = ParsePortableRsaPublicKeyFormat(ref l_Data, out l_ModLen, out l_ExpLen);

            a_Modulus = Util.CopyBytes(f_BinaryPubKey, 2 * sizeof(ushort), l_ModLen);
            a_Exponent = Util.CopyBytes(f_BinaryPubKey, (3 * sizeof(ushort)) + l_ModLen, l_ExpLen);

            if (true == a_BE)
            {
                Util.ReverseArray<byte>(a_Modulus);
                Util.ReverseArray<byte>(a_Exponent);
            }
        }

        public string AsString(string a_Format)
        {
            // Outputs a string with values in BE.

            byte[] l_Modulus;
            byte[] l_Exponent;
            GetModAndExp(true, out l_Modulus, out l_Exponent);

            return string.Format(Util.s_InvariantCulture, a_Format, GetBitSize(), Util.ToShortString(Util.BytesToString(l_Modulus), 12), Util.ToShortString(Util.BytesToString(l_Exponent), 12));
        }

        public void ToDotNetRsaParams(out RSAParameters a_Params)
        {
            a_Params = new RSAParameters();
            GetModAndExp(true, out a_Params.Modulus, out a_Params.Exponent);
        }

        public byte[] ToiMuleCompatibleFormat()
        {
            // This routine will return null for large keys.

            byte[] l_Modulus;
            byte[] l_Exponent;
            GetModAndExp(true, out l_Modulus, out l_Exponent);

            return ASN1.ToiMuleReadablePubKey(l_Modulus, l_Exponent);
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      out RsaPublicKey a_Dest)
        {
            uint l_Len = a_Buffer.ReadUInt16LE();
            ByteArray l_Data = new ByteArray(a_Buffer.ReadBytes(l_Len));

            if (false == ValidatePortableRsaPublicKeyFormat(ref l_Data))
            {
                a_Dest = new RsaPublicKey();

                return false;
            }

            a_Dest = new RsaPublicKey(l_Data.f_Arr);

            return true;
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            a_Buffer.WriteUInt16LE((ushort)f_BinaryPubKey.Length);
            a_Buffer.WriteBytes(f_BinaryPubKey);
        }

        public void ToNetworkBuffer(BufferIO a_Buffer,
                                    uint a_SecIdentVer,
                                    bool a_PrependLength)
        {
            if (2 >= a_SecIdentVer)
            {
                byte[] l_PubKey = ToiMuleCompatibleFormat();

                if (true == a_PrependLength)
                {
                    a_Buffer.WriteByte((byte)l_PubKey.Length);
                }

                a_Buffer.WriteBytes(l_PubKey);
            }
            else
            {
                if (true == a_PrependLength)
                {
                    a_Buffer.WriteUInt16LE((ushort)f_BinaryPubKey.Length);
                }

                a_Buffer.WriteBytes(f_BinaryPubKey);
            }
        }

        public bool VerifySignature(ref SignatureVerificationContext a_VerCntxt)
        {
            a_VerCntxt.f_KeyStrength = 0;

            byte[] l_EffHash;
            byte[] l_EffSig;
            uint l_ExpectedEffSigLen = GetExpectedSignatureLength();

            if (false == a_VerCntxt.f_ExpectSalt)
            {
                if (a_VerCntxt.f_Signature.Length != l_ExpectedEffSigLen)
                {
                    return false;
                }

                l_EffHash = a_VerCntxt.f_Hash;
                l_EffSig = a_VerCntxt.f_Signature;
            }
            else
            {
                const byte lc_HeaderWithSaltLen = (sizeof(uint) + c_SignableHashByteLen);

                if (lc_HeaderWithSaltLen >= a_VerCntxt.f_Signature.Length ||
                    (a_VerCntxt.f_Signature.Length - lc_HeaderWithSaltLen) != l_ExpectedEffSigLen)
                {
                    return false;
                }

                if (RsaKeyPair.c_SignatureWithSaltHeaderMagic != Util.ReadUInt16LE(a_VerCntxt.f_Signature, 0) ||
                    Util.ReadUInt16LE(a_VerCntxt.f_Signature, sizeof(ushort)) != l_ExpectedEffSigLen)
                {
                    return false;
                }

                byte[] l_Salt = new byte[c_SignableHashByteLen];
                Buffer.BlockCopy(a_VerCntxt.f_Signature, sizeof(uint), l_Salt, 0, c_SignableHashByteLen);
                HMACWrapper l_HMAC = new HMACWrapper(a_VerCntxt.f_SaltHashAlgo, l_Salt);
                l_HMAC.Update(a_VerCntxt.f_Hash);

                uint l_RawHashLen = a_VerCntxt.f_SaltHashAlgo.p_HashByteSize;
                l_EffHash = new byte[l_RawHashLen];
                l_HMAC.GetDigest(l_EffHash, 0);

                if (c_SignableHashByteLen != l_RawHashLen)
                {
                    Array.Resize(ref l_EffHash, c_SignableHashByteLen);
                }

                l_EffSig = new byte[l_ExpectedEffSigLen];
                Buffer.BlockCopy(a_VerCntxt.f_Signature, lc_HeaderWithSaltLen, l_EffSig, 0, (int)l_ExpectedEffSigLen);
            }

            RSAParameters l_Params;
            ToDotNetRsaParams(out l_Params);
            a_VerCntxt.f_Rsa.ImportParameters(l_Params);

            if (true == a_VerCntxt.f_Rsa.VerifyHash(l_EffHash, RsaKeyPair.c_AllegedHashAlgoName, l_EffSig))
            {
                a_VerCntxt.f_KeyStrength = GetBitSize();

                return true;
            }

            return false;
        }

        // Since there are so many incompatible RSA public key formats
        // (like ASN.1 and PKCS #8) out there that everyone seems to be
        // incompatible with everyone (the RSACryptoServiceProvider.ExportCspBlob()
        // of .NET 2 is incompatible with Crypto++ 6.5.1), I decided to
        // specify my own compact, easy to understand/use format. The
        // PORTABLE RSA PUBLIC KEY FORMAT (it can be used with up to 524,280
        // (that is 0xFFFF * 8) bit keys).

        // MAGIC <USHORT>, always 0x1E4C.
        // MOD LENGTH <USHORT>, the byte length of the modulus (little endian).
        // MOD <BYTE ARRAY>, the modulus (little endian).
        // EXP LENGTH <USHORT>, the byte length of the exponent (little endian).
        // MOD <BYTE ARRAY>, the exponent (little endian).

        public byte[] ExposePortableRsaPublicKeyFormat()
        {
            return f_BinaryPubKey;
        }

        public static bool FromRsaParams(bool a_BE,
                                         byte[] a_Modulus,
                                         byte[] a_Exponent,
                                         out RsaPublicKey a_RsaPubKey)
        {
            int l_ModLen = a_Modulus.Length;
            int l_ExpLen = a_Exponent.Length;

            if (1 > l_ModLen ||
                ushort.MaxValue < l_ModLen ||
                1 > l_ExpLen ||
                ushort.MaxValue < l_ExpLen)
            {
                a_RsaPubKey = new RsaPublicKey();

                return false;
            }

            byte[] l_Key = new byte[(3 * sizeof(ushort)) + l_ModLen + l_ExpLen];

            Util.WriteUInt16LE(c_PortableRsaPublicKeyFormatMagic, l_Key, 0);
            Util.WriteUInt16LE((ushort)l_ModLen, l_Key, sizeof(ushort));
            Util.WriteUInt16LE((ushort)l_ExpLen, l_Key, (2 * sizeof(ushort)) + l_ModLen);

            int l_ModIdx = (2 * sizeof(ushort));
            int l_ExpIdx = ((3 * sizeof(ushort)) + l_ModLen);
            Buffer.BlockCopy(a_Modulus, 0, l_Key, l_ModIdx, l_ModLen);
            Buffer.BlockCopy(a_Exponent, 0, l_Key, l_ExpIdx, l_ExpLen);

            if (true == a_BE)
            {
                // BE to LE.
                Util.ReverseArray(l_Key, l_ModIdx, l_ModLen);
                Util.ReverseArray(l_Key, l_ExpIdx, l_ExpLen);
            }

            a_RsaPubKey = new RsaPublicKey(l_Key);

            return true;
        }

        public static bool ParsePortableRsaPublicKeyFormat(ref ByteArray a_Data,
                                                           out int a_ModLen,
                                                           out int a_ExpLen)
        {
            // Returns true if the format is valid.

            if (c_PortableRsaPublicKeyFormatMinSize > a_Data.f_Len ||
                c_PortableRsaPublicKeyFormatMagic != Util.ReadUInt16LE(a_Data.f_Arr, a_Data.f_Idx))
            {
                a_ModLen = 0;
                a_ExpLen = 0;

                return false;
            }

            a_ModLen = Util.ReadUInt16LE(a_Data.f_Arr, a_Data.f_Idx + sizeof(ushort));
            a_ExpLen = Util.ReadUInt16LE(a_Data.f_Arr, a_Data.f_Idx + (2 * sizeof(ushort)) + a_ModLen);

            return (((3 * sizeof(ushort)) + a_ModLen + a_ExpLen) == a_Data.f_Len);
        }

        public static bool ValidatePortableRsaPublicKeyFormat(ref ByteArray a_Data)
        {
            // This routine validates just the key format, not the key
            // itself.

            int l_Junk1;
            int l_Junk2;
            return ParsePortableRsaPublicKeyFormat(ref a_Data, out l_Junk1, out l_Junk2);
        }

        public static bool FromPortableRsaPublicKeyFormat(RSACryptoServiceProvider a_RSA,
                                                          ref ByteArray a_Data,
                                                          out RsaPublicKey a_RsaPubKey)
        {
            if (true == ValidatePortableRsaPublicKeyFormat(ref a_Data))
            {
                byte[] l_TightArr = a_Data.ToArray(false);
                a_RsaPubKey = new RsaPublicKey(l_TightArr);

                return a_RsaPubKey.IsValid(a_RSA);
            }

            a_RsaPubKey = new RsaPublicKey();

            return false;
        }

        public static bool FromUnknownFormat(RSACryptoServiceProvider a_RSA,
                                             ref ByteArray a_DataInAnyFormat,
                                             out RsaPublicKey a_RsaPubKey)
        {
            if (0 >= a_DataInAnyFormat.f_Len)
            {
                a_RsaPubKey = new RsaPublicKey();

                return false;
            }

            if (ASN1.c_SupposedFirstASN1Byte == a_DataInAnyFormat.f_Arr[a_DataInAnyFormat.f_Idx])
            {
                // Try to intepret it as PKCS #1 data.

                byte[] l_ModBE;
                byte[] l_ExpBE;
                if (true == ASN1.ToPubKeyParams(ref a_DataInAnyFormat, out l_ModBE, out l_ExpBE))
                {
                    if (false == FromRsaParams(true, l_ModBE, l_ExpBE, out a_RsaPubKey))
                    {
                        a_RsaPubKey = new RsaPublicKey();

                        return false;
                    }

                    return a_RsaPubKey.IsValid(a_RSA);
                }
            }
            else if (true == FromPortableRsaPublicKeyFormat(a_RSA, ref a_DataInAnyFormat, out a_RsaPubKey))
            {
                return true;
            }

            // Maybe .NET can read it.

            try
            {
                byte[] l_TightArr = a_DataInAnyFormat.ToArray(false);
                a_RSA.ImportCspBlob(l_TightArr);
            }
            catch (CryptographicException)
            {
                a_RsaPubKey = new RsaPublicKey();

                return false;
            }

            RSAParameters l_Params;
            l_Params = a_RSA.ExportParameters(false);

            return FromRsaParams(true, l_Params.Modulus, l_Params.Exponent, out a_RsaPubKey);
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct SignatureVerificationContext
        {
            public readonly RSACryptoServiceProvider f_Rsa;
            public readonly byte[] f_Hash;
            public readonly byte[] f_Signature;
            public readonly bool f_ExpectSalt;
            public readonly IHashFunction f_SaltHashAlgo;

            // f_KeyStrength will be set only if the signature was valid.
            public uint f_KeyStrength;

            public SignatureVerificationContext(RSACryptoServiceProvider a_Rsa,
                                                byte[] a_Hash,
                                                byte[] a_Signature,
                                                bool a_ExpectSalt,
                                                IHashFunction a_SaltHashAlgo)
            {
                f_Rsa = a_Rsa;
                f_Hash = a_Hash;
                f_Signature = a_Signature;
                f_ExpectSalt = a_ExpectSalt;
                f_SaltHashAlgo = a_SaltHashAlgo;
                f_KeyStrength = 0;
            }
        }
    }

    public static class WinShell
    {
        private const byte c_S_OK = 0;

        [DllImport("shell32.dll", CallingConvention = CallingConvention.Winapi, CharSet = CharSet.Unicode)]
        private static extern int SHOpenFolderAndSelectItems([In] IntPtr pidlFolder,
                                                             [In] uint cidl,
                                                             [In] IntPtr apidl,
                                                             [In] uint dwFlags);

        [DllImport("shell32.dll", CallingConvention = CallingConvention.Winapi, CharSet = CharSet.Unicode)]
        private static extern int SHParseDisplayName([In] string pszName,
                                                     [In] IntPtr pbc,
                                                     [Out] out IntPtr ppidl,
                                                     [In] uint sfgaoIn,
                                                     [Out] out uint psfgaoOut);

        public static bool OpenFolderAndSelectItem(string a_Name)
        {
            if (false == File.Exists(a_Name) &&
                false == Directory.Exists(a_Name))
            {
                return false;
            }

            IntPtr l_Pidl = IntPtr.Zero;
            uint l_Junk;

            try
            {
                if (c_S_OK != SHParseDisplayName(a_Name, IntPtr.Zero, out l_Pidl, 0, out l_Junk))
                {
                    return false;
                }
            }
            catch (BadImageFormatException)
            {
            }
            catch (TypeLoadException)
            {
            }

            bool l_Succ = false;

            try
            {
                if (c_S_OK == SHOpenFolderAndSelectItems(l_Pidl, 0, IntPtr.Zero, 0))
                {
                    l_Succ = true;
                }
            }
            catch (BadImageFormatException)
            {
            }
            catch (TypeLoadException)
            {
            }

            Marshal.FreeCoTaskMem(l_Pidl);

            return l_Succ;
        }
    }
}

namespace Support.TaskBar
{
    public sealed class TaskBarProgress : IDisposable
    {
        private const int c_S_OK = 0;

        private readonly ITaskbarList3 f_TaskbarList;
        private readonly IntPtr f_TargetWindow;
        private bool f_Disposed;

        private bool f_Enabled;
        private TaskBarProgressContext f_Context;
        // The last values submitted to the OS.
        private ulong f_LastCompleted;
        private ulong f_LastTotal;

        private TaskBarProgress(ITaskbarList3 a_TaskbarList,
                                IntPtr a_TargetWindow,
                                bool a_InitiallyEnabled)
        {
            f_TaskbarList = a_TaskbarList;
            f_TargetWindow = a_TargetWindow;
            f_Enabled = a_InitiallyEnabled;
        }

        public static TaskBarProgress TryCreate(IntPtr a_TargetWindowHandle,
                                                bool a_InitiallyEnabled)
        {
            // Return can be null.

            ITaskbarList3 l_TaskbarList;

            try
            {
                // This will fail on Windows XP and lower.
                l_TaskbarList = (ITaskbarList3)new CTaskbarList();

                if (c_S_OK != l_TaskbarList.HrInit())
                {
                    int l_RefCount = Marshal.ReleaseComObject(l_TaskbarList);

                    return null;
                }
            }
            catch
            {
                return null;
            }

            return new TaskBarProgress(l_TaskbarList, a_TargetWindowHandle, a_InitiallyEnabled);
        }

        public void Dispose()
        {
            if (false == f_Disposed)
            {
                f_Disposed = true;

                int l_RefCount = Marshal.ReleaseComObject(f_TaskbarList);
            }
        }

        public void AdjustEnabledFlag(bool a_Enabled)
        {
            if (a_Enabled == f_Enabled)
            {
                return;
            }

            f_Enabled = a_Enabled;

            if (true == f_Context.IsVisible())
            {
                if (false == a_Enabled)
                {
                    // Turning off.

                    ComSetProgressKind(TaskBarProgressKinds.None);
                }
                else
                {
                    // Turning on.

                    SetProgressCore(ref f_Context, true);
                }
            }
        }

        public void SetNoProgress()
        {
            SetKindWithLastValues(TaskBarProgressKinds.None);
        }

        public void SetMarqueeEffect()
        {
            SetKindWithLastValues(TaskBarProgressKinds.GreenMarquee);
        }

        private void SetKindWithLastValues(TaskBarProgressKinds a_Kind)
        {
            if (f_Context.f_Kind != a_Kind)
            {
                TaskBarProgressContext l_Cntxt = new TaskBarProgressContext(a_Kind, f_LastCompleted, f_LastTotal);
                SetProgress(ref l_Cntxt);
            }
        }

        public void SetFilled(TaskBarProgressKinds a_Kind)
        {
            if (f_Context.f_Kind != a_Kind ||
                f_LastCompleted != f_LastTotal)
            {
                TaskBarProgressContext l_Cntxt = new TaskBarProgressContext(a_Kind);
                SetProgress(ref l_Cntxt);
            }
        }

        public void SetProgress(ref TaskBarProgressContext a_Cntxt)
        {
            if (true == f_Enabled)
            {
                bool l_KindChanges = (a_Cntxt.f_Kind != f_Context.f_Kind);
                SetProgressCore(ref a_Cntxt, l_KindChanges);
            }

            f_Context = a_Cntxt;
        }

        private void SetProgressCore(ref TaskBarProgressContext a_Cntxt,
                                     bool a_UpdateKind)
        {
            if (true == a_UpdateKind)
            {
                ComSetProgressKind(a_Cntxt.f_Kind);
            }

            if (true == a_Cntxt.NeedsValues() &&
                (true == a_UpdateKind ||
                 a_Cntxt.f_Completed != f_LastCompleted ||
                 a_Cntxt.f_Total != f_LastTotal))
            {
                // We only need to do this for certain kinds.

                ComSetProgressValue(ref a_Cntxt);

                f_LastCompleted = a_Cntxt.f_Completed;
                f_LastTotal = a_Cntxt.f_Total;
            }
        }

        private bool ComSetProgressKind(TaskBarProgressKinds a_Kind)
        {
            return (c_S_OK == f_TaskbarList.SetProgressState(f_TargetWindow, (uint)a_Kind));
        }

        private bool ComSetProgressValue(ref TaskBarProgressContext a_Cntxt)
        {
            return (c_S_OK == f_TaskbarList.SetProgressValue(f_TargetWindow, a_Cntxt.f_Completed, a_Cntxt.f_Total));
        }

        [ComImport()]
        [Guid("EA1AFB91-9E28-4B86-90E9-9E9F8A5EEFAF")]
        [InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
        private interface ITaskbarList3
        {
            // Keep the order of declarations, otherwise the interface
            // can get useless. Really.

            // ITaskbarList

            [PreserveSig]
            int HrInit();

            [PreserveSig]
            int AddTab([In] IntPtr hwnd);

            [PreserveSig]
            int DeleteTab([In] IntPtr hwnd);

            [PreserveSig]
            int ActivateTab([In] IntPtr hwnd);

            [PreserveSig]
            int SetActiveAlt([In] IntPtr hwnd);

            // ITaskbarList2

            [PreserveSig]
            int MarkFullscreenWindow([In] IntPtr hwnd,
                                     [In] int fFullscreen);

            // ITaskbarList3

            [PreserveSig]
            int SetProgressValue([In] IntPtr hwnd,
                                 [In] ulong ullCompleted,
                                 [In] ulong ullTotal);

            [PreserveSig]
            int SetProgressState([In] IntPtr hwnd,
                                 [In] uint tbpFlags);
        }

        [ComImport()]
        [Guid("56FDF344-FD6D-11D0-958A-006097C9A090")]
        [ClassInterface(ClassInterfaceType.None)]
        private class CTaskbarList
        {
        }
    }

    public enum TaskBarProgressKinds : byte
    {
        // NOPROGRESS
        None = 0,

        // INDETERMINATE
        GreenMarquee = 0x1,
        // NORMAL
        Green = 0x2,
        // ERROR
        Red = 0x4,
        // PAUSED
        Yellow = 0x8,
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    public struct TaskBarProgressContext
    {
        // The default constructor may be used.

        public readonly TaskBarProgressKinds f_Kind;
        public readonly ulong f_Completed;
        public readonly ulong f_Total;

        public TaskBarProgressContext(TaskBarProgressKinds a_Kind,
                                      ulong a_Completed,
                                      ulong a_Total)
        {
            f_Kind = a_Kind;
            f_Completed = a_Completed;
            f_Total = a_Total;
        }

        public TaskBarProgressContext(TaskBarProgressKinds a_Kind) :
                                 this(a_Kind, 0, 0)
        {
        }

        public bool IsVisible()
        {
            return (TaskBarProgressKinds.None != f_Kind);
        }

        public bool NeedsValues()
        {
            // Returns true if the kind supports progress values.

            return (TaskBarProgressKinds.Green <= f_Kind);
        }
    }
}

// End of file.
