/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\GUI.cs
 * PURPOSE: Graphical user interface
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using NB.Content;
using NB.I2P;
using NB.Import;
using NB.Language;
using NB.Network;
using NB.Network.eD2k;
using Support;
using Support.HashFunctions;

namespace NB.Gui
{
    class SamBridgeTestDialog : ButtonDialog
    {
        private readonly Instance f_Inst;

        public readonly GuiGroupBox f_GroupBox;
        public readonly GuiLabel f_InfoLabel;
        public readonly LabeledNumberBox f_PortBox;
        public readonly GuiLabel f_ResultTitleLabel;
        public readonly GuiLabel f_ResultLabel;

        private readonly Font f_BoldFont;

        private SamBridgeManager.SamTestContext f_Cntxt;

        public SamBridgeTestDialog(Instance a_Inst) :
                              base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            const int l_SuggestedWidth = 350;
            const ushort l_DefaultSamPort = NB.I2P.Internal.SamControlClient.c_DefaultSamPort;

            f_Inst = a_Inst;

            f_GroupBox = new GuiGroupBox(this);
            f_InfoLabel = new GuiLabel(f_GroupBox);
            int l_DlgWidth = Lyt.CalcX(l_SuggestedWidth);
            int l_Width = (l_DlgWidth - Lyt.s_Space2);
            int l_InnerWidth = (l_Width - f_GroupBox.Padding.Horizontal);
            f_PortBox = new LabeledNumberBox(f_GroupBox, l_InnerWidth);
            f_ResultTitleLabel = new GuiLabel(f_GroupBox);
            f_ResultLabel = new GuiLabel(f_GroupBox);

            f_BoldFont = MainForm.RecreateFont(f_ResultLabel.Font, 0, FontStyle.Bold);

            TextSet l_TextSet = a_Inst.f_TextSet;
            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            int l_Left = f_GroupBox.Padding.Left;
            f_InfoLabel.Location = new Point(l_Left, f_GroupBox.Padding.Top);
            f_InfoLabel.Text = l_TextSet.Format(TextKeys.SamConnTestDlgInfoText, l_TextSet.f_AppName, Software.s_ReleaseInfo.BuildReleaseNotesFileName(), System.Net.IPAddress.Loopback, l_DefaultSamPort);
            f_InfoLabel.Width = l_InnerWidth;
            Lyt.SetFontAndHeight(f_InfoLabel, f_InfoLabel.Font);
            f_PortBox.SetLocation(l_Left, f_InfoLabel.GetBottom() + Lyt.s_SmallSpace);
            f_PortBox.SetIPPortLimits();
            f_PortBox.SetValueClipped(l_DefaultSamPort);
            f_ResultTitleLabel.Location = new Point(l_Left, f_PortBox.GetBottom());
            f_ResultTitleLabel.Width = l_InnerWidth;
            f_ResultLabel.Location = new Point(l_Left, f_ResultTitleLabel.GetBottom());
            f_ResultLabel.Width = l_InnerWidth;
            // We want enough space to display two lines.
            f_ResultLabel.Text = "_\n_";
            Lyt.SetFontAndHeight(f_ResultLabel, f_BoldFont);
            f_ResultLabel.Text = string.Empty;
            f_GroupBox.SetSizePlusBottomPadding(l_Width, f_ResultLabel.Bottom);
            base.AsFixedButtonDialog(l_DlgWidth, f_GroupBox);

            AdjustTestButtonDirtyness(true);

            f_PortBox.f_Box.ValueChanged += PortBox_ValueChanged;
            base.f_OKButton.Click += OKButton_Click;
            base.f_CancelButton.Click += base.DefaultOKCloseHandler;

            this.Text = l_TextSet[TextKeys.SamConnTestDlgTitle];
            f_GroupBox.Text = l_TextSet[TextKeys.SamConnTestDlgGroupBox];
            f_PortBox.SetCaption(l_TextSet, TextKeys.OptionsDlgAdvancedSamTcpPort);
            f_ResultTitleLabel.Text = l_TextSet.GetAsInputControlCaption(TextKeys.SamConnTestDlgTestResultLabel);
            base.f_OKButton.Text = l_TextSet[TextKeys.SamConnTestDlgTestButton];
            base.f_CancelButton.Text = l_TextSet[TextKeys.ButtonTextClose];
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                EndAnyTest();

                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();

                f_BoldFont.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        private void EndAnyTest()
        {
            if (null != f_Cntxt)
            {
                f_Cntxt.Abort();
                f_Cntxt = null;
            }
        }

        private void AdjustTestButtonDirtyness(bool a_Dirty)
        {
            GuiButton l_Button = base.f_OKButton;

            if (false == a_Dirty)
            {
                l_Button.Font = base.f_CancelButton.Font;
                l_Button.BackColor = Color.Transparent;
            }
            else
            {
                l_Button.Font = f_BoldFont;
                l_Button.BackColor = Color.FromArgb(100, Color.Orange);
            }
        }

        private void PortBox_ValueChanged(object sender, EventArgs e)
        {
            AdjustTestButtonDirtyness(true);
        }

        private void OKButton_Click(object a_Sender,
                                    EventArgs e)
        {
            const uint l_TimeoutMs = (20 * 1000);
            System.Net.IPEndPoint l_SamEndPoint = new System.Net.IPEndPoint(System.Net.IPAddress.Loopback, f_PortBox.p_ValueAsInt);

            DoTest(l_SamEndPoint, l_TimeoutMs);
        }

        private void DoTest(System.Net.IPEndPoint a_SamEndPoint,
                            uint a_TimeoutMs)
        {
            const string l_SamTestWorkerName = "SamTestWorker";

            TextSet l_TextSet = f_Inst.f_TextSet;
            string l_RemoteStr = l_TextSet.Format(TextKeys.IPEndPointFormat, a_SamEndPoint.Address, a_SamEndPoint.Port);
            f_ResultLabel.Text = string.Empty;
            f_ResultLabel.ForeColor = SystemColors.ControlText;

            // Create controller context.

            SamBridgeManager.SamTestContext l_Cntxt = new SamBridgeManager.SamTestContext(a_SamEndPoint, a_TimeoutMs);
            f_Cntxt = l_Cntxt;

            // Start worker.

            Util.StartThreadWithParam(l_SamTestWorkerName, SamTestWorker, System.Threading.ThreadPriority.Normal, l_Cntxt);

            // Open wait message dialog (abortable).

            string l_ProgressMsg = l_TextSet.Format(TextKeys.MsgSAMBridgeConnecting, l_RemoteStr);
            bool l_UIRunning = WaitMessageDialog.ShowDialog(l_TextSet, this, l_Cntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.SamConnTestDlgWaitDlgTitle, l_ProgressMsg);

            SamBridgeManager.SamTestContext.States l_State = (false == l_Cntxt.p_AbortRequested ? l_Cntxt.p_State : SamBridgeManager.SamTestContext.States.AbortedViaController);
            EndAnyTest();

            if (false == l_UIRunning)
            {
                // The user interface quitted in the meantime.

                return;
            }

            // Print the result.

            string l_ResStr;

            switch (l_State)
            {
                case SamBridgeManager.SamTestContext.States.AbortedViaController:
                    l_ResStr = l_TextSet[TextKeys.MsgSAMBridgeConnAbortedAsDemanded];
                    break;

                case SamBridgeManager.SamTestContext.States.FailedToConnect:
                case SamBridgeManager.SamTestContext.States.ConnectionError:
                    l_ResStr = l_TextSet[TextKeys.MsgSAMBridgeConnAttemptFailed];
                    break;

                case SamBridgeManager.SamTestContext.States.Timeout:
                    l_ResStr = "Remote host did not respond.";
                    break;

                case SamBridgeManager.SamTestContext.States.UnexpectedAnswer:
                    l_ResStr = "Remote host is not a (compatible) SAM bridge.";
                    break;

                case SamBridgeManager.SamTestContext.States.GoodAnswer:
                    l_ResStr = l_TextSet.Format(TextKeys.SamConnTestDlgSuccessMsg, l_RemoteStr);
                    f_ResultLabel.ForeColor = Color.DarkGreen;
                    break;

                default: throw (new UnexpectedStateException());
            }

            f_ResultLabel.Text = l_ResStr;
            AdjustTestButtonDirtyness(false);
        }

        private static void SamTestWorker(object a_Param)
        {
            SamBridgeManager.SamTestContext l_Cntxt = (SamBridgeManager.SamTestContext)a_Param;
            SamBridgeManager.SamTest(l_Cntxt);
        }

        public static DialogResult ShowModal(Instance a_Inst,
                                             IWin32Window a_Owner)
        {
            using (SamBridgeTestDialog l_Dlg = new SamBridgeTestDialog(a_Inst))
            {
                return l_Dlg.ShowDialog(a_Owner);
            }
        }
    }

    class I2PRouterStatusDialog : ButtonDialog
    {
        // This dialog provides a widget-based wrapper around a tiny
        // portion of the I2P router's console (the HTTP engine and
        // HTML parser are *very* crude however). It will probably
        // break when I2P changes the console's HTML for the next time.

        private readonly TextSet f_TextSet;

        private readonly GuiGroupBox f_GroupBox;
        private readonly GuiLabel f_InfoLabel;
        private readonly TextGroup f_VersionTextGroup;
        private readonly TextGroup f_UptimeTextGroup;
        private readonly TextGroup f_StatusTextGroup;
        private readonly GuiButton f_ShutdownButton;

        private readonly GuiTimer f_Timer;
        private bool f_PreventReentry;

        private bool f_Ready;
        private string f_VersionStr;
        private string f_UptimeStr;
        private string f_StatusStr;
        private string f_ConsoleNonce;

        public I2PRouterStatusDialog(TextSet a_TextSet) :
                                base(a_TextSet, DialogButtons.Cancel)
        {
            f_TextSet = a_TextSet;

            const int lc_DialogWidth = 400;
            int l_DialogWidth = Lyt.CalcX(lc_DialogWidth);
            int l_GroupBoxWidth = (l_DialogWidth - Lyt.s_Space2);

            f_GroupBox = new GuiGroupBox(this);
            int l_Left = f_GroupBox.Padding.Left;
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            f_InfoLabel = new GuiLabel(f_GroupBox);
            f_InfoLabel.Location = new Point(l_Left, f_GroupBox.Padding.Top);
            f_InfoLabel.Text = "This feature was tested with I2P 0.9.17 and may or may not work with future versions of I2P. The router must listen at port 7657 (as is default) for this to work.";
            f_InfoLabel.Size = new Size(l_InnerWidth, f_InfoLabel.GetPreferredSize(new Size(l_InnerWidth, 0)).Height);
            f_VersionTextGroup = TextGroup.Create(f_GroupBox, l_Left, f_InfoLabel.GetBottom());
            f_UptimeTextGroup = TextGroup.Create(f_GroupBox, l_Left, f_VersionTextGroup.GetBottom());
            f_StatusTextGroup = TextGroup.Create(f_GroupBox, l_Left, f_UptimeTextGroup.GetBottom());
            f_ShutdownButton = new GuiButton(f_GroupBox, true);
            f_Timer = new GuiTimer(true);

            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_VersionTextGroup.SetWidth(l_InnerWidth);
            f_UptimeTextGroup.SetWidth(l_InnerWidth);
            f_StatusTextGroup.SetWidth(l_InnerWidth);
            f_ShutdownButton.Location = new Point((l_GroupBoxWidth - f_ShutdownButton.Width) / 2, f_StatusTextGroup.GetBottom() + Lyt.s_SmallSpace);
            f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_ShutdownButton.Bottom);
            base.AsFixedButtonDialog(l_DialogWidth, f_GroupBox);

            f_ShutdownButton.Click += ShutdownButton_Click;
            f_Timer.e_AdvancedTick += Timer_e_AdvancedTick;
            f_Timer.Interval = 7500;

            this.Text = "I2P Router Status";
            f_GroupBox.Text = "I2P Router";
            f_VersionTextGroup.SetLeftText("Version:");
            f_UptimeTextGroup.SetLeftText("Uptime:");
            f_StatusTextGroup.SetLeftText("Status:");
            f_ShutdownButton.Text = "Request Router Shutdown";
            f_CancelButton.Text = a_TextSet[TextKeys.ButtonTextClose];

            ClearGUI();
        }

        protected override void Dispose(bool a_Disposing)
        {
            if (true == a_Disposing)
            {
                MainForm.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
                f_Timer.Dispose();
            }

            base.Dispose(a_Disposing);
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            this.Refresh();

            f_Timer.Start();

            base.OnShown(e);
        }

        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            // Need to call base class implementation.

            f_Timer.Stop();

            base.OnClosing(e);
        }

        private void  Timer_e_AdvancedTick(GuiTimer a_Sender,
                                           uint a_Times)
        {
            if (false == this.Visible)
            {
                return;
            }

            UpdateStatus();
        }

        private void ShutdownButton_Click(object a_Sender,
                                          EventArgs e)
        {
            MainForm l_MainWindow = Software.sp_MainForm;
            bool l_ShowMsgBox = f_Ready;

            if (true == l_ShowMsgBox &&
                (false == l_MainWindow.YesNoMsgBox("Really request that the I2P router shuts down?") ||
                 false == Software.sp_MainForm.p_IsRunning))
            {
                return;
            }

            if (false == f_Ready)
            {
                if (true == l_ShowMsgBox)
                {
                    // The status has changed while the message box got
                    // shown.

                    l_MainWindow.ShowOperationFailedMsg();
                }

                return;
            }

            f_ShutdownButton.Enabled = false;

            if (false == TrySendTermSignal())
            {
                l_MainWindow.ShowOperationFailedMsg();
            }

            UpdateStatus();
        }

        private bool TrySendTermSignal()
        {
            string l_PostPayload = string.Format("consoleNonce={0}&action=shutdown", f_ConsoleNonce);
            string l_RequestStr = BuildHTTPRequestStr(true, l_PostPayload);
            string l_AnswerStr = HandleHTTPRequest(l_RequestStr);

            return (null != l_AnswerStr &&
                    true == l_AnswerStr.StartsWith("HTTP/1.1 200 OK"));
        }

        private void UpdateStatus()
        {
            if (true == f_PreventReentry)
            {
                return;
            }

            string l_RequestStr = BuildHTTPRequestStr(false, null);
            string l_AnswerStr = HandleHTTPRequest(l_RequestStr);

            if (null == l_AnswerStr ||
                false == l_AnswerStr.StartsWith("HTTP/1.1 200 OK"))
            {
                goto FAIL;
            }

            if (false == ParseFields(l_AnswerStr, out f_VersionStr, out f_UptimeStr, out f_StatusStr, out f_ConsoleNonce))
            {
                goto FAIL;
            }

            f_Ready = true;
            f_VersionTextGroup.SetRightText(f_VersionStr);
            f_UptimeTextGroup.SetRightText(f_UptimeStr);
            f_StatusTextGroup.SetRightText(f_StatusStr);
            f_ShutdownButton.Enabled = true;

            return;

            FAIL:

            f_Ready = false;
            f_ConsoleNonce = f_StatusStr = f_UptimeStr = f_VersionStr = null;

            ClearGUI();
        }

        private void ClearGUI()
        {
            string l_UnknownValStr = f_TextSet[TextKeys.UnknownValue];
            f_VersionTextGroup.SetRightText(l_UnknownValStr);
            f_UptimeTextGroup.SetRightText(l_UnknownValStr);
            f_StatusTextGroup.SetRightText(l_UnknownValStr);
            f_ShutdownButton.Enabled = false;
        }

        private string HandleHTTPRequest(string a_Request)
        {
            // NOTE: This is quick and dirty. Return can be null on
            // error.

            if (true == f_PreventReentry)
            {
                goto FAIL;
            }

            const ushort lc_MaxWaitMs = 500;
            const uint lc_MiB = (1024 * 1024);
            Instance l_Inst = Software.sp_MainForm.f_Inst;
            System.Net.Sockets.TcpClient l_Client = new System.Net.Sockets.TcpClient();
            Time l_Start = l_Inst.GetTime();
            StringBuilder l_StrBld = new StringBuilder();
            Encoding l_Encoding = Encoding.UTF8;

            try
            {
                l_Client.Connect(System.Net.IPAddress.Loopback, 7657);

                {
                    byte[] l_RequestBytes = l_Encoding.GetBytes(a_Request);
                    l_Client.GetStream().Write(l_RequestBytes, 0, l_RequestBytes.Length);
                }

                while (true)
                {
                    f_PreventReentry = true;
                    {
                        Application.DoEvents();
                    }
                    f_PreventReentry = false;

                    if (false == Software.sp_MainForm.p_IsRunning)
                    {
                        break;
                    }

                    int l_Avail = l_Client.Available;

                    if (lc_MiB < l_Avail)
                    {
                        break;
                    }

                    if (0 >= l_Avail)
                    {
                        Util.MiniSleep();
                    }
                    else
                    {
                        byte[] l_Bytes = new byte[l_Avail];
                        l_Client.GetStream().Read(l_Bytes, 0, l_Bytes.Length);

                        try
                        {
                            l_StrBld.Append(l_Encoding.GetString(l_Bytes));
                        }
                        catch (DecoderFallbackException)
                        {
                            break;
                        }

                        if (lc_MiB <= l_StrBld.Length)
                        {
                            break;
                        }
                    }

                    if ((true == l_Client.Connected ||
                         0 < l_Client.Available) &&
                        lc_MaxWaitMs > l_Inst.GetTime().MillisecondsSince(l_Start))
                    {
                        continue;
                    }

                    return l_StrBld.ToString();
                }
            }
            catch (System.InvalidOperationException)
            {
            }
            catch (System.Net.Sockets.SocketException)
            {
            }
            catch (System.IO.IOException)
            {
            }
            finally
            {
                if (true == l_Client.Connected)
                {
                    try
                    {
                        l_Client.Client.Shutdown(System.Net.Sockets.SocketShutdown.Both);
                    }
                    catch (System.Net.Sockets.SocketException)
                    {
                    }

                    l_Client.Client.Close();
                }

                l_Client.Close();
            }

            FAIL:

            return null;
        }

        private static string BuildHTTPRequestStr(bool a_Post,
                                                  string a_PostPayload)
        {
            const string l_Path = "/home";
            const string lc_LF = "\r\n";

            StringBuilder l_Bld = new StringBuilder();
            l_Bld.Append(false == a_Post ? "GET" : "POST");
            l_Bld.Append(' ' + l_Path + " HTTP/1.1" + lc_LF +
                         "Host: 127.0.0.1:7657" + lc_LF +
                         "User-Agent: " + Software.c_AppName + lc_LF +
                         "Accept: text/html" + lc_LF +
                         "Accept-Language: en" + lc_LF +
                         "DNT: 1" + lc_LF +
                         "Referer: http://127.0.0.1:7657" + l_Path + lc_LF +
                         "Connection: close" + lc_LF);

            if (false == a_Post)
            {
                l_Bld.Append(lc_LF);
            }
            else
            {
                l_Bld.Append("Content-Type: application/x-www-form-urlencoded" + lc_LF + "Content-Length: ");
                l_Bld.Append(Util.NumberToInvString(a_PostPayload.Length) + lc_LF + lc_LF);
                l_Bld.Append(a_PostPayload);
            }

            return l_Bld.ToString();
        }

        private static bool ParseFields(string a_HTTPResponse,
                                        out string a_VersionStr,
                                        out string a_UptimeStr,
                                        out string a_StatusStr,
                                        out string a_ConsoleNonce)
        {
            // This is quick and dirty.

            int l_Idx = a_HTTPResponse.IndexOf("<table><tr title=");

            if (0 > l_Idx)
            {
                goto FAIL;
            }

            if (false == ParseNextTag(a_HTTPResponse, ref l_Idx, 1, "td", out a_VersionStr) ||
                false == ParseNextTag(a_HTTPResponse, ref l_Idx, 1, "td", out a_UptimeStr) ||
                false == ParseNextTag(a_HTTPResponse, ref l_Idx, 0, "a", out a_StatusStr))
            {
                goto FAIL;
            }

            l_Idx = a_HTTPResponse.IndexOf("name=\"consoleNonce\" value=\"");

            if (0 > l_Idx)
            {
                goto FAIL;
            }

            l_Idx += 27;

            if (a_HTTPResponse.Length <= l_Idx)
            {
                goto FAIL;
            }

            int l_FinalIdx = a_HTTPResponse.IndexOf('"', l_Idx);

            if (0 > l_FinalIdx)
            {
                goto FAIL;
            }

            int l_Len = (l_FinalIdx - l_Idx);

            if (256 < l_Len)
            {
                goto FAIL;
            }

            a_ConsoleNonce = a_HTTPResponse.Substring(l_Idx, l_FinalIdx - l_Idx);

            return true;

            FAIL:

            a_ConsoleNonce = a_StatusStr = a_UptimeStr = a_VersionStr = null;

            return false;
        }

        private static bool ParseNextTag(string a_HTML,
                                         ref int a_Idx,
                                         uint a_SkipCount,
                                         string a_Tag,
                                         out string a_Value)
        {
            // Helper for ParseFields().

            for (uint i = 0; true;)
            {
                a_Idx = a_HTML.IndexOf(a_Tag, a_Idx);

                if (0 > a_Idx)
                {
                    goto FAIL;
                }

                // We expect no space before the tag identifier.

                int l_FinalTagIdx = (a_Idx + a_Tag.Length);

                if (a_HTML.Length <= l_FinalTagIdx)
                {
                    goto FAIL;
                }

                if (0 == a_Idx ||
                    '<' != a_HTML[a_Idx - 1])
                {
                    goto CONTINUE;
                }

                char l_NextChar = a_HTML[l_FinalTagIdx];

                if (' ' != l_NextChar &&
                    '>' != l_NextChar)
                {
                    goto CONTINUE;
                }

                if (i++ == a_SkipCount)
                {
                    break;
                }

                CONTINUE:

                a_Idx = l_FinalTagIdx;
            }

            a_Idx = a_HTML.IndexOf('>', a_Idx);

            if (0 > a_Idx)
            {
                goto FAIL;
            }

            a_Idx++;

            if (a_HTML.Length <= a_Idx)
            {
                goto FAIL;
            }

            int l_FinalIdx = a_HTML.IndexOf('<', a_Idx);

            if (0 > l_FinalIdx)
            {
                goto FAIL;
            }

            const byte lc_MaxExpectedLen = 128;
            int l_Len = (l_FinalIdx - a_Idx);

            if (lc_MaxExpectedLen < l_Len)
            {
                goto FAIL;
            }

            a_Value = Support.Encodings.HtmlEncoding.HtmlDecode(a_HTML.Substring(a_Idx, l_Len));
            a_Idx = l_FinalIdx;

            return true;

            FAIL:

            a_Idx = (-1);
            a_Value = null;

            return false;
        }
    }

    sealed class I2PAddressInputDialog : ButtonDialog
    {
        private const ushort c_DlgWidth = 550;
        private const ushort c_TextBoxHeight = 250;

        private readonly Instance f_Inst;

        private readonly GuiGroupBox f_GroupBox;
        private readonly LabeledTextBox f_AddressBox;

        private I2PAddress f_Address;

        public I2PAddressInputDialog(Instance a_Inst,
                                     bool a_StreamAddr) :
                                base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            f_Inst = a_Inst;

            this.SuspendLayout();

            f_GroupBox = new GuiGroupBox(this);
            int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
            int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            f_AddressBox = new LabeledTextBox(f_GroupBox, l_InnerWidth, true);

            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_AddressBox.SetLocation(new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top));
            f_AddressBox.f_Box.Height = Lyt.CalcY(c_TextBoxHeight);
            f_AddressBox.f_Box.ScrollBars = ScrollBars.Vertical;
            f_GroupBox.Size = new Size(l_GroupBoxWidth, f_AddressBox.f_Box.Bottom + f_GroupBox.Padding.Bottom);

            AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
            base.SetSelectOnShowControl(f_AddressBox.f_Box);

            f_AddressBox.f_Box.MaxLength = (I2PAddress.c_MaximumBase64CharLen + 100);
            f_OKButton.Click += OKButton_Click;

            TextSet l_TextSet = a_Inst.f_TextSet;
            this.Text = l_TextSet[TextKeys.FilesTabAddrDlgTitle];
            f_GroupBox.Text = l_TextSet[TextKeys.FilesTabAddrDlgGroupBoxCaption];

            if (false == a_StreamAddr)
            {
                // FIXME: Make translatable.
                f_AddressBox.SetCaption("Datagram Address:");
            }
            else
            {
                f_AddressBox.SetCaption(l_TextSet, TextKeys.FilesTabAddrDlgStreamAddr);
            }

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(disposing);
        }

        private void OKButton_Click(object sender, EventArgs e)
        {
            string l_Str = f_AddressBox.f_Box.Text.Trim();

            f_Address = I2PAddress.TryCreateFromString(l_Str, 0, l_Str.Length);

            if (null != f_Address)
            {
                DialogResult = DialogResult.OK;

                this.Close();

                return;
            }

            f_Inst.f_MainForm.ErrMsgBox(f_Inst[TextKeys.FilesTabAddrDlgTextNotRecognizedAsAddr]);
        }

        public bool ShowDialog(IWin32Window a_Owner,
                               out I2PAddress a_Address)
        {
            if (DialogResult.OK != this.ShowDialog(a_Owner))
            {
                a_Address = null;

                return false;
            }

            a_Address = f_Address;

            return true;
        }
    }

    class FilesPanel : AdvancedTabPage
    {
        private const ushort c_FileUpdateInterval = 750;
        private const byte c_GenericTopTabCount = 7;

        private static readonly ushort[] s_SuggestedSpeedLimits = new ushort[]
        {
            5, 10, 20, 30, 40, 50, 75, 100, 150, 200, 300, 400, 500,
        };

        private static string[] s_GenericCategoryNames;
        private static string s_TabNameWithCountFormat;
        private static string[] s_RatingStrs;

        private static string s_UnknownValueStr;
        private static string s_IncompleteLocStr;
        private static string s_MissingLocStr;

        private static string s_NoSourcesStr;
        private static string s_SourcesShortFormat;
        private static string s_SourcesLongFormat;

        private static string[] s_DirectionStrs;

        private static string s_SecIdentNoKeyStr;
        private static string s_SecIdentRsaKeyFormat;

        private static string s_CompletenessEverythingStr;
        private static string s_CompletenessPercFormat;

        private static string s_I2PBusyStr;

        private static string s_TransportStylePlainStr;
        private static string s_TransportStyleComprStr;

        private static string s_FileStatusIncomplStartedStr;
        private static string s_FileStatusIncomplStoppedStr;
        private static string s_FileStatusMissingStr;
        private static string s_FileStatusCompleteStr;

        private readonly Instance f_Inst;

        private readonly HorSplitContainer f_Splitter;
        private readonly AdvancedTabControl f_TopTabCntrl;
        private readonly AdvancedTabControl f_BottomTabCntrl;
        private readonly ContextMenuStripEx f_TopTabCntxtMenu;

        private readonly DetailsPanel f_Details;
        private readonly FilePeersPanel f_Peers;
        private readonly TransfersPanel f_Transfers;
        private readonly FileNamesPanel f_FileNames;
        private readonly CommentsPanel f_Comments;

        private readonly int f_BottomPanelExtraY;

        // f_SelectedTab and f_MenuFilesTab can be null.
        private CategoryTab f_SelectedTab;
        private CategoryTab f_MenuFilesTab;
        private SelectedFilesInfo f_MenuSelFilesInfo;

        // f_SubscriptionActCntxt can be null.
        private KnownFile.ActivityContext f_SubscriptionActCntxt;

        private bool f_IsFilesPanelSelected;
        private bool f_BottomTabsAreClear;

        private CategoryList p_Categories
        {
            get { return f_Inst.f_MainForm.f_Categories; }
        }

        public FilesPanel(Instance a_Inst)
        {
            f_Inst = a_Inst;

            f_Splitter = new HorSplitContainer(this);
            f_Splitter.p_FixedPanel = FixedPanel.Panel2;
            f_Splitter.Dock = DockStyle.Fill;
            f_Splitter.p_UpdateWhileMoving = true;

            f_TopTabCntrl = new AdvancedTabControl(f_Splitter);
            f_BottomTabCntrl = new AdvancedTabControl(f_Splitter);
            f_Splitter.SetControls(f_TopTabCntrl, f_BottomTabCntrl);

            f_TopTabCntxtMenu = new ContextMenuStripEx(f_TopTabCntrl, true);

            f_TopTabCntrl.SetDefaultPadding(true);
            f_TopTabCntrl.p_AllowHidingThroughUI = true;
            f_TopTabCntrl.p_AlwaysShowPageListingButton = true;
            f_TopTabCntrl.p_AllowReorderingThroughUI = true;

            f_BottomTabCntrl.SetDefaultPadding(false);

            f_Details = new DetailsPanel(f_Inst, this);
            f_Peers = new FilePeersPanel(f_Inst);
            f_Transfers = new TransfersPanel(f_Inst);
            f_FileNames = new FileNamesPanel(f_Inst);
            f_Comments = new CommentsPanel(f_Inst);

            f_BottomTabCntrl.AddPageWithoutSelecting(f_Details, IconTypes.Sheet);
            f_BottomTabCntrl.AddPageWithoutSelecting(f_Peers, IconTypes.Person);
            f_BottomTabCntrl.AddPageWithoutSelecting(f_Transfers, IconTypes.Progress);
            f_BottomTabCntrl.AddPageWithoutSelecting(f_FileNames, IconTypes.Pendant);
            f_BottomTabCntrl.AddPageWithoutSelecting(f_Comments, IconTypes.Balloon);
            f_BottomTabCntrl.SetSelectedPageIndex(0, SelectionChangeOrigins.Programmatic);

            int l_TypicalTabHeight = f_BottomTabCntrl.p_TabHeight;
            f_Splitter.p_MinUpperHeight = (l_TypicalTabHeight + Lyt.s_LargeSpace2);
            f_BottomPanelExtraY = (l_TypicalTabHeight + Lyt.CalcY(11));

            TextSet l_TextSet = f_Inst.f_TextSet;
            f_Details.p_Caption = l_TextSet[TextKeys.FilesTabDetailsTitle];
            f_Peers.p_Caption = l_TextSet[TextKeys.FilesTabPeersTitle];
            f_Transfers.p_Caption = l_TextSet[TextKeys.FilesTabTransfersTitle];
            f_FileNames.p_Caption = l_TextSet[TextKeys.FilesTabFileNamesTitle];
            f_Comments.p_Caption = l_TextSet[TextKeys.FilesTabCommentsTitle];

            RequestNextFileListNow();
            SetDetailsPanelBarBackgroundImages(null);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_GenericCategoryNames = a_TextSet.ReadArray(TextKeys.FilesTabSubTabAll, c_GenericTopTabCount);
            s_TabNameWithCountFormat = a_TextSet[TextKeys.NameWithCountFormat];
            s_RatingStrs = a_TextSet.ReadArray(TextKeys.FileRatingNotRated, CommentsPanel.c_FileRatingCount);

            s_UnknownValueStr = a_TextSet[TextKeys.UnknownValue];
            s_IncompleteLocStr = a_TextSet[TextKeys.FilesTabDetailsIncompleteLoc];
            s_MissingLocStr = a_TextSet[TextKeys.FilesTabDetailsMissingLoc];

            s_NoSourcesStr = a_TextSet[TextKeys.SourcesNone];
            s_SourcesShortFormat = a_TextSet[TextKeys.SourcesShort];
            s_SourcesLongFormat = a_TextSet[TextKeys.SourcesLong];

            s_DirectionStrs = a_TextSet.ReadArray(TextKeys.DirectionDown, 3);

            s_SecIdentNoKeyStr = a_TextSet[TextKeys.RsaKeyNo];
            s_SecIdentRsaKeyFormat = a_TextSet[TextKeys.RsaKeyYesFormat];

            s_CompletenessEverythingStr = a_TextSet[TextKeys.CompletenessEverything];
            s_CompletenessPercFormat = a_TextSet[TextKeys.CompletenessPercFormat];

            s_I2PBusyStr = a_TextSet[TextKeys.SpeedI2PBusy];

            s_TransportStylePlainStr = a_TextSet[TextKeys.TransportStylePlain];
            s_TransportStyleComprStr = a_TextSet[TextKeys.TransportStyleCompressed];

            s_FileStatusIncomplStartedStr = a_TextSet[TextKeys.FileStatusIncomplEnabled];
            s_FileStatusIncomplStoppedStr = a_TextSet[TextKeys.FileStatusIncomplDisabled];
            s_FileStatusMissingStr = a_TextSet[TextKeys.FileStatusMissing];
            s_FileStatusCompleteStr = a_TextSet[TextKeys.FileStatusComplete];
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                SetOrRemoveTopTabSelectionChangedHandler(false);
                f_SelectedTab = null;

                f_TopTabCntrl.Dispose();
                f_BottomTabCntrl.Dispose();
                f_TopTabCntxtMenu.Dispose();
                f_Splitter.Dispose();
            }

            base.Dispose(disposing);
        }

        public void ArrangeLayoutOnShown()
        {
            // Make the lower panel as small as possible.

            f_Splitter.SetLowerHeight(f_Splitter.p_MinLowerHeight);
        }

        public void UpdateImagesOfAllPages()
        {
            f_TopTabCntrl.UpdateImagesOfAllPages();
            f_BottomTabCntrl.UpdateImagesOfAllPages();
        }

        private void BottomPanelMinHeightChanged(int a_MinHeight)
        {
            f_Splitter.p_MinLowerHeight = (f_BottomPanelExtraY + a_MinHeight);
        }

        public void FinishLoading()
        {
            ReloadAllTabs();

            f_Peers.FinishLoading();
            f_Transfers.FinishLoading();
            f_FileNames.FinishLoading();
            f_Comments.FinishLoading();

            SetOrRemoveTopTabSelectionChangedHandler(true);
            f_TopTabCntrl.e_BeforeDraggingTab += TopTabCntrl_BeforeDraggingTab;
            f_TopTabCntrl.e_DraggingTab += TopTabCntrl_DraggingTab;
            f_TopTabCntrl.e_AfterDraggingTab += TopTabCntrl_AfterDraggingTab;
            f_TopTabCntrl.e_HidingSelPageThroughUI += TopTabCntrl_HidingSelPageThroughUI;
            f_TopTabCntrl.e_PrepareListingPages += TopTabCntrl_PrepareListingPages;
            f_TopTabCntrl.e_ItemSelectedFromPageList += TopTabCntrl_ItemSelectedFromPageList;
            f_BottomTabCntrl.e_SelectedPageChanged += BottomTabCntrl_SelectedPageChanged;
            f_TopTabCntxtMenu.Opening += TopTabCntxtMenu_Opening;

            // Now that the handler is there, select the first tab.
            f_TopTabCntrl.SetSelectedPageIndex(0, SelectionChangeOrigins.Programmatic);
        }

        public void PrepareSaving()
        {
            if (null != f_SelectedTab)
            {
                f_SelectedTab.PrepareSaving();
            }

            f_Peers.PrepareSaving();
            f_Transfers.PrepareSaving();
            f_FileNames.PrepareSaving();
            f_Comments.PrepareSaving();
        }

        public Category GetSelectedCategory()
        {
            if (null != f_SelectedTab)
            {
                Category l_Category = f_SelectedTab.f_Category;

                if (true == l_Category.IsNonGeneric())
                {
                    return l_Category;
                }
            }

            return Category.s_Unassigned;
        }

        public bool HasDetailsPanelBarBackgroundImages()
        {
            return f_Details.HasBarBackgroundImages();
        }

        public void SetDetailsPanelBarBackgroundImages(BarBackgroundImages a_BarBgImages)
        {
            f_Details.SetBarBackgroundImages(a_BarBgImages);
        }

        private CategoryTab FindTab(Category a_Category)
        {
            return FindTab(a_Category, 0);
        }

        private CategoryTab FindTab(Category a_Category,
                                    int a_ScanStartingIndex)
        {
            // Return can be null.

            for (uint i = (uint)a_ScanStartingIndex, lim = f_TopTabCntrl.p_PageCount; i != lim; i++)
            {
                CategoryTab l_TabPage = (CategoryTab)f_TopTabCntrl.GetPageByIndex(i);

                if (l_TabPage.f_Category == a_Category)
                {
                    return l_TabPage;
                }
            }

            return null;
        }

        public void UpdateFiles(MainForm.FormInput.Files a_Files)
        {
            // This routine is called periodically through a timer.

            if (true == f_MenuSelFilesInfo.IsValid() &&
                null == f_MenuFilesTab)
            {
                // It's now safe to release this unneeded memory,
                // since the menu has closed and any handler was
                // executed.

                f_MenuSelFilesInfo = new SelectedFilesInfo();
            }

            if (null == f_SelectedTab)
            {
                return;
            }

            bool l_UpdateTaskBar;
            bool l_UpdateControls = f_Inst.f_MainForm.ShouldTabUpdateControls(f_IsFilesPanelSelected, out l_UpdateTaskBar);

            if (f_SelectedTab.f_Category == a_Files.f_Category)
            {
                if (true == l_UpdateControls)
                {
                    f_SelectedTab.UpdateFiles(a_Files.f_FileArr);
                }

                UpdateBottomTabs(f_SelectedTab.f_GridView, l_UpdateControls, l_UpdateTaskBar);

                RequestNextFileList(c_FileUpdateInterval);
            }
            else
            {
                // We've got input, but it's already obsolete, so we'll
                // request new input right away.

                if (true == l_UpdateControls)
                {
                    CategoryTab l_Tab = FindTab(a_Files.f_Category);

                    if (null != l_Tab)
                    {
                        l_Tab.UpdateFiles(a_Files.f_FileArr);
                    }
                }

                RequestNextFileListNow();
            }
        }

        private void TopTabCntxtMenu_Opening(object sender, System.ComponentModel.CancelEventArgs e)
        {
            f_TopTabCntxtMenu.DisposeItems();

            int l_Idx = f_TopTabCntxtMenu.p_ClickedTabPageIndex;
            CategoryTab l_Page = (CategoryTab)f_TopTabCntrl.GetPageByIndex((uint)l_Idx);
            f_TopTabCntrl.SetSelectedPage(l_Page, SelectionChangeOrigins.UserInteraction);

            Category l_Category = l_Page.f_Category;
            bool l_IsNotAllCategory = (0 != l_Category.f_Val);
            bool l_CurIsCustom = l_Category.IsNonGeneric();
            bool l_HasAnyCustom = (0 < f_Inst.f_MainForm.f_Categories.p_CustomCount);

            GenericMenuItem<Category>.CreateItem(f_TopTabCntxtMenu, f_Inst[TextKeys.FilesTabCntxtMenuHideTab], TopTabCntxtMenuHideTab_Click, l_Category).Enabled = l_IsNotAllCategory;
            GenericMenuItem<Category>.CreateItem(f_TopTabCntxtMenu, f_Inst[TextKeys.FilesTabCntxtMenuModifyTab], TopTabCntxtMenuModifyTab_Click, IconTypes.Pendant, l_Category).Enabled = l_CurIsCustom;
            GenericMenuItem<Category>.CreateItem(f_TopTabCntxtMenu, f_Inst[TextKeys.FilesTabCntxtMenuDeleteTab], TopTabCntxtMenuDeleteTab_Click, IconTypes.TrashBin, l_Category).Enabled = l_CurIsCustom;
            MainForm.AddMenuSeparator(f_TopTabCntxtMenu);
            MainForm.AddMenuItem(f_TopTabCntxtMenu, f_Inst[TextKeys.FilesTabCntxtMenuCreateNewTab], TopTabCntxtMenuCreateTab_Click, IconTypes.Star);
            MainForm.AddMenuItem(f_TopTabCntxtMenu, f_Inst[TextKeys.FilesTabCntxtMenuReorderCustomTabs], TopTabCntxtMenuReorderTabs_Click, IconTypes.File).Enabled = l_HasAnyCustom;
            p_Categories.FillMenuWithHiddenCategories(f_TopTabCntxtMenu, s_GenericCategoryNames, f_Inst.f_TextSet[TextKeys.FilesTabCntxtMenuUnhideTab], TopTabCntxtMenuShowTab_Click, true);
        }

        private static Category GetCategoryFromMenuItem(object a_Obj)
        {
            return ((GenericMenuItem<Category>)a_Obj).f_GenericTag;
        }

        private bool RemoveFilesFromCategoryAndUpdate(Category a_Category)
        {
            // Do not continue processing if false is returned.

            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiClearCategory, a_Category);
            if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
            {
                return false;
            }

            UpdateSpecificTab(a_Category);

            return true;
        }

        private void ReloadAllTabs()
        {
            // This routine will update the selected tab in case one is
            // already selected.

            CategoryList l_Categories = p_Categories;

            int l_TotalCount = (Category.c_GenericCount + l_Categories.p_CustomCount);
            int l_TabIndex = 0;
            CategoryList.CategoryInfo[] l_Arr = l_Categories.ExposeArray();
            for (uint i = 0; i < l_TotalCount; i++)
            {
                bool l_Generic = (Category.c_GenericCount > i);
                int l_CustomIdx;
                Category l_Category;

                if (true == l_Generic)
                {
                    l_CustomIdx = (-1);
                    l_Category = new Category(i);

                    if (false == l_Categories.IsGenericCategoryVisible(l_Category))
                    {
                        continue;
                    }
                }
                else
                {
                    l_CustomIdx = (int)(i - Category.c_GenericCount);

                    if (false == l_Arr[l_CustomIdx].f_Visibile)
                    {
                        continue;
                    }

                    l_Category = l_Arr[l_CustomIdx].f_Category;
                }

                CategoryTab l_ExistingTab = FindTab(l_Category, l_TabIndex);

                if (null == l_ExistingTab)
                {
                    string l_Name;
                    Image l_Icon;

                    if (true == l_Generic)
                    {
                        l_Name = s_GenericCategoryNames[i];
                        l_Icon = null;
                    }
                    else
                    {
                        l_Name = l_Arr[l_CustomIdx].f_Name;
                        l_Icon = l_Arr[l_CustomIdx].p_Icon;
                    }

                    // NOTE: Keep this order of initialization, since
                    // the TabControl class is tricky.

                    CategoryTab l_NewTab = new CategoryTab(this, l_Category);
                    l_NewTab.p_Image = l_Icon;
                    f_TopTabCntrl.InsertPage(l_NewTab, (uint)l_TabIndex, false);
                    l_NewTab.SetCaption(l_Name);
                }
                else
                {
                    f_TopTabCntrl.ChangePageIndex(l_ExistingTab, (uint)l_TabIndex, SelectionChangeOrigins.Programmatic);
                }

                l_TabIndex++;
            }

            // Update selection.

            if (null != f_SelectedTab)
            {
                int l_PredecessorIndex;
                bool l_Visible = l_Categories.FindVisibilityIndex(f_SelectedTab.f_Category, out l_PredecessorIndex);

                if (false == l_Visible)
                {
                    CategoryTab l_OldSelTab = f_SelectedTab;

                    f_TopTabCntrl.SetSelectedPageIndex((uint)l_PredecessorIndex, SelectionChangeOrigins.Programmatic);
                    f_TopTabCntrl.RemovePage(l_OldSelTab, true, SelectionChangeOrigins.Programmatic);
                }
            }

            // Delete any surplus tabs.

            f_TopTabCntrl.RemovePageRange((uint)l_TabIndex, f_TopTabCntrl.p_PageCount - (uint)l_TabIndex, true, SelectionChangeOrigins.Programmatic);
        }

        private void UpdateSpecificTab(Category a_Category)
        {
            // It's okay if l_Tab is null after the call.
            CategoryTab l_Tab = FindTab(a_Category);
            CategoryTab l_HideTab = null;
            bool l_ShowTab = false;
            string l_Name = null;
            Bitmap l_Icon = null;

            if (true == a_Category.IsGeneric())
            {
                bool l_Visible = p_Categories.IsGenericCategoryVisible(a_Category);

                if (false == l_Visible)
                {
                    l_HideTab = l_Tab;
                }
                else if (null == l_Tab)
                {
                    l_ShowTab = true;
                    l_Name = s_GenericCategoryNames[a_Category.f_Val];
                }
            }
            else
            {
                bool l_Visible;
                if (false == p_Categories.GetAttributes(a_Category, out l_Name, out l_Visible, out l_Icon) ||
                    false == l_Visible)
                {
                    l_HideTab = l_Tab;
                }
                else
                {
                    if (null == l_Tab)
                    {
                        l_ShowTab = true;
                    }
                    else
                    {
                        // The tab exists, only make sure name and
                        // icon fit.

                        l_Tab.SetCaptionAndImage(l_Name, l_Icon);
                    }
                }
            }

            if (null != l_HideTab)
            {
                // Hide a currently visible tab.

                if (f_SelectedTab == l_HideTab)
                {
                    // Select the page before the page that hides.
                    // NOTE: There will always be one.

                    uint l_HideIdx;
                    bool l_OK = f_TopTabCntrl.GetIndexOfPage(l_HideTab, out l_HideIdx);
                    f_TopTabCntrl.SetSelectedPageIndex(l_HideIdx - 1, SelectionChangeOrigins.UserInteraction);
                }

                f_TopTabCntrl.RemovePage(l_HideTab, true, SelectionChangeOrigins.Programmatic);
            }
            else if (true == l_ShowTab)
            {
                // Unhide a currently invisible tab.

                int l_InsertIdx;
                bool l_Vis = p_Categories.FindVisibilityIndex(a_Category, out l_InsertIdx);

                // NOTE: Keep this order of initialization, since
                // the TabControl class is tricky.

                l_Tab = new CategoryTab(this, a_Category);
                l_Tab.p_Image = l_Icon;
                f_TopTabCntrl.InsertPage(l_Tab, (uint)l_InsertIdx, false);
                l_Tab.SetCaption(l_Name);
                f_TopTabCntrl.SetSelectedPage(l_Tab, SelectionChangeOrigins.UserInteraction);
            }
        }

        private void TopTabCntxtMenuHideTab_Click(object sender, EventArgs e)
        {
            Category l_Category = GetCategoryFromMenuItem(sender);

            if (true == p_Categories.Hide(l_Category))
            {
                UpdateSpecificTab(l_Category);
            }
        }

        private void TopTabCntxtMenuShowTab_Click(object sender, EventArgs e)
        {
            Category l_Category = GetCategoryFromMenuItem(sender);

            if (true == p_Categories.Show(l_Category))
            {
                UpdateSpecificTab(l_Category);
            }
        }

        private void TopTabCntxtMenuDeleteTab_Click(object sender, EventArgs e)
        {
            Category l_Category = GetCategoryFromMenuItem(sender);

            string l_Name;
            if (false == p_Categories.GetName(l_Category, out l_Name) ||
                false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst.f_TextSet.Format(TextKeys.FilesTabCntxtMenuQuestionReallyDeleteTab, l_Name)))
            {
                return;
            }

            if (true == p_Categories.Remove(l_Category))
            {
                RemoveFilesFromCategoryAndUpdate(l_Category);
            }
        }

        private void TopTabCntxtMenuModifyTab_Click(object sender, EventArgs e)
        {
            Category l_Category = GetCategoryFromMenuItem(sender);
            bool l_OK;

            using (CustomCategoryDialog l_Dlg = new CustomCategoryDialog(f_Inst))
            {
                l_OK = l_Dlg.ShowModifyDialog(this, l_Category);
            }

            if (true == l_OK)
            {
                UpdateSpecificTab(l_Category);
            }
        }

        private void TopTabCntxtMenuCreateTab_Click(object sender, EventArgs e)
        {
            if (CategoryList.c_MaxCustomCategories <= p_Categories.p_CustomCount)
            {
                f_Inst.f_MainForm.WarnMsgBox(f_Inst[TextKeys.FilesTabCntxtMenuMsgMaxTabCountReached]);

                return;
            }

            Category l_Category;
            bool l_OK;

            using (CustomCategoryDialog l_Dlg = new CustomCategoryDialog(f_Inst))
            {
                l_OK = l_Dlg.ShowCreateDialog(this, out l_Category);
            }

            if (true == l_OK)
            {
                // First remove any files from that category that are
                // still assigned to it (possibly leaked from previous
                // sessions).

                RemoveFilesFromCategoryAndUpdate(l_Category);
            }
        }

        private void TopTabCntxtMenuReorderTabs_Click(object sender, EventArgs e)
        {
            CategoryList l_Categories = p_Categories;
            int l_CustomTabs = l_Categories.p_CustomCount;

            if (0 >= l_CustomTabs)
            {
                return;
            }

            ReorderItemsDialog<Category>.Item[] l_Items = l_Categories.CreateItemsArrayForReordering(s_GenericCategoryNames);
            bool l_OrderChanged;

            using (ReorderItemsDialog<Category> l_Dlg = new ReorderItemsDialog<Category>(f_Inst, TextKeys.FilesTabCntxtMenuReorderCustomTabs))
            {
                if (false == l_Dlg.ShowDialog(f_Inst.f_MainForm, Category.c_GenericCount, ModifyItemVisibilityMode.FirstItemMustStayVisible, l_Items, out l_OrderChanged))
                {
                    return;
                }
            }

            l_Categories.ApplyItemsArrayAfterReordering(l_Items, l_OrderChanged);
            ReloadAllTabs();
        }

        private void HideCategoryThroughUI(Category a_Category)
        {
            if (true == p_Categories.Hide(a_Category))
            {
                UpdateSpecificTab(a_Category);
            }
        }

        private void SetOrRemoveTopTabSelectionChangedHandler(bool a_Set)
        {
            if (false == a_Set)
            {
                f_TopTabCntrl.e_SelectedPageChanged -= TopTabCntrl_SelectedPageChanged;
            }
            else
            {
                f_TopTabCntrl.e_SelectedPageChanged += TopTabCntrl_SelectedPageChanged;
            }
        }

        private void TopTabCntrl_SelectedPageChanged(AdvancedTabControl a_Sender)
        {
            // We never expect p_SelectedPage to be null.
            f_SelectedTab = (CategoryTab)f_TopTabCntrl.p_SelectedPage;

            Util.VolatileWrite<Category>(out f_Inst.f_MainForm.f_Input.f_SubscribedFileCategory, f_SelectedTab.f_Category);
            RequestNextFileListNow();

            UpdateBottomTabs(f_SelectedTab.f_GridView);
        }

        private void TopTabCntrl_BeforeDraggingTab(AdvancedTabControl a_Sender,
                                                   AdvancedTabPage a_Page,
                                                   ref bool a_Handled)
        {
            if (false == p_Categories.IsMoreThanOneCustomCategoryVisible())
            {
                a_Handled = true;
            }
        }

        private void TopTabCntrl_DraggingTab(AdvancedTabControl a_Sender,
                                             AdvancedTabPage a_Page,
                                             ref uint a_NewIndex)
        {
            uint l_MinIdx = p_Categories.GetGenericVisibleCount();

            a_NewIndex = Math.Max(l_MinIdx, a_NewIndex);
        }

        private void TopTabCntrl_AfterDraggingTab(AdvancedTabControl a_Sender,
                                                  AdvancedTabPage a_Page,
                                                  bool a_DidDrag,
                                                  bool a_Commited)
        {
            if (false == a_DidDrag ||
                false == a_Commited)
            {
                return;
            }

            uint l_DisplIdx;

            if (true == f_TopTabCntrl.GetIndexOfPage(a_Page, out l_DisplIdx))
            {
                p_Categories.AdjustIndex(((CategoryTab)a_Page).f_Category, ((CategoryTab)f_TopTabCntrl.GetPageByIndex(l_DisplIdx - 1)).f_Category);
            }
        }

        private void TopTabCntrl_HidingSelPageThroughUI(AdvancedTabControl a_Sender,
                                                        AdvancedTabPage a_Page,
                                                        ref bool a_Handled)
        {
            a_Handled = true;

            HideCategoryThroughUI(f_SelectedTab.f_Category);
        }

        private void TopTabCntrl_PrepareListingPages(AdvancedTabControl a_Sender,
                                                     ref AdvancedTabControl.PageListItem[] a_Items)
        {
            p_Categories.PrepareListingPages(null == f_SelectedTab ? Category.s_Unassigned : f_SelectedTab.f_Category, s_GenericCategoryNames, ref a_Items);
        }

        private void TopTabCntrl_ItemSelectedFromPageList(AdvancedTabControl a_Sender,
                                                          uint a_Index,
                                                          ref bool a_Handled)
        {
            // a_Index may be outside bounds, in which case nothing
            // happens.

            a_Handled = true;

            CategoryList l_Categories = p_Categories;
            Category l_Category;

            if (false == l_Categories.FindCategoryByIndex(a_Index, out l_Category))
            {
                return;
            }

            if (false == l_Categories.IsCategoryVisible(l_Category))
            {
                bool l_Accepted = l_Categories.Show(l_Category);

                UpdateSpecificTab(l_Category);
            }
            else if (null == f_SelectedTab ||
                     l_Category != f_SelectedTab.f_Category)
            {
                CategoryTab l_Tab = FindTab(l_Category);
                f_TopTabCntrl.SetSelectedPage(l_Tab, SelectionChangeOrigins.UserInteraction);
            }
        }

        protected override void OnSelectPage(SelectionChangeOrigins a_Origin)
        {
            // Need to call base class implementation.

            f_IsFilesPanelSelected = true;
            RequestNextFileListNow();

            base.OnSelectPage(a_Origin);
        }

        protected override void OnDeselectPage(SelectionChangeOrigins a_Origin)
        {
            // Need to call base class implementation.

            f_IsFilesPanelSelected = false;
            f_Inst.f_MainForm.ResetTaskBarState();

            base.OnDeselectPage(a_Origin);
        }

        private void RequestNextFileListNow()
        {
            RequestNextFileList(0);
        }

        private void RequestNextFileList(uint a_MillisecsUntil)
        {
            if (false == f_IsFilesPanelSelected)
            {
                return;
            }

            Time l_NextTime = f_Inst.GetTime();

            if (0 != a_MillisecsUntil)
            {
                l_NextTime = l_NextTime.AddMilliseconds(a_MillisecsUntil);
            }

            System.Threading.Interlocked.Exchange(ref f_Inst.f_MainForm.f_Input.f_NextFileUpdateTime, (long)l_NextTime.ToBinary());
        }

        private void BottomTabCntrl_SelectedPageChanged(AdvancedTabControl a_Sender)
        {
            if (null != f_SelectedTab)
            {
                UpdateBottomTabs(f_SelectedTab.f_GridView);
            }
        }

        private bool HasBottomTabPageInputFocus()
        {
            AdvancedTabPage l_SelPage = f_BottomTabCntrl.p_SelectedPage;

            if (null == l_SelPage)
            {
                return false;
            }

            return ((BottomTabPage)l_SelPage).HasInputFocus();
        }

        private void ChildGridView_SelectionChanged(CategoryTab a_CategoryTab)
        {
            if (a_CategoryTab == f_SelectedTab)
            {
                UpdateBottomTabs(f_SelectedTab.f_GridView);
            }
        }

        private void UpdateBottomTabs(GridView<KnownFile, bool> a_GridView)
        {
            bool l_UpdateTaskBar;
            bool l_UpdateControls = f_Inst.f_MainForm.ShouldTabUpdateControls(f_IsFilesPanelSelected, out l_UpdateTaskBar);
            UpdateBottomTabs(a_GridView, l_UpdateControls, l_UpdateTaskBar);
        }

        private void UpdateBottomTabs(GridView<KnownFile, bool> a_GridView,
                                      bool a_UpdateControls,
                                      bool a_UpdateTaskBar)
        {
            if (true == a_UpdateTaskBar)
            {
                CategoryTab.SetTaskBarStateFromSelection(f_Inst, a_GridView);
            }

            if (false == a_UpdateControls)
            {
                return;
            }

            if (1 != a_GridView.p_SelectedCount)
            {
                if (true == f_BottomTabsAreClear)
                {
                    return;
                }

                f_Details.ClearDetails();
                f_Peers.Clear();
                f_Transfers.Clear();
                f_FileNames.Clear();
                f_Comments.Clear();

                f_BottomTabsAreClear = true;
            }
            else
            {
                GridViewItem<KnownFile, bool> l_Item = a_GridView.GetSelectedItem(0);
                KnownFileUpdateContext l_UpdateCntxt = new KnownFileUpdateContext(f_Inst.GetTime(), l_Item.f_Ref);

                if (true == l_UpdateCntxt.f_IsActive &&
                    l_UpdateCntxt.f_ActCntxt != f_SubscriptionActCntxt)
                {
                    if (null != f_SubscriptionActCntxt)
                    {
                        f_SubscriptionActCntxt.SetGuiPeerContextsSubscription(false);
                    }

                    f_SubscriptionActCntxt = l_UpdateCntxt.f_ActCntxt;
                    f_SubscriptionActCntxt.SetGuiPeerContextsSubscription(true);

                    // Speed things up by calling into the network
                    // manager.

                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiEnableAccessToFileAndCallBack, new object[] { l_UpdateCntxt.f_Known, new OneArgHandler(NetworkCallbackRoutineForUpdatingBottomTabs) });
                    f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);

                    // Do not wait for the APC to complete.
                }

                if (true == f_BottomTabsAreClear)
                {
                    f_BottomTabsAreClear = false;
                }

                switch ((BottomTabs)f_BottomTabCntrl.p_SelectedPageIndex)
                {
                    case BottomTabs.Details:
                        f_Details.Update(ref l_UpdateCntxt);
                        break;

                    case BottomTabs.Peers:
                        f_Peers.Update(ref l_UpdateCntxt);
                        break;

                    case BottomTabs.Transfers:
                        f_Transfers.Update(ref l_UpdateCntxt);
                        break;

                    case BottomTabs.FileNames:
                        f_FileNames.Update(ref l_UpdateCntxt);
                        break;

                    case BottomTabs.Comments:
                        f_Comments.Update(ref l_UpdateCntxt);
                        break;

                    default: throw (new ApplicationException());
                }
            }
        }

        private void NetworkCallbackRoutineForUpdatingBottomTabs(object a_Obj)
        {
            // This routine gets called as soon as possible after the
            // network thread was able to satisfy a data request.

            if (null != f_SelectedTab &&
                true == f_Inst.f_MainForm.p_IsRunning)
            {
                GridView<KnownFile, bool> l_GridView = f_SelectedTab.f_GridView;

                if (1 == l_GridView.p_SelectedCount &&
                    l_GridView.GetSelectedItem(0).f_Ref == a_Obj)
                {
                    UpdateBottomTabs(l_GridView);
                }
            }
        }

        private void TryToMoveFilesWithMessage(KnownFile[] a_Files,
                                               string a_TargetDirPath)
        {
            ComplStatus l_Status;
            string l_ContainingDirPath;
            string l_FileName = a_Files[0].GetFileNameVolatile(true, out l_Status, out l_ContainingDirPath);

            if (ComplStatus.Complete != l_Status ||
                null == l_ContainingDirPath)
            {
                return;
            }

            TextSet l_TextSet = f_Inst.f_TextSet;
            MainForm l_MainForm = f_Inst.f_MainForm;

            if (true == FileNames.AreNamesIdentical(a_TargetDirPath, l_ContainingDirPath))
            {
                return;
            }

            string l_Quest = l_TextSet.Format("Should the file \"{0}\" be moved?\n\nCurrent location: \"{1}\"\n\nDestination: \"{2}\"", l_FileName, l_ContainingDirPath, a_TargetDirPath);
            SharedDirectories l_SharedDirs = l_MainForm.p_Prefs.f_IOPrefs.f_SharedDirs;

            if (ShareKinds.None == l_SharedDirs.GetShareKind(a_TargetDirPath))
            {
                l_Quest += "\n\nNOTE: The destination directory is not a shared directory currently. The file will not be uploadable from there.";
            }

            if (false == l_MainForm.YesNoMsgBox(l_Quest))
            {
                return;
            }

            UnabortableControllerContext l_Cntxt = new UnabortableControllerContext();
            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiMoveCompleteFile, new object[] { a_Files, a_TargetDirPath, l_Cntxt });
            f_Inst.f_ContentMngr.EnqueueAPC(l_Cookie);

            if (false == WaitMessageDialog.ShowDialog(f_Inst.f_TextSet, f_Inst.f_MainForm, l_Cntxt, WaitMessageDialog.ProgressBarStyles.DummyProgressBar, TextKeys.MoveCompleteFilesDlgTitle, TextKeys.MoveCompleteFilesDlgMsg))
            {
                return;
            }

            bool l_Succ = (CookieStates.Success == l_Cookie.p_State);
            uint l_SuccCount = (false == l_Succ ? 0 : (uint)l_Cookie.GetReturnObject());

            if (0 == l_SuccCount)
            {
                l_MainForm.ErrMsgBox("Failed to move the file.");
            }
            else
            {
                // Nop. Don't nag the user if everything went well.
            }

            return;
        }

        public static string GetDlControlStr(bool a_Enabled)
        {
            return (false == a_Enabled ? FilesPanel.s_FileStatusIncomplStoppedStr : FilesPanel.s_FileStatusIncomplStartedStr);
        }

        private enum BottomTabs : byte
        {
            Details = 0,
            Peers = 1,
            Transfers = 2,
            FileNames = 3,
            Comments = 4,
        }

        private abstract class BottomTabPage : AdvancedTabPage
        {
            public abstract bool HasInputFocus();
        }

        private class CustomCategoryDialog : ButtonDialog
        {
            private const ushort c_DlgWidth = 320;
            private const ushort c_LeftIconWidth = 70;

            private readonly Instance f_Inst;

            private readonly GuiGroupBox f_GroupBox;
            private readonly LabeledTextBox f_NameBox;
            private readonly IconPicker f_IconPicker;
            private readonly ColorPicker f_ColorPicker;

            private bool f_Create;
            private Category f_Category;

            public CustomCategoryDialog(Instance a_Inst) :
                                   base(a_Inst.f_TextSet, DialogButtons.OKCancel)
            {
                f_Inst = a_Inst;

                this.SuspendLayout();

                f_GroupBox = new GuiGroupBox(this);
                int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
                int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
                int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
                int l_IconLeftWidth = Lyt.CalcX(c_LeftIconWidth);
                int l_IconRightWidth = (l_InnerWidth - Lyt.s_Space - l_IconLeftWidth);
                f_NameBox = new LabeledTextBox(f_GroupBox, l_InnerWidth);
                f_IconPicker = new IconPicker(f_Inst, f_GroupBox, l_IconLeftWidth);
                f_ColorPicker = new ColorPicker(f_Inst, f_GroupBox, l_IconRightWidth, false);

                int l_Left = f_GroupBox.Padding.Left;
                f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                f_NameBox.SetLocation(l_Left, f_GroupBox.Padding.Top);
                int l_IconTop = f_NameBox.GetBottom();
                f_IconPicker.SetLocation(l_Left, l_IconTop);
                f_ColorPicker.Location = new Point(l_Left + l_IconLeftWidth + Lyt.s_Space, l_IconTop);
                f_GroupBox.Size = new Size(l_GroupBoxWidth, Math.Max(f_ColorPicker.GetBottom(true), f_IconPicker.GetBottom()));

                AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
                base.SetSelectOnShowControl(f_NameBox.f_Box);

                f_NameBox.f_Box.MaxLength = CategoryList.c_MaxCategoryNameChars;
                f_ColorPicker.e_ColorChanged += ColorPicker_ColorChanged;
                f_OKButton.Click += OKButton_Click;

                TextSet l_TextSet = f_Inst.f_TextSet;
                f_GroupBox.Text = l_TextSet[TextKeys.FilesTabCustomTabDlgGroupBoxCaption];
                f_NameBox.SetCaption(l_TextSet, TextKeys.FilesTabCustomTabDlgTabName);

                this.ResumeLayout();
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    f_IconPicker.Dispose();
                    Dialog.DisposeAllChildren(f_GroupBox);
                    f_GroupBox.Dispose();
                }

                base.Dispose(disposing);
            }

            private void ColorPicker_ColorChanged(ColorPicker a_Sender,
                                                  bool a_Default,
                                                  Color a_Color)
            {
                f_IconPicker.SetBackColor(false == a_Default ? a_Color : Software.s_IconStorage.p_BackgroundColor);
            }

            private void OKButton_Click(object sender, EventArgs e)
            {
                string l_NewName = f_NameBox.f_Box.Text.Trim().Normalize();
                string l_ErrMsg = null;
                bool l_SelTextBoxOnErr = true;

                if (0 == l_NewName.Length)
                {
                    l_ErrMsg = f_Inst[TextKeys.FilesTabCustomTabDlgMsgBlockEmptyName];
                }
                else if (CategoryList.c_MaxCategoryNameChars < l_NewName.Length)
                {
                    // This might happen as a result of normalization.

                    l_ErrMsg = f_Inst[TextKeys.FilesTabCustomTabDlgMsgBlockLongName];
                }
                else
                {
                    bool l_OK = true;

                    foreach (string l_GenericName in FilesPanel.s_GenericCategoryNames)
                    {
                        if (true == CategoryList.IsSameName(l_GenericName, l_NewName))
                        {
                            l_OK = false;

                            break;
                        }
                    }

                    if (true == l_OK)
                    {
                        CategoryList l_Categories = f_Inst.f_MainForm.f_Categories;
                        IconTypes l_IconType = f_IconPicker.p_IconType;
                        Color l_IconColor = f_ColorPicker.GetColorEasy();

                        if (false == f_Create)
                        {
                            l_OK = l_Categories.Modify(f_Category, l_NewName, l_IconType, l_IconColor);
                        }
                        else
                        {
                            l_OK = l_Categories.Add(l_NewName, l_IconType, l_IconColor, out f_Category);
                        }

                        l_SelTextBoxOnErr = false;
                    }

                    if (false == l_OK)
                    {
                        l_ErrMsg = f_Inst[TextKeys.FilesTabCustomTabDlgMsgBlockExistingName];
                    }
                }

                if (null != l_ErrMsg)
                {
                    f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);

                    if (true == l_SelTextBoxOnErr)
                    {
                        f_NameBox.f_Box.Select();
                    }

                    return;
                }

                this.DialogResult = DialogResult.OK;
                this.Close();
            }

            public bool ShowCreateDialog(FilesPanel a_Panel,
                                         out Category a_Category)
            {
                f_Create = true;
                f_Category = new Category();

                this.Text = f_Inst[TextKeys.FilesTabCustomTabDlgTitleCreate];
                f_IconPicker.SetIconType(IconTypes.NoIcon);
                f_ColorPicker.ResetAttributes();

                if (DialogResult.OK == this.ShowDialog(f_Inst.f_MainForm))
                {
                    a_Category = f_Category;

                    return true;
                }

                a_Category = new Category();

                return false;
            }

            public bool ShowModifyDialog(FilesPanel a_Panel,
                                         Category a_Category)
            {
                string l_Name;
                IconTypes f_IconType;
                Color f_IconColor;

                if (false == f_Inst.f_MainForm.f_Categories.GetAttributes(a_Category, out l_Name, out f_IconType, out f_IconColor))
                {
                    // No such category.

                    throw (new ApplicationException());
                }

                f_Create = false;
                f_Category = a_Category;

                this.Text = f_Inst[TextKeys.FilesTabCustomTabDlgTitleRename];
                f_NameBox.f_Box.Text = l_Name;
                f_IconPicker.SetIconType(f_IconType);
                f_ColorPicker.SetColorEasy(f_IconColor, true);

                return (DialogResult.OK == this.ShowDialog(f_Inst.f_MainForm));
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct KnownFileUpdateContext
        {
            public readonly Time f_Time;
            public readonly KnownFile f_Known;
            public readonly bool f_IsActive;
            public readonly KnownFile.ActivityContext f_ActCntxt;
            public readonly bool f_HasPeers;
            public readonly DlPeerContext[] f_DlPeers;
            public readonly UlPeerContext[] f_UlPeers;

            public KnownFileUpdateContext(Time a_Time,
                                          KnownFile a_Known)
            {
                f_Time = a_Time;
                f_Known = a_Known;
                f_IsActive = f_Known.GetActivityContextVolatile(out f_ActCntxt);

                if (true == f_IsActive)
                {
                    f_HasPeers = f_ActCntxt.LoadGuiAccessablePeerContexts(out f_DlPeers, out f_UlPeers);
                }
                else
                {
                    f_HasPeers = false;
                    f_DlPeers = null;
                    f_UlPeers = null;
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct SelectedFilesInfo
        {
            // f_MenuFiles can be null.
            public readonly KnownFile[] f_Files;
            public readonly int f_IncomplFiles;
            public readonly int f_MissingFiles;
            public readonly int f_ComplFiles;

            public int p_Count
            {
                get { return (null == f_Files ? 0 : f_Files.Length); }
            }

            public SelectedFilesInfo(GridView<KnownFile, bool> a_GridView)
            {
                DataGridViewSelectedRowCollection l_Coll = a_GridView.SelectedRows;
                int l_Count = l_Coll.Count;

                f_MissingFiles = 0;
                f_ComplFiles = 0;

                if (0 == l_Count)
                {
                    f_Files = null;
                    f_IncomplFiles = 0;

                    return;
                }

                f_Files = new KnownFile[l_Count];

                int l_Idx = 0;
                foreach (GridViewItem<KnownFile, bool> l_Item in l_Coll)
                {
                    ComplStatus l_Status = l_Item.f_Ref.p_Status;

                    if (ComplStatus.Complete == l_Status)
                    {
                        f_ComplFiles++;
                    }
                    else if (ComplStatus.Missing == l_Status)
                    {
                        f_MissingFiles++;
                    }

                    f_Files[l_Idx++] = l_Item.f_Ref;
                }

                f_IncomplFiles = (l_Count - f_MissingFiles - f_ComplFiles);
            }

            public bool IsValid()
            {
                return (null != f_Files);
            }

            public DownloadControl GetCommonDlControl(out bool a_AllHaveSameControl)
            {
                a_AllHaveSameControl = true;
                DownloadControl l_DlControl;

                if (null == f_Files)
                {
                    l_DlControl = DownloadControl.DlDisabled;
                }
                else
                {
                    int l_Idx = (f_Files.Length - 1);
                    l_DlControl = f_Files[l_Idx].f_DlControl;

                    for (--l_Idx; 0 <= l_Idx; l_Idx--)
                    {
                        if (l_DlControl != f_Files[l_Idx].f_DlControl)
                        {
                            a_AllHaveSameControl = false;

                            return DownloadControl.DlDisabled;
                        }
                    }
                }

                return l_DlControl;
            }
        }

        private class CategoryTab : AdvancedTabPage
        {
            private const byte c_MaxPreviewChunks = 4;
            private const Keys c_ActionShortcutKey = Keys.Space;
            private const Keys c_DisableDlShortcutKey = Keys.F12;
            private const Keys c_EnableDlShortcutKey = Keys.F11;
            private const Keys c_DropFilesShortcutKey = Keys.Delete;

            private readonly Instance f_Inst;
            private readonly FilesPanel f_FilesPanel;
            public readonly Category f_Category;

            // f_Tag is unused (so use the bool type).
            public readonly GridView<KnownFile, bool> f_GridView;

            private string f_Caption;
            private bool f_MayDisplayCount;

            public CategoryTab(FilesPanel a_Panel,
                               Category a_Category)
            {
                f_Inst = a_Panel.f_Inst;
                f_FilesPanel = a_Panel;
                f_Category = a_Category;

                f_GridView = new GridView<KnownFile, bool>(a_Panel.f_Inst, this, a_Panel.f_Inst.f_MainForm.f_VisualCust.f_FilesTabColSet);

                if (0 != f_Category.f_Val)
                {
                    this.p_AllowHidingThroughUI = true;
                }

                if (true == f_Category.IsNonGeneric())
                {
                    this.p_AllowReorderingThroughUI = true;
                }

                f_GridView.SelectionChanged += GridView_SelectionChanged;
                f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
                f_GridView.e_FillContextMenu += FillContextMenu;
                f_GridView.e_ClosedContextMenu += GridView_ClosedContextMenu;
                f_GridView.e_UpdateItem += GridView_UpdateItem;
                f_GridView.KeyDown += GridView_KeyDown;
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    f_GridView.Dispose();
                }

                base.Dispose(disposing);
            }

            public void SetCaptionAndImage(string a_Caption,
                                           Image a_Image)
            {
                SetCaption(a_Caption);
                SetImage(a_Image, true);
            }

            public void SetCaption(string a_Caption)
            {
                if (a_Caption == f_Caption)
                {
                    return;
                }

                f_Caption = a_Caption;

                UpdateCaption();
            }

            private void UpdateCaption()
            {
                string l_Str = f_Caption;

                if (true == f_MayDisplayCount)
                {
                    int l_Count = f_GridView.p_ItemCount;

                    if (0 < l_Count)
                    {
                        l_Str = f_Inst.f_TextSet.Format(FilesPanel.s_TabNameWithCountFormat, l_Str, l_Count);
                    }
                }

                base.p_Caption = l_Str;
            }

            public void PrepareSaving()
            {
                f_GridView.StoreColumnAdjustments();
            }

            protected override void OnSelectPage(SelectionChangeOrigins a_Origin)
            {
                // Need to call base class implementation.

                f_MayDisplayCount = true;
                UpdateCaption();

                f_GridView.LoadColumnAdjustments();
                f_GridView.Select();

                base.OnSelectPage(a_Origin);
            }

            protected override void OnDeselectPage(SelectionChangeOrigins a_Origin)
            {
                // Need to call base class implementation.

                f_MayDisplayCount = false;
                UpdateCaption();

                f_GridView.StoreColumnAdjustments();

                base.OnDeselectPage(a_Origin);
            }

            private void GridView_SelectionChanged(object sender, EventArgs e)
            {
                f_FilesPanel.ChildGridView_SelectionChanged(this);
            }

            private void GridView_ItemDoubleClicked(GridViewItem<KnownFile, bool> a_Item)
            {
                HandleActionKeyPressed(a_Item.f_Ref);
            }

            private void HandleActionKeyPressed(KnownFile a_Known)
            {
                bool l_RunDirectly = f_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_RunFilesOnDblClick;

                if (false == l_RunDirectly)
                {
                    OpenContainingFolder(a_Known);
                }
                else
                {
                    OpenOrPreviewFile(a_Known);
                }
            }

#region Menu

            private KnownFile[] GetMenuFiles()
            {
                // NOTE: This is quite tricky and will probably change.
                // For the moment: If this routine is called after the
                // menu has closed, it must happen right in the click
                // handler of the menu and before control is transferred
                // to other parts of the GUI (e.g. via message box).

                SelectedFilesInfo l_MenuSelFilesInfo = f_FilesPanel.f_MenuSelFilesInfo;

                return l_MenuSelFilesInfo.f_Files;
            }

            private bool FillContextMenu(GridView<KnownFile, bool> a_Sender,
                                         ContextMenuStripEx a_Menu,
                                         bool a_ItemHit)
            {
                if (false == a_ItemHit)
                {
                    return false;
                }

                SelectedFilesInfo l_SelFilesInfo = new SelectedFilesInfo(f_GridView);

                if (false == FillContextMenuCore(a_Menu, ref l_SelFilesInfo))
                {
                    f_FilesPanel.f_MenuFilesTab = null;
                    f_FilesPanel.f_MenuSelFilesInfo = new SelectedFilesInfo();

                    return false;
                }

                f_FilesPanel.f_MenuFilesTab = this;
                f_FilesPanel.f_MenuSelFilesInfo = l_SelFilesInfo;

                return true;
            }

            private bool FillContextMenuCore(ContextMenuStripEx a_Menu,
                                             ref SelectedFilesInfo a_SelFilesInfo)
            {
                int l_Count = a_SelFilesInfo.p_Count;

                if (0 == l_Count)
                {
                    return false;
                }

                TextSet l_TextSet = f_Inst.f_TextSet;

                // Open and preview.

                if (1 == l_Count)
                {
                    if (0 == a_SelFilesInfo.f_IncomplFiles)
                    {
                        MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuOpen], MenuOpenOrPreview_Click, IconTypes.Lightning).Enabled = (0 < a_SelFilesInfo.f_ComplFiles);
                    }
                    else
                    {
                        IncomplFileProgress l_IncomplProgress = a_SelFilesInfo.f_Files[0].p_IncomplProgressVolatile;

                        bool l_Enabled = (null != l_IncomplProgress &&
                                          true == l_IncomplProgress.CanBePreviewed());

                        MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuPreview], MenuOpenOrPreview_Click, IconTypes.Camera).Enabled = l_Enabled;
                    }

                    MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuOpenFolder], MenuOpenContainingFolder_Click, IconTypes.Box).Enabled = (0 < a_SelFilesInfo.f_ComplFiles);
                    MainForm.AddMenuSeparator(a_Menu);
                }

                // Download control.

                if (0 < a_SelFilesInfo.f_IncomplFiles)
                {
                    bool l_CanDisable = false;
                    bool l_CanEnable = false;
                    bool l_AllHaveSameControl;
                    DownloadControl l_DlControl = a_SelFilesInfo.GetCommonDlControl(out l_AllHaveSameControl);

                    l_CanDisable = (false == l_AllHaveSameControl ||
                                    DownloadControl.DlDisabled != l_DlControl);
                    l_CanEnable = (false == l_AllHaveSameControl ||
                                   DownloadControl.DlEnabled != l_DlControl);

                    MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuEnableDl], MenuStartFile_Click, c_EnableDlShortcutKey, IconTypes.Progress).Enabled = l_CanEnable;
                    MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuDisableDl], MenuStopFile_Click, c_DisableDlShortcutKey, IconTypes.Forbidden).Enabled = l_CanDisable;
                    MainForm.AddMenuSeparator(a_Menu);
                }

                ToolStripMenuItem l_Download = MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuDlBandwidth], null, IconTypes.Download);
                ToolStripMenuItem l_Upload = MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuUlBandwidth], null, IconTypes.CloudUp);
                Speed l_CommonDlLimit;
                Speed l_CommonUlLimit;
                Priority l_CommonDlPriority;
                Priority l_CommonUlPriority;
                FileSpeedAndPriorityDialog.AnalyzeFiles(a_SelFilesInfo.f_Files, out l_CommonDlLimit, out l_CommonUlLimit, out l_CommonDlPriority, out l_CommonUlPriority);

                if (0 == a_SelFilesInfo.f_IncomplFiles)
                {
                    l_Download.Enabled = false;
                }
                else
                {
                    CreateDlUlSubMenuItems(true, l_Download, DlSubMenu_Click, l_CommonDlPriority, l_CommonDlLimit);
                }

                CreateDlUlSubMenuItems(false, l_Upload, UlSubMenu_Click, l_CommonUlPriority, l_CommonUlLimit);

                MainForm.AddMenuItem(a_Menu, l_TextSet.GetAsDialogOpener(TextKeys.FilesTabListCntxtMenuAdjustSpeedAndPriority), MenuOpenSpeedAndPriorityDlg_Click, IconTypes.Gear);
                MainForm.AddMenuSeparator(a_Menu);

                // Category handling.

                f_Inst.f_MainForm.f_Categories.FillMenuWithCustomCategorySubMenu(l_TextSet, a_Menu, TextKeys.FilesTabListCntxtMenuAssignToCustomTab, true, MenuChangeFileCategory_Click);

                // Advanced sub menu.

                ToolStripMenuItem l_Advanced = new ToolStripMenuItem(l_TextSet[TextKeys.FilesTabListCntxtMenuAdvanced]);
                GenericMenuItem<DownloadQueue.ResetOperation>.CreateSubItem(l_Advanced, l_TextSet[TextKeys.FilesTabListCntxtMenuRequerySources], MenuResetFiles_Click, DownloadQueue.ResetOperation.RequerySources).Enabled = (0 < a_SelFilesInfo.f_IncomplFiles);
                MainForm.AddMenuItem(l_Advanced, l_TextSet.GetAsDialogOpener(TextKeys.FilesTabListCntxtMenuManuallyAddAddr), MenuManuallyAddSource_Click).Enabled = (0 < a_SelFilesInfo.f_IncomplFiles);
                MainForm.AddMenuSeparator(l_Advanced);
                GenericMenuItem<DownloadQueue.ResetOperation>.CreateSubItem(l_Advanced, l_TextSet[TextKeys.FilesTabListCntxtMenuResetPeerStates], MenuResetFiles_Click, DownloadQueue.ResetOperation.ResetPeerStates).Enabled = (0 < a_SelFilesInfo.f_IncomplFiles);
                a_Menu.Items.Add(l_Advanced);

                MainForm.AddMenuSeparator(a_Menu);

                // Link and collection handling.

                GenericMenuItem<FileLink.CopyLinkToClipboardOps>.AddLinkMenu(l_TextSet, a_Menu, 1 == l_Count, MenuCopyLink_Click);
                MainForm.AddMenuItem(a_Menu, l_TextSet.GetAsDialogOpener(TextKeys.FilesTabListCntxtMenuCreateColl), MenuCreateCollection_Click, IconTypes.LetterClosed).Enabled = (a_SelFilesInfo.f_ComplFiles == l_Count);
                MainForm.AddMenuSeparator(a_Menu);

                // Removal and rechecking.

                bool l_AllowComplFilesDeletion = f_Inst.f_MainForm.p_AllowComplFilesDeletion;
                TextKeys l_RemovalTextKey = (false == l_AllowComplFilesDeletion &&
                                             0 == a_SelFilesInfo.f_ComplFiles ? TextKeys.FilesTabListCntxtMenuRemoveIncomplOrMissing : TextKeys.FilesTabListCntxtMenuRemoveAnyFiles);
                ToolStripMenuItem l_RemovalMenuItem = MainForm.AddMenuItem(a_Menu, l_TextSet[l_RemovalTextKey], MenuDropFiles_Click, c_DropFilesShortcutKey, IconTypes.TrashBin);

                if (false == l_AllowComplFilesDeletion)
                {
                    l_RemovalMenuItem.Enabled = (a_SelFilesInfo.f_ComplFiles < l_Count);
                }

                MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuRemoveTurnMissingIntoIncompl], MenuTurnMissingIntoIncompleteFiles_Click).Enabled = (0 < a_SelFilesInfo.f_MissingFiles);
                MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FilesTabListCntxtMenuRemoveForceRecheckOfCompl], MenuRecheckFiles_Click, IconTypes.Recycle).Enabled = (0 < a_SelFilesInfo.f_ComplFiles);

                return true;
            }

            private void GridView_ClosedContextMenu(object sender, ToolStripDropDownClosedEventArgs e)
            {
                if (this == f_FilesPanel.f_MenuFilesTab)
                {
                    // This panel has contributed the most recent files,
                    // but we may not nullify the menu files yet, since
                    // a click handler might need them just now.

                    f_FilesPanel.f_MenuFilesTab = null;
                }
            }

            private void CreateDlUlSubMenuItems(bool a_ForDl,
                                                ToolStripMenuItem a_Parent,
                                                EventHandler a_ClickHandler,
                                                Priority a_Priority,
                                                Speed a_SpeedLimit)
            {
                string[] l_Texts = PriorityPickerControl.s_TitleCasedTexts;
                for (int i = 0; i < Priority.c_GuiPriorityStages; i++)
                {
                    MainForm.AddMenuItem(a_Parent, l_Texts[i], a_ClickHandler);
                }

                int l_Idx = a_Priority.EncodeIntoIndex();

                if (0 <= l_Idx)
                {
                    ((ToolStripMenuItem)a_Parent.DropDownItems[l_Idx]).CheckState = CheckState.Checked;
                }

                MainForm.AddMenuSeparator(a_Parent);

                MainForm l_Form = f_Inst.f_MainForm;
                bool l_SI = l_Form.p_DisplaySISizes;
                uint l_Granularity = SizeFormatter.GetGranularity(l_SI);
                uint l_EffLimit = a_SpeedLimit.p_BytesPerSec;
                bool l_LimitHandled = false;

                GenericMenuItem<Speed> l_MinItem = GenericMenuItem<Speed>.CreateSubItem(a_Parent, f_Inst[false == a_ForDl ? TextKeys.SpeedDontUl : TextKeys.SpeedDontDl], a_ClickHandler, Speed.s_NoSpeed);

                if (true == a_SpeedLimit.IsMax())
                {
                    // The max item will be added and checked below.

                    l_LimitHandled = true;
                }
                else if (true == a_SpeedLimit.IsZero())
                {
                    l_MinItem.CheckState = CheckState.Checked;
                    l_LimitHandled = true;
                }
                else if (false == a_SpeedLimit.IsValid())
                {
                    l_LimitHandled = true;
                }

                bool l_SISizes = f_Inst.f_MainForm.p_DisplaySISizes;
                for (int i = 0, lim = s_SuggestedSpeedLimits.Length; i <= lim; i++)
                {
                    uint l_CurLimit;
                    bool l_AddSpecial;

                    if (i < lim)
                    {
                        l_CurLimit = (s_SuggestedSpeedLimits[i] * l_Granularity);
                        l_AddSpecial = (false == l_LimitHandled &&
                                        l_CurLimit > l_EffLimit);
                    }
                    else
                    {
                        l_CurLimit = 0;
                        l_AddSpecial = (false == l_LimitHandled);
                    }

                    if (true == l_AddSpecial)
                    {
                        Speed l_SpeedEffLimit = new Speed(l_EffLimit);
                        ToolStripMenuItem l_CustomItem = GenericMenuItem<Speed>.CreateSubItem(a_Parent, l_SpeedEffLimit.ToMenuString(l_SISizes), a_ClickHandler, l_SpeedEffLimit);
                        l_CustomItem.CheckState = CheckState.Checked;
                        l_LimitHandled = true;
                    }

                    if (i < lim)
                    {
                        Speed l_SpeedCurLimit = new Speed(l_CurLimit);
                        ToolStripMenuItem l_PredefItem = GenericMenuItem<Speed>.CreateSubItem(a_Parent, l_SpeedCurLimit.ToMenuString(l_SISizes), a_ClickHandler, l_SpeedCurLimit);

                        if (false == l_LimitHandled &&
                            l_CurLimit == l_EffLimit)
                        {
                            l_PredefItem.CheckState = CheckState.Checked;
                            l_LimitHandled = true;
                        }
                    }
                }

                GenericMenuItem<Speed> l_MaxItem = GenericMenuItem<Speed>.CreateSubItem(a_Parent, f_Inst[TextKeys.SpeedNoLimit], a_ClickHandler, Speed.s_MaxSpeed);

                if (true == a_SpeedLimit.IsMax())
                {
                    l_MaxItem.CheckState = CheckState.Checked;
                }
            }

            private void DlSubMenu_Click(object sender, EventArgs e)
            {
                InterpretDlUlSubMenuClick(true, sender);
            }

            private void UlSubMenu_Click(object sender, EventArgs e)
            {
                InterpretDlUlSubMenuClick(false, sender);
            }

            private void InterpretDlUlSubMenuClick(bool a_Dl,
                                                   object a_Sender)
            {
                ToolStripMenuItem l_Item = (ToolStripMenuItem)a_Sender;
                int l_Idx = l_Item.Owner.Items.IndexOf(l_Item);

                if (Priority.c_GuiPriorityStages > l_Idx)
                {
                    Priority l_Prio = Priority.DecodeFromIndex(l_Idx);

                    foreach (GridViewItem<KnownFile, bool> l_GridViewItem in f_GridView.SelectedRows)
                    {
                        if (false == a_Dl)
                        {
                            l_GridViewItem.f_Ref.f_UlPriority = l_Prio;
                        }
                        else
                        {
                            l_GridViewItem.f_Ref.f_DlPriority = l_Prio;
                        }
                    }

                    return;
                }

                Speed l_SpeedLimit = ((GenericMenuItem<Speed>)a_Sender).f_GenericTag;

                foreach (GridViewItem<KnownFile, bool> l_GridViewItem in f_GridView.SelectedRows)
                {
                    if (false == a_Dl)
                    {
                        l_GridViewItem.f_Ref.f_UlLimit = l_SpeedLimit;
                    }
                    else
                    {
                        l_GridViewItem.f_Ref.f_DlLimit = l_SpeedLimit;
                    }
                }
            }

            private void GridView_KeyDown(object sender, KeyEventArgs e)
            {
                // This is for forwarding menu shortcuts in case the
                // corresponding menu item doesn't exist or is disabled.

                Keys l_Modifiers = e.Modifiers;

                if (true == Dialog.HasNonShiftModifiers(l_Modifiers))
                {
                    return;
                }

                Keys l_KeyCode = e.KeyCode;

                if (c_ActionShortcutKey != l_KeyCode &&
                    c_DisableDlShortcutKey != l_KeyCode &&
                    c_EnableDlShortcutKey != l_KeyCode &&
                    c_DropFilesShortcutKey != l_KeyCode)
                {
                    return;
                }

                // We will handle this.

                e.Handled = true;
                int l_SelCount = f_GridView.p_SelectedCount;

                if (c_ActionShortcutKey == l_KeyCode)
                {
                    if (1 != l_SelCount)
                    {
                        return;
                    }
                }
                else
                {
                    if (0 >= l_SelCount)
                    {
                        return;
                    }
                }

                // Do not rely on menu files here.
                SelectedFilesInfo l_SelFilesInfo = new SelectedFilesInfo(f_GridView);

                switch (l_KeyCode)
                {
                    case c_ActionShortcutKey:
                        HandleActionKeyPressed(l_SelFilesInfo.f_Files[0]);
                        break;

                    case c_DisableDlShortcutKey:
                    case c_EnableDlShortcutKey:
                        DownloadControl l_Control = (c_DisableDlShortcutKey == l_KeyCode ? DownloadControl.DlDisabled : DownloadControl.DlEnabled);
                        SetDlControlForFiles(ref l_SelFilesInfo, l_Control);
                        break;

                    case c_DropFilesShortcutKey:
                        MenuDropFiles_Invoke(ref l_SelFilesInfo, l_Modifiers);
                        break;
                }
            }

            private void MenuOpenSpeedAndPriorityDlg_Click(object sender, EventArgs e)
            {
                using (FileSpeedAndPriorityDialog l_Dlg = new FileSpeedAndPriorityDialog(f_Inst, GetMenuFiles()))
                {
                    l_Dlg.ShowDialog(f_Inst.f_MainForm);
                }
            }

            private void MenuResetFiles_Click(object sender, EventArgs e)
            {
                GenericMenuItem<DownloadQueue.ResetOperation> l_Item = (GenericMenuItem<DownloadQueue.ResetOperation>)sender;
                DownloadQueue.ResetOperation l_ResetOp = l_Item.f_GenericTag;

                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.f_DlQueue.GuiResetFile, new object[] { GetMenuFiles(), l_ResetOp });
                f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
            }

            private void MenuManuallyAddSource_Click(object sender, EventArgs e)
            {
                KnownFile[] l_MenuFiles = GetMenuFiles();
                bool l_OK;
                I2PAddress l_Addr;

                using (I2PAddressInputDialog l_Dlg = new I2PAddressInputDialog(f_Inst, true))
                {
                    l_OK = l_Dlg.ShowDialog(f_Inst.f_MainForm, out l_Addr);
                }

                if (true == l_OK)
                {
                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.f_DlQueue.GuiAddSource, new object[] { l_MenuFiles, l_Addr });
                    f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
                }
            }

            private void MenuOpenOrPreview_Click(object sender, EventArgs e)
            {
                OpenOrPreviewFile(GetMenuFiles()[0]);
            }

            private void OpenOrPreviewFile(KnownFile a_Known)
            {
                TextSet l_TextSet = f_Inst.f_TextSet;
                ComplStatus l_Status;
                string l_ParentPath;
                string l_FileName = a_Known.GetFileNameVolatile(true, out l_Status, out l_ParentPath);

                if (ComplStatus.Complete == l_Status)
                {
                    // Open the file.

                    string l_FullFileName = LongPath.Combine(l_ParentPath, l_FileName);

                    if (FileType.FT_PROGRAM == FileTypeDetector.GuessFileType(l_FullFileName))
                    {
                        string l_Question = l_TextSet.Format(TextKeys.FilesTabListLaunchQuestReallyExecuteStrangeFile, l_FullFileName);

                        if (false == f_Inst.f_MainForm.YesNoMsgBox(l_Question))
                        {
                            return;
                        }
                    }

                    if (false == Util.ShellOpenOrExecuteFile(l_FullFileName))
                    {
                        string l_ErrMsg = l_TextSet.Format(TextKeys.FilesTabListLaunchMsgFailedToLaunch, l_FullFileName);

                        f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);
                    }
                }
                else if (ComplStatus.Incomplete == l_Status)
                {
                    // Preview the file.

                    IncomplFileProgress l_IncomplProgress = a_Known.p_IncomplProgressVolatile;
                    string l_ErrMsg = null;

                    if (null == l_IncomplProgress)
                    {
                        l_ErrMsg = l_TextSet[TextKeys.FilesTabListLaunchMsgBeginningNotDled];

                        goto PRINT_PREVIEW_ERR_MSG;
                    }

                    uint l_ReadyChunks = 0;
                    uint l_ChunksToCheck = Math.Min(l_IncomplProgress.f_DoneMap.f_BitCount, c_MaxPreviewChunks);

                    for (; l_ReadyChunks != l_ChunksToCheck; l_ReadyChunks++)
                    {
                        if (false == l_IncomplProgress.f_DoneMap.Get(l_ReadyChunks))
                        {
                            break;
                        }
                    }

                    if (0 == l_ReadyChunks)
                    {
                        l_ErrMsg = l_TextSet[TextKeys.FilesTabListLaunchMsgBeginningNotDled];

                        goto PRINT_PREVIEW_ERR_MSG;
                    }

                    uint l_Size = (l_ReadyChunks * Content.Internal.IncompleteFile.c_ChunkSize);
                    ExtractionCookie l_Cookie;
                    ExtractionCookie.Result l_Result = ExtractionCookie.CreateForPreviewing(f_Inst, l_FileName, l_Size, out l_Cookie);

                    if (ExtractionCookie.Result.Success == l_Result)
                    {
                        // We will now register the GUI as a source to
                        // this incomplete file, begin extracting its
                        // beginning and then unregister again.

                        SourceCookie l_SourceCookie = new SourceCookie(UserNum.s_PreviewExtraction);

                        if (false == f_Inst.f_ContentMngr.RegisterSourceAsync(a_Known, l_SourceCookie))
                        {
                            goto PRINT_PREVIEW_ERR_MSG;
                        }

                        // Wait right here until the content manager
                        // answers.

                        if (false == l_SourceCookie.GuiWaitForCompletion())
                        {
                            return;
                        }

                        if (CookieStates.Success != l_SourceCookie.p_ContentMngrState)
                        {
                            goto PRINT_PREVIEW_ERR_MSG;
                        }

                        bool l_ExtractionRunning = f_Inst.f_ContentMngr.ExtractAsync(a_Known, l_Cookie);
                        bool l_Unregistered = f_Inst.f_ContentMngr.UnregisterSourceAsync(a_Known, l_SourceCookie);

                        if (true == l_ExtractionRunning)
                        {
                            // We're done so far.

                            f_Inst.f_MainForm.AddExtractionCookie(l_Cookie);

                            return;
                        }

                        goto PRINT_PREVIEW_ERR_MSG;
                    }

                    TextKeys l_ErrMsgKey;

                    switch (l_Result)
                    {
                        case ExtractionCookie.Result.UnsuitableExtension:
                            l_ErrMsgKey = TextKeys.FilesTabListLaunchMsgNotFitForPreviewing;
                            break;

                        case ExtractionCookie.Result.NotEnoughDiskSpace:
                            l_ErrMsgKey = TextKeys.FilesTabListLaunchMsgLackOfFreeSpace;
                            break;

                        case ExtractionCookie.Result.LowDiskSpace:
                            l_ErrMsgKey = TextKeys.FilesTabListLaunchMsgLowFreeSpace;
                            break;

                        case ExtractionCookie.Result.FailedToCreateFile:
                            l_ErrMsgKey = TextKeys.FilesTabListLaunchMsgErrAccessingTempFile;
                            break;

                        case ExtractionCookie.Result.PreviewFolderDoesntExist:
                            l_ErrMsgKey = TextKeys.Empty;
                            l_ErrMsg = l_TextSet.Format(TextKeys.FilesTabListLaunchMsgPreviewFolderMissing, f_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_PreviewFolder);
                            break;

                        default:
                            l_ErrMsgKey = TextKeys.FilesTabListLaunchMsgUnspecifiedErr;
                            break;
                    }

                    if (TextKeys.Empty != l_ErrMsgKey)
                    {
                        l_ErrMsg = l_TextSet[l_ErrMsgKey];
                    }

                    PRINT_PREVIEW_ERR_MSG:

                    l_ErrMsg = l_TextSet.Format(TextKeys.FilesTabListLaunchMsgFailedToPreview, l_FileName, l_ErrMsg);

                    f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);
                }
            }

            private void MenuOpenContainingFolder_Click(object sender, EventArgs e)
            {
                OpenContainingFolder(GetMenuFiles()[0]);
            }

            private void OpenContainingFolder(KnownFile a_Known)
            {
                // Works only for complete files (of course).

                ComplStatus l_Status;
                string l_ParentPath;
                string l_FileName = a_Known.GetFileNameVolatile(true, out l_Status, out l_ParentPath);

                if (ComplStatus.Complete == l_Status)
                {
                    if (false == f_Inst.f_MainForm.NavigateToObject(LongPath.Combine(l_ParentPath, l_FileName)))
                    {
                        bool l_Succ = f_Inst.f_MainForm.NavigateToFolder(l_ParentPath);
                    }
                }
            }

            private void MenuStartFile_Click(object sender, EventArgs e)
            {
                SetDlControlFromMenu(DownloadControl.DlEnabled);
            }

            private void MenuStopFile_Click(object sender, EventArgs e)
            {
                SetDlControlFromMenu(DownloadControl.DlDisabled);
            }

            private void SetDlControlFromMenu(DownloadControl a_Control)
            {
                // Do not rely on menu files here. They can be obsolete
                // as there is a keyboard shortcut.
                SelectedFilesInfo l_SelFilesInfo = new SelectedFilesInfo(f_GridView);
                SetDlControlForFiles(ref l_SelFilesInfo, a_Control);
            }

            private void SetDlControlForFiles(ref SelectedFilesInfo a_SelFilesInfo,
                                              DownloadControl a_Control)
            {
                if (0 >= a_SelFilesInfo.f_IncomplFiles)
                {
                    return;
                }

                foreach (KnownFile l_Known in a_SelFilesInfo.f_Files)
                {
                    l_Known.f_DlControl = a_Control;
                }

                Util.MemoryBarrier();
            }

            private void MenuChangeFileCategory_Click(object sender, EventArgs e)
            {
                GenericMenuItem<Category> l_Item = (GenericMenuItem<Category>)sender;
                Category l_Category = l_Item.f_GenericTag;

                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiChangeFilesCategory, new object[2] { GetMenuFiles(), l_Category });
                f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
            }

            private void MenuCopyLink_Click(object sender, EventArgs e)
            {
                KnownFile[] l_MenuFiles = GetMenuFiles();

                GenericMenuItem<FileLink.CopyLinkToClipboardOps> l_Item = (GenericMenuItem<FileLink.CopyLinkToClipboardOps>)sender;
                KnownFile l_Known = l_MenuFiles[0];
                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiGetFileLink, l_Known);
                if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
                {
                    return;
                }

                if (CookieStates.Success != l_Cookie.p_StateVolatile)
                {
                    // Should never happen.

                    return;
                }

                FileLink l_Link = (FileLink)l_Cookie.GetReturnObject();

                if (ComplStatus.Complete == l_Known.p_Status &&
                    false == l_Link.f_MasterHashes.HasAllHashes(P2PHashSets._ALL_HASHSETS_))
                {
                    // If this is a complete file, we don't export
                    // until we know all about it.

                    f_Inst.f_MainForm.WarnMsgBox(f_Inst[TextKeys.FilesTabListCopyLinkMsgFileNotHashedYet]);

                    return;
                }

                if (false == l_Link.HasEnoughForExport())
                {
                    f_Inst.f_MainForm.WarnMsgBox(f_Inst[TextKeys.FilesTabListCopyLinkMsgMissingParams]);

                    return;
                }

                FileLink.CopyLinkToClipboardOps l_Op = l_Item.f_GenericTag;
                bool l_OK = l_Link.CopyLinkToClipboard(l_Op);

                if (false == l_OK)
                {
                    // Can happen if the file was deleted in the meantime.

                    f_Inst.f_MainForm.ErrMsgBox(f_Inst[TextKeys.FilesTabListCopyLinkMsgFailedToCopy]);
                }
            }

            private void MenuCreateCollection_Click(object sender, EventArgs e)
            {
                FileCollection l_Collection;
                FileCollection.LoadResult l_Result = FileCollection.CreateFromFileArray(GetMenuFiles(), out l_Collection);

                if (FileCollection.LoadResult.Success != l_Result)
                {
                    TextKeys l_MsgKey;

                    switch (l_Result)
                    {
                        case FileCollection.LoadResult.NoFiles:
                            l_MsgKey = TextKeys.FilesTabListCreateCollMsgNoFilesSelected;
                            break;

                        case FileCollection.LoadResult.TooManyFiles:
                            l_MsgKey = TextKeys.FilesTabListCreateCollMsgTooManyFiles;
                            break;

                        case FileCollection.LoadResult.NeedMoreHashing:
                            l_MsgKey = TextKeys.FilesTabListCreateCollMsgFilesNotHashedYet;
                            break;

                        default:
                            Util.DebugBreak();
                            return;
                    }

                    f_Inst.f_MainForm.ErrMsgBox(f_Inst[l_MsgKey]);

                    return;
                }

                using (CreateCollectionDialog l_Dlg = new CreateCollectionDialog(f_Inst, l_Collection))
                {
                    l_Dlg.ShowDialog(f_Inst.f_MainForm);
                }
            }

            private void MenuDropFiles_Click(object sender, EventArgs e)
            {
                // Do not rely on menu files here. They can be obsolete
                // as there is a keyboard shortcut.
                SelectedFilesInfo l_SelFilesInfo = new SelectedFilesInfo(f_GridView);
                MenuDropFiles_Invoke(ref l_SelFilesInfo, Form.ModifierKeys);
            }

            private void MenuDropFiles_Invoke(ref SelectedFilesInfo a_SelFilesInfo,
                                              Keys a_Modifiers)
            {
                int l_Count = a_SelFilesInfo.p_Count;

                if (0 == l_Count)
                {
                    return;
                }

                FileRemovalModes l_ComplFilesRemovalMode;

                if (false == f_Inst.f_MainForm.p_AllowComplFilesDeletion)
                {
                    // In this case we can safely pass a selection that
                    // contains complete files to the content manager,
                    // since those files will be ignored.

                    l_ComplFilesRemovalMode = FileRemovalModes.DoNothing;
                }
                else
                {
                    // The shift modifier indicates the kind of action.

                    bool l_Shift = (0 != (Keys.Shift & a_Modifiers));
                    l_ComplFilesRemovalMode = (false == l_Shift ? FileRemovalModes.MoveToRecycleBin : FileRemovalModes.DeletePermanently);
                }

                // Build message.

                string l_SubMsg;

                switch (l_ComplFilesRemovalMode)
                {
                    case FileRemovalModes.DoNothing:
                        l_SubMsg = string.Empty;
                        break;

                    case FileRemovalModes.MoveToRecycleBin:
                        l_SubMsg = f_Inst[TextKeys.FilesTabDropFilesSubMsgCompleteFilesWillBeMovedToRecycleBin];
                        break;

                    case FileRemovalModes.DeletePermanently:
                        l_SubMsg = f_Inst[TextKeys.FilesTabDropFilesSubMsgCompleteFilesWillBeDeleted];
                        break;

                    default: throw(new ApplicationException());
                }

                string l_Quest = f_Inst.f_TextSet.Format(TextKeys.FilesTabDropFilesQuestionReally, l_SubMsg);

                if (false == f_Inst.f_MainForm.YesNoMsgBox(l_Quest))
                {
                    return;
                }

                FileRemovalContext l_Cntxt = new FileRemovalContext(a_SelFilesInfo.f_Files, l_ComplFilesRemovalMode);

                f_Inst.f_ContentMngr.GuiBeginFileRemoval(l_Cntxt);

                if (false == WaitMessageDialog.ShowDialog(f_Inst.f_TextSet, f_Inst.f_MainForm, l_Cntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.RemovingFilesDlgTitle, TextKeys.RemovingFilesDlgMsgInGeneral))
                {
                    return;
                }

                f_Inst.f_MainForm.ShowDeletedFilesMsgBoxIfNecessary((int)l_Cntxt.GetSuccCount(), l_Count, l_Cntxt.p_Aborted);
            }

            private void MenuTurnMissingIntoIncompleteFiles_Click(object sender, EventArgs e)
            {
                KnownFile[] l_MenuFiles = GetMenuFiles();

                if (false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.FilesTabQuestionReallyTurnMissingIntoIncomplFiles]))
                {
                    return;
                }

                GeneralAPCCookie l_Cookie = f_Inst.f_ContentMngr.GuiBeginToTurnMissingIntoIncomplFiles(l_MenuFiles);
                if (false == l_Cookie.GuiWaitForCompletion())
                {
                    return;
                }

                if (CookieStates.Success != l_Cookie.p_StateVolatile)
                {
                    // Should never happen.

                    f_Inst.f_MainForm.ShowOperationFailedMsg();
                }
            }

            private void MenuRecheckFiles_Click(object sender, EventArgs e)
            {
                KnownFile[] l_MenuFiles = GetMenuFiles();

                if (false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.FilesTabQuestionReallyRecheckComplFiles]))
                {
                    return;
                }

                GeneralAPCCookie l_Cookie = f_Inst.f_ContentMngr.GuiBeginRecheckOfCompleteFiles(l_MenuFiles);
                if (false == l_Cookie.GuiWaitForCompletion())
                {
                    return;
                }

                if (CookieStates.Success != l_Cookie.p_StateVolatile)
                {
                    // Should never happen.

                    f_Inst.f_MainForm.ShowOperationFailedMsg();
                }
            }

#endregion

            public static void SetTaskBarStateFromSelection(Instance a_Inst,
                                                            GridView<KnownFile, bool> a_GridView)
            {
                // Show the combined progress of up to l_MaxFiles
                // incomplete files on the task bar.

                const byte l_MaxFiles = 64;

                int l_Count = a_GridView.p_SelectedCount;
                MainForm l_Form = a_Inst.f_MainForm;

                if (0 < l_Count &&
                    l_MaxFiles >= l_Count)
                {
                    const ComplStatus l_UnknownStatus = (ComplStatus)126;
                    const ComplStatus l_ConflictingStatus = (ComplStatus)127;

                    ComplStatus l_CommonStatus = l_UnknownStatus;
                    int l_IncomplFiles = 0;
                    ulong l_CompletedSize = 0;
                    ulong l_TotalSize = 0;
                    GeneralPrefs.ProgessBarKinds l_ProgressBarKind = l_Form.p_Prefs.f_GeneralPrefs.f_ProgressBarKind;

                    foreach (GridViewItem<KnownFile, bool> l_Item in a_GridView.SelectedRows)
                    {
                        KnownFile l_Known = l_Item.f_Ref;
                        ComplStatus l_Status = l_Known.p_Status;
                        IncomplFileProgress l_IncomplProgress = l_Known.f_IncomplProgress;
                        Util.MemoryBarrier();

                        if (l_UnknownStatus == l_CommonStatus)
                        {
                            l_CommonStatus = l_Status;
                        }
                        else if (l_ConflictingStatus != l_CommonStatus &&
                                 l_Status != l_CommonStatus)
                        {
                            // No common status.

                            l_CommonStatus = l_ConflictingStatus;
                        }

                        if (ComplStatus.Incomplete == l_Status)
                        {
                            l_IncomplFiles++;

                            if (null != l_IncomplProgress)
                            {
                                l_CompletedSize += (GeneralPrefs.ProgessBarKinds.VerifiedOnly == l_ProgressBarKind ? l_IncomplProgress.f_CompletedBytes : l_IncomplProgress.f_CompletedBytesInclUnverified);
                            }

                            l_TotalSize += l_Known.f_FileSize;
                        }
                    }

                    switch (l_CommonStatus)
                    {
                        case ComplStatus.Complete:
                            l_Form.SetTaskBarStateFilled(Support.TaskBar.TaskBarProgressKinds.Green);
                            return;

                        case ComplStatus.Missing:
                            l_Form.SetTaskBarStateFilled(Support.TaskBar.TaskBarProgressKinds.Red);
                            return;
                    }

                    if (0 < l_IncomplFiles)
                    {
                        // Regard only selected incomplete files.

                        Support.TaskBar.TaskBarProgressContext l_Cntxt = new Support.TaskBar.TaskBarProgressContext(Support.TaskBar.TaskBarProgressKinds.Yellow, l_CompletedSize, l_TotalSize);
                        l_Form.SetTaskBarStateAndValues(ref l_Cntxt);

                        return;
                    }

                    // Disable progress indication, since no meaningful
                    // state can be figured out.
                }

                l_Form.ResetTaskBarState();
            }

            public void UpdateFiles(KnownFile[] a_Files)
            {
                // This routine is called periodically through a timer.

                if (null == a_Files)
                {
                    f_GridView.ClearItems();
                }
                else
                {
                    f_GridView.StartChanging();

                    foreach (KnownFile l_Known in a_Files)
                    {
                        GridViewItem<KnownFile, bool> l_Item = f_GridView.CreateItem(l_Known);
                    }

                    f_GridView.EndChanging();
                }

                UpdateCaption();

                // We cannot always focus the grid view right in the
                // moment the corresponding tab page is selected (since
                // the page isn't enabled in that moment), so try
                // periodically.
                if (false == f_GridView.Focused &&
                    true == f_FilesPanel.IsPageSelected() &&
                    false == f_FilesPanel.HasBottomTabPageInputFocus())
                {
                    f_GridView.Select();
                }
            }

            private void GridView_UpdateItem(Time a_Time,
                                             GridViewItem<KnownFile, bool> a_Item)
            {
                TextSet l_TextSet = f_Inst.f_TextSet;
                MainForm l_Form = f_Inst.f_MainForm;
                bool l_SISizes = l_Form.p_DisplaySISizes;

                KnownFile l_Known = a_Item.f_Ref;
                KnownFile.ActivityContext l_ActCntxt;
                bool l_IsActive = l_Known.GetActivityContextVolatile(out l_ActCntxt);

                ComplStatus l_Status;
                string l_ParentPath;
                string l_FileName = l_Known.GetFileNameVolatile(true, out l_Status, out l_ParentPath);

                ulong l_Have;
                ulong l_HaveInclUnver;
                ulong l_Remaining;
                float l_Progress;
                FileStatus l_FileStatus;
                string l_FileStatusStr;

                if (ComplStatus.Incomplete == l_Status)
                {
                    IncomplFileProgress l_IncomplProgress = l_Known.p_IncomplProgressVolatile;

                    if (null == l_IncomplProgress)
                    {
                        l_Have = 0;
                        l_HaveInclUnver = 0;
                        l_Remaining = l_Known.f_FileSize;
                        l_Progress = 0;
                    }
                    else
                    {
                        l_Have = l_IncomplProgress.f_CompletedBytes;
                        l_HaveInclUnver = l_IncomplProgress.f_CompletedBytesInclUnverified;
                        l_Remaining = (l_Known.f_FileSize - l_HaveInclUnver);
                        ulong l_Achieved = (GeneralPrefs.ProgessBarKinds.VerifiedOnly == l_Form.p_Prefs.f_GeneralPrefs.f_ProgressBarKind ? l_Have : l_HaveInclUnver);
                        l_Progress = IncomplFileProgress.CalcQuotient(l_Achieved, l_Known.f_FileSize);
                    }

                    l_FileStatus = (FileStatus)Util.VolatileRead(ref l_Known.f_DlControl);
                    l_FileStatusStr = GetDlControlStr(FileStatus.IncomplEnabled == l_FileStatus);

                    string l_ProgressStr = l_TextSet.PercStr(l_Progress);
                    a_Item.SetDrawContext(6, OwnerDrawContext.CreateIncomplFileProgressBarContext(l_IsActive, l_Progress));
                }
                else
                {
                    // This file was completed.

                    l_Have = l_Known.f_FileSize;
                    l_HaveInclUnver = l_Known.f_FileSize;
                    l_Remaining = 0;
                    l_Progress = 1f;

                    if (ComplStatus.Missing == l_Status)
                    {
                        l_FileStatus = FileStatus.Missing;
                        l_FileStatusStr = FilesPanel.s_FileStatusMissingStr;
                        a_Item.SetDrawContext(6, OwnerDrawContext.CreateMissingFileProgressBarContext());
                    }
                    else
                    {
                        l_FileStatus = FileStatus.Finished;
                        l_FileStatusStr = FilesPanel.s_FileStatusCompleteStr;
                        a_Item.SetDrawContext(6, OwnerDrawContext.CreateComplFileProgressBarContext(l_IsActive));
                    }
                }

                a_Item.SetTextWithCompObj(0, l_FileName, l_FileName);
                a_Item.SetText(1, l_Known.f_FileSize, l_Form.BuildDefaultSizeStr(l_Known.f_FileSize));
                a_Item.SetText(2, (byte)l_FileStatus, l_FileStatusStr);
                a_Item.SetText(3, l_Have, l_Form.BuildDefaultSizeStr(l_Have));
                a_Item.SetText(4, l_HaveInclUnver, l_Form.BuildDefaultSizeStr(l_HaveInclUnver));
                a_Item.SetText(5, l_Remaining, l_Form.BuildDefaultSizeStr(l_Remaining));

                // ETA, sources and speed.

                const uint lc_ETANotApplicable = uint.MaxValue;
                const uint lc_ETALongTime = (uint.MaxValue - 1);

                bool l_NeedETA = (ComplStatus.Incomplete == l_Status);
                uint l_SecsToGo;
                string l_ETAStr = string.Empty;
                uint l_SourcesValue;
                string l_SourcesStr;
                bool l_HasComplSource;
                Speed l_DlSpeed;
                Speed l_UlSpeed;
                float l_AvailRatio;
                uint l_SecsActive = l_Known.GetBaseSecsActive();

                if (false == l_IsActive)
                {
                    l_SecsToGo = (false == l_NeedETA ? lc_ETANotApplicable : lc_ETALongTime);
                    l_SourcesValue = 0;
                    l_SourcesStr = string.Empty;
                    l_HasComplSource = false;
                    l_DlSpeed = Speed.s_NoSpeed;
                    l_UlSpeed = Speed.s_NoSpeed;
                    l_AvailRatio = 0f;
                }
                else
                {
                    uint l_CompleteNodes;
                    uint l_PartialNodes;
                    l_ActCntxt.GetFilesPanelAttribsVolatile(a_Time, out l_SecsToGo, out l_DlSpeed, out l_UlSpeed, out l_CompleteNodes, out l_PartialNodes, out l_AvailRatio);

                    if (false == l_NeedETA)
                    {
                        l_SecsToGo = lc_ETANotApplicable;
                    }
                    else
                    {
                        if (false == ArrivalTimeEstimator.IsUnknownTime(l_SecsToGo))
                        {
                            if ((30 * Time.c_SecsPerDay) < l_SecsToGo)
                            {
                                l_ETAStr = TimeStrHelper.s_LongerThanMonthStr;
                            }
                            else
                            {
                                uint l_Rem = (l_SecsToGo % 60);
                                ulong l_SecsToGoRoundedUp = ((ulong)l_SecsToGo + (0 == l_Rem ? 0u : (60 - l_Rem)));
                                l_ETAStr = l_TextSet.TimeStr((long)l_SecsToGoRoundedUp);
                            }
                        }
                        else
                        {
                            l_SecsToGo = lc_ETALongTime;
                        }
                    }

                    l_SourcesValue = ((l_CompleteNodes << Util.c_BitsPerByte) | l_PartialNodes);
                    l_SourcesStr = l_TextSet.Format(FilesPanel.s_SourcesShortFormat, l_CompleteNodes, l_PartialNodes);
                    l_HasComplSource = (0 != l_CompleteNodes);
                    l_SecsActive += l_ActCntxt.GetSecsActive(a_Time);
                }

                a_Item.SetText(7, l_SecsToGo, l_ETAStr);
                a_Item.SetText(8, l_SourcesValue, l_SourcesStr);
                a_Item.SetText(9, l_DlSpeed.p_BytesPerSec, l_DlSpeed.ToGUIString(l_SISizes));
                a_Item.SetText(10, l_UlSpeed.p_BytesPerSec, l_UlSpeed.ToGUIString(l_SISizes));

                // Transfer volumes.

                TransferVolume l_TransferVol;
                l_Known.CopyTransferVolumeVolatile(out l_TransferVol);

                a_Item.SetText(11, l_TransferVol.p_BytesReceived, l_Form.BuildDefaultSizeStr(l_TransferVol.p_BytesReceived));
                a_Item.SetText(12, l_TransferVol.p_BytesSent, l_Form.BuildDefaultSizeStr(l_TransferVol.p_BytesSent));

                float l_Ratio = ((float)l_TransferVol.p_BytesSent / l_Known.f_FileSize);
                float l_UlDlRatio = l_TransferVol.GetUlDlRatio();

                a_Item.SetText(13, l_Ratio, l_TextSet.FloatStr(l_Ratio));
                a_Item.SetText(14, l_UlDlRatio, SizeFormatter.BuildUlDlRatioStr(l_UlDlRatio));
                a_Item.SetText(15, l_AvailRatio, false == l_IsActive ? string.Empty : l_TextSet.PercStr(l_AvailRatio));
                a_Item.SetText(16, l_SecsActive, l_TextSet.TimeStr((long)l_SecsActive));
                a_Item.SetTime(17, l_TextSet, a_Time, l_Known.f_AddedOn);

                bool l_HandledLastSeenTime = false;
                bool l_HandledLastRcvdTime = false;

                if (true == l_IsActive)
                {
                    if (true == l_HasComplSource)
                    {
                        a_Item.SetText(18, a_Time.ToBinary(), TimeStrHelper.s_NowStr);
                        l_HandledLastSeenTime = true;
                    }

                    if (true == l_ActCntxt.IsDownloading())
                    {
                        a_Item.SetText(19, a_Time.ToBinary(), TimeStrHelper.s_NowStr);
                        l_HandledLastRcvdTime = true;
                    }
                }

                if (false == l_HandledLastSeenTime)
                {
                    a_Item.SetTime(18, l_TextSet, a_Time, l_Known.f_LastSeenComplete, TimeStrHelper.s_NeverStr);
                }

                if (false == l_HandledLastRcvdTime)
                {
                    a_Item.SetTime(19, l_TextSet, a_Time, l_Known.f_LastReception, TimeStrHelper.s_NeverStr);
                }
            }

            private enum FileStatus : byte
            {
                // Compatible with the DownloadControl enumeration.

                IncomplDisabled = DownloadControl.DlDisabled,
                IncomplEnabled = DownloadControl.DlEnabled,
                Missing = 2,
                Finished = 3,
            }
        }

        sealed private class FileSpeedAndPriorityDialog : ButtonDialog
        {
            private const ushort c_DialogWidth = 350;

            private readonly Instance f_Inst;
            private readonly KnownFile[] f_KnownFiles;

            private readonly GuiGroupBox f_DlGroupBox;
            private readonly GuiGroupBox f_UlGroupBox;
            private readonly BandwidthLimitControls f_DlSpeedLimit;
            private readonly BandwidthLimitControls f_UlSpeedLimit;
            private readonly PriorityPickerControl f_DlPriority;
            private readonly PriorityPickerControl f_UlPriority;

            public FileSpeedAndPriorityDialog(Instance a_Inst,
                                              KnownFile[] a_KnownFiles) :
                                         base(a_Inst.f_TextSet, DialogButtons.OKCancel)
            {
                f_Inst = a_Inst;
                f_KnownFiles = a_KnownFiles;

                this.SuspendLayout();

                TextSet l_TextSet = a_Inst.f_TextSet;
                bool l_SISizes = a_Inst.f_MainForm.p_DisplaySISizes;
                int l_DlgWidth = Lyt.CalcX(c_DialogWidth);
                f_DlGroupBox = new GuiGroupBox(this);
                f_UlGroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_DlgWidth - (5 * Lyt.s_Space));
                // Use 60 percent for the bandwidth limit controls.
                int l_LeftWidth = ((l_InnerWidth / 100) * 60);
                int l_RightWidth = (l_InnerWidth - l_LeftWidth);
                f_DlSpeedLimit = new BandwidthLimitControls(l_TextSet, f_DlGroupBox, l_LeftWidth, l_SISizes);
                f_UlSpeedLimit = new BandwidthLimitControls(l_TextSet, f_UlGroupBox, l_LeftWidth, l_SISizes);
                f_DlPriority = new PriorityPickerControl(f_DlGroupBox, l_RightWidth);
                f_UlPriority = new PriorityPickerControl(f_UlGroupBox, l_RightWidth);

                f_DlGroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                Point l_LimitLoc = new Point(f_DlGroupBox.Padding.Left, f_DlGroupBox.Padding.Top);
                f_DlSpeedLimit.SetLocation(l_LimitLoc);
                Point l_PriorityLoc = new Point(f_DlSpeedLimit.GetRight(), f_DlGroupBox.Padding.Top);
                f_DlPriority.SetLocation(l_PriorityLoc);
                f_DlGroupBox.Size = new Size(l_DlgWidth - Lyt.s_Space2, f_DlSpeedLimit.GetBottom());

                f_UlGroupBox.Location = new Point(Lyt.s_Space, f_DlGroupBox.GetBottom());
                f_UlSpeedLimit.SetLocation(l_LimitLoc);
                f_UlPriority.SetLocation(l_PriorityLoc);
                f_UlGroupBox.Size = f_DlGroupBox.Size;

                AsFixedButtonDialog(l_DlgWidth, f_UlGroupBox);

                f_OKButton.Click += OKButton_Click;

                this.Text = l_TextSet[TextKeys.FilesTabBwLimitDlgTitle];
                f_DlGroupBox.Text = l_TextSet[TextKeys.FilesTabBwLimitDlgDlCaption];
                f_UlGroupBox.Text = l_TextSet[TextKeys.FilesTabBwLimitDlgUlCaption];
                f_DlSpeedLimit.SetText(l_TextSet[TextKeys.FilesTabBwLimitDlgDlLimitSpeed]);
                f_UlSpeedLimit.SetText(l_TextSet[TextKeys.FilesTabBwLimitDlgUlLimitSpeed]);
                f_DlPriority.SetCaption(l_TextSet, TextKeys.FilesTabBwLimitDlgDlPriority);
                f_UlPriority.SetCaption(l_TextSet, TextKeys.FilesTabBwLimitDlgUlPriority);

                this.ResumeLayout();

                FilesToGui();
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    Dialog.DisposeAllChildren(f_DlGroupBox);
                    Dialog.DisposeAllChildren(f_UlGroupBox);
                    Dialog.DisposeAllChildren(this);
                }

                base.Dispose(disposing);
            }

            public static void AnalyzeFiles(KnownFile[] a_Files,
                                            out Speed a_CommonDlLimit,
                                            out Speed a_CommonUlLimit,
                                            out Priority a_CommonDlPriority,
                                            out Priority a_CommonUlPriority)
            {
                a_CommonDlLimit = Speed.s_UnknownSpeed;
                a_CommonUlLimit = Speed.s_UnknownSpeed;
                a_CommonDlPriority = new Priority();
                a_CommonUlPriority = new Priority();

                int l_Idx = 0;
                foreach (KnownFile l_Known in a_Files)
                {
                    Speed l_DlLimit = l_Known.f_DlLimit;
                    Speed l_UlLimit = l_Known.f_UlLimit;
                    Priority l_DlPriority = l_Known.f_DlPriority;
                    Priority l_UlPriority = l_Known.f_UlPriority;

                    if (0 == l_Idx)
                    {
                        a_CommonDlLimit = l_DlLimit;
                        a_CommonUlLimit = l_UlLimit;
                        a_CommonDlPriority = l_DlPriority;
                        a_CommonUlPriority = l_UlPriority;
                    }
                    else
                    {
                        if (a_CommonDlLimit != l_DlLimit) a_CommonDlLimit = Speed.s_UnknownSpeed;
                        if (a_CommonUlLimit != l_UlLimit) a_CommonUlLimit = Speed.s_UnknownSpeed;
                        if (a_CommonDlPriority != l_DlPriority) a_CommonDlPriority = new Priority();
                        if (a_CommonUlPriority != l_UlPriority) a_CommonUlPriority = new Priority();
                    }

                    l_Idx++;

                    if (false == a_CommonDlLimit.IsValid() &&
                        false == a_CommonUlLimit.IsValid() &&
                        false == a_CommonDlPriority.IsValid() &&
                        false == a_CommonUlPriority.IsValid())
                    {
                        break;
                    }
                }
            }

            private void FilesToGui()
            {
                Speed l_CommonDlLimit;
                Speed l_CommonUlLimit;
                Priority l_CommonDlPriority;
                Priority l_CommonUlPriority;
                AnalyzeFiles(f_KnownFiles, out l_CommonDlLimit, out l_CommonUlLimit, out l_CommonDlPriority, out l_CommonUlPriority);

                f_DlSpeedLimit.SetLimit(l_CommonDlLimit);
                f_UlSpeedLimit.SetLimit(l_CommonUlLimit);
                f_DlPriority.SetPriority(l_CommonDlPriority);
                f_UlPriority.SetPriority(l_CommonUlPriority);
            }

            private bool GuiToFiles(out string a_ErrMsg,
                                    out ArglessHandler a_SetFocus)
            {
                Speed l_DlLimit;
                if (false == f_DlSpeedLimit.LoadSpeed(out l_DlLimit, out a_ErrMsg))
                {
                    a_SetFocus = f_DlSpeedLimit.SetFocus;

                    return false;
                }

                Speed l_UlLimit;
                if (false == f_UlSpeedLimit.LoadSpeed(out l_UlLimit, out a_ErrMsg))
                {
                    a_SetFocus = f_UlSpeedLimit.SetFocus;

                    return false;
                }

                Priority l_DlPriority;
                if (false == f_DlPriority.GetPriority(out l_DlPriority, out a_ErrMsg))
                {
                    a_SetFocus = f_DlPriority.SetFocus;

                    return false;
                }

                Priority l_UlPriority;
                if (false == f_UlPriority.GetPriority(out l_UlPriority, out a_ErrMsg))
                {
                    a_SetFocus = f_UlPriority.SetFocus;

                    return false;
                }

                foreach (KnownFile l_Known in f_KnownFiles)
                {
                    if (true == l_DlLimit.IsValid())
                    {
                        l_Known.f_DlLimit = l_DlLimit;
                    }

                    if (true == l_UlLimit.IsValid())
                    {
                        l_Known.f_UlLimit = l_UlLimit;
                    }

                    if (true == l_DlPriority.IsValid())
                    {
                        l_Known.f_DlPriority = l_DlPriority;
                    }

                    if (true == l_UlPriority.IsValid())
                    {
                        l_Known.f_UlPriority = l_UlPriority;
                    }
                }

                a_SetFocus = null;

                return true;
            }

            private void OKButton_Click(object sender, EventArgs e)
            {
                string l_ErrMsg;
                ArglessHandler l_SetFocus;
                if (false == GuiToFiles(out l_ErrMsg, out l_SetFocus))
                {
                    f_Inst.f_MainForm.WarnMsgBox(l_ErrMsg);
                    l_SetFocus();

                    return;
                }

                this.Close();
            }
        }

        private class DetailsPanel : BottomTabPage
        {
            private const int c_NoComplPicIndex = (-1);

            private static readonly Color s_PicBgColor = Color.White;

            private readonly Instance f_Inst;
            private readonly BarDrawer f_BarDrawer;
            private readonly FilesPanel f_FilesPanel;

            private readonly GuiToolTip f_ToolTip;
            private readonly GuiLabel f_AvailLab;
            private readonly GuiLabel f_ComplLab;
            private readonly GuiPictureBox f_AvailPic;
            private readonly GuiPictureBox f_ComplPic;
            private readonly GuiLabel f_AvailValLab;
            private readonly GuiLabel f_ComplValLab;

            private readonly Panel f_TextGroupPanel;
            private readonly TextGroup f_FileName;
            private readonly TextGroup f_Location;
            private readonly TextGroup f_FileSize;
            private readonly TextGroup f_eD2kHash;
            private readonly TextGroup f_AICHHash;
            private readonly TextGroup f_TAICHHash;
            private readonly TextGroup f_Sources;

            private readonly GuiButton f_MoveToDirButton;
            private readonly ContextMenuStripEx f_MoveToDirCntxtMenu;
            private readonly Font f_ItalicMenuFont;
            private FileNum f_MenuFileNum;

            private CompletionPicMode f_ComplPicMode;
            private int f_ComplPicIndex;
            // f_BarBgImages can be null.
            private BarBackgroundImages f_BarBgImages;
            private KnownFile f_Known;

            public DetailsPanel(Instance a_Inst,
                                FilesPanel a_FilesPanel)
            {
                f_Inst = a_Inst;
                f_BarDrawer = f_Inst.f_MainForm.f_BarDrawer;
                f_FilesPanel = a_FilesPanel;

                f_ToolTip = new GuiToolTip();
                f_AvailLab = new GuiLabel(this);
                f_ComplLab = new GuiLabel(this);
                f_AvailPic = new GuiPictureBox(this, s_PicBgColor);
                f_ComplPic = new GuiPictureBox(this, s_PicBgColor);
                f_AvailValLab = new GuiLabel(this);
                f_ComplValLab = new GuiLabel(this);

                int l_Height = Math.Max(Lyt.CalcY(18), f_AvailLab.Height);
                f_AvailLab.TextAlign = ContentAlignment.MiddleRight;
                f_AvailLab.Location = new Point(Lyt.s_SmallSpace, Lyt.s_SmallSpace);
                f_AvailLab.Size = new Size(TextGroup.s_DefLeftWidth, l_Height);
                f_AvailPic.SizeMode = PictureBoxSizeMode.StretchImage;
                f_AvailPic.BorderStyle = BorderStyle.Fixed3D;
                f_AvailPic.Location = new Point(f_AvailLab.Right + Lyt.s_SmallSpace, Lyt.s_SmallSpace);
                f_AvailPic.Height = l_Height;
                f_AvailValLab.TextAlign = ContentAlignment.MiddleLeft;
                f_AvailValLab.Location = new Point(f_AvailPic.Right + Lyt.s_SmallSpace, Lyt.s_SmallSpace);
                f_AvailValLab.Size = new Size(Lyt.CalcX(50), l_Height);

                f_ComplLab.TextAlign = ContentAlignment.MiddleRight;
                f_ComplLab.Location = new Point(f_AvailLab.Left, f_AvailLab.Bottom + Lyt.s_SmallSpace);
                f_ComplLab.Size = f_AvailLab.Size;
                f_ComplPic.SizeMode = PictureBoxSizeMode.StretchImage;
                f_ComplPic.BorderStyle = BorderStyle.Fixed3D;
                f_ComplPic.Location = new Point(f_AvailPic.Left, f_AvailPic.Bottom + Lyt.s_SmallSpace);
                f_ComplPic.Size = f_AvailPic.Size;
                f_ComplValLab.TextAlign = ContentAlignment.MiddleLeft;
                f_ComplValLab.Location = new Point(f_AvailValLab.Left, f_AvailValLab.Bottom + Lyt.s_SmallSpace);
                f_ComplValLab.Size = f_AvailValLab.Size;

                f_TextGroupPanel = new GuiPanel(this);
                f_FileName = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, 0);
                f_Location = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_FileName.GetBottom());
                f_FileSize = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_Location.GetBottom());
                f_eD2kHash = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_FileSize.GetBottom());
                f_AICHHash = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_eD2kHash.GetBottom());
                f_TAICHHash = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_AICHHash.GetBottom());
                f_Sources = TextGroup.CreateLargeRight(f_TextGroupPanel, Lyt.s_SmallSpace, f_TAICHHash.GetBottom());
                f_TextGroupPanel.Size = new Size(0, f_Sources.GetBottom());
                f_MoveToDirButton = new GuiButton(f_TextGroupPanel, true);
                f_MoveToDirButton.Enabled = false;
                f_MoveToDirCntxtMenu = ContextMenuStripEx.CreateUnassociated(false);
                f_ItalicMenuFont = MainForm.RecreateFont(f_MoveToDirCntxtMenu.Font, 0, FontStyle.Italic);

                f_MoveToDirButton.MouseDown += MoveToDirButton_MouseDown;

                TextSet l_TextSet = f_Inst.f_TextSet;
                f_AvailLab.Text = l_TextSet.GetAsKeyPortion(TextKeys.FilesTabDetailsAvailability);
                f_ComplLab.Text = l_TextSet.GetAsKeyPortion(TextKeys.FilesTabDetailsDone);
                f_FileName.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsFileName);
                f_Location.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsLocation);
                f_FileSize.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsFileSize);
                f_eD2kHash.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsFileID);
                f_AICHHash.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsAICHHash);
                f_TAICHHash.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsTAICHHash);
                f_Sources.SetLeftText(l_TextSet, TextKeys.FilesTabDetailsSources);
                f_MoveToDirButton.BringToFront();
                f_MoveToDirButton.Top = (f_Sources.GetBottom() - Lyt.s_SmallSpace - f_MoveToDirButton.Height);
                f_MoveToDirButton.Text = "Move to Directory";

                f_ComplPicIndex = c_NoComplPicIndex;
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    // We don't use TextGroup.Dispose() for this, since
                    // we don't have to.

                    f_ToolTip.Dispose();
                    Dialog.DisposeAllChildren(f_TextGroupPanel);
                    Dialog.DisposeAllChildren(this);
                    f_MoveToDirCntxtMenu.Dispose();
                    f_ItalicMenuFont.Dispose();

                    if (null != f_BarBgImages)
                    {
                        f_BarBgImages.Dispose();
                        f_BarBgImages = null;
                    }
                }

                base.Dispose(disposing);
            }

            public override bool HasInputFocus()
            {
                // None of our child controls can have the focus.

                return false;
            }

            public bool HasBarBackgroundImages()
            {
                return (null != f_BarBgImages);
            }

            public void SetBarBackgroundImages(BarBackgroundImages a_BarBgImages)
            {
                // a_BarBgImages may be null to disable background
                // images. If it's not null, this instance takes
                // ownership.

                if (a_BarBgImages == f_BarBgImages)
                {
                    return;
                }

                if (null != f_BarBgImages)
                {
                    f_BarBgImages.Dispose();
                }

                f_ComplPicMode = CompletionPicMode.Undefined;
                f_ComplPicIndex = c_NoComplPicIndex;
                f_BarBgImages = a_BarBgImages;

                UpdateSizes();
            }

            protected override void OnResize(EventArgs eventargs)
            {
                // Need to call base class implementation.

                UpdateSizes();

                base.OnResize(eventargs);
            }

            private void UpdateSizes()
            {
                f_FilesPanel.SuspendLayout();

                int l_Width = this.Width;

                f_AvailValLab.Left = (l_Width - f_AvailValLab.Width);
                f_ComplValLab.Left = f_AvailValLab.Left;
                f_AvailPic.Width = (f_AvailValLab.Left - f_AvailPic.Left - Lyt.s_SmallSpace);
                f_ComplPic.Size = new Size(f_AvailPic.Width, null == f_BarBgImages ? f_AvailPic.Height : (f_AvailPic.Width / BarBackgroundImages.c_AspectDivisor));
                f_TextGroupPanel.Top = (f_ComplPic.Bottom + Lyt.s_SmallSpace);
                f_TextGroupPanel.Width = l_Width;

                bool l_ShowButton = ((f_AvailPic.Width - Lyt.s_LargeSpace2) >= f_MoveToDirButton.Width);
                f_MoveToDirButton.Visible = l_ShowButton;

                if (true == l_ShowButton)
                {
                    f_MoveToDirButton.Left = (f_AvailPic.Right - f_MoveToDirButton.Width);
                }

                f_FilesPanel.BottomPanelMinHeightChanged(f_TextGroupPanel.Bottom);
                f_FilesPanel.ResumeLayout();
            }

            public void ClearDetails()
            {
                if (null == f_Known)
                {
                    return;
                }

                NotifySelectedFileChanged(null);

                f_AvailPic.Clear();
                SetComplPicSimple(CompletionPicMode.Cleared, null);

                f_ToolTip.Active = false;
                f_AvailValLab.Text = string.Empty;
                f_ComplValLab.Text = string.Empty;

                f_FileName.ClearRightText();
                f_Location.ClearRightText();
                f_FileSize.ClearRightText();
                f_eD2kHash.ClearRightText();
                f_AICHHash.ClearRightText();
                f_TAICHHash.ClearRightText();
                f_Sources.ClearRightText();
                f_MoveToDirButton.Enabled = false;
            }

            public void Update(ref KnownFileUpdateContext a_UpdateCntxt)
            {
                TextSet l_TextSet = f_Inst.f_TextSet;
                MainForm l_Form = f_Inst.f_MainForm;
                KnownFile l_Known = a_UpdateCntxt.f_Known;

                if (f_Known != l_Known)
                {
                    NotifySelectedFileChanged(l_Known);
                }

                f_ToolTip.Active = true;

                uint l_CompleteNodes;
                uint l_PartialNodes;
                NetworkAvailability l_Avail;
                float l_AvailRatio;

                if (false == a_UpdateCntxt.f_IsActive)
                {
                    l_CompleteNodes = 0;
                    l_PartialNodes = 0;
                    l_Avail = null;
                    l_AvailRatio = 0;
                }
                else
                {
                    a_UpdateCntxt.f_ActCntxt.GetDetailsPanelAttribsVolatile(out l_CompleteNodes, out l_PartialNodes, out l_Avail, out l_AvailRatio);
                }

                if (null == l_Avail ||
                    true == l_Avail.IsEmpty())
                {
                    f_AvailPic.Clear();
                    f_AvailValLab.Text = FilesPanel.s_UnknownValueStr;
                    f_ToolTip.SetToolTip(f_AvailValLab, null);
                }
                else
                {
                    Bitmap l_AvailImg = f_BarDrawer.DrawAvailabilityBarImageScaled(l_Avail, f_AvailPic.Height);

                    f_AvailPic.SetPicture(l_AvailImg);
                    f_AvailValLab.Text = l_TextSet.PercStr(l_AvailRatio);
                    f_ToolTip.SetToolTip(f_AvailValLab, string.Format("The currently connected peers can together provide {0:P1} of the file.", l_AvailRatio));
                }

                ComplStatus l_Status;
                string l_ParentPath;
                string l_FileName = l_Known.GetFileNameVolatile(true, out l_Status, out l_ParentPath);
                string l_ComplValLabText;

                f_MoveToDirButton.Enabled = (ComplStatus.Complete == l_Status);

                if (ComplStatus.Incomplete == l_Status)
                {
                    l_ParentPath = FilesPanel.s_IncompleteLocStr;

                    IncomplFileProgress l_IncomplProgress = l_Known.p_IncomplProgressVolatile;
                    ulong l_Have;
                    ulong l_HaveInclUnver;
                    float l_ComplPerc;
                    float l_ComplPercInclUnver;

                    if (null == l_IncomplProgress)
                    {
                        SetComplPicSimple(CompletionPicMode.NothingDone, l_Known);
                        l_ComplValLabText = l_TextSet.PercStr(0f);
                        l_Have = l_HaveInclUnver = 0;
                        l_ComplPerc = l_ComplPercInclUnver = 0f;
                    }
                    else
                    {
                        f_ComplPicMode = CompletionPicMode.Structured;
                        Bitmap l_ComplImg;

                        if (null == f_BarBgImages)
                        {
                            l_ComplImg = f_BarDrawer.DrawCompletenessBarImageScaled(l_IncomplProgress.f_DoneMap, l_IncomplProgress.f_HotIncomplChunks, f_ComplPic.Height);
                        }
                        else
                        {
                            // First update the index.
                            f_ComplPicIndex = f_BarBgImages.GetBitmapIndex(l_Known);

                            l_ComplImg = f_BarDrawer.DrawCompletenessBarImageScaledOnBackground(l_IncomplProgress.f_DoneMap, l_IncomplProgress.f_HotIncomplChunks, f_ComplPic.Height, f_BarBgImages.GetBitmap(f_ComplPicIndex));
                        }

                        f_ComplPic.SetPicture(l_ComplImg);

                        l_Have = l_IncomplProgress.f_CompletedBytes;
                        l_HaveInclUnver = l_IncomplProgress.f_CompletedBytesInclUnverified;
                        l_ComplPerc = IncomplFileProgress.CalcQuotient(l_Have, l_Known.f_FileSize);
                        l_ComplPercInclUnver = IncomplFileProgress.CalcQuotient(l_HaveInclUnver, l_Known.f_FileSize);

                        float l_AchievedRatio = (GeneralPrefs.ProgessBarKinds.VerifiedOnly == l_Form.p_Prefs.f_GeneralPrefs.f_ProgressBarKind ? l_ComplPerc : l_ComplPercInclUnver);
                        l_ComplValLabText = l_TextSet.PercStr(l_AchievedRatio);
                    }

                    f_ToolTip.SetToolTip(f_ComplValLab, l_TextSet.Format("Share of total size present: {0:P1} ({1})\nShare of total size verified: {2:P1} ({3})", l_ComplPercInclUnver, l_Form.BuildDefaultSizeStr(l_HaveInclUnver), l_ComplPerc, l_Form.BuildDefaultSizeStr(l_Have)));
                }
                else
                {
                    if (ComplStatus.Complete == l_Status)
                    {
                        SetComplPicSimple(CompletionPicMode.AllDone, l_Known);
                        l_ComplValLabText = l_TextSet.PercStr(1f);
                    }
                    else
                    {
                        // Missing or dropped.

                        SetComplPicSimple(CompletionPicMode.NothingDone, l_Known);
                        l_ComplValLabText = FilesPanel.s_UnknownValueStr;
                        l_ParentPath = FilesPanel.s_MissingLocStr;
                    }

                    f_ToolTip.SetToolTip(f_ComplValLab, null);
                }

                f_ComplValLab.Text = l_ComplValLabText;

                MasterHashes l_MasterHashes;
                MasterHashes.CloneVolatile(ref l_Known.f_MasterHashes, out l_MasterHashes);

                f_FileName.SetRightText(l_FileName);
                f_Location.SetRightText(l_ParentPath);
                f_FileSize.SetRightText( f_Inst.f_MainForm.BuildLongSizeStr(l_Known.f_FileSize));
                f_eD2kHash.SetRightText(l_MasterHashes.eD2kHashAsStr(FilesPanel.s_UnknownValueStr));
                f_AICHHash.SetRightText(l_MasterHashes.AICHHashAsStr(FilesPanel.s_UnknownValueStr));
                f_TAICHHash.SetRightText(l_MasterHashes.TAICHHashAsStr(FilesPanel.s_UnknownValueStr));

                string l_SourcesStr;

                if (0 == l_CompleteNodes &&
                    0 == l_PartialNodes)
                {
                    l_SourcesStr = FilesPanel.s_NoSourcesStr;
                }
                else
                {
                    l_SourcesStr = l_TextSet.Format(FilesPanel.s_SourcesLongFormat, l_CompleteNodes, l_PartialNodes);
                }

                f_Sources.SetRightText(l_SourcesStr);
            }

            private void NotifySelectedFileChanged(KnownFile a_Known)
            {
                // a_Known may be null.

                f_Known = a_Known;
                f_MoveToDirCntxtMenu.Close();
            }

            private void SetComplPicSimple(CompletionPicMode a_Mode,
                                           KnownFile a_Known)
            {
                // a_Known may be null if a_Mode is
                // CompletionPicMode.Cleared.

                if (null == f_BarBgImages)
                {
                    if (a_Mode == f_ComplPicMode)
                    {
                        // Nothing to do.

                        return;
                    }

                    f_ComplPicMode = a_Mode;

                    switch (f_ComplPicMode)
                    {
                        case CompletionPicMode.NothingDone:
                            f_ComplPic.DisplayColor(f_BarDrawer.p_SolidClearColor);
                            return;

                        case CompletionPicMode.AllDone:
                            f_ComplPic.DisplayColor(f_BarDrawer.p_SolidDoneColor);
                            return;
                    }
                }
                else
                {
                    int l_BitmapIdx = (CompletionPicMode.Cleared == a_Mode ? c_NoComplPicIndex : f_BarBgImages.GetBitmapIndex(a_Known));

                    if (a_Mode == f_ComplPicMode &&
                        l_BitmapIdx == f_ComplPicIndex)
                    {
                        // Nothing to do.

                        return;
                    }

                    f_ComplPicMode = a_Mode;
                    f_ComplPicIndex = l_BitmapIdx;

                    switch (f_ComplPicMode)
                    {
                        case CompletionPicMode.NothingDone:
                            SetComplPicWithOverlay(f_ComplPicIndex, f_BarDrawer.p_OverlayClearBrush);
                            return;

                        case CompletionPicMode.AllDone:
                            SetComplPicWithOverlay(f_ComplPicIndex, f_BarDrawer.p_OverlayDoneBrush);
                            return;
                    }
                }

                f_ComplPic.Clear();
            }

            private void SetComplPicWithOverlay(int a_BitmapIdx,
                                                Brush a_OverlayBrush)
            {
                // Should only get called by SetComplPicSimple().
                // a_OverlayBrush may be null.

                Bitmap l_MapClone = IconStorage.CloneBitmap(f_BarBgImages.GetBitmap(a_BitmapIdx));

                if (null != a_OverlayBrush)
                {
                    Graphics l_G = Graphics.FromImage(l_MapClone);
                    {
                        l_G.FillRectangle(a_OverlayBrush, 0, 0, l_MapClone.Width, l_MapClone.Height);
                    }
                    l_G.Dispose();
                }

                f_ComplPic.SetPicture(l_MapClone);
            }

            private void MoveToDirButton_MouseDown(object a_Sender,
                                                   MouseEventArgs e)
            {
                f_MoveToDirCntxtMenu.Close();
                f_MenuFileNum = new FileNum();
                KnownFile l_Known = f_Known;

                if (MouseButtons.Left != e.Button ||
                    null == l_Known)
                {
                    return;
                }

                ComplStatus l_Status;
                string l_ContainingDirPath;
                string l_Junk = l_Known.GetFileNameVolatile(true, out l_Status, out l_ContainingDirPath);

                if (ComplStatus.Complete != l_Status ||
                    null == l_ContainingDirPath)
                {
                    return;
                }

                string[] l_Subdirs = null;

                try
                {
                    l_Subdirs = System.IO.Directory.GetDirectories(l_ContainingDirPath);

                    if (0 >= l_Subdirs.Length)
                    {
                        l_Subdirs = null;
                    }
                }
                catch (ArgumentException)
                {
                }
                catch (System.IO.IOException)
                {
                }
                catch (UnauthorizedAccessException)
                {
                }

                ContextMenuStripEx l_Menu = f_MoveToDirCntxtMenu;
                l_Menu.DisposeItems();

                bool l_IncludeUnsharedDirs = true;
                bool l_IncludeParentDir = !LongPath.IsRoot(l_ContainingDirPath);
                SharedDirectories l_SharedDirs = f_Inst.f_MainForm.p_Prefs.f_IOPrefs.f_SharedDirs;

                if (true == l_IncludeParentDir)
                {
                    string l_ParentDirPath = LongPath.GetDirectoryName(l_ContainingDirPath);

                    if (false == l_IncludeUnsharedDirs &&
                        ShareKinds.None == l_SharedDirs.GetShareKind(l_ParentDirPath))
                    {
                        l_IncludeParentDir = false;
                    }
                    else
                    {
                        string l_Text = string.Format(Util.s_InvariantCulture, "<Parent Directory> ({0})", l_ParentDirPath);
                        GenericMenuItem<string> l_Item = GenericMenuItem<string>.CreateItem(l_Menu, l_Text, Menu_MoveFileToDirectoryClicked, l_ParentDirPath);
                        l_Item.Font = f_ItalicMenuFont;
                    }
                }

                if (null != l_Subdirs)
                {
                    const byte lc_MaxDisplayedSubdirs = 64;
                    bool l_NeedToAddSep = l_IncludeParentDir;
                    uint l_EffCount = 0;

                    foreach (string l_SubdirPath in l_Subdirs)
                    {
                        if (false == l_IncludeUnsharedDirs &&
                            ShareKinds.None == l_SharedDirs.GetShareKind(l_SubdirPath))
                        {
                            continue;
                        }

                        if (true == l_NeedToAddSep)
                        {
                            l_NeedToAddSep = false;
                            MainForm.AddMenuSeparator(l_Menu);
                        }

                        string l_SubdirName = LongPath.SplitFullPath(l_SubdirPath, out l_Junk);
                        GenericMenuItem<string>.CreateItem(l_Menu, l_SubdirName, Menu_MoveFileToDirectoryClicked, l_SubdirPath);

                        if (lc_MaxDisplayedSubdirs <= ++l_EffCount)
                        {
                            break;
                        }
                    }
                }

                if (0 >= l_Menu.Items.Count)
                {
                    f_Inst.f_MainForm.InfoMsgBox("No suitable directories detected to move the selected file into.");

                    return;
                }

                f_MenuFileNum = l_Known.f_FileNum;

                Point l_Pos = f_MoveToDirButton.PointToClient(Form.MousePosition);
                f_MoveToDirCntxtMenu.Show(f_MoveToDirButton, l_Pos, ToolStripDropDownDirection.AboveLeft);
            }

            private void Menu_MoveFileToDirectoryClicked(object a_Sender,
                                                         EventArgs e)
            {
                FileNum l_FileNum = f_MenuFileNum;
                KnownFile l_Known = f_Known;

                f_MenuFileNum = new FileNum();

                if (null == l_Known ||
                    0 == l_FileNum.f_Val)
                {
                    return;
                }

                if (l_Known.f_FileNum != l_FileNum)
                {
                    f_Inst.f_MainForm.ShowOperationFailedMsg();

                    return;
                }

                GenericMenuItem<string> l_Item = (GenericMenuItem<string>)a_Sender;
                string l_TargetDirPath = l_Item.f_GenericTag;

                if (true == System.IO.File.Exists(LongPath.Combine(l_TargetDirPath, l_Known.GetFileNameVolatile(true))))
                {
                    f_Inst.f_MainForm.ShowOperationFailedMsg();

                    return;
                }

                f_FilesPanel.TryToMoveFilesWithMessage(new KnownFile[1] { l_Known }, l_TargetDirPath);
            }

            private enum CompletionPicMode : byte
            {
                Undefined = 0,

                // Cleared is used when no file is selected, or multiple files.
                Cleared = 1,
                NothingDone = 2,
                AllDone = 3,
                // Structured is used if some portions are done, some not.
                Structured = 4,
            }
        }

        private class FilePeersPanel : BottomTabPage
        {
            private readonly Instance f_Inst;

            private readonly GridView<I2PAddress, PeerData> f_GridView;

            public FilePeersPanel(Instance a_Inst)
            {
                f_Inst = a_Inst;

                f_GridView = new GridView<I2PAddress, PeerData>(f_Inst, this, f_Inst.f_MainForm.f_VisualCust.f_FilePeersPanelColSet);

                f_GridView.e_UpdateItem += GridView_UpdateItem;
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    f_GridView.Dispose();
                }

                base.Dispose(disposing);
            }

            public override bool HasInputFocus()
            {
                return f_GridView.Focused;
            }

            public void FinishLoading()
            {
                f_GridView.LoadColumnAdjustments();
            }

            public void PrepareSaving()
            {
                f_GridView.StoreColumnAdjustments();
            }

            public void Clear()
            {
                f_GridView.ClearItems();
            }

            public void Update(ref KnownFileUpdateContext a_UpdateCntxt)
            {
                if (false == a_UpdateCntxt.f_HasPeers)
                {
                    f_GridView.ClearItems();

                    return;
                }

                f_GridView.StartChanging();

                if (null != a_UpdateCntxt.f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in a_UpdateCntxt.f_DlPeers)
                    {
                        GridViewItem<I2PAddress, PeerData> l_Item = f_GridView.CreateItem(l_DlPeerCntxt.f_Clnt.p_Addr);
                        l_Item.f_Tag.f_Directions = Directions.Down;
                        l_Item.f_Tag.f_PeerCntxt = l_DlPeerCntxt;
                    }
                }

                if (null != a_UpdateCntxt.f_UlPeers)
                {
                    foreach (UlPeerContext l_UlPeerCntxt in a_UpdateCntxt.f_UlPeers)
                    {
                        bool l_AlreadyAdded;
                        GridViewItem<I2PAddress, PeerData> l_Item = f_GridView.CreateItem(l_UlPeerCntxt.f_Clnt.p_Addr, out l_AlreadyAdded);

                        if (false == l_AlreadyAdded)
                        {
                            l_Item.f_Tag.f_Directions = Directions.Up;
                            l_Item.f_Tag.f_PeerCntxt = l_UlPeerCntxt;
                        }
                        else
                        {
                            // If there are both contexts, use the
                            // DlPeerContext, since it usually has the
                            // more up to date completion bitmap.

                            l_Item.f_Tag.f_Directions = Directions.UpAndDown;
                        }
                    }
                }

                f_GridView.EndChanging();
            }

            private void GridView_UpdateItem(Time a_Time,
                                             GridViewItem<I2PAddress, PeerData> a_Item)
            {
                TextSet l_TextSet = f_Inst.f_TextSet;
                MainForm l_Form = f_Inst.f_MainForm;
                PeerContext l_PeerCntxt = a_Item.f_Tag.f_PeerCntxt;
                UserNum l_UserNum = l_PeerCntxt.f_Clnt.p_UserCntxt.f_UserNum;
                I2PAddress l_StreamAddr = l_PeerCntxt.f_Clnt.p_Addr;
                string l_DirectionStr = FilesPanel.s_DirectionStrs[(int)a_Item.f_Tag.f_Directions];

                a_Item.SetText(0, l_UserNum.f_Val, l_TextSet.NumStr(l_UserNum.f_Val));
                a_Item.SetTextWithCompObj(1, l_StreamAddr, l_StreamAddr.ToShortString());
                a_Item.SetText(2, (byte)a_Item.f_Tag.f_Directions, l_DirectionStr);

                DonkeyClient.ShortPeerInfo l_PeerInfo;
                l_PeerCntxt.f_Clnt.GetPeerInfoVolatile(out l_PeerInfo);
                string l_AuthStat = (0 == l_PeerInfo.f_AuthKeyStrength ? FilesPanel.s_SecIdentNoKeyStr : l_TextSet.Format(FilesPanel.s_SecIdentRsaKeyFormat, l_PeerInfo.f_AuthKeyStrength));

                a_Item.SetTextWithCompObj(4, l_PeerInfo.f_UserName, l_PeerInfo.f_UserName);
                a_Item.SetTextWithCompObj(5, l_PeerInfo.f_ClientStr, l_PeerInfo.f_ClientStr);
                a_Item.SetText(6, l_PeerInfo.f_AuthKeyStrength, l_AuthStat);

                string l_DlStatusStr = a_Item.f_Tag.BuildDlStatusStr();
                a_Item.SetText(7, 0, l_DlStatusStr);

                float l_ComplNum = l_PeerCntxt.f_eD2kComplNum;
                OwnerDrawContext l_AvailBarImgDrawCntxt = l_PeerCntxt.f_AvailBarImgDrawCntxt;
                Util.MemoryBarrier();

                a_Item.SetText(8, l_ComplNum, null == l_AvailBarImgDrawCntxt ? string.Empty : (1f <= l_ComplNum ? FilesPanel.s_CompletenessEverythingStr : l_TextSet.Format(FilesPanel.s_CompletenessPercFormat, l_ComplNum)));
                // l_AvailBarImgDrawCntxt can be null.
                a_Item.SetDrawContext(9, l_AvailBarImgDrawCntxt);

                UpdateSpeed(a_Time, f_GridView.p_UpdateGeneration, l_PeerCntxt, a_Item);

                TransferVolume l_ConnTransferVol;
                l_PeerCntxt.CopyTransferVolumeVolatile(out l_ConnTransferVol);

                a_Item.SetText(12, l_ConnTransferVol.p_BytesReceived, l_Form.BuildDefaultSizeStr(l_ConnTransferVol.p_BytesReceived));
                a_Item.SetText(13, l_ConnTransferVol.p_BytesSent, l_Form.BuildDefaultSizeStr(l_ConnTransferVol.p_BytesSent));

                UserContext l_UserCntxt = l_PeerCntxt.f_Clnt.p_UserCntxt;
                TransferVolume l_TotalUserTransferVol;
                string l_UserIDStr;

                if (null == l_UserCntxt)
                {
                    // Should never happen.

                    l_TotalUserTransferVol = new TransferVolume();
                    l_UserIDStr = string.Empty;
                }
                else
                {
                    UserID l_UserID = l_UserCntxt.p_UserID;
                    l_UserCntxt.CopyTransferVolumeVolatile(out l_TotalUserTransferVol);

                    l_UserIDStr = l_UserID.AsStringNullSafe(string.Empty);
                }

                a_Item.SetText(14, l_TotalUserTransferVol.p_BytesReceived, l_Form.BuildDefaultSizeStr(l_TotalUserTransferVol.p_BytesReceived));
                a_Item.SetText(15, l_TotalUserTransferVol.p_BytesSent, l_Form.BuildDefaultSizeStr(l_TotalUserTransferVol.p_BytesSent));
                a_Item.SetTextWithCompObj(3, l_UserIDStr, l_UserIDStr);
            }

            private void UpdateSpeed(Time a_Time,
                                     uint a_Generation,
                                     PeerContext a_PeerCntxt,
                                     GridViewItem<I2PAddress, PeerData> a_Item)
            {
                Speed l_DlSpeed;
                Speed l_UlSpeed;
                a_PeerCntxt.GetSpeed(out l_DlSpeed, out l_UlSpeed);

                ItemContext f_ItemCntxt = (ItemContext)a_Item.Tag;

                if (null != f_ItemCntxt &&
                    f_ItemCntxt.f_Generation == a_Generation)
                {
                    // We've already assigned a value in this round
                    // and now we add up.

                    l_DlSpeed.Add(f_ItemCntxt.f_DlSpeed);
                    l_UlSpeed.Add(f_ItemCntxt.f_UlSpeed);
                }

                bool l_SISizes = f_Inst.f_MainForm.p_DisplaySISizes;
                string l_UlSpeedStr;

                if (true == l_UlSpeed.IsZero() &&
                    true == a_PeerCntxt.f_Clnt.p_IsBusyWriting)
                {
                    // The send buffer is full. Filling it even more
                    // would only make the OS buffer more data, but
                    // the problem is that the I2P router (SAM bridge)
                    // doesn't read any faster.

                    l_UlSpeedStr = FilesPanel.s_I2PBusyStr;
                }
                else
                {
                    l_UlSpeedStr = l_UlSpeed.ToGUIString(l_SISizes);
                }

                a_Item.SetText(10, l_DlSpeed.p_BytesPerSec, l_DlSpeed.ToGUIString(l_SISizes));
                a_Item.SetText(11, l_UlSpeed.p_BytesPerSec, l_UlSpeedStr);

                a_Item.Tag = (typeof(DlPeerContext) == a_PeerCntxt.GetType() ?
                              new ItemContext(a_Generation, l_DlSpeed, l_UlSpeed) : null);
            }

            private class ItemContext
            {
                public readonly uint f_Generation;
                public readonly Speed f_DlSpeed;
                public readonly Speed f_UlSpeed;

                public ItemContext(uint a_Generation,
                                   Speed a_DlSpeed,
                                   Speed a_UlSpeed)
                {
                    f_Generation = a_Generation;
                    f_DlSpeed = a_DlSpeed;
                    f_UlSpeed = a_UlSpeed;
                }
            }

            private enum Directions : byte
            {
                Down = 0,
                Up = 1,
                UpAndDown = 2,
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct PeerData
            {
                public Directions f_Directions;
                public PeerContext f_PeerCntxt;

                public string BuildDlStatusStr()
                {
                    if (Directions.Up == f_Directions)
                    {
                        return string.Empty;
                    }

                    return ((DlPeerContext)f_PeerCntxt).BuildDlStatusStrVolatile();
                }
            }
        }

        class TransfersPanel : BottomTabPage
        {
            private readonly Instance f_Inst;

            private readonly GridView<PeerContext, TransferInfo> f_GridView;

            public TransfersPanel(Instance a_Inst)
            {
                f_Inst = a_Inst;

                f_GridView = new GridView<PeerContext, TransferInfo>(f_Inst, this, f_Inst.f_MainForm.f_VisualCust.f_TransferPanelColSet);

                f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
                f_GridView.e_UpdateItem += GridView_UpdateItem;
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    f_GridView.Dispose();
                }

                base.Dispose(disposing);
            }

            public override bool HasInputFocus()
            {
                return f_GridView.Focused;
            }

            public void FinishLoading()
            {
                f_GridView.LoadColumnAdjustments();
            }

            public void PrepareSaving()
            {
                f_GridView.StoreColumnAdjustments();
            }

            public void Clear()
            {
                f_GridView.ClearItems();
            }

            private void GridView_ItemDoubleClicked(GridViewItem<PeerContext, TransferInfo> a_Item)
            {
                f_Inst.f_MainForm.SelectUserContextOfPeerContext(a_Item.f_Ref);
            }

            public void Update(ref KnownFileUpdateContext a_UpdateCntxt)
            {
                if (false == a_UpdateCntxt.f_HasPeers)
                {
                    f_GridView.ClearItems();

                    return;
                }

                f_GridView.StartChanging();

                if (null != a_UpdateCntxt.f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in a_UpdateCntxt.f_DlPeers)
                    {
                        TransferInfo l_Info;
                        if (true == l_DlPeerCntxt.f_Clnt.GetDlTransferInfo(out l_Info))
                        {
                            GridViewItem<PeerContext, TransferInfo> l_Item = f_GridView.CreateItem(l_DlPeerCntxt);
                            l_Item.f_Tag = l_Info;
                        }
                    }
                }

                if (null != a_UpdateCntxt.f_UlPeers)
                {
                    foreach (UlPeerContext l_UlPeerCntxt in a_UpdateCntxt.f_UlPeers)
                    {
                        TransferInfo l_Info;
                        if (true == l_UlPeerCntxt.f_Clnt.GetUlTransferInfo(out l_Info))
                        {
                            GridViewItem<PeerContext, TransferInfo> l_Item = f_GridView.CreateItem(l_UlPeerCntxt);
                            l_Item.f_Tag = l_Info;
                        }
                    }
                }

                f_GridView.EndChanging();
            }

            private void GridView_UpdateItem(Time a_Time,
                                             GridViewItem<PeerContext, TransferInfo> a_Item)
            {
                MainForm l_Form = f_Inst.f_MainForm;
                TextSet l_TextSet = f_Inst.f_TextSet;
                UserNum l_UserNum = a_Item.f_Ref.f_Clnt.p_UserCntxt.f_UserNum;
                string l_DirStr = FilesPanel.s_DirectionStrs[false == a_Item.f_Tag.f_Outgoing ? 0 : 1];
                string l_OffsetStr = l_TextSet.NumStr(a_Item.f_Tag.f_Range.f_StartingOffset);
                string l_FormStr = (false == a_Item.f_Tag.f_Compressed ? s_TransportStylePlainStr : s_TransportStyleComprStr);

                a_Item.SetText(0, l_UserNum.f_Val, l_TextSet.NumStr(l_UserNum.f_Val));
                a_Item.SetTextWithCompObj(1, a_Item.f_Ref.f_Clnt.p_Addr, a_Item.f_Ref.f_Clnt.p_Addr.ToShortString());
                a_Item.SetText(2, (false == a_Item.f_Tag.f_Outgoing ? 0u : 1u), l_DirStr);
                a_Item.SetText(3, a_Item.f_Tag.f_Range.f_StartingOffset, l_OffsetStr);
                a_Item.SetText(4, a_Item.f_Tag.f_Range.f_Length, l_Form.BuildDefaultSizeStr(a_Item.f_Tag.f_Range.f_Length));
                a_Item.SetText(5, (false == a_Item.f_Tag.f_Compressed ? 0u : 1u), l_FormStr);
                a_Item.SetDrawContext(6, OwnerDrawContext.CreateIncomplFileProgressBarContext(true, a_Item.f_Tag.f_Progress));
            }
        }

        private class FileNamesPanel : BottomTabPage
        {
            private readonly Instance f_Inst;

            // f_Tag is unused (so use the bool type).
            private readonly GridView<string, bool> f_GridView;
            private readonly TextBox f_NameBox;
            private readonly GuiButton f_CleanupBtn;
            private readonly GuiButton f_TakeOverBtn;

            private KnownFile f_Known;

            public FileNamesPanel(Instance a_Inst)
            {
                f_Inst = a_Inst;

                f_GridView = new GridView<string, bool>(f_Inst, this, f_Inst.f_MainForm.f_VisualCust.f_FileNamesPanelColSet);
                f_NameBox = new TextBox();
                f_CleanupBtn = new GuiButton(this);
                f_TakeOverBtn = new GuiButton(this);

                this.Controls.Add(f_NameBox);

                f_GridView.Dock = DockStyle.None;
                f_GridView.MultiSelect = false;
                f_NameBox.MaxLength = FileNames.c_MaxFileNameChars;
                f_NameBox.Left = Lyt.s_SmallSpace;
                Lyt.SetFontAndHeight(f_CleanupBtn, f_CleanupBtn.Font);
                f_TakeOverBtn.Height = f_CleanupBtn.Height;
                f_CleanupBtn.Width = f_TakeOverBtn.Width = Lyt.s_DefaultButtonWidth;

                f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
                f_GridView.e_UpdateItem += GridView_UpdateItem;
                f_TakeOverBtn.Click += TakeOverBtn_Click;
                f_CleanupBtn.Click += CleanupBtn_Click;

                TextSet l_TextSet = f_Inst.f_TextSet;
                f_CleanupBtn.Text = l_TextSet[TextKeys.FilesTabFileNamesCleanup];
                f_TakeOverBtn.Text = l_TextSet[TextKeys.FilesTabFileNamesTakeOver];
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    Dialog.DisposeAllChildren(this);
                }

                base.Dispose(disposing);
            }

            public override bool HasInputFocus()
            {
                return (true == f_GridView.Focused ||
                        true == f_NameBox.Focused ||
                        true == f_CleanupBtn.Focused ||
                        true == f_TakeOverBtn.Focused);
            }

            protected override void OnResize(EventArgs eventargs)
            {
                base.OnResize(eventargs);

                f_NameBox.Top = f_CleanupBtn.Top = f_TakeOverBtn.Top = (this.Height - f_CleanupBtn.Height - Lyt.s_SmallSpace);
                f_GridView.Size = new Size(this.Width, f_NameBox.Top - Lyt.s_Space);
                f_TakeOverBtn.Left = (this.Width - f_TakeOverBtn.Width - Lyt.s_SmallSpace);
                f_CleanupBtn.Left = (f_TakeOverBtn.Left - f_CleanupBtn.Width - Lyt.s_SmallSpace);
                f_NameBox.Width = (f_CleanupBtn.Left - Lyt.s_Space);
            }

            public void FinishLoading()
            {
                f_GridView.LoadColumnAdjustments();
            }

            public void PrepareSaving()
            {
                f_GridView.StoreColumnAdjustments();
            }

            public void Clear()
            {
                f_GridView.ClearItems();
                f_NameBox.Clear();
                f_NameBox.Enabled = false;
                f_TakeOverBtn.Enabled = false;

                f_Known = null;
            }

            public void Update(ref KnownFileUpdateContext a_UpdateCntxt)
            {
                f_GridView.StartChanging();

                if (null != a_UpdateCntxt.f_DlPeers)
                {
                    foreach (DlPeerContext l_DlPeerCntxt in a_UpdateCntxt.f_DlPeers)
                    {
                        string l_FileName = Util.VolatileRead(ref l_DlPeerCntxt.f_PeersFileName);

                        if (null != l_FileName)
                        {
                            GridViewItem<string, bool> l_Item = f_GridView.CreateItem(l_FileName);
                        }
                    }
                }

                f_GridView.EndChanging();

                if (f_Known != a_UpdateCntxt.f_Known)
                {
                    f_Known = a_UpdateCntxt.f_Known;

                    f_NameBox.Text = f_Known.GetFileNameVolatile(true);
                    f_NameBox.Enabled = true;
                    f_TakeOverBtn.Enabled = true;
                }
            }

            private void GridView_ItemDoubleClicked(GridViewItem<string, bool> a_Item)
            {
                f_NameBox.Text = a_Item.f_Ref;
            }

            private void GridView_UpdateItem(Time a_Time, GridViewItem<string, bool> a_Item)
            {
                string l_FileName = a_Item.f_Ref;

                a_Item.SetTextWithCompObj(0, l_FileName, l_FileName);
            }

            private void TakeOverBtn_Click(object sender, EventArgs e)
            {
                if (null == f_Known)
                {
                    return;
                }

                string l_NewFileName = f_NameBox.Text.Trim();

                if (0 == l_NewFileName.Length)
                {
                    f_Inst.f_MainForm.ErrMsgBox(f_Inst[TextKeys.FilesTabFileNamesMsgNeedFileName]);

                    return;
                }

                if (false == FileNames.IsLegalFileName(l_NewFileName) ||
                    true == FileNames.StartsWithDOSDeviceName(l_NewFileName))
                {
                    string l_Msg = f_Inst.f_TextSet.Format(TextKeys.FilesTabFileNamesMsgNameHasIllegalChars, FileNames.ListIllegalFileNameChars());
                    f_Inst.f_MainForm.ErrMsgBox(l_Msg);

                    return;
                }

                GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiRenameFile, new object[] { f_Known, l_NewFileName });
                if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
                {
                    return;
                }

                if (CookieStates.Success != l_Cookie.p_State)
                {
                    f_Inst.f_MainForm.ErrMsgBox(f_Inst[TextKeys.FilesTabFileNamesMsgFailedToRename]);
                }
            }

            private void CleanupBtn_Click(object sender, EventArgs e)
            {
                string l_NewText = Util.FixUTF8Latin1ConversionErrors(f_NameBox.Text);
                l_NewText = FileNames.CleanUpFileName(l_NewText);

                if (true == FileNames.StartsWithDOSDeviceName(l_NewText))
                {
                    l_NewText = l_NewText.Substring(1);
                }

                f_NameBox.Text = (null == l_NewText ? string.Empty : l_NewText);
            }
        }

        private class CommentsPanel : BottomTabPage
        {
            public const byte c_FileRatingCount = 6;

            private readonly Instance f_Inst;

            private readonly GridView<object, DlPeerContext> f_GridView;
            private readonly GuiButton f_AddOwnBtn;
            private readonly GuiButton f_EditFilterBtn;
            private readonly GuiButton f_ReloadBtn;

            private KnownFile f_Known;

            private string f_CachedCommentFilterStr;
            private string[] f_CachedCommentFilterArray;

            public CommentsPanel(Instance a_Inst)
            {
                f_Inst = a_Inst;

                f_GridView = new GridView<object, DlPeerContext>(f_Inst, this, f_Inst.f_MainForm.f_VisualCust.f_CommentsPanelColSet);
                f_AddOwnBtn = new GuiButton(this);
                f_EditFilterBtn = new GuiButton(this);
                f_ReloadBtn = new GuiButton(this);

                f_GridView.Dock = DockStyle.None;
                Lyt.SetFontAndHeight(f_AddOwnBtn, f_AddOwnBtn.Font);
                f_EditFilterBtn.Height = f_ReloadBtn.Height = f_AddOwnBtn.Height;
                f_EditFilterBtn.Width = f_ReloadBtn.Width = f_AddOwnBtn.Width = Lyt.s_GenerousButtonWidth;

                f_GridView.e_UpdateItem += GridView_UpdateItem;
                f_AddOwnBtn.Click += AddOwnBtn_Click;
                f_EditFilterBtn.Click += EditFilterBtn_Click;
                f_ReloadBtn.Click += ReloadBtn_Click;

                TextSet l_TextSet = f_Inst.f_TextSet;
                f_AddOwnBtn.Text = l_TextSet.GetAsDialogOpener(TextKeys.FilesTabCommentsSetOwn);
                f_EditFilterBtn.Text = l_TextSet.GetAsDialogOpener(TextKeys.FilesTabCommentsEditFilter);
                f_ReloadBtn.Text = l_TextSet[TextKeys.FilesTabCommentsQueryFromKad];
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    Dialog.DisposeAllChildren(this);
                }

                base.Dispose(disposing);
            }

            public override bool HasInputFocus()
            {
                return (true == f_GridView.Focused ||
                        true == f_AddOwnBtn.Focused ||
                        true == f_EditFilterBtn.Focused ||
                        true == f_ReloadBtn.Focused);
            }

            protected override void OnResize(EventArgs eventargs)
            {
                f_AddOwnBtn.Top = f_EditFilterBtn.Top = f_ReloadBtn.Top = (this.Height - f_AddOwnBtn.Height - Lyt.s_SmallSpace);
                f_GridView.Size = new Size(this.Width, f_AddOwnBtn.Top - Lyt.s_Space);
                f_ReloadBtn.Left = (this.Width - f_ReloadBtn.Width - Lyt.s_SmallSpace);
                f_EditFilterBtn.Left = (f_ReloadBtn.Left - f_EditFilterBtn.Width - Lyt.s_SmallSpace);
                f_AddOwnBtn.Left = (f_EditFilterBtn.Left - f_AddOwnBtn.Width - Lyt.s_SmallSpace);

                base.OnResize(eventargs);
            }

            private void AddOwnBtn_Click(object sender, EventArgs e)
            {
                if (null == f_Known)
                {
                    return;
                }

                using (SetOwnCommentDialog l_Dlg = new SetOwnCommentDialog(f_Inst))
                {
                    l_Dlg.ShowDialog(f_Known);

                    if (DialogResult.OK == l_Dlg.DialogResult)
                    {
                        return;
                    }
                }
            }

            private void EditFilterBtn_Click(object sender, EventArgs e)
            {
                f_Inst.f_MainForm.ShowOptionsDialog(MainForm.ShowOptionsDlgReasons.EditCommentFilter);
            }

            private void ReloadBtn_Click(object sender, EventArgs e)
            {
                f_Inst.f_MainForm.InfoMsgBox("Sorry, this feature won't be installed until tuesday.");
            }

            public void FinishLoading()
            {
                f_GridView.LoadColumnAdjustments();
            }

            public void PrepareSaving()
            {
                f_GridView.StoreColumnAdjustments();
            }

            public void Clear()
            {
                f_Known = null;

                f_GridView.ClearItems();
                f_AddOwnBtn.Enabled = f_ReloadBtn.Enabled = false;
            }

            public void Update(ref KnownFileUpdateContext a_UpdateCntxt)
            {
                f_Known = a_UpdateCntxt.f_Known;

                f_GridView.StartChanging();

                if (null != a_UpdateCntxt.f_DlPeers)
                {
                    if (f_Inst.f_MainForm.p_Prefs.f_KadPrefs.f_CommentsFilter != f_CachedCommentFilterStr)
                    {
                        // Update our cached filter array for
                        // GridView_UpdateItem().

                        f_CachedCommentFilterStr = f_Inst.f_MainForm.p_Prefs.f_KadPrefs.f_CommentsFilter;
                        f_CachedCommentFilterArray = KadPrefs.BuildCommentFilterArray(f_CachedCommentFilterStr);
                    }

                    foreach (DlPeerContext l_DlPeerCntxt in a_UpdateCntxt.f_DlPeers)
                    {
                        if (FileRatings.NotRated != l_DlPeerCntxt.f_PeersRating ||
                            null != l_DlPeerCntxt.f_PeersDescription)
                        {
                            GridViewItem<object, DlPeerContext> l_Item = f_GridView.CreateItem(l_DlPeerCntxt);
                            l_Item.f_Tag = l_DlPeerCntxt;
                        }
                    }
                }

                if (FileRatings.NotRated != f_Known.f_OwnRating ||
                    null != f_Known.f_OwnComment)
                {
                    // We don't use a DlPeerContext as key here, but
                    // this panel itself.
                    GridViewItem<object, DlPeerContext> l_Item = f_GridView.CreateItem(this);
                }

                f_GridView.EndChanging();

                f_AddOwnBtn.Enabled = f_ReloadBtn.Enabled = true;
            }

            private void GridView_UpdateItem(Time a_Time, GridViewItem<object, DlPeerContext> a_Item)
            {
                bool l_OwnComment = (a_Item.f_Ref == this);
                string l_FileName;
                FileRatings l_Rating;
                string l_Description;

                if (false == l_OwnComment)
                {
                    // This is a peer's comment.

                    l_FileName = a_Item.f_Tag.f_PeersFileName;
                    l_Rating = a_Item.f_Tag.f_PeersRating;
                    l_Description = a_Item.f_Tag.f_PeersDescription;
                }
                else
                {
                    // This is our own comment. a_Item.f_Tag is null.

                    if (null == f_Known)
                    {
                        // Just to be on the safe side.

                        return;
                    }

                    l_FileName = f_Known.GetFileNameVolatile(true);
                    l_Rating = f_Known.f_OwnRating;
                    l_Description = f_Known.f_OwnComment;
                }

                Util.MemoryBarrier();

                if (null == l_FileName)
                {
                    l_FileName = string.Empty;
                }

                if (null == l_Description)
                {
                    l_Description = string.Empty;
                }
                else if (false == l_OwnComment &&
                         null != f_CachedCommentFilterArray &&
                         true == KadPrefs.IsCommentFiltered(l_Description, f_CachedCommentFilterArray))
                {
                    // Filtered.

                    l_Description = string.Empty;
                }

                a_Item.SetTextWithCompObj(0, l_FileName, l_FileName);
                a_Item.SetText(1, (byte)l_Rating, FilesPanel.s_RatingStrs[(int)l_Rating].ToString());
                a_Item.SetTextWithCompObj(2, l_Description, l_Description);

                if (true == l_OwnComment)
                {
                    a_Item.DefaultCellStyle = f_Inst.f_MainForm.f_BoldCellStyle;
                }
            }

            sealed private class SetOwnCommentDialog : ButtonDialog
            {
                private const ushort c_DlgWidth = 400;
                private const ushort c_RatingBoxWidth = 170;

                private readonly Instance f_Inst;

                private readonly GuiGroupBox f_GroupBox;
                private readonly LabeledTextBox f_DescriptionBox;
                private readonly LabeledComboBox f_RatingBox;

                private KnownFile f_TargetFile;

                public SetOwnCommentDialog(Instance a_Inst) :
                                      base(a_Inst.f_TextSet, DialogButtons.OKCancel)
                {
                    f_Inst = a_Inst;

                    this.SuspendLayout();

                    f_GroupBox = new GuiGroupBox(this);
                    int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
                    int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
                    int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
                    f_DescriptionBox = new LabeledTextBox(f_GroupBox, l_InnerWidth);
                    f_RatingBox = new LabeledComboBox(f_GroupBox, l_InnerWidth, Lyt.CalcX(c_RatingBoxWidth), false);

                    f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                    f_DescriptionBox.SetLocation(new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top));
                    f_RatingBox.SetLocation(new Point(f_GroupBox.Padding.Left, f_DescriptionBox.GetBottom()));
                    f_GroupBox.Size = new Size(l_GroupBoxWidth, f_RatingBox.GetBottom());

                    AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
                    base.SetSelectOnShowControl(f_DescriptionBox.f_Box);

                    f_DescriptionBox.f_Box.MaxLength = Kademlia.c_MaxCommentChars;
                    f_OKButton.Click += OKButton_Click;

                    TextSet l_TextSet = f_Inst.f_TextSet;
                    this.Text = l_TextSet[TextKeys.FilesTabCommentsSetDlgTitle];
                    f_GroupBox.Text = l_TextSet[TextKeys.FilesTabCommentsSetDlgGroupBoxCaption];
                    f_DescriptionBox.SetCaption(l_TextSet, TextKeys.FilesTabCommentsSetDlgDescription);
                    f_RatingBox.SetTexts(l_TextSet, TextKeys.FilesTabCommentsSetDlgRating, FilesPanel.s_RatingStrs);

                    this.ResumeLayout();
                }

                protected override void Dispose(bool disposing)
                {
                    if (true == disposing)
                    {
                        Dialog.DisposeAllChildren(f_GroupBox);
                        Dialog.DisposeAllChildren(this);
                    }

                    base.Dispose(disposing);
                }

                private void OKButton_Click(object sender, EventArgs e)
                {
                    string l_Description = f_DescriptionBox.f_Box.Text.Trim();

                    if (0 == l_Description.Length)
                    {
                        l_Description = null;
                    }
                    else
                    {
                        string[] l_Arr = f_Inst.f_MainForm.p_Prefs.f_KadPrefs.BuildCommentFilterArray();

                        if (true == KadPrefs.IsCommentFiltered(l_Description, l_Arr) &&
                            false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.FilesTabCommentsSetDlgWarnMsgFilter]))
                        {
                            return;
                        }
                    }

                    f_TargetFile.f_OwnRating = (FileRatings)f_RatingBox.f_Box.SelectedIndex;
                    f_TargetFile.f_OwnComment = l_Description;

                    this.DialogResult = DialogResult.OK;
                    this.Close();
                }

                public void ShowDialog(KnownFile a_Known)
                {
                    f_TargetFile = a_Known;

                    f_DescriptionBox.f_Box.Text = f_TargetFile.f_OwnComment;
                    f_RatingBox.f_Box.SelectedIndex = (int)f_TargetFile.f_OwnRating;

                    this.ShowDialog(f_Inst.f_MainForm);
                }
            }
        }

        public class FileRemovalContext : AbortableControllerContext
        {
            public readonly KnownFile[] f_Files;
            public readonly FileRemovalModes f_RemovalMode;

            private uint f_SuccCount;

            public FileRemovalContext(KnownFile[] a_Files,
                                      FileRemovalModes a_RemovalMode)
            {
                f_Files = a_Files;
                f_RemovalMode = a_RemovalMode;
            }

            public void SetProgress(uint a_Done,
                                    uint a_Total)
            {
                base.SetPercentage(a_Done, a_Total);
            }

            public void SetFinished(uint a_SuccCount)
            {
                f_SuccCount = a_SuccCount;

                base.SetFinishedBase();
            }

            public uint GetSuccCount()
            {
                return f_SuccCount;
            }
        }
    }

    class GuiSearchTerms
    {
        public const ushort c_MaxKeywordStrChars = 512;

        // NOTE:
        // The result must contain all of the f_AndKeywords.
        // The result must contain at least one of the f_OrKeywords.
        // The result must not contain any of the f_NotKeywords.

        public readonly string[] f_AndKeywords;
        // The following three pointers can be null.
        public readonly string[] f_OrKeywords;
        public readonly string[] f_NotKeywords;
        public readonly string f_Extension;
        public readonly FileType f_FileType;
        public readonly ulong f_MinFileSize;
        public readonly ulong f_MaxFileSize;

        public readonly ulong f_HashCode;

        private GuiSearchTerms(string[] a_AndKeywords,
                               string[] a_OrKeywords,
                               string[] a_NotKeywords,
                               string a_Extension,
                               FileType a_FileType,
                               ulong a_MinFileSize,
                               ulong a_MaxFileSize)
        {
            f_AndKeywords = a_AndKeywords;
            f_OrKeywords = a_OrKeywords;
            f_NotKeywords = a_NotKeywords;
            f_Extension = a_Extension;
            f_FileType = a_FileType;
            f_MinFileSize = a_MinFileSize;
            f_MaxFileSize = a_MaxFileSize;

            f_HashCode = CalcHashCode();
        }

        public static GuiSearchTerms Create(TextSet a_TextSet,
                                            string a_AndKeywords,
                                            string a_OrKeywords,
                                            string a_NotKeywords,
                                            string a_Extension,
                                            FileType a_FileType,
                                            ulong a_MinFileSize,
                                            ulong a_MaxFileSize,
                                            out string a_ErrMsg)
        {
            string[] l_AndKeywords = SearchManager.IdentifyKeywords(a_AndKeywords, true);

            if (null == l_AndKeywords)
            {
                a_ErrMsg = a_TextSet[TextKeys.FileSearchTabErrMustHaveOneKeyword];

                return null;
            }

            string[] l_OrKeywords = SearchManager.IdentifyKeywords(a_OrKeywords, true);
            string[] l_NotKeywords = SearchManager.IdentifyKeywords(a_NotKeywords, true);

            if (null != l_NotKeywords)
            {
                foreach (string l_Not in l_NotKeywords)
                {
                    // If there is only one or-keyword, make sure it
                    // isn't excluded.
                    bool l_Conflict = (null != l_OrKeywords &&
                                       1 == l_OrKeywords.Length &&
                                       l_OrKeywords[0] == l_Not);

                    if (false == l_Conflict)
                    {
                        l_Conflict = (0 <= Array.IndexOf(l_AndKeywords, l_Not));
                    }

                    if (true == l_Conflict)
                    {
                        a_ErrMsg = a_TextSet.Format(TextKeys.FileSearchTabErrProblemWithNOT, l_Not);

                        return null;
                    }
                }
            }

            string l_Extension = (0 == a_Extension.Length ? null : a_Extension);

            if (a_MinFileSize > a_MaxFileSize)
            {
                a_ErrMsg = a_TextSet[TextKeys.FileSearchTabErrProblemWithSizes];

                return null;
            }

            a_ErrMsg = null;

            return new GuiSearchTerms(l_AndKeywords, l_OrKeywords, l_NotKeywords, l_Extension, a_FileType, a_MinFileSize, a_MaxFileSize);
        }

        public void ToBuffer(BufferIO a_Buffer)
        {
            WriteKeywordArray(f_AndKeywords, a_Buffer);
            WriteKeywordArray(f_OrKeywords, a_Buffer);
            WriteKeywordArray(f_NotKeywords, a_Buffer);
            a_Buffer.WriteUTF16LEStringEncLen(null == f_Extension ? string.Empty : f_Extension);
            a_Buffer.WriteByte((byte)f_FileType);
            a_Buffer.WriteUInt64LE(f_MinFileSize);
            a_Buffer.WriteUInt64LE(f_MaxFileSize);
        }

        public static bool FromBuffer(BufferIO a_Buffer,
                                      out GuiSearchTerms a_SearchTerms)
        {
            string[] l_AndKeywords;
            string[] l_OrKeywords;
            string[] l_NotKeywords;

            if (false == ReadKeywordArray(a_Buffer, out l_AndKeywords) ||
                null == l_AndKeywords ||
                false == ReadKeywordArray(a_Buffer, out l_OrKeywords) ||
                false == ReadKeywordArray(a_Buffer, out l_NotKeywords))
            {
                a_SearchTerms = null;

                return false;
            }

            string l_Extension = a_Buffer.ReadUTF16LEStringEncLen();

            if (0 == l_Extension.Length)
            {
                l_Extension = null;
            }

            FileType l_FileType = (FileType)a_Buffer.ReadByte();
            ulong l_MinFileSize = a_Buffer.ReadUInt64LE();
            ulong l_MaxFileSize = a_Buffer.ReadUInt64LE();

            if (FileType.FT_COLLECTION < l_FileType ||
                l_MinFileSize > l_MaxFileSize)
            {
                a_SearchTerms = null;

                return false;
            }

            a_SearchTerms = new GuiSearchTerms(l_AndKeywords,
                                               l_OrKeywords,
                                               l_NotKeywords,
                                               l_Extension,
                                               l_FileType,
                                               l_MinFileSize,
                                               l_MaxFileSize);

            return true;
        }

        public string GetCaption()
        {
            return Join(f_AndKeywords);
        }

        public bool IsSimple()
        {
            return (1 == f_AndKeywords.Length &&
                    null == f_OrKeywords &&
                    null == f_NotKeywords &&
                    null == f_Extension &&
                    FileType.FT_ANY == f_FileType &&
                    false == HasFileSizeConstraints());
        }

        public bool HasFileSizeConstraints()
        {
            return (0 != f_MinFileSize ||
                    ulong.MaxValue != f_MaxFileSize);
        }

        private ulong CalcHashCode()
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            l_Buffer.WriteByte(0);
            WriteKeywordArray(f_AndKeywords, l_Buffer);

            if (null != f_OrKeywords)
            {
                l_Buffer.WriteByte(1);
                WriteKeywordArray(f_OrKeywords, l_Buffer);
            }

            if (null != f_NotKeywords)
            {
                l_Buffer.WriteByte(2);
                WriteKeywordArray(f_NotKeywords, l_Buffer);
            }

            if (null != f_Extension)
            {
                l_Buffer.WriteByte(3);
                l_Buffer.WriteUTF16LEStringEncLen(f_Extension);
            }

            if (FileType.FT_ANY == f_FileType)
            {
                l_Buffer.WriteByte(4);
                l_Buffer.WriteByte((byte)f_FileType);
            }

            if (0 != f_MinFileSize)
            {
                l_Buffer.WriteByte(5);
                l_Buffer.WriteUInt64LE(f_MinFileSize);
            }

            if (ulong.MaxValue != f_MaxFileSize)
            {
                l_Buffer.WriteByte(6);
                l_Buffer.WriteUInt64LE(f_MaxFileSize);
            }

            Lookup3 l_HashAlgo = new Lookup3();
            l_Buffer.HashRange(l_HashAlgo, 0, l_Buffer.p_Length);

            return l_HashAlgo.GetDigestAsULong();
        }

        private static void WriteKeywordArray(string[] a_Strs,
                                              BufferIO a_Buffer)
        {
            // a_Strs may be null.

            string l_CommonStr = Join(a_Strs);
            a_Buffer.WriteUTF16LEStringEncLen(l_CommonStr);
        }

        private static bool ReadKeywordArray(BufferIO a_Buffer,
                                             out string[] a_Keywords)
        {
            // a_Keywords can be null on return regardless of success.

            string l_RawStr = a_Buffer.ReadUTF16LEStringEncLen();

            if (c_MaxKeywordStrChars < l_RawStr.Length)
            {
                a_Keywords = null;

                return false;
            }

            a_Keywords = SearchManager.IdentifyKeywords(l_RawStr, false);

            return true;
        }

        public static string Join(string[] a_Strs)
        {
            // a_Strs may be null.

            if (null != a_Strs)
            {
                return string.Join(" ", a_Strs);
            }

            return string.Empty;
        }
    }

    class GuiSearchCookie
    {
        public readonly GuiSearchTerms f_SearchTerms;
        // Volatile.
        private CookieStates f_GuiState;

        private readonly Dictionary<FileID, KadFoundFile> f_Results;
        private bool f_ResultsDirty;
        // The following two fields may be safely read by the GUI.
        public uint f_GuiResultsGeneration;
        public Dictionary<FileID, KadFoundFile> f_GuiResults;
        // Volatile.
        private CookieStates f_NetworkState;

        public bool f_KadFiltered;

        public CookieStates p_GuiStateVolatile
        {
            get { return Util.VolatileRead(ref f_GuiState); }
        }

        public CookieStates p_NetworkState
        {
            get { return f_NetworkState; }
        }

        public CookieStates p_NetworkStateVolatile
        {
            get { return Util.VolatileRead(ref f_NetworkState); }
        }

        public GuiSearchCookie(GuiSearchTerms a_SearchTerms)
        {
            f_Results = new Dictionary<FileID, KadFoundFile>();
            f_SearchTerms = a_SearchTerms;
        }

        public void Aborted()
        {
            // Used by the SearchCacheItem class.

            Util.VolatileWrite(out f_GuiState, CookieStates.Failure);
        }

        public void Rejected()
        {
            // Used by the SearchManager class.

            Util.VolatileWrite(out f_NetworkState, CookieStates.Failure);
        }

        public void SearchStarted()
        {
            // Used by the SearchManager class.

            Util.VolatileWrite(out f_NetworkState, CookieStates.SearchRunning);
        }

        public void Finished()
        {
            // Used by the SearchManager class.

            Util.VolatileWrite(out f_NetworkState, CookieStates.Success);
        }

        public void AddFile(KademliaFilter a_KadFilter,
                            FileID a_FileID,
                            string a_FileName,
                            ulong a_FileSize,
                            ushort a_CompleteSources,
                            ushort a_Sources)
        {
            // Used by the SearchManager class.

            if (true == a_KadFilter.RunFileNameThroughFilter(a_FileName, KademliaFilterScopes.ScopeSearchResultsForSelf))
            {
                f_KadFiltered = true;

                return;
            }

            KadFoundFile l_KadFile;

            if (false == f_Results.TryGetValue(a_FileID, out l_KadFile))
            {
                f_Results.Add(a_FileID, new KadFoundFile(a_FileID, a_FileName, a_FileSize, a_CompleteSources, a_Sources));
            }
            else
            {
                l_KadFile.AddSources(a_CompleteSources, a_Sources);
            }

            f_ResultsDirty = true;
        }

        public void MakeResultsAvailableToGui()
        {
            // Used by the SearchManager class.

            if (true == f_ResultsDirty)
            {
                f_ResultsDirty = false;

                f_GuiResults = new Dictionary<FileID, KadFoundFile>(f_Results);
                Util.VolatileWrite(out f_GuiResultsGeneration, f_GuiResultsGeneration + 1);
            }
        }
    }

    class LicenseDialog : ButtonDialog
    {
        private const byte c_InfoTextHeight = 30;

        private readonly DialogMode f_Mode;
        private readonly string f_LicenseText;

        private readonly GuiLabel f_InfoLabel;
        private readonly ReadOnlyTextBox f_TextBox;
        // f_AcceptBox can be null.
        private readonly GuiCheckBox f_AcceptBox;

        private readonly Font f_TextBoxFont;
        // f_AcceptBoxFont can be null.
        private readonly Font f_AcceptBoxFont;
        private readonly ContextMenuStripEx f_CntxtMenu;

        private readonly bool f_MayHandleResize;

        public LicenseDialog(TextSet a_TextSet,
                             DialogMode a_Mode) :
                        base(a_TextSet, DialogMode.ViewLicense == a_Mode ? DialogButtons.OK : DialogButtons.OKCancel)
        {
            f_Mode = a_Mode;
            f_LicenseText = GetLicense();

            this.SuspendLayout();

            f_InfoLabel = new GuiLabel(this);
            f_TextBox = new ReadOnlyTextBox(this);

            // Enforce a minimum font size.
            f_TextBoxFont = new Font(FontFamily.GenericMonospace, Math.Max(f_TextBox.Font.SizeInPoints, 9f));

            const byte l_MaxExpectedCharsPerLine = 80;
            int l_ReqWidth = (MainForm.MeasureTextWidth(new string('*', l_MaxExpectedCharsPerLine), f_TextBoxFont) + (3 * Lyt.s_Space));
            Size l_AimedDlgSize = Lyt.TranformSize(650, 400);
            l_AimedDlgSize.Width = Math.Max(l_AimedDlgSize.Width, l_ReqWidth);

            base.AsSizeableDialog(l_AimedDlgSize);
            base.SetSelectOnShowControl(f_TextBox);

            f_InfoLabel.Location = f_InnerRect.Location;
            f_InfoLabel.Size = new Size(f_InnerRect.Width, Lyt.CalcY(c_InfoTextHeight));
            f_TextBox.Location = new Point(f_InnerRect.Left, f_InfoLabel.GetBottom() + Lyt.s_SmallSpace);
            f_TextBox.Font = f_TextBoxFont;
            f_TextBox.WordWrap = false;
            f_TextBox.ScrollBars = RichTextBoxScrollBars.Both;

            if (DialogMode.ViewLicense == a_Mode)
            {
                // Since the user cannot decide anything, use the okay
                // button as cancel button as well.
                this.CancelButton = f_OKButton;
            }
            else
            {
                f_AcceptBox = new GuiCheckBox(this, int.MaxValue);
                f_AcceptBox.Font = f_AcceptBoxFont = MainForm.RecreateFont(f_AcceptBox.Font, f_AcceptBox.Font.SizeInPoints, FontStyle.Bold);
                f_AcceptBox.Click += AcceptBox_Click;
                f_OKButton.Enabled = false;

                f_AcceptBox.SetTextAndUpdateSize(a_TextSet[TextKeys.LicenseDlgIAccept]);
                this.f_OKButton.Text = a_TextSet[TextKeys.ButtonTextAccept];
            }

            f_CntxtMenu = new ContextMenuStripEx(f_TextBox, false);
            MainForm.AddMenuItem(f_CntxtMenu, a_TextSet[TextKeys.LicenseDlgCopyLicenseToClipboard], MenuCopyLicenseToClipboard_Click);

            this.MinimizeBox = false;
            this.f_OKButton.Click += DefaultOKCloseHandler;

            this.Text = a_TextSet.FormatWithAppName(TextKeys.LicenseDlgTitle);
            f_InfoLabel.Text = a_TextSet[TextKeys.LicenseDlgComment];
            f_TextBox.Text = f_LicenseText;

            UpdateSizes();

            this.ResumeLayout();

            f_MayHandleResize = true;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);
                f_TextBoxFont.Dispose();

                if (null != f_AcceptBoxFont)
                {
                    f_AcceptBoxFont.Dispose();
                }

                f_CntxtMenu.Dispose();
            }

            base.Dispose(disposing);
        }

        protected override void OnResize(EventArgs e)
        {
            // Need to call base class implementation.

            // Update the base class first, since we rely f_InnerRect.
            base.OnResize(e);

            if (true == f_MayHandleResize)
            {
                UpdateSizes();
            }
        }

        private void UpdateSizes()
        {
            this.SuspendLayout();

            if (DialogMode.ViewLicense == f_Mode)
            {
                f_TextBox.Size = new Size(f_InnerRect.Width, f_InnerRect.Bottom - f_TextBox.Top);
            }
            else
            {
                f_AcceptBox.Location = new Point(f_InnerRect.Left, f_InnerRect.Bottom - f_AcceptBox.Height);
                f_TextBox.Size = new Size(f_InnerRect.Width, f_AcceptBox.Top - f_TextBox.Top - Lyt.s_Space);
            }

            this.ResumeLayout();
        }

        private void AcceptBox_Click(object sender, EventArgs e)
        {
            f_OKButton.Enabled = f_AcceptBox.Checked;
        }

        private void MenuCopyLicenseToClipboard_Click(object sender, EventArgs e)
        {
            Util.SetClipboard(f_LicenseText);
        }

        private static string GetLicense()
        {
            const int l_LicenseBinarySize = 79346;
            const uint l_LicenseHash = 0xCB480501;
            ByteArray l_Data = new ByteArray(l_LicenseBinarySize);

            System.IO.Stream l_Strm = Software.OpenResource("license.txt", true);
            {
                if (l_LicenseBinarySize != l_Strm.Length ||
                    l_LicenseBinarySize != Util.ReadFromStream(l_Strm, ref l_Data) ||
                    l_LicenseHash != Adler32.ComputeHash(l_Data.f_Arr))
                {
                    throw (new ApplicationException());
                }
            }
            l_Strm.Dispose();

            return Util.s_Latin1BestFit.GetString(l_Data.f_Arr);
        }

        public enum DialogMode : byte
        {
            ViewLicense = 0,
            AcceptOrRejectLicense = 1,
        }
    }

    sealed class AboutDialog : Dialog
    {
        private readonly Instance f_Inst;
        private readonly TimeOfYear f_TimeOfYear;
        private readonly int f_LabelWidth;

        private readonly GuiGroupBox f_TopGroupBox;
        private readonly GuiLabel f_AppLabel;

        private readonly GuiGroupBox f_BottomGroupBox;
        private readonly GuiLabel f_DescriptionLabel;
        private readonly GuiLabel f_DeveloperLabel;
        private readonly GuiLabel f_VersionLabel;
        private readonly GuiLabel f_DateLabel;
        private readonly GuiLabel f_CopyrightLabel;
        private readonly GuiLabel f_FeedbackLabel;
        private readonly GuiLinkLabel f_MailToLink;
        private readonly GuiLabel f_CheckForUpdatesLabel;
        private readonly GuiLinkLabel f_VisitSiteLink;

        private readonly GuiLabel f_GreetingLabel;
        private readonly GuiButton f_OKButton;

        private readonly Font f_AppFont;
        private readonly Font f_GreetingFont;

        private WinterInNB f_WinterInNB;
        private bool f_IgnorePaintReentry;

        public AboutDialog(Instance a_Inst)
        {
            const ushort l_DlgWidth = 325;
            const float l_NormalFormOpacity = .90f;
            const float l_FlakesFormOpacity = .87f;

            f_Inst = a_Inst;
            f_TimeOfYear = DetermineTimeOfYear();

            this.SuspendLayout();

            f_TopGroupBox = new GuiGroupBox(this, false);
            Color l_BackCol = f_TopGroupBox.BackColor;
            f_AppLabel = new GuiLabel(f_TopGroupBox);
            f_BottomGroupBox = new GuiGroupBox(this, false);
            f_DescriptionLabel = new GuiLabel(f_BottomGroupBox);
            f_DeveloperLabel = new GuiLabel(f_BottomGroupBox);
            f_VersionLabel = new GuiLabel(f_BottomGroupBox);
            f_DateLabel = new GuiLabel(f_BottomGroupBox);
            f_CopyrightLabel = new GuiLabel(f_BottomGroupBox);
            f_FeedbackLabel = new GuiLabel(f_BottomGroupBox);
            f_MailToLink = new GuiLinkLabel(f_BottomGroupBox);
            f_CheckForUpdatesLabel = new GuiLabel(f_BottomGroupBox);
            f_VisitSiteLink = new GuiLinkLabel(f_BottomGroupBox);
            f_GreetingLabel = new GuiLabel(this);
            f_OKButton = new GuiButton(this);

            // NOTE: Text that has this font will not appear 100%
            // vertically centered if demanded so.
            bool l_SpecialAppFont = MainForm.CreateFont("Colonna MT", Math.Min(4f * this.Font.SizeInPoints, 33f), FontStyle.Regular, out f_AppFont);

            if (false == l_SpecialAppFont)
            {
                f_AppFont = MainForm.RecreateFont(this.Font, 1.75f * this.Font.SizeInPoints, FontStyle.Bold);
            }

            f_GreetingFont = f_GreetingLabel.Font;
            f_GreetingFont = MainForm.RecreateFont(f_GreetingFont, f_GreetingFont.SizeInPoints, FontStyle.Bold);

            int l_Width = Lyt.CalcX(l_DlgWidth);
            int l_GroupBoxWidth = (l_Width - Lyt.s_Space2);
            f_LabelWidth = (l_GroupBoxWidth - f_BottomGroupBox.Padding.Horizontal);

            PrepareLabel(f_AppLabel);
            PrepareLabel(f_DescriptionLabel);
            PrepareLabel(f_DeveloperLabel);
            PrepareLabel(f_VersionLabel);
            PrepareLabel(f_DateLabel);
            PrepareLabel(f_CopyrightLabel);
            PrepareLabel(f_FeedbackLabel);
            PrepareLabel(f_MailToLink);
            PrepareLabel(f_CheckForUpdatesLabel);
            PrepareLabel(f_VisitSiteLink);
            PrepareLabel(f_GreetingLabel);

            TextSet l_TextSet = f_Inst.f_TextSet;
            // Textless group boxes behave different depending on
            // the framework.
            int l_GroupBoxVertSpace =
#if MONO
            0;
#else
 Lyt.s_SmallSpace;
#endif
            int l_Left = f_TopGroupBox.Padding.Left;
            f_TopGroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space - l_GroupBoxVertSpace);
            f_AppLabel.Location = new Point(f_TopGroupBox.Padding.Left, f_TopGroupBox.Padding.Top - l_GroupBoxVertSpace);
            Lyt.SetFontAndHeight(f_AppLabel, f_AppFont);
            f_TopGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_AppLabel.Bottom);
            f_BottomGroupBox.Location = new Point(f_TopGroupBox.Left, f_TopGroupBox.GetBottom() - l_GroupBoxVertSpace);
            f_DescriptionLabel.Location = new Point(l_Left, f_BottomGroupBox.Padding.Top - l_GroupBoxVertSpace);
            f_DescriptionLabel.Text = TextSet.EnsureHasLineBreak(l_TextSet[TextKeys.AboutDlgAppDescription], 40);
            Lyt.SetFontAndHeight(f_DescriptionLabel, f_DescriptionLabel.Font);
            f_DeveloperLabel.Location = new Point(l_Left, f_DescriptionLabel.GetBottom());
            f_VersionLabel.Location = new Point(l_Left, f_DeveloperLabel.GetBottom());
            f_DateLabel.Location = new Point(l_Left, f_VersionLabel.GetBottom());
            f_CopyrightLabel.Location = new Point(l_Left, f_DateLabel.GetBottom());
            f_CopyrightLabel.Height = f_DescriptionLabel.Height;
            f_FeedbackLabel.Location = new Point(l_Left, f_CopyrightLabel.GetBottom() + Lyt.s_SmallSpace);
            f_MailToLink.Location = new Point(l_Left, f_FeedbackLabel.GetBottom());
            f_CheckForUpdatesLabel.Location = new Point(l_Left, f_MailToLink.GetBottom() + Lyt.s_SmallSpace);
            f_VisitSiteLink.Location = new Point(l_Left, f_CheckForUpdatesLabel.GetBottom());
            f_BottomGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_VisitSiteLink.Bottom);
            f_GreetingLabel.Location = new Point(f_BottomGroupBox.Left + l_Left, f_BottomGroupBox.GetBottom());
            f_GreetingLabel.Visible = (TimeOfYear.NeutralButSnowy < f_TimeOfYear);
            Lyt.SetFontAndHeight(f_GreetingLabel, f_GreetingFont);
            f_OKButton.Location = new Point(f_BottomGroupBox.Left + ((l_GroupBoxWidth - Lyt.s_DefaultButtonWidth) / 2), TimeOfYear.NeutralButSnowy < f_TimeOfYear ? f_GreetingLabel.GetBottom() : f_BottomGroupBox.GetBottom());
            this.AcceptButton = f_OKButton;
            this.CancelButton = f_OKButton;
            this.Opacity = l_NormalFormOpacity;

            if (TimeOfYear.Neutral != f_TimeOfYear)
            {
                this.DoubleBuffered = true;
                this.Opacity = l_FlakesFormOpacity;
                this.BackColor = Color.DarkRed;

                f_TopGroupBox.BackColor = Color.Transparent;
                f_VisitSiteLink.LinkColor = f_MailToLink.LinkColor = this.ForeColor = Color.White;
                f_BottomGroupBox.BackColor = Color.Transparent;
                f_OKButton.ForeColor = Color.Black;

                if (TimeOfYear.NeutralButSnowy < f_TimeOfYear)
                {
                    f_GreetingLabel.Text = (TimeOfYear.NewGregorianYear == f_TimeOfYear ? "Happy New Year!!!" : "Happy Holidays!!!");
                }
            }

            AsFixedDialog(new Size(l_Width, f_OKButton.Bottom + Lyt.s_Space));

            f_MailToLink.LinkClicked += MailToLink_Clicked;
            f_VisitSiteLink.LinkClicked += VisitSiteLink_Clicked;
            f_OKButton.Click += DefaultOKCloseHandler;

            this.Text = l_TextSet.FormatWithAppName(TextKeys.AboutDlgTitle);
            int l_BuildAgeInDays = Software.s_ReleaseInfo.GetReleaseAgeInDays();
            string l_BuildAgeStr = l_TextSet.Format(TextKeys.AboutDlgAppBuildDate, Software.s_ReleaseInfo.f_BuildDate);

            if (2 <= l_BuildAgeInDays)
            {
                // Our text is in plural form, so 2 days at least.

                l_BuildAgeStr = l_TextSet.Format(TextKeys.AboutDlgAppBuildAgeFormat, l_BuildAgeStr, l_BuildAgeInDays);
            }

            f_AppLabel.Text = (false == l_SpecialAppFont ? l_TextSet.f_AppName : l_TextSet.f_AppName.ToUpper());
            f_DeveloperLabel.Text = l_TextSet.Format(TextKeys.AboutDlgAppCredits, "Stephan R\x00F6ger");
            f_VersionLabel.Text = (l_TextSet.Format(TextKeys.AboutDlgAppVersion, Software.s_ReleaseInfo.f_VersionStr) + " \x201E" + Software.s_ReleaseInfo.f_ReleaseCodename + "\x201C");
            f_DateLabel.Text = l_BuildAgeStr;
            f_CopyrightLabel.Text = ("Copyright \x00A9 2012-2015 " + Software.c_AppName + " Authors.\nAll rights reserved.");
            f_FeedbackLabel.Text = l_TextSet.GetAsInputControlCaption(TextKeys.AboutDlgProvideFeedback);
            f_MailToLink.Text = Software.c_ProjectEmailAddr;
            f_CheckForUpdatesLabel.Text = l_TextSet.GetAsInputControlCaption(TextKeys.AboutDlgCheckForUpdates);
            f_VisitSiteLink.Text = Software.c_ProjectURL;
            f_OKButton.Text = f_Inst[TextKeys.ButtonTextOK];

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_WinterInNB.Dispose();
                Dialog.DisposeAllChildren(f_TopGroupBox);
                Dialog.DisposeAllChildren(f_BottomGroupBox);
                Dialog.DisposeAllChildren(this);
                f_AppFont.Dispose();
                f_GreetingFont.Dispose();
            }

            base.Dispose(disposing);
        }

        protected override void OnShown(EventArgs e)
        {
            f_OKButton.Focus();

            base.OnShown(e);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            // Need to call base class implementation.

            if (TimeOfYear.Neutral != f_TimeOfYear &&
                false == f_IgnorePaintReentry)
            {
                if (false == f_WinterInNB.IsValid())
                {
                    f_IgnorePaintReentry = true;

                    // We must initialize this here, since construction
                    // involves painting on a bitmap.
                    f_WinterInNB = new WinterInNB(this);
                    f_WinterInNB.AdjustSnow(true);

                    f_IgnorePaintReentry = false;

                    f_OKButton.Focus();
                }
                else
                {
                    f_WinterInNB.Draw(e.Graphics);
                }
            }

            base.OnPaint(e);
        }

        protected override void OnClosed(EventArgs e)
        {
            f_WinterInNB.Dispose();
            f_WinterInNB = new WinterInNB();

            base.OnClosed(e);
        }

        private void FlakesTimer_AdvTick(GuiTimer a_Sender,
                                         uint a_Times)
        {
            f_WinterInNB.HandleTick(a_Times);
        }

        private void AdjustVisibleForFlakes(bool a_Sensitive)
        {
            f_BottomGroupBox.Visible = f_TopGroupBox.Visible = !a_Sensitive;
            f_GreetingLabel.Visible = (false == a_Sensitive &&
                                       TimeOfYear.NeutralButSnowy < f_TimeOfYear);
            f_OKButton.Visible = a_Sensitive;
        }

        private void MailToLink_Clicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            // We need this check, since the MS infrastructure seems
            // to be buggy and allows the keyboard to trigger this
            // event if the control is invisible.
            if (false == f_MailToLink.Visible)
            {
                return;
            }

            Util.SetClipboard(Software.c_ProjectEmailAddr);

            string l_Msg = f_Inst.f_TextSet.Format(TextKeys.AboutDlgMailAddrCopiedToClipboardMsg, Software.c_AppName, Software.c_ProjectEmailAddr);
            f_Inst.f_MainForm.InfoMsgBox(l_Msg);
        }

        private void VisitSiteLink_Clicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            // See comment in MailToLink_Clicked().
            if (false == f_VisitSiteLink.Visible)
            {
                return;
            }

            ExecuteLinkOrPrintFailMessage(Software.c_ProjectURL);
        }

        private void ExecuteLinkOrPrintFailMessage(string a_Link)
        {
            try
            {
                Process.Start(a_Link);
            }
            catch
            {
                f_Inst.f_MainForm.ShowOperationFailedMsg();
            }
        }

        private void PrepareLabel(Label a_Label)
        {
            a_Label.TextAlign = ContentAlignment.MiddleCenter;
            a_Label.Width = f_LabelWidth;
        }

        private static TimeOfYear DetermineTimeOfYear()
        {
            if (null != Software.TryGetWineVersion())
            {
                // Effects don't look any good on Wine (as of Wine
                // v1.6.2).

                return TimeOfYear.Neutral;
            }

            DateTime l_Date = DateTime.Now.Date;

            System.Globalization.GregorianCalendar l_Greg = new System.Globalization.GregorianCalendar(System.Globalization.GregorianCalendarTypes.USEnglish);
            uint l_GregYear = (uint)l_Greg.GetYear(l_Date);
            uint l_GregMonth = (uint)l_Greg.GetMonth(l_Date);
            uint l_GregDay = (uint)l_Greg.GetDayOfMonth(l_Date);
            TimeOfYear l_TimeOfYear = TimeOfYear.Neutral;

            if (11 <= l_GregMonth)
            {
                if (12 == l_GregMonth &&
                    24 <= l_GregDay)
                {
                    l_TimeOfYear = TimeOfYear.Xmas;
                }
                else if (GetBeginOfAdventSeason(l_GregYear) <= l_Date)
                {
                    l_TimeOfYear = TimeOfYear.NeutralButSnowy;
                }
            }
            else if (1 == l_GregMonth &&
                     7 >= l_GregDay)
            {
                l_TimeOfYear = TimeOfYear.NewGregorianYear;
            }

            bool l_UsesGreg = (typeof(System.Globalization.GregorianCalendar) == System.Globalization.CultureInfo.CurrentCulture.Calendar.GetType());

            if (false == l_UsesGreg &&
                TimeOfYear.NeutralButSnowy < l_TimeOfYear)
            {
                l_TimeOfYear = TimeOfYear.NeutralButSnowy;
            }

            return l_TimeOfYear;
        }

        private static DateTime GetBeginOfAdventSeason(uint a_GregorianYear)
        {
            // Go back four sundays from and including 24-dec-y.
            DateTime l_BeginDate = new DateTime((int)a_GregorianYear, 12, 24, 0, 0, 0, 0, Util.s_InvariantCulture.Calendar, DateTimeKind.Local);
            uint l_DaysToGoBack = ((3 * 7) + (uint)l_BeginDate.DayOfWeek);

            return l_BeginDate.AddTicks(-(l_DaysToGoBack * TimeSpan.TicksPerDay));
        }

        private enum TimeOfYear : byte
        {
            Neutral = 0,
            NeutralButSnowy = 1,
            Xmas = 2,
            NewGregorianYear = 3,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct WinterInNB : IDisposable
        {
            // Don't use the struct default constructor. Structure
            // must be disposed.

            private readonly AboutDialog f_Dlg;
            // f_Brushes only contains framework brushes that do not
            // require disposing.
            private readonly Brush[] f_Brushes;
            private FlexArray<Flake> f_Flakes;
            private readonly GuiTimer f_Timer;
            // f_MinFlakes and f_MaxExtraFlakes may not be zero (after
            // initialization).
            private readonly uint f_MinFlakes;
            private readonly uint f_MaxExtraFlakes;
            private float f_WindCycle;
            private SizeF f_WindStrength;

            // f_InsensitiveMap is owned by this thread.
            private readonly Bitmap f_InsensitiveMap;
            private Point f_MapPos;

            public WinterInNB(AboutDialog a_Dlg)
            {
                const ushort l_TimerInterval = 50;
                const float l_SinkSpeed = 3.75f;
                uint l_InvDensity = (50 - ((uint)DateTime.UtcNow.Hour / 2));

                f_Dlg = a_Dlg;
                Size l_DlgSize = f_Dlg.Size;
                Size l_DlgClientSize = f_Dlg.ClientSize;
                f_Brushes = new Brush[] { Brushes.Snow, Brushes.White, Brushes.WhiteSmoke, Brushes.Silver, Brushes.FloralWhite, Brushes.Cornsilk, Brushes.AntiqueWhite };
                FlexArray<Flake>.Create(out f_Flakes);
                f_Timer = new GuiTimer(false);
                f_Timer.Interval = l_TimerInterval;
                f_Timer.e_AdvancedTick += f_Dlg.FlakesTimer_AdvTick;
                uint l_FlakeDensitiy = ((uint)f_Dlg.ClientSize.Width / l_InvDensity);
                f_MinFlakes = Math.Max(1, l_FlakeDensitiy / 2);
                f_MaxExtraFlakes = Math.Max(1, l_FlakeDensitiy);
                f_WindCycle = (Util.GetRandomFactorFast() * 6);
                f_WindStrength = new SizeF(Util.GetRandomFactorFast(), l_SinkSpeed);

                // Since performance problems force us to speed up
                // frame presentation, we will not actually draw
                // insensitive controls in the normal fashion with every
                // frame. Instead we will present a stock image of the
                // (insensitive) background. So capture the background.

                f_Dlg.AdjustVisibleForFlakes(false);
                {
                    f_InsensitiveMap = new Bitmap(l_DlgSize.Width, l_DlgSize.Height);
                    f_Dlg.DrawToBitmap(f_InsensitiveMap, new Rectangle(Point.Empty, f_InsensitiveMap.Size));
                }
                f_Dlg.AdjustVisibleForFlakes(true);

                // We need to apply a correction when drawing the
                // background, basically we need to subtract window
                // title bar and the (left/top) window border.

                int l_FrameBorderWidth = ((l_DlgSize.Width - l_DlgClientSize.Width) / 2);
                f_MapPos = new Point(-l_FrameBorderWidth, -(l_DlgSize.Height - l_DlgClientSize.Height) + l_FrameBorderWidth);
            }

            public void Dispose()
            {
                // Can be regardless of initilization.

                if (true == IsValid())
                {
                    f_Timer.Dispose();
                    f_InsensitiveMap.Dispose();
                }
            }

            public bool IsValid()
            {
                return (null != f_Dlg);
            }

            public void AdjustSnow(bool a_Enabled)
            {
                f_Timer.Enabled = a_Enabled;
            }

            public void HandleTick(uint a_Times)
            {
                for (uint i = 0; i != a_Times; i++)
                {
                    MoveFlakes();

                    uint l_NewFlakes = (f_MinFlakes + (uint)Util.GetRandomIndexFast((int)f_MaxExtraFlakes));
                    AddSomeFlakes(l_NewFlakes);
                }

                f_Dlg.Invalidate();
            }

            private void AddSomeFlakes(uint a_Count)
            {
                if (0 == a_Count)
                {
                    return;
                }

                uint l_Idx = (uint)f_Flakes.p_Count;
                f_Flakes.ReserveRange((int)l_Idx, (int)a_Count);
                Flake[] l_Arr = f_Flakes.p_Arr;
                uint l_RelevantWidth = (uint)f_Dlg.ClientSize.Width;
                int l_BrushCount = f_Brushes.Length;

                for (uint i = 0, k = 0, l_Rand = 0; i != a_Count; i++, k--)
                {
                    if (0 == k)
                    {
                        k = (Util.c_BitsPerUInt / 2);
                        l_Rand = Util.GetRandomUIntFast();
                    }

                    Brush l_Brush = f_Brushes[Util.GetRandomIndexFast(l_BrushCount)];
                    uint l_Width = (0 == (l_Rand & 1) ? 1u : 2u);
                    uint l_Height = (0 == (l_Rand & 2) ? 1u : 2u);
                    l_Rand >>= 2;
                    uint l_X = (Util.GetRandomUIntFast() % l_RelevantWidth);
                    l_Arr[l_Idx + i] = new Flake(l_Brush, l_Width, l_Height, new Point((int)l_X, 1 - (int)l_Height));
                }
            }

            private void MoveFlakes()
            {
                uint l_Count = (uint)f_Flakes.p_Count;

                if (0 == l_Count)
                {
                    return;
                }

                // Adjust wind.

                const float l_2xPI = 6.2831853072f;

                f_WindCycle += (l_2xPI / 2513f);
                uint l_RandVal = Util.GetRandomUIntFast();

                if ((uint.MaxValue / 100) > l_RandVal)
                {
                    if ((uint.MaxValue / 200) > l_RandVal)
                    {
                        f_WindCycle += (l_2xPI / 12f);
                    }
                    else
                    {
                        f_WindCycle -= (l_2xPI / 24f);
                    }
                }

                if (f_WindCycle >= l_2xPI)
                {
                    f_WindCycle -= l_2xPI;
                }

                f_WindStrength.Width = ((float)Math.Sin(f_WindCycle) * .5f);

                // Move flakes and drop flakes that have fallen beyond the
                // visible area.

                Flake[] l_Arr = f_Flakes.p_Arr;
                Size l_DlgClientSize = f_Dlg.ClientSize;

                for (uint i = l_Count; 0 != i--; )
                {
                    int l_Top = l_Arr[i].ApplyWind(ref f_WindStrength, (uint)l_DlgClientSize.Width);

                    if (l_Top >= l_DlgClientSize.Height)
                    {
                        f_Flakes.DeleteAt((int)i);
                        l_Arr = f_Flakes.p_Arr;
                    }
                }
            }

            public void Draw(Graphics a_G)
            {
                // Gets called from f_Dlg.OnPaint().

                a_G.DrawImage(f_InsensitiveMap, f_MapPos);

                uint l_Count = (uint)f_Flakes.p_Count;
                Flake[] l_Arr = f_Flakes.p_Arr;

                for (uint i = 0; i != l_Count; i++)
                {
                    Rectangle l_Rect;
                    l_Arr[i].GetRect(out l_Rect);
                    a_G.FillRectangle(l_Arr[i].f_Brush, l_Rect);
                }
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct Flake
            {
                // f_Brush is not owned by this structure.
                public readonly Brush f_Brush;
                public readonly byte f_Width;
                public readonly byte f_Height;
                public readonly byte f_WindFactor;
                private PointF f_Pos;

                public Flake(Brush a_Brush,
                                 uint a_Width,
                                 uint a_Height,
                                 Point a_Pos)
                {
                    f_Brush = a_Brush;
                    f_Width = (byte)a_Width;
                    f_Height = (byte)a_Height;
                    f_WindFactor = (byte)((uint)a_Width * a_Height);
                    f_Pos = a_Pos;
                }

                public void GetRect(out Rectangle a_Rect)
                {
                    a_Rect = new Rectangle((int)f_Pos.X, (int)f_Pos.Y, f_Width, f_Height);
                }

                public int ApplyWind(ref SizeF a_WindStrength,
                                     uint a_RelevantWidth)
                {
                    float l_RandFactor = Math.Max(.75f, Util.GetRandomFactorFast());
                    f_Pos.X += (a_WindStrength.Width * l_RandFactor * f_WindFactor);

                    // We assume three iteration suffice in any case.
                    for (uint k = 0; 3 != k; k++)
                    {
                        if (0 > f_Pos.X)
                        {
                            f_Pos.X += a_RelevantWidth;
                        }
                        else if (a_RelevantWidth < f_Pos.X)
                        {
                            f_Pos.X -= a_RelevantWidth;
                        }
                        else
                        {
                            break;
                        }
                    }

                    f_Pos.Y += (a_WindStrength.Height * l_RandFactor);

                    return (int)f_Pos.Y;
                }
            }
        }
    }

    abstract class CollectionAndImportDialog : ButtonDialog
    {
        private const ushort c_RightSideWidth = 250;

        protected readonly Instance f_Inst;

        private readonly GuiGroupBox f_GroupBox;
        protected readonly TextGroup f_TitleText;
        private readonly TextGroup f_FileCountText;
        private readonly TextGroup f_SizeText;
        private readonly GuiButton f_SelectNoneButton;
        private readonly GuiButton f_SelectAllButton;
        private readonly LabeledIconComboBox<Category> f_CategoryBox;
        protected readonly GuiCheckBox f_CopyAttributesBox;
        protected readonly GridView<int, bool> f_GridView;

        private uint f_ItemCount;
        private ulong f_TotalSize;

        private uint f_SelCount;
        private ulong f_SelSize;

        public CollectionAndImportDialog(Instance a_Inst,
                                         ColumnSet a_ColSet,
                                         TextKeys a_DlgTitleTextKeys,
                                         bool a_ShowCopyAttrbsBox) :
                                    base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            f_Inst = a_Inst;

            this.SuspendLayout();

            Size l_DlgSize = base.SuggestWindowSize(Lyt.TranformSize(550, 400));

            f_GroupBox = new GuiGroupBox(this);
            int l_GroupBoxWidth = (l_DlgSize.Width - Lyt.s_Space2);
            Size l_InnerSize = new Size(l_GroupBoxWidth - f_GroupBox.Padding.Horizontal, l_DlgSize.Height - Lyt.s_Space2 - f_GroupBox.Padding.Vertical);
            int l_SecondWidth = Lyt.CalcX(c_RightSideWidth);
            int l_FirstWidth = (l_InnerSize.Width - l_SecondWidth - Lyt.s_Space);
            int l_Left = f_GroupBox.Padding.Left;
            int l_Right = (l_GroupBoxWidth - f_GroupBox.Padding.Right);
            int l_SecondLeft = (l_Right - l_SecondWidth);

            f_TitleText = TextGroup.CreateLargeLeft(f_GroupBox, l_Left, f_GroupBox.Padding.Top);
            int l_TopLevelTwo = f_TitleText.GetBottom();
            f_FileCountText = TextGroup.CreateLargeLeft(f_GroupBox, l_Left, l_TopLevelTwo);
            f_SizeText = TextGroup.CreateLargeLeft(f_GroupBox, l_Left, f_FileCountText.GetBottom());
            f_SelectNoneButton = new GuiButton(f_GroupBox);
            f_SelectAllButton = new GuiButton(f_GroupBox);
            f_CategoryBox = new LabeledIconComboBox<Category>(f_GroupBox, l_SecondWidth, l_SecondWidth);
            f_CopyAttributesBox = new GuiCheckBox(f_GroupBox, l_SecondWidth);
            f_GridView = new GridView<int, bool>(f_Inst, f_GroupBox, a_ColSet);

            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_TitleText.SetWidthAndAlignLeft(l_InnerSize.Width);
            f_FileCountText.SetWidthAndAlignLeft(l_FirstWidth);
            f_SizeText.SetWidthAndAlignLeft(l_FirstWidth);
            f_SelectNoneButton.Location = new Point(l_Left, f_SizeText.GetBottom() + Lyt.s_SmallSpace);
            f_SelectAllButton.Location = new Point(f_SelectNoneButton.GetRight(), f_SelectNoneButton.Top);
            f_CategoryBox.SetLocation(l_SecondLeft, l_TopLevelTwo);
            f_CopyAttributesBox.Location = new Point(l_SecondLeft, f_CategoryBox.GetBottom());
            int l_RightBottom;

            if (false == a_ShowCopyAttrbsBox)
            {
                f_CopyAttributesBox.Visible = false;

                l_RightBottom = f_CopyAttributesBox.Top;
            }
            else
            {
                l_RightBottom = f_CopyAttributesBox.GetBottom();
            }

            f_GridView.Dock = DockStyle.None;
            f_GridView.Location = new Point(l_Left, Math.Max(f_SelectAllButton.GetBottom() + Lyt.s_SmallSpace, l_RightBottom));
            f_GridView.Size = new Size(l_InnerSize.Width, l_InnerSize.Height - f_GridView.Top);
            f_GridView.LoadColumnAdjustments();
            f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_GridView.Bottom);

            AsFixedButtonDialog(l_DlgSize.Width, f_GroupBox);
            base.SetSelectOnShowControl(f_GridView);

            f_OKButton.Click += OKButton_Click;
            f_SelectNoneButton.Click += SelectNoneButton_Click;
            f_SelectAllButton.Click += SelectAllButton_Click;
            f_GridView.CellClick += GridView_CellClick;

            TextSet l_TextSet = f_Inst.f_TextSet;
            this.Text = l_TextSet[a_DlgTitleTextKeys];
            // NOTE: f_TitleText is managed by the child classes.
            f_GroupBox.Text = l_TextSet[TextKeys.OpenCollDlgGroupBoxCaption];
            f_FileCountText.SetLeftText(l_TextSet, TextKeys.OpenCollDlgNumOfFiles);
            f_SizeText.SetLeftText(l_TextSet, TextKeys.OpenCollDlgSelectedSize);
            f_SelectNoneButton.Text = l_TextSet[TextKeys.OpenCollDlgSelectNone];
            f_SelectAllButton.Text = l_TextSet[TextKeys.OpenCollDlgSelectAll];
            f_CategoryBox.SetCaption(l_TextSet, TextKeys.OpenCollDlgAddToCustomTab);
            f_CopyAttributesBox.SetTextAndUpdateSize(l_TextSet[TextKeys.ImportPartialDlsDlgCopyAttributes]);

            // Finish initialization.

            f_Inst.f_MainForm.f_Categories.FillComboWithCustomCategories(f_CategoryBox);

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_GridView.StoreColumnAdjustments();

                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(disposing);
        }

        public DialogResult LoadAndShowDialog(IWin32Window a_Owner)
        {
            if (false == LoadItems(out f_ItemCount, out f_TotalSize))
            {
                return DialogResult.Cancel;
            }

            // Create items.

            f_GridView.StartChanging();

            for (int i = ((int)f_ItemCount) - 1; 0 <= i; i--)
            {
                f_GridView.CreateItem(i);
            }

            f_GridView.EndChanging();

            AdjustEnabledState(true, true);

            return ShowDialog(a_Owner);
        }

        // Returns true if loading succeeded.
        protected abstract bool LoadItems(out uint a_ItemCount,
                                          out ulong a_TotalSize);

        private void UpdateLabels()
        {
            TextSet l_TextSet = f_Inst.f_TextSet;
            MainForm l_Form = f_Inst.f_MainForm;
            f_FileCountText.SetRightText(l_TextSet.Format(TextKeys.OpenCollDlgFormatXOfN, l_TextSet.NumStr(f_SelCount), l_TextSet.NumStr(f_ItemCount)));
            f_SizeText.SetRightText(l_TextSet.Format(TextKeys.OpenCollDlgFormatXOfN, l_Form.BuildDefaultSizeStr(f_SelSize), l_Form.BuildDefaultSizeStr(f_TotalSize)));
        }

        private void SelectNoneButton_Click(object sender, EventArgs e)
        {
            AdjustEnabledState(false, false);
        }

        private void SelectAllButton_Click(object sender, EventArgs e)
        {
            AdjustEnabledState(true, false);
        }

        private void AdjustEnabledState(bool a_Set,
                                        bool a_Force)
        {
            if (false == a_Force)
            {
                if (false == a_Set)
                {
                    if (0 == f_ItemCount)
                    {
                        return;
                    }
                }
                else
                {
                    if (f_SelCount == f_ItemCount)
                    {
                        return;
                    }
                }
            }

            AdjustStateOfAllItems(a_Set);
            f_GridView.NotifyValuesChanged(true);

            if (false == a_Set)
            {
                f_SelCount = 0;
                f_SelSize = 0;
            }
            else
            {
                f_SelCount = f_ItemCount;
                f_SelSize = f_TotalSize;
            }

            UpdateLabels();
        }

        protected abstract void AdjustStateOfAllItems(bool a_State);

        private void GridView_CellClick(object sender, DataGridViewCellEventArgs e)
        {
            int l_RowIdx = e.RowIndex;

            if (0 != e.ColumnIndex ||
                0 > l_RowIdx)
            {
                return;
            }

            int l_ArrIdx = f_GridView.GetItem(l_RowIdx).f_Ref;
            ulong l_Size;
            bool l_NewState = FlipItemState(l_ArrIdx, out l_Size);

            f_GridView.NotifyValuesChanged(false);
            f_GridView.UpdateCellValue(0, l_RowIdx);

            if (false == l_NewState)
            {
                f_SelCount--;
                f_SelSize -= l_Size;
            }
            else
            {
                f_SelCount++;
                f_SelSize += l_Size;
            }

            UpdateLabels();
        }

        // Returns the new state of the item.
        protected abstract bool FlipItemState(int a_ArrIdx,
                                              out ulong a_Size);

        private void OKButton_Click(object sender, EventArgs e)
        {
            LabeledIconComboBox<Category>.Item l_Item = (LabeledIconComboBox<Category>.Item)f_CategoryBox.f_Box.SelectedItem;
            bool l_Close = OKButtonClicked(f_SelCount, f_SelSize, l_Item.f_Tag);

            if (true == l_Close)
            {
                this.DialogResult = DialogResult.OK;
                this.Close();
            }
        }

        // Returns true if the dialog should close.
        protected abstract bool OKButtonClicked(uint a_SelCount,
                                                ulong a_SelSize,
                                                Category a_Category);
    }

    sealed class ImportPartialDownloadsDialog : CollectionAndImportDialog
    {
        private readonly string f_SuggestedPath;

        private string f_PartFilesDir;
        private ImportableFile[] f_Files;

        public ImportPartialDownloadsDialog(Instance a_Inst,
                                            string a_SuggestedPath) :
                                       base(a_Inst, a_Inst.f_MainForm.f_VisualCust.f_ImportPartialDlsDlgColSet, TextKeys.ImportPartialDlsDlgTitle, true)
        {
            f_SuggestedPath = a_SuggestedPath;

            f_CopyAttributesBox.Checked = true;

            TextSet l_TextSet = a_Inst.f_TextSet;
            this.f_OKButton.Text = l_TextSet[TextKeys.ImportPartialDlsDlgImport];
            this.f_CancelButton.Text = l_TextSet[TextKeys.ButtonTextClose];

            f_GridView.e_UpdateItem += GridView_UpdateItem;

            f_TitleText.SetLeftText(l_TextSet, TextKeys.ImportPartialDlsDlgSourceFolder);
        }

        protected override bool LoadItems(out uint a_ItemCount,
                                          out ulong a_TotalSize)
        {
            LoadMetFilesContext l_Cntxt = new LoadMetFilesContext(f_SuggestedPath);
            l_Cntxt.StartAsync();

            if (true == WaitMessageDialog.ShowDialog(f_Inst.f_TextSet, f_Inst.f_MainForm, l_Cntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.ImportPartialDlsLoadingDlgTitle, TextKeys.ImportPartialDlsLoadingDlgMsg))
            {
                LoadMetFilesContext.ErrorCodes l_ErrCode = l_Cntxt.GetErrorCode();

                if (LoadMetFilesContext.ErrorCodes.Success == l_ErrCode)
                {
                    // We did it.

                    f_Files = l_Cntxt.GetFiles(out f_PartFilesDir, out a_TotalSize);
                    a_ItemCount = (uint)f_Files.Length;

                    f_TitleText.SetRightText(f_PartFilesDir);

                    return true;
                }

                TextSet l_TextSet = f_Inst.f_TextSet;
                string l_ErrMsg;

                switch (l_ErrCode)
                {
                    case LoadMetFilesContext.ErrorCodes.AbortedByUser:
                        l_ErrMsg = l_TextSet[TextKeys.MsgOperationAbortedByUser];
                        break;

                    case LoadMetFilesContext.ErrorCodes.TempDirContainsNoUseableFiles:
                        l_ErrMsg = "No files detected that might be imported.";
                        break;

                    case LoadMetFilesContext.ErrorCodes.FailedToScanTempPath:
                        l_ErrMsg = "An error occurred while scanning for importable files.";
                        break;

                    default:
                        l_ErrMsg = l_TextSet[TextKeys.MsgOperationFailed];
                        break;
                }

                f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);
            }

            a_ItemCount = 0;
            a_TotalSize = 0;

            return false;
        }

        protected override void AdjustStateOfAllItems(bool a_State)
        {
            foreach (ImportableFile l_File in f_Files)
            {
                l_File.f_ImportFlag = a_State;
            }
        }

        protected override bool FlipItemState(int a_ArrIdx,
                                              out ulong a_Size)
        {
            ImportableFile l_File = f_Files[a_ArrIdx];
            a_Size = l_File.p_SizeForImportDlg;

            bool l_NewState = !l_File.f_ImportFlag;
            l_File.f_ImportFlag = l_NewState;

            return l_NewState;
        }

        protected override bool OKButtonClicked(uint a_SelCount,
                                                ulong a_SelSize,
                                                Category a_Category)
        {
            if (0 == a_SelCount)
            {
                return false;
            }

            MainForm l_Form = f_Inst.f_MainForm;

            if (true == l_Form.CheckIfiMuleIsRunningWithErrMsg())
            {
                return false;
            }

            bool l_CopyAttributes = f_CopyAttributesBox.Checked;
            ContentManager.MultiFileImportContext l_Cntxt = new ContentManager.MultiFileImportContext(f_PartFilesDir, f_Files, a_SelSize, l_CopyAttributes, a_Category);
            f_Inst.f_ContentMngr.GuiBeginImport(l_Cntxt);

            if (false == WaitMessageDialog.ShowDialog(f_Inst.f_TextSet, f_Inst.f_MainForm, l_Cntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.ImportPartialDlsImportingDlgTitle, TextKeys.ImportPartialDlsImportingDlgMsg))
            {
                // Exit.

                return true;
            }

            ulong l_ElapsedSize;
            ulong l_TranslatedSize;
            uint l_FilesCreated;
            uint l_FilesSuccessfullyImported;
            ContentManager.MultiFileImportContext.ResultCodes l_Res = l_Cntxt.QueryStats(out l_ElapsedSize, out l_TranslatedSize, out l_FilesCreated, out l_FilesSuccessfullyImported);
            bool l_Succ = (ContentManager.MultiFileImportContext.ResultCodes.Done == l_Res);

            TextSet l_TextSet = f_Inst.f_TextSet;
            StringBuilder l_Bld = new StringBuilder();

            switch (l_Res)
            {
                case ContentManager.MultiFileImportContext.ResultCodes.Done:
                    l_Bld.Append("Import finished successfully.");
                    break;

                case ContentManager.MultiFileImportContext.ResultCodes.AbortedThroughController:
                    l_Bld.Append(l_TextSet[TextKeys.MsgOperationAbortedByUser]);
                    break;

                case ContentManager.MultiFileImportContext.ResultCodes.LackOfAvailDiskSpace:
                    l_Bld.Append("Import couldn't finish successfully, since the level of available disk space is critical.");
                    break;

                default:
                    l_Bld.Append("Import was unsuccessful.");
                    break;
            }

            ulong l_BytesNotTakenOver = (l_ElapsedSize - l_TranslatedSize);
            uint l_FilesNotImported = (a_SelCount - l_FilesSuccessfullyImported);

            l_Bld.AppendLine();
            l_Bld.AppendLine();
            DebugAddTextKey(l_Bld, "Relevant size processed", l_Form.BuildDefaultSizeStr(l_ElapsedSize));

            if (true == l_Succ &&
                0 != l_BytesNotTakenOver)
            {
                DebugAddTextKey(l_Bld, "Amount that was not imported due to incompatibility", l_Form.BuildDefaultSizeStr(l_BytesNotTakenOver));
            }

            DebugAddTextKey(l_Bld, "Files that were successfully processed", l_TextSet.NumStr(l_FilesSuccessfullyImported));

            if (0 != l_FilesNotImported)
            {
                DebugAddTextKey(l_Bld, "Files that could not be processed successfully", l_TextSet.NumStr(l_FilesNotImported));
            }

            DebugAddTextKey(l_Bld, "Number of files created", l_TextSet.NumStr(l_FilesCreated));

            string l_Msg = l_Bld.ToString();

            if (false == l_Succ)
            {
                l_Form.WarnMsgBox(l_Msg);
            }
            else
            {
                l_Form.InfoMsgBox(l_Msg);
            }

            // Do not return true here, since we don't want to make
            // the dialog close.
            return false;
        }

        private void DebugAddTextKey(StringBuilder a_Bld,
                                     string a_Key,
                                     string a_Value)
        {
            a_Bld.AppendLine(a_Key + ": " + a_Value);
        }

        private void GridView_UpdateItem(Time a_Time,
                                         GridViewItem<int, bool> a_Item)
        {
            ImportableFile l_File = f_Files[a_Item.f_Ref];
            UInt128 l_FileID = l_File.f_FileID.f_Val;

            a_Item.SetBool(0, l_File.f_ImportFlag);
            a_Item.SetTextWithCompObj(1, l_File.f_FileName, l_File.f_FileName);
            a_Item.SetTextWithCompObj(2, l_FileID, l_FileID.ToHexString());
            a_Item.SetTextWithCompObj(3, l_File.f_ContentFileName, l_File.f_ContentFileName);
            a_Item.SetText(4, l_File.f_FileSize, f_Inst.f_MainForm.BuildDefaultSizeStr(l_File.f_FileSize));
            a_Item.SetText(5, l_File.f_Done, f_Inst.f_MainForm.BuildDefaultSizeStr(l_File.f_Done));
            a_Item.SetDrawContext(6, OwnerDrawContext.CreateIncomplFileProgressBarContext(false, l_File.f_Progress));
            a_Item.SetText(7, l_File.f_Fragments, f_Inst.f_TextSet.NumStr(l_File.f_Fragments));
            a_Item.SetBool(8, l_File.p_HasPartHashes);
            a_Item.SetText(9, (byte)l_File.f_DlControl, FilesPanel.GetDlControlStr(DownloadControl.DlEnabled == l_File.f_DlControl));
            a_Item.SetText(10, l_File.f_DlPriority.f_Value, PriorityPickerControl.PriorityToString(l_File.f_DlPriority, string.Empty));
            a_Item.SetText(11, l_File.f_UlPriority.f_Value, PriorityPickerControl.PriorityToString(l_File.f_UlPriority, string.Empty));
        }
    }

    sealed class OpenCollectionDialog : CollectionAndImportDialog
    {
        private readonly string f_FileName;

        private FileCollection f_Collection;
        private FileCollection.FileInfo[] f_Infos;

        public OpenCollectionDialog(Instance a_Inst,
                                    string a_FileName) :
                               base(a_Inst, a_Inst.f_MainForm.f_VisualCust.f_OpenCollectionDlgColSet, TextKeys.OpenCollDlgTitle, false)
        {
            f_FileName = a_FileName;

            f_GridView.e_UpdateItem += GridView_UpdateItem;

            f_TitleText.SetLeftText(a_Inst.f_TextSet, TextKeys.OpenCollDlgTitleLabel);
        }

        protected override bool LoadItems(out uint a_ItemCount,
                                          out ulong a_TotalSize)
        {
            string l_SubErrMsg;
            f_Collection = OpenCollectionFileCore(out l_SubErrMsg);

            if (null == f_Collection)
            {
                f_Inst.f_MainForm.ErrMsgBox(f_Inst.f_TextSet.BuildFileErrMsg(TextKeys.MsgCollErrLoadingFromFile, f_FileName, l_SubErrMsg));

                a_ItemCount = 0;
                a_TotalSize = 0;

                return false;
            }

            f_Infos = f_Collection.ExposeArray();

            a_ItemCount = f_Collection.p_TotalCount;
            a_TotalSize = f_Collection.f_TotalSize;

            f_TitleText.SetRightText(f_Collection.p_Title);

            return true;
        }

        private FileCollection OpenCollectionFileCore(out string a_SubErrMsg)
        {
            // Return can be null, in which case a_SubErrMsg is valid.

            TextSet l_TextSet = f_Inst.f_TextSet;

            Exception l_Exc;
            FHandle l_SrcFile = FHandle.OpenR(f_FileName, false, System.IO.FileOptions.None, 0, FHandle.ThreadSafetyMode.NotThreadSafe, out l_Exc);

            if (null == l_SrcFile)
            {
                a_SubErrMsg = l_TextSet[TextKeys.MsgFileErrAccessFailed];

                return null;
            }

            FileCollection.LoadResult l_Result;
            FileCollection l_Collection = null;

            try
            {
                FileIO l_File = FileIO.CreateReadable(l_SrcFile);
                l_Result = FileCollection.CreateFromBuffer(l_File, out l_Collection);
            }
            catch (BufferIO.InterpretBufferException)
            {
                l_Result = FileCollection.LoadResult.Corruption;
            }
            catch (BufferIO.FileAccessException)
            {
                a_SubErrMsg = l_TextSet[TextKeys.MsgFileErrReadFailed];

                return null;
            }
            finally
            {
                l_SrcFile.Close();
            }

            if (FileCollection.LoadResult.Success != l_Result)
            {
                switch (l_Result)
                {
                    case FileCollection.LoadResult.NoCollection:
                        a_SubErrMsg = l_TextSet[TextKeys.MsgCollErrCantInterpretFile];
                        break;

                    case FileCollection.LoadResult.TooAdvanced:
                        a_SubErrMsg = l_TextSet[TextKeys.MsgCollErrTooAdvancedFile];
                        break;

                    default:
                        a_SubErrMsg = string.Empty;
                        break;
                }

                return null;
            }

            a_SubErrMsg = null;

            return l_Collection;
        }

        protected override void AdjustStateOfAllItems(bool a_State)
        {
            f_Collection.AdjustEnabledState(a_State);
        }

        protected override bool FlipItemState(int a_ArrIdx,
                                              out ulong a_Size)
        {
            a_Size = f_Infos[a_ArrIdx].f_FileSize;

            bool l_NewState = !f_Infos[a_ArrIdx].f_Enabled;
            f_Infos[a_ArrIdx].f_Enabled = l_NewState;

            return l_NewState;
        }

        protected override bool OKButtonClicked(uint a_SelCount,
                                                ulong a_SelSize,
                                                Category a_Category)
        {
            if (0 == a_SelCount)
            {
                return true;
            }

            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.GuiAddFilesFromCollection, new object[] { f_Infos, a_Category });
            if (false == f_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
            {
                return false;
            }

            if (CookieStates.Success != l_Cookie.p_State)
            {
                // Should never happen.

                return false;
            }

            uint l_Done = (uint)l_Cookie.GetReturnObject();

            if (a_SelCount > l_Done)
            {
                string l_Str = f_Inst.f_TextSet.Format(TextKeys.OpenCollDlgFailedToTakeOver, a_SelCount - l_Done);
                f_Inst.f_MainForm.WarnMsgBox(l_Str);
            }

            return true;
        }

        private void GridView_UpdateItem(Time a_Time,
                                         GridViewItem<int, bool> a_Item)
        {
            int l_Idx = a_Item.f_Ref;

            a_Item.SetBool(0, f_Infos[l_Idx].f_Enabled);
            a_Item.SetTextWithCompObj(1, f_Infos[l_Idx].f_FileName, f_Infos[l_Idx].f_FileName);
            a_Item.SetText(2, f_Infos[l_Idx].f_FileSize, f_Inst.f_MainForm.BuildDefaultSizeStr(f_Infos[l_Idx].f_FileSize));
            a_Item.SetText(3, 0, f_Infos[l_Idx].f_MasterHashes.eD2kHashAsStr(string.Empty));
            a_Item.SetText(4, 0, f_Infos[l_Idx].f_MasterHashes.AICHHashAsStr(string.Empty));
            a_Item.SetText(5, 0, f_Infos[l_Idx].f_MasterHashes.TAICHHashAsStr(string.Empty));
        }
    }

    sealed class CreateCollectionDialog : ButtonDialog
    {
        private const ushort c_DlgWidth = 400;

        private readonly Instance f_Inst;
        private readonly FileCollection f_Collection;

        private readonly GuiGroupBox f_GroupBox;
        private readonly LabeledTextBox f_TitleBox;
        private readonly TextGroup f_FileCountText;
        private readonly TextGroup f_SizeText;

        public CreateCollectionDialog(Instance a_Inst,
                                      FileCollection a_Collection) :
                                 base(a_Inst.f_TextSet, DialogButtons.OKCancel)
        {
            f_Inst = a_Inst;
            f_Collection = a_Collection;

            this.SuspendLayout();

            f_GroupBox = new GuiGroupBox(this);
            int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
            int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            int l_LeftPadding = f_GroupBox.Padding.Left;
            f_TitleBox = new LabeledTextBox(f_GroupBox, l_InnerWidth);
            f_TitleBox.SetLocation(l_LeftPadding, f_GroupBox.Padding.Top);
            f_FileCountText = TextGroup.CreateLargeLeft(f_GroupBox, l_LeftPadding, f_TitleBox.GetBottom());
            f_SizeText = TextGroup.CreateLargeLeft(f_GroupBox, l_LeftPadding, f_FileCountText.GetBottom());

            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_FileCountText.SetWidthAndAlignLeft(l_InnerWidth);
            f_SizeText.SetWidthAndAlignLeft(l_InnerWidth);
            f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_SizeText.GetBottom() - Lyt.s_SmallSpace);

            AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
            base.SetSelectOnShowControl(f_TitleBox.f_Box);

            f_TitleBox.f_Box.MaxLength = FileCollection.c_MaxTitleChars;
            f_OKButton.Click += OKButton_Click;

            TextSet l_TextSet = f_Inst.f_TextSet;
            this.Text = l_TextSet[TextKeys.CreateCollDlgTitle];
            f_GroupBox.Text = l_TextSet[TextKeys.CreateCollDlgGroupBoxCaption];
            f_TitleBox.SetCaption(l_TextSet, TextKeys.CreateCollDlgTitleLabel);
            f_FileCountText.SetLeftText(l_TextSet, TextKeys.CreateCollDlgNumOfFiles);
            f_SizeText.SetLeftText(l_TextSet, TextKeys.CreateCollDlgTotalSize);

            this.ResumeLayout();

            UpdateLabels();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();
            }

            base.Dispose(disposing);
        }

        private void UpdateLabels()
        {
            f_TitleBox.f_Box.Text = f_Collection.p_Title;
            f_FileCountText.SetRightText(f_Inst.f_TextSet.NumStr(f_Collection.p_TotalCount));
            f_SizeText.SetRightText(f_Inst.f_MainForm.BuildLongSizeStr(f_Collection.f_TotalSize));
        }

        private void OKButton_Click(object sender, EventArgs e)
        {
            string l_ErrMsg;
            if (false == HandleOK(out l_ErrMsg))
            {
                if (null != l_ErrMsg)
                {
                    f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);
                }

                return;
            }

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        private bool HandleOK(out string a_ErrMsg)
        {
            // a_ErrMsg can be null on return, regardless of the
            // return value.

            string l_Title = f_TitleBox.f_Box.Text.Trim();

            if (0 == l_Title.Length)
            {
                f_TitleBox.f_Box.Select();

                a_ErrMsg = f_Inst[TextKeys.CreateCollDlgMsgInsertTitle];

                return false;
            }

            f_Collection.SetTitle(l_Title);

            using (SaveFileDialog l_Dlg = new SaveFileDialog())
            {
                l_Dlg.Filter = f_Inst.f_MainForm.BuildCollectionFilesFilterStr();

                // Try to turn the title into a suggested file name.

                string l_SuggFileName = FileNames.CleanUpFileName(l_Title);

                if (null != l_SuggFileName)
                {
                    l_Dlg.FileName = l_SuggFileName;
                }

                if (DialogResult.OK != l_Dlg.ShowDialog(this))
                {
                    a_ErrMsg = null;

                    return false;
                }

                string l_FileName = l_Dlg.FileName;
                MemoryIO l_Mem = MemoryIO.CreateWritable();

                if (false == f_Collection.ToBuffer(l_Mem))
                {
                    a_ErrMsg = f_Inst[TextKeys.CreateCollDlgMsgFailedToCreate];

                    return false;
                }

                byte[] l_Arr = l_Mem.ToArray(true);

                try
                {
                    System.IO.File.WriteAllBytes(l_FileName, l_Arr);
                }
                catch (Exception l_Exc)
                {
                    a_ErrMsg = f_Inst.f_TextSet.BuildFileErrMsg(TextKeys.CreateCollDlgMsgFailedToWriteToFile, l_FileName, l_Exc.Message);

                    return false;
                }
            }

            a_ErrMsg = null;

            return true;
        }
    }

    sealed class OptionsDialog : ButtonDialog
    {
        public const byte c_PageCount = 15;
        private const ushort c_ListViewWidth = 210;
        private const ushort c_CaptionHeight = 35;
        private const ushort c_DescriptionHeight = 35;

        private readonly Instance f_Inst;
        private readonly Preferences f_DefaultPrefs;

        private readonly Font f_ListBoxFont;
        private readonly Font f_CaptionBoxFont;

        private readonly ListBox f_ListBox;
        private readonly GuiLabel f_Caption;
        private readonly GuiLabel f_Description;
        private readonly GuiButton f_DefaultsButton;

        private readonly GeneralPanel f_GeneralPanel;
        private readonly LanguagePanel f_LanguagePanel;
        private readonly OSIntegrationPanel f_OSIntegrationPanel;
        private readonly AppDirsPanel f_AppDirsPanel;
        private readonly DisplayPanel f_DisplPanel;
        private readonly ListViewsPanel f_ListViewsPanel;
        private readonly TunnelsPanel f_TunnelsPanel;
        private readonly AddressesPanel f_AddrPanel;
        private readonly BandwidthPanel f_BandwidthPanel;
        private readonly KademliaPanel f_KadPanel;
        private readonly KadFilterPanel f_KadFilterPanel;
        private readonly eD2kPanel f_eD2kPanel;
        private readonly SharedDirsPanel f_SharedDirsPanel;
        private readonly SharedFilesPanel f_SharedFilesPanel;
        private readonly AdvancedPanel f_AdvPanel;
        private readonly OptionsPanel[] f_Panels;

        private int f_SelectedIdx;
        private bool f_IgnoreIndexEvent;

        public OptionsDialog(Instance a_Inst,
                             int[] a_ScrollStates) :
                        base(a_Inst.f_TextSet, DialogButtons.All)
        {
            f_Inst = a_Inst;
            f_DefaultPrefs = Preferences.CreateDefault(f_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_OwnPersona);

            this.SuspendLayout();

            Size l_SuggSz = base.SuggestWindowSize(Lyt.TranformSize(750, 450));
            // Don't make the dialog too broad, since that looks odd.
            l_SuggSz.Width = Math.Min(l_SuggSz.Width, Lyt.CalcX(900));
            AsFixedDialog(l_SuggSz);

            f_ListBox = new ListBox();
            f_Caption = new GuiLabel(this);
            f_Description = new GuiLabel(this);
            f_DefaultsButton = new GuiButton(this);
            this.Controls.Add(f_ListBox);

            f_ListBoxFont = MainForm.RecreateFont(f_ListBox.Font, 10f, FontStyle.Regular);
            f_CaptionBoxFont = MainForm.RecreateFont(f_ListBox.Font, 11.5f, FontStyle.Bold);

            // Unregister f_CancelButton as cancel button, since we
            // don't want this dialog to close when escape is pressed.
            this.CancelButton = null;
            this.DoubleBuffered = true;
            f_ListBox.Font = f_ListBoxFont;
            f_ListBox.Location = f_InnerRect.Location;
            f_ListBox.Size = new Size(Lyt.CalcX(c_ListViewWidth), f_InnerRect.Height);
            f_ListBox.IntegralHeight = false;
            f_Caption.Location = new Point(f_ListBox.Right + Lyt.s_Space, f_InnerRect.Top);
            f_Caption.Size = new Size(f_InnerRect.Width - f_ListBox.Right, Lyt.CalcY(c_CaptionHeight));
            f_Caption.BorderStyle = BorderStyle.Fixed3D;
            f_Caption.Font = f_CaptionBoxFont;
            f_Caption.ForeColor = SystemColors.HighlightText;
            f_Caption.BackColor = SystemColors.MenuHighlight;
            f_Caption.TextAlign = ContentAlignment.MiddleCenter;
            f_Description.Location = new Point(f_Caption.Left, f_Caption.Bottom + Lyt.s_Space);
            f_Description.Size = new Size(f_Caption.Width, Lyt.CalcY(c_DescriptionHeight));
            f_DefaultsButton.Location = new Point(f_InnerRect.Left, f_InnerRect.Bottom + Lyt.s_Space);
            f_DefaultsButton.Width = f_ListBox.Width;

            int l_PanelTop = (f_Description.Bottom + Lyt.s_Space);
            Rectangle l_PanelRect = new Rectangle(f_Caption.Left, l_PanelTop, f_Description.Width, f_InnerRect.Bottom - l_PanelTop);

            f_GeneralPanel = new GeneralPanel(this, ref l_PanelRect);
            f_LanguagePanel = new LanguagePanel(this, ref l_PanelRect);
            f_OSIntegrationPanel = new OSIntegrationPanel(this, ref l_PanelRect);
            f_AppDirsPanel = new AppDirsPanel(this, ref l_PanelRect);
            f_DisplPanel = new DisplayPanel(this, ref l_PanelRect);
            f_ListViewsPanel = new ListViewsPanel(this, ref l_PanelRect);
            f_TunnelsPanel = new TunnelsPanel(this, ref l_PanelRect);
            f_AddrPanel = new AddressesPanel(this, ref l_PanelRect);
            f_BandwidthPanel = new BandwidthPanel(this, ref l_PanelRect);
            f_KadPanel = new KademliaPanel(this, ref l_PanelRect);
            f_KadFilterPanel = new KadFilterPanel(this, ref l_PanelRect);
            f_eD2kPanel = new eD2kPanel(this, ref l_PanelRect);
            f_SharedDirsPanel = new SharedDirsPanel(this, ref l_PanelRect);
            f_SharedFilesPanel = new SharedFilesPanel(this, ref l_PanelRect);
            f_AdvPanel = new AdvancedPanel(this, ref l_PanelRect);
            f_Panels = new OptionsPanel[c_PageCount]
            {
                f_GeneralPanel,
                f_LanguagePanel,
                f_OSIntegrationPanel,
                f_AppDirsPanel,
                f_DisplPanel,
                f_ListViewsPanel,
                f_TunnelsPanel,
                f_AddrPanel,
                f_BandwidthPanel,
                f_KadPanel,
                f_KadFilterPanel,
                f_eD2kPanel,
                f_SharedDirsPanel,
                f_SharedFilesPanel,
                f_AdvPanel,
            };

            f_ListBox.SelectedIndexChanged += ListBox_SelectedIndexChanged;
            f_DefaultsButton.Click += DefaultsButton_Click;
            f_OKButton.Click += OKButton_Click;
            f_CancelButton.Click += DefaultCancelCloseHandler;
            f_ApplyButton.Click += ApplyButton_Click;

            this.Text = f_Inst[TextKeys.OptionsDlgTitle];
            f_Panels[0].SetTexts(TextKeys.OptionsDlgGeneralTitle, TextKeys.OptionsDlgGeneralComment);
            f_Panels[1].SetTexts(TextKeys.OptionsDlgLangTitle, TextKeys.OptionsDlgLangComment);
            f_Panels[2].SetTexts(TextKeys.OptionsDlgOSIntgrTitle, TextKeys.OptionsDlgOSIntgrComment);
            f_Panels[3].SetTexts(TextKeys.OptionsDlgAppDirsTitle, TextKeys.OptionsDlgAppDirsComment);
            f_Panels[4].SetTexts(TextKeys.OptionsDlgDisplTitle, TextKeys.OptionsDlgDisplComment);
            f_Panels[5].SetTexts(TextKeys.OptionsDlgListViewsTitle, TextKeys.OptionsDlgListViewsComment);
            f_Panels[6].SetTexts(TextKeys.OptionsDlgTunnelsTitle, TextKeys.OptionsDlgTunnelsComment);
            f_Panels[7].SetTexts(TextKeys.OptionsDlgAddressTitle, TextKeys.OptionsDlgAddressComment);
            f_Panels[8].SetTexts(TextKeys.OptionsDlgBandwidthTitle, TextKeys.OptionsDlgBandwidthComment);
            f_Panels[9].SetTexts(TextKeys.OptionsDlgKademliaTitle, TextKeys.OptionsDlgKademliaComment);
            f_Panels[10].SetTexts(TextKeys.OptionsDlgKadFilterTitle, TextKeys.OptionsDlgKadFilterComment);
            f_Panels[11].SetTexts(TextKeys.OptionsDlgeD2kTitle, TextKeys.OptionsDlgeD2kComments);
            f_Panels[12].SetTexts(TextKeys.OptionsDlgSharedDirsTitle, TextKeys.OptionsDlgSharedDirsComments);
            f_Panels[13].SetTexts(TextKeys.OptionsDlgSharedFilesTitle, TextKeys.OptionsDlgSharedFilesComment);
            f_Panels[14].SetTexts(TextKeys.OptionsDlgAdvancedTitle, TextKeys.OptionsDlgAdvancedComment);

            for (int i = 0; i < c_PageCount; i++)
            {
                f_Panels[i].p_ScrollPosition = a_ScrollStates[i];
                f_ListBox.Items.Add(f_Panels[i].p_Caption);
            }

            int l_MaxListBoxHeight = (f_ListBox.GetItemRectangle(c_PageCount - 1).Bottom + Lyt.s_SmallSpace);
            f_ListBox.Height = Math.Min(l_MaxListBoxHeight, f_ListBox.Height);

            f_DefaultsButton.Text = f_Inst[TextKeys.OptionsDlgResetToDefaults];

            this.ResumeLayout();

            SetAll(f_Inst.f_MainForm.p_Prefs);
            f_SelectedIdx = (-1);
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                foreach (OptionsPanel l_Cntrl in f_Panels)
                {
                    l_Cntrl.Dispose();
                }

                Dialog.DisposeAllChildren(this);

                f_ListBoxFont.Dispose();
                f_CaptionBoxFont.Dispose();
            }

            base.Dispose(disposing);
        }

        public void CopyScrollStates(int[] a_ScrollStates)
        {
            for (int i = 0; i < c_PageCount; i++)
            {
                a_ScrollStates[i] = f_Panels[i].p_ScrollPosition;
            }
        }

        public void ShowDialog(IWin32Window a_Parent,
                               MainForm.ShowOptionsDlgReasons a_Reason)
        {
            Control l_SelCntrl = null;

            switch (a_Reason)
            {
                case MainForm.ShowOptionsDlgReasons.EditCommentFilter:
                    SetSelectedPanel(PageTypes.Kademlia);
                    l_SelCntrl = f_KadPanel.f_CommentFilterBox.f_Box;
                    break;

                case MainForm.ShowOptionsDlgReasons.SetOutputDir:
                    SetSelectedPanel(PageTypes.SharedDirs);
                    break;

                default:
                    SetSelectedPanel(f_Inst.f_MainForm.f_VisualCust.f_OptionsDlgPage);
                    break;
            }

            base.SetSelectOnShowControl(l_SelCntrl);
            this.ShowDialog(a_Parent);

            f_Inst.f_MainForm.f_VisualCust.f_OptionsDlgPage = (PageTypes)f_SelectedIdx;
            // NOTE: Sometimes the main window loses focus when the
            // options dialog closes.
            f_Inst.f_MainForm.Select();
        }

        private void ListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (false == f_IgnoreIndexEvent)
            {
                SetSelectedPanel((PageTypes)f_ListBox.SelectedIndex);
            }
        }

        private void SetSelectedPanel(PageTypes a_PageType)
        {
            int l_NewSelIdx = (int)a_PageType;

            if (0 <= f_SelectedIdx)
            {
                if (f_SelectedIdx == l_NewSelIdx)
                {
                    return;
                }

                f_Panels[f_SelectedIdx].Hide();
            }

            f_SelectedIdx = l_NewSelIdx;

            f_IgnoreIndexEvent = true;
            f_ListBox.SelectedIndex = f_SelectedIdx;
            f_IgnoreIndexEvent = false;

            f_Panels[f_SelectedIdx].Show();

            f_Caption.Text = f_Panels[f_SelectedIdx].p_Caption;
            f_Description.Text = f_Panels[f_SelectedIdx].p_Description;
        }

        private void NotifyChanges()
        {
            f_ApplyButton.Enabled = true;
        }

        private void DefaultsButton_Click(object sender, EventArgs e)
        {
            if (true == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.OptionsDlgQuestionReallyReset]))
            {
                f_Panels[f_SelectedIdx].Set(f_DefaultPrefs, true);
            }
        }

        public void DeliverKeysFromNewConnection(I2PKeyPair a_DatagramKeyPair,
                                                 I2PKeyPair a_StreamKeyPair)
        {
            // This routine gets called when a connection is established
            // while the options dialog is shown.

            AddressesPanel l_Panel = (AddressesPanel)f_Panels[(int)PageTypes.Addresses];
            l_Panel.SetKeyPairs(a_DatagramKeyPair, a_StreamKeyPair);
        }

        private void OKButton_Click(object sender, EventArgs e)
        {
            AcceptStates l_State = TryAcceptChanges();

            if (AcceptStates.Accepted != l_State)
            {
                return;
            }

            this.DialogResult = DialogResult.OK;
            this.Close();
        }

        private void ApplyButton_Click(object sender, EventArgs e)
        {
            AcceptStates l_State = TryAcceptChanges();
        }

        private void SetAll(Preferences a_Prefs)
        {
            foreach (OptionsPanel l_Cntrl in f_Panels)
            {
                l_Cntrl.Set(a_Prefs, false);
            }
        }

        public Preferences ParsePrefs()
        {
            // Return can be null.

            MainForm l_Form = f_Inst.f_MainForm;
            string l_ErrMsg;

            // Get application directories.

            string l_AppDataDir;
            string l_PreviewFolder;
            string l_TempDir;

            if (false == f_AppDirsPanel.GetDirectories(out l_AppDataDir,
                                                       out l_PreviewFolder,
                                                       out l_TempDir,
                                                       out l_ErrMsg))
            {
                SetSelectedPanel(PageTypes.AppDirs);
                l_Form.ErrMsgBox(l_ErrMsg);

                return null;
            }

            // Get bandwidth limits.

            Speed l_DlLimit;

            if (false == f_BandwidthPanel.TryGetDlLimit(out l_DlLimit, out l_ErrMsg))
            {
                SetSelectedPanel(PageTypes.Bandwidth);
                f_BandwidthPanel.FocusDlLimitControl();
                l_Form.ErrMsgBox(l_ErrMsg);

                return null;
            }

            Speed l_UlLimit;

            if (false == f_BandwidthPanel.TryGetUlLimit(out l_UlLimit, out l_ErrMsg))
            {
                SetSelectedPanel(PageTypes.Bandwidth);
                f_BandwidthPanel.FocusUlLimitControl();
                l_Form.ErrMsgBox(l_ErrMsg);

                return null;
            }

            // Get shared directories.

            SharedDirectories l_SharedFiles;

            if (false == f_SharedDirsPanel.TryCreateSharedDirs(out l_SharedFiles, out l_ErrMsg))
            {
                SetSelectedPanel(PageTypes.SharedDirs);
                l_Form.InfoMsgBox(l_ErrMsg);

                return null;
            }

            // Get SAM end point.

            System.Net.IPEndPoint l_SamEndPoint;

            if (false == f_AdvPanel.TryGetSamEndPoint(out l_SamEndPoint, out l_ErrMsg))
            {
                SetSelectedPanel(PageTypes.Advanced);
                f_AdvPanel.FocusSamEndPointControl();
                l_Form.ErrMsgBox(l_ErrMsg);

                return null;
            }

            // Check tunnel settings and warn if a short tunnel is
            // used.

            bool l_UseInboundSettingsForOutbound = f_TunnelsPanel.p_UseInboundSettingsForOutbound;
            TunnelSettings l_InboundTunnelSettings = f_TunnelsPanel.p_InboundTunnels;
            TunnelSettings l_OutboundTunnelSettings = f_TunnelsPanel.p_OutboundTunnels;

            // Short tunnel warning.

            if (true == l_InboundTunnelSettings.IsShortTunnel() ||
                (false == l_UseInboundSettingsForOutbound &&
                 true == l_OutboundTunnelSettings.IsShortTunnel()))
            {
                if (false == l_Form.YesNoMsgBox(f_Inst[TextKeys.OptionsDlgTunnelsMsgWarnShortTunnels]))
                {
                    SetSelectedPanel(PageTypes.Tunnels);

                    return null;
                }
            }

            // Save boot config data.

            f_LanguagePanel.Store();
            BootConfig l_BootConfig = Software.s_BootConfig;
            l_BootConfig.SetNewApplicationDataDirectory(l_AppDataDir);
            l_BootConfig.SetTempDirectory(l_TempDir);
            bool l_Saved = l_BootConfig.Save();

            // Create and return new preferences.
            GeneralPrefs l_GeneralPrefs = new GeneralPrefs(f_GeneralPanel.p_SaveSearchHistory, f_DisplPanel.p_ShowSpeedOnTitle, f_DisplPanel.p_ShowTrayIcon, f_DisplPanel.p_ShowBalloonNotifications, f_DisplPanel.p_MinimizeToTray, f_GeneralPanel.p_PromptOnExit, f_GeneralPanel.p_BeepWhenDlIsDone, f_DisplPanel.p_DisplaySISizes, f_ListViewsPanel.p_ProgressBarKind, f_AdvPanel.p_CompactOnExit, f_GeneralPanel.p_ConnStartKind, f_eD2kPanel.p_OwnPersona, l_UseInboundSettingsForOutbound, l_SamEndPoint, l_InboundTunnelSettings, l_OutboundTunnelSettings, f_AddrPanel.p_DatagramKeyPair, f_AddrPanel.p_StreamKeyPair, f_AddrPanel.p_AutoManageDatagramKeys, f_AddrPanel.p_AutoManageStreamKeys, l_PreviewFolder, f_ListViewsPanel.p_GridViewTextSizeRatio, f_ListViewsPanel.p_GridViewBorderStyle, f_AdvPanel.p_AllowComplFilesDeletion, f_DisplPanel.p_UseBackgroundImagesForComplBar, f_ListViewsPanel.p_AutoSortThreshold, f_AdvPanel.p_ShowPeersPanelAdvSubMenu, f_DisplPanel.p_IconBgColor, f_AdvPanel.p_RunFilesOnDblClick, f_GeneralPanel.p_UIModeAdvanced, f_DisplPanel.p_PreferredMainTabBox, f_AdvPanel.p_AcceptAdditionalKeyCombinationForExitingBox);
            AssocPrefs l_AssocPrefs = new AssocPrefs(f_OSIntegrationPanel.p_LaunchWithOS, f_OSIntegrationPanel.p_AssocWitheD2kURIs, f_OSIntegrationPanel.p_AssocWithMagnetURIs, f_OSIntegrationPanel.p_AssocWithNBCollections);
            NetworkPrefs l_NetworkPrefs = new NetworkPrefs(l_DlLimit, l_UlLimit, f_BandwidthPanel.p_MaxConnsGlobal, f_BandwidthPanel.p_MaxIngoingConnsPerAddr, f_BandwidthPanel.p_MaxOutgoingConnsPerAddr, f_BandwidthPanel.p_MaxDlFiles, f_BandwidthPanel.p_MaxDlSlotsPerFile, f_BandwidthPanel.p_MaxGlobalUlSlots, f_eD2kPanel.p_UserName, f_AdvPanel.p_AllowDataCompression, f_AdvPanel.p_PoisonPeers);
            KadPrefs l_KadPrefs = new KadPrefs(KadPrefs.c_PublMinAcceptReqFitness, KadPrefs.c_PublMinOwnFitness, f_AdvPanel.p_OnlyTrustedUsersMayPublish, f_KadPanel.p_MaxKadMapSize, f_KadPanel.p_CommentFilter, f_KadPanel.p_RandomizeClientID, f_KadPanel.p_ClientID, f_KadPanel.p_PublFlags, f_AdvPanel.p_PublUnicodeTextsInNBStyleOnly, f_KadFilterPanel.CreateFilter());
            IOPrefs l_IOPrefs = new IOPrefs(f_GeneralPanel.p_InsertIncomplFilesStarted, f_SharedFilesPanel.p_OpenFilesAsMoveable, f_SharedFilesPanel.p_WatchFileSystem, l_SharedFiles, false, f_SharedFilesPanel.GetExclFileAttributes(), f_SharedFilesPanel.p_HashingHDLoadThresholdPercentage);

            return new Preferences(ref l_GeneralPrefs, ref l_AssocPrefs, ref l_NetworkPrefs, ref l_KadPrefs, ref l_IOPrefs);
        }

        private AcceptStates TryAcceptChanges()
        {
            Preferences l_Prefs = ParsePrefs();

            if (null == l_Prefs)
            {
                return AcceptStates.InvalidInput;
            }

            MainForm l_Form = f_Inst.f_MainForm;
            l_Form.UpdatePreferences(this, l_Prefs, false);

            // We never expect the GUI to be closed while the options
            // dialog is shown, but in any case.

            return (false == l_Form.p_IsRunning ? AcceptStates.GUIClosed : AcceptStates.Accepted);
        }

        public enum PageTypes : byte
        {
            General,
            Language,
            OSIntegration,
            AppDirs,
            Display,
            ListView,
            Tunnels,
            Addresses,
            Bandwidth,
            Kademlia,
            KadFilter,
            eD2k,
            SharedDirs,
            SharedFiles,
            Advanced,
        }

        private enum AcceptStates : byte
        {
            Accepted = 0,
            InvalidInput = 1,
            GUIClosed = 2,
        }

        private abstract class OptionsPanel : GuiAutoPanel
        {
            protected readonly OptionsDialog f_OptionsDlg;

            private string f_Caption;
            private string f_Description;

            public string p_Caption
            {
                get { return f_Caption; }
            }

            public string p_Description
            {
                get { return f_Description; }
            }

            public OptionsPanel(OptionsDialog a_OptionsDlg,
                                ref Rectangle a_Rect) :
                           base(true)
            {
                f_OptionsDlg = a_OptionsDlg;

                f_OptionsDlg.Controls.Add(this);
                this.Location = a_Rect.Location;
                this.Size = a_Rect.Size;

                this.Hide();
            }

            public void SetTexts(TextKeys a_CaptionKey,
                                 TextKeys a_DescKey)
            {
                f_Caption = f_OptionsDlg.f_Inst[a_CaptionKey];
                f_Description = f_OptionsDlg.f_Inst[a_DescKey];
            }

            public void Set(Preferences a_Prefs,
                            bool a_IsResetToDefault)
            {
                SetCore(a_Prefs, a_IsResetToDefault);
            }

            protected abstract void SetCore(Preferences a_Prefs,
                                            bool a_Default);

            public virtual void BeforeDisposal()
            {
            }
        }

        private class GeneralPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 450;

            private readonly GuiGroupBox f_AppGroupBox;
            private readonly GuiCheckBox f_PromptOnExitBox;
            private readonly GuiCheckBox f_BeepWhenDoneBox;
            private readonly GuiCheckBox f_InsertIncomplFilesStartedBox;
            private readonly GuiCheckBox f_SaveSearchHistoryBox;

            private readonly RadioButtonBattery f_ConnStartKind;

            private readonly RadioButtonBattery f_UIMode;

            public bool p_PromptOnExit
            {
                get { return f_PromptOnExitBox.Checked; }
            }

            public bool p_BeepWhenDlIsDone
            {
                get { return f_BeepWhenDoneBox.Checked; }
            }

            public bool p_InsertIncomplFilesStarted
            {
                get { return f_InsertIncomplFilesStartedBox.Checked; }
            }

            public bool p_SaveSearchHistory
            {
                get { return f_SaveSearchHistoryBox.Checked; }
            }

            public GeneralPrefs.ConnStartKinds p_ConnStartKind
            {
                get { return (GeneralPrefs.ConnStartKinds)f_ConnStartKind.GetSelectedIndex(); }
            }

            public bool p_UIModeAdvanced
            {
                get { return (1 == f_UIMode.GetSelectedIndex()); }
            }

            public GeneralPanel(OptionsDialog a_OptionsDlg,
                                ref Rectangle a_Rect) :
                base(a_OptionsDlg, ref a_Rect)
            {
                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                f_AppGroupBox = new GuiGroupBox(this);
                f_PromptOnExitBox = new GuiCheckBox(f_AppGroupBox, l_Width);
                f_BeepWhenDoneBox = new GuiCheckBox(f_AppGroupBox, l_Width);
                f_InsertIncomplFilesStartedBox = new GuiCheckBox(f_AppGroupBox, l_Width);
                f_SaveSearchHistoryBox = new GuiCheckBox(f_AppGroupBox, l_Width);
                f_ConnStartKind = new RadioButtonBattery(this, l_Width, 3);
                f_UIMode = new RadioButtonBattery(this, l_Width, 2);

                int l_Left = f_AppGroupBox.Padding.Left;
                f_PromptOnExitBox.Location = new Point(l_Left, f_AppGroupBox.Padding.Top);
                f_BeepWhenDoneBox.Location = new Point(l_Left, f_PromptOnExitBox.GetBottom());
                f_InsertIncomplFilesStartedBox.Location = new Point(l_Left, f_BeepWhenDoneBox.GetBottom());
                f_SaveSearchHistoryBox.Location = new Point(l_Left, f_InsertIncomplFilesStartedBox.GetBottom());
                f_AppGroupBox.Size = new Size(l_Width, f_SaveSearchHistoryBox.Bottom + f_AppGroupBox.Padding.Bottom);
                f_ConnStartKind.Top = f_AppGroupBox.GetBottom();
                f_UIMode.Top = f_ConnStartKind.GetBottom();
                this.ChildBoundsUpdated();

                Instance l_Inst = a_OptionsDlg.f_Inst;
                f_AppGroupBox.Text = l_Inst[TextKeys.OptionsDlgGeneralAppTitle];
                f_PromptOnExitBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgGeneralAppPromptOnExit]);
                f_BeepWhenDoneBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgGeneralAppBeepWhenDlIsDone]);
                f_InsertIncomplFilesStartedBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgGeneralAppAutoStartDl]);
                f_SaveSearchHistoryBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgGeneralAppSaveSearchHist]);
                f_ConnStartKind.SetTexts(l_Inst.f_TextSet, TextKeys.OptionsDlgGeneralConnMode, TextKeys.OptionsDlgGeneralConnManual, TextKeys.OptionsDlgGeneralConnManualButRepair, TextKeys.OptionsDlgGeneralConnAuto);
                f_UIMode.SetTexts(l_Inst.f_TextSet, TextKeys.OptionsDlgGeneralUIMode, TextKeys.OptionsDlgGeneralUIModeSimple, TextKeys.OptionsDlgGeneralUIModeAdvanced);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_AppGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_PromptOnExitBox.Checked = a_Prefs.f_GeneralPrefs.f_PromptOnExit;
                f_BeepWhenDoneBox.Checked = a_Prefs.f_GeneralPrefs.f_BeepWhenDlIsDone;
                f_InsertIncomplFilesStartedBox.Checked = a_Prefs.f_IOPrefs.f_InsertIncomplFilesStarted;
                f_SaveSearchHistoryBox.Checked = a_Prefs.f_GeneralPrefs.f_SaveSearchHistory;
                f_ConnStartKind.SetSelectedIndex((int)a_Prefs.f_GeneralPrefs.f_ConnStartKind);
                f_UIMode.SetSelectedIndex(false == a_Prefs.f_GeneralPrefs.f_UIModeAdvanced ? 0 : 1);
            }
        }

        private class LanguagePanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 450;

            private readonly CulturePickerControl f_CulturePicker;

            public LanguagePanel(OptionsDialog a_OptionsDlg,
                                 ref Rectangle a_Rect) :
                            base(a_OptionsDlg, ref a_Rect)
            {
                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                f_CulturePicker = new CulturePickerControl(f_OptionsDlg.f_Inst, this, Lyt.CalcX(325), l_Width);
                this.ChildBoundsUpdated();
            }

            protected override void DisposeAutoPanelChildren()
            {
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == a_Default)
                {
                    f_CulturePicker.Load();
                }
                else
                {
                    f_CulturePicker.ResetToDefaults();
                }
            }

            public void Store()
            {
                f_CulturePicker.Store();
            }
        }

        private class OSIntegrationPanel : OptionsPanel
        {
            private const ushort c_GroupBoxWidth = 350;

            private readonly GuiGroupBox f_GroupBox;
            private readonly GuiCheckBox f_LaunchWithOSBox;
            private readonly GuiCheckBox f_AssocWitheD2kURIsBox;
            private readonly GuiCheckBox f_AssocWithMagnetURIsBox;
            private readonly GuiCheckBox f_AssocWithNBCollectionsBox;

            public bool p_LaunchWithOS
            {
                get { return f_LaunchWithOSBox.Checked; }
            }

            public bool p_AssocWitheD2kURIs
            {
                get { return f_AssocWitheD2kURIsBox.Checked; }
            }

            public bool p_AssocWithMagnetURIs
            {
                get { return f_AssocWithMagnetURIsBox.Checked; }
            }

            public bool p_AssocWithNBCollections
            {
                get { return f_AssocWithNBCollectionsBox.Checked; }
            }

            public OSIntegrationPanel(OptionsDialog a_OptionsDlg,
                                      ref Rectangle a_Rect) :
                                 base(a_OptionsDlg, ref a_Rect)
            {
                f_GroupBox = new GuiGroupBox(this);
                int l_GroupBoxWidth = Lyt.CalcX(c_GroupBoxWidth);
                int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
                f_LaunchWithOSBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);
                f_AssocWitheD2kURIsBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);
                f_AssocWithMagnetURIsBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);
                f_AssocWithNBCollectionsBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);

                f_GroupBox.Enabled = Software.sp_WindowsDetected;
                f_LaunchWithOSBox.Location = new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top);
                f_AssocWitheD2kURIsBox.Location = new Point(f_LaunchWithOSBox.Left, f_LaunchWithOSBox.GetBottom());
                f_AssocWithMagnetURIsBox.Location = new Point(f_LaunchWithOSBox.Left, f_AssocWitheD2kURIsBox.GetBottom());
                f_AssocWithNBCollectionsBox.Location = new Point(f_LaunchWithOSBox.Left, f_AssocWithMagnetURIsBox.GetBottom());
                f_LaunchWithOSBox.Width = l_InnerWidth;
                f_AssocWitheD2kURIsBox.Width = l_InnerWidth;
                f_AssocWithMagnetURIsBox.Width = l_InnerWidth;
                f_AssocWithNBCollectionsBox.Width = l_InnerWidth;
                f_GroupBox.Size = new Size(l_GroupBoxWidth, f_AssocWithNBCollectionsBox.Bottom + f_GroupBox.Padding.Bottom);
                this.ChildBoundsUpdated();

                Instance l_Inst = a_OptionsDlg.f_Inst;
                f_GroupBox.Text = l_Inst[TextKeys.OptionsDlgOSIntgrWinTitle];
                f_LaunchWithOSBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgOSIntgrStartOnLogon]);
                f_AssocWitheD2kURIsBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgOSIntgrAssoceD2kURIs]);
                f_AssocWithMagnetURIsBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgOSIntgrAssocMagnetURIs]);
                f_AssocWithNBCollectionsBox.SetTextAndUpdateSize(l_Inst[TextKeys.OptionsDlgOSIntgrAssocWithCollections]);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_GroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == Software.sp_WindowsDetected)
                {
                    return;
                }

                f_LaunchWithOSBox.Checked = a_Prefs.f_AssocPrefs.f_LaunchWithOS;
                f_AssocWitheD2kURIsBox.Checked = a_Prefs.f_AssocPrefs.f_AssocWitheD2kURIs;
                f_AssocWithMagnetURIsBox.Checked = a_Prefs.f_AssocPrefs.f_AssocWithMagnetURIs;
                f_AssocWithNBCollectionsBox.Checked = a_Prefs.f_AssocPrefs.f_AssocWithNBCollections;
            }
        }

        private class AppDirsPanel : OptionsPanel
        {
            private const ushort c_GroupBoxWidth = 400;

            private readonly BootConfig f_BootConfig;

            private readonly GuiGroupBox f_AppDataPathGroupBox;
            private readonly DirectoryControl f_AppDataDirCntrl;
            private readonly GuiLabel f_AppDataLabel;

            private readonly GuiGroupBox f_TempDataGroupBox;
            private readonly DirectoryControl f_PreviewDirCntrl;
            private readonly GuiButton f_DeletePreviewFilesButton;
            private readonly DirectoryControl f_TempPathCntrl;

            public AppDirsPanel(OptionsDialog a_OptionsDlg,
                                ref Rectangle a_Rect) :
                           base(a_OptionsDlg, ref a_Rect)
            {
                f_BootConfig = Software.s_BootConfig;

                Instance l_Inst = a_OptionsDlg.f_Inst;
                f_AppDataPathGroupBox = new GuiGroupBox(this);
                int l_Width = Lyt.CalcX(c_GroupBoxWidth);
                int l_InnerWidth = (l_Width - f_AppDataPathGroupBox.Padding.Horizontal);
                f_AppDataDirCntrl = new DirectoryControl(l_Inst, f_AppDataPathGroupBox, l_InnerWidth, TextKeys.OptionsDlgAppDataDirPickADir);
                f_AppDataLabel = new GuiLabel(f_AppDataPathGroupBox);
                f_TempDataGroupBox = new GuiGroupBox(this);
                f_PreviewDirCntrl = new DirectoryControl(l_Inst, f_TempDataGroupBox, l_InnerWidth, TextKeys.OptionsDlgTempDataPreviewFolderPickADir);
                f_DeletePreviewFilesButton = new GuiButton(f_TempDataGroupBox);
                f_TempPathCntrl = new DirectoryControl(l_Inst, f_TempDataGroupBox, l_InnerWidth, TextKeys.OptionsDlgTempDataTempFolderPickADir);

                // We won't initialize completely here, since
                // f_AppDataLabel has a dynamic height.
                int l_Left = f_AppDataPathGroupBox.Padding.Left;
                f_AppDataDirCntrl.Location = new Point(l_Left, f_AppDataPathGroupBox.Padding.Top);
                f_AppDataLabel.Location = new Point(l_Left, f_AppDataDirCntrl.GetBottom());
                f_AppDataLabel.Width = l_InnerWidth;
                f_PreviewDirCntrl.Location = new Point(l_Left, f_TempDataGroupBox.Padding.Top);
                f_DeletePreviewFilesButton.Location = new Point(l_Left, f_PreviewDirCntrl.GetBottom());
                f_DeletePreviewFilesButton.Width = Lyt.s_GenerousButtonWidth;
                f_TempPathCntrl.Location = new Point(l_Left, f_DeletePreviewFilesButton.GetBottom() + Lyt.s_SmallSpace);
                f_TempDataGroupBox.SetSizePlusBottomPadding(l_Width, f_TempPathCntrl.Bottom);
                this.ChildBoundsUpdated();

                f_AppDataDirCntrl.e_TextChanged += AppDataPathDirCntrl_TextChanged;
                f_DeletePreviewFilesButton.Click += DeletePreviewFilesButton_Click;

                TextSet l_TextSet = l_Inst.f_TextSet;
                f_AppDataPathGroupBox.Text = l_TextSet[TextKeys.OptionsDlgAppDataGroupBox];
                f_AppDataDirCntrl.SetCaption(TextKeys.OptionsDlgAppDataDir);
                f_TempDataGroupBox.Text = l_TextSet[TextKeys.OptionsDlgTempDataGroupBox];
                f_PreviewDirCntrl.SetCaption(TextKeys.OptionsDlgTempDataPreviewFolder);
                f_DeletePreviewFilesButton.Text = l_TextSet[TextKeys.OptionsDlgTempDataDelPreviewFiles];
                f_TempPathCntrl.SetCaption(TextKeys.OptionsDlgTempDataTempFolder);

                SetAppDataLabelText(true, string.Empty);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_AppDataPathGroupBox);
                Dialog.DisposeAllChildren(f_TempDataGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            public void SetAppDataLabelText(bool a_Force,
                                            string a_Str)
            {
                bool l_Visible = (0 < a_Str.Length);

                if (false == a_Force &&
                    f_AppDataLabel.Visible == l_Visible)
                {
                    return;
                }

                this.SuspendLayout();

                int l_Bottom;

                if (false == l_Visible)
                {
                    l_Bottom = f_AppDataDirCntrl.Bottom;
                }
                else
                {
                    f_AppDataLabel.Text = a_Str;
                    Lyt.SetFontAndHeight(f_AppDataLabel, f_AppDataLabel.Font);
                    l_Bottom = f_AppDataLabel.Bottom;
                }

                f_AppDataPathGroupBox.SetSizePlusBottomPadding(f_TempDataGroupBox.Width, l_Bottom);
                f_AppDataLabel.Visible = l_Visible;
                f_TempDataGroupBox.Top = f_AppDataPathGroupBox.GetBottom();
                this.ChildBoundsUpdated();

                this.ResumeLayout();
            }

            private void AppDataPathDirCntrl_TextChanged(object sender, EventArgs e)
            {
                if (f_AppDataDirCntrl.GetPath(true) == f_BootConfig.p_AppDataDir)
                {
                    SetAppDataLabelText(false, string.Empty);

                    return;
                }

                string l_Msg = f_OptionsDlg.f_Inst.f_TextSet.Format(TextKeys.OptionsDlgAppDataMsgPromptLater, f_BootConfig.p_AppDataDir);
                SetAppDataLabelText(false, l_Msg);
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_AppDataDirCntrl.SetPath(f_BootConfig.p_NewAppDataDir);
                f_PreviewDirCntrl.SetPath(a_Prefs.f_GeneralPrefs.f_PreviewFolder);
                f_TempPathCntrl.SetPath(f_BootConfig.p_TempDir);
            }

            public bool GetDirectories(out string a_AppDataDir,
                                       out string a_PreviewFolder,
                                       out string a_TempDir,
                                       out string a_ErrMsg)
            {
                if (false == f_AppDataDirCntrl.GetPathWithCheck(true, out a_AppDataDir, out a_ErrMsg))
                {
                    a_PreviewFolder = null;
                    a_TempDir = null;
                    f_AppDataDirCntrl.SetFocus();

                    return false;
                }

                if (false == f_PreviewDirCntrl.GetPathWithCheck(true, out a_PreviewFolder, out a_ErrMsg))
                {
                    a_TempDir = null;
                    f_PreviewDirCntrl.SetFocus();

                    return false;
                }

                if (false == f_TempPathCntrl.GetPathWithCheck(true, out a_TempDir, out a_ErrMsg))
                {
                    f_TempPathCntrl.SetFocus();

                    return false;
                }

                a_ErrMsg = null;

                return true;
            }

            private void DeletePreviewFilesButton_Click(object sender, EventArgs e)
            {
                MainForm l_Form = f_OptionsDlg.f_Inst.f_MainForm;
                TextSet l_TextSet = f_OptionsDlg.f_Inst.f_TextSet;
                string l_Folder = f_PreviewDirCntrl.GetPath(true);

                if (null == GetFiles(l_Folder))
                {
                    // No files.

                    string l_Msg = l_TextSet.Format(TextKeys.OptionsDlgMsgNoFilesInDir, l_Folder);
                    l_Form.InfoMsgBox(l_Msg);

                    return;
                }

                string l_Quest = l_TextSet.Format(TextKeys.OptionsDlgQuestReallyDelFilesInDir, l_Folder);
                if (false == l_Form.SeriousYesNoChoice(l_Quest))
                {
                    return;
                }

                // Refresh the available files (after the message box).
                string[] l_FileNames = GetFiles(l_Folder);

                if (null == l_FileNames)
                {
                    return;
                }

                // Since moving to recycle can block, we use an extra
                // thread.

                DeletionWorkerContext l_Cntxt = new DeletionWorkerContext(l_FileNames);
                l_Cntxt.Start();

                if (false == WaitMessageDialog.ShowDialog(f_OptionsDlg.f_Inst.f_TextSet, f_OptionsDlg, l_Cntxt, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.RemovingFilesDlgTitle, TextKeys.RemovingFilesDlgMsgMovingToTrash))
                {
                    return;
                }

                l_Form.ShowDeletedFilesMsgBoxIfNecessary((int)l_Cntxt.p_SuccCount, l_FileNames.Length, l_Cntxt.p_Aborted);
            }

            private static string[] GetFiles(string a_ParentDirectory)
            {
                // Return can is null if there are no files or an error
                // occurred.

                string[] l_FileNames;

                try
                {
                    l_FileNames = System.IO.Directory.GetFiles(a_ParentDirectory);
                }
                catch (ArgumentNullException l_Exc)
                {
                    throw (l_Exc);
                }
                catch
                {
                    return null;
                }

                return (0 == l_FileNames.Length ? null : l_FileNames);
            }

            private class DeletionWorkerContext : AbortableControllerContext
            {
                private const string c_DeletionWorkerName = "PreviewFilesDeletionWorker";

                public readonly string[] f_FileNames;

                private uint f_SuccCount;

                public uint p_SuccCount
                {
                    get { return f_SuccCount; }
                }

                public DeletionWorkerContext(string[] a_FileNames)
                {
                    f_FileNames = a_FileNames;
                }

                public void Start()
                {
                    Util.StartThread(c_DeletionWorkerName, DeletionWorker);
                }

                private void DeletionWorker()
                {
                    f_SuccCount = 0;

                    uint l_Pos = 0;
                    uint l_Total = (uint)f_FileNames.Length;
                    foreach (string l_FileName in f_FileNames)
                    {
                        if (true == p_Aborted)
                        {
                            break;
                        }

                        if (true == FHandle.Delete(l_FileName, true))
                        {
                            f_SuccCount++;
                        }

                        base.SetPercentage(++l_Pos, l_Total);
                    }

                    base.SetFinishedBase();
                }
            }
        }

        private class DisplayPanel : OptionsPanel
        {
            private const ushort c_GroupBoxWidth = 350;

            private readonly GuiGroupBox f_TopGroupBox;
            private readonly GuiCheckBox f_TitleSpeedBox;
            private readonly GuiCheckBox f_ShowTrayIconBox;
            private readonly GuiCheckBox f_ShowBalloonNotificationsBox;
            private readonly GuiCheckBox f_MinimizeToTrayBox;
            private readonly ColorPicker f_IconBgColorPicker;
            private readonly LabeledComboBox f_PreferredMainTabBox;

            private readonly RadioButtonBattery f_SizeDisplKind;
            private readonly RadioButtonBattery f_CompletionBarKind;

            public bool p_ShowSpeedOnTitle
            {
                get { return f_TitleSpeedBox.Checked; }
            }

            public bool p_ShowTrayIcon
            {
                get { return f_ShowTrayIconBox.Checked; }
            }

            public bool p_ShowBalloonNotifications
            {
                get { return f_ShowBalloonNotificationsBox.Checked; }
            }

            public bool p_MinimizeToTray
            {
                get { return f_MinimizeToTrayBox.Checked; }
            }

            public Color p_IconBgColor
            {
                get { return f_IconBgColorPicker.GetColorEasy(); }
            }

            public MainForm.MainTabs p_PreferredMainTabBox
            {
                get
                {
                    int l_Idx = f_PreferredMainTabBox.f_Box.SelectedIndex;

                    return (0 >= l_Idx ? GeneralPrefs.c_InvalidPreferredMainTab : (MainForm.MainTabs)(l_Idx - 1));
                }
            }

            public bool p_DisplaySISizes
            {
                get { return (0 != f_SizeDisplKind.GetSelectedIndex()); }
            }

            public bool p_UseBackgroundImagesForComplBar
            {
                get { return (1 == f_CompletionBarKind.GetSelectedIndex()); }
            }

            public DisplayPanel(OptionsDialog a_OptionsDlg,
                                ref Rectangle a_Rect) :
                           base(a_OptionsDlg, ref a_Rect)
            {
                f_TopGroupBox = new GuiGroupBox(this);
                int l_Width = Lyt.CalcX(c_GroupBoxWidth);
                int l_InnerWidth = (l_Width - f_TopGroupBox.Padding.Horizontal);
                f_TitleSpeedBox = new GuiCheckBox(f_TopGroupBox, l_InnerWidth);
                f_ShowTrayIconBox = new GuiCheckBox(f_TopGroupBox, l_InnerWidth);
                f_ShowBalloonNotificationsBox = new GuiCheckBox(f_TopGroupBox, l_InnerWidth);
                f_MinimizeToTrayBox = new GuiCheckBox(f_TopGroupBox, l_InnerWidth);
                f_IconBgColorPicker = new ColorPicker(a_OptionsDlg.f_Inst, f_TopGroupBox, l_InnerWidth, true);
                f_PreferredMainTabBox = new LabeledComboBox(f_TopGroupBox, l_InnerWidth, Math.Min(l_InnerWidth, Lyt.CalcX(200)), false);
                f_SizeDisplKind = new RadioButtonBattery(this, l_Width, 2);
                f_CompletionBarKind = new RadioButtonBattery(this, l_Width, 2);

                int l_Left = f_TopGroupBox.Padding.Left;
                f_TitleSpeedBox.Location = new Point(l_Left, f_TopGroupBox.Padding.Top);
                f_ShowTrayIconBox.Location = new Point(l_Left, f_TitleSpeedBox.GetBottom());
                f_ShowBalloonNotificationsBox.Location = new Point(l_Left + Lyt.s_LargeSpace, f_ShowTrayIconBox.GetBottom());
                f_MinimizeToTrayBox.Location = new Point(l_Left + Lyt.s_LargeSpace, f_ShowBalloonNotificationsBox.GetBottom());
                f_IconBgColorPicker.Location = new Point(l_Left, f_MinimizeToTrayBox.GetBottom());
                f_PreferredMainTabBox.SetLocation(l_Left, f_IconBgColorPicker.GetBottom(true));
                f_TopGroupBox.SetSizePlusBottomPadding(l_Width, f_PreferredMainTabBox.f_Box.Bottom);
                f_SizeDisplKind.Top = f_TopGroupBox.GetBottom();
                f_CompletionBarKind.Top = f_SizeDisplKind.GetBottom();
                this.ChildBoundsUpdated();

                f_ShowTrayIconBox.Click += ShowTrayIconBox_Click;

                TextSet l_TextSet = a_OptionsDlg.f_Inst.f_TextSet;
                f_TopGroupBox.Text = l_TextSet[TextKeys.OptionsDlgDisplGeneralTitle];
                f_TitleSpeedBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgDisplGeneralShowSpeedOnTitle]);
                f_ShowTrayIconBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgDisplGeneralShowTrayIcon]);
                f_ShowBalloonNotificationsBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgDisplGeneralShowBalloons]);
                f_MinimizeToTrayBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgDisplGeneralMinimizeToTray]);
                // FIXME: Make translatable.
                f_PreferredMainTabBox.SetCaption("Preferred tab (gets pre-selected on startup):");
                f_PreferredMainTabBox.SetTexts(new string[] { "<tab selected on last shutdown>", l_TextSet[TextKeys.KademliaTabTitle], l_TextSet[TextKeys.FileSearchTabTitle], l_TextSet[TextKeys.FilesTabTitle], l_TextSet[TextKeys.PeersTabTitle], l_TextSet[TextKeys.StatsTabTitle], l_TextSet[TextKeys.LogTabTitle]});
                f_SizeDisplKind.SetTexts(l_TextSet, TextKeys.OptionsDlgDisplDataFormatKindTitle, TextKeys.OptionsDlgDisplDataFormatIEC, TextKeys.OptionsDlgDisplDataFormatSI);
                f_CompletionBarKind.SetTexts(l_TextSet, TextKeys.OptionsDlgDisplComplBarBgImgsTitle, TextKeys.OptionsDlgDisplComplBarBgImgsNo, TextKeys.OptionsDlgDisplComplBarBgImgsYes);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_TopGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_TitleSpeedBox.Checked = a_Prefs.f_GeneralPrefs.f_ShowSpeedOnTitle;
                f_ShowTrayIconBox.Checked = a_Prefs.f_GeneralPrefs.f_ShowTrayIcon;
                f_ShowBalloonNotificationsBox.Checked = a_Prefs.f_GeneralPrefs.f_ShowBalloonNotifications;
                f_MinimizeToTrayBox.Checked = a_Prefs.f_GeneralPrefs.f_MinimizeToTray;

                // Run the handler of f_ShowTrayIconBox.
                ShowTrayIconBox_Click(null, null);

                f_IconBgColorPicker.SetColorEasy(a_Prefs.f_GeneralPrefs.f_IconBgColor, true);
                MainForm.MainTabs l_PreferredTab = a_Prefs.f_GeneralPrefs.f_PreferredMainTab;
                f_PreferredMainTabBox.f_Box.SelectedIndex = (MainForm.c_MainTabsCount <= (int)l_PreferredTab ? 0 : (int)(l_PreferredTab + 1));
                f_SizeDisplKind.SetSelectedIndexFromFlag(a_Prefs.f_GeneralPrefs.f_DisplaySISizes);
                f_CompletionBarKind.SetSelectedIndexFromFlag(a_Prefs.f_GeneralPrefs.f_UseBackgroundImagesForComplBar);
            }

            private void ShowTrayIconBox_Click(object sender, EventArgs e)
            {
                bool l_Enabled = f_ShowTrayIconBox.Checked;
                f_ShowBalloonNotificationsBox.Enabled = l_Enabled;
                f_MinimizeToTrayBox.Enabled = l_Enabled;

                if (false == l_Enabled)
                {
                    f_ShowBalloonNotificationsBox.Checked = false;
                    f_MinimizeToTrayBox.Checked = false;
                }
            }
        }

        private class ListViewsPanel : OptionsPanel
        {
            private const ushort c_GroupBoxWidth = 300;

            private readonly ListAttributesControl f_ListAttributes;

            private readonly GuiGroupBox f_PerfGroupBox;
            private readonly LabeledNumberBoxWithUnit f_AutoSortThresholdBox;

            private readonly RadioButtonBattery f_ProgressBarKind;

            public byte p_GridViewTextSizeRatio
            {
                get { return f_ListAttributes.GetTextSizeRatio(); }
            }

            public GridViewBorderStyles p_GridViewBorderStyle
            {
                get { return f_ListAttributes.GetBorderStyle(); }
            }

            public uint p_AutoSortThreshold
            {
                get { return f_AutoSortThresholdBox.p_ValueAsUInt; }
            }

            public GeneralPrefs.ProgessBarKinds p_ProgressBarKind
            {
                get { return (GeneralPrefs.ProgessBarKinds)f_ProgressBarKind.GetSelectedIndex(); }
            }

            public ListViewsPanel(OptionsDialog a_OptionsDlg,
                                  ref Rectangle a_Rect) :
                             base(a_OptionsDlg, ref a_Rect)
            {
                int l_Width = Lyt.CalcX(c_GroupBoxWidth);
                f_ListAttributes = new ListAttributesControl(this, l_Width);
                f_PerfGroupBox = new GuiGroupBox(this);
                f_AutoSortThresholdBox = new LabeledNumberBoxWithUnit(f_PerfGroupBox, l_Width - f_PerfGroupBox.Padding.Horizontal);
                f_ProgressBarKind = new RadioButtonBattery(this, l_Width, 2);

                f_PerfGroupBox.Top = f_ListAttributes.GetBottom();
                f_AutoSortThresholdBox.SetLocation(f_PerfGroupBox.Padding.Left, f_PerfGroupBox.Padding.Top);
                f_AutoSortThresholdBox.SetLimits(0, (int)GeneralPrefs.c_MaxGridViewAutoSortItemThreshold);
                f_PerfGroupBox.SetSizePlusBottomPadding(l_Width, f_AutoSortThresholdBox.f_Box.Bottom);
                f_ProgressBarKind.Top = f_PerfGroupBox.GetBottom();
                this.ChildBoundsUpdated();

                TextSet l_TextSet = a_OptionsDlg.f_Inst.f_TextSet;
                f_ListAttributes.SetTexts(l_TextSet);
                f_PerfGroupBox.Text = l_TextSet[TextKeys.OptionsDlgListViewsPerfGroupBox];
                f_AutoSortThresholdBox.SetCaption(l_TextSet, TextKeys.OptionsDlgListViewsAutoSortThreshold);
                f_AutoSortThresholdBox.SetUnitText(l_TextSet, TextKeys.OptionsDlgListViewsItems);
                f_ProgressBarKind.SetTexts(l_TextSet, TextKeys.OptionsDlgDisplProgressBarTitle, TextKeys.OptionsDlgDisplProgressBarVerifiedOnly, TextKeys.OptionsDlgDisplProgressBarInclUnverified);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_PerfGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_ProgressBarKind.SetSelectedIndex((int)a_Prefs.f_GeneralPrefs.f_ProgressBarKind);
                f_AutoSortThresholdBox.f_Box.Value = a_Prefs.f_GeneralPrefs.f_GridViewAutoSortItemThreshold;
                f_ListAttributes.SetValues(a_Prefs.f_GeneralPrefs.f_GridViewTextSizeRatio, a_Prefs.f_GeneralPrefs.f_GridViewBorderStyle);
            }
        }

        private class TunnelsPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 350;

            private readonly TunnelControl f_InboundTunnels;
            private readonly TunnelControl f_OutboundTunnels;

            public TunnelSettings p_InboundTunnels
            {
                get { return f_InboundTunnels.Get(); }
            }

            public bool p_UseInboundSettingsForOutbound
            {
                get { return f_OutboundTunnels.p_UseInboundSettings; }
            }

            public TunnelSettings p_OutboundTunnels
            {
                get { return f_OutboundTunnels.Get(); }
            }

            public TunnelsPanel(OptionsDialog a_OptionsDlg,
                                ref Rectangle a_Rect) :
                           base(a_OptionsDlg, ref a_Rect)
            {
                TextSet l_TextSet = a_OptionsDlg.f_Inst.f_TextSet;
                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                f_InboundTunnels = new TunnelControl(l_TextSet, this, l_Width, false);
                f_OutboundTunnels = new TunnelControl(l_TextSet, this, l_Width, true);

                f_InboundTunnels.Location = new Point(0, 0);
                f_OutboundTunnels.Location = new Point(0, f_InboundTunnels.GetBottom());
                this.ChildBoundsUpdated();

                f_InboundTunnels.Text = l_TextSet[TextKeys.OptionsDlgTunnelsInboundSettings];
                f_OutboundTunnels.Text = l_TextSet[TextKeys.OptionsDlgTunnelsOutboundSettings];
            }

            protected override void DisposeAutoPanelChildren()
            {
                f_InboundTunnels.Dispose();
                f_OutboundTunnels.Dispose();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_InboundTunnels.Set(false, a_Prefs.f_GeneralPrefs.f_InboundTunnelSettings);
                f_OutboundTunnels.Set(a_Prefs.f_GeneralPrefs.f_UseInboundTunnelSettingsForOutboundTunnels, a_Prefs.f_GeneralPrefs.f_OutboundTunnelSettings);
            }
        }

        private class AddressesPanel : OptionsPanel
        {
            private readonly I2PAddressControl f_DatagramAddr;
            private readonly I2PAddressControl f_StreamAddr;

            public I2PKeyPair p_DatagramKeyPair
            {
                get { return f_DatagramAddr.p_KeyPair; }
            }

            public I2PKeyPair p_StreamKeyPair
            {
                get { return f_StreamAddr.p_KeyPair; }
            }

            public bool p_AutoManageDatagramKeys
            {
                get { return f_DatagramAddr.p_AutoManage; }
            }

            public bool p_AutoManageStreamKeys
            {
                get { return f_StreamAddr.p_AutoManage; }
            }

            public AddressesPanel(OptionsDialog a_OptionsDlg,
                                  ref Rectangle a_Rect) :
                             base(a_OptionsDlg, ref a_Rect)
            {
                Instance l_Inst = a_OptionsDlg.f_Inst;
                int l_Width = (a_Rect.Width - base.p_ScrollBarWidthInclSpace);
                f_DatagramAddr = new I2PAddressControl(l_Inst, this, l_Width, GeneralPrefs.c_AutoDropDatagramKeyPairDays);
                f_StreamAddr = new I2PAddressControl(l_Inst, this, l_Width, GeneralPrefs.c_AutoDropStreamKeyPairDays);

                f_StreamAddr.Location = new Point(0, f_StreamAddr.GetBottom());
                this.ChildBoundsUpdated();

                f_DatagramAddr.e_ClearClick += Addr_ClearClick;
                f_StreamAddr.e_ClearClick += Addr_ClearClick;

                f_DatagramAddr.Text = l_Inst[TextKeys.OptionsDlgAddressYourDatagramAddr];
                f_StreamAddr.Text = l_Inst[TextKeys.OptionsDlgAddressYourStreamAddr];
            }

            protected override void DisposeAutoPanelChildren()
            {
                f_DatagramAddr.Dispose();
                f_StreamAddr.Dispose();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == a_Default)
                {
                    SetKeyPairs(a_Prefs.f_GeneralPrefs.p_DatagramKeyPair, a_Prefs.f_GeneralPrefs.p_StreamKeyPair);
                }

                f_DatagramAddr.AdjustAutoManagement(a_Prefs.f_GeneralPrefs.f_AutoManageDatagramKeys);
                f_StreamAddr.AdjustAutoManagement(a_Prefs.f_GeneralPrefs.f_AutoManageStreamKeys);
            }

            public void ClearKeyPairs()
            {
                f_DatagramAddr.ClearKeyPair();
                f_StreamAddr.ClearKeyPair();
            }

            public void SetKeyPairs(I2PKeyPair a_DatagramKeyPair,
                                    I2PKeyPair a_StreamKeyPair)
            {
                f_DatagramAddr.SetKeyPair(a_DatagramKeyPair);
                f_StreamAddr.SetKeyPair(a_StreamKeyPair);
            }

            private void Addr_ClearClick(I2PAddressControl a_Cntrl)
            {
                if (true == f_OptionsDlg.f_Inst.f_MainForm.YesNoMsgBox(f_OptionsDlg.f_Inst[TextKeys.OptionsDlgAddressReallyDeleteAddr]))
                {
                    a_Cntrl.ClearKeyPair();
                }
            }
        }

        private class BandwidthPanel : OptionsPanel
        {
            private readonly GuiGroupBox f_DlGroupBox;
            private readonly BandwidthLimitControls f_DlLimitCntrl;
            private readonly LabeledNumberBox f_MaxDlFilesBox;
            private readonly LabeledNumberBox f_MaxDlSlotsPerFileBox;

            private readonly GuiGroupBox f_UlGroupBox;
            private readonly BandwidthLimitControls f_UlLimitCntrl;
            private readonly LabeledNumberBox f_MaxGlobalUlSlotsBox;

            private readonly GuiGroupBox f_NumberOfConnGroupBox;
            private readonly LabeledNumberBox f_MaxConnsGlobalBox;
            private readonly LabeledNumberBox f_MaxIngoingConnsPerAddrBox;
            private readonly LabeledNumberBox f_MaxOutgoingConnsPerAddrBox;

            public uint p_MaxDlFiles
            {
                get { return f_MaxDlFilesBox.p_ValueAsUInt; }
            }

            public uint p_MaxDlSlotsPerFile
            {
                get { return f_MaxDlSlotsPerFileBox.p_ValueAsUInt; }
            }

            public uint p_MaxGlobalUlSlots
            {
                get { return f_MaxGlobalUlSlotsBox.p_ValueAsUInt; }
            }

            public uint p_MaxConnsGlobal
            {
                get { return f_MaxConnsGlobalBox.p_ValueAsUInt; }
            }

            public uint p_MaxIngoingConnsPerAddr
            {
                get { return f_MaxIngoingConnsPerAddrBox.p_ValueAsUInt; }
            }

            public uint p_MaxOutgoingConnsPerAddr
            {
                get { return f_MaxOutgoingConnsPerAddrBox.p_ValueAsUInt; }
            }

            public BandwidthPanel(OptionsDialog a_OptionsDlg,
                                  ref Rectangle a_Rect) :
                             base(a_OptionsDlg, ref a_Rect)
            {
                TextSet l_TextSet = a_OptionsDlg.f_Inst.f_TextSet;
                bool l_DisplaySISizes = a_OptionsDlg.f_Inst.f_MainForm.p_DisplaySISizes;
                int l_GroupBoxWidth = ((a_Rect.Width - base.p_ScrollBarWidthInclSpace - Lyt.s_Space) / 2);
                f_DlGroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_GroupBoxWidth - f_DlGroupBox.Padding.Horizontal);
                f_DlLimitCntrl = new BandwidthLimitControls(l_TextSet, f_DlGroupBox, l_InnerWidth, l_DisplaySISizes);
                f_MaxDlFilesBox = new LabeledNumberBox(f_DlGroupBox, l_InnerWidth);
                f_MaxDlSlotsPerFileBox = new LabeledNumberBox(f_DlGroupBox, l_InnerWidth);
                f_UlGroupBox = new GuiGroupBox(this);
                f_UlLimitCntrl = new BandwidthLimitControls(l_TextSet, f_UlGroupBox, l_InnerWidth, l_DisplaySISizes);
                f_MaxGlobalUlSlotsBox = new LabeledNumberBox(f_UlGroupBox, l_InnerWidth);
                f_NumberOfConnGroupBox = new GuiGroupBox(this);
                f_MaxConnsGlobalBox = new LabeledNumberBox(f_NumberOfConnGroupBox, l_InnerWidth);
                f_MaxIngoingConnsPerAddrBox = new LabeledNumberBox(f_NumberOfConnGroupBox, l_InnerWidth);
                f_MaxOutgoingConnsPerAddrBox = new LabeledNumberBox(f_NumberOfConnGroupBox, l_InnerWidth);

                int l_Left = f_DlGroupBox.Padding.Left;
                f_DlLimitCntrl.SetLocation(l_Left, f_DlGroupBox.Padding.Top);
                f_MaxDlFilesBox.SetLocation(l_Left, f_DlLimitCntrl.GetBottom());
                f_MaxDlSlotsPerFileBox.SetLocation(l_Left, f_MaxDlFilesBox.GetBottom());
                f_DlGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_MaxDlSlotsPerFileBox.f_Box.Bottom);
                f_UlGroupBox.Left = (l_GroupBoxWidth + Lyt.s_Space);
                f_UlLimitCntrl.SetLocation(l_Left, f_UlGroupBox.Padding.Top);
                f_MaxGlobalUlSlotsBox.SetLocation(l_Left, f_UlLimitCntrl.GetBottom());
                f_UlGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_MaxGlobalUlSlotsBox.f_Box.Bottom);
                f_NumberOfConnGroupBox.Location = new Point(f_UlGroupBox.Left, f_UlGroupBox.GetBottom());
                f_MaxConnsGlobalBox.SetLocation(l_Left, f_NumberOfConnGroupBox.Padding.Top);
                f_MaxIngoingConnsPerAddrBox.SetLocation(l_Left, f_MaxConnsGlobalBox.GetBottom());
                f_MaxOutgoingConnsPerAddrBox.SetLocation(l_Left, f_MaxIngoingConnsPerAddrBox.GetBottom());
                f_NumberOfConnGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_MaxOutgoingConnsPerAddrBox.f_Box.Bottom);
                this.ChildBoundsUpdated();

                f_MaxDlFilesBox.SetLimits(0, NetworkPrefs.c_MaxDlFiles);
                f_MaxDlSlotsPerFileBox.SetLimits(0, NetworkPrefs.c_MaxDlOrUlConnsPerFile);
                f_MaxGlobalUlSlotsBox.SetLimits(0, NetworkPrefs.c_MaxGlobalUlSlots);
                f_MaxConnsGlobalBox.SetLimits(0, NetworkPrefs.c_MaxConns);
                f_MaxIngoingConnsPerAddrBox.SetLimits(0, NetworkPrefs.c_MaxConns);
                f_MaxOutgoingConnsPerAddrBox.SetLimits(0, NetworkPrefs.c_MaxConns);

                f_DlGroupBox.Text = l_TextSet[TextKeys.OptionsDlgBandwidthDlLimits];
                f_DlLimitCntrl.SetText(l_TextSet[TextKeys.OptionsDlgBandwidthDlGlobalSpeedLimit]);
                f_MaxDlFilesBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthDlNumOfFiles);
                f_MaxDlSlotsPerFileBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthDlNumOfSlotsPerFile);
                f_UlGroupBox.Text = l_TextSet[TextKeys.OptionsDlgBandwidthUlLimits];
                f_UlLimitCntrl.SetText(l_TextSet[TextKeys.OptionsDlgBandwidthUlGlobalSpeedLimit]);
                f_MaxGlobalUlSlotsBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthUlNumOfGlobalSlots);
                f_NumberOfConnGroupBox.Text = l_TextSet[TextKeys.OptionsDlgBandwidthConnsLimits];
                f_MaxConnsGlobalBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthConnsTotal);
                f_MaxIngoingConnsPerAddrBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthConnsInPerAddr);
                f_MaxOutgoingConnsPerAddrBox.SetCaption(l_TextSet, TextKeys.OptionsDlgBandwidthConnsOutPerAddr);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_DlGroupBox);
                Dialog.DisposeAllChildren(f_UlGroupBox);
                Dialog.DisposeAllChildren(f_NumberOfConnGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_DlLimitCntrl.SetLimit(a_Prefs.f_NetworkPrefs.f_DlLimit);
                f_MaxDlFilesBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxDlFiles);
                f_MaxDlSlotsPerFileBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxDownloadConnsPerFile);

                f_UlLimitCntrl.SetLimit(a_Prefs.f_NetworkPrefs.f_UlLimit);
                f_MaxGlobalUlSlotsBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxGlobalUlSlots);

                f_MaxConnsGlobalBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxGlobalConns);
                f_MaxIngoingConnsPerAddrBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxIngoingConnsPerAddr);
                f_MaxOutgoingConnsPerAddrBox.SetValueClipped(a_Prefs.f_NetworkPrefs.f_MaxOutgoingConnsPerAddr);
            }

            public bool TryGetDlLimit(out Speed a_DlLimit,
                                      out string a_ErrMsg)
            {
                return f_DlLimitCntrl.LoadSpeed(out a_DlLimit, out a_ErrMsg);
            }

            public bool TryGetUlLimit(out Speed a_UlLimit,
                                      out string a_ErrMsg)
            {
                return f_UlLimitCntrl.LoadSpeed(out a_UlLimit, out a_ErrMsg);
            }

            public void FocusDlLimitControl()
            {
                f_DlLimitCntrl.SetFocus();
            }

            public void FocusUlLimitControl()
            {
                f_UlLimitCntrl.SetFocus();
            }
        }

        private class KademliaPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 400;

            private readonly KadPublishControl f_PublishCntrl;

            private readonly GuiGroupBox f_CommentsGroupBox;
            public readonly LabeledTextBox f_CommentFilterBox;

            private readonly GuiGroupBox f_MiscGroupBox;
            private readonly DataSizePicker f_KadMapSizePicker;
            private readonly GuiCheckBox f_RandClientIDBox;

            private ClientID f_ClientID;

            public PublishFlags p_PublFlags
            {
                get { return f_PublishCntrl.GetFlags(); }
            }

            public bool p_RandomizeClientID
            {
                get { return f_RandClientIDBox.Checked; }
            }

            public ClientID p_ClientID
            {
                get { return f_ClientID; }
            }

            public string p_CommentFilter
            {
                get { return f_CommentFilterBox.f_Box.Text; }
            }

            public uint p_MaxKadMapSize
            {
                get { return f_KadMapSizePicker.GetDataSize(); }
            }

            public KademliaPanel(OptionsDialog a_OptionsDlg,
                                 ref Rectangle a_Rect) :
                            base(a_OptionsDlg, ref a_Rect)
            {
                bool l_DisplaySISizes = a_OptionsDlg.f_Inst.f_MainForm.p_DisplaySISizes;
                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                f_PublishCntrl = new KadPublishControl(this, l_Width);
                f_CommentsGroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_Width - f_CommentsGroupBox.Padding.Horizontal);
                f_CommentFilterBox = new LabeledTextBox(f_CommentsGroupBox, l_InnerWidth);
                f_MiscGroupBox = new GuiGroupBox(this);
                f_RandClientIDBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_KadMapSizePicker = new DataSizePicker(f_MiscGroupBox, l_InnerWidth, l_DisplaySISizes);

                int l_Left = f_CommentsGroupBox.Padding.Left;
                f_CommentsGroupBox.Top = f_PublishCntrl.GetBottom();
                f_CommentFilterBox.SetLocation(l_Left, f_CommentsGroupBox.Padding.Top);
                f_CommentsGroupBox.SetSizePlusBottomPadding(l_Width, f_CommentFilterBox.f_Box.Bottom);
                f_MiscGroupBox.Top = f_CommentsGroupBox.GetBottom();
                f_RandClientIDBox.Location = new Point(l_Left, f_MiscGroupBox.Padding.Top);
                f_KadMapSizePicker.SetLocation(l_Left, f_RandClientIDBox.GetBottom());
                f_MiscGroupBox.SetSizePlusBottomPadding(l_Width, f_KadMapSizePicker.GetBottom() - Lyt.s_Space);
                this.ChildBoundsUpdated();

                f_CommentFilterBox.f_Box.MaxLength = KadPrefs.c_MaxCommentsFilterStrChars;
                f_KadMapSizePicker.SetDataSizeLimits(KadPrefs.c_MinKadMapSize, KadPrefs.c_MaxKadMapSize);

                TextSet l_TextSet = a_OptionsDlg.f_Inst.f_TextSet;
                f_PublishCntrl.SetTexts(l_TextSet);
                f_CommentsGroupBox.Text = l_TextSet[TextKeys.OptionsDlgKademliaCommentsCaption];
                f_CommentFilterBox.SetCaption(l_TextSet, TextKeys.OptionsDlgKademliaFilter);
                f_MiscGroupBox.Text = l_TextSet[TextKeys.OptionsDlgKademliaMiscCaption];
                f_RandClientIDBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgKademliaMiscChangeClientID]);
                f_KadMapSizePicker.SetCaption(l_TextSet, TextKeys.OptionsDlgKademliaMiscMaxNetSizeOfKadIndex);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_CommentsGroupBox);
                Dialog.DisposeAllChildren(f_MiscGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_PublishCntrl.SetFlags(a_Prefs.f_KadPrefs.f_PublFlags);

                if (false == a_Default)
                {
                    f_ClientID = a_Prefs.f_KadPrefs.f_ClientID;
                }

                f_RandClientIDBox.Checked = a_Prefs.f_KadPrefs.f_RandomizeClientID;
                f_CommentFilterBox.f_Box.Text = a_Prefs.f_KadPrefs.f_CommentsFilter;
                f_KadMapSizePicker.SetDataSize(a_Prefs.f_KadPrefs.f_MaxKadMapSize);
            }
        }

        private class KadFilterPanel : OptionsPanel
        {
            private const string c_FileDlgFilter = "All Files|*.*|Nachtblitz Filter Rules|*.nfr";

            private readonly MainForm f_MainForm;

            private readonly GuiGroupBox f_GeneralGroupBox;
            private readonly GuiCheckBox f_GeneralEnableFilteringCheckBox;
            private readonly GuiCheckBox f_GeneralEnforceDefaultScopesCheckBox;
            private readonly GuiButton f_TestRulesButton;

            private readonly GuiGroupBox f_ScopeGroupBox;
            private readonly ScopeCheckBoxes f_ScopeCheckBoxes;

            private readonly GuiGroupBox f_RulesGroupBox;
            private readonly GridView<uint, DummyStruct> f_GridView;
            private readonly GuiButton f_AddButton;
            private readonly GuiButton f_ModifySelButton;
            private readonly GuiButton f_RemoveSelButton;
            private readonly GuiButton f_ImportButton;
            private readonly GuiButton f_ExportSelButton;

            private KademliaFilter f_Filter;

            public KadFilterPanel(OptionsDialog a_OptionsDlg,
                                     ref Rectangle a_Rect) :
                                base(a_OptionsDlg, ref a_Rect)
            {
                Instance l_Inst = a_OptionsDlg.f_Inst;
                f_MainForm = l_Inst.f_MainForm;
                this.AdjustVertAutoScroll(false);
                int l_RulesGroupBoxWidth = (a_Rect.Width - base.f_ScrollBarWidth - Lyt.s_SmallSpace);
                int l_GroupBoxWidthWithoutSpacer = (l_RulesGroupBoxWidth - Lyt.s_Space);
                int l_GeneralGroupBoxWidth = (l_GroupBoxWidthWithoutSpacer / 2);
                int l_ScopeGroupBoxWidth = (l_GroupBoxWidthWithoutSpacer - l_GeneralGroupBoxWidth);
                f_GeneralGroupBox = new GuiGroupBox(this);
                int l_GeneralInnerWidth = (l_GeneralGroupBoxWidth - f_GeneralGroupBox.Padding.Horizontal);
                f_GeneralEnableFilteringCheckBox = new GuiCheckBox(f_GeneralGroupBox, l_GeneralInnerWidth);
                f_GeneralEnforceDefaultScopesCheckBox = new GuiCheckBox(f_GeneralGroupBox, l_GeneralInnerWidth);
                f_TestRulesButton = new GuiButton(f_GeneralGroupBox, true);
                f_ScopeGroupBox = new GuiGroupBox(this);
                f_ScopeCheckBoxes = new ScopeCheckBoxes(f_ScopeGroupBox, l_ScopeGroupBoxWidth - f_ScopeGroupBox.Padding.Horizontal);
                f_RulesGroupBox = new GuiGroupBox(this);
                f_GridView = new GridView<uint, DummyStruct>(l_Inst, f_RulesGroupBox, l_Inst.f_MainForm.f_VisualCust.f_KadFilterPanelColSet);
                f_AddButton = new GuiButton(f_RulesGroupBox, true);
                f_ModifySelButton = new GuiButton(f_RulesGroupBox, true);
                f_RemoveSelButton = new GuiButton(f_RulesGroupBox, true);
                f_ImportButton = new GuiButton(f_RulesGroupBox, true);
                f_ExportSelButton = new GuiButton(f_RulesGroupBox, true);

                int l_LeftPadding = f_GeneralGroupBox.Padding.Left;
                f_GeneralEnableFilteringCheckBox.Location = new Point(l_LeftPadding, f_GeneralGroupBox.Padding.Top);
                f_GeneralEnforceDefaultScopesCheckBox.Location = new Point(l_LeftPadding, f_GeneralEnableFilteringCheckBox.GetBottom());
                f_TestRulesButton.Location = new Point(l_LeftPadding, f_GeneralEnforceDefaultScopesCheckBox.GetBottom());
                f_GeneralGroupBox.SetSizePlusBottomPadding(l_GeneralGroupBoxWidth, f_TestRulesButton.Bottom);
                f_ScopeGroupBox.Left = (f_GeneralGroupBox.Right + Lyt.s_Space);
                f_ScopeCheckBoxes.SetLocation(l_LeftPadding, f_GeneralGroupBox.Padding.Top);
                f_ScopeGroupBox.SetSizePlusBottomPadding(l_ScopeGroupBoxWidth, f_ScopeCheckBoxes.GetRawBottom());
                int l_RightControlsWidth = Lyt.s_GenerousButtonWidth;
                int l_LeftControlsWidth = (l_RulesGroupBoxWidth - f_RulesGroupBox.Padding.Horizontal - Lyt.s_Space - l_RightControlsWidth);
                int l_RightButtonsX = (l_LeftPadding + l_LeftControlsWidth + Lyt.s_Space);
                int l_RulesTop = Math.Max(f_GeneralGroupBox.GetBottom(), f_ScopeGroupBox.GetBottom());
                f_RulesGroupBox.Top = l_RulesTop;
                f_RulesGroupBox.Size = new Size(l_RulesGroupBoxWidth, a_Rect.Height - l_RulesTop);
                f_GridView.Dock = DockStyle.None;
                f_GridView.Location = new Point(l_LeftPadding, f_RulesGroupBox.Padding.Top);
                f_GridView.Size = new Size(l_LeftControlsWidth, f_RulesGroupBox.Height - f_RulesGroupBox.Padding.Vertical);
                f_GridView.LoadColumnAdjustments();
                // These are the upper buttons.
                f_AddButton.Location = new Point(l_RightButtonsX, f_GridView.Top);
                f_ModifySelButton.Location = new Point(l_RightButtonsX, f_AddButton.GetBottom());
                f_RemoveSelButton.Location = new Point(l_RightButtonsX, f_ModifySelButton.GetBottom());
                // These are the lower buttons.
                f_ImportButton.Location = new Point(l_RightButtonsX, f_GridView.Bottom - (2 * f_ImportButton.Height) - Lyt.s_SmallSpace);
                f_ExportSelButton.Location = new Point(l_RightButtonsX, f_ImportButton.GetBottom());

                f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
                f_GridView.e_UpdateItem += GridView_UpdateItem;
                f_GridView.SelectionChanged += GridView_SelectionChanged;
                f_GridView.CellClick += GridView_CellClick;
                f_AddButton.Click += AddButton_Click;
                f_ModifySelButton.Click += ModifySelButton_Click;
                f_RemoveSelButton.Click += RemoveSelButton_Click;
                f_ImportButton.Click += ImportButton_Click;
                f_ExportSelButton.Click += ExportSelButton_Click;
                f_TestRulesButton.Click += TestRulesButton_Click;

                TextSet l_TextSet = l_Inst.f_TextSet;
                this.Text = "Filter Dialog";
                f_GeneralGroupBox.Text = "General";
                f_GeneralEnableFilteringCheckBox.SetTextAndUpdateSize("Enable filtering");
                f_GeneralEnforceDefaultScopesCheckBox.SetTextAndUpdateSize("Enforce the default scope of application for all rules");
                f_TestRulesButton.Text = l_TextSet.GetAsDialogOpener(TextKeys.ButtonTextTest);
                f_ScopeGroupBox.Text = "Default Scope of Application";
                f_RulesGroupBox.Text = "Rules";
                f_AddButton.Text = "Add...";
                f_ModifySelButton.Text = "Modify Selected...";
                f_RemoveSelButton.Text = "Remove Selected";
                f_ImportButton.Text = "Import...";
                f_ExportSelButton.Text = "Export Selected...";
            }

            protected override void DisposeAutoPanelChildren()
            {
                f_GridView.StoreColumnAdjustments();

                Dialog.DisposeAllChildren(f_GeneralGroupBox);
                f_ScopeCheckBoxes.Dispose();
                Dialog.DisposeAllChildren(f_ScopeGroupBox);
                Dialog.DisposeAllChildren(f_RulesGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            private void LoadEnabledFlagAndDefaultScopesFromUI()
            {
                KademliaFilterScopes l_Scopes = f_ScopeCheckBoxes.p_Scopes;

                if (true == f_GeneralEnforceDefaultScopesCheckBox.Checked)
                {
                    l_Scopes |= KademliaFilterScopes.UseDefault;
                }

                f_Filter.SetEnabledFlagAndDefaultScopes(f_GeneralEnableFilteringCheckBox.Checked, l_Scopes);
            }

            public KademliaFilter CreateFilter()
            {
                LoadEnabledFlagAndDefaultScopesFromUI();

                return f_Filter.Clone();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == a_Default)
                {
                    f_Filter = a_Prefs.f_KadPrefs.f_KadFilter.Clone();
                }
                else
                {
                    KademliaFilter l_Blank = KademliaFilter.Create();
                    f_Filter.SetEnabledFlagAndDefaultScopes(l_Blank.p_Enabled, l_Blank.p_DefaultScopes);
                }

                f_GeneralEnableFilteringCheckBox.Checked = f_Filter.p_Enabled;
                KademliaFilterScopes l_Scopes = f_Filter.p_DefaultScopes;
                f_GeneralEnforceDefaultScopesCheckBox.Checked = KademliaFilter.IsDefaultScopesFlagSet(l_Scopes);
                f_ScopeCheckBoxes.p_Scopes = l_Scopes;

                UpdatedRules();
            }

            private bool GetRule(uint a_RuleID,
                                 out bool a_Enabled,
                                 out string a_Title,
                                 out KademliaFilterScopes a_Scopes,
                                 out string a_Rules)
            {
                return f_Filter.GetRule(a_RuleID, out a_Enabled, out a_Title, out a_Scopes, out a_Rules);
            }

            private void AddButton_Click(object a_Sender,
                                         EventArgs e)
            {
                if (KademliaFilter.c_MaxRuleCount <= f_Filter.p_Count)
                {
                    f_MainForm.ErrMsgBox("Maximum number of items reached.");

                    return;
                }

                using (FilterRuleDialog l_Dlg = new FilterRuleDialog(f_OptionsDlg.f_Inst, this, f_ScopeCheckBoxes.p_Scopes, KademliaFilter.c_NoRuleID))
                {
                    l_Dlg.ShowDialog(f_OptionsDlg);
                }
            }

            private void ModifySelButton_Click(object a_Sender,
                                               EventArgs e)
            {
                if (1 != f_GridView.p_SelectedCount)
                {
                    return;
                }

                uint l_RuleID = f_GridView.GetSelectedItem(0).f_Ref;
                ModifyRuleItem(l_RuleID);
            }

            private void ModifyRuleItem(uint a_RuleID)
            {
                using (FilterRuleDialog l_Dlg = new FilterRuleDialog(f_OptionsDlg.f_Inst, this, KademliaFilterScopes.None, a_RuleID))
                {
                    l_Dlg.ShowDialog(f_OptionsDlg);
                }
            }

            private bool TryAddOrModifyRule(uint a_RuleID,
                                            bool a_Enabled,
                                            string a_Title,
                                            KademliaFilterScopes a_Scopes,
                                            string a_Rules)
            {
                // a_RuleID may be KademliaFilter.c_NoRuleID if a new
                // rule should be created.

                bool l_Succ;
                uint l_SelRuleID;

                if (KademliaFilter.c_NoRuleID == a_RuleID)
                {
                    l_Succ = f_Filter.TryAddRule(a_Enabled, a_Title, a_Scopes, a_Rules, out l_SelRuleID);
                }
                else
                {
                    l_Succ = f_Filter.TryModifyRule(a_RuleID, a_Enabled, a_Title, a_Scopes, a_Rules);
                    l_SelRuleID = (false == l_Succ ? KademliaFilter.c_NoRuleID : a_RuleID);
                }

                if (true == l_Succ)
                {
                    UpdatedRules();

                    if (KademliaFilter.c_NoRuleID != l_SelRuleID)
                    {
                        SelectRule(l_SelRuleID);
                    }
                }

                return l_Succ;
            }

            private void RemoveSelButton_Click(object a_Sender,
                                               EventArgs e)
            {
                uint l_SelCount = (uint)f_GridView.p_SelectedCount;

                if (0 == l_SelCount ||
                    false == f_OptionsDlg.f_Inst.f_MainForm.YesNoMsgBox("Do you really want to remove the selected items?"))
                {
                    return;
                }

                if (f_Filter.p_Count == l_SelCount)
                {
                    f_Filter.ClearItems();
                }
                else
                {
                    foreach (GridViewItem<uint, DummyStruct> l_Item in f_GridView.SelectedRows)
                    {
                        bool l_Removed = f_Filter.TryRemoveRule(l_Item.f_Ref);
                    }
                }

                UpdatedRules();
            }

            private void ImportButton_Click(object a_Sender,
                                            EventArgs e)
            {
                string l_FilePath;

                using (OpenFileDialog l_Dlg = new OpenFileDialog())
                {
                    l_Dlg.Filter = c_FileDlgFilter;
                    l_Dlg.FilterIndex = 2;

                    if (DialogResult.OK != l_Dlg.ShowDialog(f_OptionsDlg))
                    {
                        return;
                    }

                    l_FilePath = l_Dlg.FileName;
                }

                ImportWorkerContext l_Cntxt = new ImportWorkerContext(f_Filter, l_FilePath);
                l_Cntxt.Start();

                if (false == WaitMessageDialog.ShowDialog(f_OptionsDlg.f_Inst.f_TextSet, f_OptionsDlg, l_Cntxt, WaitMessageDialog.ProgressBarStyles.DummyProgressBar, TextKeys.ImportKadFilterRulesDlgTitle, TextKeys.ImportKadFilterRulesDlgMsg))
                {
                    return;
                }

                KademliaFilter l_ImpFilter;
                bool l_ConflictingTitles;
                StorageStates l_State = l_Cntxt.GetReadFileResult(out l_ImpFilter, out l_ConflictingTitles);

                if (StorageStates.Success != l_State)
                {
                    string l_ErrMsg;

                    if (StorageStates.TooAdvanced == l_State)
                    {
                        l_ErrMsg = "The input seems to be from a newer version of this software and cannot be used by this version.";
                    }
                    else if (StorageStates.Corruption == l_State)
                    {
                        l_ErrMsg = "The input could not be interpreted and might be corrupt or there was an error while reading.";
                    }
                    else
                    {
                        l_ErrMsg = "The input file could not be accessed or is too large.";
                    }

                    f_MainForm.ErrMsgBox(l_ErrMsg);

                    return;
                }

                KademliaFilter.CollisionMergeMode l_Mode;

                if (false == l_ConflictingTitles)
                {
                    l_Mode = KademliaFilter.CollisionMergeMode.IgnoreImported;
                }
                else
                {
                    using (PickImportModeDialog l_Dlg = new PickImportModeDialog(f_OptionsDlg, f_OptionsDlg.f_Inst.f_TextSet))
                    {
                        if (DialogResult.OK != l_Dlg.ShowDialog(f_OptionsDlg))
                        {
                            return;
                        }

                        l_Mode = l_Dlg.p_Mode;
                    }
                }

                l_Cntxt = new ImportWorkerContext(f_Filter, l_ImpFilter, l_Mode);
                l_Cntxt.Start();

                if (false == WaitMessageDialog.ShowDialog(f_OptionsDlg.f_Inst.f_TextSet, f_OptionsDlg, l_Cntxt, WaitMessageDialog.ProgressBarStyles.DummyProgressBar, TextKeys.ImportKadFilterRulesDlgTitle, TextKeys.ImportKadFilterRulesDlgMsg))
                {
                    return;
                }

                bool l_HadOverflow;
                l_Cntxt.GetMergeResult(out l_HadOverflow);

                if (true == l_HadOverflow)
                {
                    f_MainForm.InfoMsgBox("Not all of the input could be taken over, as the maximum extent of the filter was reached in one or more cases.");
                }

                UpdatedRules();
            }

            private void ExportSelButton_Click(object a_Sender,
                                               EventArgs e)
            {
                uint l_SelCount = (uint)f_GridView.p_SelectedCount;

                if (0 == l_SelCount)
                {
                    return;
                }

                string l_FilePath;

                using (SaveFileDialog l_Dlg = new SaveFileDialog())
                {
                    l_Dlg.Filter = c_FileDlgFilter;
                    l_Dlg.FilterIndex = 2;

                    if (DialogResult.OK != l_Dlg.ShowDialog(f_ExportSelButton))
                    {
                        return;
                    }

                    l_FilePath = l_Dlg.FileName;
                }

                uint[] l_RuleIDs = f_GridView.SelectedItemsToKeyArr();
                MemoryIO l_Mem = MemoryIO.CreateWritable();

                if (false == f_Filter.ExportToBuffer(l_Mem, l_RuleIDs))
                {
                    f_MainForm.ShowOperationFailedMsg();
                }

                byte[] l_Bytes = l_Mem.ToArray(true);

                try
                {
                    System.IO.File.WriteAllBytes(l_FilePath, l_Bytes);
                }
                catch
                {
                    f_MainForm.ErrMsgBox("An error happened while exporting.");

                    return;
                }

                TextSet l_TextSet = f_OptionsDlg.f_Inst.f_TextSet;
                uint l_Count = (uint)f_GridView.p_ItemCount;
                string l_RatioStr = (l_Count <= l_SelCount ? null : l_TextSet.Format("\n\nNOTE: There are {0} of {1} rules selected. Only the selected rules were exported.", l_SelCount, l_Count));

                f_MainForm.InfoMsgBox(l_TextSet.Format("Successfully exported the selected rules to \"{0}\".{1}", l_FilePath, l_RatioStr));
            }

            private void TestRulesButton_Click(object a_Sender,
                                               EventArgs e)
            {
                LoadEnabledFlagAndDefaultScopesFromUI();

                if (false == f_Filter.HasActiveRules())
                {
                    f_MainForm.InfoMsgBox("Filtering is disabled globally, there are no rule items or all items are disabled.\n\nAs a consequence, nothing gets filtered out.");

                    return;
                }

                using (TextBoxDialog l_Dlg = new TextBoxDialog(f_OptionsDlg.f_Inst.f_TextSet, TextKeys.OptionsDlgKadFilterTestDlgTitle, TextKeys.OptionsDlgKadFilterTestDlgGroupBox, TextKeys.OptionsDlgKadFilterTestDlgLabel, TextKeys.OptionsDlgKadFilterTestDlgTextBox, FileNames.c_MaxFileNameChars + 10))
                {
                    l_Dlg.AdjustButtonTexts(TextKeys.ButtonTextTest, TextKeys.ButtonTextClose);
                    l_Dlg.e_OKClick += TestDlg_TestClick;
                    DialogResult l_DlgRes = l_Dlg.ShowDialog(f_TestRulesButton);
                }
            }

            private void TestDlg_TestClick(object a_Sender,
                                           EventArgs e)
            {
                TextBoxDialog l_Dlg = (TextBoxDialog)a_Sender;
                string l_FileName = l_Dlg.p_TextBoxText;
                string[] l_Keywords = null;

                if (0 < l_FileName.Length)
                {
                    l_Keywords = SearchManager.IdentifyKeywords(l_FileName, true);
                }

                if (null == l_Keywords)
                {
                    f_MainForm.InfoMsgBox("The specified file name does contain any keywords. It will PASS the filter in any case.");

                    return;
                }

                KademliaFilterScopes l_Scopes = f_Filter.TestIfFileNameGetsFilteredOut(l_FileName, KademliaFilterScopes.AllSpecificScopes);
                StringBuilder l_Bld = new StringBuilder();
                bool l_FilterIsInEffect = f_MainForm.p_Prefs.f_KadPrefs.f_KadFilter.Equals(f_Filter, true);

                if (false == l_FilterIsInEffect)
                {
                    l_Bld.AppendLine("Note that this summary is based on the current configuration as seen on the options dialog. This configuration was not applied yet and hence is not currently in effect throughout the application.");
                    l_Bld.AppendLine();
                }

                l_Bld.AppendLine("The specified file name contains the following keyword(s):");
                l_Bld.AppendLine();

                foreach (string l_Keyword in l_Keywords)
                {
                    l_Bld.Append("- ");
                    l_Bld.AppendLine(l_Keyword);
                }

                l_Bld.AppendLine();

                if (KademliaFilterScopes.None == l_Scopes)
                {
                    l_Bld.AppendLine("The file name PASSED the filter. It's not filtered out in any scope.");
                }
                else
                {
                    l_Bld.AppendLine("The file name gets FILTERED OUT concerning the following scope(s):");
                    l_Bld.AppendLine();

                    ScopeCheckBoxes.WriteScopes(l_Scopes, l_Bld);

                    if (KademliaFilterScopes.AllSpecificScopes != l_Scopes)
                    {
                        l_Bld.AppendLine();
                        l_Bld.AppendLine("The file name PASSED concerning all other scopes.");
                    }
                }

                f_MainForm.InfoMsgBox(l_Bld.ToString());
            }

            private void GridView_SelectionChanged(object a_Sender,
                                                   EventArgs e)
            {
                UpdateButtons();
            }

            private void UpdatedRules()
            {
                uint[] l_RuleIDs = f_Filter.CreateArrayOfRuleIDs();

                if (null == l_RuleIDs)
                {
                    f_GridView.ClearItems();
                }
                else
                {
                    f_GridView.StartChanging();

                    foreach (uint l_RuleID in l_RuleIDs)
                    {
                        GridViewItem<uint, DummyStruct> l_Item = f_GridView.CreateItem(l_RuleID);
                    }

                    f_GridView.EndChanging();
                }

                UpdateButtons();
            }

            private void UpdateButtons()
            {
                uint l_SelCount = (uint)f_GridView.p_SelectedCount;

                f_ModifySelButton.Enabled = (1 == l_SelCount);
                f_ExportSelButton.Enabled = f_RemoveSelButton.Enabled = (0 != l_SelCount);
            }

            private void SelectRule(uint a_RuleID)
            {
                // Selects the rule in the list view.

                GridViewItem<uint, DummyStruct> l_Item = f_GridView.GetItemByKey(a_RuleID);

                f_GridView.SelectAndDisplayItem(l_Item);
            }

            private void GridView_CellClick(object a_Sender,
                                            DataGridViewCellEventArgs e)
            {
                int l_RowIdx = e.RowIndex;

                if (0 != e.ColumnIndex ||
                    0 > l_RowIdx)
                {
                    return;
                }

                uint l_RuleID = f_GridView.GetItem(l_RowIdx).f_Ref;
                f_Filter.ToggleRuleEnabledFlag(l_RuleID);

                f_GridView.NotifyValuesChanged(false);
                f_GridView.UpdateCellValue(0, l_RowIdx);
            }

            private void GridView_ItemDoubleClicked(GridViewItem<uint, DummyStruct> a_Item)
            {
                ModifyRuleItem(a_Item.f_Ref);
            }

            private void GridView_UpdateItem(Time a_Time,
                                             GridViewItem<uint, DummyStruct> a_Item)
            {
                uint l_RuleID = a_Item.f_Ref;
                bool l_Active;
                string l_Title;
                KademliaFilterScopes l_Junk0;
                string l_Junk1;
                bool l_OK = GetRule(l_RuleID, out l_Active, out l_Title, out l_Junk0, out l_Junk1);

                a_Item.SetBool(0, l_Active);
                a_Item.SetTextWithCompObj(1, l_Title, l_Title);
            }

            private class ImportWorkerContext : UnabortableControllerContext
            {
                private const string c_ImportKadFilterWorkerName = "ImportKadFilterWorker";

                private readonly KademliaFilter f_OriginalFilter;

                private readonly string f_SrcFilePath;
                private StorageStates f_StorageState;
                private KademliaFilter f_OutFilter;
                private bool f_ConflictingTitles;

                private readonly KademliaFilter f_ImportFilter;
                private readonly KademliaFilter.CollisionMergeMode f_CollisionMergeMode;
                private bool f_HadOverflow;

                private bool p_ReadFileMode
                {
                    get { return (null != f_SrcFilePath); }
                }

                public ImportWorkerContext(KademliaFilter a_OriginalFilter,
                                           string a_SrcFilePath)
                {
                    f_OriginalFilter = a_OriginalFilter;
                    f_SrcFilePath = a_SrcFilePath;
                }

                public ImportWorkerContext(KademliaFilter a_OriginalFilter,
                                           KademliaFilter a_ImportFilter,
                                           KademliaFilter.CollisionMergeMode a_CollisionMergeMode)
                {
                    f_OriginalFilter = a_OriginalFilter;
                    f_ImportFilter = a_ImportFilter;
                    f_CollisionMergeMode = a_CollisionMergeMode;
                }

                public void Start()
                {
                    Util.StartThread(c_ImportKadFilterWorkerName, ImportWorker);
                }

                private void ImportWorker()
                {
                    if (false == p_ReadFileMode)
                    {
                        Merge();
                    }
                    else
                    {
                        ReadFile();
                    }

                    base.SetFinishedBase();
                }

                private void ReadFile()
                {
                    ByteArray l_Data = new ByteArray();
                    bool l_IOErr = true;

                    try
                    {
                        using (System.IO.FileStream l_Strm = System.IO.File.OpenRead(f_SrcFilePath))
                        {
                            long l_FileSize = l_Strm.Length;

                            if ((KademliaFilter.c_MaxUncompressedByteSize + 1024) >= l_FileSize)
                            {
                                l_Data = new ByteArray((int)l_FileSize);

                                if (l_Data.f_Len == l_Strm.Read(l_Data.f_Arr, 0, l_Data.f_Len))
                                {
                                    l_IOErr = false;
                                }
                            }
                        }
                    }
                    catch (ArgumentException)
                    {
                    }
                    catch (System.IO.IOException)
                    {
                    }
                    catch (UnauthorizedAccessException)
                    {
                    }
                    catch (NotSupportedException)
                    {
                    }
                    catch (ObjectDisposedException)
                    {
                    }

                    if (true == l_IOErr)
                    {
                        f_StorageState = StorageStates.ReadFailed;

                        return;
                    }

                    MemoryIO l_Mem = MemoryIO.CreateReadable(ref l_Data);
                    f_StorageState = KademliaFilter.CreateFromBuffer(l_Mem, true, out f_OutFilter);
                    f_ConflictingTitles = (StorageStates.Success == f_StorageState &&
                                           true == f_OriginalFilter.CheckForConflictingTitles(f_OutFilter));
                }

                public StorageStates GetReadFileResult(out KademliaFilter a_Filter,
                                                       out bool a_ConflictingTitles)
                {
                    a_Filter = f_OutFilter;
                    a_ConflictingTitles = f_ConflictingTitles;

                    return f_StorageState;
                }

                private void Merge()
                {
                    f_OriginalFilter.MergeRules(f_ImportFilter, f_CollisionMergeMode, out f_HadOverflow);
                }

                public void GetMergeResult(out bool a_HadOverflow)
                {
                    a_HadOverflow = f_HadOverflow;
                }
            }

            private sealed class ScopeCheckBoxes : IDisposable
            {
                private static string[] s_ScopeTextsForWriting = new string[]
                {
                    "Local Kademlia index",
                    "Own publishing",
                    "Search responses for others",
                    "Own search results",
                };

                private readonly GuiCheckBox f_ScopeKadIndexCheckBox;
                private readonly GuiCheckBox f_ScopePublishingCheckBox;
                private readonly GuiCheckBox f_ScopeSearchResultsForOthersCheckBox;
                private readonly GuiCheckBox f_ScopeSearchResultsForSelfCheckBox;

                private bool f_ScopesDirty;
                private KademliaFilterScopes f_Scopes;

                public KademliaFilterScopes p_Scopes
                {
                    get { return GetScopesInternal(); }
                    set { SetScopesInternal(value); }
                }

                public ScopeCheckBoxes(Control a_Owner,
                                       int a_MaxWidth)
                {
                    f_ScopeKadIndexCheckBox = new GuiCheckBox(a_Owner, a_MaxWidth);
                    f_ScopePublishingCheckBox = new GuiCheckBox(a_Owner, a_MaxWidth);
                    f_ScopeSearchResultsForOthersCheckBox = new GuiCheckBox(a_Owner, a_MaxWidth);
                    f_ScopeSearchResultsForSelfCheckBox = new GuiCheckBox(a_Owner, a_MaxWidth);

                    f_ScopeKadIndexCheckBox.CheckedChanged += CheckBox_CheckedChanged;
                    f_ScopePublishingCheckBox.CheckedChanged += CheckBox_CheckedChanged;
                    f_ScopeSearchResultsForOthersCheckBox.CheckedChanged += CheckBox_CheckedChanged;
                    f_ScopeSearchResultsForSelfCheckBox.CheckedChanged += CheckBox_CheckedChanged;

                    f_ScopeKadIndexCheckBox.SetTextAndUpdateSize("Filter what I index (local Kademlia index)");
                    f_ScopePublishingCheckBox.SetTextAndUpdateSize("Filter what I publish to others");
                    f_ScopeSearchResultsForOthersCheckBox.SetTextAndUpdateSize("Filter what others can find through me");
                    f_ScopeSearchResultsForSelfCheckBox.SetTextAndUpdateSize("Filter what I find (search results)");
                }

                public void Dispose()
                {
                    f_ScopeKadIndexCheckBox.Dispose();
                    f_ScopePublishingCheckBox.Dispose();
                    f_ScopeSearchResultsForOthersCheckBox.Dispose();
                    f_ScopeSearchResultsForSelfCheckBox.Dispose();
                }

                private void CheckBox_CheckedChanged(object a_Sender,
                                                     EventArgs e)
                {
                    MarkScopesAsDirty();
                }

                private void MarkScopesAsDirty()
                {
                    f_ScopesDirty = true;
                }

                private KademliaFilterScopes GetScopesInternal()
                {
                    if (true == f_ScopesDirty)
                    {
                        f_ScopesDirty = false;
                        f_Scopes = KademliaFilterScopes.None;

                        SetFlagIfCheckBoxIsSet(f_ScopeKadIndexCheckBox, KademliaFilterScopes.ScopeLocalKademliaIndex, ref f_Scopes);
                        SetFlagIfCheckBoxIsSet(f_ScopePublishingCheckBox, KademliaFilterScopes.ScopeOwnPublishingToOthers, ref f_Scopes);
                        SetFlagIfCheckBoxIsSet(f_ScopeSearchResultsForOthersCheckBox, KademliaFilterScopes.ScopeSearchResultsForOthers, ref f_Scopes);
                        SetFlagIfCheckBoxIsSet(f_ScopeSearchResultsForSelfCheckBox, KademliaFilterScopes.ScopeSearchResultsForSelf, ref f_Scopes);
                    }

                    return f_Scopes;
                }

                private void SetScopesInternal(KademliaFilterScopes a_Scopes)
                {
                    // a_Scopes may have the KademliaFilterScopes.UseDefault
                    // flag set. That flag will be ignored.

                    KademliaFilterScopes l_EffScopes = (a_Scopes & KademliaFilterScopes.AllSpecificScopes);

                    if (false == f_ScopesDirty &&
                        l_EffScopes == f_Scopes)
                    {
                        return;
                    }

                    f_ScopesDirty = false;
                    f_Scopes = l_EffScopes;

                    UpdateCheckBoxState(f_ScopeKadIndexCheckBox, KademliaFilterScopes.ScopeLocalKademliaIndex);
                    UpdateCheckBoxState(f_ScopePublishingCheckBox, KademliaFilterScopes.ScopeOwnPublishingToOthers);
                    UpdateCheckBoxState(f_ScopeSearchResultsForOthersCheckBox, KademliaFilterScopes.ScopeSearchResultsForOthers);
                    UpdateCheckBoxState(f_ScopeSearchResultsForSelfCheckBox, KademliaFilterScopes.ScopeSearchResultsForSelf);
                }

                private static void SetFlagIfCheckBoxIsSet(GuiCheckBox a_CheckBox,
                                                           KademliaFilterScopes a_Scope,
                                                           ref KademliaFilterScopes a_DestScopes)
                {
                    // This is a helper for GetScopesInternal().

                    if (true == a_CheckBox.Checked)
                    {
                        a_DestScopes |= a_Scope;
                    }
                }

                private void UpdateCheckBoxState(GuiCheckBox a_CheckBox,
                                                 KademliaFilterScopes a_Scope)
                {
                    // This is a helper for SetScopesInternal().

                    a_CheckBox.Checked = (KademliaFilterScopes.None != (f_Scopes & a_Scope));
                }

                public void AdjustEnabledFlag(bool a_Enable)
                {
                    f_ScopeSearchResultsForSelfCheckBox.Enabled = f_ScopeSearchResultsForOthersCheckBox.Enabled = f_ScopePublishingCheckBox.Enabled = f_ScopeKadIndexCheckBox.Enabled = a_Enable;
                }

                public void SetLocation(int a_x,
                                        int a_y)
                {
                    f_ScopeKadIndexCheckBox.Location = new Point(a_x, a_y);
                    f_ScopePublishingCheckBox.Location = new Point(a_x, f_ScopeKadIndexCheckBox.GetBottom());
                    f_ScopeSearchResultsForOthersCheckBox.Location = new Point(a_x, f_ScopePublishingCheckBox.GetBottom());
                    f_ScopeSearchResultsForSelfCheckBox.Location = new Point(a_x, f_ScopeSearchResultsForOthersCheckBox.GetBottom());
                }

                public int GetRawBottom()
                {
                    return f_ScopeSearchResultsForSelfCheckBox.Bottom;
                }

                public static void WriteScopes(KademliaFilterScopes a_Scopes,
                                               StringBuilder a_Bld)
                {
                    for (uint i = 0, l_Mask = 1; true; i++, l_Mask <<= 1)
                    {
                        if (0 == (l_Mask & (uint)KademliaFilterScopes.AllSpecificScopes))
                        {
                            break;
                        }

                        if (0 == (l_Mask & (uint)a_Scopes))
                        {
                            continue;
                        }

                        a_Bld.Append("- ");
                        a_Bld.AppendLine(s_ScopeTextsForWriting[i]);
                    }
                }
            }

            private sealed class PickImportModeDialog : ButtonDialog
            {
                public static KademliaFilter.CollisionMergeMode s_LastSelMode = KademliaFilter.CollisionMergeMode.MergeImportedIntoExisting;

                private RadioButtonBattery f_RadioButtons;

                private KademliaFilter.CollisionMergeMode f_Mode;

                public KademliaFilter.CollisionMergeMode p_Mode
                {
                    get
                    {
                        return f_Mode;
                    }
                }

                public PickImportModeDialog(Dialog a_Owner,
                                            TextSet a_TextSet) :
                                       base(a_TextSet, DialogButtons.OKCancel)
                {
                    int l_DlgWidth = Lyt.CalcX(500);
                    f_RadioButtons = new RadioButtonBattery(this, l_DlgWidth - Lyt.s_Space2, 4);
                    f_RadioButtons.Location = new Point(Lyt.s_Space, Lyt.s_Space);

                    base.AsFixedButtonDialog(l_DlgWidth, f_RadioButtons);
                    base.SetSelectOnShowControl(f_OKButton);

                    f_OKButton.Click += OKButton_Click;

                    f_RadioButtons.SetTexts("Mode for dealing with an imported rule that has the same title as an existing rule", "Do not import that rule and leave the existing one", "Overwrite the existing rule with the imported one", "Merge the imported rule into the existing rule", "Import the rule with a different, auto-generated title");
                    f_CancelButton.Text = a_TextSet[TextKeys.ButtonTextAbort];
                }

                protected override void Dispose(bool a_Disposing)
                {
                    if (true == a_Disposing)
                    {
                        f_RadioButtons.Dispose();
                    }

                    base.Dispose(a_Disposing);
                }

                protected override void OnShown(EventArgs e)
                {
                    // Need to call base class implementation.

                    f_Mode = KademliaFilter.CollisionMergeMode.Undefined;
                    f_RadioButtons.SetSelectedIndex((int)s_LastSelMode - 1);

                    base.OnShown(e);
                }

                private void OKButton_Click(object a_Sender,
                                            EventArgs e)
                {
                    KademliaFilter.CollisionMergeMode l_Mode = (KademliaFilter.CollisionMergeMode)(f_RadioButtons.GetSelectedIndex() + 1);

                    f_Mode = s_LastSelMode = l_Mode;

                    base.DefaultOKCloseHandler(a_Sender, e);
                }
            }

            private sealed class FilterRuleDialog : ButtonDialog
            {
                private readonly Instance f_Inst;
                private readonly KadFilterPanel f_OwnerPanel;
                private readonly KademliaFilterScopes f_DefaultScopes;
                private readonly uint f_RuleID;

                private readonly GuiGroupBox f_GeneralGroupBox;
                private readonly GuiCheckBox f_GeneralEnabledCheckBox;
                private readonly LabeledTextBox f_GeneralTitleTextBox;

                private readonly GuiGroupBox f_ScopesGroupBox;
                private readonly GuiRadioButton f_ScopesUseDefaultRadioButton;
                private readonly GuiRadioButton f_ScopesUseCustomRadioButton;
                private readonly ScopeCheckBoxes f_ScopesCheckBoxes;

                private readonly GuiGroupBox f_RulesGroupBox;
                private readonly GuiFilterTextBox f_RulesAndTextBox;
                private readonly GuiFilterTextBox f_RulesOrTextBox;
                private readonly GuiFilterTextBox f_RulesNotTextBox;

                public FilterRuleDialog(Instance a_Inst,
                                        KadFilterPanel a_OwnerPanel,
                                        KademliaFilterScopes a_DefaultScopes,
                                        uint a_RuleID) :
                                   base(a_Inst.f_TextSet, DialogButtons.OKCancel)
                {
                    f_Inst = a_Inst;
                    f_OwnerPanel = a_OwnerPanel;
                    f_DefaultScopes = a_DefaultScopes;
                    f_RuleID = a_RuleID;

                    const int lc_TitleNativeDlgWidth = 350;
                    const int lc_RulesNativeDlgWidth = 550;
                    int l_TitleCausedDlgWidth = Lyt.CalcX(lc_TitleNativeDlgWidth);
                    int l_RulesCausedDlgWidth = Lyt.CalcX(lc_RulesNativeDlgWidth);
                    int l_DlgWidth = Math.Max(l_TitleCausedDlgWidth, l_RulesCausedDlgWidth);
                    int l_RulesGroupBoxWidth = (l_RulesCausedDlgWidth - Lyt.s_Space2);
                    int l_GroupBoxWidthWithoutSpacer = (l_RulesGroupBoxWidth - Lyt.s_Space);
                    int l_GeneralGroupBoxWidth = (l_GroupBoxWidthWithoutSpacer / 2);
                    int l_ScopeGroupBoxWidth = (l_GroupBoxWidthWithoutSpacer - l_GeneralGroupBoxWidth);
                    int l_ControlIndent = Lyt.s_LargeSpace;
                    f_GeneralGroupBox = new GuiGroupBox(this);
                    int l_GeneralInnerWidth = (l_GeneralGroupBoxWidth - f_GeneralGroupBox.Padding.Horizontal);
                    f_GeneralEnabledCheckBox = new GuiCheckBox(f_GeneralGroupBox, l_GeneralInnerWidth);
                    f_GeneralTitleTextBox = new LabeledTextBox(f_GeneralGroupBox, l_GeneralInnerWidth);
                    f_ScopesGroupBox = new GuiGroupBox(this);
                    int l_ScopeInnerWidth = (l_GeneralGroupBoxWidth - f_GeneralGroupBox.Padding.Horizontal);
                    f_ScopesUseDefaultRadioButton = new GuiRadioButton(f_ScopesGroupBox, l_ScopeInnerWidth);
                    f_ScopesUseCustomRadioButton = new GuiRadioButton(f_ScopesGroupBox, l_ScopeInnerWidth);
                    f_ScopesCheckBoxes = new ScopeCheckBoxes(f_ScopesGroupBox, l_ScopeInnerWidth - l_ControlIndent);
                    f_RulesGroupBox = new GuiGroupBox(this);
                    int l_RulesInnerWidth = (l_RulesGroupBoxWidth - f_RulesGroupBox.Padding.Horizontal);
                    f_RulesAndTextBox = new GuiFilterTextBox(f_RulesGroupBox, l_RulesInnerWidth);
                    f_RulesOrTextBox = new GuiFilterTextBox(f_RulesGroupBox, l_RulesInnerWidth);
                    f_RulesNotTextBox = new GuiFilterTextBox(f_RulesGroupBox, l_RulesInnerWidth);

                    int l_LeftPadding = f_GeneralGroupBox.Padding.Left;
                    f_GeneralGroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                    f_GeneralEnabledCheckBox.Location = new Point(l_LeftPadding, f_GeneralGroupBox.Padding.Top);
                    f_GeneralTitleTextBox.SetLocation(l_LeftPadding, f_GeneralEnabledCheckBox.GetBottom());
                    f_GeneralGroupBox.SetSizePlusBottomPadding(l_GeneralGroupBoxWidth, f_GeneralTitleTextBox.f_Box.Bottom);
                    f_ScopesGroupBox.Location = new Point(f_GeneralGroupBox.Right + Lyt.s_Space, Lyt.s_Space);
                    f_ScopesUseDefaultRadioButton.Location = new Point(l_LeftPadding, f_ScopesGroupBox.Padding.Top);
                    f_ScopesUseCustomRadioButton.Location = new Point(l_LeftPadding, f_ScopesUseDefaultRadioButton.GetBottom());
                    int l_CheckBoxLeft = (l_LeftPadding + l_ControlIndent);
                    f_ScopesCheckBoxes.SetLocation(l_CheckBoxLeft, f_ScopesUseCustomRadioButton.GetBottom() + Lyt.s_SmallSpace);
                    f_ScopesGroupBox.SetSizePlusBottomPadding(l_ScopeGroupBoxWidth, f_ScopesCheckBoxes.GetRawBottom());
                    int l_RightControlsWidth = Lyt.s_GenerousButtonWidth;
                    int l_LeftControlsWidth = (l_RulesGroupBoxWidth - f_RulesGroupBox.Padding.Horizontal - Lyt.s_Space - l_RightControlsWidth);
                    int l_RightButtonsX = (l_LeftPadding + l_LeftControlsWidth + Lyt.s_Space);
                    int l_RulesTop = Math.Max(f_GeneralGroupBox.GetBottom(), f_ScopesGroupBox.GetBottom());
                    f_RulesGroupBox.Location = new Point(Lyt.s_Space, l_RulesTop);
                    f_RulesAndTextBox.SetLocation(l_LeftPadding, f_RulesGroupBox.Padding.Top);
                    f_RulesOrTextBox.SetLocation(l_LeftPadding, f_RulesAndTextBox.GetBottom());
                    f_RulesNotTextBox.SetLocation(l_LeftPadding, f_RulesOrTextBox.GetBottom());
                    f_RulesGroupBox.SetSizePlusBottomPadding(l_RulesGroupBoxWidth, f_RulesNotTextBox.f_Box.Bottom);

                    base.AsFixedButtonDialog(l_DlgWidth, f_RulesGroupBox);
                    base.SetSelectOnShowControl(f_GeneralTitleTextBox.f_Box);
                    this.AcceptButton = null;
                    base.CancelButton = null;

                    f_ScopesUseDefaultRadioButton.CheckedChanged += ScopeUseDefaultRadioButton_CheckedChanged;
                    f_RulesAndTextBox.e_ButtonClick += FilterTextBox_ButtonClick;
                    f_RulesOrTextBox.e_ButtonClick += FilterTextBox_ButtonClick;
                    f_RulesNotTextBox.e_ButtonClick += FilterTextBox_ButtonClick;
                    f_OKButton.Click += OKButton_Click;

                    f_RulesAndTextBox.SetFilterText("Block file names containing all of the following words Block file names containing all of the following words all of the following words all of the following words");
                    f_RulesOrTextBox.SetFilterText("Block file names containing all of the following words Block file names containing all of the following");
                    f_RulesNotTextBox.SetFilterText("Block file names containing all of the following words Block file names containing all of the");

                    this.Text = "Add/Modify Rule";
                    f_GeneralGroupBox.Text = "General";
                    f_GeneralEnabledCheckBox.SetTextAndUpdateSize("Enabled");
                    f_GeneralTitleTextBox.SetCaption("Rule title:");
                    f_ScopesGroupBox.Text = "Scope of Application";
                    f_ScopesUseDefaultRadioButton.SetTextAndUpdateSize("Use the global default scope");
                    f_ScopesUseCustomRadioButton.SetTextAndUpdateSize("Use the following scope");
                    f_RulesGroupBox.Text = "Rules";
                    f_RulesAndTextBox.SetCaption("Block file names containing all of the following words (AND):");
                    f_RulesOrTextBox.SetCaption("Block file names containing at least one of the following words (OR):");
                    f_RulesNotTextBox.SetCaption("But don't block file names regardless of the above if they contain any of these words (NOT):");

                    Reset();
                }

                protected override void Dispose(bool a_Disposing)
                {
                    if (true == a_Disposing)
                    {
                        Dialog.DisposeAllChildren(f_GeneralGroupBox);
                        f_ScopesCheckBoxes.Dispose();
                        Dialog.DisposeAllChildren(f_ScopesGroupBox);
                        Dialog.DisposeAllChildren(f_RulesGroupBox);
                        f_GeneralGroupBox.Dispose();
                        f_ScopesGroupBox.Dispose();
                        f_RulesGroupBox.Dispose();
                    }

                    base.Dispose(a_Disposing);
                }

                private void Reset()
                {
                    if (KademliaFilter.c_NoRuleID == f_RuleID)
                    {
                        ClearDialog();

                        return;
                    }

                    bool l_Active;
                    string l_Title;
                    KademliaFilterScopes l_Scopes;
                    string l_Rules;
                    bool l_OK = f_OwnerPanel.GetRule(f_RuleID, out l_Active, out l_Title, out l_Scopes, out l_Rules);

                    f_GeneralEnabledCheckBox.Checked = l_Active;
                    f_GeneralTitleTextBox.f_Box.Text = l_Title;
                    f_GeneralTitleTextBox.f_Box.Select(0, 0);

                    if (false == KademliaFilter.IsDefaultScopesFlagSet(l_Scopes))
                    {
                        f_ScopesUseCustomRadioButton.Checked = true;
                    }
                    else
                    {
                        f_ScopesUseDefaultRadioButton.Checked = true;
                    }

                    f_ScopesCheckBoxes.p_Scopes = l_Scopes;

                    string l_AndSeg;
                    string l_OrSeg;
                    string l_NotSeg;
                    KademliaFilter.SplitRuleSequences(l_Rules, out l_AndSeg, out l_OrSeg, out l_NotSeg);

                    f_RulesAndTextBox.SetFilterText(l_AndSeg);
                    f_RulesOrTextBox.SetFilterText(l_OrSeg);
                    f_RulesNotTextBox.SetFilterText(l_NotSeg);
                }

                private void ClearDialog()
                {
                    f_GeneralEnabledCheckBox.Checked = true;
                    f_GeneralTitleTextBox.f_Box.Clear();
                    f_ScopesUseDefaultRadioButton.Checked = true;
                    f_ScopesCheckBoxes.p_Scopes = f_DefaultScopes;
                    f_RulesAndTextBox.SetFilterText(null);
                    f_RulesOrTextBox.SetFilterText(null);
                    f_RulesNotTextBox.SetFilterText(null);
                }

                private void ScopeUseDefaultRadioButton_CheckedChanged(object a_Sender,
                                                                       EventArgs e)
                {
                    f_ScopesCheckBoxes.AdjustEnabledFlag(f_ScopesUseCustomRadioButton.Checked);
                }

                private void FilterTextBox_ButtonClick(object a_Sender,
                                                       EventArgs e)
                {
                    GuiFilterTextBox l_FilterTextBox = (GuiFilterTextBox)a_Sender;
                    string l_RuleSequence;

                    if (false == TextItemsDialog.ShowDialog(f_Inst, uint.MaxValue, KademliaFilter.c_MaxPatternChars, KademliaFilter.c_MaxRuleSequenceChars, KademliaFilter.c_RulePatternSeparatorChar, TextItemsDialog_AcceptText, f_OwnerPanel.f_OptionsDlg, l_FilterTextBox.p_FilterText, out l_RuleSequence))
                    {
                        return;
                    }

                    // l_RuleSequence can be null.

                    l_FilterTextBox.SetFilterText(l_RuleSequence);
                }

                private static bool TextItemsDialog_AcceptText(TextItemsDialog a_Sender,
                                                               ref string a_ProposedText,
                                                               out string a_ErrMsg)
                {
                    string l_Text = a_ProposedText.ToLowerInvariant();
                    uint l_TextLen = Util.GetStringLength(l_Text);
                    StringBuilder l_Bld = new StringBuilder((int)l_TextLen);

                    bool l_WasWritten;
                    bool l_IsValidDirectly = KademliaFilter.IsValidPattern(l_Text, 0, l_TextLen, l_Bld, out l_WasWritten);

                    if (true == l_IsValidDirectly ||
                        true == l_WasWritten)
                    {
                        a_ProposedText = (false == l_IsValidDirectly ? l_Bld.ToString() : l_Text);
                        a_ErrMsg = null;

                        return true;
                    }

                    a_ErrMsg = "The specified pattern is invalid and cannot be used.";

                    return false;
                }

                private void OKButton_Click(object a_Sender,
                                            EventArgs e)
                {
                    string l_Title = f_GeneralTitleTextBox.GetBoxTextTrimmedAndNormalized();

                    if (false == KademliaFilter.IsValidTitle(l_Title))
                    {
                        f_OwnerPanel.f_MainForm.ErrMsgBox("The specified title is not valid.");
                        f_GeneralTitleTextBox.f_Box.Focus();

                        return;
                    }

                    KademliaFilterScopes l_Scopes = f_ScopesCheckBoxes.p_Scopes;

                    if (true == f_ScopesUseDefaultRadioButton.Checked)
                    {
                        l_Scopes |= KademliaFilterScopes.UseDefault;
                    }

                    bool l_Enabled = f_GeneralEnabledCheckBox.Checked;
                    string l_Rules = KademliaFilter.JoinRuleSequences(f_RulesAndTextBox.p_FilterText, f_RulesOrTextBox.p_FilterText, f_RulesNotTextBox.p_FilterText);
                    bool l_Accepted = f_OwnerPanel.TryAddOrModifyRule(f_RuleID, l_Enabled, l_Title, l_Scopes, l_Rules);

                    if (true == l_Accepted)
                    {
                        base.DefaultOKCloseHandler(a_Sender, e);

                        return;
                    }

                    if (KademliaFilter.c_NoRuleID == f_RuleID &&
                        KademliaFilter.c_MaxRuleCount <= f_OwnerPanel.f_Filter.p_Count)
                    {
                        f_OwnerPanel.f_MainForm.ShowOperationFailedMsg();

                        return;
                    }

                    f_OwnerPanel.f_MainForm.ErrMsgBox("The specified title is already occupied by another rule.");
                    f_GeneralTitleTextBox.f_Box.Focus();
                }
            }
        }

        private class eD2kPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 350;

            private readonly OwnPersonaControl f_PersonaCntrl;

            private readonly GuiGroupBox f_GroupBox;
            private readonly LabeledTextBox f_UserNameBox;

            public OwnUserPersona p_OwnPersona
            {
                get { return f_PersonaCntrl.p_OwnPersona; }
            }

            public string p_UserName
            {
                get { return f_UserNameBox.f_Box.Text.Trim(); }
            }

            public eD2kPanel(OptionsDialog a_OptionsDlg,
                             ref Rectangle a_Rect) :
                        base(a_OptionsDlg, ref a_Rect)
            {
                TextSet l_TextSet = f_OptionsDlg.f_Inst.f_TextSet;
                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                f_PersonaCntrl = new OwnPersonaControl(l_TextSet, this, l_Width);
                f_GroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_Width - f_GroupBox.Padding.Horizontal);
                f_UserNameBox = new LabeledTextBox(f_GroupBox, l_InnerWidth);

                f_GroupBox.Top = f_PersonaCntrl.GetBottom();
                f_UserNameBox.SetLocation(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top);
                f_GroupBox.Size = new Size(l_Width, f_UserNameBox.f_Box.Bottom + f_GroupBox.Padding.Bottom);
                this.ChildBoundsUpdated();

                f_UserNameBox.f_Box.MaxLength = Network.eD2k.PeerInfo.c_MaxUserNameChars;

                f_PersonaCntrl.e_Reset += PersonaCntrl_Reset;

                f_GroupBox.Text = l_TextSet[TextKeys.OptionsDlgeD2kAppearance];
                f_UserNameBox.SetCaption(l_TextSet, TextKeys.OptionsDlgeD2kAppearanceNickname);
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_GroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == a_Default)
                {
                    f_PersonaCntrl.Set(f_OptionsDlg.f_Inst.f_TextSet, a_Prefs.f_GeneralPrefs.f_OwnPersona);
                }

                f_UserNameBox.f_Box.Text = a_Prefs.f_NetworkPrefs.f_UserName;
            }

            private void PersonaCntrl_Reset()
            {
                if (false == f_OptionsDlg.f_Inst.f_MainForm.SeriousYesNoChoice(f_OptionsDlg.f_Inst[TextKeys.OptionsDlgeD2kUserMsgWarnChange]))
                {
                    return;
                }

                TextSet l_TextSet = f_OptionsDlg.f_Inst.f_TextSet;
                OwnUserPersona l_NewPersona;

                if (false == KeyGenerationWaitDialog.ShowDialog(l_TextSet, this, true, out l_NewPersona))
                {
                    return;
                }

                // We have to check if l_NewPersona is zero in any
                // case.

                if (null == l_NewPersona)
                {
                    f_OptionsDlg.f_Inst.f_MainForm.ErrMsgBox(l_TextSet[TextKeys.MsgOperationAbortedByUser]);

                    return;
                }

                // We always need to delete the addresses we used when
                // changing the user.
                f_OptionsDlg.f_AddrPanel.ClearKeyPairs();
                f_PersonaCntrl.Set(l_TextSet, l_NewPersona);
            }
        }

        private class SharedDirsPanel : OptionsPanel
        {
            private readonly Instance f_Inst;
            private readonly TextSet f_TextSet;
            private readonly Dictionary<RefDir, RefDir> f_Dirs;

            // f_Tag is unused (so use the bool type).
            private readonly GridView<RefDir, bool> f_GridView;
            private readonly GuiButton f_AddButton;
            private readonly GuiButton f_EditButton;
            private readonly GuiButton f_RemoveButton;
            private readonly GuiButton f_MarkAsOutputButton;
            private readonly GuiButton f_ImportButton;

            private readonly Font f_BoldFont;

            private RefDir f_ComplTrgtDir;

            public SharedDirsPanel(OptionsDialog a_OptionsDlg,
                                   ref Rectangle a_Rect) :
                              base(a_OptionsDlg, ref a_Rect)
            {
                f_Inst = a_OptionsDlg.f_Inst;
                f_TextSet = f_Inst.f_TextSet;
                f_Dirs = new Dictionary<RefDir, RefDir>();

                int l_Width = (a_Rect.Width - base.f_ScrollBarWidth - Lyt.s_SmallSpace);
                f_GridView = new GridView<RefDir, bool>(f_Inst, this, f_Inst.f_MainForm.f_VisualCust.f_SharedDirsPanelColSet);
                f_AddButton = new GuiButton(this, true);
                f_EditButton = new GuiButton(this, true);
                f_RemoveButton = new GuiButton(this, true);
                f_MarkAsOutputButton = new GuiButton(this, true);
                f_ImportButton = new GuiButton(this, true);
                f_BoldFont = MainForm.RecreateFont(f_MarkAsOutputButton.Font, 0, FontStyle.Bold);

                this.AdjustVertAutoScroll(false);
                f_GridView.Dock = DockStyle.None;
                f_GridView.Size = new Size(l_Width - f_AddButton.Width - Lyt.s_Space, this.Height - f_GridView.Top);
                f_GridView.LoadColumnAdjustments();
                int l_BtnLeft = (f_GridView.Right + Lyt.s_Space);
                f_AddButton.Location = new Point(l_BtnLeft, f_GridView.Top);
                f_EditButton.Location = new Point(l_BtnLeft, f_AddButton.GetBottom());
                f_RemoveButton.Location = new Point(l_BtnLeft, f_EditButton.GetBottom());
                f_MarkAsOutputButton.Location = new Point(l_BtnLeft, f_RemoveButton.GetBottom() + Lyt.s_Space);
                f_ImportButton.Location = new Point(l_BtnLeft, this.Height - f_ImportButton.Height);

                f_GridView.e_FillContextMenu += GridView_FillContextMenu;
                f_GridView.e_ItemDoubleClicked += GridView_ItemDoubleClicked;
                f_GridView.e_UpdateItem += GridView_UpdateItem;
                f_GridView.SelectionChanged += GridView_SelectionChanged;
                f_AddButton.Click += AddButton_Click;
                f_EditButton.Click += EditButton_Click;
                f_RemoveButton.Click += RemoveButton_Click;
                f_MarkAsOutputButton.Click += MarkAsOutputButton_Click;
                f_ImportButton.Click += ImportButton_Click;

                f_AddButton.Text = f_TextSet.GetAsDialogOpener(TextKeys.OptionsDlgSharedDirsAdd);
                f_EditButton.Text = f_TextSet.GetAsDialogOpener(TextKeys.OptionsDlgSharedDirsModify);
                f_RemoveButton.Text = f_TextSet[TextKeys.OptionsDlgSharedDirsRemove];
                f_MarkAsOutputButton.Text = f_TextSet[TextKeys.OptionsDlgSharedDirsSetOutputDir];
                f_ImportButton.Text = f_TextSet[TextKeys.OptionsDlgSharedDirsImport];
            }

            protected override void DisposeAutoPanelChildren()
            {
                f_GridView.StoreColumnAdjustments();

                base.DisposeSubClassCreatedChildren();
                f_BoldFont.Dispose();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                if (false == a_Default)
                {
                    int l_ComplTrgtDirIdx;
                    SharedDirectories.Dir[] l_Dirs = a_Prefs.f_IOPrefs.f_SharedDirs.Expose(out l_ComplTrgtDirIdx);
                    SetGui(l_Dirs, l_ComplTrgtDirIdx);
                }
            }

            public bool TryCreateSharedDirs(out SharedDirectories a_SharedDirs,
                                            out string a_ErrMsg)
            {
                int l_ComplTrgtDirIdx;
                SharedDirectories.Dir[] l_Dirs = LoadFromGui(out l_ComplTrgtDirIdx);

                if (0 > l_ComplTrgtDirIdx)
                {
                    a_SharedDirs = null;
                    a_ErrMsg = f_TextSet[TextKeys.OptionsDlgSharedDirsMissingOutputDir];

                    return false;
                }

                a_SharedDirs = new SharedDirectories(l_Dirs, l_ComplTrgtDirIdx);
                a_ErrMsg = null;

                return true;
            }

            private void SetGui(SharedDirectories.Dir[] a_Dirs,
                                int a_ComplTrgtDirIdx)
            {
                f_Dirs.Clear();
                f_ComplTrgtDir = null;

                for (int i = (a_Dirs.Length - 1); 0 <= i; i--)
                {
                    RefDir l_RefDir = new RefDir();
                    l_RefDir.f_Val = a_Dirs[i];

                    f_Dirs.Add(l_RefDir, l_RefDir);

                    if (a_ComplTrgtDirIdx == i)
                    {
                        f_ComplTrgtDir = l_RefDir;
                    }
                }

                UpdateGridView();
            }

            private SharedDirectories.Dir[] LoadFromGui(out int a_ComplTrgtDirIdx)
            {
                a_ComplTrgtDirIdx = SharedDirectories.c_NoTrgtDirIdx;
                SharedDirectories.Dir[] l_Dirs = new SharedDirectories.Dir[f_Dirs.Count];

                int l_Idx = 0;
                foreach (RefDir l_Dir in f_Dirs.Values)
                {
                    if (l_Dir == f_ComplTrgtDir)
                    {
                        a_ComplTrgtDirIdx = l_Idx;
                    }

                    l_Dirs[l_Idx++] = l_Dir.f_Val;
                }

                return l_Dirs;
            }

            private void UpdateGridView()
            {
                f_GridView.StartChanging();

                foreach (RefDir l_Dir in f_Dirs.Values)
                {
                    GridViewItem<RefDir, bool> l_Item = f_GridView.CreateItem(l_Dir);
                }

                f_GridView.EndChanging();

                f_MarkAsOutputButton.Font = (null == f_ComplTrgtDir ? f_BoldFont : f_AddButton.Font);
                f_MarkAsOutputButton.BackColor = (null == f_ComplTrgtDir ? Color.FromArgb(100, Color.Orange) : Color.Transparent);
            }

            private bool GridView_FillContextMenu(GridView<RefDir, bool> a_Sender,
                                                  ContextMenuStrip a_Menu,
                                                  bool a_ItemHit)
            {
                if (false == a_ItemHit ||
                    1 != f_GridView.p_SelectedCount)
                {
                    return false;
                }

                RefDir l_RefDir = f_GridView.GetSelectedItem(0).f_Ref;

                ToolStripItem l_Item = MainForm.AddMenuItem(a_Menu, f_TextSet[TextKeys.DirectoryOpen], MenuOpenFolder_Click, IconTypes.Box);
                l_Item.Tag = l_RefDir;
                MainForm.AddMenuSeparator(a_Menu);

                l_Item = MainForm.AddMenuItem(a_Menu, f_TextSet[TextKeys.OptionsDlgSharedDirsCntxtMenuSetOutputDir], MenuSetOutputDir_Click, IconTypes.Star);
                l_Item.Enabled = !IsCompletionTargetDir(l_RefDir);
                l_Item.Tag = l_RefDir;

                return true;
            }

            private void GridView_ItemDoubleClicked(GridViewItem<SharedDirsPanel.RefDir, bool> a_Item)
            {
                string l_ClickedFolder = a_Item.f_Ref.f_Val.f_Path;
                bool l_OK = f_Inst.f_MainForm.NavigateToFolder(l_ClickedFolder);
            }

            private void GridView_UpdateItem(Time a_Time,
                                             GridViewItem<RefDir, bool> a_Item)
            {
                RefDir l_RefDir = a_Item.f_Ref;
                bool l_FriendsOnly = l_RefDir.f_Val.p_FriendsOnly;

                a_Item.SetTextWithCompObj(0, l_RefDir.f_Val.f_Path, l_RefDir.f_Val.f_Path);
                a_Item.SetBool(1, l_RefDir.f_Val.f_Recursive);
                a_Item.SetBool(2, l_FriendsOnly);

                a_Item.DefaultCellStyle = (a_Item.f_Ref == f_ComplTrgtDir ? f_Inst.f_MainForm.f_BoldCellStyle : f_GridView.DefaultCellStyle);
            }

            private void GridView_SelectionChanged(object sender, EventArgs e)
            {
                UpdateAllButtons();
            }

            private void UpdateAllButtons()
            {
                int l_SelCount = f_GridView.p_SelectedCount;

                f_EditButton.Enabled = (1 == l_SelCount);
                f_RemoveButton.Enabled = (0 < l_SelCount);

                UpdateMarkAsOutputButton();
            }

            private void UpdateMarkAsOutputButton()
            {
                f_MarkAsOutputButton.Enabled = (1 == f_GridView.p_SelectedCount &&
                                                false == IsCompletionTargetDir(f_GridView.GetSelectedItem(0).f_Ref));
            }

            private bool IsCompatible(RefDir a_NewDir,
                                      RefDir a_IgnoreDir,
                                      out string a_ErrMsg)
            {
                // a_IgnoreDir can be an existing item that will not
                // be tested for conflicts with a_TestDir.

                string l_NewDirName = a_NewDir.f_Val.f_Path;

                if (false == DirectoryControl.CheckPath(f_Inst, l_NewDirName, false, out a_ErrMsg))
                {
                    return false;
                }

                foreach (RefDir l_Dir in f_Dirs.Values)
                {
                    if (l_Dir == a_IgnoreDir)
                    {
                        continue;
                    }

                    if (true == SharedDirectories.Dir.DoOverlap(ref a_NewDir.f_Val, ref l_Dir.f_Val))
                    {
                        a_ErrMsg = f_TextSet.Format(TextKeys.OptionsDlgSharedDirsConflictOfSharedDirsFormat, l_NewDirName);

                        return false;
                    }
                }

                a_ErrMsg = null;

                return true;
            }

            private void AddButton_Click(object sender, EventArgs e)
            {
                if (SharedDirectories.c_MaxSharedDirs <= f_Dirs.Count)
                {
                    f_Inst.f_MainForm.WarnMsgBox(f_TextSet[TextKeys.OptionsDlgSharedDirsMaxNumberOfDirsReached]);

                    return;
                }

                using (PickDirDialog l_Dlg = new PickDirDialog(this))
                {
                    l_Dlg.ShowAddDialog();
                }
            }

            private void EditButton_Click(object sender, EventArgs e)
            {
                if (1 != f_GridView.p_SelectedCount)
                {
                    f_Inst.f_MainForm.WarnMsgBox(f_TextSet[TextKeys.OptionsDlgSharedDirsSelectOneItemForModification]);

                    return;
                }

                GridViewItem<RefDir, bool> l_Item = f_GridView.GetSelectedItem(0);

                using (PickDirDialog l_Dlg = new PickDirDialog(this))
                {
                    l_Dlg.ShowEditDialog(l_Item.f_Ref);
                }
            }

            private void RemoveButton_Click(object sender, EventArgs e)
            {
                if (0 == f_GridView.p_SelectedCount)
                {
                    f_Inst.f_MainForm.WarnMsgBox(f_TextSet[TextKeys.OptionsDlgSharedDirsNothingSelectedForRemoval]);

                    return;
                }

                if (false == f_Inst.f_MainForm.YesNoMsgBox(f_TextSet[TextKeys.OptionsDlgSharedDirsReallyRemove]))
                {
                    return;
                }

                foreach (GridViewItem<RefDir, bool> l_Item in f_GridView.SelectedRows)
                {
                    bool l_OK = f_Dirs.Remove(l_Item.f_Ref);

                    if (l_Item.f_Ref == f_ComplTrgtDir)
                    {
                        f_ComplTrgtDir = null;
                    }
                }

                UpdateGridView();
            }

            private void MarkAsOutputButton_Click(object sender, EventArgs e)
            {
                if (1 != f_GridView.p_SelectedCount)
                {
                    return;
                }

                SetCompletionTargetDir(f_GridView.GetSelectedItem(0).f_Ref);
            }

            private void MenuSetOutputDir_Click(object sender, EventArgs e)
            {
                RefDir l_RefDir = (RefDir)((ToolStripItem)sender).Tag;
                SetCompletionTargetDir(l_RefDir);
            }

            private void MenuOpenFolder_Click(object sender, EventArgs e)
            {
                RefDir l_RefDir = (RefDir)((ToolStripItem)sender).Tag;
                f_Inst.f_MainForm.NavigateToFolder(l_RefDir.f_Val.f_Path);
            }

            private bool IsCompletionTargetDir(RefDir a_RefDir)
            {
                return (a_RefDir == f_ComplTrgtDir);
            }

            private void SetCompletionTargetDir(RefDir a_RefDir)
            {
                if (a_RefDir != f_ComplTrgtDir)
                {
                    f_ComplTrgtDir = a_RefDir;
                    UpdateMarkAsOutputButton();
                    UpdateGridView();
                }
            }

            private void ImportButton_Click(object sender, EventArgs e)
            {
                TextSet l_TextSet = f_Inst.f_TextSet;
                MainForm l_Form = f_Inst.f_MainForm;
                string a_Folder;

                if (false == l_Form.ShowPickiMuleFolderDialog(false, out a_Folder))
                {
                    return;
                }

                bool l_Exists;
                string l_FullFileName = SharedDirsImporter.BuildFullConfigFileName(a_Folder, out l_Exists);

                if (false == l_Exists)
                {
                    // From iMule version 2.3 on the config file is in
                    // a different location.

                    string l_CnfgDir;
                    if (LoadMetFilesContext.ErrorCodes.Success == LoadMetFilesContext.FindiMuleConfigDir(a_Folder, out l_CnfgDir))
                    {
                        string l_Candidate = SharedDirsImporter.BuildFullConfigFileName(l_CnfgDir, out l_Exists);

                        if (true == l_Exists)
                        {
                            l_FullFileName = l_Candidate;
                        }
                    }

                    if (false == l_Exists)
                    {
                        l_Form.ErrMsgBox(l_TextSet.Format(TextKeys.UnableToDetectiMuleCnfgMsg, a_Folder));

                        return;
                    }
                }

                if (false == l_Form.ShowDontRuniMuleNowMsg())
                {
                    return;
                }

                int l_Junk;
                SharedDirectories.Dir[] l_Dirs = LoadFromGui(out l_Junk);

                SharedDirsImporter l_Importer = new SharedDirsImporter(l_Dirs, l_FullFileName);
                l_Importer.StartAsync();

                if (false == WaitMessageDialog.ShowDialog(f_TextSet, f_OptionsDlg, l_Importer, WaitMessageDialog.ProgressBarStyles.TrueProgressBar, TextKeys.ImportSharedDirsDlgTitle, TextKeys.ImportSharedDirsDlgMsg))
                {
                    return;
                }

                SharedDirsImporter.ResultCode l_ResCode = l_Importer.GetResult(out l_Dirs);
                string l_ErrMsg;

                switch (l_ResCode)
                {
                    case SharedDirsImporter.ResultCode.Success:
                    {
                        int l_ComplTrgtDirIdx;

                        if (null == f_ComplTrgtDir)
                        {
                            l_ComplTrgtDirIdx = SharedDirectories.c_NoTrgtDirIdx;
                        }
                        else
                        {
                            l_ComplTrgtDirIdx = SharedDirectories.FindCompletionTargetPath(l_Dirs, f_ComplTrgtDir.f_Val.f_Path);
                        }

                        SetGui(l_Dirs, l_ComplTrgtDirIdx);
                        l_ErrMsg = null;

                        f_Inst.f_MainForm.InfoMsgBox("Import was successful.");
                    }
                    break;

                    case SharedDirsImporter.ResultCode.AbortedByUser:
                        l_ErrMsg = f_TextSet[TextKeys.MsgOperationAbortedByUser];
                        break;

                    case SharedDirsImporter.ResultCode.SrcFileTooLarge:
                        l_ErrMsg = "Import source file too large.";
                        break;

                    case SharedDirsImporter.ResultCode.IOError:
                        l_ErrMsg = "An io error occurred while trying to read from the source file.";
                        break;

                    case SharedDirsImporter.ResultCode.EmptyInputFile:
                        l_ErrMsg = "Import source file is empty.";
                        break;

                    default:
                        Util.DebugBreak();
                        l_ErrMsg = f_TextSet[TextKeys.MsgOperationFailed];
                        break;
                }

                if (null != l_ErrMsg)
                {
                    f_Inst.f_MainForm.ErrMsgBox("Import failed. " + l_ErrMsg);
                }
            }

            private class RefDir
            {
                public SharedDirectories.Dir f_Val;
            }

            sealed private class PickDirDialog : ButtonDialog
            {
                private const ushort c_DlgWidth = 450;

                private readonly SharedDirsPanel f_Parent;

                private readonly GuiGroupBox f_GroupBox;
                private readonly DirectoryControl f_DirCntrl;
                private readonly GuiCheckBox f_RecursiveCheckBox;
                private readonly GuiCheckBox f_FriendsOnlyCheckBox;

                private RefDir f_EditedDir;

                public PickDirDialog(SharedDirsPanel a_SharedDirsPanel) :
                                base(a_SharedDirsPanel.f_TextSet, DialogButtons.OKCancel)
                {
                    TextSet l_TextSet = a_SharedDirsPanel.f_TextSet;

                    f_Parent = a_SharedDirsPanel;

                    this.SuspendLayout();

                    f_GroupBox = new GuiGroupBox(this);
                    int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
                    int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
                    int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
                    f_DirCntrl = new DirectoryControl(a_SharedDirsPanel.f_Inst, f_GroupBox, l_InnerWidth, TextKeys.OptionsDlgSharedDirsPickDlgPickADir);
                    f_RecursiveCheckBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);
                    f_FriendsOnlyCheckBox = new GuiCheckBox(f_GroupBox, l_InnerWidth);

                    f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                    f_DirCntrl.Location = new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top);
                    f_RecursiveCheckBox.Location = new Point(Lyt.s_Space, f_DirCntrl.GetBottom());
                    f_FriendsOnlyCheckBox.Location = new Point(Lyt.s_Space, f_RecursiveCheckBox.GetBottom());
                    f_GroupBox.Size = new Size(l_GroupBoxWidth, f_FriendsOnlyCheckBox.Bottom + f_GroupBox.Padding.Bottom);

                    AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
                    base.SetSelectOnShowControl(f_DirCntrl);

                    f_DirCntrl.SetMaxPathChars(SharedDirectories.Dir.c_MaxPathChars);
                    f_OKButton.Click += OKButton_Click;

                    this.Text = l_TextSet[TextKeys.OptionsDlgSharedDirsPickDlgTitle];
                    f_GroupBox.Text = l_TextSet[TextKeys.OptionsDlgSharedDirsPickDlgSharedDir];
                    f_DirCntrl.SetCaption(TextKeys.OptionsDlgSharedDirsPickDlgPath);
                    f_RecursiveCheckBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedDirsPickDlgIncludeSubDirs]);
                    f_FriendsOnlyCheckBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedDirsPickDlgFriendsOnly]);

                    this.ResumeLayout();
                }

                protected override void Dispose(bool disposing)
                {
                    if (true == disposing)
                    {
                        Dialog.DisposeAllChildren(f_GroupBox);
                        f_GroupBox.Dispose();
                    }

                    base.Dispose(disposing);
                }

                private RefDir LoadFromGui()
                {
                    // First make sure we use a standard naming
                    // convention for paths.

                    string l_DirName = f_DirCntrl.GetPath(true);
                    ShareKinds l_ShareKind = (false == f_FriendsOnlyCheckBox.Checked ? SharedDirectories.c_DefaultShareKind : (SharedDirectories.c_DefaultShareKind | ShareKinds.FriendsOnly));

                    RefDir l_Dir = new RefDir();
                    l_Dir.f_Val = new SharedDirectories.Dir(l_DirName, f_RecursiveCheckBox.Checked, l_ShareKind);

                    return l_Dir;
                }

                private void SetGui(RefDir a_Dir)
                {
                    f_DirCntrl.SetPath(a_Dir.f_Val.f_Path);
                    f_RecursiveCheckBox.Checked = a_Dir.f_Val.f_Recursive;
                    f_FriendsOnlyCheckBox.Checked = a_Dir.f_Val.p_FriendsOnly;
                }

                private void OKButton_Click(object sender, EventArgs e)
                {
                    RefDir l_NewDir = LoadFromGui();

                    string l_ErrMsg;
                    if (false == f_Parent.IsCompatible(l_NewDir, f_EditedDir, out l_ErrMsg))
                    {
                        f_Parent.f_Inst.f_MainForm.ErrMsgBox(l_ErrMsg);

                        return;
                    }

                    if (null != f_EditedDir)
                    {
                        if (true == l_NewDir.f_Val.ValueEqualsMindCase(ref f_EditedDir.f_Val))
                        {
                            // Nothing was modified.

                            this.Close();

                            return;
                        }

                        bool l_OK = f_Parent.f_Dirs.Remove(f_EditedDir);

                        if (f_EditedDir == f_Parent.f_ComplTrgtDir)
                        {
                            // Update the completion dir.

                            f_Parent.f_ComplTrgtDir = l_NewDir;
                        }
                    }

                    f_Parent.f_Dirs.Add(l_NewDir, l_NewDir);
                    f_Parent.UpdateGridView();

                    this.Close();
                }

                public void ShowAddDialog()
                {
                    f_EditedDir = null;

                    ShowDialogAndSelectTextBox();
                }

                public void ShowEditDialog(RefDir a_Dir)
                {
                    f_EditedDir = a_Dir;
                    SetGui(f_EditedDir);

                    ShowDialogAndSelectTextBox();
                }

                private void ShowDialogAndSelectTextBox()
                {
                    f_DirCntrl.SetFocus();
                    this.ShowDialog(f_Parent.f_OptionsDlg);
                }
            }
        }

        private class SharedFilesPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 420;

            private readonly Instance f_Inst;

            private readonly GuiGroupBox f_AccessGroupBox;
            private readonly GuiCheckBox f_WatchFsBox;
            private readonly GuiCheckBox f_OpenFilesAsMoveableBox;
            private readonly GuiLabel f_HashingHDLoadLabel;
            private readonly TrackBar f_HashingHDLoadThresholdBar;

            private readonly GuiGroupBox f_FileAttrbsGroupBox;
            private readonly GuiCheckBox f_InclHiddenBox;
            private readonly GuiCheckBox f_InclSystemBox;

            public bool p_WatchFileSystem
            {
                get { return f_WatchFsBox.Checked; }
            }

            public bool p_OpenFilesAsMoveable
            {
                get { return f_OpenFilesAsMoveableBox.Checked; }
            }

            public byte p_HashingHDLoadThresholdPercentage
            {
                get { return (byte)f_HashingHDLoadThresholdBar.Value; }
            }

            public SharedFilesPanel(OptionsDialog a_OptionsDlg,
                                    ref Rectangle a_Rect) :
                               base(a_OptionsDlg, ref a_Rect)
            {
                f_Inst = a_OptionsDlg.f_Inst;

                int l_Width = Lyt.CalcX(c_SuggestedWidth);
                int l_LongWidth = (a_Rect.Width - base.f_ScrollBarWidth - Lyt.s_SmallSpace);
                f_AccessGroupBox = new GuiGroupBox(this, false);
                f_WatchFsBox = new GuiCheckBox(f_AccessGroupBox, l_LongWidth - Lyt.s_Space2);
                f_OpenFilesAsMoveableBox = new GuiCheckBox(f_AccessGroupBox, l_LongWidth - Lyt.s_Space2);
                f_HashingHDLoadLabel = new GuiLabel(f_AccessGroupBox);
                f_HashingHDLoadThresholdBar = new TrackBar();
                f_AccessGroupBox.Controls.Add(f_HashingHDLoadThresholdBar);

                f_FileAttrbsGroupBox = new GuiGroupBox(this);
                f_InclHiddenBox = new GuiCheckBox(f_FileAttrbsGroupBox, l_Width);
                f_InclSystemBox = new GuiCheckBox(f_FileAttrbsGroupBox, l_Width);

                int l_Left = f_AccessGroupBox.Padding.Left;
                f_WatchFsBox.Location = new Point(l_Left, f_AccessGroupBox.Padding.Top);
                f_OpenFilesAsMoveableBox.Location = new Point(l_Left, f_WatchFsBox.GetBottom());
                f_HashingHDLoadLabel.Location = new Point(l_Left, f_OpenFilesAsMoveableBox.GetBottom());
                f_HashingHDLoadThresholdBar.Location = new Point(l_Left, f_HashingHDLoadLabel.GetBottom());
                f_HashingHDLoadThresholdBar.Width = f_HashingHDLoadLabel.Width = (l_LongWidth - Lyt.s_Space2);
                f_HashingHDLoadThresholdBar.SetRange(1, 100);
                f_HashingHDLoadThresholdBar.LargeChange = 20;
                f_HashingHDLoadThresholdBar.TickStyle = TickStyle.Both;
                f_HashingHDLoadThresholdBar.BackColor = GuiGroupBox.s_SuggestedTrackBarBgColor;
                f_AccessGroupBox.SetSize(l_LongWidth, f_HashingHDLoadThresholdBar.Bottom + Lyt.s_TinySpace);

                f_FileAttrbsGroupBox.Top = f_AccessGroupBox.GetBottom();
                f_InclHiddenBox.Location = new Point(l_Left, f_FileAttrbsGroupBox.Padding.Top);
                f_InclSystemBox.Location = new Point(l_Left, f_InclHiddenBox.GetBottom());
                f_FileAttrbsGroupBox.SetSizePlusBottomPadding(l_Width, f_InclSystemBox.Bottom);
                this.ChildBoundsUpdated();

                f_HashingHDLoadThresholdBar.ValueChanged += HashingHDLoadThresholdBar_ValueChanged;

                TextSet l_TextSet = f_Inst.f_TextSet;
                f_AccessGroupBox.Text = l_TextSet[TextKeys.OptionsDlgSharedFilesAccessGroupBox];
                f_WatchFsBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedDirsWatchDirs]);
                f_OpenFilesAsMoveableBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedDirsOpenFilesWithSharedAccess]);
                f_FileAttrbsGroupBox.Text = l_TextSet[TextKeys.OptionsDlgSharedFilesAttribsGroupBox];
                f_InclHiddenBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedFilesAttribsInclHidden]);
                f_InclSystemBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgSharedFilesAttribsInclSystem]);

                HashingHDLoadThresholdBar_ValueChanged(null, null);
            }

            protected override void DisposeAutoPanelChildren()
            {
                MainForm.DisposeAllChildren(f_AccessGroupBox);
                MainForm.DisposeAllChildren(f_FileAttrbsGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_WatchFsBox.Checked = a_Prefs.f_IOPrefs.f_WatchFileSystem;
                f_OpenFilesAsMoveableBox.Checked = a_Prefs.f_IOPrefs.f_OpenFilesAsMoveable;
                f_HashingHDLoadThresholdBar.Value = a_Prefs.f_IOPrefs.f_HashingHDLoadThresholdPercentage;

                System.IO.FileAttributes l_InclAttrbs = ~a_Prefs.f_IOPrefs.f_ExclusionFileAttrbs;
                f_InclHiddenBox.Checked = (0 != (l_InclAttrbs & System.IO.FileAttributes.Hidden));
                f_InclSystemBox.Checked = (0 != (l_InclAttrbs & System.IO.FileAttributes.System));
            }

            public System.IO.FileAttributes GetExclFileAttributes()
            {
                System.IO.FileAttributes l_ExclAttrbs = IOPrefs.c_NoFileAttributes;

                if (false == f_InclHiddenBox.Checked)
                {
                    l_ExclAttrbs |= System.IO.FileAttributes.Hidden;
                }

                if (false == f_InclSystemBox.Checked)
                {
                    l_ExclAttrbs |= System.IO.FileAttributes.System;
                }

                return l_ExclAttrbs;
            }

            private void HashingHDLoadThresholdBar_ValueChanged(object sender, EventArgs e)
            {
                string l_PercStr = string.Format(f_Inst.f_TextSet.f_FormatCulture, "Hard Drive Load Threshold (Hashing): {0:P0}", f_HashingHDLoadThresholdBar.Value / 100f);
                f_HashingHDLoadLabel.Text = l_PercStr;
            }
        }

        private class AdvancedPanel : OptionsPanel
        {
            private const ushort c_SuggestedWidth = 400;

            private readonly SamControl f_SamCntrl;

            private readonly GuiGroupBox f_MiscGroupBox;
            private readonly GuiCheckBox f_AvoidComprBox;
            private readonly GuiCheckBox f_TrustedPublishingBox;
            private readonly GuiCheckBox f_PublUnicodeTextsInNBStyleOnlyBox;
            private readonly GuiCheckBox f_DontCompactBox;
            private readonly GuiCheckBox f_ComplFilesDelBox;
            private readonly GuiCheckBox f_ShowPeersPanelAdvSubMenuBox;
            private readonly GuiCheckBox f_RunOrPreviewFileOnDblClickBox;
            private readonly GuiCheckBox f_AcceptAdditionalKeyCombinationForExitingBox;
            // NOTE: The poison feature is exclusive to debug builds.
            private readonly GuiCheckBox f_PoisonPeersBox;

            public bool p_AllowDataCompression
            {
                get { return !f_AvoidComprBox.Checked; }
            }

            public bool p_OnlyTrustedUsersMayPublish
            {
                get { return f_TrustedPublishingBox.Checked; }
            }

            public bool p_PublUnicodeTextsInNBStyleOnly
            {
                get { return f_PublUnicodeTextsInNBStyleOnlyBox.Checked; }
            }

            public bool p_CompactOnExit
            {
                get { return !f_DontCompactBox.Checked; }
            }

            public bool p_AllowComplFilesDeletion
            {
                get { return f_ComplFilesDelBox.Checked; }
            }

            public bool p_ShowPeersPanelAdvSubMenu
            {
                get { return f_ShowPeersPanelAdvSubMenuBox.Checked; }
            }

            public bool p_RunFilesOnDblClick
            {
                get { return f_RunOrPreviewFileOnDblClickBox.Checked; }
            }

            public bool p_AcceptAdditionalKeyCombinationForExitingBox
            {
                get { return f_AcceptAdditionalKeyCombinationForExitingBox.Checked; }
            }

            public bool p_PoisonPeers
            {
                get { return f_PoisonPeersBox.Checked; }
            }

            public AdvancedPanel(OptionsDialog a_OptionsDlg,
                                 ref Rectangle a_Rect) :
                            base(a_OptionsDlg, ref a_Rect)
            {
                int l_Width = (a_Rect.Width - base.f_ScrollBarWidth - Lyt.s_SmallSpace);
                f_SamCntrl = new SamControl(f_OptionsDlg.f_Inst.f_TextSet, this, l_Width);
                f_MiscGroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_Width - f_MiscGroupBox.Padding.Horizontal);
                f_AvoidComprBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_TrustedPublishingBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_PublUnicodeTextsInNBStyleOnlyBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_DontCompactBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_ComplFilesDelBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_ShowPeersPanelAdvSubMenuBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_RunOrPreviewFileOnDblClickBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_AcceptAdditionalKeyCombinationForExitingBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);
                f_PoisonPeersBox = new GuiCheckBox(f_MiscGroupBox, l_InnerWidth);

                int l_Left = f_MiscGroupBox.Padding.Left;
                f_MiscGroupBox.Top = f_SamCntrl.GetBottom();
                f_AvoidComprBox.Location = new Point(l_Left, f_MiscGroupBox.Padding.Top);
                f_TrustedPublishingBox.Location = new Point(l_Left, f_AvoidComprBox.GetBottom());
                f_PublUnicodeTextsInNBStyleOnlyBox.Location = new Point(l_Left, f_TrustedPublishingBox.GetBottom());
                f_DontCompactBox.Location = new Point(l_Left, f_PublUnicodeTextsInNBStyleOnlyBox.GetBottom());
                f_ComplFilesDelBox.Location = new Point(l_Left, f_DontCompactBox.GetBottom());
                f_ShowPeersPanelAdvSubMenuBox.Location = new Point(l_Left, f_ComplFilesDelBox.GetBottom());
                f_RunOrPreviewFileOnDblClickBox.Location = new Point(l_Left, f_ShowPeersPanelAdvSubMenuBox.GetBottom());
                f_AcceptAdditionalKeyCombinationForExitingBox.Location = new Point(l_Left, f_RunOrPreviewFileOnDblClickBox.GetBottom());
                f_PoisonPeersBox.Location = new Point(l_Left, f_AcceptAdditionalKeyCombinationForExitingBox.GetBottom());
                f_PoisonPeersBox.Visible = false;
                f_PoisonPeersBox.Enabled = false;

                int l_Bottom = (f_PoisonPeersBox.Top - Lyt.s_Space);

                f_MiscGroupBox.Size = new Size(l_Width, l_Bottom + f_MiscGroupBox.Padding.Bottom);
                this.ChildBoundsUpdated();

                TextSet l_TextSet = f_OptionsDlg.f_Inst.f_TextSet;
                f_MiscGroupBox.Text = l_TextSet[TextKeys.OptionsDlgAdvancedMisc];
                f_AvoidComprBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscAvoidCompression]);
                f_TrustedPublishingBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscRequireHighRepForAcceptingPubRequests]);
                f_PublUnicodeTextsInNBStyleOnlyBox.SetTextAndUpdateSize(l_TextSet.FormatWithAppName(TextKeys.OptionsDlgAdvancedMiscPublUnicodeTextsInNBStyleOnly));
                f_DontCompactBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscDontCompact]);
                f_ComplFilesDelBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscAllowComplFilesDeletion]);
                f_ShowPeersPanelAdvSubMenuBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscShowAdvPeersPanelSubMenu]);
                f_RunOrPreviewFileOnDblClickBox.SetTextAndUpdateSize(l_TextSet[TextKeys.OptionsDlgAdvancedMiscRunFilesOnDblClick]);
                // FIXME: Make translatable.
                f_AcceptAdditionalKeyCombinationForExitingBox.SetTextAndUpdateSize("CTRL+Q on main window triggers application shutdown");
            }

            protected override void DisposeAutoPanelChildren()
            {
                Dialog.DisposeAllChildren(f_MiscGroupBox);
                base.DisposeSubClassCreatedChildren();
            }

            protected override void SetCore(Preferences a_Prefs,
                                            bool a_Default)
            {
                f_SamCntrl.Set(a_Prefs.f_GeneralPrefs.f_SamEndPoint);
                f_AvoidComprBox.Checked = !a_Prefs.f_NetworkPrefs.f_AllowDataCompression;
                f_TrustedPublishingBox.Checked = a_Prefs.f_KadPrefs.f_PublAcceptOnlyTrustedUsers;
                f_PublUnicodeTextsInNBStyleOnlyBox.Checked = a_Prefs.f_KadPrefs.f_PublUnicodeTextsInNBStyle;
                f_DontCompactBox.Checked = !a_Prefs.f_GeneralPrefs.f_CompactOnExit;
                f_ComplFilesDelBox.Checked = a_Prefs.f_GeneralPrefs.f_AllowComplFilesDeletion;
                f_ShowPeersPanelAdvSubMenuBox.Checked = a_Prefs.f_GeneralPrefs.f_ShowPeersPanelAdvSubMenu;
                f_RunOrPreviewFileOnDblClickBox.Checked = a_Prefs.f_GeneralPrefs.f_RunFilesOnDblClick;
                f_AcceptAdditionalKeyCombinationForExitingBox.Checked = a_Prefs.f_GeneralPrefs.f_AdditionalKeyCombinationForExiting;
            }

            public bool TryGetSamEndPoint(out System.Net.IPEndPoint a_EndPoint,
                                          out string a_ErrMsg)
            {
                a_EndPoint = f_SamCntrl.TryParse(out a_ErrMsg);

                return (null != a_EndPoint);
            }

            public void FocusSamEndPointControl()
            {
                f_SamCntrl.SetFocus();
            }
        }
    }

    class WaitMessageDialog : Dialog
    {
        private const ushort c_DialogWidth = 250;
        private const ushort c_TextHeight = 60;
        private const ushort c_TimerInterval = 50;

        private readonly TextSet f_TextSet;
        private readonly IDialogController f_Controller;
        private readonly ProgressBarStyles f_Style;

        private readonly string f_LabelMessage;
        private readonly GuiLabel f_Label;
        // f_ProgressBar and f_Button can be null.
        private readonly GuiProgressBar f_ProgressBar;
        private readonly GuiButton f_Button;
        private readonly System.Windows.Forms.Timer f_Timer;

        private bool f_MayClose;

        private WaitMessageDialog(TextSet a_TextSet,
                                  IDialogController a_Controller,
                                  ProgressBarStyles a_Style,
                                  TextKeys a_TitleMsgKey,
                                  string a_Msg)
        {
            f_TextSet = a_TextSet;
            f_Controller = a_Controller;
            f_Style = a_Style;

            this.SuspendLayout();

            f_Label = new GuiLabel(this);

            int l_TextWidth = Lyt.CalcX(c_DialogWidth);
            f_Label.TextAlign = ContentAlignment.MiddleCenter;
            f_Label.Size = new Size(l_TextWidth, Lyt.CalcY(c_TextHeight));

            Control l_LowerBottomCntrl;

            if (ProgressBarStyles.NoProgressBar == f_Style)
            {
                l_LowerBottomCntrl = f_Label;
            }
            else
            {
                f_ProgressBar = new GuiProgressBar(this);

                f_ProgressBar.Top = f_Label.Bottom;
                f_ProgressBar.AdjustOuterFrameColor(true);

                if (ProgressBarStyles.DummyProgressBar == f_Style)
                {
                    f_ProgressBar.AdjustDisplayText(false);
                }
                else
                {
                    f_ProgressBar.AdjustDecimalsDisplay(0);
                }

                if (false == f_Controller.IsAbortable())
                {
                    f_ProgressBar.Width = f_Label.Width;

                    l_LowerBottomCntrl = f_ProgressBar;
                }
                else
                {
                    f_Button = new GuiButton(this);
                    f_Button.Top = f_ProgressBar.Top;
                    f_Button.Left = f_ProgressBar.Width = (f_Label.Width - f_Button.Width);
                    int l_CommonHeight = Math.Max(f_Button.Height, f_ProgressBar.Height);
                    f_Button.Height = f_ProgressBar.Height = l_CommonHeight;
                    f_Button.Click += Button_Click;
                    f_Button.Text = (typeof(MainForm.GeneralShutdownContext) == f_Controller.GetType() ? "Exit Now" : f_TextSet[TextKeys.ButtonTextCancel]);
                    this.CancelButton = f_Button;

                    l_LowerBottomCntrl = f_Button;
                }
            }

            base.AsFixedDialog(new Size(l_LowerBottomCntrl.Right, l_LowerBottomCntrl.Bottom));
            this.ControlBox = false;

            f_Timer = new Timer();
            f_Timer.Interval = c_TimerInterval;
            f_Timer.Tick += Timer_Tick;

            this.Text = a_TextSet[a_TitleMsgKey];
            f_Label.Text = f_LabelMessage = a_Msg;

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(this);

                f_Timer.Dispose();
            }

            base.Dispose(disposing);
        }

        private void Timer_Tick(object sender, EventArgs e)
        {
            uint l_Percentage;

            if (false == f_Controller.GetProgress(out l_Percentage))
            {
                f_Timer.Stop();

                f_MayClose = true;
                this.Close();

                return;
            }

            if (ProgressBarStyles.NoProgressBar == f_Style)
            {
                return;
            }

            if (ProgressBarStyles.TrueProgressBar == f_Style)
            {
#if !DEBUG
                l_Percentage = Math.Min(100, l_Percentage);
#endif
            }
            else
            {
                if ((uint.MaxValue / Util.c_BitsPerByte) <= Util.GetRandomUIntFast())
                {
                    // Pause one fourth of the time.

                    return;
                }

                l_Percentage = (f_ProgressBar.GetProgressPercent() + (Util.GetRandomUIntFast() % 10));
                l_Percentage %= 100;
            }

            f_ProgressBar.SetProgressPercent(l_Percentage);
        }

        private void Button_Click(object sender, EventArgs e)
        {
            // We cannot use MainForm here, since it might not exist.
            if (typeof(MainForm.GeneralShutdownContext) == f_Controller.GetType() ||
                DialogResult.Yes == CustomMsgBox.ShowAbortableMsgBox(f_TextSet, this, f_Controller, f_TextSet.f_AppName, f_TextSet[TextKeys.QuestOperationReallyAbort], MessageBoxIcon.Question, MessageBoxButtons.YesNo, DialogResult.Yes))
            {
                f_Controller.Abort();
                f_Button.Enabled = false;
            }
        }

        protected override void OnShown(EventArgs e)
        {
            // Need to call base class implementation.

            f_Timer.Start();

            base.OnShown(e);
        }

        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            // Need to call base class implementation.

            if (false == f_MayClose)
            {
                e.Cancel = true;
            }
            else
            {
                f_Timer.Stop();
            }

            base.OnClosing(e);
        }

        public static bool ShowDialog(TextSet a_TextSet,
                                      IWin32Window a_Owner,
                                      IDialogController a_Controller,
                                      ProgressBarStyles a_Style,
                                      TextKeys a_TitleMsgKey,
                                      TextKeys a_MsgKey)
        {
            // Constructor for pure text key use.

            return ShowDialog(a_TextSet, a_Owner, a_Controller, a_Style, a_TitleMsgKey, a_TextSet[a_MsgKey]);
        }

        public static bool ShowDialog(TextSet a_TextSet,
                                      IWin32Window a_Owner,
                                      IDialogController a_Controller,
                                      ProgressBarStyles a_Style,
                                      TextKeys a_TitleMsgKey,
                                      string a_Msg)
        {
            // See APCCookie.GuiWaitForCompletion() for the meaning
            // of the return value.

            uint l_Junk;
            if (true == a_Controller.GetProgress(out l_Junk))
            {
                using (WaitMessageDialog l_Dlg = new WaitMessageDialog(a_TextSet, a_Controller, a_Style, a_TitleMsgKey, a_Msg))
                {
                    l_Dlg.ShowDialog(a_Owner);
                }
            }

            return Software.sp_MainForm.p_IsRunning;
        }

        public enum ProgressBarStyles : byte
        {
            NoProgressBar = 0,
            TrueProgressBar = 1,
            DummyProgressBar = 2,
        }
    }

    static class KeyGenerationWaitDialog
    {
        public static bool ShowDialog(TextSet a_TextSet,
                                      IWin32Window a_Owner,
                                      bool a_Abortable,
                                      out OwnUserPersona a_OwnUserPersona)
        {
            // If a_Abortable is true, a_OwnUserPersona can be null on
            // return and must be checked regardless of the return
            // value.

            PersGenContext l_Cntxt = new PersGenContext(a_Abortable);
            l_Cntxt.Start();

            bool l_Running = WaitMessageDialog.ShowDialog(a_TextSet, a_Owner, l_Cntxt, WaitMessageDialog.ProgressBarStyles.DummyProgressBar, TextKeys.KeyGenDlgTitle, TextKeys.KeyGenDlgMsgPleaseWait);
            a_OwnUserPersona = l_Cntxt.p_NewPersonaVolatile;

            return l_Running;
        }

        private class PersGenContext : ControllerContext
        {
            private const string c_PersonaGenerationThreadName = "PersonaGenerator";

            private OwnUserPersona f_NewPersona;

            public OwnUserPersona p_NewPersonaVolatile
            {
                // The return value can be null.
                get { return Util.VolatileRead(ref f_NewPersona); }
            }

            public PersGenContext(bool a_Abortable) :
                             base(a_Abortable)
            {
            }

            public void Start()
            {
                Util.StartThread(c_PersonaGenerationThreadName, PersonaGenerator);
            }

            public override void Abort()
            {
                // Since we have no real way of aborting the
                // operation, we will just let the thread run
                // without caring any further on abortion.
                base.SetFinishedBase();
            }

            private void PersonaGenerator()
            {
                // The following line can take several minutes to
                // execute.
                f_NewPersona = OwnUserPersona.CreateNew();

                base.SetFinishedBase();
            }
        }
    }

    class FileSearchPanel : AdvancedTabPage
    {
        private const ushort c_MaxFilterStrChars = 128;

        private const ushort c_MaxWidth = 850;
        private const ushort c_FileSizeComboWidth = 170;

        private const uint c_CacheMagic = 0xB93D0B6E;
        private const uint c_CacheHeaderSize = 512;

        private const Keys c_ActionShortcutKey = Keys.Enter;

        private static string s_SourcesFormat;

        private readonly Instance f_Inst;
        private readonly SearchCache f_Cache;

        private readonly HorSplitContainer f_Splitter;

        private readonly GuiPanel f_UpperPanel;
        private readonly GuiGroupBox f_RemoteGroupBox;
        private readonly LabeledComboBox f_AndComboBox;
        private readonly LabeledComboBox f_OrComboBox;
        private readonly LabeledComboBox f_NotComboBox;
        private readonly LabeledComboBox f_FileTypeComboBox;
        private readonly LabeledComboBox f_ExtensionComboBox;
        private readonly SearchSizePicker f_MinSizePicker;
        private readonly SearchSizePicker f_MaxSizePicker;

        private readonly GuiGroupBox f_FilterGroupBox;
        private readonly LabeledComboBox f_FilterComboBox;
        private readonly GuiButton f_ApplyFilterBtn;
        private readonly GuiButton f_ApplyInvFilterBtn;
        private readonly GuiCheckBox f_HideKnownFilesBox;

        private readonly GuiButton f_ExpandFileSizeBtn;
        private readonly GuiButton f_ExpandFilterBtn;
        private readonly GuiButton f_ResetBtn;
        private readonly GuiButton f_StopAllBtn;
        private readonly GuiButton f_StartBtn;
        private readonly GuiLabel f_KadFilterInfoLabel;

        private readonly ContextMenuStripEx f_ResetCntxtMenu;

        private readonly AdvancedTabControl f_TabCntrl;
        private readonly ContextMenuStripEx f_TabCntxtMenu;

        private readonly int f_MinWidth;
        private readonly int f_MaxWidth;
        private readonly int f_FileTypeComboWidth;
        private int f_CurrentWidth;
        private bool f_ShowFileSizeOptions;
        private bool f_ShowFilterOptions;

        private SearchCache.SearchCacheItem f_SelectedItem;

        public FileSearchPanel(Instance a_Inst)
        {
            f_Inst = a_Inst;

            f_Splitter = HorSplitContainer.CreateFixed(this);
            f_MinWidth = (Lyt.s_SmallSpace2 + (4 * Lyt.s_SmallSpace) + (2 * Lyt.s_GenerousButtonWidth) + (3 * Lyt.s_DefaultButtonWidth));
            f_MaxWidth = Lyt.CalcX(c_MaxWidth);
            f_FileTypeComboWidth = Lyt.CalcX(c_FileSizeComboWidth);

            int l_DummyWidth = 100;
            f_UpperPanel = new GuiPanel(f_Splitter);
            f_RemoteGroupBox = new GuiGroupBox(f_UpperPanel);
            f_AndComboBox = new LabeledComboBox(f_RemoteGroupBox, l_DummyWidth, true);
            f_OrComboBox = new LabeledComboBox(f_RemoteGroupBox, l_DummyWidth, true);
            f_NotComboBox = new LabeledComboBox(f_RemoteGroupBox, l_DummyWidth, true);
            f_FileTypeComboBox = new LabeledComboBox(f_RemoteGroupBox, f_FileTypeComboWidth, false);
            f_ExtensionComboBox = new LabeledComboBox(f_RemoteGroupBox, l_DummyWidth, true);
            f_MinSizePicker = new SearchSizePicker(f_RemoteGroupBox, int.MaxValue);
            f_MaxSizePicker = new SearchSizePicker(f_RemoteGroupBox, int.MaxValue);

            f_FilterGroupBox = new GuiGroupBox(f_UpperPanel);
            f_FilterComboBox = new LabeledComboBox(f_FilterGroupBox, l_DummyWidth, true);
            f_HideKnownFilesBox = new GuiCheckBox(f_FilterGroupBox, int.MaxValue);
            f_ApplyFilterBtn = new GuiButton(f_FilterGroupBox);
            f_ApplyInvFilterBtn = new GuiButton(f_FilterGroupBox);

            f_ExpandFileSizeBtn = new GuiButton(f_UpperPanel);
            f_ExpandFilterBtn = new GuiButton(f_UpperPanel);
            f_ResetBtn = new GuiButton(f_UpperPanel);
            f_StopAllBtn = new GuiButton(f_UpperPanel);
            f_StartBtn = new GuiButton(f_UpperPanel);
            f_KadFilterInfoLabel = new GuiLabel(f_UpperPanel);

            f_ResetCntxtMenu = new ContextMenuStripEx(f_ResetBtn, false);

            f_TabCntrl = new AdvancedTabControl(f_Splitter);
            f_TabCntxtMenu = new ContextMenuStripEx(f_TabCntrl, false);

            f_Cache = new SearchCache(f_Inst, f_TabCntrl);
            f_Splitter.SetControls(f_UpperPanel, f_TabCntrl);

            int l_Left = f_RemoteGroupBox.Padding.Left;
            f_RemoteGroupBox.Location = new Point(Lyt.s_SmallSpace, Lyt.s_SmallSpace);
            int l_Top = f_RemoteGroupBox.Padding.Top;
            f_AndComboBox.SetLocation(l_Left, l_Top);
            f_OrComboBox.SetTop(l_Top);
            l_Top = f_AndComboBox.GetBottom();
            f_NotComboBox.SetLocation(l_Left, l_Top);
            f_FileTypeComboBox.SetTop(l_Top);
            f_ExtensionComboBox.SetTop(l_Top);
            l_Top = f_NotComboBox.GetBottom();
            f_MinSizePicker.SetLocation(l_Left, l_Top);
            f_MaxSizePicker.SetTop(l_Top);

            f_FilterGroupBox.Left = Lyt.s_SmallSpace;
            l_Top = f_FilterGroupBox.Padding.Top;
            f_FilterComboBox.SetLocation(l_Left, l_Top);
            l_Top = (f_FilterComboBox.f_Box.Bottom - f_ApplyFilterBtn.Height);
            f_ApplyFilterBtn.Top = l_Top;
            f_ApplyInvFilterBtn.Top = l_Top;
            l_Top = (f_FilterComboBox.f_Box.Bottom - f_HideKnownFilesBox.Height);
            f_HideKnownFilesBox.Top = l_Top;
            f_FilterGroupBox.Height = (f_FilterComboBox.f_Box.Bottom + f_FilterGroupBox.Padding.Bottom);

            f_ExpandFileSizeBtn.Left = f_RemoteGroupBox.Left;
            f_ExpandFileSizeBtn.Width = Lyt.s_GenerousButtonWidth;
            f_ExpandFilterBtn.Width = Lyt.s_GenerousButtonWidth;
            f_KadFilterInfoLabel.Top = f_ExpandFilterBtn.GetBottom();
            f_KadFilterInfoLabel.Width = int.MaxValue;
            f_KadFilterInfoLabel.Visible = false;

            f_AndComboBox.f_Box.MaxLength = GuiSearchTerms.c_MaxKeywordStrChars;
            f_OrComboBox.f_Box.MaxLength = GuiSearchTerms.c_MaxKeywordStrChars;
            f_NotComboBox.f_Box.MaxLength = GuiSearchTerms.c_MaxKeywordStrChars;
            f_ExtensionComboBox.f_Box.MaxLength = GuiSearchTerms.c_MaxKeywordStrChars;
            f_FilterComboBox.f_Box.MaxLength = c_MaxFilterStrChars;

            TextSet l_TextSet = f_Inst.f_TextSet;
            MainForm.AddMenuItem(f_ResetCntxtMenu, l_TextSet[TextKeys.FileSearchTabLocalResetForm], ResetCntxtMenu_ResetFormClick, IconTypes.Sheet);
            MainForm.AddMenuItem(f_ResetCntxtMenu, l_TextSet[TextKeys.FileSearchTabLocalDelKeywordHist], ResetCntxtMenu_DeleteKeywordHistoryClick, IconTypes.TrashBin);

            f_TabCntrl.SetDefaultPadding(false);
            f_TabCntrl.p_AllowClosingThroughUI = true;
            f_TabCntrl.p_AllowReorderingThroughUI = true;

            // When changing the order of items, also update the code
            // in CntxtMenu_Opening().
            f_TabCntxtMenu.Opening += CntxtMenu_Opening;
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuRestoreSearchParams], CntxtMenu_RestoreParams, IconTypes.SheetPencil);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuClearSearchResults], CntxtMenu_ClearSearchResults, IconTypes.Sheet);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuRerunSearch], CntxtMenu_StartSearch, IconTypes.Lightning);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuStopSearch], CntxtMenu_StopSearch, IconTypes.Forbidden);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuCloseSearch], CntxtMenu_CloseSearch, IconTypes.TrashBin);
            MainForm.AddMenuSeparator(f_TabCntxtMenu);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuStopAllSearches], CntxtMenu_StopAllSearches);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuReorderSearches], CntxtMenu_ReorderSearches, IconTypes.File);
            MainForm.AddMenuItem(f_TabCntxtMenu, l_TextSet[TextKeys.FileSearchTabCntxtMenuCloseAllSearches], CntxtMenu_CloseAllSearches);

            f_Splitter.Resize += Splitter_Resize;
            f_AndComboBox.f_Box.KeyDown += SearchTextBox_KeyDown;
            f_OrComboBox.f_Box.KeyDown += SearchTextBox_KeyDown;
            f_NotComboBox.f_Box.KeyDown += SearchTextBox_KeyDown;
            f_ExtensionComboBox.f_Box.KeyDown += SearchTextBox_KeyDown;
            f_MinSizePicker.f_NumBox.KeyDown += SearchTextBox_KeyDown;
            f_MaxSizePicker.f_NumBox.KeyDown += SearchTextBox_KeyDown;
            f_FilterComboBox.f_Box.KeyDown += FilterTextBox_KeyDown;
            f_ApplyFilterBtn.Click += ApplyFilterBtn_Click;
            f_ApplyInvFilterBtn.Click += ApplyInvFilterBtn_Click;
            f_HideKnownFilesBox.Click += HideKnownFilesBox_Click;
            f_ExpandFileSizeBtn.Click += ExpandFileSizeBtn_Click;
            f_ExpandFilterBtn.Click += ExpandFilterBtn_Click;
            f_ResetBtn.Click += ResetBtn_Click;
            f_StopAllBtn.Click += StopAllBtn_Click;
            f_StartBtn.Click += StartBtn_Click;
            f_TabCntrl.e_SelectedPageChanged += TabCntrl_SelectedPageChanged;
            f_TabCntrl.e_ClosingSelPageThroughUI += TabCntrl_ClosingSelPageThroughUI;

            f_RemoteGroupBox.Text = l_TextSet[TextKeys.FileSearchTabRemoteCaption];
            f_AndComboBox.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileNameAND);
            f_OrComboBox.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileNameOR);
            f_NotComboBox.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileNameNOT);
            f_FileTypeComboBox.SetTexts(l_TextSet, TextKeys.FileSearchTabRemoteFileType, FileTypeDetector.GetFileTypesStrings());
            f_ExtensionComboBox.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileExtension);
            f_MinSizePicker.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileMinSize);
            f_MaxSizePicker.SetCaption(l_TextSet, TextKeys.FileSearchTabRemoteFileMaxSize);
            f_FileTypeComboBox.f_Box.SelectedIndex = 0;

            f_FilterGroupBox.Text = l_TextSet[TextKeys.FileSearchTabLocalCaption];
            f_FilterComboBox.SetCaption(l_TextSet, TextKeys.FileSearchTabLocalTextFilter);
            f_HideKnownFilesBox.SetTextAndUpdateSize(l_TextSet[TextKeys.FileSearchTabLocalHideKnown]);

            f_ResetBtn.Text = l_TextSet[TextKeys.FileSearchTabLocalReset];
            f_StopAllBtn.Text = l_TextSet[TextKeys.FileSearchTabLocalStopAll];
            f_StartBtn.Text = l_TextSet[TextKeys.FileSearchTabLocalStart];
            f_KadFilterInfoLabel.Text = "Some results were filtered out during the last run of this search, due to your global Kademlia filter settings.";

            UpdateLayout(true);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_SourcesFormat = a_TextSet[TextKeys.SourcesShort];
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_TabCntrl.e_SelectedPageChanged -= TabCntrl_SelectedPageChanged;
                f_SelectedItem = null;

                f_Cache.CloseAllSearches();

                Dialog.DisposeAllChildren(f_RemoteGroupBox);
                Dialog.DisposeAllChildren(f_FilterGroupBox);
                Dialog.DisposeAllChildren(f_UpperPanel);
                f_ResetCntxtMenu.Dispose();

                f_UpperPanel.Dispose();
                f_TabCntrl.Dispose();
                f_TabCntxtMenu.Dispose();
                f_Splitter.Dispose();
            }

            base.Dispose(disposing);
        }

        public void ArrangeLayoutOnShown()
        {
            UpdateAfterShrinkOrExpand();
        }

        public void UpdateImagesOfAllPages()
        {
            f_TabCntrl.UpdateImagesOfAllPages();
        }

        public void SetSizeFormat(bool a_SI)
        {
            f_MinSizePicker.SetSizeFormat(a_SI);
            f_MaxSizePicker.SetSizeFormat(a_SI);
        }

        public void FinishLoading()
        {
            if (null != f_SelectedItem)
            {
                f_SelectedItem.FinishLoading();
            }
        }

        public void PrepareSaving()
        {
            if (null != f_SelectedItem)
            {
                f_SelectedItem.PrepareSaving();
            }
        }

        protected override void OnSelectPage(SelectionChangeOrigins a_Origin)
        {
            f_AndComboBox.SetFocus(false);

            base.OnSelectPage(a_Origin);
        }

        private void TabCntrl_SelectedPageChanged(AdvancedTabControl a_Sender)
        {
            // The assigned reference can be null.
            f_SelectedItem = (SearchCache.SearchCacheItem)f_TabCntrl.p_SelectedPage;
            UpdateKadFilterTextVisibilty();
        }

        private void TabCntrl_ClosingSelPageThroughUI(AdvancedTabControl a_Sender,
                                                      AdvancedTabPage a_Page,
                                                      ref bool a_Handled,
                                                      ref bool a_DisposePage)
        {
            a_Handled = true;

            f_Cache.CloseSearch((int)f_SelectedItem.p_PageIndex, true);
        }

        public void Process(Time a_Time)
        {
            f_Cache.Process(a_Time);
            UpdateKadFilterTextVisibilty();
        }

        public void SetSearchResultsFilter(KademliaFilter a_KadFilter)
        {
            f_Cache.SetSearchResultsFilter(a_KadFilter);
        }

        public StorageStates SaveSearchCache(BufferIO a_Buffer)
        {
            // May only get called through the MainForm class.

            a_Buffer.WriteUInt32LE(c_CacheMagic);
            a_Buffer.SetPosition(c_CacheHeaderSize);
            f_Cache.ToBuffer(a_Buffer);

            return StorageStates.Success;
        }

        public StorageStates LoadSearchCache(BufferIO a_Buffer)
        {
            // May only get called through the MainForm class.

            if (c_CacheHeaderSize > a_Buffer.GetRemaining() ||
                c_CacheMagic != a_Buffer.ReadUInt32LE())
            {
                return StorageStates.Corruption;
            }

            a_Buffer.SetPosition(c_CacheHeaderSize);

            try
            {
                if (false == f_Cache.FromBuffer(a_Buffer))
                {
                    return StorageStates.Corruption;
                }
            }
            catch (BufferIO.InterpretBufferException)
            {
                return StorageStates.Corruption;
            }

            if (false == a_Buffer.EndReached())
            {
                return StorageStates.Corruption;
            }

            ApplyHistory();

            return StorageStates.Success;
        }

        public void ResetSearchCache()
        {
            // Use if a load failed.

            f_Cache.DeleteStringHistory();
            f_Cache.CloseAllSearches();
            ApplyHistory();
            ResetCntxtMenu_ResetFormClick(null, null);
        }

        private void ApplyKeywordHistory()
        {
            f_Cache.ApplyKeywordHistory(f_AndComboBox, f_OrComboBox, f_NotComboBox);
        }

        private void ApplyFilterHistory()
        {
            f_Cache.ApplyFilterHistory(f_FilterComboBox);
        }

        private void ApplyHistory()
        {
            ApplyKeywordHistory();
            ApplyFilterHistory();
        }

        private void ResetFileSizeControls()
        {
            f_MinSizePicker.Reset();
            f_MaxSizePicker.Reset();
        }

        private void Splitter_Resize(object sender, EventArgs e)
        {
            UpdateLayout(false);
        }

        private void UpdateKadFilterTextVisibilty()
        {
            bool l_Vis = (null != f_SelectedItem &&
                          true == f_SelectedItem.p_ResultsKadFilteredDuringLastSearchRun);

            f_KadFilterInfoLabel.Visible = l_Vis;
        }

        private void UpdateAfterShrinkOrExpand()
        {
            TextSet l_TextSet = f_Inst.f_TextSet;

            if (false == f_ShowFileSizeOptions)
            {
                f_ExpandFileSizeBtn.Text = l_TextSet.GetAsExpandButtonCaption(TextKeys.FileSearchTabLocalShowSizeCntrls);
            }
            else
            {
                f_ExpandFileSizeBtn.Text = l_TextSet.GetAsCollapseButtonCaption(TextKeys.FileSearchTabLocalHideSizeCntrls);
            }

            if (false == f_ShowFilterOptions)
            {
                f_ExpandFilterBtn.Text = l_TextSet.GetAsExpandButtonCaption(TextKeys.FileSearchTabLocalShowFilterCntrls);
                SetFilter(StrFilterMode.Off);
            }
            else
            {
                f_ExpandFilterBtn.Text = l_TextSet.GetAsCollapseButtonCaption(TextKeys.FileSearchTabLocalHideFilterCntrls);
            }

            f_MinSizePicker.SetVisibilty(f_ShowFileSizeOptions);
            f_MaxSizePicker.SetVisibilty(f_ShowFileSizeOptions);
            f_FilterGroupBox.Visible = f_ShowFilterOptions;

            UpdateLayout(true);
        }

        private void UpdateLayout(bool a_Force)
        {
            int l_NewWidth = f_UpperPanel.Width;

            if (l_NewWidth != f_CurrentWidth)
            {
                f_CurrentWidth = l_NewWidth;
            }
            else if (false == a_Force)
            {
                // Nothing to do.

                return;
            }

            l_NewWidth = Math.Max(Math.Min(f_CurrentWidth, f_MaxWidth), f_MinWidth);

            f_Splitter.SuspendLayout();

            int l_GroupBoxWidth = (l_NewWidth - Lyt.s_SmallSpace2);
            int l_HalfWidth = ((l_GroupBoxWidth - f_RemoteGroupBox.Padding.Horizontal - Lyt.s_Space) / 2);
            f_AndComboBox.SetWidth(l_HalfWidth);
            f_OrComboBox.SetWidth(l_HalfWidth);
            f_NotComboBox.SetWidth(l_HalfWidth);
            f_ExtensionComboBox.SetWidth(l_HalfWidth - f_FileTypeComboWidth - Lyt.s_Space);
            f_MinSizePicker.SetWidth(l_HalfWidth);
            f_MaxSizePicker.SetWidth(l_HalfWidth);

            int l_Mid = (f_RemoteGroupBox.Padding.Left + l_HalfWidth + Lyt.s_Space);
            f_OrComboBox.SetLeft(l_Mid);
            f_FileTypeComboBox.SetLeft(l_Mid);
            f_ExtensionComboBox.SetLeft(f_FileTypeComboBox.GetRight());
            f_MaxSizePicker.SetLeft(l_Mid);
            int l_RemoteGroupBoxInnerHeight = (false == f_ShowFileSizeOptions ? f_ExtensionComboBox.f_Box.Bottom : f_MaxSizePicker.f_NumBox.Bottom);
            f_RemoteGroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, l_RemoteGroupBoxInnerHeight);

            int l_ButtonTop;

            if (false == f_ShowFilterOptions)
            {
                l_ButtonTop = f_RemoteGroupBox.GetBottom();
            }
            else
            {
                f_FilterComboBox.SetWidth(l_HalfWidth);
                f_ApplyFilterBtn.Left = l_Mid;
                f_ApplyInvFilterBtn.Left = f_ApplyFilterBtn.GetRight();
                f_HideKnownFilesBox.Left = f_ApplyInvFilterBtn.GetRight();
                f_FilterGroupBox.Width = l_GroupBoxWidth;
                f_FilterGroupBox.Top = f_RemoteGroupBox.GetBottom();

                l_ButtonTop = f_FilterGroupBox.GetBottom();
            }

            f_ExpandFileSizeBtn.Top = l_ButtonTop;
            f_ExpandFilterBtn.Location = new Point(f_ExpandFileSizeBtn.GetRight(), l_ButtonTop);
            f_ResetBtn.Location = new Point(f_RemoteGroupBox.Right - (3 * Lyt.s_DefaultButtonWidth) - (2 * Lyt.s_SmallSpace), l_ButtonTop);
            f_StopAllBtn.Location = new Point(f_ResetBtn.GetRight(), l_ButtonTop);
            f_StartBtn.Location = new Point(f_StopAllBtn.GetRight(), l_ButtonTop);
            f_KadFilterInfoLabel.Location = new Point(Lyt.s_SmallSpace, f_StartBtn.GetBottom());

            f_Splitter.SetUpperHeight(f_KadFilterInfoLabel.GetBottom());
            f_Splitter.ResumeLayout();
        }

        private void CntxtMenu_Opening(object sender, System.ComponentModel.CancelEventArgs e)
        {
            uint l_Count = f_TabCntrl.p_PageCount;
            int l_Idx = f_TabCntxtMenu.p_ClickedTabPageIndex;
            f_TabCntrl.SetSelectedPageIndex((uint)l_Idx, SelectionChangeOrigins.UserInteraction);
            bool l_Runs = f_Cache.IsSearchRunning(l_Idx);

            // Clear results.
            f_TabCntxtMenu.Items[1].Enabled = !l_Runs;
            // Run this search.
            f_TabCntxtMenu.Items[2].Enabled = !l_Runs;
            // Stop this search.
            f_TabCntxtMenu.Items[3].Enabled = l_Runs;
            // Reorder.
            f_TabCntxtMenu.Items[7].Enabled = (1 < l_Count);
        }

        private void CntxtMenu_RestoreParams(object sender, EventArgs e)
        {
            GuiSearchTerms l_SearchTerms = f_Cache.GetSearchTerms(f_TabCntxtMenu.p_ClickedTabPageIndex);

            f_AndComboBox.f_Box.Text = GuiSearchTerms.Join(l_SearchTerms.f_AndKeywords);
            f_OrComboBox.f_Box.Text = GuiSearchTerms.Join(l_SearchTerms.f_OrKeywords);
            f_NotComboBox.f_Box.Text = GuiSearchTerms.Join(l_SearchTerms.f_NotKeywords);
            f_FileTypeComboBox.f_Box.SelectedIndex = (int)l_SearchTerms.f_FileType;
            f_ExtensionComboBox.f_Box.Text = l_SearchTerms.f_Extension;

            if (true == f_ShowFileSizeOptions)
            {
                ResetFileSizeControls();
            }

            if (true == l_SearchTerms.HasFileSizeConstraints())
            {
                if (false == f_ShowFileSizeOptions)
                {
                    // Show the file size options, since they belong to
                    // the used search terms.

                    f_ShowFileSizeOptions = true;

                    UpdateAfterShrinkOrExpand();
                }

                if (0 != l_SearchTerms.f_MinFileSize)
                {
                    f_MinSizePicker.SetSize(l_SearchTerms.f_MinFileSize);
                }

                if (ulong.MaxValue != l_SearchTerms.f_MaxFileSize)
                {
                    f_MaxSizePicker.SetSize(l_SearchTerms.f_MaxFileSize);
                }
            }

            f_AndComboBox.SetFocus(false);
        }

        private void CntxtMenu_ClearSearchResults(object sender, EventArgs e)
        {
            f_Cache.ClearSearchResults(f_TabCntxtMenu.p_ClickedTabPageIndex);
        }

        private void CntxtMenu_StartSearch(object sender, EventArgs e)
        {
            if (true == f_Cache.StartSearch(f_TabCntxtMenu.p_ClickedTabPageIndex))
            {
                ApplyKeywordHistory();
            }
        }

        private void CntxtMenu_StopSearch(object sender, EventArgs e)
        {
            f_Cache.StopSearch(f_TabCntxtMenu.p_ClickedTabPageIndex);
        }

        private void CntxtMenu_CloseSearch(object sender, EventArgs e)
        {
            f_Cache.CloseSearch(f_TabCntxtMenu.p_ClickedTabPageIndex, false);
        }

        private void CntxtMenu_StopAllSearches(object sender, EventArgs e)
        {
            f_Cache.StopAllSearches();
        }

        private void CntxtMenu_ReorderSearches(object sender, EventArgs e)
        {
            uint l_Count = f_TabCntrl.p_PageCount;

            if (1 >= l_Count)
            {
                return;
            }

            // Load the search cache items into our array.

            ReorderItemsDialog<SearchCache.SearchCacheItem>.Item[] l_Arr = new ReorderItemsDialog<SearchCache.SearchCacheItem>.Item[l_Count];

            int i = 0;

            foreach (SearchCache.SearchCacheItem l_CacheItem in f_TabCntrl)
            {
                l_Arr[i] = new ReorderItemsDialog<SearchCache.SearchCacheItem>.Item(l_CacheItem.f_BaseCaption, l_CacheItem, true, i);
                i++;
            }

            // Show dialog.

            using (ReorderItemsDialog<SearchCache.SearchCacheItem> l_Dlg = new ReorderItemsDialog<SearchCache.SearchCacheItem>(f_Inst, TextKeys.FileSearchTabCntxtMenuReorderSearches))
            {
                bool l_OrderChanged;
                if (false == l_Dlg.ShowDialog(f_Inst.f_MainForm, 0, ModifyItemVisibilityMode.Disabled, l_Arr, out l_OrderChanged))
                {
                    return;
                }
            }

            // Evaluate results.

            ReorderItemsDialog<SearchCache.SearchCacheItem>.Item.SortArrayByDisplayOrderIndex(l_Arr);

            foreach (ReorderItemsDialog<SearchCache.SearchCacheItem>.Item l_ReorderItem in l_Arr)
            {
                f_TabCntrl.ChangePageIndex(l_ReorderItem.f_Tag, (uint)l_ReorderItem.p_DisplayOrderIndex, SelectionChangeOrigins.UserInteraction);
            }
        }

        private void CntxtMenu_CloseAllSearches(object sender, EventArgs e)
        {
            if (false == f_Cache.HasAnySearch() ||
                false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.FileSearchTabReallyCloseAllSearches]))
            {
                return;
            }

            f_Cache.CloseAllSearches();
        }

        private GuiSearchTerms FormToSearchTerms(out string a_ErrMsg)
        {
            // Return can be null, in which case l_ErrMsg will be set.

            FileType l_FileType = (FileType)f_FileTypeComboBox.f_Box.SelectedIndex;
            ulong l_MinFileSize;
            ulong l_MaxFileSize;

            if (false == f_MinSizePicker.GetSize(out l_MinFileSize))
            {
                l_MinFileSize = 0;
            }

            if (false == f_MaxSizePicker.GetSize(out l_MaxFileSize))
            {
                l_MaxFileSize = ulong.MaxValue;
            }

            return GuiSearchTerms.Create(f_Inst.f_TextSet, f_AndComboBox.f_Box.Text, f_OrComboBox.f_Box.Text, f_NotComboBox.f_Box.Text, f_ExtensionComboBox.f_Box.Text, l_FileType, l_MinFileSize, l_MaxFileSize, out a_ErrMsg);
        }

        private void ExpandFileSizeBtn_Click(object sender, EventArgs e)
        {
            f_ShowFileSizeOptions = !f_ShowFileSizeOptions;

            if (false == f_ShowFileSizeOptions)
            {
                ResetFileSizeControls();
            }

            UpdateAfterShrinkOrExpand();
        }

        private void ExpandFilterBtn_Click(object sender, EventArgs e)
        {
            f_ShowFilterOptions = !f_ShowFilterOptions;

            UpdateAfterShrinkOrExpand();
        }

        private void FilterTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (c_ActionShortcutKey == e.KeyCode)
            {
                SetFilter(StrFilterMode.On);
                f_ApplyFilterBtn.Select();

                e.Handled = true;
            }
        }

        private void ApplyFilterBtn_Click(object sender, EventArgs e)
        {
            StrFilterMode l_Mode = (StrFilterMode.On == f_Cache.p_StrFilterMode ? StrFilterMode.Off : StrFilterMode.On);
            SetFilter(l_Mode);
        }

        private void ApplyInvFilterBtn_Click(object sender, EventArgs e)
        {
            StrFilterMode l_Mode = (StrFilterMode.InvertedOn == f_Cache.p_StrFilterMode ? StrFilterMode.Off : StrFilterMode.InvertedOn);
            SetFilter(l_Mode);
        }

        private void HideKnownFilesBox_Click(object sender, EventArgs e)
        {
            bool l_HideKnownFiles = f_HideKnownFilesBox.Checked;
            f_Cache.AdjustFilteringOfKnownFiles(l_HideKnownFiles);
        }

        private void SetFilter(StrFilterMode a_Mode)
        {
            string l_FilterStr = f_FilterComboBox.f_Box.Text.Trim();
            bool l_HideKnownFiles = f_HideKnownFilesBox.Checked;

            string l_ApplyStr = f_Inst[TextKeys.FileSearchTabLocalApply];
            string l_ApplyInvStr = f_Inst[TextKeys.FileSearchTabLocalApplyInverted];

            if (StrFilterMode.Off != a_Mode)
            {
                if (0 == l_FilterStr.Length)
                {
                    // No valid string for string filtering.

                    a_Mode = StrFilterMode.Off;
                }
                else
                {
                    string l_ResetStr = f_Inst[TextKeys.FileSearchTabLocalEndFiltering];

                    if (StrFilterMode.On == a_Mode)
                    {
                        l_ApplyStr = l_ResetStr;
                    }
                    else
                    {
                        l_ApplyInvStr = l_ResetStr;
                    }
                }
            }

            f_FilterComboBox.f_Box.Enabled = (StrFilterMode.Off == a_Mode);
            f_ApplyFilterBtn.Text = l_ApplyStr;
            f_ApplyInvFilterBtn.Text = l_ApplyInvStr;

            f_Cache.SetFilter(a_Mode, l_FilterStr, l_HideKnownFiles);

            if (StrFilterMode.Off != a_Mode)
            {
                ApplyFilterHistory();
            }
        }

        private void ResetBtn_Click(object sender, EventArgs e)
        {
            Point l_Pos = f_ResetBtn.PointToClient(Form.MousePosition);
            f_ResetCntxtMenu.Show(f_ResetBtn, l_Pos);
        }

        private void ResetCntxtMenu_ResetFormClick(object sender, EventArgs e)
        {
            f_AndComboBox.ClearText();
            f_OrComboBox.ClearText();
            f_NotComboBox.ClearText();
            f_FileTypeComboBox.f_Box.SelectedIndex = 0;
            f_ExtensionComboBox.ClearText();
            ResetFileSizeControls();

            f_FilterComboBox.ClearText();
            f_HideKnownFilesBox.Checked = false;
        }

        private void ResetCntxtMenu_DeleteKeywordHistoryClick(object sender, EventArgs e)
        {
            if (false == f_Cache.HasStringHistory() ||
                false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.FileSearchTabReallyDeleteKeywordHist]))
            {
                return;
            }

            f_Cache.DeleteStringHistory();
            ApplyHistory();
        }

        private void StopAllBtn_Click(object sender, EventArgs e)
        {
            f_Cache.StopAllSearches();
        }

        private void SearchTextBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (c_ActionShortcutKey == e.KeyCode)
            {
                StartSearch();

                e.Handled = true;
            }
        }

        private void StartBtn_Click(object sender, EventArgs e)
        {
            StartSearch();
        }

        private void StartSearch()
        {
            if (false == f_Cache.CheckIfStartIsPossibleIfNotPrintMsg())
            {
                return;
            }

            string l_ErrMsg;
            GuiSearchTerms l_SearchTerms = FormToSearchTerms(out l_ErrMsg);

            if (null == l_SearchTerms ||
                false == f_Cache.AddSearch(l_SearchTerms, true, out l_ErrMsg))
            {
                f_Inst.f_MainForm.WarnMsgBox(l_ErrMsg);
            }

            ApplyKeywordHistory();
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct FoundFile
        {
            public const byte c_MinTempBufferSize = (24);

            private const ushort c_FoundFileMagic = 0x1C62;

            private static SipHash.Key s_SipHashKey;

            public readonly UInt128Value f_FileID;
            // f_FileName is guaranteed to be legal.
            public readonly string f_FileName;
            public readonly ulong f_FileSize;
            public readonly FileType f_FileType;

            public readonly uint f_HashValue;

            public ushort f_CompleteSources;
            public ushort f_Sources;

            // Both times can be not set. f_FirstAppearanceTime will
            // not be greater than f_LastAppearanceTime.
            public StorableTime f_FirstAppearanceTime;
            public StorableTime f_LastAppearanceTime;

            public ComplStatus f_ComplStatus;

            public uint p_PartialSources
            {
                get { return ((uint)f_Sources - f_CompleteSources); }
            }

            static FoundFile()
            {
                SipHash.Key.CreateRandom(out s_SipHashKey);
            }

            public FoundFile(ref UInt128Value a_FileID,
                             string a_FileName,
                             ulong a_FileSize,
                             ushort a_CompleteSources,
                             ushort a_Sources,
                             StorableTime a_FirstAppearanceTime,
                             StorableTime a_LastAppearanceTime,
                             byte[] a_TempBuffer)
            {
                f_FileID = a_FileID;
                f_FileName = a_FileName;
                f_FileSize = a_FileSize;
                f_FileType = FileTypeDetector.GuessFileType(f_FileName);

                f_HashValue = CalcHashValue(ref a_FileID, a_FileSize, a_TempBuffer);

                f_CompleteSources = a_CompleteSources;
                f_Sources = a_Sources;

                f_FirstAppearanceTime = a_FirstAppearanceTime;
                f_LastAppearanceTime = a_LastAppearanceTime;

                f_ComplStatus = ComplStatus.Dropped;
            }

            public void ResetSourceCounts()
            {
                f_CompleteSources = 0;
                f_Sources = 0;
            }

            public static uint CalcHashValue(ref UInt128Value a_FileID,
                                             ulong a_FileSize,
                                             byte[] a_TempBuffer)
            {
                Util.WriteUInt64LE(a_FileID.f_ULong0, a_TempBuffer, 0);
                Util.WriteUInt64LE(a_FileID.f_ULong1, a_TempBuffer, sizeof(ulong));
                Util.WriteUInt64LE(a_FileSize, a_TempBuffer, 2 * sizeof(ulong));

                return (uint)SipHash.CalculateHashCode(ref s_SipHashKey, a_TempBuffer, 0, c_MinTempBufferSize);
            }

            public bool SameFile(ref FoundFile a_Other)
            {
                return (this.f_FileID.f_ULong0 == a_Other.f_FileID.f_ULong0 &&
                        this.f_FileID.f_ULong1 == a_Other.f_FileID.f_ULong1 &&
                        this.f_FileSize == a_Other.f_FileSize);
            }

            public void TakeOverSourceCountAndLastAppearanceTime(ref FoundFile a_Src)
            {
                this.f_CompleteSources = a_Src.f_CompleteSources;
                this.f_Sources = a_Src.f_Sources;

                if (false == this.f_FirstAppearanceTime.IsSet() ||
                    this.f_FirstAppearanceTime > a_Src.f_LastAppearanceTime)
                {
                    this.f_FirstAppearanceTime = a_Src.f_LastAppearanceTime;
                }

                this.f_LastAppearanceTime = a_Src.f_LastAppearanceTime;
            }

            public string GetFileTypeStr()
            {
                if (FileType.FT_ANY == f_FileType)
                {
                    return string.Empty;
                }

                return FileTypeDetector.FileTypeToPrettyString(f_FileType);
            }

            public string GetSourcesStr(string a_Format)
            {
                return string.Format(a_Format, f_CompleteSources, (f_Sources - f_CompleteSources));
            }

            public FileLink CreateLink()
            {
                return new FileLink(f_FileName, f_FileSize, new FileID(f_FileID.ToRef()));
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                a_Buffer.WriteUInt32LE(c_FoundFileMagic);
                f_FileID.ToBufferLE(a_Buffer);
                a_Buffer.WriteUTF16LEStringEncLen(f_FileName);
                a_Buffer.WriteUInt64LE(f_FileSize);
                a_Buffer.WriteUInt16LE(f_CompleteSources);
                a_Buffer.WriteUInt16LE(f_Sources);
                a_Buffer.WriteUInt32LE(f_FirstAppearanceTime.GetAsUInt32());
                a_Buffer.WriteUInt32LE(f_LastAppearanceTime.GetAsUInt32());
            }

            public static bool FromBuffer(BufferIO a_Buffer,
                                          bool a_NewFormat,
                                          byte[] a_TempBuffer,
                                          out FoundFile a_Dest)
            {
                uint l_ExtraSize;

                if (false == a_NewFormat)
                {
                    l_ExtraSize = 0;
                }
                else
                {
                    uint l_Val = a_Buffer.ReadUInt32LE();
                    ushort l_Magic = (ushort)l_Val;

                    if (c_FoundFileMagic != l_Magic)
                    {
                        a_Dest = new FoundFile();

                        return false;
                    }

                    l_ExtraSize = (l_Val >> Util.c_BitsPerUShort);
                }

                UInt128Value l_FileID;
                UInt128Value.FromBufferLE(a_Buffer, out l_FileID);
                string l_FileName = a_Buffer.ReadUTF16LEStringEncLen();
                ulong l_FileSize = a_Buffer.ReadUInt64LE();
                ushort l_CompletedSources = a_Buffer.ReadUInt16LE();
                ushort l_Sources = a_Buffer.ReadUInt16LE();

                // Check and repair.

                l_FileName = Util.FixUTF8Latin1ConversionErrors(l_FileName);
                l_FileName = FileNames.MakeFileNameLegalFailsafe(l_FileName);

                if (false == ContentManager.IsValidFileSize(l_FileSize))
                {
                    Util.DebugBreak();

                    l_FileSize = 1;
                }

                if (l_Sources < l_CompletedSources)
                {
                    Util.DebugBreak();

                    l_CompletedSources = l_Sources;
                }

                StorableTime l_FirstAppearanceTime;
                StorableTime l_LastAppearanceTime;

                if (false == a_NewFormat)
                {
                    l_FirstAppearanceTime = new StorableTime();
                    l_LastAppearanceTime = new StorableTime();
                }
                else
                {
                    ulong l_Val = a_Buffer.ReadUInt64LE();
                    l_FirstAppearanceTime = StorableTime.FromUInt32((uint)l_Val);
                    l_LastAppearanceTime = StorableTime.FromUInt32((uint)(l_Val >> Util.c_BitsPerUInt));
                }

                if (false == l_FirstAppearanceTime.IsSet())
                {
                    l_FirstAppearanceTime = l_LastAppearanceTime;
                }
                else if (false == l_LastAppearanceTime.IsSet())
                {
                    l_LastAppearanceTime = l_FirstAppearanceTime;
                }
                else
                {
                    l_FirstAppearanceTime = StorableTime.Min(l_FirstAppearanceTime, l_LastAppearanceTime);
                }

                if (0 != l_ExtraSize)
                {
                    a_Buffer.JumpAhead(l_ExtraSize);
                }

                a_Dest = new FoundFile(ref l_FileID, l_FileName, l_FileSize, l_CompletedSources, l_Sources, l_FirstAppearanceTime, l_LastAppearanceTime, a_TempBuffer);

                return true;
            }
        }

        private enum StrFilterMode : byte
        {
            Off = 0,
            On = 1,
            InvertedOn = 2,
        }

        private class SearchCache
        {
            // Used before v1.4.0.
            public const ushort c_ObsoleteMaxSearchCacheItems = 256;
            // Used from v1.4.0 on.
            public const ushort c_MaxSearchCacheItems = 64;

            public readonly Instance f_Inst;
            private readonly DataGridViewCellStyle[] f_FileSearchCellStyles;
            private readonly AdvancedTabControl f_TabCntrl;

            private StringHistory f_AndKeywordHistory;
            private StringHistory f_OrKeywordHistory;
            private StringHistory f_NotKeywordHistory;
            private StringHistory f_FilterKeywordHistory;

            private FilterConfig f_Filter;
            private bool f_ConflictMsgBoxShown;

            public StrFilterMode p_StrFilterMode
            {
                get { return f_Filter.f_StrFilterMode; }
            }

            public SearchCache(Instance a_Inst,
                               AdvancedTabControl a_TabCntrl)
            {
                f_Inst = a_Inst;
                f_FileSearchCellStyles = f_Inst.f_MainForm.f_FileSearchCellStyles;
                f_TabCntrl = a_TabCntrl;

                StringHistory.Create(out f_AndKeywordHistory);
                StringHistory.Create(out f_OrKeywordHistory);
                StringHistory.Create(out f_NotKeywordHistory);
                StringHistory.Create(out f_FilterKeywordHistory);
            }

            public void ToBuffer(BufferIO a_Buffer)
            {
                // First write the format revision byte.
                a_Buffer.WriteByte(1);
                f_AndKeywordHistory.ToBuffer(a_Buffer);
                f_OrKeywordHistory.ToBuffer(a_Buffer);
                f_NotKeywordHistory.ToBuffer(a_Buffer);
                f_FilterKeywordHistory.ToBuffer(a_Buffer);

                a_Buffer.WriteUInt16LE((ushort)f_TabCntrl.p_PageCount);

                foreach (SearchCacheItem l_Item in f_TabCntrl)
                {
                    l_Item.ToBuffer(a_Buffer);
                }
            }

            public bool FromBuffer(BufferIO a_Buffer)
            {
                CloseAllSearches();

                byte l_FormatRevision = a_Buffer.ReadByte();

                if (1 < l_FormatRevision)
                {
                    return false;
                }

                if (false == StringHistory.FromBuffer(a_Buffer, out f_AndKeywordHistory) ||
                    false == StringHistory.FromBuffer(a_Buffer, out f_OrKeywordHistory) ||
                    false == StringHistory.FromBuffer(a_Buffer, out f_NotKeywordHistory) ||
                    false == StringHistory.FromBuffer(a_Buffer, out f_FilterKeywordHistory))
                {
                    return false;
                }

                uint l_Count = a_Buffer.ReadUInt16LE();

                if (c_ObsoleteMaxSearchCacheItems < l_Count)
                {
                    // No version ever allowed that many.

                    return false;
                }

                if (0 != l_Count)
                {
                    bool l_NewFileItemFormat = (1 <= l_FormatRevision);

                    for (uint i = 0; i < l_Count; i++)
                    {
                        SearchCacheItem l_Item;

                        if (false == SearchCacheItem.FromBuffer(this, a_Buffer, l_NewFileItemFormat, out l_Item))
                        {
                            return false;
                        }

                        if (i < c_MaxSearchCacheItems)
                        {
                            f_TabCntrl.AddPageWithoutSelecting(l_Item);
                        }
                        else
                        {
                            // Surplus item, so drop.

                            l_Item.Dispose();
                        }
                    }

                    f_TabCntrl.SetSelectedPageIndex(0, SelectionChangeOrigins.Programmatic);
                }

                return true;
            }

            public bool HasStringHistory()
            {
                return (true == f_AndKeywordHistory.HasAnyString() ||
                        true == f_OrKeywordHistory.HasAnyString() ||
                        true == f_NotKeywordHistory.HasAnyString() ||
                        true == f_FilterKeywordHistory.HasAnyString());
            }

            public void DeleteStringHistory()
            {
                f_AndKeywordHistory.Clear();
                f_OrKeywordHistory.Clear();
                f_NotKeywordHistory.Clear();
                f_FilterKeywordHistory.Clear();
            }

            public void ApplyKeywordHistory(LabeledComboBox a_AndComboBox,
                                            LabeledComboBox a_OrComboBox,
                                            LabeledComboBox a_NotComboBox)
            {
                f_AndKeywordHistory.FillComboBox(a_AndComboBox);
                f_OrKeywordHistory.FillComboBox(a_OrComboBox);
                f_NotKeywordHistory.FillComboBox(a_NotComboBox);
            }

            public void ApplyFilterHistory(LabeledComboBox a_FilterComboBox)
            {
                f_FilterKeywordHistory.FillComboBox(a_FilterComboBox);
            }

            private void AddToHistory(GuiSearchTerms a_SearchTerms)
            {
                f_AndKeywordHistory.Add(GuiSearchTerms.Join(a_SearchTerms.f_AndKeywords));

                if (null != a_SearchTerms.f_OrKeywords)
                {
                    f_OrKeywordHistory.Add(GuiSearchTerms.Join(a_SearchTerms.f_OrKeywords));
                }

                if (null != a_SearchTerms.f_NotKeywords)
                {
                    f_NotKeywordHistory.Add(GuiSearchTerms.Join(a_SearchTerms.f_NotKeywords));
                }
            }

            public bool HasAnySearch()
            {
                return (0 < f_TabCntrl.p_PageCount);
            }

            private SearchCacheItem GetItem(int a_Idx)
            {
                return (SearchCacheItem)f_TabCntrl.GetPageByIndex((uint)a_Idx);
            }

            public void CloseAllSearches()
            {
                if (false == HasAnySearch())
                {
                    return;
                }

                f_TabCntrl.RemoveAllPages(true, SelectionChangeOrigins.Programmatic);
            }

            public DataGridViewCellStyle GetCellStyle(ComplStatus a_Status)
            {
                return f_FileSearchCellStyles[(int)a_Status];
            }

            public bool CheckIfStartIsPossibleIfNotPrintMsg()
            {
                if (KademliaStatus.Connected == f_Inst.f_Kad.p_KadStatusVolatile)
                {
                    return true;
                }

                f_Inst.f_MainForm.WarnMsgBox(f_Inst[TextKeys.FileSearchTabKadConnRequired]);

                return false;
            }

            public bool AddSearch(GuiSearchTerms a_SearchTerms,
                                  bool a_StartSearch,
                                  out string a_ErrMsg)
            {
                SearchCacheItem l_Item = null;

                foreach (SearchCacheItem l_ExistingItem in f_TabCntrl)
                {
                    if (a_SearchTerms.f_HashCode == l_ExistingItem.f_SearchTerms.f_HashCode)
                    {
                        // This very search exists already.

                        l_Item = l_ExistingItem;

                        break;
                    }
                }

                if (null == l_Item)
                {
                    // We need to create a new search item.

                    if (f_TabCntrl.p_PageCount >= c_MaxSearchCacheItems)
                    {
                        a_ErrMsg = f_Inst[TextKeys.FileSearchTabMaxSearchTabCountReached];

                        return false;
                    }

                    l_Item = new SearchCacheItem(this, a_SearchTerms);
                    f_TabCntrl.AddPage(l_Item);
                }

                f_TabCntrl.SetSelectedPage(l_Item, SelectionChangeOrigins.UserInteraction);

                if (true == a_StartSearch)
                {
                    l_Item.StartSearch();
                    AddToHistory(l_Item.f_SearchTerms);
                }

                a_ErrMsg = null;

                return true;
            }

            public GuiSearchTerms GetSearchTerms(int a_Idx)
            {
                return GetItem(a_Idx).f_SearchTerms;
            }

            public bool IsSearchRunning(int a_Idx)
            {
                return GetItem(a_Idx).IsSearchRunning();
            }

            public bool ClearSearchResults(int a_Idx)
            {
                return GetItem(a_Idx).ClearSearchResults();
            }

            public bool StartSearch(int a_Idx)
            {
                if (false == CheckIfStartIsPossibleIfNotPrintMsg())
                {
                    return false;
                }

                SearchCacheItem l_Item = GetItem(a_Idx);

                if (false == l_Item.StartSearch())
                {
                    return false;
                }

                AddToHistory(l_Item.f_SearchTerms);

                return true;
            }

            public void StopSearch(int a_Idx)
            {
                bool l_Stopped = GetItem(a_Idx).StopSearch();
            }

            public void StopAllSearches()
            {
                foreach (SearchCacheItem l_Item in f_TabCntrl)
                {
                    bool l_Stopped = l_Item.StopSearch();
                }
            }

            public void CloseSearch(int a_Idx,
                                    bool a_PromptForConfirmation)
            {
                string l_TabName = GetItem(a_Idx).f_BaseCaption;

                if (false == a_PromptForConfirmation ||
                    true == f_Inst.f_MainForm.YesNoMsgBox("Do you really want to remove the search tab \"" + l_TabName + "\"?"))
                {
                    f_TabCntrl.RemovePageAt((uint)a_Idx, true, SelectionChangeOrigins.UserInteraction);
                }
            }

            public void SetSearchResultsFilter(KademliaFilter a_KadFilter)
            {
                foreach (SearchCacheItem l_Item in f_TabCntrl)
                {
                    l_Item.SetSearchResultsFilter(a_KadFilter);
                }
            }

            public void Process(Time a_Time)
            {
                bool l_Conflict = false;

                foreach (SearchCacheItem l_Item in f_TabCntrl)
                {
                    l_Item.Process(a_Time, ref l_Conflict);
                }

                if (true == l_Conflict &&
                    false == f_ConflictMsgBoxShown)
                {
                    // Show the message box and prevent the GUI thread
                    // from doing this over and over.

                    f_ConflictMsgBoxShown = true;
                    f_Inst.f_MainForm.InfoMsgBox(f_Inst[TextKeys.FileSearchTabSearchStartFailed]);
                    f_ConflictMsgBoxShown = false;
                }
            }

            public void SetFilter(StrFilterMode a_StrFilterMode,
                                  string a_FilterStr,
                                  bool a_HideKnownFiles)
            {
                f_Filter = new FilterConfig(a_StrFilterMode, a_FilterStr, a_HideKnownFiles);

                if (StrFilterMode.Off != a_StrFilterMode)
                {
                    f_FilterKeywordHistory.Add(a_FilterStr);
                }

                NotifyFilterChange();
            }

            public void AdjustFilteringOfKnownFiles(bool a_HideKnownFiles)
            {
                f_Filter = new FilterConfig(f_Filter.f_StrFilterMode, f_Filter.f_FilterStr, a_HideKnownFiles);

                NotifyFilterChange();
            }

            private void NotifyFilterChange()
            {
                foreach (SearchCacheItem l_Item in f_TabCntrl)
                {
                    l_Item.MarkAsDirtyAndUpdateResults();
                }
            }

            public struct StringHistory
            {
                private const byte c_MaxCachedKeywords = 30;

                private FlexArray<string> f_Keywords;

                private StringHistory(int a_Capacity)
                {
                    FlexArray<string>.Create(a_Capacity, out f_Keywords);
                }

                public static void Create(out StringHistory a_Dest)
                {
                    a_Dest = new StringHistory(0);
                }

                public bool HasAnyString()
                {
                    return (0 < f_Keywords.p_Count);
                }

                public void Clear()
                {
                    f_Keywords.Clear();
                }

                public bool Add(string a_NewText)
                {
                    // Returns true if the cache has changed.

                    int l_Count = f_Keywords.p_Count;
                    string[] l_Arr = f_Keywords.p_Arr;
                    for (int i = 0; i < l_Count; i++)
                    {
                        if (false == Util.StringsEqualIgnoreCase(l_Arr[i], a_NewText))
                        {
                            continue;
                        }

                        if (0 == i)
                        {
                            // Nothing to do.

                            return false;
                        }

                        f_Keywords.DeleteAt(i);
                        l_Count--;

                        break;
                    }

                    if (c_MaxCachedKeywords <= l_Count)
                    {
                        f_Keywords.DeleteAt(c_MaxCachedKeywords - 1);
                    }

                    f_Keywords.ReserveAt(0);
                    f_Keywords.p_Arr[0] = a_NewText;

                    return true;
                }

                public void ToBuffer(BufferIO a_Buffer)
                {
                    int l_Count = f_Keywords.p_Count;

                    a_Buffer.WriteByte((byte)l_Count);

                    string[] l_Arr = f_Keywords.p_Arr;
                    for (int i = 0; i < l_Count; i++)
                    {
                        a_Buffer.WriteUTF16LEStringEncLen(l_Arr[i]);
                    }
                }

                public static bool FromBuffer(BufferIO a_Buffer,
                                              out StringHistory a_Dest)
                {
                    int l_PresentCount = a_Buffer.ReadByte();

                    if (c_MaxCachedKeywords < l_PresentCount)
                    {
                        a_Dest = new StringHistory();

                        return false;
                    }

                    a_Dest = new StringHistory(l_PresentCount);
                    a_Dest.f_Keywords.ReserveRange(0, l_PresentCount);

                    string[] l_Arr = a_Dest.f_Keywords.p_Arr;
                    for (int i = 0; i < l_PresentCount; i++)
                    {
                        string l_Text = a_Buffer.ReadUTF16LEStringEncLen();

                        if (GuiSearchTerms.c_MaxKeywordStrChars < l_Text.Length)
                        {
                            return false;
                        }

                        l_Arr[i] = l_Text;
                    }

                    return true;
                }

                public void FillComboBox(LabeledComboBox a_ComboBox)
                {
                    string l_StoredText = a_ComboBox.f_Box.Text;

                    a_ComboBox.SetTexts(f_Keywords.ToArray(false));

                    a_ComboBox.f_Box.Text = l_StoredText;
                }
            }

#if !MONO
            [StructLayout(LayoutKind.Auto)]
#endif
            private struct FilterConfig
            {
                public readonly StrFilterMode f_StrFilterMode;
                public readonly string f_FilterStr;

                public readonly bool f_HideKnownFiles;

                public FilterConfig(StrFilterMode a_StrFilterMode,
                                    string a_FilterStr,
                                    bool a_HideKnownFiles)
                {
                    f_StrFilterMode = a_StrFilterMode;
                    f_FilterStr = a_FilterStr;
                    f_HideKnownFiles = a_HideKnownFiles;
                }

                public bool IsFiltered(string a_FileName)
                {
                    switch (f_StrFilterMode)
                    {
                        case StrFilterMode.On:
                            return (0 > Util.IndexOfIgnoreCase(a_FileName, f_FilterStr));

                        case StrFilterMode.InvertedOn:
                            return (0 <= Util.IndexOfIgnoreCase(a_FileName, f_FilterStr));

                        default:
                            return false;
                    }
                }
            }

            public class SearchCacheItem : AdvancedTabPage
            {
                private const ushort c_ItemMagic = 0xD412;
                private const ushort c_MaxFileCount = 10000;
                private const byte c_ResultsStatusUpdateIntervalSecs = 3;
                private const Keys c_DownloadShortcutKey = Keys.F11;

                public readonly SearchCache f_Cache;
                public readonly GuiSearchTerms f_SearchTerms;
                public readonly string f_BaseCaption;
                private readonly GridView<int, bool> f_GridView;

                private bool f_Entered;
                private string f_CaptionWithCount;

                private FlexArray<FoundFile> f_Files;
                private bool f_ResultsDirty;

                private GuiSearchCookie f_SearchCookie;
                private uint f_CookieResultGeneration;
                private Dictionary<uint, int> f_IndexMap;

                private bool f_NeedStatusUpdate;
                private Time f_BlockStatusUpdateTime;
                private Time f_StatusExpirationTime;
                private GeneralAPCCookie f_ResultsStatusUpdateCookie;

                // f_KadFilter can be null.
                private KademliaFilter f_KadFilter;
                private bool f_ResultsKadFilteredDuringLastSearchRun;

                public bool p_ResultsKadFilteredDuringLastSearchRun
                {
                    get { return f_ResultsKadFilteredDuringLastSearchRun; }
                }

                public SearchCacheItem(SearchCache a_Cache,
                                       GuiSearchTerms a_SearchTerms)
                {
                    f_Cache = a_Cache;
                    f_SearchTerms = a_SearchTerms;
                    f_BaseCaption = f_SearchTerms.GetCaption();
                    f_GridView = new GridView<int, bool>(f_Cache.f_Inst, this, f_Cache.f_Inst.f_MainForm.f_VisualCust.f_FileSearchPanelColSet);

                    FlexArray<FoundFile>.Create(out f_Files);

                    this.p_AllowClosingThroughUI = true;
                    this.p_AllowReorderingThroughUI = true;

                    f_GridView.e_FillContextMenu += FillContextMenu;
                    f_GridView.e_UpdateItem += GridView_UpdateItem;
                    f_GridView.KeyDown += GridView_KeyDown;

                    UpdateCaption();
                }

                protected override void Dispose(bool disposing)
                {
                    if (true == disposing)
                    {
                        bool l_Stopped = StopSearch();

                        f_GridView.Dispose();
                    }

                    base.Dispose(disposing);
                }

                public void ToBuffer(BufferIO a_Buffer)
                {
                    a_Buffer.WriteUInt16LE(c_ItemMagic);
                    f_SearchTerms.ToBuffer(a_Buffer);

                    int l_Count = f_Files.p_Count;

                    a_Buffer.WriteUInt32LE((uint)l_Count);

                    FoundFile[] l_Arr = f_Files.p_Arr;
                    for (int i = 0; i < l_Count; i++)
                    {
                        l_Arr[i].ToBuffer(a_Buffer);
                    }
                }

                public static bool FromBuffer(SearchCache a_Cache,
                                              BufferIO a_Buffer,
                                              bool a_NewFileItemFormat,
                                              out SearchCacheItem a_Item)
                {
                    GuiSearchTerms l_SearchTerms;

                    if (c_ItemMagic != a_Buffer.ReadUInt16LE() ||
                        false == GuiSearchTerms.FromBuffer(a_Buffer, out l_SearchTerms))
                    {
                        a_Item = null;

                        return false;
                    }

                    int l_Count = (int)a_Buffer.ReadUInt32LE();

                    if (c_MaxFileCount < l_Count)
                    {
                        a_Item = null;

                        return false;
                    }

                    a_Item = new SearchCacheItem(a_Cache, l_SearchTerms);

                    // NOTE: Use an instance routine for reading, since
                    // this is a marshal-by-ref class.

                    if (false == a_Item.ReadFilesCore(a_Buffer, a_NewFileItemFormat, l_Count))
                    {
                        a_Item = null;

                        return false;
                    }

                    return true;
                }

                private bool ReadFilesCore(BufferIO a_Buffer,
                                           bool a_NewFormat,
                                           int a_Count)
                {
                    // Even on success this routine may have produced
                    // less than a_Count files.

                    if (0 == a_Count)
                    {
                        return true;
                    }

                    f_Files.ReserveRange(0, a_Count);

                    byte[] l_TempBuffer = new byte[FoundFile.c_MinTempBufferSize];

                    FoundFile[] l_Arr = f_Files.p_Arr;
                    for (int i = 0; i < a_Count; i++)
                    {
                        if (false == FoundFile.FromBuffer(a_Buffer, a_NewFormat, l_TempBuffer, out l_Arr[i]))
                        {
                            f_Files.Clear(false);

                            return false;
                        }
                    }

                    // Check for duplicates.

                    {
                        Dictionary<uint, DummyStruct> l_Dic = new Dictionary<uint, DummyStruct>(a_Count);
                        int l_EffCount = a_Count;

                        for (int i = (l_EffCount - 1); 0 <= i; i--)
                        {
                            try
                            {
                                l_Dic.Add(l_Arr[i].f_HashValue, new DummyStruct());

                                continue;
                            }
                            catch (ArgumentNullException l_Exc)
                            {
                                throw (l_Exc);
                            }
                            catch (ArgumentException)
                            {
                                // Duplicate key.
                            }

                            for (int k = (l_EffCount - 1); i < k; k--)
                            {
                                if (true == l_Arr[k].SameFile(ref l_Arr[i]))
                                {
                                    f_Files.DeleteAt(k);
                                    l_EffCount--;

                                    break;
                                }
                            }
                        }
                    }

                    MarkAsDirty();

                    return true;
                }

                public void FinishLoading()
                {
                    f_GridView.LoadColumnAdjustments();
                }

                public void PrepareSaving()
                {
                    f_GridView.StoreColumnAdjustments();
                }

                protected override void OnSelectPage(SelectionChangeOrigins a_Origin)
                {
                    // Need to call base class implementation.

                    f_Entered = true;
                    UpdateResults();
                    UpdateCaption();

                    if (null == f_ResultsStatusUpdateCookie)
                    {
                        // Refresh the file status.

                        f_NeedStatusUpdate = true;
                    }

                    f_GridView.LoadColumnAdjustments();
                    f_GridView.Select();

                    base.OnSelectPage(a_Origin);
                }

                protected override void OnDeselectPage(SelectionChangeOrigins a_Origin)
                {
                    // Need to call base class implementation.

                    f_Entered = false;
                    UpdateCaption();

                    f_GridView.StoreColumnAdjustments();

                    base.OnDeselectPage(a_Origin);
                }

                private void GridView_KeyDown(object sender, KeyEventArgs e)
                {
                    // This is for forwarding menu shortcuts in case the
                    // corresponding menu item doesn't exist or is disabled.

                    if (c_DownloadShortcutKey == e.KeyCode &&
                        false == Dialog.HasNonShiftModifiers(e.Modifiers))
                    {
                        AddToIncomplFilesUnassigned_Click(null, null);

                        e.Handled = true;
                    }
                }

                private bool FillContextMenu(GridView<int, bool> a_Sender,
                                             ContextMenuStrip a_Menu,
                                             bool a_ItemHit)
                {
                    int l_SelCount = f_GridView.p_SelectedCount;

                    if (0 == l_SelCount)
                    {
                        return false;
                    }

                    // Download.

                    TextSet l_TextSet = f_Cache.f_Inst.f_TextSet;
                    MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FileSearchTabResultsCntxtMenuAddToIncompl], AddToIncomplFilesUnassigned_Click, c_DownloadShortcutKey, IconTypes.Download);
                    f_Cache.f_Inst.f_MainForm.f_Categories.FillMenuWithCustomCategorySubMenu(l_TextSet, a_Menu, TextKeys.FileSearchTabResultsCntxtMenuAddAndAssignToCategory, false, AddToIncomplFilesCustomCategory_Click);
                    MainForm.AddMenuSeparator(a_Menu);

                    // Links.

                    GenericMenuItem<FileLink.CopyLinkToClipboardOps>.AddLinkMenu(l_TextSet, a_Menu, 1 == l_SelCount, CopyLinkToClipboard_Click);

                    // Copy.

                    MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.FileSearchTabResultsCntxtMenuCopySelItemsToClipboardAsText], CopyResultsToClipboard_Click, IconTypes.SheetPencil);

                    return true;
                }

                private void AddToIncomplFilesUnassigned_Click(object sender, EventArgs e)
                {
                    // We may use a keyboard shortcut without extra
                    // caution here, since AddToIncompleteFiles() may
                    // be called in any situation (not just when the
                    // menu closes).

                    AddToIncompleteFiles(Category.s_Unassigned);
                }

                private void AddToIncomplFilesCustomCategory_Click(object sender, EventArgs e)
                {
                    GenericMenuItem<Category> l_Item = (GenericMenuItem<Category>)sender;
                    Category l_Category = l_Item.f_GenericTag;

                    AddToIncompleteFiles(l_Category);
                }

                private void AddToIncompleteFiles(Category a_TargetCategory)
                {
                    // a_TargetCategory may be Category.s_Unassigned
                    // (of course).

                    int l_Count = f_GridView.SelectedRows.Count;

                    if (0 == l_Count)
                    {
                        return;
                    }

                    FoundFile[] l_TrgtFiles = new FoundFile[l_Count];

                    FoundFile[] l_SrcArr = f_Files.p_Arr;
                    for (int i = 0; i < l_Count; i++)
                    {
                        int l_SrcIdx = f_GridView.GetSelectedItem(i).f_Ref;
                        l_TrgtFiles[i] = l_SrcArr[l_SrcIdx];
                    }

                    Instance l_Inst = f_Cache.f_Inst;
                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(l_Inst.f_ContentMngr.GuiAddFoundFiles, new object[] { l_TrgtFiles, a_TargetCategory });
                    if (false == l_Inst.f_ContentMngr.EnqueueAPCAndWaitAsGUI(l_Cookie))
                    {
                        return;
                    }

                    if (CookieStates.Success != l_Cookie.p_State)
                    {
                        // Should never happen.

                        return;
                    }

                    uint l_Done = (uint)l_Cookie.GetReturnObject();
                    string l_Str;

                    if (1 == l_Count)
                    {
                        l_Str = l_Inst.f_TextSet[1 == l_Done ? TextKeys.FileSearchTabFileWasTakenOver : TextKeys.FileSearchTabFileWasNotTakenOver];
                    }
                    else
                    {
                        l_Str = l_Inst.f_TextSet.Format(TextKeys.FileSearchTabNFilesWereTakenOver, l_Done, l_Count);
                    }

                    if (0 < l_Done)
                    {
                        // We've at least changed the status of one file.

                        f_NeedStatusUpdate = true;
                    }

                    l_Inst.f_MainForm.InfoMsgBox(l_Str);
                }

                private void CopyResultsToClipboard_Click(object sender, EventArgs e)
                {
                    string l_Str = f_GridView.CopyToString();
                    Util.SetClipboard(l_Str);
                }

                private void CopyLinkToClipboard_Click(object sender, EventArgs e)
                {
                    if (1 == f_GridView.p_SelectedCount)
                    {
                        int l_Idx = f_GridView.GetSelectedItem(0).f_Ref;
                        FileLink l_Link = f_Files.p_Arr[l_Idx].CreateLink();

                        GenericMenuItem<FileLink.CopyLinkToClipboardOps> l_Item = (GenericMenuItem<FileLink.CopyLinkToClipboardOps>)sender;
                        FileLink.CopyLinkToClipboardOps l_Op = l_Item.f_GenericTag;

                        l_Link.CopyLinkToClipboard(l_Op);
                    }
                }

                private void UpdateCaption()
                {
                    string l_Caption;

                    if (false == f_Entered ||
                        null == f_CaptionWithCount)
                    {
                        l_Caption = f_BaseCaption;
                    }
                    else
                    {
                        l_Caption = f_CaptionWithCount;
                    }

                    base.p_Caption = l_Caption;
                    base.SetImage(false == IsSearchRunning() ? IconTypes.NoIcon : IconTypes.Lightning);
                }

                private void SetCaptionWithCount(string a_CaptionWithCount)
                {
                    // a_CaptionWithCount may be null.

                    f_CaptionWithCount = a_CaptionWithCount;

                    UpdateCaption();
                }

                private void NotifySearchStarted()
                {
                    UpdateCaption();
                }

                private void NotifySearchEnded()
                {
                    if (true == f_SearchCookie.f_KadFiltered &&
                        false == f_ResultsKadFilteredDuringLastSearchRun)
                    {
                        f_ResultsKadFilteredDuringLastSearchRun = true;
                    }

                    f_SearchCookie.Aborted();
                    f_SearchCookie = null;
                    f_CookieResultGeneration = 0;
                    f_IndexMap = null;

                    UpdateCaption();
                }

                public bool ClearSearchResults()
                {
                    // We will only do this while the search isn't
                    // running.

                    if (null == f_SearchCookie)
                    {
                        if (0 < f_Files.p_Count)
                        {
                            f_Files.Clear(false);
                            MarkAsDirtyAndUpdateResults();
                        }

                        return true;
                    }

                    return false;
                }

                public bool StartSearch()
                {
                    // Returns false if the search already runs.

                    if (null != f_SearchCookie)
                    {
                        return false;
                    }

                    f_SearchCookie = new GuiSearchCookie(f_SearchTerms);

                    GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Cache.f_Inst.f_Kad.f_SearchMngr.GuiStartKeywordSearch, f_SearchCookie);
                    f_Cache.f_Inst.f_NetworkMngr.EnqueueAPC(l_Cookie);

                    NotifySearchStarted();

                    f_ResultsKadFilteredDuringLastSearchRun = false;

                    return true;
                }

                public bool StopSearch()
                {
                    if (null != f_SearchCookie)
                    {
                        NotifySearchEnded();

                        return true;
                    }

                    return false;
                }

                public bool IsSearchRunning()
                {
                    if (null != f_SearchCookie)
                    {
                        CookieStates l_Status = f_SearchCookie.p_NetworkStateVolatile;

                        return (CookieStates.Pending == l_Status ||
                                CookieStates.SearchRunning == l_Status);
                    }

                    return false;
                }

                public void Process(Time a_Time,
                                    ref bool a_Conflict)
                {
                    HandleSearchResults(ref a_Conflict);
                    HandleResultsStatusUpdate(a_Time);

                    UpdateResults();
                }

                public void SetSearchResultsFilter(KademliaFilter a_KadFilter)
                {
                    f_KadFilter = a_KadFilter;
                }

                private void HandleKadFilterOfResults()
                {
                    if (null == f_KadFilter ||
                        null != f_ResultsStatusUpdateCookie)
                    {
                        return;
                    }

                    KademliaFilter l_Filter = f_KadFilter;

                    f_KadFilter = null;
                    int l_Count = f_Files.p_Count;

                    if (0 < l_Count)
                    {
                        bool l_Dirty = false;
                        FoundFile[] l_Arr = f_Files.p_Arr;

                        for (int i = (l_Count - 1); 0 <= i; i--)
                        {
                            if (false == l_Filter.RunFileNameThroughFilter(l_Arr[i].f_FileName, KademliaFilterScopes.ScopeSearchResultsForSelf))
                            {
                                continue;
                            }

                            f_Files.DeleteAt(i);

                            l_Arr = f_Files.p_Arr;
                            l_Dirty = true;
                        }

                        if (true == l_Dirty)
                        {
                            MarkAsDirty();
                        }
                    }
                }

                private void HandleSearchResults(ref bool a_Conflict)
                {
                    if (null == f_SearchCookie)
                    {
                        HandleKadFilterOfResults();

                        return;
                    }

                    if (true == f_SearchCookie.f_KadFiltered &&
                        false == f_ResultsKadFilteredDuringLastSearchRun)
                    {
                        f_ResultsKadFilteredDuringLastSearchRun = true;
                    }

                    CookieStates l_Status = f_SearchCookie.p_NetworkStateVolatile;

                    if (CookieStates.Pending == l_Status)
                    {
                        return;
                    }

                    uint l_CurrentGen = Util.VolatileRead(ref f_SearchCookie.f_GuiResultsGeneration);

                    if (f_CookieResultGeneration != l_CurrentGen)
                    {
                        if (null != f_ResultsStatusUpdateCookie)
                        {
                            // There is a status update in progress
                            // that should finish first (since it uses
                            // our array).

                            return;
                        }

                        f_CookieResultGeneration = l_CurrentGen;
                        Dictionary<FileID, KadFoundFile> l_Input = Util.VolatileRead(ref f_SearchCookie.f_GuiResults);

                        if (null != l_Input)
                        {
                            IntegrateResults(l_Input);

                            HandleKadFilterOfResults();
                        }
                    }

                    if (CookieStates.SearchRunning != l_Status)
                    {
                        if (CookieStates.Failure == l_Status)
                        {
                            a_Conflict = true;
                        }

                        NotifySearchEnded();
                    }
                }

                private void IntegrateResults(Dictionary<FileID, KadFoundFile> a_Input)
                {
                    FoundFile[] l_Arr = f_Files.p_Arr;

                    if (null == f_IndexMap)
                    {
                        // Build the index.

                        int l_Count = f_Files.p_Count;
                        f_IndexMap = new Dictionary<uint, int>(l_Count);

                        if (0 < l_Count)
                        {
                            for (int i = 0; i < l_Count; i++)
                            {
                                l_Arr[i].ResetSourceCounts();

                                f_IndexMap[l_Arr[i].f_HashValue] = i;
                            }

                            MarkAsDirty();
                        }
                    }

                    if (0 >= a_Input.Count)
                    {
                        return;
                    }

                    StorableTime l_StorableTime = Software.GetStorableTime();
                    byte[] l_TempBuffer = new byte[FoundFile.c_MinTempBufferSize];

                    foreach (KadFoundFile l_InputFile in a_Input.Values)
                    {
                        FoundFile l_File;
                        l_InputFile.ToValue(l_StorableTime, l_TempBuffer, out l_File);
                        int l_Idx;

                        if (true == f_IndexMap.TryGetValue(l_File.f_HashValue, out l_Idx))
                        {
                            bool l_UpdateExistingFile = l_File.SameFile(ref l_Arr[l_Idx]);

                            if (false == l_UpdateExistingFile)
                            {
                                int l_Count = f_Files.p_Count;

                                for (l_Idx = 0; l_Idx < l_Count; l_Idx++)
                                {
                                    if (true == l_File.SameFile(ref l_Arr[l_Idx]))
                                    {
                                        l_UpdateExistingFile = true;

                                        break;
                                    }
                                }
                            }

                            if (true == l_UpdateExistingFile)
                            {
                                l_Arr[l_Idx].TakeOverSourceCountAndLastAppearanceTime(ref l_File);

                                continue;
                            }
                        }

                        // A file with this hash is not indexed.

                        if (c_MaxFileCount > f_Files.p_Count)
                        {
                            l_Idx = f_Files.ReserveAtEnd();
                            l_Arr = f_Files.p_Arr;
                            l_Arr[l_Idx] = l_File;
                            f_IndexMap[l_File.f_HashValue] = l_Idx;
                            f_NeedStatusUpdate = true;
                        }
                    }

                    MarkAsDirty();
                }

                private void HandleResultsStatusUpdate(Time a_Time)
                {
                    if (null == f_ResultsStatusUpdateCookie)
                    {
                        if (true == f_Entered &&
                            a_Time >= f_StatusExpirationTime)
                        {
                            f_NeedStatusUpdate = true;
                        }
                    }
                    else
                    {
                        CookieStates l_Status = f_ResultsStatusUpdateCookie.p_StateVolatile;

                        if (CookieStates.Pending == l_Status)
                        {
                            return;
                        }

                        f_ResultsStatusUpdateCookie = null;

                        if (CookieStates.Success == l_Status)
                        {
                            f_BlockStatusUpdateTime = a_Time.AddSeconds(c_ResultsStatusUpdateIntervalSecs);
                            f_StatusExpirationTime = a_Time.AddSeconds(5);

                            if (true == f_Entered &&
                                0 < f_Files.p_Count)
                            {
                                f_GridView.Invalidate();
                            }

                            return;
                        }
                    }

                    if (true == f_NeedStatusUpdate)
                    {
                        int l_Count = f_Files.p_Count;

                        if (0 == l_Count)
                        {
                            f_NeedStatusUpdate = false;

                            return;
                        }

                        if (true == f_BlockStatusUpdateTime.IsEarlierThan(a_Time))
                        {
                            FoundFile[] l_Arr = f_Files.p_Arr;

                            f_NeedStatusUpdate = false;
                            f_ResultsStatusUpdateCookie = new GeneralAPCCookie(f_Cache.f_Inst.f_ContentMngr.GuiUpdateFoundFilesCompletionStatus, new object[2] { l_Arr, l_Count });
                            f_Cache.f_Inst.f_ContentMngr.EnqueueAPC(f_ResultsStatusUpdateCookie);

                            MarkAsDirty();
                        }
                    }
                }

                public void MarkAsDirtyAndUpdateResults()
                {
                    MarkAsDirty();
                    UpdateResults();
                }

                private void MarkAsDirty()
                {
                    f_ResultsDirty = true;
                }

                private void UpdateResults()
                {
                    int l_Total = f_Files.p_Count;

                    if (0 == l_Total)
                    {
                        f_ResultsDirty = false;

                        f_GridView.ClearItems();
                        SetCaptionWithCount(null);

                        return;
                    }

                    if (false == f_Entered ||
                        false == f_ResultsDirty)
                    {
                        return;
                    }

                    f_ResultsDirty = false;

                    bool l_HideKnownFiles = f_Cache.f_Filter.f_HideKnownFiles;
                    bool l_HasStrFilter = (StrFilterMode.Off != f_Cache.f_Filter.f_StrFilterMode);
                    FoundFile[] l_Arr = f_Files.p_Arr;

                    f_GridView.StartChanging();

                    for (int i = 0; i < l_Total; i++)
                    {
                        if (true == l_HideKnownFiles &&
                            ComplStatus.Dropped != l_Arr[i].f_ComplStatus)
                        {
                            continue;
                        }

                        if (true == l_HasStrFilter &&
                            true == f_Cache.f_Filter.IsFiltered(l_Arr[i].f_FileName))
                        {
                            continue;
                        }

                        GridViewItem<int, bool> l_Item = f_GridView.CreateItem(i);
                    }

                    f_GridView.EndChanging();

                    TextSet l_TextSet = f_Cache.f_Inst.f_TextSet;
                    int l_SubsetCount = f_GridView.p_ItemCount;
                    string l_CaptionWithCount;

                    if (l_SubsetCount == l_Total)
                    {
                        l_CaptionWithCount = l_TextSet.Format(TextKeys.NameWithCountFormat, f_BaseCaption, l_Total);
                    }
                    else
                    {
                        l_CaptionWithCount = l_TextSet.Format(TextKeys.NameWithTwoCountsFormat, f_BaseCaption, l_SubsetCount, l_Total);
                    }

                    SetCaptionWithCount(l_CaptionWithCount);
                }

                private void GridView_UpdateItem(Time a_Time,
                                                 GridViewItem<int, bool> a_Item)
                {
                    FoundFile[] l_Arr = f_Files.p_Arr;
                    int i = a_Item.f_Ref;

                    if (f_Files.p_Count <= i)
                    {
                        return;
                    }

                    FileID l_FileID = new FileID(l_Arr[i].f_FileID.ToRef());
                    TextSet l_TextSet = f_Cache.f_Inst.f_TextSet;
                    MainForm l_Form = f_Cache.f_Inst.f_MainForm;

                    a_Item.SetTextWithCompObj(0, l_Arr[i].f_FileName, l_Arr[i].f_FileName);
                    a_Item.SetText(1, l_Arr[i].f_FileSize, l_Form.BuildDefaultSizeStr(l_Arr[i].f_FileSize));
                    a_Item.SetText(2, (((uint)l_Arr[i].f_CompleteSources << Util.c_BitsPerUShort) | l_Arr[i].p_PartialSources), l_Arr[i].GetSourcesStr(FileSearchPanel.s_SourcesFormat));
                    a_Item.SetText(3, (byte)l_Arr[i].f_FileType, l_Arr[i].GetFileTypeStr());
                    a_Item.SetTextWithCompObj(4, l_FileID.f_Val, l_FileID.f_Val.ToHexString());
                    a_Item.SetTime(5, l_TextSet, a_Time, l_Arr[i].f_FirstAppearanceTime.ToTime(), null);
                    a_Item.SetTime(6, l_TextSet, a_Time, l_Arr[i].f_LastAppearanceTime.ToTime(), null);

                    a_Item.DefaultCellStyle = f_Cache.GetCellStyle(l_Arr[i].f_ComplStatus);
                }
            }
        }
    }

    class StatsPanel : AdvancedTabPage
    {
        private readonly Instance f_Inst;

        private readonly ReadOnlyTextBox f_TextBox;

        public StatsPanel(Instance a_Inst)
        {
            f_Inst = a_Inst;

            f_TextBox = new ReadOnlyTextBox(this);
            f_TextBox.Dock = DockStyle.Fill;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_TextBox.Dispose();
            }

            base.Dispose(disposing);
        }

        public void SetTextBoxFont(Font a_Font)
        {
            f_TextBox.Font = a_Font;
        }

        public void SetText(string a_Text)
        {
            f_TextBox.SetTextAndHoldPosition(a_Text);
        }
    }

    abstract class SplitPanel<TKey, TVal> : AdvancedTabPage
    {
        protected readonly Instance f_Inst;

        private readonly HorSplitContainer f_Splitter;
        protected readonly GuiLabel f_Label;
        protected readonly GridView<TKey, TVal> f_GridView;

        protected SplitPanel(Instance a_Inst,
                             ColumnSet a_ColSet)
        {
            f_Inst = a_Inst;

            f_Splitter = HorSplitContainer.CreateFixed(this);
            f_Label = new GuiLabel(f_Splitter);
            f_GridView = new GridView<TKey, TVal>(f_Inst, f_Splitter, a_ColSet);
            f_Splitter.SetControls(f_Label, f_GridView);

            f_Label.Padding = new Padding(Lyt.s_SmallSpace);
            f_Label.TextAlign = ContentAlignment.MiddleLeft;
            f_GridView.Dock = DockStyle.None;
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                f_Label.Dispose();
                f_GridView.Dispose();
                f_Splitter.Dispose();
            }

            base.Dispose(disposing);
        }

        public void ArrangeLayoutOnShown()
        {
            f_Splitter.SetUpperHeight(f_Label.Font.Height + Lyt.s_Space);
        }

        public void FinishLoading()
        {
            f_GridView.LoadColumnAdjustments();
        }

        public void PrepareSaving()
        {
            f_GridView.StoreColumnAdjustments();
        }
    }

    class KademliaPanel : SplitPanel<UInt128, int>
    {
        private const byte c_DistanceBitsToDisplay = (UInt128.c_BitSize / 4);

        private TimeStrHelper f_TimeStrHelper;
        private GuiContactInfo[] f_ContactInfos;

        public KademliaPanel(Instance a_Inst) :
                        base(a_Inst, a_Inst.f_MainForm.f_VisualCust.f_KademliaPanelColSet)
        {
            f_GridView.e_UpdateItem += GridView_UpdateItem;
        }

        public void Update(GuiContactInfo[] a_ContactInfos)
        {
            // a_ContactInfos may be null.

            TextKeys l_KadStatusKey;

            switch (f_Inst.f_Kad.p_KadStatusVolatile)
            {
                case KademliaStatus.Connected:
                    l_KadStatusKey = TextKeys.KademliaTabConn;
                    break;

                case KademliaStatus.Disconnected:
                    l_KadStatusKey = TextKeys.KademliaTabNotConn;
                    break;

                case KademliaStatus.TryingToConnect:
                    l_KadStatusKey = TextKeys.KademliaTabTryingToConn;
                    break;

                default: throw (new ApplicationException());
            }

            string l_KadStatusStr = f_Inst[l_KadStatusKey];
            f_ContactInfos = a_ContactInfos;

            if (null == f_ContactInfos)
            {
                f_GridView.ClearItems();
            }
            else
            {
                TimeStrHelper.Create(f_Inst.GetTime(), out f_TimeStrHelper);
                int l_Count = f_ContactInfos.Length;
                int l_VerifiedContacts = 0;

                f_GridView.StartChanging();

                for (int i = 0; i < l_Count; i++)
                {
                    GridViewItem<UInt128, int> l_Item = f_GridView.CreateItem(f_ContactInfos[i].f_ClientID.f_Val);
                    l_Item.f_Tag = i;

                    bool l_Verified = f_ContactInfos[i].f_LastActivity.IsSet();

                    if (true == l_Verified)
                    {
                        l_VerifiedContacts++;
                    }
                }

                f_GridView.EndChanging();

                l_KadStatusStr = f_Inst.f_TextSet.Format(TextKeys.KademliaTabContactCountFormat, l_KadStatusStr, l_VerifiedContacts, l_Count);
            }

            f_Label.Text = l_KadStatusStr;
            f_GridView.SelectIfTabPageIsSelected(this);
        }

        private void GridView_UpdateItem(Time a_Time,
                                         GridViewItem<UInt128, int> a_Item)
        {
            int l_Idx = a_Item.f_Tag;

            if (f_ContactInfos.Length <= l_Idx)
            {
                return;
            }

            a_Item.SetTextWithCompObj(0, f_ContactInfos[l_Idx].f_ClientID.f_Val, f_ContactInfos[l_Idx].f_ClientID.f_Val.ToHexString());
            a_Item.SetTime(1, f_Inst.f_TextSet, a_Time, f_ContactInfos[l_Idx].f_LastActivity, string.Empty);
            a_Item.SetTextWithCompObj(2, f_ContactInfos[l_Idx].f_Distance.f_Val, f_ContactInfos[l_Idx].f_Distance.f_Val.ToBinaryString(c_DistanceBitsToDisplay));
        }
    }

    class GeneralPeersPanel : SplitPanel<UserContext, bool>
    {
        private static string s_InConnCount;
        private static string s_OutConnCount;
        private static string s_JoinedFormat;

        // The following cell styles are not owned by this class.
        private readonly DataGridViewCellStyle f_NormalCellStyle;
        private readonly DataGridViewCellStyle f_FriendCellStyle;

        // f_AutoSelectUserCntxt can be null.
        private UserContext f_AutoSelectUserCntxt;

        public GeneralPeersPanel(Instance a_Inst,
                                 DataGridViewCellStyle a_NormalCellStyle,
                                 DataGridViewCellStyle a_FriendCellStyle) :
                            base(a_Inst, a_Inst.f_MainForm.f_VisualCust.f_GeneralPeersColSet)
        {
            f_NormalCellStyle = a_NormalCellStyle;
            f_FriendCellStyle = a_FriendCellStyle;

            f_GridView.e_FillContextMenu += FillContextMenu;
            f_GridView.e_UpdateItem += GridView_UpdateItem;
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_InConnCount = a_TextSet[TextKeys.PeersTabInConnCountFormat];
            s_OutConnCount = a_TextSet[TextKeys.PeersTabOutConnCountFormat];
            s_JoinedFormat = a_TextSet[TextKeys.PeersTabConnCountJoinedFormat];
        }

        private bool FillContextMenu(GridView<UserContext, bool> a_Sender,
                                     ContextMenuStrip a_Menu,
                                     bool a_ItemHit)
        {
            int l_SelCount = f_GridView.p_SelectedCount;

            if (0 == l_SelCount)
            {
                return false;
            }

            TextSet l_TextSet = f_Inst.f_TextSet;

            if (1 == l_SelCount)
            {
                MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.PeersTabCntxtMenuCopyStrmAddr], MenuCopyStreamAddress_Click, IconTypes.Pendant);
                MainForm.AddMenuSeparator(a_Menu);
            }

            GenericMenuItem<bool>.CreateItem(a_Menu, l_TextSet[TextKeys.PeersTabCntxtMenuMarkAsFriend], MenuAdjustFriendStatus_Click, IconTypes.Heart, true);
            GenericMenuItem<bool>.CreateItem(a_Menu, l_TextSet[TextKeys.PeersTabCntxtMenuResetFriendStatus], MenuAdjustFriendStatus_Click, IconTypes.Recycle, false);
            MainForm.AddMenuSeparator(a_Menu);
            GenericMenuItem<bool>.CreateItem(a_Menu, l_TextSet[TextKeys.PeersTabCntxtMenuParoleBannedPeers], MenuBanOrUnban_Click, IconTypes.Shirt, false);

            if (true == f_Inst.f_MainForm.p_Prefs.f_GeneralPrefs.f_ShowPeersPanelAdvSubMenu)
            {
                ToolStripMenuItem l_Advanced = MainForm.AddMenuItem(a_Menu, l_TextSet[TextKeys.PeersTabCntxtMenuAdvanced], null);
                GenericMenuItem<bool>.CreateSubItem(l_Advanced, l_TextSet[TextKeys.PeersTabCntxtMenuBan], MenuBanOrUnban_Click, true);
                MainForm.AddMenuSeparator(l_Advanced);
                MainForm.AddMenuItem(l_Advanced, l_TextSet.GetAsDialogOpener(TextKeys.PeersTabCntxtMenuAdjustReputation), MenuSetReputation_Click);
                MainForm.AddMenuItem(l_Advanced, l_TextSet[TextKeys.PeersTabCntxtMenuSetNegRepToZero], MenuNullifyRepIfNegative_Click);
            }

            return true;
        }

        private void MenuCopyStreamAddress_Click(object sender, EventArgs e)
        {
            if (1 != f_GridView.p_SelectedCount)
            {
                return;
            }

            I2PAddress l_Addr = f_GridView.GetSelectedItem(0).f_Ref.f_StreamAddr;

            Util.SetClipboard(l_Addr.ToFullString());
        }

        private void MenuAdjustFriendStatus_Click(object sender, EventArgs e)
        {
            GenericMenuItem<bool> l_MenuItem = (GenericMenuItem<bool>)sender;
            bool l_Friend = l_MenuItem.f_GenericTag;

            foreach (GridViewItem<UserContext, bool> l_Item in f_GridView.SelectedRows)
            {
                if (l_Item.f_Ref.IsFriend() != l_Friend)
                {
                    l_Item.f_Ref.GuiAdjustFriendStatus(l_Friend);
                }
            }
        }

        private void MenuBanOrUnban_Click(object sender, EventArgs e)
        {
            GenericMenuItem<bool> l_MenuItem = (GenericMenuItem<bool>)sender;
            bool l_Ban = l_MenuItem.f_GenericTag;

            BanOrUnbanSelectedUsers(l_Ban);
        }

        private void BanOrUnbanSelectedUsers(bool a_Ban)
        {
            UserContext[] l_Arr = f_GridView.SelectedItemsToKeyArr();

            if (null == l_Arr)
            {
                return;
            }

            if (true == a_Ban &&
                false == f_Inst.f_MainForm.YesNoMsgBox(f_Inst[TextKeys.PeersTabQuestionReallyBanPeers]))
            {
                return;
            }

            object[] l_Params = new object[] { l_Arr, a_Ban };

            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiBanOrUnbanUsers, l_Params);
            f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
        }

        private void MenuSetReputation_Click(object sender, EventArgs e)
        {
            UserContext[] l_Arr = f_GridView.SelectedItemsToKeyArr();

            if (null == l_Arr)
            {
                return;
            }

            using (SetReputationDialog l_Dlg = new SetReputationDialog(f_Inst))
            {
                if (1 == l_Arr.Length)
                {
                    // Volatile read.

                    int l_SuggRep = l_Arr[0].p_Reputation;
                    Util.MemoryBarrier();

                    l_Dlg.SetReputation(l_SuggRep);
                }

                if (DialogResult.OK == l_Dlg.ShowDialog())
                {
                    AdjustReputationOfSelectedUsers(false, l_Dlg.GetReputation());
                }
            }
        }

        private void MenuNullifyRepIfNegative_Click(object sender, EventArgs e)
        {
            AdjustReputationOfSelectedUsers(true, 0);
        }

        public void AdjustReputationOfSelectedUsers(bool a_NullifyIfNegative,
                                                    int a_TargetReputation)
        {
            UserContext[] l_Arr = f_GridView.SelectedItemsToKeyArr();

            if (null == l_Arr)
            {
                return;
            }

            object[] l_Params = new object[] { l_Arr, a_NullifyIfNegative, a_TargetReputation };

            GeneralAPCCookie l_Cookie = new GeneralAPCCookie(f_Inst.f_NetworkMngr.GuiAdjustReputationOfUsers, l_Params);
            f_Inst.f_NetworkMngr.EnqueueAPCAndWaitAsGUI(l_Cookie);
        }

        public void SetAutoSelectUserContext(UserContext a_UserCntxt)
        {
            GridViewItem<UserContext, bool> l_Item = f_GridView.GetItemByKey(a_UserCntxt);

            if (null == l_Item)
            {
                // Try to do this in the next update cycle.

                f_AutoSelectUserCntxt = a_UserCntxt;
            }
            else
            {
                f_GridView.SelectAndDisplayItem(l_Item);

                f_AutoSelectUserCntxt = null;
            }
        }

        public void UpdatePeers(MainForm.FormInput.GeneralPeers a_Peers)
        {
            int l_Count = (null == a_Peers.f_UserCntxts ? 0 : a_Peers.f_UserCntxts.Length);
            f_Label.Text = f_Inst.f_TextSet.Format(TextKeys.PeersTabCountFormat, a_Peers.f_ConnectedPeers, l_Count);

            if (0 == l_Count)
            {
                f_GridView.ClearItems();
            }
            else
            {
                UserContext[] l_UserCntxts = a_Peers.f_UserCntxts;
                GridViewItem<UserContext, bool> l_AutoSelItem = null;

                f_GridView.StartChanging();

                for (int i = 0; i < l_Count; i++)
                {
                    GridViewItem<UserContext, bool> l_Item = f_GridView.CreateItem(l_UserCntxts[i]);

                    if (l_Item.f_Ref == f_AutoSelectUserCntxt)
                    {
                        l_AutoSelItem = l_Item;
                    }
                }

                f_GridView.EndChanging();

                if (null != l_AutoSelItem)
                {
                    f_GridView.SelectAndDisplayItem(l_AutoSelItem);
                }

                f_GridView.SelectIfTabPageIsSelected(this);
            }

            f_AutoSelectUserCntxt = null;
        }

        private void GridView_UpdateItem(Time a_Time,
                                         GridViewItem<UserContext, bool> a_Item)
        {
            TextSet l_TextSet = f_Inst.f_TextSet;
            UserContext l_UserCntxt = a_Item.f_Ref;
            PeerInfo l_PeerInfo;
            l_UserCntxt.GetInfoForGeneralPeersPanelVolatile(out l_PeerInfo);
            RsaPublicKey l_RsaPubKey = l_PeerInfo.f_RsaPubKey;
            uint l_KeySize = (false == l_RsaPubKey.IsValid() ? 0 : l_RsaPubKey.GetBitSize());
            string l_PubKeyStr = (0 == l_KeySize ? l_TextSet[TextKeys.RsaKeyNo] : l_TextSet.Format(TextKeys.RsaKeyYesFormat, l_KeySize));
            string l_UserIDStr = l_PeerInfo.f_UserID.AsStringNullSafe(string.Empty);
            string l_UserName = (null == l_PeerInfo.f_UserName ? string.Empty : l_PeerInfo.f_UserName);

            a_Item.SetText(0, l_UserCntxt.f_UserNum.f_Val, l_TextSet.NumStr(l_UserCntxt.f_UserNum.f_Val));
            a_Item.SetTextWithCompObj(1, l_UserCntxt.f_StreamAddr, l_UserCntxt.f_StreamAddr.ToShortString());
            a_Item.SetText(2, l_KeySize, l_PubKeyStr);
            a_Item.SetTextWithCompObj(3, l_UserIDStr, l_UserIDStr);
            a_Item.SetTextWithCompObj(4, l_UserName, l_UserName);
            a_Item.SetBool(5, l_PeerInfo.f_Friend);

            if (false == l_PeerInfo.f_Connected)
            {
                a_Item.SetText(6, 0, string.Empty);
                a_Item.SetTime(7, l_TextSet, a_Time, l_PeerInfo.f_LastConnectedTime, TimeStrHelper.s_NeverStr);
            }
            else
            {
                string l_ConnStr = BuildConnectionsCountString(l_TextSet, l_PeerInfo.f_IngoingEstabConns, l_PeerInfo.f_OutgoingEstabConns);

                if (null == l_ConnStr)
                {
                    l_ConnStr = string.Empty;
                }

                a_Item.SetText(6, GridViewItem<UserContext, bool>.BuildCompareWeight(l_PeerInfo.f_IngoingEstabConns, l_PeerInfo.f_OutgoingEstabConns), l_ConnStr);
                a_Item.SetText(7, ulong.MaxValue, TimeStrHelper.s_NowStr);
            }

            a_Item.SetText(8, l_PeerInfo.f_TransferVol.p_BytesReceived, f_Inst.f_MainForm.BuildDefaultSizeStr(l_PeerInfo.f_TransferVol.p_BytesReceived));
            a_Item.SetText(9, l_PeerInfo.f_TransferVol.p_BytesSent, f_Inst.f_MainForm.BuildDefaultSizeStr(l_PeerInfo.f_TransferVol.p_BytesSent));
            a_Item.SetText(10, l_PeerInfo.f_Reputation, l_TextSet.NumStr(l_PeerInfo.f_Reputation));
            a_Item.SetBool(11, l_PeerInfo.f_Banned);

            a_Item.DefaultCellStyle = (false == l_PeerInfo.f_Friend ? f_NormalCellStyle : f_FriendCellStyle);
        }

        public static string BuildConnectionsCountString(TextSet a_TextSet,
                                                         uint a_IngoingEstabConns,
                                                         uint a_OutgoingEstabConns)
        {
            // Returns null if both counts are zero.

            string l_ConnStr = null;

            if (0 != a_IngoingEstabConns)
            {
                l_ConnStr = a_TextSet.Format(s_InConnCount, a_IngoingEstabConns);
            }

            if (0 != a_OutgoingEstabConns)
            {
                string l_OutConnStr = a_TextSet.Format(s_OutConnCount, a_OutgoingEstabConns);

                if (null == l_ConnStr)
                {
                    l_ConnStr = l_OutConnStr;
                }
                else
                {
                    l_ConnStr = a_TextSet.Format(s_JoinedFormat, l_ConnStr, l_OutConnStr);
                }
            }

            return l_ConnStr;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct PeerInfo
        {
            // f_RsaPubKey, f_UserID.f_Val and f_UserName can be null.
            public readonly RsaPublicKey f_RsaPubKey;
            public readonly UserID f_UserID;
            public readonly string f_UserName;
            public readonly bool f_Friend;
            public readonly bool f_Connected;
            public readonly ushort f_IngoingEstabConns;
            public readonly ushort f_OutgoingEstabConns;
            public readonly Time f_LastConnectedTime;
            public readonly TransferVolume f_TransferVol;
            public readonly int f_Reputation;
            public readonly bool f_Banned;

            public PeerInfo(RsaPublicKey a_RsaPubKey,
                            UserID a_UserID,
                            string a_UserName,
                            bool a_Friend,
                            bool a_Connected,
                            ushort a_IngoingEstabConns,
                            ushort a_OutgoingEstabConns,
                            Time a_LastConnectedTime,
                            ref TransferVolume a_TransferVol,
                            int a_Reputation,
                            bool a_Banned)
            {
                // This constructor can be safely used with volatile
                // data.

                f_RsaPubKey = a_RsaPubKey;
                f_UserID = a_UserID;
                f_UserName = a_UserName;
                f_Friend = a_Friend;
                f_Connected = a_Connected;
                f_IngoingEstabConns = a_IngoingEstabConns;
                f_OutgoingEstabConns = a_OutgoingEstabConns;
                f_LastConnectedTime = a_LastConnectedTime;
                f_TransferVol = a_TransferVol;
                f_Reputation = a_Reputation;
                f_Banned = a_Banned;
            }
        }

        private class SetReputationDialog : ButtonDialog
        {
            private const ushort c_DlgWidth = 350;

            private readonly Instance f_Inst;

            private readonly GuiGroupBox f_GroupBox;
            private readonly GuiLabel f_InfoLabel;
            private readonly LabeledNumberBoxWithUnit f_NumBox;

            public SetReputationDialog(Instance a_Inst) :
                                  base(a_Inst.f_TextSet, DialogButtons.OKCancel)
            {
                f_Inst = a_Inst;

                this.SuspendLayout();

                int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
                int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
                f_GroupBox = new GuiGroupBox(this);
                int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
                f_InfoLabel = new GuiLabel(f_GroupBox);
                f_NumBox = new LabeledNumberBoxWithUnit(f_GroupBox, l_InnerWidth, Lyt.CalcX(95));

                TextSet l_TextSet = a_Inst.f_TextSet;
                f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
                f_InfoLabel.Location = new Point(f_GroupBox.Padding.Left, f_GroupBox.Padding.Top);
                f_InfoLabel.Width = l_InnerWidth;
                f_InfoLabel.Text = l_TextSet[TextKeys.PeersTabSetRepDlgInfoLabel];
                Lyt.SetFontAndHeight(f_InfoLabel, f_InfoLabel.Font);
                f_NumBox.SetLocation(f_InfoLabel.Left, f_InfoLabel.GetBottom() + Lyt.s_SmallSpace);
                f_NumBox.SetLimits(int.MinValue, int.MaxValue);
                f_GroupBox.SetSizePlusBottomPadding(l_GroupBoxWidth, f_NumBox.f_Box.Bottom);

                AsFixedButtonDialog(l_DlgWidth, f_GroupBox);
                base.SetSelectOnShowControl(f_NumBox.f_Box);

                this.f_OKButton.Click += DefaultOKCloseHandler;

                this.Text = l_TextSet[TextKeys.PeersTabSetRepDlgTitle];
                f_GroupBox.Text = l_TextSet[TextKeys.PeersTabSetRepDlgGroupBox];
                f_NumBox.SetCaption(l_TextSet, TextKeys.PeersTabSetRepDlgNumBox);
                f_NumBox.SetUnitText(l_TextSet, TextKeys.PeersTabSetRepDlgUnits);

                this.ResumeLayout();
            }

            protected override void Dispose(bool disposing)
            {
                if (true == disposing)
                {
                    Dialog.DisposeAllChildren(f_GroupBox);
                    f_GroupBox.Dispose();
                }

                base.Dispose(disposing);
            }

            public int GetReputation()
            {
                return f_NumBox.p_ValueAsInt;
            }

            public void SetReputation(int a_Reputation)
            {
                f_NumBox.f_Box.Value = a_Reputation;
            }
        }
    }

    sealed class MoveDataFilesDialog : ButtonDialog
    {
        private const byte c_FilesToCopy = 4;
        private const string c_FileMoveThreadName = "FileMover";

        // This dialog is used while the program terminates, so this
        // class must rely on itself.

        private const ushort c_DlgWidth = 500;
        private const ushort c_TimerInterval = 100;

        private static string s_ProgressFormat;

        private readonly TextSet f_TextSet;
        // f_MainForm should only be used for message boxes, since
        // large parts of it have already shut down when this dialog
        // is shown.
        private readonly MainForm f_MainForm;
        private readonly bool f_DisplaySISizes;

        private readonly GuiGroupBox f_GroupBox;
        private readonly TextGroup f_SrcDirText;
        private readonly TextGroup f_DestDirText;
        private readonly TextGroup f_ProgressText;
        private readonly GuiProgressBar f_ProgressBar;
        private readonly System.Windows.Forms.Timer f_Timer;

        private MoveContext f_MoveCntxt;
        private bool f_MayClose;
        private bool f_UserWantsToAbort;
        private bool f_ReallyAbortQuestionIsShown;

        public MoveDataFilesDialog(TextSet a_TextSet,
                                   MainForm a_MainForm,
                                   bool a_DisplaySISizes) :
                              base(a_TextSet, DialogButtons.Cancel)
        {
            f_TextSet = a_TextSet;
            f_MainForm = a_MainForm;
            f_DisplaySISizes = a_DisplaySISizes;

            this.SuspendLayout();

            f_GroupBox = new GuiGroupBox(this);
            int l_DlgWidth = Lyt.CalcX(c_DlgWidth);
            int l_GroupBoxWidth = (l_DlgWidth - Lyt.s_Space2);
            int l_InnerWidth = (l_GroupBoxWidth - f_GroupBox.Padding.Horizontal);
            int l_LeftPadding = f_GroupBox.Padding.Left;
            f_SrcDirText = TextGroup.Create(f_GroupBox, l_LeftPadding, f_GroupBox.Padding.Top);
            f_DestDirText = TextGroup.Create(f_GroupBox, l_LeftPadding, f_SrcDirText.GetBottom());
            f_ProgressText = TextGroup.Create(f_GroupBox, l_LeftPadding, f_DestDirText.GetBottom());
            f_ProgressBar = new GuiProgressBar(f_GroupBox);

            f_GroupBox.Location = new Point(Lyt.s_Space, Lyt.s_Space);
            f_SrcDirText.SetWidthAndAlignLeft(l_InnerWidth);
            f_DestDirText.SetWidthAndAlignLeft(l_InnerWidth);
            f_ProgressText.SetWidthAndAlignLeft(l_InnerWidth);
            f_ProgressBar.Location = new Point(l_LeftPadding, f_ProgressText.GetBottom() + Lyt.s_SmallSpace);
            f_ProgressBar.Width = l_InnerWidth;
            f_GroupBox.Size = new Size(l_GroupBoxWidth, f_ProgressBar.Bottom + f_GroupBox.Padding.Bottom);

            AsFixedButtonDialog(l_DlgWidth, f_GroupBox);

            f_Timer = new Timer();
            f_Timer.Interval = c_TimerInterval;
            f_Timer.Tick += Timer_Tick;

            Shown += Form_Shown;
            FormClosing += Form_Closing;

            this.Text = f_TextSet[TextKeys.MoveDataFilesDlgTitle];
            f_GroupBox.Text = f_TextSet[TextKeys.MoveDataFilesDlgGroupBoxCaption];
            f_SrcDirText.SetLeftText(f_TextSet, TextKeys.MoveDataFilesDlgSrcDir);
            f_DestDirText.SetLeftText(f_TextSet, TextKeys.MoveDataFilesDlgDestDir);
            f_ProgressText.SetLeftText(f_TextSet, TextKeys.MoveDataFilesDlgProgress);

            this.ResumeLayout();
        }

        protected override void Dispose(bool disposing)
        {
            if (true == disposing)
            {
                Dialog.DisposeAllChildren(f_GroupBox);
                f_GroupBox.Dispose();

                f_Timer.Dispose();
            }

            base.Dispose(disposing);
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_ProgressFormat = a_TextSet[TextKeys.MoveDataFilesDlgProgressFormat];
        }

        public void HandleMoving(IWin32Window a_Parent)
        {
            BootConfig l_BootConfig = Software.s_BootConfig;
            string l_SrcDir = l_BootConfig.p_AppDataDir;
            string l_DestDir = l_BootConfig.p_NewAppDataDir;

            if (l_SrcDir == l_DestDir)
            {
                Util.DebugBreak();

                return;
            }

            string l_Quest = f_TextSet.Format(TextKeys.MoveDataFilesDlgQuestReallyMoveFiles, l_SrcDir, l_DestDir);
            if (false == f_MainForm.YesNoMsgBox(l_Quest))
            {
                return;
            }

            string l_ErrMsg;
            f_MoveCntxt = MoveContext.Create(f_TextSet, l_SrcDir, l_DestDir, out l_ErrMsg);

            if (null == f_MoveCntxt)
            {
                f_MainForm.ErrMsgBox(l_ErrMsg);

                return;
            }

            f_SrcDirText.SetRightText(l_SrcDir);
            f_DestDirText.SetRightText(l_DestDir);

            this.ShowDialog(a_Parent);
        }

        private void Form_Shown(object sender, EventArgs e)
        {
            Timer_Tick(null, null);
            f_Timer.Start();

            Util.StartThread(c_FileMoveThreadName, FileMoveWorker);
        }

        private void Form_Closing(object sender, FormClosingEventArgs e)
        {
            if (true == f_MayClose)
            {
                return;
            }

            // We need to ask the user if this is really
            // wanted first.
            f_UserWantsToAbort = true;
            e.Cancel = true;
        }

        private void Timer_Tick(object sender, EventArgs e)
        {
            ulong l_Done;
            string l_ErrMsg;
            bool l_Finished = f_MoveCntxt.QueryVolatileFields(out l_Done, out l_ErrMsg);

            ulong l_Total = f_MoveCntxt.f_TotalSize;
            float f_Perc = (((float)l_Done / l_Total));
            string l_ProgressStr = string.Format(s_ProgressFormat, f_Perc, BuildSizeString(l_Done), BuildSizeString(l_Total));

            f_ProgressText.SetRightText(l_ProgressStr);
            f_ProgressBar.SetProgress(l_Done, l_Total);

            if (false == l_Finished)
            {
                if (true == f_UserWantsToAbort)
                {
                    f_UserWantsToAbort = false;
                    f_ReallyAbortQuestionIsShown = true;
                    {
                        if (DialogResult.Yes == f_MainForm.AbortableYesNoMsgBox(f_MoveCntxt, f_TextSet[TextKeys.MoveDataFilesDlgQuestionReallyAbort]))
                        {
                            f_MoveCntxt.Abort();
                            f_CancelButton.Enabled = false;
                        }
                    }
                    f_ReallyAbortQuestionIsShown = false;
                }

                return;
            }

            if (true == f_ReallyAbortQuestionIsShown)
            {
                // Wait until the message box is closed.

                return;
            }

            f_CancelButton.Enabled = false;
            f_Timer.Stop();

            if (null == l_ErrMsg)
            {
                f_MainForm.InfoMsgBox(f_TextSet[TextKeys.MoveDataFilesDlgMsgMoveSuccess]);
                this.DialogResult = DialogResult.OK;
            }
            else
            {
                f_MainForm.ErrMsgBox(l_ErrMsg);
            }

            PrepareClose();
        }

        private void PrepareClose()
        {
            f_MayClose = true;
            Close();
        }

        private string BuildSizeString(ulong a_Bytes)
        {
            // We don't use MainForm for this, since it's in
            // termination while this executes.

            return SizeFormatter.BuildSizeStr(f_DisplaySISizes, false, false, true, a_Bytes);
        }

        private void FileMoveWorker()
        {
            string l_ErrMsg;

            for (int i = 0; i < c_FilesToCopy; i++)
            {
                FHandle.AsyncCopyResult l_Res = FHandle.CopyAsync(f_MoveCntxt.f_SrcFiles[i], f_MoveCntxt.f_DestFiles[i], ref f_MoveCntxt.f_Run, ref f_MoveCntxt.f_Done);

                if (FHandle.AsyncCopyResult.Success == l_Res)
                {
                    continue;
                }

                if (false == f_MoveCntxt.f_Run)
                {
                    // Aborted by user.

                    l_ErrMsg = f_TextSet[TextKeys.MsgOperationAbortedByUser];
                }
                else
                {
                    // Error.

                    l_ErrMsg = f_TextSet[TextKeys.MoveDataFilesDlgMsgErrWhileMoving];
                }

                goto ERROR;
            }

            if (false == Software.s_BootConfig.TakeOverNewApplicationDataDirectoryAndSave())
            {
                l_ErrMsg = f_TextSet[TextKeys.MsgErrWhileWritingBootConfig];

                goto ERROR;
            }

            f_MoveCntxt.CloseSourceFiles(true);
            f_MoveCntxt.CloseDestinationFiles(false);
            f_MoveCntxt.NotifySuccess();

            return;

            ERROR:

            f_MoveCntxt.CloseSourceFiles(false);
            f_MoveCntxt.CloseDestinationFiles(true);
            f_MoveCntxt.NotifyError(l_ErrMsg);
        }

        private class MoveContext : ControllerContext
        {
            private const uint c_MoveFilesRecommendedExtraSpace = (8 * 1024 * 1024);

            private readonly string[] f_SrcFileNames;
            private readonly string[] f_DestFileNames;
            public readonly FHandle[] f_SrcFiles;
            public readonly FHandle[] f_DestFiles;
            public readonly ulong f_TotalSize;

            public bool f_Run;
            public ulong f_Done;

            private string f_ErrMsg;

            private MoveContext(string[] a_SrcFileNames,
                                string[] a_DestFileNames,
                                FHandle[] a_SrcFiles,
                                FHandle[] a_DestFiles,
                                ulong a_TotalSize) :
                           base(true)
            {
                f_SrcFileNames = a_SrcFileNames;
                f_DestFileNames = a_DestFileNames;
                f_SrcFiles = a_SrcFiles;
                f_DestFiles = a_DestFiles;
                f_TotalSize = a_TotalSize;

                f_Run = true;
            }

            public static MoveContext Create(TextSet a_TextSet,
                                             string a_SrcDir,
                                             string a_DestDir,
                                             out string a_ErrMsg)
            {
                string[] l_SrcFileNames = Software.BuildAppDataFileNames(a_SrcDir);
                FHandle[] l_SrcFiles = new FHandle[c_FilesToCopy];
                ulong l_TotalSize = 0;

                for (int i = 0; i < c_FilesToCopy; i++)
                {
                    Exception l_Exc;
                    l_SrcFiles[i] = FHandle.OpenR(l_SrcFileNames[i], false, System.IO.FileOptions.Asynchronous, 0, FHandle.ThreadSafetyMode.NeedSynchronization, out l_Exc);
                    ulong l_FileSize;

                    if (null == l_SrcFiles[i] ||
                        false == l_SrcFiles[i].GetSize(out l_FileSize))
                    {
                        CloseFiles(l_SrcFileNames, l_SrcFiles, false);

                        a_ErrMsg = a_TextSet.BuildFileErrMsg(TextKeys.MoveDataFilesDlgMsgErrPreparingFiles, l_SrcFileNames[i], TextKeys.MsgFileErrAccessFailed);

                        return null;
                    }

                    l_TotalSize += l_FileSize;
                }

                ulong l_AvailSpace;
                if (true == FHandle.GetAvailableVolumeSpace(a_DestDir, out l_AvailSpace) &&
                    l_AvailSpace < (l_TotalSize + c_MoveFilesRecommendedExtraSpace))
                {
                    CloseFiles(l_SrcFileNames, l_SrcFiles, false);

                    a_ErrMsg = a_TextSet[TextKeys.MoveDataFilesDlgMsgErrInsuffSpace];

                    return null;
                }

                string[] l_DestFileNames = Software.BuildAppDataFileNames(a_DestDir);
                FHandle[] l_DestFiles = new FHandle[c_FilesToCopy];

                for (int i = 0; i < c_FilesToCopy; i++)
                {
                    Exception l_Exc;
                    l_DestFiles[i] = FHandle.OpenRWRITE(typeof(MoveContext), l_DestFileNames[i], System.IO.FileMode.CreateNew, System.IO.FileOptions.Asynchronous, 0, FHandle.ThreadSafetyMode.NeedSynchronization, out l_Exc);

                    if (null == l_DestFiles[i])
                    {
                        CloseFiles(l_SrcFileNames, l_SrcFiles, false);
                        CloseFiles(l_DestFileNames, l_DestFiles, true);

                        a_ErrMsg = a_TextSet.BuildFileErrMsg(TextKeys.MoveDataFilesDlgMsgErrPreparingFiles, l_DestFileNames[i], TextKeys.MsgFileErrCreationFailed);

                        return null;
                    }
                }

                a_ErrMsg = null;

                return new MoveContext(l_SrcFileNames, l_DestFileNames, l_SrcFiles, l_DestFiles, l_TotalSize);
            }

            public bool CloseSourceFiles(bool a_Delete)
            {
                return CloseFiles(f_SrcFileNames, f_SrcFiles, a_Delete);
            }

            public bool CloseDestinationFiles(bool a_Delete)
            {
                return CloseFiles(f_DestFileNames, f_DestFiles, a_Delete);
            }

            public override void Abort()
            {
                f_Run = false;
            }

            public void NotifyError(string a_ErrMsg)
            {
                f_ErrMsg = a_ErrMsg;
                SetFinishedBase();
            }

            public void NotifySuccess()
            {
                SetFinishedBase();
            }

            public bool QueryVolatileFields(out ulong a_Done,
                                            out string a_ErrMsg)
            {
                a_Done = f_Done;
                a_ErrMsg = f_ErrMsg;
                bool l_Finished = p_Finished;
                Util.MemoryBarrier();

                return l_Finished;
            }

            private static bool CloseFiles(string[] a_FileNames,
                                           FHandle[] a_Files,
                                           bool a_Delete)
            {
                // Each element of a_Files may be null.

                bool l_Succ = true;

                for (int i = 0; i < c_FilesToCopy; i++)
                {
                    FHandle l_File = a_Files[i];

                    if (null != l_File)
                    {
                        l_File.Close(false);
                    }

                    if (true == a_Delete &&
                        false == FHandle.Delete(a_FileNames[i]))
                    {
                        l_Succ = false;
                    }
                }

                return l_Succ;
            }
        }
    }

    class OwnerDrawContext
    {
        private readonly static OwnerDrawContext s_IncompletePassive0f;
        private readonly static OwnerDrawContext s_IncompleteActive0f;
        private readonly static OwnerDrawContext s_MissingActive1f;
        private readonly static OwnerDrawContext s_CompletePassive1f;
        private readonly static OwnerDrawContext s_CompleteActive1f;

        public readonly bool f_DrawProgressBar;
        public readonly bool f_DrawFrame;

        // If f_DrawProgressBar is true.
        public readonly ProgressBarStyles f_ProgressBarStyle;
        public readonly float f_Progress;

        // If f_DrawProgressBar is false.
        public Bitmap f_Map;

        public ulong f_CompareWeight;

        static OwnerDrawContext()
        {
            s_IncompletePassive0f = new OwnerDrawContext(ProgressBarStyles.IncompletePassive, 0f);
            s_IncompleteActive0f = new OwnerDrawContext(ProgressBarStyles.Active, 0f);
            s_MissingActive1f = new OwnerDrawContext(ProgressBarStyles.Missing, 1f);
            s_CompletePassive1f = new OwnerDrawContext(ProgressBarStyles.CompletePassive, 1f);
            s_CompleteActive1f = new OwnerDrawContext(ProgressBarStyles.Active, 1f);
        }

        public OwnerDrawContext(Bitmap a_Map)
        {
            f_DrawFrame = true;
            f_Map = a_Map;
            f_CompareWeight = 0;
        }

        private OwnerDrawContext(ProgressBarStyles a_ProgressBarStyle,
                                 float a_Progress)
        {
            f_DrawProgressBar = true;
            f_DrawFrame = true;
            f_ProgressBarStyle = a_ProgressBarStyle;
            f_Progress = a_Progress;
            f_CompareWeight = (uint)(a_Progress * 10000);

            GC.SuppressFinalize(this);
        }

        public static OwnerDrawContext CreateIncomplFileProgressBarContext(bool a_IsActive,
                                                                       float a_Progress)
        {
            if (0 >= a_Progress)
            {
                return (false == a_IsActive ? s_IncompletePassive0f : s_IncompleteActive0f);
            }

            // We'll have to create a custom instance.

            ProgressBarStyles l_Style = (false == a_IsActive ? ProgressBarStyles.IncompletePassive : ProgressBarStyles.Active);

            return new OwnerDrawContext(l_Style, a_Progress);
        }

        public static OwnerDrawContext CreateMissingFileProgressBarContext()
        {
            return s_MissingActive1f;
        }

        public static OwnerDrawContext CreateComplFileProgressBarContext(bool a_IsActive)
        {
            return (false == a_IsActive ? s_CompletePassive1f : s_CompleteActive1f);
        }

        ~OwnerDrawContext()
        {
            if (null != f_Map)
            {
                f_Map.Dispose();
            }
        }

        public enum ProgressBarStyles : byte
        {
            IncompletePassive = 0,
            Missing = 1,
            CompletePassive = 2,
            Active = 3,
        }
    }

    class BarDrawer : IDisposable
    {
        private const int c_AvailLevels = 11;

        private static readonly StringFormat s_StrFrmt;
        private static readonly Brush[] s_AvailBrushes;

        // Used for the progress bar.
        private readonly Font f_Fnt;
        private readonly Brush f_FontBrush;
        // The following two arrays match GridViewDrawContext.ProgressBarStyles.
        private readonly Brush[] f_ProgressBarForeBrushes;
        private readonly Brush[] f_LightProgressBarForeBrushes;
        // This is a quick fix for thread safety of s_SolidComplBarColors.
        [ThreadStatic]
        private static BarColors s_SolidComplBarColors;
        private BarColors f_OverlayComplBarColors;

        // Used for the availability bar.
        public readonly int f_MaxAvailWidth;

        public Color p_SolidClearColor
        {
            get { EnsureSolidComplBarColorsCreation(); return s_SolidComplBarColors.f_ClearColor; }
        }

        public Color p_SolidDoneColor
        {
            get { EnsureSolidComplBarColorsCreation(); return s_SolidComplBarColors.f_DoneColor; }
        }

        public Brush p_OverlayClearBrush
        {
            // Return can be null.
            get { return (0 == f_OverlayComplBarColors.f_ClearColor.A ? null : f_OverlayComplBarColors.f_ClearBrush); }
        }

        public Brush p_OverlayDoneBrush
        {
            // Return can be null.
            get { return (0 == f_OverlayComplBarColors.f_DoneColor.A ? null : f_OverlayComplBarColors.f_DoneBrush); }
        }

        static BarDrawer()
        {
            s_StrFrmt = CreateCenterStringFormat();

            s_AvailBrushes = new Brush[c_AvailLevels];
            s_AvailBrushes[0] = (Brush)Brushes.Red.Clone();

            float l_Perc = .1f;
            for (int i = 1; i < c_AvailLevels; i++, l_Perc += .1f)
            {
                s_AvailBrushes[i] = new SolidBrush(MixColors(Color.White, Color.Blue, l_Perc));
            }
        }

        public BarDrawer(Font a_FntTemplate,
                         Color a_FontCol,
                         Color a_PassiveIncomplForeBrush,
                         Color a_MissingForeBrush,
                         Color a_PassiveComplForeBrush,
                         Color a_ActiveForeCol,
                         int a_MaxAvailWidth)
        {
            const byte l_ProgressBarStyleCount = 4;
            const byte l_DefaultDelta = 15;
            const byte l_BigDelta = 45;

            f_Fnt = new Font(a_FntTemplate.FontFamily, a_FntTemplate.SizeInPoints, FontStyle.Bold);
            f_FontBrush = new SolidBrush(a_FontCol);
            f_ProgressBarForeBrushes = new Brush[l_ProgressBarStyleCount];
            f_LightProgressBarForeBrushes = new Brush[l_ProgressBarStyleCount];
            CreateBrushes(0, a_PassiveIncomplForeBrush, l_DefaultDelta);
            CreateBrushes(1, a_MissingForeBrush, l_BigDelta);
            CreateBrushes(2, a_PassiveComplForeBrush, l_DefaultDelta);
            CreateBrushes(3, a_ActiveForeCol, l_DefaultDelta);
            f_OverlayComplBarColors = new BarColors(Color.FromArgb(170, Color.Black), Color.FromArgb(170, 195, 105, 0), Color.FromArgb(0, Color.White));
            f_MaxAvailWidth = a_MaxAvailWidth;
        }

        public void Dispose()
        {
            f_Fnt.Dispose();
            f_FontBrush.Dispose();

            foreach (Brush l_Brush in f_ProgressBarForeBrushes)
            {
                l_Brush.Dispose();
            }

            foreach (Brush l_Brush in f_LightProgressBarForeBrushes)
            {
                l_Brush.Dispose();
            }

            f_OverlayComplBarColors.Dispose();
        }

        private static void EnsureSolidComplBarColorsCreation()
        {
            // Create the thread static struct.

            if (false == s_SolidComplBarColors.IsValid())
            {
                s_SolidComplBarColors = new BarColors(Color.White, Color.Orange, Color.Blue);
            }
        }

        private void CreateBrushes(int a_Idx,
                                   Color a_Col,
                                   int a_Delta)
        {
            f_ProgressBarForeBrushes[a_Idx] = new SolidBrush(a_Col);
            f_LightProgressBarForeBrushes[a_Idx] = new SolidBrush(ChangeColor(a_Col, a_Delta));
        }

        private static Color ChangeColor(Color a_Col,
                                         int a_Delta)
        {
            return Color.FromArgb(Math.Max(0, Math.Min(byte.MaxValue, a_Col.R + a_Delta)),
                                  Math.Max(0, Math.Min(byte.MaxValue, a_Col.G + a_Delta)),
                                  Math.Max(0, Math.Min(byte.MaxValue, a_Col.B + a_Delta)));
        }

        public void DrawProgressBar(TextSet a_TextSet,
                                    Graphics a_Graphics,
                                    Rectangle a_DestRect,
                                    OwnerDrawContext.ProgressBarStyles a_BarStyle,
                                    float a_Progress)
        {
            // Switch to highest text rendering quality.
            System.Drawing.Text.TextRenderingHint l_StoredHint = a_Graphics.TextRenderingHint;
            RectangleF l_StoredClip = a_Graphics.ClipBounds;
            {
                a_Graphics.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;
                int l_LeftWidth = Util.RoundDoubleToInt(a_DestRect.Width * a_Progress);
                Rectangle l_LeftRect = new Rectangle(a_DestRect.X, a_DestRect.Y, l_LeftWidth, a_DestRect.Height);
                Rectangle l_RightRect = new Rectangle(a_DestRect.X + l_LeftWidth, a_DestRect.Y, a_DestRect.Width - l_LeftWidth, a_DestRect.Height);
                string l_ProgressStr = a_TextSet.PercStr(a_Progress);

                if (0 < l_LeftWidth)
                {
                    int l_Idx = (int)a_BarStyle;
                    Brush l_ForeBrush = f_ProgressBarForeBrushes[l_Idx];
                    Brush l_LightForeBrush = f_LightProgressBarForeBrushes[l_Idx];
                    int l_TriangleY = ((l_LeftRect.Height / 8) * 3);

                    a_Graphics.SetClip(l_LeftRect);
                    a_Graphics.FillRectangle(l_ForeBrush, l_LeftRect);
                    a_Graphics.FillPolygon(l_LightForeBrush, new Point[] { l_LeftRect.Location, new Point(l_LeftRect.X, l_LeftRect.Y + l_TriangleY), new Point(l_RightRect.Right, l_RightRect.Top) });
                    a_Graphics.DrawString(l_ProgressStr, f_Fnt, SystemBrushes.Window, a_DestRect, s_StrFrmt);
                }

                if (0 < l_RightRect.Width)
                {
                    a_Graphics.SetClip(l_RightRect);
                    a_Graphics.FillRectangle(SystemBrushes.Window, l_RightRect);
                    a_Graphics.DrawString(l_ProgressStr, f_Fnt, f_FontBrush, a_DestRect, s_StrFrmt);
                }
            }
            a_Graphics.TextRenderingHint = l_StoredHint;
            a_Graphics.SetClip(l_StoredClip);
        }

        public Bitmap DrawAvailabilityBarImageScaled(NetworkAvailability a_VolatileAvail,
                                                     int a_PixelHeight)
        {
            // a_VolatileCompl can safely be modified while processed
            // by this routine.

            uint l_Bytes = (uint)a_VolatileAvail.f_Arr.Length;
            Size l_Size = new Size(Math.Max(1, f_MaxAvailWidth), Math.Max(1, a_PixelHeight));
            Bitmap l_Img;
            bool l_Shrink;
            int l_Factor = CalcFactor(l_Bytes, ref l_Size, out l_Img, out l_Shrink);

            Graphics l_G = Graphics.FromImage(l_Img);
            {
                uint l_NonzeroItems = a_VolatileAvail.p_NonzeroItems;
                int l_MaxAvail = (int)a_VolatileAvail.p_Contributors;
                Util.MemoryBarrier();

                if (0 == l_MaxAvail)
                {
                    l_G.Clear(Color.White);
                }
                else if (0 == l_NonzeroItems)
                {
                    l_G.Clear(Color.Red);
                }
                else if (1 == l_MaxAvail &&
                         l_Bytes == l_NonzeroItems)
                {
                    l_G.Clear(Color.Blue);
                }
                else
                {
                    int t = 0;
                    for (int i = 0; i < l_Bytes; i++, t += l_Factor)
                    {
                        Brush l_Brush = GetAvailabilityBrush(a_VolatileAvail.f_Arr[i], l_MaxAvail);

                        l_G.FillRectangle(l_Brush, new Rectangle(t, 0, l_Factor, a_PixelHeight));
                    }
                }
            }
            l_G.Dispose();

            if (true == l_Shrink)
            {
                // Shrink this large image, to make it easier to handle.
                ShrinkBitmap(ref l_Img, l_Size);
            }

            return l_Img;
        }

        public Bitmap DrawCompletenessBarImageScaledOnBackground(BitMap a_VolatileCompl,
                                                                 Dictionary<uint, ElmKey> a_HotBits,
                                                                 int a_PixelHeight,
                                                                 Bitmap a_Background)
        {
            Size l_Size = new Size(Math.Max(1, f_MaxAvailWidth), Math.Max(1, a_PixelHeight));
            Bitmap l_BarMap = DrawCompletenessBarImageScaled(a_VolatileCompl, a_HotBits, ref l_Size, ref f_OverlayComplBarColors);
            Bitmap l_FinalMap = new Bitmap(l_BarMap.Width, l_BarMap.Height);
            Graphics l_G = Graphics.FromImage(l_FinalMap);
            {
                l_G.DrawImage(a_Background, new Rectangle(0, 0, l_FinalMap.Width, l_FinalMap.Height));
                l_G.DrawImageUnscaled(l_BarMap, Point.Empty);
            }
            l_BarMap.Dispose();
            l_G.Dispose();

            return l_FinalMap;
        }

        public Bitmap DrawCompletenessBarImageScaled(BitMap a_VolatileCompl,
                                                     Dictionary<uint, ElmKey> a_HotBits,
                                                     int a_PixelHeight)
        {
            EnsureSolidComplBarColorsCreation();

            Size l_Size = new Size(Math.Max(1, f_MaxAvailWidth), Math.Max(1, a_PixelHeight));

            return DrawCompletenessBarImageScaled(a_VolatileCompl, a_HotBits, ref l_Size, ref s_SolidComplBarColors);
        }

        private static Bitmap DrawCompletenessBarImageScaled(BitMap a_VolatileCompl,
                                                             Dictionary<uint, ElmKey> a_HotBits,
                                                             ref Size a_Size,
                                                             ref BarColors a_ComplBarColors)
        {
            // a_VolatileCompl can safely be modified while processed
            // by this routine. a_HotBits may be null.

            uint l_Bits = a_VolatileCompl.f_BitCount;
            Bitmap l_Img;
            bool l_Shrink;
            int l_Factor = CalcFactor(l_Bits, ref a_Size, out l_Img, out l_Shrink);

            // This is just a performance hint.
            bool l_DoneColHasNonzeroAlpha = (0 != a_ComplBarColors.f_DoneColor.A);

            Graphics l_G = Graphics.FromImage(l_Img);
            {
                if (true == a_VolatileCompl.p_AllSet)
                {
                    if (true == l_DoneColHasNonzeroAlpha)
                    {
                        l_G.Clear(a_ComplBarColors.f_DoneColor);
                    }
                }
                else if (true == a_VolatileCompl.p_AllClear &&
                         (null == a_HotBits ||
                          0 == a_HotBits.Count))
                {
                    l_G.Clear(a_ComplBarColors.f_ClearColor);
                }
                else
                {
                    int t = 0;
                    for (uint i = 0; i < l_Bits; i++, t += l_Factor)
                    {
                        Brush l_Brush;

                        if (false == a_VolatileCompl.Get(i))
                        {
                            if (null == a_HotBits ||
                                false == a_HotBits.ContainsKey(i))
                            {
                                l_Brush = a_ComplBarColors.f_ClearBrush;
                            }
                            else
                            {
                                l_Brush = a_ComplBarColors.f_InProgressBrush;
                            }
                        }
                        else
                        {
                            if (false == l_DoneColHasNonzeroAlpha)
                            {
                                continue;
                            }

                            l_Brush = a_ComplBarColors.f_DoneBrush;
                        }

                        l_G.FillRectangle(l_Brush, new Rectangle(t, 0, l_Factor, a_Size.Height));
                    }
                }
            }
            l_G.Dispose();

            if (true == l_Shrink)
            {
                // Shrink this large image, to make it easier to handle.
                ShrinkBitmap(ref l_Img, a_Size);
            }

            return l_Img;
        }

        private static void ShrinkBitmap(ref Bitmap a_Map,
                                         Size a_Size)
        {
            Bitmap l_OldMap = a_Map;
            {
                a_Map = new Bitmap(a_Map, a_Size);
            }
            l_OldMap.Dispose();
        }

        private static int CalcFactor(uint a_DataWith,
                                      ref Size a_Size,
                                      out Bitmap a_Bitmap,
                                      out bool a_Shrink)
        {
            if ((uint)a_Size.Width <= a_DataWith)
            {
                // In this case we draw a huge graphic and shrink it
                // later.

                a_Bitmap = new Bitmap(checked((int)a_DataWith), a_Size.Height);
                a_Shrink = ((uint)a_Size.Width < a_DataWith);

                return 1;
            }

            int l_Factor = ((a_Size.Width / (int)a_DataWith) +
                            (0 == (a_Size.Width % (int)a_DataWith) ? 0 : 1));

            a_Bitmap = new Bitmap(checked((int)(l_Factor * a_DataWith)), a_Size.Height);
            a_Shrink = false;

            return l_Factor;
        }

        private static Brush GetAvailabilityBrush(int a_Avail,
                                                  int a_Max)
        {
            // This routine is robust enough to process invalid input
            // as possible from a volatile source (e.g. a_Avail > a_Max).

            if (0 >= a_Avail)
            {
                return s_AvailBrushes[0];
            }

            // This check also safes us from a division by zero.
            if (a_Max <= a_Avail)
            {
                return s_AvailBrushes[c_AvailLevels - 1];
            }

            int l_Idx = ((int)(((float)a_Avail / a_Max) * (c_AvailLevels - 2)) + 1);

            return s_AvailBrushes[l_Idx];
        }

        public static void DrawFrame(Graphics a_G,
                                     ref Rectangle a_Bounds,
                                     Pen a_Pen)
        {
            // a_Bounds describes the inner rectangle on return.

            a_Bounds.Width--;
            a_Bounds.Height--;

            if (1 > a_Bounds.Width ||
                1 > a_Bounds.Height)
            {
                return;
            }

            // Draw a full rectangle.
            //a_G.DrawRectangle(a_Pen, a_Bounds);

            // Draw a rectangle with soft edges.
            a_G.DrawLine(a_Pen, a_Bounds.X + 1, a_Bounds.Y, a_Bounds.Right - 1, a_Bounds.Y);
            a_G.DrawLine(a_Pen, a_Bounds.X + 1, a_Bounds.Bottom, a_Bounds.Right - 1, a_Bounds.Bottom);
            a_G.DrawLine(a_Pen, a_Bounds.X, a_Bounds.Y + 1, a_Bounds.X, a_Bounds.Bottom - 1);
            a_G.DrawLine(a_Pen, a_Bounds.Right, a_Bounds.Y + 1, a_Bounds.Right, a_Bounds.Bottom - 1);

            a_Bounds.X++;
            a_Bounds.Y++;
            a_Bounds.Width--;
            a_Bounds.Height--;
        }

        private static Color MixColors(Color a_Col1,
                                       Color a_Col2,
                                       float a_Perc)
        {
            return Color.FromArgb(MixBytes(a_Col1.R, a_Col2.R, a_Perc),
                                  MixBytes(a_Col1.G, a_Col2.G, a_Perc),
                                  MixBytes(a_Col1.B, a_Col2.B, a_Perc));
        }

        private static byte MixBytes(int a_A,
                                     int a_B,
                                     float a_Perc)
        {
            int l_Diff = (a_B - a_A);
            int l_Num = (a_A + Util.RoundDoubleToInt(a_Perc * l_Diff));

            return (byte)Math.Max(0, Math.Min(byte.MaxValue, l_Num));
        }

        public static StringFormat CreateCenterStringFormat()
        {
            StringFormat l_StrFrmt = (StringFormat)StringFormat.GenericDefault.Clone();
            l_StrFrmt.Alignment = StringAlignment.Center;
            l_StrFrmt.LineAlignment = StringAlignment.Center;
            l_StrFrmt.FormatFlags |= StringFormatFlags.NoWrap;

            return l_StrFrmt;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct BarColors : IDisposable
        {
            // Don't use the struct default constructor.

            public readonly Color f_ClearColor;
            public readonly Color f_InProgressColor;
            public readonly Color f_DoneColor;

            public readonly Brush f_ClearBrush;
            public readonly Brush f_InProgressBrush;
            public readonly Brush f_DoneBrush;

            public BarColors(Color a_ClearColor,
                             Color a_InProgressColor,
                             Color a_DoneColor)
            {
                f_ClearColor = a_ClearColor;
                f_InProgressColor = a_InProgressColor;
                f_DoneColor = a_DoneColor;

                f_ClearBrush = new SolidBrush(f_ClearColor);
                f_InProgressBrush = new SolidBrush(f_InProgressColor);
                f_DoneBrush = new SolidBrush(f_DoneColor);
            }

            public void Dispose()
            {
                f_ClearBrush.Dispose();
                f_InProgressBrush.Dispose();
                f_DoneBrush.Dispose();
            }

            public bool IsValid()
            {
                return (null != f_ClearBrush);
            }
        }
    }

    static class SizeFormatter
    {
        public const ushort c_SIGranularity = 1000;
        public const ushort c_IECGranularity = 1024;
        private const byte c_Levels = 7;
        private const byte c_AboveByteLevels = (c_Levels - 1);
        private const string c_ByteSymbol = "B";

        private static readonly string[] s_SINames;
        private static readonly string[] s_IECNames;
        private static readonly ulong[] s_SI;
        private static readonly ulong[] s_IEC;

        private static TextSet s_TextSet;
        private static string s_BytesStr;
        private static string s_DefaultSizeFormat;
        private static string s_LongSizeFormat;
        private static string s_SpeedUnitFormat;
        private static string s_UlDlRatioFormat;
        private static string s_OnlyDlStr;
        private static string s_OnlyUlStr;
        private static string s_OneStr;

        static SizeFormatter()
        {
            // The SI style has a lowercase k.
            s_SINames = new string[] { c_ByteSymbol, "kB", "MB", "GB", "TB", "PB", "EB" };
            s_IECNames = new string[] { c_ByteSymbol, "KiB", "MiB", "GiB", "TiB", "PiB", "EiB" };
            s_SI = new ulong[c_AboveByteLevels];
            s_IEC = new ulong[c_AboveByteLevels];

            s_SI[0] = c_SIGranularity;
            s_IEC[0] = c_IECGranularity;

            for (int i = 1; i < c_AboveByteLevels; i++)
            {
                s_SI[i] = (s_SI[i - 1] * s_SI[0]);
                s_IEC[i] = (s_IEC[i - 1] * s_IEC[0]);
            }
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            s_TextSet = a_TextSet;
            s_BytesStr = s_TextSet[TextKeys.BytePlurStr];
            s_DefaultSizeFormat = s_TextSet[TextKeys.DefaultSizeFormat];
            s_LongSizeFormat = s_TextSet[TextKeys.LongSizeFormat];
            s_SpeedUnitFormat = s_TextSet[TextKeys.SpeedUnitFormat];
            s_UlDlRatioFormat = s_TextSet[TextKeys.UlDlRatioFormat];
            s_OnlyDlStr = s_TextSet[TextKeys.OnlyDlStr];
            s_OnlyUlStr = s_TextSet[TextKeys.OnlyUlStr];
            s_OneStr = s_TextSet.NumStr(1u);
        }

        private static string[] GetNameArray(bool a_SI)
        {
            return (false == a_SI ? s_IECNames : s_SINames);
        }

        public static string GetUnitStr(bool a_SI,
                                        int a_Magnitude)
        {
            return GetNameArray(a_SI)[a_Magnitude];
        }

        public static string GetUnitStr(bool a_SI,
                                        bool a_Speed,
                                        int a_Magnitude)
        {
            return WrapUnitStrWithSpeed(a_Speed, GetUnitStr(a_SI, a_Magnitude));
        }

        private static string WrapUnitStrWithSpeed(bool a_Speed,
                                                   string a_UnitStr)
        {
            if (false == a_Speed)
            {
                return a_UnitStr;
            }

            return string.Format(s_SpeedUnitFormat, a_UnitStr);
        }

        public static void FillComboBox(bool a_SI,
                                        ComboBox a_ComboBox)
        {
            a_ComboBox.Items.Clear();
            a_ComboBox.Items.Add(s_BytesStr);

            // Do not include the highest two levels, since they are
            // not meaningful in file sharing currently.
            string[] l_Names = GetNameArray(a_SI);
            for (int i = 1; i < (c_Levels - 2); i++)
            {
                a_ComboBox.Items.Add(l_Names[i]);
            }
        }

        public static ushort GetGranularity(bool a_SI)
        {
            return (false == a_SI ? c_IECGranularity : c_SIGranularity);
        }

        public static ulong GetSize(bool a_SI,
                                    int a_Magnitude)
        {
            return (false == a_SI ? s_IEC[a_Magnitude] : s_SI[a_Magnitude]);
        }

        public static int FindMagnitude(bool a_SI,
                                        ulong a_Bytes)
        {
            ulong[] l_Junk;
            return FindMagnitude(a_SI, a_Bytes, out l_Junk);
        }

        private static int FindMagnitude(bool a_SI,
                                         ulong a_Bytes,
                                         out ulong[] a_Arr)
        {
            a_Arr = (false == a_SI ? s_IEC : s_SI);

            for (int i = 0; i < c_AboveByteLevels; i++)
            {
                if (a_Arr[i] > a_Bytes)
                {
                    return i;
                }
            }

            return c_AboveByteLevels;
        }

        public static string BuildKiloStr(bool a_SI,
                                          ulong a_Bytes)
        {
            ulong l_Num = (a_Bytes / (false == a_SI ? c_IECGranularity : c_SIGranularity));
            string l_Unit = (false == a_SI ? "K" : "k");

            return string.Format(s_DefaultSizeFormat, l_Num, l_Unit);
        }

        public static string BuildSizeStr(bool a_SI,
                                          bool a_Long,
                                          bool a_Speed,
                                          bool a_WithDecimals,
                                          ulong a_Bytes)
        {
            ulong[] l_Arr;
            int l_Magnitude = FindMagnitude(a_SI, a_Bytes, out l_Arr);

            if (0 == l_Magnitude)
            {
                return BuildByteSizeStr(a_Speed, a_Bytes, false == a_Long ? c_ByteSymbol : s_BytesStr);
            }

            ulong l_NormBytes = (2 > l_Magnitude ? a_Bytes : (a_Bytes / l_Arr[l_Magnitude - 2]));
            double l_Val;
            int l_Prec;

            if (false == a_WithDecimals)
            {
                // Always round down.
                l_Prec = 0;
                l_Val = (l_NormBytes / l_Arr[0]);
            }
            else
            {
                l_Prec = (false == a_Speed ? 2 : 1);
                l_Val = Math.Round((double)l_NormBytes / l_Arr[0], l_Prec);

                // Remove unneeded precision.

                if (false == a_Speed &&
                    0 == ((int)(l_Val * 100) % 10))
                {
                    l_Prec = 1;
                }

                if (1 == l_Prec &&
                    0 == ((int)(l_Val * 10) % 10))
                {
                    l_Prec = 0;
                }
            }

            string l_ValStr = s_TextSet.FloatStr(l_Val, l_Prec);
            string l_MagStr = GetUnitStr(a_SI, a_Speed, l_Magnitude);
            string l_ResStr = string.Format(s_DefaultSizeFormat, l_ValStr, l_MagStr);

            if (true == a_Long)
            {
                l_ResStr = string.Format(s_LongSizeFormat, l_ResStr, BuildByteSizeStr(a_Speed, a_Bytes, s_BytesStr));
            }

            return l_ResStr;
        }

        private static string BuildByteSizeStr(bool a_Speed,
                                               ulong a_Bytes,
                                               string a_ByteStr)
        {
            return string.Format(s_DefaultSizeFormat, s_TextSet.NumStr(a_Bytes), WrapUnitStrWithSpeed(a_Speed, a_ByteStr));
        }

        public static string BuildUlDlRatioStr(float a_UlDlRatio)
        {
            if (1 <= a_UlDlRatio)
            {
                if (1f == a_UlDlRatio)
                {
                    return string.Format(s_UlDlRatioFormat, s_OneStr, s_OneStr);
                }

                if (float.MaxValue == a_UlDlRatio)
                {
                    return s_OnlyUlStr;
                }

                return string.Format(s_UlDlRatioFormat, s_TextSet.FloatStr(a_UlDlRatio), s_OneStr);
            }

            if (float.MinValue == a_UlDlRatio)
            {
                return s_OnlyDlStr;
            }

            return string.Format(s_UlDlRatioFormat, s_OneStr, s_TextSet.FloatStr(1 / a_UlDlRatio));
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TimeStrHelper
    {
        public static string s_NeverStr;
        public static string s_NowStr;
        public static string s_LongerThanMonthStr;
        private static Period[] s_Periods;
        private static string s_NegStrFormat;
        private static string s_Format;
        private static string s_SpaceStr;

        private readonly Time f_RefTime;
        private readonly StringBuilder f_Bld;

        private TimeStrHelper(Time a_RefTime,
                              StringBuilder a_Bld)
        {
            f_RefTime = a_RefTime;
            f_Bld = a_Bld;
        }

        public static void ApplyTextSet(TextSet a_TextSet)
        {
            Period[] l_Periods = new Period[5];
            l_Periods[0] = new Period(Time.c_Second, a_TextSet[TextKeys.TimeStrSecs]);
            l_Periods[1] = new Period(Time.c_SecsPerMinute, a_TextSet[TextKeys.TimeStrMins]);
            l_Periods[2] = new Period(Time.c_SecsPerHour, a_TextSet[TextKeys.TimeStrHours]);
            l_Periods[3] = new Period(Time.c_SecsPerDay, a_TextSet[TextKeys.TimeStrDays]);
            l_Periods[4] = new Period(7 * Time.c_SecsPerDay, a_TextSet[TextKeys.TimeStrWeeks]);
            s_Periods = l_Periods;

            s_NeverStr = a_TextSet[TextKeys.TimeStrNever];
            s_NowStr = a_TextSet[TextKeys.TimeStrNow];
            s_LongerThanMonthStr = a_TextSet[TextKeys.TimeStrLongerThanMonth];
            s_NegStrFormat = a_TextSet[TextKeys.TimeStrNegFormat];
            s_Format = a_TextSet[TextKeys.TimeStrFormat];
            s_SpaceStr = a_TextSet[TextKeys.TimeStrSpace];
        }

        public static void Create(Time a_RefTime,
                                  out TimeStrHelper a_Dest)
        {
            a_Dest = new TimeStrHelper(a_RefTime, new StringBuilder(16));
        }

        public static void Create(out TimeStrHelper a_Dest)
        {
            Create(new Time(), out a_Dest);
        }

        public string GetRelativeTimeString(Time a_Time,
                                            string a_FallbackStr)
        {
            if (false == a_Time.IsSet())
            {
                return a_FallbackStr;
            }

            long l_SecsAgo = f_RefTime.SecondsSince(a_Time);

            return GetTimeString(l_SecsAgo);
        }

        public string GetTimeString(long a_Secs)
        {
            f_Bld.Length = 0;
            AppendTimeString(f_Bld, a_Secs);

            return f_Bld.ToString();
        }

        public static void AppendTimeString(StringBuilder a_Bld,
                                            long a_Secs)
        {
            ulong l_Secs;
            int l_Pos;

            if (0 <= a_Secs)
            {
                l_Secs = (ulong)a_Secs;
                l_Pos = 0;
            }
            else
            {
                if (long.MinValue == a_Secs)
                {
                    l_Secs = ((ulong)long.MaxValue + 1);
                }
                else
                {
                    l_Secs = (ulong)-a_Secs;
                }

                l_Pos = a_Bld.Length;
            }

            AppendTimeStringCore(a_Bld, l_Secs);

            if (0 > a_Secs)
            {
                int l_Len = (a_Bld.Length - l_Pos);
                string l_Str = a_Bld.ToString(l_Pos, l_Len);
                a_Bld.Remove(l_Pos, l_Len);
                a_Bld.AppendFormat(s_NegStrFormat, l_Str);
            }
        }

        private static void AppendTimeStringCore(StringBuilder a_Bld,
                                                 ulong a_Secs)
        {
            int l_Count = 0;

            for (int i = (s_Periods.Length - 1); 0 <= i; i--)
            {
                uint l_SecsPerPeriod = s_Periods[i].f_SecsPerPeriod;

                if (l_SecsPerPeriod <= a_Secs ||
                    0 == i)
                {
                    ulong l_FullPeriods = (a_Secs / l_SecsPerPeriod);
                    a_Secs %= l_SecsPerPeriod;

                    if (0 < l_Count)
                    {
                        a_Bld.Append(s_SpaceStr);
                    }

                    a_Bld.AppendFormat(s_Format, l_FullPeriods, s_Periods[i].f_PeriodStr);

                    l_Count++;

                    if (2 == l_Count ||
                        0 == a_Secs)
                    {
                        break;
                    }
                }
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct Period
        {
            public readonly uint f_SecsPerPeriod;
            public readonly string f_PeriodStr;

            public Period(uint a_SecsPerPeriod,
                          string a_PeriodStr)
            {
                f_SecsPerPeriod = a_SecsPerPeriod;

                const byte lc_MaxStrLen = 3;
                int l_StrLen = a_PeriodStr.Length;
                f_PeriodStr = (0 == l_StrLen ? "?" : (lc_MaxStrLen >= l_StrLen ? a_PeriodStr : a_PeriodStr.Substring(0, lc_MaxStrLen)));
            }
        }
    }
}

// End of file.
