/*
 * PROJECT: Nachtblitz File Sharing Client
 * PROJECT-URL: http://echelon.i2p/nachtblitz/
 * LICENSE: GNU Affero General Public License, version 3 (AGPLv3)
 * FILE: Nachtblitz-1.4.26-src\Nachtblitz\eDProt.cs
 * PURPOSE: eD2k client
 * VERSION: 1.4.26
 * DATE: April 2, 2015
 * COPYRIGHT: Copyright (C) 2012-2015 Stephan A. Rger
 *            (nachtblitz@i2pmail.org). All rights reserved.
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of version 3 - no other version - of the
 * GNU Affero General Public License as published by the Free
 * Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public
 * License along with this program (agpl-3.0.txt in the parent
 * directory). If not, see <http://www.gnu.org/licenses/>.
 *
 *         __           _     _   _     _ _ _         _  _  _
 *      /\ \ \__ _  ___| |__ | |_| |__ | (_) |_ ____ / || || |
 *     /  \/ / _` |/ __| '_ \| __| '_ \| | | __|_  / | || || |_
 *    / /\  / (_| | (__| | | | |_| |_) | | | |_ / /  | ||__   _|
 *    \_\ \/ \__,_|\___|_| |_|\__|_.__/|_|_|\__/___| |_(_) |_|
 *
 */

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;
using NB.Compression;
using NB.Content;
using NB.I2P;
using NB.Language;
using Support;
using Support.HashFunctions;

namespace NB.Network.eD2k
{
    enum Protocols : byte
    {
        NONE = 0,

        // OP_EDONKEYHEADER is in effect the very same
        // as OP_EDONKEYPROT.

        OP_EDONKEYHEADER = 1, // supported
        OP_EDONKEYPROT = 2, // supported
        OP_PACKEDPROT = 3, // supported (compressed OP_EMULEPROT)
        OP_EMULEPROT = 4, // supported
        OP_KADEMLIAHEADER = 5, // supported
        OP_KADEMLIAPACKEDPROT = 6, // supported (compressed OP_KADEMLIAHEADER)
        OP_MLDONKEYPROT = 7,

        PROT_NB_PLAIN = 154, // supported
        PROT_NB_COMPRESSED = 155, // supported
    }

    enum Opcodes : byte
    {
        NONE = 0,

        // OP_EDONKEYPROT stream
        OP_HELLO = 0x29,
        OP_SENDINGPART,
        OP_REQUESTPARTS,
        OP_FILEREQANSNOFIL,
        OP_END_OF_DOWNLOAD,
        OP_ASKSHAREDFILES,
        OP_ASKSHAREDFILESANSWER,
        OP_HELLOANSWER,
        OP_CHANGE_CLIENT_ID,
        OP_MESSAGE,
        OP_SETREQFILEID,
        OP_FILESTATUS,
        OP_HASHSETREQUEST,
        OP_HASHSETANSWER,
        OP_STARTUPLOADREQ,
        OP_ACCEPTUPLOADREQ,
        OP_CANCELTRANSFER,
        OP_OUTOFPARTREQS,
        OP_REQUESTFILENAME,
        OP_REQFILENAMEANSWER,
        OP_CHANGE_SLOT,
        OP_QUEUERANK,
        OP_ASKSHAREDDIRS,
        OP_ASKSHAREDFILESDIR,
        OP_ASKSHAREDDIRSANS,
        OP_ASKSHAREDFILESDIRANS,
        OP_ASKSHAREDDENIEDANS,

        // OP_EMULEPROT stream
        OP_EMULEINFO = 0x44,
        OP_EMULEINFOANSWER,
        OP_COMPRESSEDPART,
        OP_QUEUERANKING,
        OP_FILEDESC,
        OP_VERIFYUPSREQ,
        OP_VERIFYUPSANSWER,
        OP_UDPVERIFYUPREQ,
        OP_UDPVERIFYUPA,
        OP_REQUESTSOURCES,
        OP_ANSWERSOURCES,
        OP_PUBLICKEY,
        OP_SIGNATURE,
        OP_SECIDENTSTATE,
        OP_REQUESTPREVIEW,
        OP_PREVIEWANSWER,
        OP_MULTIPACKET,
        OP_MULTIPACKETANSWER,
        OP_PEERCACHE_QUERY,
        OP_PEERCACHE_ANSWER,
        OP_PEERCACHE_ACK,
        OP_PUBLICIP_REQ,
        OP_PUBLICIP_ANSWER,
        OP_CALLBACK,
        OP_AICHREQUEST,
        OP_AICHANSWER,
        OP_AICHFILEHASHANS,
        OP_AICHFILEHASHREQ,

        OP_LEGACY_START = 0x80, // OP_NOOP
        OP_LEGACY_END = 0x8B, // OP_VERSIONSIGNATURE_ANS
    }

    enum OpcodesEx : byte
    {
        Undefined = 0,

        LAST_LEGACY_EXT_OPCODE = OP_VERSIONSIGNATURE_ANS,

        LAST_LOW_EXT_OPCODE = OP_FILE_SELECTION_NOT_ALLOWED_INFO,

        // The following 12 opcodes can also be used ranging from
        // 128 to 139 in the OP_EDONKEYPROT for compatibility with
        // NB 1.4.23 and before.

        // No answer expected, will be ignored entirely.
        OP_NOOP = 1,
        // Pong expected.
        OP_PING = 2,
        OP_PONG = 3,
        OP_FILEMETADATA_REQ = 4,
        OP_FILEMETADATA_ANS = 5,
        // Has nothing to do with OP_FILEDESC (file comment).
        OP_FILEDESCRIPTOR_REQ = 6,
        OP_FILEDESCRIPTOR_ANS = 7,
        OP_DESELECT_UPLOAD_FILE = 8,
        // OP_COMPLBITMAP_REQ is outdated and will not be supported
        // forever. Use OP_COMPLBITMAP_EXT_REQ instead.
        OP_COMPLBITMAP_REQ = 9,
        OP_COMPLBITMAP_ANS = 10,
        OP_VERSIONSIGNATURE_REQ = 11,
        OP_VERSIONSIGNATURE_ANS = 12,
        OP_SECIDENTSTATE_EXT = 13,
        OP_COMPLBITMAP_EXT_REQ = 14,
        OP_FILE_SELECTION_NOT_ALLOWED_INFO = 15,

        TI_DL_OPCODES_START = TI_UPLOAD_REQUEST,
        TI_DL_OPCODES_END = TI_END_OF_DOWNLOAD_INFO,

        TI_UL_OPCODES_START = TI_QUEUE_RANK_INFO,
        TI_UL_OPCODES_END = TI_END_OF_UPLOAD_INFO,

        // Opcodes supporting transfer id.

        TI_UPLOAD_REQUEST = 100, // Like OP_STARTUPLOADREQ.
        TI_QUEUE_RANK_REQUEST = 101, // Has no counterpart in eD2k.
        TI_RANGE_REQUEST = 102, // Like OP_REQUESTPARTS.
        // The downloading client received a TI_REFUSE_UPLOAD_INFO
        // message and confirms it. The arguments are a new auxil
        // transfer id that substitutes the old auxil transfer id.
        TI_REFUSE_UPLOAD_ACKNOWLEDGEMENT = 103, // Has no counterpart in eD2k.
        TI_END_OF_DOWNLOAD_INFO = 104, // Like OP_END_OF_DOWNLOAD.

        TI_QUEUE_RANK_INFO = 110, // Like OP_QUEUERANKING.
        TI_ACCEPT_UPLOAD_INFO = 111, // Like OP_ACCEPTUPLOADREQ.
        TI_REFUSE_UPLOAD_INFO = 112, // Like OP_OUTOFPARTREQS.
        TI_END_OF_UPLOAD_INFO = 113, // Has no counterpart in eD2k.
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct PeerInfo
    {
        // Limit is in UTF-16 code units.
        public const ushort c_MaxUserNameChars = 0x7F;
        public const Tag.UnicodeSupport c_ExpectedUnicodeSupportForOutgoingHello = Tag.UnicodeSupport.ImplicitUTF8;
        public const bool c_ExpectNormalStrsAsUTF8InIngoingHello = true;

        public readonly string f_UserName;
        public readonly UserID f_UserID;
        public readonly byte f_eDonkeyVersion;
        public readonly I2PAddress f_DatagramAddr;

        public readonly ClientSoftware f_CompatibleClient;
        public readonly bool f_IsMule;
        public readonly byte f_ClientMajVer;
        public readonly byte f_ClientMinVer;
        public readonly byte f_ClientUpdate;

        public readonly byte f_AICHVer; // 3 bits used
        public readonly bool f_UnicodeSupport;
        public readonly byte f_UdpVer;
        public readonly byte f_DataCompVer;
        public readonly byte f_SupportSecIdent;
        public readonly byte f_SourceExchangeVer;
        public readonly byte f_ExtendedRequestsVer;
        public readonly byte f_AcceptCommentVer;
        public readonly bool f_PeerCache;
        public readonly bool f_NoViewSharedFiles;
        public readonly bool f_MultiPacket;
        public readonly bool f_SupportsPreview;

        public readonly byte f_KadVersion;
        public readonly bool f_OsInfoSupport;
        public readonly bool f_SupportsSharedDirs;

        // True for iMule v2.3 and later.
        public readonly bool f_IsSecondiMulePort;
        public readonly byte f_SuggestedSecIdentVer;

        private readonly uint f_EncodedVersion;

        public bool p_IsiMule
        {
            get { return (ClientSoftware.SO_IMULE == f_CompatibleClient); }
        }

        public bool p_IsThisClient
        {
            get { return (ClientSoftware.SO_NACHTBLITZ == f_CompatibleClient); }
        }

        public PeerInfo(string a_UserName,
                        UserID a_UserID,
                        I2PAddress a_DatagramAddr,
                        bool a_AllowDataCompression)
        {
            f_UserName = a_UserName;
            f_UserID = a_UserID;
            f_eDonkeyVersion = 0x3C;
            f_DatagramAddr = a_DatagramAddr;

            f_CompatibleClient = ClientSoftware.SO_NACHTBLITZ;
            // Well, at least we claim it's a mule to make some protocol
            // features work that we'd like to use.
            f_IsMule = true;
            f_ClientMajVer = Software.s_ReleaseInfo.f_MajorVersion;
            f_ClientMinVer = Software.s_ReleaseInfo.f_MinorVersion;
            f_ClientUpdate = Software.s_ReleaseInfo.f_Revision;

            f_AICHVer = 0x1;
            f_UnicodeSupport = true;
            f_UdpVer = 0x1;
            f_DataCompVer = (false == a_AllowDataCompression ? (byte)0 : (byte)0x1);
            f_SupportSecIdent = (byte)(25 > f_ClientUpdate ? 0x6 : 0xE);
            f_SourceExchangeVer = 0x3;
            f_ExtendedRequestsVer = 0;
            f_AcceptCommentVer = 0x1;
            f_PeerCache = false;
            f_NoViewSharedFiles = true;
            f_MultiPacket = false;
            f_SupportsPreview = false;

            f_KadVersion = 0x1;
            f_OsInfoSupport = true;
            f_SupportsSharedDirs = true;

            f_IsSecondiMulePort = false;
            f_SuggestedSecIdentVer = (byte)SuggestSecIdentVer(f_SupportSecIdent);

            f_EncodedVersion = (((uint)f_ClientMajVer << Util.c_BitsPerUShort) |
                                ((uint)f_ClientMinVer << Util.c_BitsPerByte) |
                                f_ClientUpdate);
        }

        public PeerInfo(string a_UserName,
                        UserID a_UserID,
                        byte a_eDonkeyVersion,
                        I2PAddress a_DatagramAddr,
                        ClientSoftware a_CompatibleClient,
                        bool a_IsMule,
                        byte a_ClientMajVer,
                        byte a_ClientMinVer,
                        byte a_ClientUpdate,
                        byte a_AICHVer,
                        bool a_UnicodeSupport,
                        byte a_UdpVer,
                        byte a_DataCompVer,
                        byte a_SupportSecIdent,
                        byte a_SourceExchangeVer,
                        byte a_ExtendedRequestsVer,
                        byte a_AcceptCommentVer,
                        bool a_PeerCache,
                        bool a_NoViewSharedFiles,
                        bool a_MultiPacket,
                        bool a_SupportsPreview,
                        byte a_KadVersion,
                        bool a_OsInfoSupport,
                        bool a_SupportsSharedDirs,
                        bool a_IsSecondiMulePort)
        {
            f_UserName = a_UserName;
            f_UserID = a_UserID;
            f_eDonkeyVersion = a_eDonkeyVersion;
            f_DatagramAddr = a_DatagramAddr;
            f_CompatibleClient = a_CompatibleClient;
            f_IsMule = a_IsMule;
            f_ClientMajVer = a_ClientMajVer;
            f_ClientMinVer = a_ClientMinVer;
            f_ClientUpdate = a_ClientUpdate;
            f_AICHVer = a_AICHVer;
            f_UnicodeSupport = a_UnicodeSupport;
            f_UdpVer = a_UdpVer;
            f_DataCompVer = a_DataCompVer;
            f_SupportSecIdent = a_SupportSecIdent;
            f_SourceExchangeVer = a_SourceExchangeVer;
            f_ExtendedRequestsVer = a_ExtendedRequestsVer;
            f_AcceptCommentVer = a_AcceptCommentVer;
            f_PeerCache = a_PeerCache;
            f_NoViewSharedFiles = a_NoViewSharedFiles;
            f_MultiPacket = a_MultiPacket;
            f_SupportsPreview = a_SupportsPreview;
            f_KadVersion = a_KadVersion;
            f_OsInfoSupport = a_OsInfoSupport;
            f_SupportsSharedDirs = a_SupportsSharedDirs;

            f_IsSecondiMulePort = a_IsSecondiMulePort;
            f_SuggestedSecIdentVer = (byte)SuggestSecIdentVer(f_SupportSecIdent);

            f_EncodedVersion = (((uint)f_ClientMajVer << Util.c_BitsPerUShort) |
                                ((uint)f_ClientMinVer << Util.c_BitsPerByte) |
                                f_ClientUpdate);
        }

        public static bool CreateFromTagList(UserID a_UserID,
                                             TagList a_TagList,
                                             out PeerInfo a_PeerInfo)
        {
            string l_UserName;
            uint l_Temp;

            if (false == a_TagList.TryGetTagAsString(TagCodes.CT_NAME, out l_UserName) ||
                c_MaxUserNameChars < l_UserName.Length ||
                false == a_TagList.TryGetTagAsUInt(TagCodes.CT_VERSION, out l_Temp) ||
                byte.MaxValue < l_Temp)
            {
                a_PeerInfo = new PeerInfo();

                return false;
            }

            byte l_eDonkeyVersion = (byte)l_Temp;
            I2PAddress l_DatagramAddr;

            if (false == a_TagList.TryGetTagAsAddress(TagCodes.ET_UDPDEST, out l_DatagramAddr))
            {
                a_PeerInfo = new PeerInfo();

                return false;
            }

            uint l_MiscOpt1;
            uint l_MiscOpt2;

            if (false == a_TagList.TryGetTagAsUInt(TagCodes.CT_EMULE_MISCOPTIONS1, out l_MiscOpt1) ||
                false == a_TagList.TryGetTagAsUInt(TagCodes.CT_EMULE_MISCOPTIONS2, out l_MiscOpt2))
            {
                a_PeerInfo = new PeerInfo();

                return false;
            }

            byte l_AICHVer = (byte)((l_MiscOpt1 >> ((7 * 4) + 1)) & 0x07);
            bool l_UnicodeSupport = (0 != ((l_MiscOpt1 >> (7 * 4)) & 0x01));
            byte l_UdpVer = (byte)((l_MiscOpt1 >> (6 * 4)) & 0x0F);
            byte l_DataCompVer = (byte)((l_MiscOpt1 >> (5 * 4)) & 0x0F);
            byte l_SupportSecIdent = (byte)((l_MiscOpt1 >> (4 * 4)) & 0x0F);
            byte l_SourceExchangeVer = (byte)((l_MiscOpt1 >> (3 * 4)) & 0x0F);
            byte l_ExtendedRequestsVer = (byte)((l_MiscOpt1 >> (2 * 4)) & 0x0F);
            byte l_AcceptCommentVer = (byte)((l_MiscOpt1 >> (1 * 4)) & 0x0F);
            bool l_PeerCache = (bool)(0 != ((l_MiscOpt1 >> 3) & 0x01));
            bool l_NoViewSharedFiles = (bool)(0 != ((l_MiscOpt1 >> 2) & 0x01));
            bool l_MultiPacket = (bool)(0 != ((l_MiscOpt1 >> 1) & 0x01));
            bool l_SupportsPreview = (bool)(0 != (l_MiscOpt1 & 0x01));

            byte l_KadVersion = (byte)(l_MiscOpt2 & 0x0f);
            bool l_OsInfoSupport = false;

            ulong l_Compat;
            if (true == a_TagList.TryGetTagAsULong(TagCodes.CT_EMULECOMPAT_OPTIONS, out l_Compat) &&
                0 != (l_Compat & 0x1))
            {
                l_OsInfoSupport = true;
            }

            uint l_eMVer;
            if (false == a_TagList.TryGetTagAsUInt(TagCodes.CT_EMULE_VERSION, out l_eMVer))
            {
                a_PeerInfo = new PeerInfo();

                return false;
            }

            ClientSoftware l_CompatibleClient = (ClientSoftware)(l_eMVer >> 24);
            bool l_IsMule = true;
            byte l_ClientMajVer;
            byte l_ClientMinVer;
            byte l_ClientUpdate;
            uint l_EncodedAppVersion;

            if (false == a_TagList.TryGetTagAsUInt(TagCodes.TAG_APPLICATION_VERSION, out l_EncodedAppVersion))
            {
                l_ClientMajVer = (byte)((l_eMVer >> 17) & 0x7F);
                l_ClientMinVer = (byte)((l_eMVer >> 10) & 0x7F);
                l_ClientUpdate = (byte)((l_eMVer >> 7) & 0x7);
            }
            else
            {
                l_ClientMajVer = (byte)((l_EncodedAppVersion >> 16) & byte.MaxValue);
                l_ClientMinVer = (byte)((l_EncodedAppVersion >> 8) & byte.MaxValue);
                l_ClientUpdate = (byte)(l_EncodedAppVersion & byte.MaxValue);
            }

            bool l_SupportsSharedDirs = true;
            bool l_IsSecondiMulePort = false;

            if (ClientSoftware.SO_IMULE == l_CompatibleClient)
            {
                if (2 > l_ClientMajVer ||
                    (2 == l_ClientMajVer &&
                     3 > l_ClientMinVer))
                {
                    l_UnicodeSupport = false;
                }
                else
                {
                    l_IsSecondiMulePort = true;
                }
            }

            a_PeerInfo = new PeerInfo(l_UserName, a_UserID, l_eDonkeyVersion, l_DatagramAddr, l_CompatibleClient, l_IsMule, l_ClientMajVer, l_ClientMinVer, l_ClientUpdate, l_AICHVer, l_UnicodeSupport, l_UdpVer, l_DataCompVer, l_SupportSecIdent, l_SourceExchangeVer, l_ExtendedRequestsVer, l_AcceptCommentVer, l_PeerCache, l_NoViewSharedFiles, l_MultiPacket, l_SupportsPreview, l_KadVersion, l_OsInfoSupport, l_SupportsSharedDirs, l_IsSecondiMulePort);

            return true;
        }

        private static uint SuggestSecIdentVer(uint a_SupportSecIdent)
        {
            // We assume that we support SecIdent v2, v3 and v4.

            if (0 != (a_SupportSecIdent & 0x8))
            {
                return 4;
            }

            if (0 != (a_SupportSecIdent & 0x4))
            {
                return 3;
            }

            if (0 != (a_SupportSecIdent & 0x2))
            {
                return 2;
            }

            return 0;
        }

        private static bool IsVersionRepresentable(uint a_MajVersion,
                                                   uint a_MinVersion,
                                                   uint a_Revision)
        {
            return (0x7F >= a_MajVersion &&
                    0x7F >= a_MinVersion &&
                    0x7 >= a_Revision);
        }

        public byte[] ToBuffer(bool a_IncludeHashSize)
        {
            const byte l_MaxRepresentableRevision = 7;

            uint l_eMVer = (uint)(((byte)f_CompatibleClient << 24) |
                                  (f_ClientMajVer << 17) |
                                  (f_ClientMinVer << 10) |
                                  (Math.Min(l_MaxRepresentableRevision, f_ClientUpdate) << 7));

            uint l_MiscOptions1 = (uint)((f_AICHVer << ((7 * 4) + 1)) |
                                         ((false == f_UnicodeSupport ? 0 : 1) << (7 * 4)) |
                                         (f_UdpVer << (6 * 4)) |
                                         (f_DataCompVer << (5 * 4)) |
                                         (f_SupportSecIdent << (4 * 4)) |
                                         (f_SourceExchangeVer << (3 * 4)) |
                                         (f_ExtendedRequestsVer << (2 * 4)) |
                                         (f_AcceptCommentVer << (1 * 4)) |
                                         ((false == f_PeerCache ? 0 : 1) << 3) |
                                         ((false == f_NoViewSharedFiles ? 0 : 1) << 2) |
                                         ((false == f_MultiPacket ? 0 : 1) << 1) |
                                         (false == f_SupportsPreview ? 0 : 1));

            uint l_Compat = (uint)(false == f_OsInfoSupport ? 0 : 1);

            TagList l_TagList = new TagList();

            l_TagList.AddTag(new StringTag(TagCodes.CT_NAME, f_UserName));
            l_TagList.AddTag(new NumberTag(TagCodes.CT_VERSION, f_eDonkeyVersion));
            l_TagList.AddTag(new AddressTag(TagCodes.ET_UDPDEST, f_DatagramAddr));
            l_TagList.AddTag(new NumberTag(TagCodes.CT_EMULE_VERSION, l_eMVer));
            l_TagList.AddTag(new NumberTag(TagCodes.CT_EMULE_MISCOPTIONS1, l_MiscOptions1));
            l_TagList.AddTag(new NumberTag(TagCodes.CT_EMULE_MISCOPTIONS2, f_KadVersion | (1u << 4)));
            l_TagList.AddTag(new NumberTag(TagCodes.CT_EMULECOMPAT_OPTIONS, l_Compat));
            l_TagList.AddTag(new StringTag(TagCodes.ET_MOD_VERSION, Software.c_AppName));

            if (false == IsVersionRepresentable(f_ClientMajVer, f_ClientMinVer, f_ClientUpdate))
            {
                // Add a special tag.

                uint l_EncodedVer = (((uint)f_ClientMajVer << 16) |
                                     ((uint)f_ClientMinVer << 8) |
                                     f_ClientUpdate); ;
                l_TagList.AddTag(new NumberTag(TagCodes.TAG_APPLICATION_VERSION, l_EncodedVer));
            }

            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            if (true == a_IncludeHashSize)
            {
                // This is the difference between the OP_HELLO and
                // the OP_HELLOANSWER.

                l_Buffer.WriteByte(UInt128.c_ByteSize);
            }

            l_Buffer.WriteUInt128BE(f_UserID.f_Val);
            l_TagList.WriteTo(l_Buffer, sizeof(uint), c_ExpectedUnicodeSupportForOutgoingHello);

            return l_Buffer.ToArray(true);
        }

        public bool HasSecIdentSupport()
        {
            // Check for ver2, ver3, ver4 (bit 1 to 3).
            return (0 != (f_SupportSecIdent & 0xE));
        }

        public bool SupportBasicOpcodeExtensions()
        {
            return p_IsThisClient;
        }

        public bool SupportsComplBitmapOpcode(out bool a_AlsoFileSelFlags)
        {
            if (true == p_IsThisClient)
            {
                a_AlsoFileSelFlags = (0x10419 <= EncodeVersion());

                return (0x10407 <= EncodeVersion());
            }

            a_AlsoFileSelFlags = false;

            return false;
        }

        public bool SupportsOP_OUTOFPARTREQS()
        {
            return (false == p_IsThisClient ||
                    0x10418 <= EncodeVersion());
        }

        public bool SupportsNBProtocol()
        {
            return (true == p_IsThisClient &&
                    0x10418 <= EncodeVersion());
        }

        public bool SupportsFastTrackAuthentication()
        {
            return SupportsNBProtocol();
        }

        public bool SupportsTransferIDOpcodes()
        {
            return SupportsNBProtocol();
        }

        public bool HasA4AFLimitation()
        {
            return p_IsiMule;
        }

        private string GetClientName(TextSet a_TextSet)
        {
            // Return can be null.

            switch (f_CompatibleClient)
            {
                case ClientSoftware.SO_IMULE:
                    return "iMule";

                case ClientSoftware.SO_NACHTBLITZ:
                    return a_TextSet.f_AppName;

                default:
                    return null;
            }
        }

        public string BuildClientString(TextSet a_TextSet)
        {
            string l_ClientName = GetClientName(a_TextSet);

            if (null != l_ClientName)
            {
                StringBuilder l_Bld = new StringBuilder(10);
                l_Bld.Append(f_ClientMajVer);
                l_Bld.Append('.');
                l_Bld.Append(f_ClientMinVer);
                l_Bld.Append('.');
                l_Bld.Append(f_ClientUpdate);

                return a_TextSet.Format(TextKeys.ClientNameWithVersion, l_ClientName, l_Bld.ToString());
            }

            return a_TextSet[TextKeys.UnknownValue];
        }

        public uint EncodeVersion()
        {
            return f_EncodedVersion;
        }
    }

#if !MONO
    [StructLayout(LayoutKind.Auto)]
#endif
    struct TransferInfo
    {
        public readonly bool f_Outgoing;
        public readonly bool f_Compressed;
        public readonly FileRange f_Range;
        public readonly float f_Progress;

        public TransferInfo(bool a_Outgoing,
                            bool a_Compressed,
                            ref FileRange a_Range,
                            float a_Progress)
        {
            f_Outgoing = a_Outgoing;
            f_Compressed = a_Compressed;
            f_Range = a_Range;
            f_Progress = a_Progress;
        }
    }

    abstract class DonkeyClientBase : ITransferable
    {
        private const byte c_PacketHeaderSize = (sizeof(byte) + sizeof(uint) + sizeof(byte));
        protected const byte c_PlainContentHeaderSize = 32;
        protected const byte c_ComprContentHeaderSize = 28;
        protected const byte c_ComprContentEssentialHeaderSize = 24;
        private const byte c_MinPacketBodySizeForInspection = c_PlainContentHeaderSize;
        private const uint c_TestLivelinessAfterIdleSecs = Time.c_SecsPerHalfMinute;
        private const uint c_KillConnectionAfterIdleSecs = (5 * Time.c_SecsPerMinute);

        public readonly Instance f_Inst;
        protected readonly UserContext f_UserCntxt;
        protected readonly I2PStream f_Strm;

        private uint f_PacketsSent;
        private uint f_PacketsReceived;
        private TransferVolume f_TotalVolume;
        private Time f_NextLivelinessCheckTime;
        private Time f_DestroyTime;

        // For receiving.
        private readonly Queue<Packet> f_PacketInQueue;
        private uint f_InQueueSize;
        private IncomingPacketInfo f_InInfo;
        private ComprPacketSeriesInfo f_ComprSeriesInfo;

        // For sending:
        private List<Packet> f_PacketOutQueue;
        private uint f_OutQueueSize;
        private uint f_FirstOutPacketOffset;

        public bool p_Outgoing
        {
            get { return f_Strm.p_Outgoing; }
        }

        public I2PAddress p_Addr
        {
            get { return f_Strm.p_RemoteAddr; }
        }

        public bool p_IsAlive
        {
            get { return f_Strm.p_IsAlive; }
        }

        public bool p_IsBusyWriting
        {
            get { return f_Strm.p_IsBusyWriting; }
        }

        public UserContext p_UserCntxt
        {
            // NOTE: The associated user context can change.
            get { return f_UserCntxt; }
        }

        protected DonkeyClientBase(Instance a_Inst,
                                   uint a_DebugConnID,
                                   UserContext a_UserCntxt,
                                   I2PStream a_Strm)
        {
            f_Inst = a_Inst;

            f_UserCntxt = a_UserCntxt;
            f_Strm = a_Strm;

            f_PacketInQueue = new Queue<Packet>();
            f_PacketOutQueue = new List<Packet>();

            SpendLife();
        }

        public void Close()
        {
            f_Strm.Close();
        }

        public void Ban()
        {
            // This will in effect close the connection and mark the
            // corresponding user context (if any) as banned.

            f_Inst.f_NetworkMngr.BanAddress(f_Strm.p_RemoteAddr);
        }

        public Packet DequeuePacket()
        {
            if (0 < f_PacketInQueue.Count)
            {
                Packet l_Packet = f_PacketInQueue.Dequeue();

                f_InQueueSize -= l_Packet.GetReceivedBodySize();

                return l_Packet;
            }

            return null;
        }

        protected bool EnqueuePacketBase(Packet a_Packet,
                                         PacketInsertionMode a_InsertionMode)
        {
            uint l_NewOutQueueSize = (f_OutQueueSize + a_Packet.f_TransferSize);

            if (l_NewOutQueueSize < f_OutQueueSize)
            {
                // Overflow.

                return false;
            }

            if ((10 * 1024 * 1024) < l_NewOutQueueSize)
            {
                return false;
            }

            int l_Idx;

            switch (a_InsertionMode)
            {
                case PacketInsertionMode.End:
                    l_Idx = f_PacketOutQueue.Count;
                    break;

                case PacketInsertionMode.MoveAheadAsFarAsPossible:
                    l_Idx = FindMoveAheadInsertionPos(a_Packet);
                    break;

                case PacketInsertionMode.NearStart:
                    l_Idx = FindHeadInsertionPos();
                    break;

                default: throw (new ApplicationException());
            }

            f_PacketOutQueue.Insert(l_Idx, a_Packet);
            f_OutQueueSize = l_NewOutQueueSize;

            return true;
        }

        private int FindMoveAheadInsertionPos(Packet a_Packet)
        {
            for (int i = (f_PacketOutQueue.Count - 1); 0 <= i; i--)
            {
                if (false == f_PacketOutQueue[i].p_CanBeReorderedBehind)
                {
                    return (i + 1);
                }
            }

            return FindHeadInsertionPos();
        }

        private int FindHeadInsertionPos()
        {
            // Take the first index if we aren't sending
            // the top packet yet, otherwise the second
            // will have to do.

            return (0 == f_FirstOutPacketOffset ? 0 : 1);
        }

        public Packet FindAnyFileContentPacketOnOutQueue(KnownFile a_TargetFile)
        {
            foreach (Packet l_Packet in f_PacketOutQueue)
            {
                if (true == l_Packet.IsUlContent(a_TargetFile))
                {
                    return l_Packet;
                }
            }

            return null;
        }

        public uint RemoveFileContentPacketClustersFromOutQueue(KnownFile a_TargetFile,
                                                                bool a_EvenIfClusterIsInTransfer)
        {
            // If a_EvenIfClusterIsInTransfer is zero, only target
            // clusters that have no piece in transfer are removed.
            // Otherwise all pieces are removed, except the one that
            // is already in transfer (if any). Returns the number of
            // pieces that remained (including the one potentially
            // already in transfer).

            if (0 == f_PacketOutQueue.Count)
            {
                return 0;
            }

            uint l_PiecesRemaining = 0;
            List<Packet> l_OldQueue = f_PacketOutQueue;
            List<Packet> l_NewQueue = null;

            bool l_SkipAllContent = false;
            bool l_KeepRecentCluster = false;
            uint l_RecentClusterNumber = 0;
            int l_Idx = 0;

            foreach (Packet l_Packet in l_OldQueue)
            {
                bool l_IsRelevantContent;

                if (null == l_Packet.f_RelatedUlFile)
                {
                    l_IsRelevantContent = false;
                }
                else
                {
                    l_IsRelevantContent = l_Packet.IsUlContent(a_TargetFile);
                }

                bool l_SavePacket;

                if (false == l_IsRelevantContent)
                {
                    l_SavePacket = true;
                }
                else if (true == l_SkipAllContent)
                {
                    l_SavePacket = false;
                }
                else
                {
                    if (0 == l_Idx &&
                        0 != f_FirstOutPacketOffset)
                    {
                        // We cannot remove this packet, as we're
                        // already sending it.

                        l_PiecesRemaining++;
                        l_SavePacket = true;

                        if (false == a_EvenIfClusterIsInTransfer)
                        {
                            l_KeepRecentCluster = true;
                            l_RecentClusterNumber = l_Packet.p_ClusterNumber;
                        }
                        else
                        {
                            if (false == l_Packet.IsLastPieceOfCluster())
                            {
                                l_Packet.SetClusterAborted();
                            }

                            l_SkipAllContent = true;
                        }
                    }
                    else
                    {
                        if (false == l_KeepRecentCluster)
                        {
                            if (false == a_EvenIfClusterIsInTransfer)
                            {
                                if (0 == l_Packet.p_PieceNumber)
                                {
                                    l_SavePacket = false;
                                    l_SkipAllContent = true;
                                }
                                else
                                {
                                    l_SavePacket = l_KeepRecentCluster = true;
                                    l_RecentClusterNumber = l_Packet.p_ClusterNumber;
                                }
                            }
                            else
                            {
                                l_SavePacket = false;
                                l_SkipAllContent = true;
                            }
                        }
                        else
                        {
                            if (l_Packet.p_ClusterNumber == l_RecentClusterNumber)
                            {
                                l_PiecesRemaining++;
                                l_SavePacket = true;
                            }
                            else
                            {
                                l_SavePacket = false;
                                l_SkipAllContent = true;
                            }
                        }
                    }
                }

                if (false == l_SavePacket)
                {
                    f_OutQueueSize -= l_Packet.f_TransferSize;

                    if (null == l_NewQueue)
                    {
                        l_NewQueue = new List<Packet>(l_OldQueue.Count - 1);

                        for (int k = 0; k != l_Idx; k++)
                        {
                            l_NewQueue.Add(l_OldQueue[k]);
                        }
                    }
                }
                else if (null != l_NewQueue)
                {
                    l_NewQueue.Add(l_Packet);
                }

                l_Idx++;
            }

            if (null != l_NewQueue)
            {
                // We had changes.

                f_PacketOutQueue = l_NewQueue;
            }

            return l_PiecesRemaining;
        }

        private uint GetRemainingInQueueSpace()
        {
            uint l_Limit = f_Inst.f_NetworkMngr.p_MaxStreamInQueueSizePerConn;
            uint l_CurSize = (f_InQueueSize + f_InInfo.p_BytesDone);

            if (l_Limit > l_CurSize)
            {
                return (l_Limit - l_CurSize);
            }

            return 0;
        }

        private uint GetSizeReadyForSending()
        {
            return (f_OutQueueSize - f_FirstOutPacketOffset);
        }

        public uint Download(uint a_MaxBytes)
        {
            // Returns the bytes received.

            uint l_Avail = Math.Min(f_Strm.p_AvailableBytes, GetRemainingInQueueSpace());
            l_Avail = Math.Min(l_Avail, a_MaxBytes);
            uint l_StoredAvail = l_Avail;
            byte[] l_HeaderBuf = null;
            bool l_NonNoop = false;

            while (true)
            {
                if (DlStates.WaitingForHeader == f_InInfo.p_RecvState)
                {
                    if (c_PacketHeaderSize > l_Avail)
                    {
                        break;
                    }

                    if (null == l_HeaderBuf)
                    {
                        l_HeaderBuf = new byte[c_PacketHeaderSize];
                    }

                    if (false == IncomingPacketInfo.CreateFromHeader(f_Strm, l_HeaderBuf, f_Inst.f_NetworkMngr.p_MaxStreamInQueueSizePerConn, out f_InInfo))
                    {
                        Close();

                        break;
                    }

                    l_Avail -= c_PacketHeaderSize;
                }

                uint l_Req = f_InInfo.p_RemBytes;
                bool l_Abort = false;

                if (0 != l_Req)
                {
                    if (0 == l_Avail)
                    {
                        l_Abort = true;
                    }
                    else
                    {
                        uint l_Target = Math.Min(l_Avail, l_Req);
                        uint l_Done = f_InInfo.ReadBytes(f_Strm, l_Target);

                        l_Avail -= l_Done;

                        if (l_Target != l_Req)
                        {
                            l_Abort = true;
                        }
                    }
                }

                HandlePacketDownloadStop();

                if (false == l_NonNoop &&
                    l_StoredAvail > l_Avail &&
                    false == f_InInfo.f_IsNoop)
                {
                    l_NonNoop = true;
                }

                if (true == l_Abort)
                {
                    break;
                }

                ByteArray l_Data = new ByteArray(f_InInfo.f_Data);
                Packet l_Packet = Packet.InterpretIncoming(f_Strm.p_RemoteAddr,
                                                           f_InInfo.f_Prot,
                                                           (byte)f_InInfo.f_Opcode,
                                                           ref l_Data,
                                                           Packet.c_MaxAllowedBodySize,
                                                           c_PacketHeaderSize + f_InInfo.f_BodySize);

                if (null == l_Packet)
                {
                    Ban();

                    break;
                }

                f_PacketsReceived++;
                f_PacketInQueue.Enqueue(l_Packet);
                f_InQueueSize += l_Packet.GetReceivedBodySize();

                f_InInfo = new IncomingPacketInfo();
            }

            uint l_BytesRead = (l_StoredAvail - l_Avail);

            if (0 != l_BytesRead)
            {
                f_TotalVolume.Received(l_BytesRead);

                if (true == l_NonNoop)
                {
                    SpendLife();
                }
            }

            return l_BytesRead;
        }

        public void HandlePacketDownloadStop()
        {
            // This routine will get called whenever an incoming
            // packet won't progress any further during the current
            // download turn. Thus it will get called at least once
            // for every packet (in the moment it is completely downloaded).

            uint l_BytesProgressed = f_InInfo.CheckProgress();

            if (0 == l_BytesProgressed ||
                false == f_InInfo.f_IsContent)
            {
                return;
            }

            NotifyIncomingContent(ref f_InInfo, l_BytesProgressed);

            // Now try to find out the cumulative offset for
            // GetDlTransferInfo().

            if (Opcodes.OP_COMPRESSEDPART == (Opcodes)f_InInfo.f_Opcode &&
                c_ComprContentEssentialHeaderSize <= f_InInfo.p_BytesDone)
            {
                // We will determine the cumulative offset of this
                // packet from a (possible) series of compressed packets.
                // A single packet itself doesn't give us this information
                // due to protocol limitations, so we have to do extra
                // bookkeeping.

                byte[] l_PacketBody = f_InInfo.f_Data;

                if (true == f_ComprSeriesInfo.IsValid() &&
                    false == Buffer2.ValueEquals(f_ComprSeriesInfo.f_ComprHeader, 0, l_PacketBody, 0, c_ComprContentEssentialHeaderSize))
                {
                    // The former series has ended.

                    f_ComprSeriesInfo = new ComprPacketSeriesInfo();
                }

                if (false == f_ComprSeriesInfo.IsValid())
                {
                    // Start a new series.

                    byte[] l_Copy = Util.CopyBytes(l_PacketBody, 0, c_ComprContentEssentialHeaderSize);
                    f_ComprSeriesInfo = new ComprPacketSeriesInfo(l_Copy);
                }

                if (0 == f_InInfo.p_RemBytes)
                {
                    // Packet completed, so increment the offset.

                    uint l_BytesDone = f_InInfo.p_BytesDone;

                    if (c_ComprContentHeaderSize <= l_BytesDone)
                    {
                        f_ComprSeriesInfo.f_CumComprOffset += (l_BytesDone - c_ComprContentHeaderSize);
                    }
                }
            }
        }

        protected abstract void NotifyIncomingContent(ref IncomingPacketInfo a_Info,
                                                      uint a_NewlyRcvdBytes);

        public uint Upload(uint a_MaxBytes)
        {
            // Returns the bytes sent.

            ByteArray l_Bytes;
            FlexArray<PacketSendInfo> l_Infos;
            FlexArray<PacketSendInfo>.Create(out l_Infos);

            Packet l_FirstPacket = f_PacketOutQueue[0];
            uint l_FirstSizeLeft = (l_FirstPacket.f_TransferSize - f_FirstOutPacketOffset);

            if (l_FirstSizeLeft > a_MaxBytes)
            {
                // We can't even send the first queued
                // packed completely.

                l_FirstPacket.BuildByteArrayForSending(f_FirstOutPacketOffset, a_MaxBytes, out l_Bytes);

                int l_Idx = l_Infos.ReserveAtEnd();
                l_Infos.p_Arr[l_Idx] = new PacketSendInfo(l_FirstPacket, a_MaxBytes, false);
                f_FirstOutPacketOffset += a_MaxBytes;
            }
            else
            {
                // We can send the whole first packet.

                if (1 == f_PacketOutQueue.Count)
                {
                    // There is only one packet left.

                    l_FirstPacket.BuildByteArrayForSending(f_FirstOutPacketOffset, out l_Bytes);

                    DequeueTopmostOutPacket(ref l_Infos, (uint)l_Bytes.f_Len);
                    f_FirstOutPacketOffset = 0;
                }
                else
                {
                    uint l_BytesToGo = Math.Min(GetSizeReadyForSending(), a_MaxBytes);

                    MemoryIO l_Buffer = MemoryIO.CreateWritable(l_BytesToGo);

                    l_FirstPacket.WriteToFileForSending(l_Buffer, f_FirstOutPacketOffset, l_FirstSizeLeft);

                    l_BytesToGo -= l_FirstSizeLeft;
                    DequeueTopmostOutPacket(ref l_Infos, l_FirstSizeLeft);
                    f_FirstOutPacketOffset = 0;

                    while (0 != l_BytesToGo)
                    {
                        Packet l_Packet = f_PacketOutQueue[0];
                        uint l_OutSize;

                        if (l_Packet.f_TransferSize <= l_BytesToGo)
                        {
                            // Write the whole packet.

                            l_OutSize = l_Packet.f_TransferSize;
                            l_BytesToGo -= l_OutSize;
                            DequeueTopmostOutPacket(ref l_Infos, l_OutSize);
                        }
                        else
                        {
                            // Write only a part of the packet and
                            // set f_FirstPacketOffset.

                            l_OutSize = l_BytesToGo;
                            l_BytesToGo = 0;
                            int l_Idx = l_Infos.ReserveAtEnd();
                            l_Infos.p_Arr[l_Idx] = new PacketSendInfo(l_Packet, l_OutSize, false);
                            f_FirstOutPacketOffset = l_OutSize;
                        }

                        l_Packet.WriteToFileForSending(l_Buffer, 0, l_OutSize);
                    }

                    l_Bytes = new ByteArray(l_Buffer.ToArray(true));
                }
            }

            if (true == f_Strm.WriteBytesAsync(ref l_Bytes))
            {
                bool l_NonNoop = false;

                PacketSendInfo[] l_Arr = l_Infos.p_Arr;
                for (int i = 0, lim = l_Infos.p_Count; i < lim; i++)
                {
                    l_Arr[i].f_Packet.NotifySendingPacket(l_Arr[i].f_Bytes, l_Arr[i].f_Done);

                    if (true == l_Arr[i].f_Done)
                    {
                        f_PacketsSent++;
                    }

                    if (false == l_NonNoop &&
                        false == l_Arr[i].f_Packet.IsNoop())
                    {
                        l_NonNoop = true;
                    }
                }

                f_TotalVolume.Sent((uint)l_Bytes.f_Len);

                if (true == l_NonNoop)
                {
                    SpendLife();
                }

                return (uint)l_Bytes.f_Len;
            }

            return 0;
        }

        private void DequeueTopmostOutPacket(ref FlexArray<PacketSendInfo> a_SendInfo,
                                             uint a_RemBytes)
        {
            Packet l_Packet = f_PacketOutQueue[0];
            f_PacketOutQueue.RemoveAt(0);

            int l_Idx = a_SendInfo.ReserveAtEnd();
            a_SendInfo.p_Arr[l_Idx] = new PacketSendInfo(l_Packet, a_RemBytes, true);

            f_OutQueueSize -= l_Packet.f_TransferSize;
        }

        private void SpendLife()
        {
            // Increases the timers.

            Time l_Time = f_Inst.GetTime();

            f_NextLivelinessCheckTime = l_Time.AddSeconds(c_TestLivelinessAfterIdleSecs);
            f_DestroyTime = l_Time.AddSeconds(c_KillConnectionAfterIdleSecs);
        }

        public bool DoLivelinessCheck(Time a_Time)
        {
            // Since .NET won't allow us to figure out if
            // a connection is dead without sending actual
            // data, we send some periodically.

            if (0 == f_PacketOutQueue.Count &&
                false == f_Strm.p_IsBusyWriting)
            {
                if (true == f_NextLivelinessCheckTime.IsEarlierThan(a_Time))
                {
                    if (true == f_DestroyTime.IsEarlierThan(a_Time) ||
                        false == ShouldTestForLife())
                    {
                        return false;
                    }

                    if (true == Util.sp_NeedToSendToTestConnLife)
                    {
                        ByteArray l_Data = new ByteArray();

                        if (false == EnqueuePacketBase(Packet.BuildOwnPacket(p_Addr, OpcodesEx.OP_NOOP, false, true, ref l_Data), PacketInsertionMode.NearStart))
                        {
                            return false;
                        }
                    }

                    f_NextLivelinessCheckTime = a_Time.AddSeconds(c_TestLivelinessAfterIdleSecs);
                }
            }

            return true;
        }

        protected abstract bool ShouldTestForLife();

        public bool NeedsToReceive()
        {
            // This is just for quick and dirty checking
            // and not as elaborate as ReinitializeForDl().

            return (0 != f_Strm.p_AvailableBytes);
        }

        public bool NeedsToSend()
        {
            // This is just for quick and dirty checking
            // and not as elaborate as ReinitializeForUl().

            return (false == f_Strm.p_IsBusyWriting &&
                    0 != GetSizeReadyForSending());
        }

        public bool IsDownloadingFileContent()
        {
            return (null != f_InInfo.f_DlPeerCntxt);
        }

        public bool IsUploadingFileContent()
        {
            if (0 < f_PacketOutQueue.Count)
            {
                Packet l_FirstPacket = f_PacketOutQueue[0];

                return (null != l_FirstPacket.f_RelatedUlFile);
            }

            return false;
        }

        public static bool IsContent(Protocols a_Prot,
                                     Opcodes a_Opcode)
        {
            if (Protocols.OP_EMULEPROT >= a_Prot)
            {
                if (Opcodes.OP_SENDINGPART == a_Opcode)
                {
                    return (Protocols.OP_EDONKEYPROT == a_Prot ||
                            Protocols.OP_EMULEPROT == a_Prot ||
                            Protocols.OP_PACKEDPROT == a_Prot);
                }

                if (Opcodes.OP_COMPRESSEDPART == a_Opcode)
                {
                    return (Protocols.OP_EMULEPROT == a_Prot ||
                            Protocols.OP_PACKEDPROT == a_Prot);
                }
            }

            return false;
        }

        public static Protocols SuggestProtocol(Opcodes a_Opcode)
        {
            return (Opcodes.OP_EMULEINFO > a_Opcode ? Protocols.OP_EDONKEYPROT : Protocols.OP_EMULEPROT);
        }

#region Transfer

        public bool ReinitializeForDl(out TransferDemand a_Demand)
        {
            if (false == f_Strm.p_IsAlive)
            {
                a_Demand = new TransferDemand();

                return false;
            }

            // We can not say for sure that we will be able to read
            // every single byte that is available (the packet
            // header is only read as one block).
            uint l_NeededBw = f_Strm.p_AvailableBytes;

            if (0 != l_NeededBw)
            {
                l_NeededBw = Math.Min(l_NeededBw, GetRemainingInQueueSpace());
            }

            if (0 == l_NeededBw)
            {
                a_Demand = new TransferDemand();

                return false;
            }

            a_Demand = new TransferDemand(l_NeededBw);

            return true;
        }

        public bool ReinitializeForUl(out TransferDemand a_Demand)
        {
            if (true == f_Strm.p_IsBusyWriting)
            {
                a_Demand = new TransferDemand();

                return false;
            }

            uint l_NeededBw = Math.Min(GetSizeReadyForSending(), NetworkManager.c_MaxBytesUploadPerConnPerTurn);

            if (0 == l_NeededBw ||
                false == f_Strm.p_IsAlive)
            {
                a_Demand = new TransferDemand();

                return false;
            }

            a_Demand = new TransferDemand(f_UserCntxt.p_Priority, l_NeededBw);

            return true;
        }

        public uint AssignDlBandwidth(uint a_ExtraBytes)
        {
            // Ignore (trust the caller).

            return a_ExtraBytes;
        }

        public uint AssignUlBandwidth(uint a_ExtraBytes)
        {
            // Ignore (trust the caller).

            return a_ExtraBytes;
        }

#endregion

#region TransferInspection

        public bool GetDlTransferInfo(out TransferInfo a_Info)
        {
            if (DlStates.WaitingForBody == f_InInfo.p_RecvState &&
                c_MinPacketBodySizeForInspection <= f_InInfo.p_BytesDone &&
                true == f_InInfo.f_IsContent)
            {
                ByteArray l_Data = new ByteArray(f_InInfo.f_Data, 0, (int)f_InInfo.p_BytesDone);

                return BuildTransferInfo(f_InInfo.f_Prot, (Opcodes)f_InInfo.f_Opcode, false, f_ComprSeriesInfo.f_CumComprOffset, ref l_Data, out a_Info);
            }

            a_Info = new TransferInfo();

            return false;
        }

        public bool GetUlTransferInfo(out TransferInfo a_Info)
        {
            if (0 < f_PacketOutQueue.Count &&
                (c_PacketHeaderSize + c_MinPacketBodySizeForInspection) <= f_FirstOutPacketOffset)
            {
                Packet l_Packet = f_PacketOutQueue[0];

                if (true == l_Packet.IsContent())
                {
                    ByteArray l_Data;
                    l_Packet.BuildByteArrayForSending(c_PacketHeaderSize, f_FirstOutPacketOffset - c_PacketHeaderSize, out l_Data);

                    return BuildTransferInfo(l_Packet.f_Prot, l_Packet.p_DonkeyOpcode, true, l_Packet.p_CumComprOffset, ref l_Data, out a_Info);
                }
            }

            a_Info = new TransferInfo();

            return false;
        }

        private static bool BuildTransferInfo(Protocols a_Prot,
                                              Opcodes a_Opcode,
                                              bool a_Outgoing,
                                              uint a_CumComprOffset,
                                              ref ByteArray a_Body,
                                              out TransferInfo a_TransferInfo)
        {
            // See DonkeyClient.ProcessFileContent().

            bool l_Compressed = (Opcodes.OP_COMPRESSEDPART == a_Opcode);
            ulong l_Offset = Util.ReadUInt64LE(a_Body.f_Arr, a_Body.f_Idx + UInt128.c_ByteSize);
            ulong l_FinalOffset;
            uint l_ContentDone;

            if (false == l_Compressed)
            {
                l_FinalOffset = Util.ReadUInt64LE(a_Body.f_Arr, a_Body.f_Idx + 24);
                l_ContentDone = ((uint)a_Body.f_Len - c_PlainContentHeaderSize);
            }
            else
            {
                l_FinalOffset = (l_Offset + Util.ReadUInt32LE(a_Body.f_Arr, a_Body.f_Idx + 24));
                l_ContentDone = (a_CumComprOffset + (uint)a_Body.f_Len - c_ComprContentHeaderSize);
            }

            FileRange l_Range;
            if (false == FileRange.CreateFromUnsafeBounds(l_Offset, l_FinalOffset, out l_Range))
            {
                a_TransferInfo = new TransferInfo();

                return false;
            }

            if (l_ContentDone > l_Range.f_Length)
            {
                l_ContentDone = l_Range.f_Length;
            }

            float l_Perc = ((float)l_ContentDone / l_Range.f_Length);
            a_TransferInfo = new TransferInfo(a_Outgoing, l_Compressed, ref l_Range, l_Perc);

            return true;
        }

#endregion

        protected enum DlStates : byte
        {
            WaitingForHeader = 0,
            WaitingForBody = 1,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        protected struct IncomingPacketInfo
        {
            private DlStates f_RecvState;
            public readonly Protocols f_Prot;
            public readonly byte f_Opcode;
            public readonly bool f_IsNoop;
            public readonly bool f_IsContent;
            public readonly uint f_BodySize;
            public readonly byte[] f_Data;

            private uint f_BytesDone;
            private uint f_OldBytesDone;

            public DlPeerContext f_DlPeerCntxt;
            public uint f_UnrecognizedSize;

            public DlStates p_RecvState
            {
                get { return f_RecvState; }
            }

            public uint p_BytesDone
            {
                get { return f_BytesDone; }
            }

            public uint p_RemBytes
            {
                get { return (f_BodySize - f_BytesDone); }
            }

            public bool IsValid()
            {
                return (DlStates.WaitingForBody == f_RecvState);
            }

            private IncomingPacketInfo(Protocols a_Prot,
                                       byte a_Opcode,
                                       uint a_BodySize)
            {
                f_RecvState = DlStates.WaitingForBody;
                f_Prot = a_Prot;
                f_Opcode = a_Opcode;
                f_IsNoop = ((Opcodes.OP_LEGACY_START == (Opcodes)f_Opcode &&
                             Protocols.OP_EMULEPROT == f_Prot) ||
                            ((Protocols.PROT_NB_PLAIN == f_Prot ||
                              Protocols.PROT_NB_COMPRESSED == f_Prot) &&
                             OpcodesEx.OP_NOOP == (OpcodesEx)f_Opcode));
                f_IsContent = DonkeyClientBase.IsContent(a_Prot, (Opcodes)a_Opcode);
                f_BodySize = a_BodySize;
                f_Data = (0 == a_BodySize ? Util.s_Empty : new byte[a_BodySize]);

                f_BytesDone = 0;
                f_OldBytesDone = 0;

                f_DlPeerCntxt = null;
                f_UnrecognizedSize = 0;
            }

            public static bool CreateFromHeader(I2PStream a_Strm,
                                                byte[] a_HeaderBuf,
                                                uint a_MaxBodySize,
                                                out IncomingPacketInfo a_Dest)
            {
                // On error, no changes are made to any members.

                ByteArray l_Data = new ByteArray(a_HeaderBuf);
                int l_HeaderBytesRead = a_Strm.ReadAvailableBytes(ref l_Data);

                if (c_PacketHeaderSize != l_HeaderBytesRead)
                {
                    // This can happen (if the other side has closed
                    // the connection in the meantime).

                    a_Dest = new IncomingPacketInfo();

                    return false;
                }

                Protocols l_Prot = (Protocols)a_HeaderBuf[0];
                bool l_ValidProt = ((Protocols.OP_EDONKEYHEADER <= l_Prot &&
                                     Protocols.OP_EMULEPROT >= l_Prot) ||
                                    (Protocols.PROT_NB_PLAIN <= l_Prot &&
                                     Protocols.PROT_NB_COMPRESSED >= l_Prot));

                if (false == l_ValidProt)
                {
                    a_Dest = new IncomingPacketInfo();

                    return false;
                }

                uint l_Size = Util.ReadUInt32LE(a_HeaderBuf, 1);
                byte l_Opcode = a_HeaderBuf[5];

                // The given size must include the opcode byte, so
                // zero is an invalid value. We'll decrement one to
                // find out the true body size.

                if (0 == l_Size ||
                    --l_Size > a_MaxBodySize ||
                    0 == l_Opcode)
                {
                    a_Dest = new IncomingPacketInfo();

                    return false;
                }

                if (Protocols.OP_EDONKEYHEADER == l_Prot)
                {
                    l_Prot = Protocols.OP_EDONKEYPROT;
                }

                a_Dest = new IncomingPacketInfo(l_Prot, l_Opcode, l_Size);

                return true;
            }

            public uint ReadBytes(I2PStream a_Strm,
                                  uint a_Length)
            {
                ByteArray l_Bytes = new ByteArray(f_Data, (int)f_BytesDone, (int)a_Length);
                int l_Done = a_Strm.ReadAvailableBytes(ref l_Bytes);

                f_BytesDone += (uint)l_Done;

                return (uint)l_Done;
            }

            public uint CheckProgress()
            {
                // Returns the overall number of new bytes
                // since the last call.

                uint l_Progress = (f_BytesDone - f_OldBytesDone);

                f_OldBytesDone = f_BytesDone;

                return l_Progress;
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct ComprPacketSeriesInfo
        {
            // Don't use the struct default constructor.

            public readonly byte[] f_ComprHeader;
            // f_CumComprOffset does not contain overhead.
            public uint f_CumComprOffset;

            public ComprPacketSeriesInfo(byte[] a_ComprHeader)
            {
                f_ComprHeader = a_ComprHeader;
                f_CumComprOffset = 0;
            }

            public bool IsValid()
            {
                return (null != f_ComprHeader);
            }
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct PacketSendInfo
        {
            // Don't use the struct default constructor.

            public readonly Packet f_Packet;
            public readonly uint f_Bytes;
            public readonly bool f_Done;

            public PacketSendInfo(Packet a_Packet,
                                  uint a_Bytes,
                                  bool a_Done)
            {
                f_Packet = a_Packet;
                f_Bytes = a_Bytes;
                f_Done = a_Done;
            }
        }

        protected enum PacketInsertionMode : byte
        {
            End = 0,
            MoveAheadAsFarAsPossible = 1,
            NearStart = 2,
        }
    }

    sealed class DonkeyClient : DonkeyClientBase
    {
        public const uint c_MaxRangeLength = (512 * 1024);
        private const bool c_AllowAuthenticationOfPeerToFail = true;

        public const byte c_MaxPiecesPerCluster = byte.MaxValue;
        private const uint c_DefaultPieceSize = (32 * 1024);
        private const uint c_MinAllowedFragementSize = (8 * 1024);

        private const uint c_SpareConnectionOnMissingPeerContextSecs = 45;

        private const byte c_ComplBitmapReqMinLen = (UInt128.c_ByteSize + sizeof(ulong) + sizeof(ushort) + sizeof(byte));
        private const byte c_ComplBitmapReqLenInclAICH = (c_ComplBitmapReqMinLen + sizeof(ushort) + SHA_1.c_HashByteSize);
        private const byte c_ComplBitmapReqLenInclTAICH = (c_ComplBitmapReqMinLen + sizeof(ushort) + Tiger.c_HashByteSize);
        private const byte c_ComplBitmapReqInclNontrivialBodyFlag = 0x1;

        public const byte c_ClassicFileRangeRequestSlots = 3;
        private const byte c_ClassicFileRangeRequestULongs = (2 * c_ClassicFileRangeRequestSlots);

        private const byte c_MaxPreAuthPackets = 7;
        private const uint c_MaxPreAuthPacketSize = (10 * 1024);

        private const ushort c_VersionSignatureBufferSize = (sizeof(byte) + (2 * sizeof(uint)) + (4128 / Util.c_BitsPerByte));

        public const byte c_NoTransferID = 0;
        public const ushort c_MaxTransferID = ushort.MaxValue;

        private NetworkManager.ConnResultContext f_ConnResCntxt;
        private Time f_MaxHandshakeTime;
        private bool f_HandshakeFinished;

        private Queue<Packet> f_DeferredPackets;
        private DlPeerContext f_DlPeerCntxt;
        private UlPeerContext f_UlPeerCntxt;
        private GeneralAPCCookie f_UlPeerCntxtCreationCookie;
        private UlPeerCntxtCreationModes f_UlPeerCntxtCreationMode;
        private DownloadHelper f_DlHelper;
        private uint f_FileContentClusterNumber;
        private PeerInfo f_PeerInfo;
        private string f_ClientStr;
        private uint f_PeerAuthKeyStrength;
        private AuthenticationState f_PeerAuthState;

        private uint f_ProcessingGeneration;
        private Time f_EarliestNextSourcesRequest;

        private Time f_ExecuseMissingDlCntxtTime;
        private Time f_ExecuseMissingUlCntxtTime;

        // Since we don't get a file id when receiving a
        // comment, we store the file id from the most
        // recent file name packet.
        private FileID f_SupposedCommentFileID;

        private BitField f_Flags;
        private byte[] f_MyChallengeForPeer;
        private byte[] f_PeersChallengeForMe;
        private byte f_SecIdentVerOfPeer;

        private ushort f_LastTransferID;

        // p_DlPeerCntxt can be null.
        public DlPeerContext p_DlPeerCntxt
        {
            get { return f_DlPeerCntxt; }
        }

        // p_UlPeerCntxt can be null.
        public UlPeerContext p_UlPeerCntxt
        {
            get { return f_UlPeerCntxt; }
        }

        private bool p_UnicodeSupport
        {
            get { return f_PeerInfo.f_UnicodeSupport; }
        }

        public bool p_RestrictToClassicAddresses
        {
            get
            {
                return ((false == f_PeerInfo.p_IsThisClient ||
                         0x10416 > f_PeerInfo.EncodeVersion()) &&
                        NetworkManager.ShouldRestrictPeerToClassicAddresses(p_Addr));
            }
        }

        [Conditional("DEBUG")]
        private void DebugFatal(int a_Code,
                                bool a_Condition)
        {
        }

        [Conditional("DEBUG")]
        private void DebugFatal(int a_Code)
        {
        }

        public DonkeyClient(Instance a_Inst,
                            uint a_DebugConnID,
                            UserContext a_UserCntxt,
                            I2PStream a_Strm,
                            NetworkManager.ConnResultContext a_ConnResCntxt) :
            base(a_Inst, a_DebugConnID, a_UserCntxt, a_Strm)
        {
            // a_Handler may be null.

            f_ConnResCntxt = a_ConnResCntxt;

            if (true == f_Strm.p_Outgoing)
            {
                bool l_OK = SendHello();
            }

            f_ClientStr = a_Inst[TextKeys.UnknownValue];
        }

        public ushort GetNextTransferID()
        {
            // Should only be used by the DlPeerContext class.

            if (false == f_PeerInfo.SupportsTransferIDOpcodes())
            {
                return c_NoTransferID;
            }

            if (c_MaxTransferID == f_LastTransferID)
            {
                f_LastTransferID = 1;
            }
            else
            {
                f_LastTransferID++;
            }

            return f_LastTransferID;
        }

        private static bool IsTransferIDOpcode(OpcodesEx a_OpcodeEx)
        {
            return ((OpcodesEx.TI_DL_OPCODES_START <= a_OpcodeEx &&
                     OpcodesEx.TI_DL_OPCODES_END >= a_OpcodeEx) ||
                    (OpcodesEx.TI_UL_OPCODES_START <= a_OpcodeEx &&
                     OpcodesEx.TI_UL_OPCODES_END >= a_OpcodeEx));
        }

        public void GetPeerInfoVolatile(out ShortPeerInfo a_Dest)
        {
            a_Dest = new ShortPeerInfo(f_PeerInfo.f_UserName, f_ClientStr, f_PeerAuthKeyStrength, f_SecIdentVerOfPeer);
            Util.MemoryBarrier();
        }

        public bool IsiMule()
        {
            // During early stages of the connection the peer's
            // client is not known. In this case false will be
            // returned.

            return f_PeerInfo.p_IsiMule;
        }

        public bool IsThisClient()
        {
            return f_PeerInfo.p_IsThisClient;
        }

        public uint GetClientVersion()
        {
            return f_PeerInfo.EncodeVersion();
        }

        public void SetDlPeerContext(DlPeerContext a_DlPeerCntxt)
        {
            f_DlPeerCntxt = a_DlPeerCntxt;
            f_ExecuseMissingDlCntxtTime = new Time();
        }

        private UlPeerCntxtCreationStates TryBeginCreateUlPeerContext(Packet a_Packet,
                                                                      out UlPeerCntxtCreationModes a_EffMode)
        {
            UlPeerCntxtCreationStates l_State = TryBeginCreateUlPeerContextCore(a_Packet, out a_EffMode);

            if (UlPeerCntxtCreationStates.Pending != l_State &&
                UlPeerCntxtCreationStates.AlreadyUp != l_State)
            {
                // a_EffMode is set to None if this routine returns
                // an error code.

                a_EffMode = UlPeerCntxtCreationModes.None;
            }

            return l_State;
        }

        private bool f_AllowCleanConstructionWhenNextCreatingUlPeerContext;

        private UlPeerCntxtCreationStates TryBeginCreateUlPeerContextCore(Packet a_Packet,
                                                                          out UlPeerCntxtCreationModes a_EffMode)
        {
            bool l_AllowConstruction = f_AllowCleanConstructionWhenNextCreatingUlPeerContext;

            if (true == l_AllowConstruction)
            {
                f_AllowCleanConstructionWhenNextCreatingUlPeerContext = false;
            }

            BufferIO l_Mem = a_Packet.p_Buffer;
            UlPeerCntxtCreationModes l_InMode = (Protocols.OP_EDONKEYPROT == a_Packet.f_Prot ? UlPeerCntxtCreationModes.SendFileStatus : UlPeerCntxtCreationModes.SendComplBitmap);

            a_EffMode = l_InMode;
            bool l_ExpectFileSelFlags = (UlPeerCntxtCreationModes.SendComplBitmap == l_InMode &&
                                         OpcodesEx.OP_COMPLBITMAP_EXT_REQ == a_Packet.p_DonkeyOpcodesEx);
            uint l_Rem = l_Mem.GetRemaining();
            bool l_SizeIsOK;

            if (UlPeerCntxtCreationModes.SendComplBitmap == l_InMode)
            {
                if (true == l_ExpectFileSelFlags)
                {
                    l_Rem--;
                }

                l_SizeIsOK = (c_ComplBitmapReqMinLen == l_Rem ||
                              c_ComplBitmapReqLenInclAICH == l_Rem ||
                              c_ComplBitmapReqLenInclTAICH == l_Rem);
            }
            else
            {
                l_SizeIsOK = (UInt128.c_ByteSize == l_Rem);
            }

            if (false == l_SizeIsOK)
            {
                return UlPeerCntxtCreationStates.Corruption;
            }

            FileID l_FileID = new FileID(l_Mem.ReadUInt128BE());
            ulong l_FileSize = 0;
            P2PHashSets l_MustNotDifferHashType = P2PHashSets.None;
            byte[] l_MustNotDifferHash = null;
            FileSelectionFlags l_FileSelFlags = FileSelectionFlags.All;

            if (UlPeerCntxtCreationModes.SendComplBitmap == l_InMode)
            {
                if (true == l_ExpectFileSelFlags)
                {
                    l_FileSelFlags = (FileSelectionFlags)l_Mem.ReadByte();

                    if (FileSelectionFlags.None != (l_FileSelFlags & ~FileSelectionFlags.All))
                    {
                        return UlPeerCntxtCreationStates.Corruption;
                    }
                }

                l_FileSize = l_Mem.ReadUInt64LE();
                P2PHashSets l_BitmapKind = (P2PHashSets)l_Mem.ReadUInt16LE();
                bool l_IncludesBody = (0 != (c_ComplBitmapReqInclNontrivialBodyFlag & l_Mem.ReadByte()));

                if (P2PHashSets.eD2kBitmap != l_BitmapKind)
                {
                    if (P2PHashSets.ClientNativeBitmap != l_BitmapKind)
                    {
                        return UlPeerCntxtCreationStates.Corruption;
                    }

                    a_EffMode |= UlPeerCntxtCreationModes.UseNativeBitmap;
                }

                if (true == l_IncludesBody)
                {
                    a_EffMode |= UlPeerCntxtCreationModes.IncludeBitmapBodyIfNontrivial;
                }

                if (c_ComplBitmapReqMinLen != l_Rem)
                {
                    // There is an (optional) extra.

                    l_MustNotDifferHashType = (P2PHashSets)l_Mem.ReadUInt16LE();
                    bool l_Sane;

                    // At the moment only two hash types are supported.

                    if (P2PHashSets.AICHTree == l_MustNotDifferHashType)
                    {
                        l_Sane = (c_ComplBitmapReqLenInclAICH == l_Rem);
                    }
                    else if (P2PHashSets.TAICHTree == l_MustNotDifferHashType)
                    {
                        l_Sane = (c_ComplBitmapReqLenInclTAICH == l_Rem);
                    }
                    else
                    {
                        l_Sane = false;
                    }

                    if (false == l_Sane)
                    {
                        return UlPeerCntxtCreationStates.Corruption;
                    }

                    uint l_HashSize = (l_Rem - c_ComplBitmapReqMinLen - sizeof(ushort));
                    l_MustNotDifferHash = l_Mem.ReadBytes(l_HashSize);
                }
            }

            if (null != f_UlPeerCntxt)
            {
                KnownFile l_Known = f_UlPeerCntxt.f_Known;

                if (true == f_UlPeerCntxt.HasFileID(l_FileID) &&
                    (0 == l_FileSize ||
                     l_Known.f_FileSize == l_FileSize))
                {
                    // The file is potentially already up.

                    bool l_AlreadyUp;

                    if (P2PHashSets.None == l_MustNotDifferHashType)
                    {
                        l_AlreadyUp = true;
                    }
                    else
                    {
                        MasterHashes l_MasterHashes;
                        l_Known.f_MasterHashes.Clone(out l_MasterHashes);

                        l_AlreadyUp = (false == l_Known.f_MasterHashes.HasAllHashes(l_MustNotDifferHashType) ||
                                       true == l_Known.f_MasterHashes.IsHashEqual(l_MustNotDifferHashType, l_MustNotDifferHash, 0));
                    }

                    if (true == l_AlreadyUp)
                    {
                        if (FileSelectionFlags.None == (l_FileSelFlags & FileSelectionFlags.AllowConfirmation))
                        {
                            return UlPeerCntxtCreationStates.NotAllowed;
                        }

                        return UlPeerCntxtCreationStates.AlreadyUp;
                    }
                }

                // Try to remove the current context.

                if (false == f_UlPeerCntxt.MaySwitchToOtherFile())
                {
                    return UlPeerCntxtCreationStates.Corruption;
                }

                if (FileSelectionFlags.None == (l_FileSelFlags & FileSelectionFlags.AllowReplacement))
                {
                    return UlPeerCntxtCreationStates.NotAllowed;
                }

                if (null != f_DeferredPackets ||
                    false == BeginRemoveUploadContext(true))
                {
                    return UlPeerCntxtCreationStates.Corruption;
                }

                a_Packet.p_Buffer.JumpToStart();
                f_DeferredPackets = new Queue<Packet>(1);
                f_DeferredPackets.Enqueue(a_Packet);

                f_AllowCleanConstructionWhenNextCreatingUlPeerContext = true;

                return UlPeerCntxtCreationStates.Pending;
            }

            if (false == l_AllowConstruction &&
                FileSelectionFlags.None == (l_FileSelFlags & FileSelectionFlags.AllowConstruction))
            {
                return UlPeerCntxtCreationStates.NotAllowed;
            }

            // Ask the content manager for this file. We expect
            // that any processing of this object halts until
            // TryEndCreateUlPeerContext() changes the status
            // again.

            f_UlPeerCntxtCreationCookie = new GeneralAPCCookie(f_Inst.f_ContentMngr.NetworkGetFileByeD2kHash, new object[4] { l_FileID.f_Val, l_FileSize, l_MustNotDifferHashType, l_MustNotDifferHash });
            f_Inst.f_ContentMngr.EnqueueAPC(f_UlPeerCntxtCreationCookie);
            f_UlPeerCntxtCreationMode = a_EffMode;

            return UlPeerCntxtCreationStates.Pending;
        }

        public bool TryEndCreateUlPeerContext()
        {
            // This routine can be called even while no
            // creation is in progress.

            if (null == f_UlPeerCntxtCreationCookie)
            {
                return true;
            }

            FileID l_FileID = new FileID((UInt128)f_UlPeerCntxtCreationCookie.GetArgAsObjArray()[0]);
            UlPeerCntxtCreationModes l_StoredMode = f_UlPeerCntxtCreationMode;
            f_UlPeerCntxtCreationMode = UlPeerCntxtCreationModes.None;

            if (CookieStates.Success != f_UlPeerCntxtCreationCookie.p_StateVolatile)
            {
                f_UlPeerCntxtCreationCookie = null;

                goto FILE_NOT_ACCESSABLE;
            }

            KnownFile l_Known = (KnownFile)f_UlPeerCntxtCreationCookie.GetReturnObject();
            f_UlPeerCntxtCreationCookie = null;

            if (null == l_Known)
            {
                goto FILE_NOT_ACCESSABLE;
            }

            f_UlPeerCntxt = l_Known.TryCreateUlPeerContext(l_FileID, this);
            f_ExecuseMissingUlCntxtTime = new Time();

            if (null == f_UlPeerCntxt)
            {
                goto FILE_NOT_ACCESSABLE;
            }

            if (false == f_UlPeerCntxt.f_Known.p_eD2kCompleted)
            {
                // Try to add the peer as a source for possible
                // downloads.

                f_Inst.f_NetworkMngr.f_DlQueue.NotifySingleSource(f_UlPeerCntxt.f_Known, p_Addr);
            }

            const UlPeerCntxtCreationModes l_BitmapModes = (UlPeerCntxtCreationModes.SendFileStatus | UlPeerCntxtCreationModes.SendComplBitmap);

            if (0 != (l_BitmapModes & l_StoredMode))
            {
                return SendFileStatusOrComplBitmap(l_StoredMode);
            }

            return SendFileNameAndComment(f_UlPeerCntxt);

        FILE_NOT_ACCESSABLE:

            if (true == f_PeerInfo.p_IsiMule)
            {
                // IMULETWEAK: iMule often sends a lot of opcodes at
                // once without waiting if the file is accessable.

                SpareIfUlContextIsMissing();
            }

            return SendNoSuchFileNotification(l_FileID);
        }

        public void DeleteDlPeerContext()
        {
            // May only be called by the KnownFile class.

            f_DlPeerCntxt = null;
            f_DlHelper.Invalidate();
        }

        public void DeleteUlPeerContext()
        {
            // May only be called by the KnownFile class.

            f_UlPeerCntxt = null;
        }

        private bool ShouldDeferProcessing()
        {
            // Returns true if a context needs to be removed.

            return ((null != f_DlPeerCntxt &&
                     true == f_DlPeerCntxt.ShouldBeRemoved()) ||
                    (null != f_UlPeerCntxt &&
                     true == f_UlPeerCntxt.ShouldBeRemoved()));
        }

        protected override bool ShouldTestForLife()
        {
            return (null != f_DlPeerCntxt ||
                    null != f_UlPeerCntxt ||
                    true == p_Outgoing);
        }

        public void Shutdown()
        {
            Close();

            if (null != f_ConnResCntxt)
            {
                f_ConnResCntxt.Invoke(f_Inst, ConnectResult.HandshakeError);
                f_ConnResCntxt = null;
            }
        }

        private PeerContext ConfirmPeerContext(Packet a_Packet,
                                               PeerContextRequirements a_Reqs,
                                               out bool a_Excusable)
        {
            if (UInt128.c_ByteSize != a_Packet.p_Buffer.GetRemaining())
            {
                a_Excusable = false;

                return null;
            }

            uint l_Junk;
            return ConfirmPeerContext(a_Packet.p_Buffer, a_Reqs, out l_Junk, out a_Excusable);
        }

        private PeerContext ConfirmPeerContext(BufferIO a_Buffer,
                                               PeerContextRequirements a_Reqs,
                                               out uint a_Remaining,
                                               out bool a_Excusable)
        {
            a_Remaining = a_Buffer.GetRemaining();
            a_Excusable = false;

            if (UInt128.c_ByteSize > a_Remaining)
            {
                return null;
            }

            FileID l_FileID = new FileID(a_Buffer.ReadUInt128BE());
            a_Remaining -= UInt128.c_ByteSize;

            // Always return the UlPeerContext first if it fits.

            if (0 != (PeerContextRequirements.Ul & a_Reqs))
            {
                if (null != f_UlPeerCntxt &&
                    true == f_UlPeerCntxt.HasFileID(l_FileID))
                {
                    return f_UlPeerCntxt;
                }

                a_Excusable = TryExcuseMissingUlContext();
            }

            if (0 != (PeerContextRequirements.Dl & a_Reqs))
            {
                if (null != f_DlPeerCntxt &&
                    true == f_DlPeerCntxt.HasFileID(l_FileID))
                {
                    return f_DlPeerCntxt;
                }

                if (false == a_Excusable)
                {
                    a_Excusable = TryExcuseMissingDlContext();
                }
            }

            return null;
        }

        private bool ConfirmPeerContexts(BufferIO a_Buffer,
                                         out DlPeerContext a_DlPeerCntxt,
                                         out UlPeerContext a_UlPeerCntxt,
                                         out uint a_Remaining,
                                         out bool a_Excusable)
        {
            // Returns false if none of the peer contexts fits.

            a_Remaining = a_Buffer.GetRemaining();

            if (UInt128.c_ByteSize > a_Remaining)
            {
                a_DlPeerCntxt = null;
                a_UlPeerCntxt = null;
                a_Excusable = false;

                return false;
            }

            FileID l_FileID = new FileID(a_Buffer.ReadUInt128BE());
            a_Remaining -= UInt128.c_ByteSize;

            if (null != f_DlPeerCntxt &&
                true == f_DlPeerCntxt.HasFileID(l_FileID))
            {
                a_DlPeerCntxt = f_DlPeerCntxt;
            }
            else
            {
                a_DlPeerCntxt = null;
            }

            if (null != f_UlPeerCntxt &&
                true == f_UlPeerCntxt.HasFileID(l_FileID))
            {
                a_UlPeerCntxt = f_UlPeerCntxt;
            }
            else
            {
                a_UlPeerCntxt = null;

                if (null == a_DlPeerCntxt)
                {
                    a_Excusable = TryExcuseAnyMissingContext();

                    return false;
                }
            }

            a_Excusable = false;

            return true;
        }

        public void SpareIfDlContextIsMissing()
        {
            f_ExecuseMissingDlCntxtTime = f_Inst.GetTime().AddSeconds(c_SpareConnectionOnMissingPeerContextSecs);
        }

        public void SpareIfUlContextIsMissing()
        {
            f_ExecuseMissingUlCntxtTime = f_Inst.GetTime().AddSeconds(c_SpareConnectionOnMissingPeerContextSecs);
        }

        private bool TryExcuseMissingDlContext()
        {
            return (true == f_ExecuseMissingDlCntxtTime.IsSet() &&
                    true == f_Inst.GetTime().IsEarlierThan(f_ExecuseMissingDlCntxtTime));
        }

        private bool TryExcuseMissingUlContext()
        {
            return (true == f_ExecuseMissingUlCntxtTime.IsSet() &&
                    true == f_Inst.GetTime().IsEarlierThan(f_ExecuseMissingUlCntxtTime));
        }

        private bool TryExcuseAnyMissingContext()
        {
            return (true == TryExcuseMissingDlContext() ||
                    true == TryExcuseMissingUlContext());
        }

        private void NotifyReceive(Packet a_Packet)
        {
            f_UserCntxt.Rcvd(a_Packet);
        }

        private void NotifyReceiveIfOutgoing(Packet a_Packet)
        {
            if (true == p_Outgoing)
            {
                f_UserCntxt.Rcvd(a_Packet);
            }
        }

        public bool SupportsFileDescriptors()
        {
            return f_PeerInfo.SupportBasicOpcodeExtensions();
        }

        public bool SupportsAdvancedFileMetaData()
        {
            return f_PeerInfo.SupportBasicOpcodeExtensions();
        }

        private bool IsDlPeerContextGenerationPending()
        {
            return (true == p_Outgoing &&
                    false == f_HandshakeFinished);
        }

        private bool AllowPeerToGoOnWithFailedAuthentication()
        {
            if (false == c_AllowAuthenticationOfPeerToFail ||
                false == f_PeerInfo.p_IsiMule)
            {
                return false;
            }

            UserContext l_UserCntxt = p_UserCntxt;
            uint l_Junk;

            return (false == l_UserCntxt.HasPublicKeyWithMinimumSize(OwnUserPersona.c_MaxKeyBitsiMule + 1) &&
                    false == l_UserCntxt.HasAnyCurrentSecIdentAuthenticatedConnection(out l_Junk));
        }

        private bool DoesPeerNeedToProvideSecIdentMaterial()
        {
            // Returns a nonzero value if the peer still has to provide
            // sec ident packets to authenticate him-/herself.

            return !f_Flags.CheckAnyCombination(FlagsToUInt(StreamFlags.RcvdPeersSignature | StreamFlags.SecIdentFastTrackIndentifiedPeer));
        }

        public bool SuccessfullyIdentifiedPeerWithSecIdent()
        {
            // The return value of this routine does not imply that
            // the peer has identified us.

            return f_Flags.CheckAnyCombination(FlagsToUInt(StreamFlags.SecIdentFullyIdentifiedPeer | StreamFlags.SecIdentFastTrackIndentifiedPeer));
        }

        private void FinishedHandshake(bool a_HadSecIdent)
        {
            // Note that a_HadSecIdent should also be nonzero if an
            // attempted SecIdent somehow failed.

            UserContext l_UserCntxt = p_UserCntxt;

            // First update the A4AF status for this user.
            l_UserCntxt.AdjustA4AFState(f_PeerInfo.HasA4AFLimitation());
            f_Inst.f_UserMngr.SetDatagramDest(l_UserCntxt, f_PeerInfo.f_DatagramAddr);
            l_UserCntxt.NotifyUserName(f_PeerInfo.f_UserName);

            f_HandshakeFinished = true;
            f_PeerAuthState = (false == SuccessfullyIdentifiedPeerWithSecIdent() ? (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SecIdentAttemptToIdentifyPeerFailed)) ? AuthenticationState.NoAuth : AuthenticationState.FailedAuth) : AuthenticationState.SuccessfulAuth);

            if (false == Software.s_LaterVersionAvailable &&
                true == f_PeerInfo.p_IsThisClient &&
                f_PeerInfo.EncodeVersion() > Software.s_ReleaseInfo.EncodeVersion())
            {
                // Try to confirm that that there is a new version
                // available.

                bool l_Junk = RequestVersionSignature();
            }

            // Now call the handler (if any).

            if (null != f_ConnResCntxt)
            {
                f_ConnResCntxt.Invoke(f_Inst, ConnectResult.Success, this);
                f_ConnResCntxt = null;
            }

            f_PeersChallengeForMe = f_MyChallengeForPeer = null;
        }

        public bool WasProcessed(uint a_Generation)
        {
            return (a_Generation == f_ProcessingGeneration);
        }

        public bool Process(Time a_Time)
        {
            // Returns false if the connections should be
            // closed.

            if (false == f_HandshakeFinished &&
                true == f_MaxHandshakeTime.IsSet() &&
                a_Time >= f_MaxHandshakeTime)
            {
                return false;
            }

            f_ProcessingGeneration = f_Inst.f_NetworkMngr.p_ProcessingGeneration;

            while (false == f_HandshakeFinished ||
                   (null == f_UlPeerCntxtCreationCookie &&
                    false == ShouldDeferProcessing()))
            {
                Packet l_Packet;
                bool l_Ready;

                if (true == f_HandshakeFinished &&
                    null != f_DeferredPackets)
                {
                    // We have already tried to handle this packet,
                    // but we weren't ready then.

                    l_Packet = f_DeferredPackets.Dequeue();

                    if (0 == f_DeferredPackets.Count)
                    {
                        f_DeferredPackets = null;
                    }
                }
                else
                {
                    l_Packet = base.DequeuePacket();

                    if (null == l_Packet)
                    {
                        break;
                    }
                }

                if (false == f_HandshakeFinished)
                {
                    l_Ready = ProcessHandshake(l_Packet);
                }
                else
                {
                    l_Ready = ProcessAfterHandshake(l_Packet);
                }

                if (false == l_Ready)
                {
                    return false;
                }
            }

            return true;
        }

        private bool ProcessHandshake(Packet a_Packet)
        {
            if (Protocols.OP_EDONKEYPROT == a_Packet.f_Prot)
            {
                switch (a_Packet.p_DonkeyOpcode)
                {
                    case Opcodes.OP_HELLO:
                    case Opcodes.OP_HELLOANSWER:
                        return ProcessHello(a_Packet);
                }
            }
            else if (Protocols.OP_EMULEPROT == a_Packet.f_Prot)
            {
                if (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.ReceivedHello)))
                {
                    switch (a_Packet.p_DonkeyOpcode)
                    {
                        case Opcodes.OP_SECIDENTSTATE:
                            return ProcessSecIdentState(a_Packet);

                        case Opcodes.OP_EMULEINFO:
                            return ProcessMuleInfoPacket(a_Packet);

                        case Opcodes.OP_PUBLICKEY:
                            return ProcessPublicKey(a_Packet);

                        case Opcodes.OP_SIGNATURE:
                            return ProcessSignature(a_Packet);
                    }
                }
            }
            else if (OpcodesEx.OP_PONG >= a_Packet.p_DonkeyOpcodesEx ||
                     OpcodesEx.OP_SECIDENTSTATE_EXT == a_Packet.p_DonkeyOpcodesEx)
            {
                bool l_Res;

                if (true == ProcessExtensionOpcode(a_Packet, out l_Res))
                {
                    return l_Res;
                }
            }

            // If we land here this packet doesn't belong to
            // the handshake and we will refrain from processing
            // it for now.

            if (null == f_DeferredPackets)
            {
                f_DeferredPackets = new Queue<Packet>(1);
            }
            else if (c_MaxPreAuthPackets <= f_DeferredPackets.Count ||
                     c_MaxPreAuthPacketSize < a_Packet.p_Buffer.p_Length)
            {
                return false;
            }

            f_DeferredPackets.Enqueue(a_Packet);

            return true;
        }

        private bool ProcessAfterHandshake(Packet a_Packet)
        {
            if (Protocols.OP_EDONKEYPROT == a_Packet.f_Prot)
            {
                switch (a_Packet.p_DonkeyOpcode)
                {
                    case Opcodes.OP_SENDINGPART:
                        return ProcessFileContent(a_Packet, false);

                    case Opcodes.OP_REQUESTPARTS:
                        return ProcessFileRangeRequest(a_Packet);

                    case Opcodes.OP_REQUESTFILENAME:
                        return ProcessFileNameRequest(a_Packet);

                    case Opcodes.OP_REQFILENAMEANSWER:
                        return ProcessFileName(a_Packet);

                    case Opcodes.OP_FILESTATUS:
                        return ProcessFileStatusOrComplBitmap(a_Packet);

                    case Opcodes.OP_FILEREQANSNOFIL:
                        return ProcessFileReqAnswerNoFile(a_Packet);

                    case Opcodes.OP_HASHSETANSWER:
                        return ProcessHashSet(a_Packet);

                    case Opcodes.OP_HASHSETREQUEST:
                        return ProcessHashSetRequest(a_Packet);

                    case Opcodes.OP_CANCELTRANSFER:
                    case Opcodes.OP_END_OF_DOWNLOAD:
                        return ProcessDeselectDlFileMessage(a_Packet);

                    case Opcodes.OP_SETREQFILEID:
                        return ProcessFileRequest(a_Packet);

                    case Opcodes.OP_STARTUPLOADREQ:
                        return ProcessStartUploadRequest(a_Packet);

                    case Opcodes.OP_QUEUERANK:
                        return ProcessQueueRanking(a_Packet);

                    case Opcodes.OP_ACCEPTUPLOADREQ:
                        return ProcessAcceptUploadRequest(a_Packet);

                    case Opcodes.OP_MESSAGE:
                        // Ignored for now.
                        return true;

                    case Opcodes.OP_OUTOFPARTREQS:
                        return ProcessOnQueueNotification(a_Packet);

                    case Opcodes.OP_ASKSHAREDFILES:
                    case Opcodes.OP_ASKSHAREDDIRS:
                    case Opcodes.OP_ASKSHAREDFILESDIR:
                        return ProcessRevealSharedObjectsRequest(a_Packet);
                }
            }
            else if (Protocols.OP_EMULEPROT == a_Packet.f_Prot)
            {
                switch (a_Packet.p_DonkeyOpcode)
                {
                    case Opcodes.OP_COMPRESSEDPART:
                        return ProcessFileContent(a_Packet, true);

                    case Opcodes.OP_SENDINGPART:
                        return ProcessFileContent(a_Packet, false);

                    case Opcodes.OP_REQUESTPARTS:
                        return ProcessFileRangeRequest(a_Packet);

                    case Opcodes.OP_QUEUERANKING:
                        return ProcessQueueRanking(a_Packet);

                    case Opcodes.OP_FILEDESC:
                        return ProcessFileComment(a_Packet);

                    case Opcodes.OP_MULTIPACKET:
                        return ProcessMultiPacket(a_Packet);

                    case Opcodes.OP_REQUESTSOURCES:
                        return ProcessSourcesRequest(a_Packet);

                    case Opcodes.OP_ANSWERSOURCES:
                        return ProcessSourcesAnswer(a_Packet);

                    // This opcode can occur quite late, especially
                    // if we don't use secure identification.
                    case Opcodes.OP_EMULEINFO:
                        return ProcessMuleInfoPacket(a_Packet);

                    case Opcodes.OP_AICHFILEHASHREQ:
                        return ProcessAICHMasterHashRequest(a_Packet);

                    case Opcodes.OP_AICHREQUEST:
                        return ProcessAICHRequest(a_Packet);

                    case Opcodes.OP_AICHANSWER:
                        return true;

                    case Opcodes.OP_AICHFILEHASHANS:
                        return true;
                }
            }
            else
            {
                bool l_Res;

                if (true == ProcessExtensionOpcode(a_Packet, out l_Res))
                {
                    return l_Res;
                }
            }

            // We don't accept any other codes for now.

            DebugFatal(99);

            return false;
        }

        private bool ProcessExtensionOpcode(Packet a_Packet,
                                            out bool a_Result)
        {
            // Returns a value indicating if the opcode was handled
            // (zero if it's unknown).

            OpcodesEx l_OpcodeEx = a_Packet.p_DonkeyOpcodesEx;

            if (OpcodesEx.LAST_LOW_EXT_OPCODE >= l_OpcodeEx)
            {
                a_Result = ProcessLowOpcodeExtension(a_Packet);

                return true;
            }

            if (true == IsTransferIDOpcode(l_OpcodeEx))
            {
                a_Result = ProcessTransferIDOpcode(a_Packet);

                return true;
            }

            a_Result = false;

            return false;
        }

        private bool ProcessLowOpcodeExtension(Packet a_Packet)
        {
            switch (a_Packet.p_DonkeyOpcodesEx)
            {
                case OpcodesEx.OP_NOOP:
                    return true;

                case OpcodesEx.OP_PING:
                    return ProcessPing(a_Packet);

                case OpcodesEx.OP_PONG:
                    return true;

                case OpcodesEx.OP_FILEMETADATA_REQ:
                    return ProcessAdvancedFileMetaDataRequest(a_Packet);

                case OpcodesEx.OP_FILEMETADATA_ANS:
                    return ProcessAdvancedFileMetaData(a_Packet);

                case OpcodesEx.OP_FILEDESCRIPTOR_REQ:
                    return ProcessFileDescriptorRequest(a_Packet);

                case OpcodesEx.OP_FILEDESCRIPTOR_ANS:
                    return ProcessFileDescriptor(a_Packet);

                case OpcodesEx.OP_DESELECT_UPLOAD_FILE:
                    return ProcessDeselectUlFileMessage(a_Packet);

                case OpcodesEx.OP_COMPLBITMAP_REQ:
                case OpcodesEx.OP_COMPLBITMAP_EXT_REQ:
                    return ProcessFileRequest(a_Packet);

                case OpcodesEx.OP_COMPLBITMAP_ANS:
                    return ProcessFileStatusOrComplBitmap(a_Packet);

                case OpcodesEx.OP_VERSIONSIGNATURE_REQ:
                    return ProcessVersionSignatureRequest(a_Packet);

                case OpcodesEx.OP_VERSIONSIGNATURE_ANS:
                    return ProcessVersionSignature(a_Packet);

                case OpcodesEx.OP_SECIDENTSTATE_EXT:
                    return ProcessSecIdentState(a_Packet);

                case OpcodesEx.OP_FILE_SELECTION_NOT_ALLOWED_INFO:
                    return a_Packet.p_Buffer.EndReached();

                default: throw (new UnexpectedStateException());
            }
        }

        private bool ProcessTransferIDOpcode(Packet a_Packet)
        {
            OpcodesEx l_OpcodeEx = a_Packet.p_DonkeyOpcodesEx;

            BufferIO l_Mem = a_Packet.p_Buffer;
            uint l_TransferID;

            if (false == TryReadTransferID(l_Mem, out l_TransferID))
            {
                return false;
            }

            bool l_ForDlCntxt = (OpcodesEx.TI_UL_OPCODES_START <= l_OpcodeEx);
            bool l_Res;

            if (false == l_ForDlCntxt)
            {
                UlPeerContext l_UlPeerCntxt = f_UlPeerCntxt;
                bool l_Excuse;

                if (null == l_UlPeerCntxt)
                {
                    l_Excuse = TryExcuseMissingUlContext();

                    return l_Excuse;
                }

                if (OpcodesEx.TI_UPLOAD_REQUEST != l_OpcodeEx &&
                    false == l_UlPeerCntxt.IsTransferIDRelevant(l_TransferID, out l_Excuse))
                {
                    return l_Excuse;
                }

                uint l_Remaining = l_Mem.GetRemaining();

                switch (l_OpcodeEx)
                {
                    case OpcodesEx.TI_UPLOAD_REQUEST:
                        {
                            uint l_AuxTransferID = c_NoTransferID;

                            if ((sizeof(ushort) + UInt128.c_ByteSize) == l_Remaining &&
                                true == TryReadTransferID(l_Mem, out l_AuxTransferID))
                            {
                                UInt128 l_FileID = l_Mem.ReadUInt128BE();

                                if (true == l_UlPeerCntxt.f_FileID.f_Val.ValueEquals(l_FileID))
                                {
                                    l_Res = l_UlPeerCntxt.NotifyStartUploadRequest(l_TransferID, l_AuxTransferID);
                                }
                                else
                                {
                                    l_Res = false;
                                }
                            }
                            else
                            {
                                l_Res = false;
                            }
                        }
                        break;

                    case OpcodesEx.TI_QUEUE_RANK_REQUEST:
                        {
                            if (0 == l_Remaining)
                            {
                                l_Res = l_UlPeerCntxt.NotifyQueueRankRequest();
                            }
                            else
                            {
                                l_Res = false;
                            }
                        }
                        break;

                    case OpcodesEx.TI_RANGE_REQUEST:
                        {
                            if ((2 * sizeof(byte)) < l_Remaining)
                            {
                                l_Res = ProcessFileRangeRequestCore(l_Mem, true);
                            }
                            else
                            {
                                l_Res = false;
                            }
                        }
                        break;

                    case OpcodesEx.TI_REFUSE_UPLOAD_ACKNOWLEDGEMENT:
                        {
                            uint l_AuxTransferID;

                            if (sizeof(ushort) == l_Remaining &&
                                true == TryReadTransferID(l_Mem, out l_AuxTransferID))
                            {
                                l_Res = l_UlPeerCntxt.NotifyRefuseUploadAcknowledgement(l_AuxTransferID);
                            }
                            else
                            {
                                l_Res = false;
                            }
                        }
                        break;

                    case OpcodesEx.TI_END_OF_DOWNLOAD_INFO:
                        {
                            if (0 == l_Remaining)
                            {
                                l_Res = l_UlPeerCntxt.NotifyEndOfDownload();
                            }
                            else
                            {
                                l_Res = false;
                            }
                        }
                        break;

                    default: throw (new UnexpectedStateException());
                }
            }
            else
            {
                DlPeerContext l_DlPeerCntxt = f_DlPeerCntxt;
                bool l_Excuse;

                if (null == l_DlPeerCntxt)
                {
                    l_Excuse = TryExcuseMissingDlContext();

                    return l_Excuse;
                }

                if (false == l_DlPeerCntxt.IsTransferIDRelevant(l_TransferID, out l_Excuse))
                {
                    return l_Excuse;
                }

                uint l_ExpectedRemSize = (OpcodesEx.TI_QUEUE_RANK_INFO == l_OpcodeEx ? sizeof(ushort) : 0u);

                if (l_Mem.GetRemaining() != l_ExpectedRemSize)
                {
                    return false;
                }

                switch (l_OpcodeEx)
                {
                    case OpcodesEx.TI_QUEUE_RANK_INFO:
                        l_Res = l_DlPeerCntxt.SetQueueRank(l_Mem.ReadUInt16LE());
                        break;

                    case OpcodesEx.TI_ACCEPT_UPLOAD_INFO:
                        l_Res = l_DlPeerCntxt.NotifyUploadRequestAccepted(false);
                        break;

                    case OpcodesEx.TI_REFUSE_UPLOAD_INFO:
                        l_Res = l_DlPeerCntxt.NotifyOnQueueAgain(false);
                        break;

                    case OpcodesEx.TI_END_OF_UPLOAD_INFO:
                        l_Res = l_DlPeerCntxt.NotifyEndOfUpload();
                        break;

                    default: throw (new UnexpectedStateException());
                }

                if (true == l_Res)
                {
                    NotifyReceive(a_Packet);
                }
            }

            return l_Res;
        }

        private bool TryReadTransferID(BufferIO a_MemIO,
                                       out uint a_TransferID)
        {
            a_TransferID = a_MemIO.ReadUInt16LE();

            return (c_NoTransferID != a_TransferID);
        }

        public bool ReadStringUntrusted(BufferIO a_Buffer,
                                        uint a_ByteLen,
                                        out string a_Str)
        {
            // If a_ByteLen is fine, this should never prove
            // problematic, but catch exceptions in any case.

            try
            {
                a_Str = a_Buffer.ReadStringWithAutoDetection(a_ByteLen, p_UnicodeSupport);

                return true;
            }
            catch (BufferIO.InterpretBufferException)
            {
                a_Str = null;

                return false;
            }
        }

        public void WriteString16(BufferIO a_Buffer,
                                  string a_Str)
        {
            if (false == p_UnicodeSupport)
            {
                a_Buffer.WriteSBStringUInt16Len(a_Str);
            }
            else
            {
                a_Buffer.WriteUTF8StringUInt16Len(a_Str);
            }
        }

        public void WriteString32(BufferIO a_Buffer,
                                  string a_Str)
        {
            if (false == p_UnicodeSupport)
            {
                a_Buffer.WriteSBStringUInt32Len(a_Str);
            }
            else
            {
                a_Buffer.WriteUTF8StringUInt32Len(a_Str);
            }
        }

        private bool SendHello()
        {
            Opcodes l_HelloOpcode;
            ByteArray l_Data;

            if (false == p_Outgoing)
            {
                l_HelloOpcode = Opcodes.OP_HELLOANSWER;
                l_Data = new ByteArray(f_Inst.f_NetworkMngr.f_HelloAnswerPacketBody);
            }
            else
            {
                l_HelloOpcode = Opcodes.OP_HELLO;
                l_Data = new ByteArray(f_Inst.f_NetworkMngr.f_HelloPacketBody);
            }

            if (false == EnqueuePacket(l_HelloOpcode, PacketSendingFlags.IncUserCountIfIncoming, ref l_Data))
            {
                return false;
            }

            if (true == f_Inst.f_NetworkMngr.f_SelfInfo.f_IsMule)
            {
                l_Data = new ByteArray(f_Inst.f_NetworkMngr.f_MuleInfoPacketBody);
                if (false == EnqueuePacket(Opcodes.OP_EMULEINFO, PacketSendingFlags.IncUserCountIfIncoming, ref l_Data))
                {
                    return false;
                }
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.SentHello));

            return true;
        }

        private bool EnqueueFileIDPacket(Opcodes a_Opcode,
                                         PacketSendingFlags a_Flags,
                                         FileID a_FileID)
        {
            return (null != EnqueueFileIDPacketCore(a_Opcode, OpcodesEx.Undefined, a_Flags, a_FileID));
        }

        private bool EnqueueFileIDPacket(OpcodesEx a_OpcodeEx,
                                         PacketSendingFlags a_Flags,
                                         FileID a_FileID)
        {
            return (null != EnqueueFileIDPacketCore(Opcodes.NONE, a_OpcodeEx, a_Flags, a_FileID));
        }

        private Packet EnqueueFileIDPacketCore(Opcodes a_Opcode,
                                               OpcodesEx a_OpcodeEx,
                                               PacketSendingFlags a_Flags,
                                               FileID a_FileID)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable(UInt128.c_ByteSize);

            l_Buffer.WriteUInt128BE(a_FileID.f_Val);

            return EnqueuePacketCore(a_Opcode, a_OpcodeEx, a_Flags, l_Buffer);
        }

        private bool EnqueueEmptyPacket(Opcodes a_Opcode,
                                        PacketSendingFlags a_Flags)
        {
            return EnqueuePacket(a_Opcode, a_Flags, null);
        }

        private bool EnqueueEmptyPacket(OpcodesEx a_OpcodeEx,
                                        PacketSendingFlags a_Flags)
        {
            return EnqueuePacket(a_OpcodeEx, a_Flags, null);
        }

        private bool EnqueueEmptyTransferIDPacket(OpcodesEx a_OpcodeEx,
                                                  PacketSendingFlags a_Flags,
                                                  uint a_TransferID)
        {
            ByteArray l_Body = new ByteArray(sizeof(ushort));
            Util.WriteUInt16LE((ushort)a_TransferID, l_Body.f_Arr, 0);

            return EnqueuePacket(a_OpcodeEx, a_Flags, ref l_Body);
        }

        private bool EnqueuePacket(Opcodes a_Opcode,
                                   PacketSendingFlags a_Flags,
                                   MemoryIO a_Buffer)
        {
            return (null != EnqueuePacketCore(a_Opcode, OpcodesEx.Undefined, a_Flags, a_Buffer));
        }

        private bool EnqueuePacket(OpcodesEx a_OpcodeEx,
                                   PacketSendingFlags a_Flags,
                                   MemoryIO a_Buffer)
        {
            return (null != EnqueuePacketCore(Opcodes.NONE, a_OpcodeEx, a_Flags, a_Buffer));
        }

        private Packet EnqueuePacketCore(Opcodes a_Opcode,
                                         OpcodesEx a_OpcodeEx,
                                         PacketSendingFlags a_Flags,
                                         MemoryIO a_Buffer)
        {
            // a_Buffer may be null, otherwise this instance takes
            // ownership. Return can be null.

            ByteArray l_Data;

            if (null == a_Buffer)
            {
                l_Data = new ByteArray(0);
            }
            else
            {
                l_Data = new ByteArray(a_Buffer.ToArray(true));
            }

            return EnqueuePacketCore(a_Opcode, a_OpcodeEx, a_Flags, ref l_Data);
        }

        private bool EnqueuePacket(Opcodes a_Opcode,
                                   PacketSendingFlags a_Flags,
                                   ref ByteArray a_Body)
        {
            return (null != EnqueuePacketCore(a_Opcode, OpcodesEx.Undefined, a_Flags, ref a_Body));
        }

        private bool EnqueuePacket(OpcodesEx a_OpcodeEx,
                                   PacketSendingFlags a_Flags,
                                   ref ByteArray a_Body)
        {
            return (null != EnqueuePacketCore(Opcodes.NONE, a_OpcodeEx, a_Flags, ref a_Body));
        }

        private Packet EnqueuePacketCore(Opcodes a_Opcode,
                                         OpcodesEx a_OpcodeEx,
                                         PacketSendingFlags a_Flags,
                                         ref ByteArray a_BodyData)
        {
            // a_BodyData.f_Len may be zero. Return can be null.

            bool l_AllowCompression = (0 == (a_Flags & PacketSendingFlags.NoCompression) &&
                                       true == f_Inst.f_NetworkMngr.f_NetworkPrefs.f_AllowDataCompression);
            Packet l_Packet;

            if (OpcodesEx.Undefined == a_OpcodeEx)
            {
                KnownFile l_Known = (null == f_UlPeerCntxt ? null : f_UlPeerCntxt.f_Known);
                l_Packet = Packet.BuildOwnPacket(p_Addr, a_Opcode, l_Known, l_AllowCompression, ref a_BodyData);
            }
            else
            {
                bool l_LegacyTranslation = (false == f_PeerInfo.SupportsNBProtocol() &&
                                            OpcodesEx.OP_NOOP <= a_OpcodeEx &&
                                            OpcodesEx.LAST_LEGACY_EXT_OPCODE >= a_OpcodeEx);

                if (true == l_LegacyTranslation)
                {
                    l_AllowCompression = false;
                }

                l_Packet = Packet.BuildOwnPacket(p_Addr, a_OpcodeEx, l_AllowCompression, l_LegacyTranslation, ref a_BodyData);
            }

            if (0 != (a_Flags & PacketSendingFlags.IncUsersCount))
            {
                l_Packet.e_SentBytes += f_UserCntxt.Sent;
            }
            else if (false == p_Outgoing &&
                     0 != (a_Flags & PacketSendingFlags.IncUserCountIfIncoming))
            {
                l_Packet.e_SentBytes += f_UserCntxt.Sent;
            }

            if (null != f_UlPeerCntxt &&
                0 != (a_Flags & PacketSendingFlags.IncFilesCount))
            {
                l_Packet.e_SentBytes += f_UlPeerCntxt.NotifySentContentBytes;
            }

            bool l_InsertAtEndOfQueue = (0 != (a_Flags & PacketSendingFlags.InsertAtEndOfQueue));
            PacketInsertionMode l_InsertionMode = (false == l_InsertAtEndOfQueue ? PacketInsertionMode.MoveAheadAsFarAsPossible : PacketInsertionMode.End);

            if (0 != (a_Flags & PacketSendingFlags.MayBeOutpaced))
            {
                l_Packet.SetReorderableToBack();
            }

            if (true == EnqueuePacketBase(l_Packet, l_InsertionMode))
            {
                return l_Packet;
            }

            return null;
        }

        private static MemoryIO CreateTransferIDBuffer(uint a_TransferID,
                                                       uint a_ExtraCapacity)
        {
            uint l_InitialCapacity = (sizeof(ushort) + a_ExtraCapacity);
            MemoryIO l_Mem = MemoryIO.CreateWritableAndGrowable(l_InitialCapacity);

            l_Mem.WriteUInt16LE((ushort)a_TransferID);

            return l_Mem;
        }

        private static MemoryIO CreateTransferIDBuffer(uint a_PrimTransferID,
                                                       uint a_AuxTransferID,
                                                       uint a_ExtraCapacity)
        {
            MemoryIO l_Mem = CreateTransferIDBuffer(a_PrimTransferID, sizeof(ushort) + a_ExtraCapacity);

            l_Mem.WriteUInt16LE((ushort)a_AuxTransferID);

            return l_Mem;
        }

        private bool SendSecIdentState()
        {
            SecureIdentStates l_State;

            if (false == f_UserCntxt.HasPublicKey())
            {
                l_State = SecureIdentStates.IS_KEYANDSIGNEEDED;
            }
            else if (true == f_PeerInfo.SupportsFastTrackAuthentication() &&
                     true == f_UserCntxt.HasAnyCurrentSecIdentAuthenticatedConnection(out f_PeerAuthKeyStrength))
            {
                // There is already a connection to that peer currently
                // that successfully handled authentication, so that we
                // can skip doing it again for this connection.

                l_State = SecureIdentStates.IS_AUTHENTICATED_INSTANTLY;

                f_Flags.SetFlag(FlagsToUInt(StreamFlags.SecIdentFastTrackIndentifiedPeer));
            }
            else if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RcvdPeersSignature)))
            {
                l_State = SecureIdentStates.IS_SIGNATURENEEDED;
            }
            else
            {
                return true;
            }

            byte l_SecIdentVer = f_PeerInfo.f_SuggestedSecIdentVer;
            bool l_ExtOpcode = (4 <= l_SecIdentVer);
            uint l_ChallengeLen = (false == l_ExtOpcode ? sizeof(uint) : 20u);

            MemoryIO l_AnsBuffer = MemoryIO.CreateWritable((uint)Util.BoolToByte(l_ExtOpcode) + sizeof(byte) + (SecureIdentStates.IS_AUTHENTICATED_INSTANTLY == l_State ? 0 : l_ChallengeLen));

            if (true == l_ExtOpcode)
            {
                l_AnsBuffer.WriteByte(l_SecIdentVer);
            }

            l_AnsBuffer.WriteByte((byte)l_State);

            if (SecureIdentStates.IS_AUTHENTICATED_INSTANTLY != l_State)
            {
                f_MyChallengeForPeer = Util.GetRandomBytes((int)l_ChallengeLen);

                l_AnsBuffer.WriteBytes(f_MyChallengeForPeer);

                f_MaxHandshakeTime = f_Inst.GetTime().AddSeconds(5 * Time.c_SecsPerMinute);
            }

            if (false == l_ExtOpcode)
            {
                if (false == EnqueuePacket(Opcodes.OP_SECIDENTSTATE, PacketSendingFlags.IncUserCountIfIncoming, l_AnsBuffer))
                {
                    return false;
                }
            }
            else if (false == EnqueuePacket(OpcodesEx.OP_SECIDENTSTATE_EXT, PacketSendingFlags.IncUserCountIfIncoming, l_AnsBuffer))
            {
                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.SentSecIdentState));

            return true;
        }

        private bool ProcessSecIdentState(Packet a_Packet)
        {
            // This opcode may be processed several times.

            bool l_ExtOpcode = (Protocols.PROT_NB_PLAIN == a_Packet.f_Prot);

            if (0 == f_PeerInfo.f_SuggestedSecIdentVer)
            {
                DebugFatal(100);

                return false;
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining = l_Buffer.GetRemaining();

            if (((uint)Util.BoolToByte(l_ExtOpcode) + sizeof(byte)) > l_Remaining)
            {
                DebugFatal(101);

                return false;
            }

            byte l_SecIdentVer;

            if (false == l_ExtOpcode)
            {
                l_SecIdentVer = f_PeerInfo.f_SuggestedSecIdentVer;
            }
            else
            {
                l_SecIdentVer = l_Buffer.ReadByte();

                if (0 == l_SecIdentVer ||
                    0 == (f_Inst.f_NetworkMngr.f_SelfInfo.f_SupportSecIdent & (1u << (l_SecIdentVer - 1))))
                {
                    DebugFatal(102);

                    return false;
                }

                l_Remaining--;
            }

            if (0 == f_SecIdentVerOfPeer)
            {
                f_SecIdentVerOfPeer = l_SecIdentVer;
            }
            else if (l_SecIdentVer != f_SecIdentVerOfPeer)
            {
                DebugFatal(103);

                return false;
            }

            SecureIdentStates l_State = (SecureIdentStates)l_Buffer.ReadByte();

            if (SecureIdentStates.IS_SIGNATURENEEDED != l_State &&
                SecureIdentStates.IS_KEYANDSIGNEEDED != l_State &&
                SecureIdentStates.IS_AUTHENTICATED_INSTANTLY != l_State)
            {
                DebugFatal(104);

                return false;
            }

            l_Remaining--;
            f_Flags.SetFlag(FlagsToUInt(StreamFlags.RcvdPeersSecIdentState));

            if (SecureIdentStates.IS_AUTHENTICATED_INSTANTLY == l_State)
            {
                if (0 != l_Remaining)
                {
                    DebugFatal(105);

                    return false;
                }

                f_Flags.SetFlag(FlagsToUInt(StreamFlags.SecIdentPeerShouldBeDone));

                if (false == DoesPeerNeedToProvideSecIdentMaterial())
                {
                    FinishedHandshake(true);
                }
            }
            else
            {
                uint l_ChallengeLen = (4 == l_SecIdentVer ? (uint)SHA_1.c_HashByteSize : sizeof(uint));

                if (l_ChallengeLen != l_Remaining)
                {
                    DebugFatal(106);

                    return false;
                }

                byte[] l_Challenge = l_Buffer.ReadBytes(l_ChallengeLen);

                if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RcvdPeersChallenge)))
                {
                    f_PeersChallengeForMe = l_Challenge;

                    f_Flags.SetFlag(FlagsToUInt(StreamFlags.RcvdPeersChallenge));
                }
                else if (0 != Buffer2.BlockCompare(l_Challenge, f_PeersChallengeForMe))
                {
                    // There was already a challenge and now there is a
                    // different one.

                    DebugFatal(107);

                    return false;
                }

                if (SecureIdentStates.IS_KEYANDSIGNEEDED == l_State)
                {
                    if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentOwnPubKey)))
                    {
                        if (false == SendPublicKey())
                        {
                            return false;
                        }
                    }
                }

                if (true == f_UserCntxt.HasPublicKey() &&
                    false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentOwnSignature)))
                {
                    if (false == SendSignature())
                    {
                        return false;
                    }
                }
            }

            NotifyReceiveIfOutgoing(a_Packet);

            return true;
        }

        private bool SendPublicKey()
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            uint l_SecIdentVer = f_PeerInfo.f_SuggestedSecIdentVer;
            RsaPublicKey l_RsaPubKey = f_Inst.f_NetworkMngr.p_OwnPersona.GetSuitableKeyPair(l_SecIdentVer).f_PublicKey;

            l_RsaPubKey.ToNetworkBuffer(l_Buffer, l_SecIdentVer, true);

            if (false == EnqueuePacket(Opcodes.OP_PUBLICKEY, PacketSendingFlags.IncUserCountIfIncoming, l_Buffer))
            {
                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.SentOwnPubKey));

            return true;
        }

        private bool ProcessPublicKey(Packet a_Packet)
        {
            // NOTE: We can land here even if f_UserCntxt.HasPublicKey()
            // returns true, since there can be multiple connections in
            // parallel that query and set the public key independently.

            if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentSecIdentState)) ||
                false == DoesPeerNeedToProvideSecIdentMaterial())
            {
                DebugFatal(110);

                return false;
            }

            if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RcvdPeersSecIdentState)))
            {
                f_SecIdentVerOfPeer = f_PeerInfo.f_SuggestedSecIdentVer;
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;

            if ((sizeof(byte) + sizeof(ushort)) > l_Buffer.GetRemaining())
            {
                DebugFatal(111);

                return false;
            }

            int l_SecIdentVer = f_SecIdentVerOfPeer;
            uint l_PubKeyLen;

            if (2 >= l_SecIdentVer)
            {
                l_PubKeyLen = l_Buffer.ReadByte();
            }
            else
            {
                l_PubKeyLen = l_Buffer.ReadUInt16LE();
            }

            if (l_Buffer.GetRemaining() != l_PubKeyLen)
            {
                DebugFatal(112);

                return false;
            }

            ByteArray l_BinaryPubKey;

            if (false == l_Buffer.ExposeRangeChecked(l_Buffer.p_Position, l_PubKeyLen, out l_BinaryPubKey))
            {
                DebugFatal(113);

                return false;
            }

            RsaPublicKey l_PubKey;

            if (false == RsaPublicKey.FromUnknownFormat(f_Inst.f_NetworkMngr.f_RSA, ref l_BinaryPubKey, out l_PubKey))
            {
                DebugFatal(114);

                return false;
            }

            uint l_KeySize = l_PubKey.GetBitSize();

            if (OwnUserPersona.c_MinKeyBits > l_KeySize ||
                OwnUserPersona.c_MaxKeyBits < l_KeySize ||
                false == f_Inst.f_UserMngr.SetPublicKey(f_UserCntxt, l_PubKey))
            {
                DebugFatal(115);

                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.RcvdPeersPubKey));

            NotifyReceiveIfOutgoing(a_Packet);

            return (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentOwnSignature)) ||
                    false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RcvdPeersChallenge)) ||
                    true == SendSignature());
        }

        public static byte[] CreateChallengeHash(NetworkManager a_NetworkMngr,
                                                 uint a_SecIdentVer,
                                                 RsaPublicKey a_PubKey,
                                                 byte[] a_Challenge,
                                                 I2PAddress a_StreamAddr,
                                                 uint a_SecIdentVer2AddressHashCode)
        {
            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            a_PubKey.ToNetworkBuffer(l_Buffer, a_SecIdentVer, false);
            l_Buffer.WriteBytes(a_Challenge);

            IHashFunction l_HashAlgo;

            switch (a_SecIdentVer)
            {
                case 1:
                    l_HashAlgo = a_NetworkMngr.f_SHA1;
                    break;

                case 2:
                    l_HashAlgo = a_NetworkMngr.f_SHA1;
                    l_Buffer.WriteUInt32LE(a_SecIdentVer2AddressHashCode);
                    l_Buffer.WriteByte((byte)CryptCips.CRYPT_CIP_LOCALCLIENT);
                    break;

                case 3:
                    l_HashAlgo = a_NetworkMngr.f_Tiger;
                    a_StreamAddr.ToBufferRaw(l_Buffer);
                    break;

                case 4:
                    l_HashAlgo = a_NetworkMngr.f_SHA3_512;
                    a_StreamAddr.ToBufferRaw(l_Buffer);
                    break;

                default: throw (new ApplicationException());
            }

            l_HashAlgo.Reset();
            l_Buffer.HashRange(l_HashAlgo, 0, l_Buffer.p_Length);

            // Regardless of the algo we use, always use a 20
            // byte array, since that's what the protocol expects.
            byte[] l_Hash = new byte[SHA_1.c_HashByteSize];

            if (2 >= a_SecIdentVer)
            {
                a_NetworkMngr.f_SHA1.GetDigest(l_Hash, 0);
            }
            else if (3 == a_SecIdentVer)
            {
                a_NetworkMngr.f_Tiger.GetDigest160(l_Hash, 0);
            }
            else
            {
                byte[] l_Temp = new byte[64];
                a_NetworkMngr.f_SHA3_512.GetDigest(l_Temp, 0);
                Buffer.BlockCopy(l_Temp, 0, l_Hash, 0, 20);
            }

            return l_Hash;
        }

        private bool SendSignature()
        {
            // Hash the peer's public key, the challenge the
            // peer gave us and (optional) our own address.
            // Then sign this.

            uint l_SecIdentVer = f_PeerInfo.f_SuggestedSecIdentVer;
            I2PAddress l_OwnAddress = f_Inst.f_NetworkMngr.p_StreamAddr;
            uint l_SecIdentVer2AddressHashCode = (2 == l_SecIdentVer ? l_OwnAddress.GetCompatibleHashCode(f_PeerInfo.f_IsSecondiMulePort) : 0);
            byte[] l_SecureIdentHash = CreateChallengeHash(f_Inst.f_NetworkMngr,
                                                           l_SecIdentVer,
                                                           f_UserCntxt.ExposePublicKey(),
                                                           f_PeersChallengeForMe,
                                                           l_OwnAddress,
                                                           l_SecIdentVer2AddressHashCode);
            RsaKeyPair l_MyKeyPair = f_Inst.f_NetworkMngr.p_OwnPersona.GetSuitableKeyPair(l_SecIdentVer);
            byte[] l_Signature = l_MyKeyPair.CreateSignature(l_SecureIdentHash, 4 == l_SecIdentVer, 4 == l_SecIdentVer ? f_Inst.f_NetworkMngr.f_SHA3_512 : null);

            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            if (3 <= l_SecIdentVer)
            {
                l_Buffer.WriteUInt16LE((ushort)l_Signature.Length);
            }
            else
            {
                l_Buffer.WriteByte((byte)l_Signature.Length);
            }

            l_Buffer.WriteBytes(l_Signature);

            if (2 == l_SecIdentVer)
            {
                l_Buffer.WriteByte((byte)CryptCips.CRYPT_CIP_LOCALCLIENT);
            }

            // NOTE: If a client doesn't answer a signature packet,
            // this can be due to its blocking behaviour (it accepts
            // only one key per destination and rejects all further keys).
            // This leads to trouble when changing the public key while
            // keeping the old destination.

            if (false == EnqueuePacket(Opcodes.OP_SIGNATURE, PacketSendingFlags.IncUserCountIfIncoming, l_Buffer))
            {
                return false;
            }

            f_Flags.SetCombination(FlagsToUInt(StreamFlags.SentOwnSignature | StreamFlags.SecIdentPeerShouldBeDone));

            if (false == DoesPeerNeedToProvideSecIdentMaterial())
            {
                FinishedHandshake(true);
            }

            return true;
        }

        private bool ProcessSignature(Packet a_Packet)
        {
            if (false == f_UserCntxt.HasPublicKey() ||
                false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentSecIdentState)) ||
                false == DoesPeerNeedToProvideSecIdentMaterial())
            {
                DebugFatal(120);

                return false;
            }

            if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RcvdPeersSecIdentState)))
            {
                f_SecIdentVerOfPeer = f_PeerInfo.f_SuggestedSecIdentVer;
            }

            uint l_SecIdentVer = f_SecIdentVerOfPeer;

            if (2 >= l_SecIdentVer &&
                OwnUserPersona.c_MaxKeyBitsiMule < f_UserCntxt.ExposePublicKey().GetBitSize())
            {
                DebugFatal(121);
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;

            if (sizeof(ushort) > l_Buffer.GetRemaining())
            {
                DebugFatal(122);

                return false;
            }

            uint l_SigLen;
            uint l_ExpRemLen;

            if (2 >= l_SecIdentVer)
            {
                l_SigLen = l_Buffer.ReadByte();
                l_ExpRemLen = (2 == l_SecIdentVer ? sizeof(byte) : 0u);
            }
            else
            {
                l_SigLen = l_Buffer.ReadUInt16LE();
                l_ExpRemLen = 0;
            }

            if (f_UserCntxt.GetMinExpectedSignatureLength() > l_SigLen ||
                l_Buffer.GetRemaining() != (l_SigLen + l_ExpRemLen))
            {
                DebugFatal(123);

                return false;
            }

            byte[] l_Signature = l_Buffer.ReadBytes(l_SigLen);

            if (2 == l_SecIdentVer &&
                CryptCips.CRYPT_CIP_LOCALCLIENT != (CryptCips)l_Buffer.ReadByte())
            {
                DebugFatal(124);

                return false;
            }

            bool l_SigIsGood = TryVerifySignature(l_SecIdentVer, l_Signature, out f_PeerAuthKeyStrength);

            if (false == l_SigIsGood &&
                false == AllowPeerToGoOnWithFailedAuthentication())
            {
                DebugFatal(125);

                return false;
            }

            StreamFlags l_SigVerFlag = (false == l_SigIsGood ? StreamFlags.SecIdentAttemptToIdentifyPeerFailed : StreamFlags.SecIdentFullyIdentifiedPeer);
            f_Flags.SetCombination(FlagsToUInt(StreamFlags.RcvdPeersSignature | l_SigVerFlag));
            NotifyReceiveIfOutgoing(a_Packet);

            if (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SecIdentPeerShouldBeDone)))
            {
                FinishedHandshake(true);
            }

            return true;
        }

        private bool TryVerifySignature(uint a_SecIdentVer,
                                        byte[] a_Signature,
                                        out uint a_PeerAuthKeyStrength)
        {
            // This is a helper for ProcessSignature().

            // NOTE: iMule 2.3.2 and below have a broken SecIdent
            // version 2 implementation in that they send a signature
            // of a zeroed address hash code. Also note that iMule 1.x
            // and 2.x use/expect a different address hash code (the
            // sum of the entire address versus directly using the
            // first 32 bits of the address). NB uses and expects the
            // same hash code that iMule 1.x used except when
            // interacting with the second iMule port.

            RsaPublicKey l_OwnPubKey = f_Inst.f_NetworkMngr.p_OwnPersona.GetSuitableKeyPair(a_SecIdentVer).f_PublicKey;
            I2PAddress l_RemoteAddress = f_Strm.p_RemoteAddr;
            uint l_SecIdentVer2AddressHashCode = (2 == a_SecIdentVer ? (false == f_PeerInfo.p_IsiMule ? l_RemoteAddress.GetCompatibleHashCode(false) : 0) : 0);

            for (uint i = 0; 2 != i; i++)
            {
                // Create the hash we expect the peer to have signed.

                byte[] l_SecureIdentHash = CreateChallengeHash(f_Inst.f_NetworkMngr,
                                                               a_SecIdentVer,
                                                               l_OwnPubKey,
                                                               f_MyChallengeForPeer,
                                                               l_RemoteAddress,
                                                               l_SecIdentVer2AddressHashCode);
                RsaPublicKey.SignatureVerificationContext l_VerCntxt = new RsaPublicKey.SignatureVerificationContext(f_Inst.f_NetworkMngr.f_RSA, l_SecureIdentHash, a_Signature, 4 == a_SecIdentVer, 4 == a_SecIdentVer ? f_Inst.f_NetworkMngr.f_SHA3_512 : null);

                if (true == f_Inst.f_UserMngr.VerifySignature(f_UserCntxt, ref l_VerCntxt))
                {
                    a_PeerAuthKeyStrength = l_VerCntxt.f_KeyStrength;

                    return true;
                }

                if (0 == i)
                {
                    if (false == f_PeerInfo.p_IsiMule)
                    {
                        break;
                    }

                    // Retry using the hash code we would have used if
                    // iMules wasn't broken (in case iMule ever gets
                    // patched).
                    l_SecIdentVer2AddressHashCode = l_RemoteAddress.GetCompatibleHashCode(f_PeerInfo.f_IsSecondiMulePort);
                }
            }

            a_PeerAuthKeyStrength = 0;

            return false;
        }

        private bool ProcessPing(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_InLength = l_Buffer.p_Length;
            MemoryIO l_AnsBuffer = null;

            if (0 != l_InLength)
            {
                // Copy up to eight bytes of requestor defined data.
                // TODO: Send up to 64 KiB of random memory.

                if (sizeof(ulong) < l_InLength)
                {
                    DebugFatal(129);

                    return false;
                }

                l_AnsBuffer = MemoryIO.CreateReadable(l_Buffer.ReadBytes(l_InLength));
            }

            return EnqueuePacket(OpcodesEx.OP_PONG, PacketSendingFlags.None, l_AnsBuffer);
        }

#region MetaData

        public bool RequestFileDescriptor()
        {
            return EnqueueFileIDPacket(OpcodesEx.OP_FILEDESCRIPTOR_REQ, PacketSendingFlags.None, f_DlPeerCntxt.f_FileID);
        }

        private bool ProcessFileDescriptorRequest(Packet a_Packet)
        {
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(a_Packet, PeerContextRequirements.Any, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                return l_Excusable;
            }

            FileDescriptorCookie l_Cookie = new FileDescriptorCookie(f_Inst.f_ContentMngr, l_PeerCntxt.f_Known);

            if (true == f_Inst.f_ContentMngr.NetworkBuildFileDescriptor(l_Cookie))
            {
                l_Cookie.WaitForCompletion();

                if (CookieStates.Success == l_Cookie.p_State)
                {
                    ByteArray l_Data = new ByteArray(l_Cookie.p_Data);

                    return EnqueuePacket(OpcodesEx.OP_FILEDESCRIPTOR_ANS, PacketSendingFlags.None, ref l_Data);
                }
            }

            return true;
        }

        private bool ProcessFileDescriptor(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;

            if (null == ConfirmPeerContext(l_Buffer, PeerContextRequirements.Dl, out l_Remaining, out l_Excusable))
            {
                DebugFatal(130, !l_Excusable);

                return l_Excusable;
            }

            if ((sizeof(ulong) + sizeof(ushort)) > l_Remaining)
            {
                DebugFatal(131);

                return false;
            }

            ulong a_FileSize = l_Buffer.ReadUInt64LE();
            P2PHashSets l_ProvidableHashSets = (P2PHashSets)l_Buffer.ReadUInt16LE();
            MasterHashes l_MasterHashes;
            if (false == MasterHashes.FromBuffer(true, l_Buffer, out l_MasterHashes) ||
                false == f_DlPeerCntxt.NotifyFileDescriptor(l_ProvidableHashSets, a_FileSize, ref l_MasterHashes))
            {
                DebugFatal(132);

                return false;
            }

            NotifyReceive(a_Packet);

            return true;
        }

        public bool RequestAdvancedFileMetaData(ref SourceCookie.HashQueryParams a_HashQueryParams,
                                                uint a_TicketID)
        {
            AdvFileMetaDataFlags l_Flags = (null == a_HashQueryParams.f_HashIndices ? AdvFileMetaDataFlags.Range : AdvFileMetaDataFlags.Indices);

            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            l_Buffer.WriteUInt128BE(f_DlPeerCntxt.f_FileID.f_Val);
            l_Buffer.WriteUInt32LE(a_TicketID);
            l_Buffer.WriteUInt16LE((ushort)l_Flags);
            l_Buffer.WriteUInt16LE((ushort)a_HashQueryParams.f_HashSet);

            switch (a_HashQueryParams.f_HashSet)
            {
                case P2PHashSets.eD2kBitmap:
                case P2PHashSets.eD2kHashSet:
                case P2PHashSets.ClientNativeBitmap:
                    break;

                case P2PHashSets.AICHTree:
                    l_Buffer.WriteBytes(a_HashQueryParams.f_ExpectedMasterHash);
                    break;

                case P2PHashSets.TAICHTree:
                    l_Buffer.WriteBytes(a_HashQueryParams.f_ExpectedMasterHash);
                    break;

                default: throw (new ApplicationException());
            }

            if (null == a_HashQueryParams.f_HashIndices)
            {
                l_Buffer.WriteUInt32LE(a_HashQueryParams.f_HashOffset);
                l_Buffer.WriteUInt32LE(a_HashQueryParams.f_HashCount);
            }
            else
            {
                l_Buffer.WriteUIntsLE(a_HashQueryParams.f_HashIndices);
            }

            return EnqueuePacket(OpcodesEx.OP_FILEMETADATA_REQ, PacketSendingFlags.None, l_Buffer);
        }

        private bool ProcessAdvancedFileMetaDataRequest(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(l_Buffer, PeerContextRequirements.Any, out l_Remaining, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                DebugFatal(140, !l_Excusable);

                return l_Excusable;
            }

            if (8 > l_Remaining)
            {
                DebugFatal(141);

                return false;
            }

            uint l_TickedID = l_Buffer.ReadUInt32LE();
            AdvFileMetaDataFlags l_Flags = (AdvFileMetaDataFlags)l_Buffer.ReadUInt16LE();
            bool l_64 = (0 != (AdvFileMetaDataFlags.IntegerSize64 & l_Flags));
            P2PHashSets l_HashSet = (P2PHashSets)l_Buffer.ReadUInt16LE();

            l_Flags &= ~AdvFileMetaDataFlags.IntegerSize64;

            if (AdvFileMetaDataFlags.Indices != l_Flags &&
                AdvFileMetaDataFlags.Range != l_Flags)
            {
                DebugFatal(142);

                return false;
            }

            byte[] l_MasterHash;

            switch (l_HashSet)
            {
                case P2PHashSets.eD2kBitmap:
                case P2PHashSets.ClientNativeBitmap:
                    l_MasterHash = null;
                    break;

                case P2PHashSets.eD2kHashSet:
                    l_MasterHash = l_PeerCntxt.f_FileID.f_Val.ToByteArrayBE();
                    break;

                case P2PHashSets.AICHTree:
                    {
                        if ((8 + SHA_1.c_HashByteSize) > l_Remaining)
                        {
                            DebugFatal(143);

                            return false;
                        }

                        l_MasterHash = l_Buffer.ReadBytes(SHA_1.c_HashByteSize);
                        break;
                    }

                case P2PHashSets.TAICHTree:
                    {
                        if ((8 + Tiger.c_HashByteSize) > l_Remaining)
                        {
                            DebugFatal(144);

                            return false;
                        }

                        l_MasterHash = l_Buffer.ReadBytes(Tiger.c_HashByteSize);
                        break;
                    }

                default:
                    // We do not allow P2PHashSets.Upper8MiBTAICHTree here.
                    DebugFatal(145);
                    return false;
            }

            l_Remaining = l_Buffer.GetRemaining();
            HashQueryCookie l_Cookie;

            if (0 != (AdvFileMetaDataFlags.Range & l_Flags))
            {
                if ((uint)(2 * (false == l_64 ? sizeof(uint) : sizeof(ulong))) != l_Remaining)
                {
                    DebugFatal(146);

                    return false;
                }

                uint l_StartingIndex;
                uint l_Length;

                if (false == l_64)
                {
                    l_StartingIndex = l_Buffer.ReadUInt32LE();
                    l_Length = l_Buffer.ReadUInt32LE();
                }
                else
                {
                    ulong l_Temp1 = l_Buffer.ReadUInt64LE();
                    ulong l_Temp2 = l_Buffer.ReadUInt64LE();

                    if (uint.MaxValue < l_Temp1 ||
                        uint.MaxValue < l_Temp2)
                    {
                        DebugFatal(147);

                        return false;
                    }

                    l_StartingIndex = (uint)l_Temp1;
                    l_Length = (uint)l_Temp2;
                }

                if (0 == l_Length)
                {
                    DebugFatal(148);

                    return false;
                }

                l_Cookie = new HashQueryCookie(f_Inst.f_ContentMngr, l_PeerCntxt.f_Known, l_HashSet, Opcodes.NONE, HashQueryKinds.Range, l_MasterHash, null, l_StartingIndex, l_Length, 0);
            }
            else if (0 != (AdvFileMetaDataFlags.Indices & l_Flags))
            {
                if (P2PHashSets.eD2kBitmap == l_HashSet ||
                    P2PHashSets.ClientNativeBitmap == l_HashSet)
                {
                    DebugFatal(149);

                    return false;
                }

                uint l_Count = (uint)(l_Remaining / (uint)(false == l_64 ? sizeof(uint) : sizeof(ulong)));

                if (0 == l_Count ||
                    HashQueryCookie.c_MaxSpecificIndices < l_Count)
                {
                    DebugFatal(150);

                    return false;
                }

                // The array may contain multiple entries.
                uint[] l_Indices;

                if (false == l_64)
                {
                    l_Indices = l_Buffer.ReadUIntsLE(l_Count);
                }
                else
                {
                    ulong[] l_Indices64 = l_Buffer.ReadBaseTypeArray<ulong>(l_Count, sizeof(ulong));

                    l_Indices = new uint[(int)l_Count];

                    for (int i = 0; i < l_Count; i++)
                    {
                        if (uint.MaxValue < l_Indices64[i])
                        {
                            DebugFatal(151);

                            return false;
                        }

                        l_Indices[i] = (uint)l_Indices64[i];
                    }
                }

                if (false == l_Buffer.EndReached())
                {
                    DebugFatal(152);

                    return false;
                }

                l_Cookie = new HashQueryCookie(f_Inst.f_ContentMngr, l_PeerCntxt.f_Known, l_HashSet, Opcodes.NONE, HashQueryKinds.Indices, l_MasterHash, l_Indices, 0, 0, 0);
            }
            else
            {
                DebugFatal(153);

                return false;
            }

            l_Cookie.f_TickedID = l_TickedID;

            return l_PeerCntxt.AddHashQuery(l_Cookie);
        }

        private bool ProcessAdvancedFileMetaData(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;

            if (null == ConfirmPeerContext(l_Buffer, PeerContextRequirements.Dl, out l_Remaining, out l_Excusable))
            {
                DebugFatal(160, !l_Excusable);

                return l_Excusable;
            }

            if (sizeof(uint) > l_Remaining)
            {
                DebugFatal(161);

                return false;
            }

            uint l_TicketID = l_Buffer.ReadUInt32LE();
            l_Remaining -= sizeof(uint);

            if (0 == l_Remaining ||
                false == f_DlPeerCntxt.HashDelivery(l_TicketID, l_Buffer.ReadBytes(l_Remaining)))
            {
                bool l_OK = f_DlPeerCntxt.AbortHashQuery(l_TicketID);

                DebugFatal(162, !l_OK);

                return l_OK;
            }

            NotifyReceive(a_Packet);

            return true;
        }

        private bool ProcessAICHMasterHashRequest(Packet a_Packet)
        {
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(a_Packet, PeerContextRequirements.Any, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                DebugFatal(170, !l_Excusable);

                return l_Excusable;
            }

            MasterHashes l_MasterHashes;
            MasterHashes.CloneVolatile(ref l_PeerCntxt.f_Known.f_MasterHashes, out l_MasterHashes);

            ByteArray l_Data;
            if (false == l_MasterHashes.ExposeHash(P2PHashSets.AICHTree, out l_Data))
            {
                // We do not have the AICH hash.

                return true;
            }

            // Write the file id and the AICH master hash (p_Data).

            MemoryIO l_AnsBuffer = MemoryIO.CreateWritable(UInt128.c_ByteSize + SHA_1.c_HashByteSize);

            l_AnsBuffer.WriteUInt128BE(l_PeerCntxt.f_FileID.f_Val);
            l_AnsBuffer.WriteBytes(ref l_Data);

            return EnqueuePacket(Opcodes.OP_AICHFILEHASHANS, PacketSendingFlags.IncUsersCount, l_AnsBuffer);
        }

        private bool ProcessAICHRequest(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(l_Buffer, PeerContextRequirements.Any, out l_Remaining, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                DebugFatal(180, !l_Excusable);

                return l_Excusable;
            }

            if ((sizeof(ulong) + SHA_1.c_HashByteSize) != l_Remaining)
            {
                DebugFatal(181);

                return false;
            }

            ulong l_PartIndex = l_Buffer.ReadUInt64LE();

            if (uint.MaxValue < l_PartIndex)
            {
                DebugFatal(182);

                return false;
            }

            byte[] l_MasterHash = l_Buffer.ReadBytes(SHA_1.c_HashByteSize);
            HashQueryCookie l_Cookie = new HashQueryCookie(f_Inst.f_ContentMngr, l_PeerCntxt.f_Known, P2PHashSets.AICHTree, Opcodes.OP_AICHANSWER, HashQueryKinds.AICHPartRecoveryData, l_MasterHash, null, 0, 0, (uint)l_PartIndex);

            return l_PeerCntxt.AddHashQuery(l_Cookie);
        }

        private bool ProcessHashSetRequest(Packet a_Packet)
        {
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(a_Packet, PeerContextRequirements.Any, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                DebugFatal(190, !l_Excusable);

                return l_Excusable;
            }

            byte[] l_MasterHash = l_PeerCntxt.f_FileID.f_Val.ToByteArrayBE();
            uint l_HashCount = l_PeerCntxt.f_Known.f_PartCountHashSet;
            HashQueryCookie l_Cookie = new HashQueryCookie(f_Inst.f_ContentMngr, l_PeerCntxt.f_Known, P2PHashSets.eD2kHashSet, Opcodes.OP_HASHSETANSWER, HashQueryKinds.Range, l_MasterHash, null, 0, l_HashCount, 0);

            return l_PeerCntxt.AddHashQuery(l_Cookie);
        }

        private bool ProcessHashSet(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;

            if (null == ConfirmPeerContext(l_Buffer, PeerContextRequirements.Dl, out l_Remaining, out l_Excusable))
            {
                DebugFatal(200, !l_Excusable);

                return l_Excusable;
            }

            if (sizeof(ulong) > l_Remaining)
            {
                DebugFatal(201);

                return false;
            }

            // They use this crazy errornous formula to
            // calculate what they believe is a part count,
            // so beware.
            ulong l_PartCountHashSet = l_Buffer.ReadUInt64LE();
            l_Remaining -= sizeof(ulong);

            if (f_DlPeerCntxt.f_Known.f_PartCountHashSet != l_PartCountHashSet)
            {
                return f_DlPeerCntxt.NotifyDisputedFileSize();
            }

            if ((l_PartCountHashSet * MD_4.c_HashByteSize) != l_Remaining ||
                false == f_DlPeerCntxt.RegisterHashSet(0, (uint)l_PartCountHashSet, l_Buffer))
            {
                DebugFatal(202);

                return false;
            }

            NotifyReceive(a_Packet);

            return true;
        }

        public bool ProcessHashQuery(PeerContext a_PeerCntxt,
                                     HashQueryCookie a_Cookie)
        {
            // This routine gets called from UlPeerContext.

            if (CookieStates.Success != a_Cookie.p_State)
            {
                if (true == a_Cookie.p_MalformedInput)
                {
                    DebugFatal(210);

                    return false;
                }

                if (Opcodes.NONE != a_Cookie.f_Opcode)
                {
                    return true;
                }
            }

            MemoryIO l_Buffer = MemoryIO.CreateWritable();
            l_Buffer.WriteUInt128BE(a_PeerCntxt.f_FileID.f_Val);

            switch (a_Cookie.f_Opcode)
            {
                case Opcodes.NONE:
                    {
                        // Handles OpcodesEx.OP_FILEMETADATA_REQ.
                        // Write the file id, the ticket id and (on
                        // success the requested data (p_Data)).
                        // Send OpcodesEx.OP_FILEMETADATA_ANS.

                        l_Buffer.WriteUInt32LE(a_Cookie.f_TickedID);
                        break;
                    }

                case Opcodes.OP_HASHSETANSWER:
                    {
                        // Handles OP_Codes.OP_HASHSETREQUEST.
                        // Write the file id and the part count (hash
                        // set specific) and then append the part
                        // hashes (p_Data).

                        l_Buffer.WriteUInt64LE(a_PeerCntxt.f_Known.f_PartCountHashSet);
                        break;
                    }

                case Opcodes.OP_AICHANSWER:
                    {
                        // Handles OP_Codes.OP_AICHREQUEST.
                        // Write file id, part index, AICH master hash
                        // and the recovery data.

                        l_Buffer.WriteUInt64LE(a_Cookie.f_AICHTrgtPartIndex);

                        // The recovery buffer (p_Data) also contains
                        // the AICH master hash for us.
                        break;
                    }

                default: throw (new ApplicationException());
            }

            if (CookieStates.Success == a_Cookie.p_State)
            {
                l_Buffer.WriteBytes(a_Cookie.p_Data);
            }

            return (null != EnqueuePacketCore(a_Cookie.f_Opcode, Opcodes.NONE == a_Cookie.f_Opcode ? OpcodesEx.OP_FILEMETADATA_ANS : OpcodesEx.Undefined, PacketSendingFlags.IncUsersCount | PacketSendingFlags.InsertOutpacableAtEndOfQueue, l_Buffer));
        }

#endregion

        private bool ProcessMultiPacket(Packet a_Packet)
        {
            // Not supported.

            DebugFatal(220);

            return false;
        }

        public bool RequestHashSet(FileID a_FileID)
        {
            return EnqueueFileIDPacket(Opcodes.OP_HASHSETREQUEST, PacketSendingFlags.None, a_FileID);
        }

        public bool RequestFileName()
        {
            MemoryIO l_ReqBuffer = MemoryIO.CreateWritable();

            l_ReqBuffer.WriteUInt128BE(f_DlPeerCntxt.f_FileID.f_Val);

            return EnqueuePacket(Opcodes.OP_REQUESTFILENAME, PacketSendingFlags.None, l_ReqBuffer);
        }

        public bool RequestFileStatusOrComplBitmap(PeerContext a_PeerCntxt,
                                                   bool a_MayConstruct)
        {
            KnownFile l_Known = a_PeerCntxt.f_Known;
            bool l_InclFileSelFlags;
            bool l_ReqComplBitmap = f_PeerInfo.SupportsComplBitmapOpcode(out l_InclFileSelFlags);
            MasterHashes l_MasterHashes;
            uint l_ReqLen;

            if (false == l_ReqComplBitmap)
            {
                l_MasterHashes = new MasterHashes();
                l_ReqLen = UInt128.c_ByteSize;
            }
            else
            {
                MasterHashes.CloneVolatile(ref l_Known.f_MasterHashes, out l_MasterHashes);

                if (true == l_MasterHashes.p_HasTAICHHash)
                {
                    l_ReqLen = c_ComplBitmapReqLenInclTAICH;
                }
                else if (true == l_MasterHashes.p_HasAICHHash)
                {
                    l_ReqLen = c_ComplBitmapReqLenInclAICH;
                }
                else
                {
                    l_ReqLen = c_ComplBitmapReqMinLen;
                }

                if (true == l_InclFileSelFlags)
                {
                    l_ReqLen++;
                }
            }

            const PacketSendingFlags lc_Flags = PacketSendingFlags.None;
            MemoryIO l_ReqBuffer = MemoryIO.CreateWritable(l_ReqLen);
            l_ReqBuffer.WriteUInt128BE(a_PeerCntxt.f_FileID.f_Val);

            if (false == l_ReqComplBitmap)
            {
                return EnqueuePacket(Opcodes.OP_SETREQFILEID, lc_Flags, l_ReqBuffer);
            }

            OpcodesEx l_OpcodeEx;

            if (false == l_InclFileSelFlags)
            {
                l_OpcodeEx = OpcodesEx.OP_COMPLBITMAP_REQ;
            }
            else
            {
                FileSelectionFlags l_FileSelFlags = ((false == a_MayConstruct ? FileSelectionFlags.None : FileSelectionFlags.AllowConstruction) | FileSelectionFlags.AllowConfirmation);
                l_ReqBuffer.WriteByte((byte)l_FileSelFlags);

                l_OpcodeEx = OpcodesEx.OP_COMPLBITMAP_EXT_REQ;
            }

            // We write the file size, the bitmap kind and a flag
            // that will tell the remote peer that we want the
            // bitmap body if it's not trivial.

            l_ReqBuffer.WriteUInt64LE(l_Known.f_FileSize);
            l_ReqBuffer.WriteUInt16LE((ushort)P2PHashSets.ClientNativeBitmap);
            l_ReqBuffer.WriteByte(c_ComplBitmapReqInclNontrivialBodyFlag);

            // The following is optional.

            ByteArray l_Data;

            if (true == l_MasterHashes.ExposeHash(P2PHashSets.TAICHTree, out l_Data))
            {
                // We don't demand that the peer knows that the
                // requested file has this TAICH hash, but we clearly
                // don't want a file with a different known TAICH hash.

                l_ReqBuffer.WriteUInt16LE((ushort)P2PHashSets.TAICHTree);
                l_ReqBuffer.WriteBytes(ref l_Data);
            }
            else if (true == l_MasterHashes.ExposeHash(P2PHashSets.AICHTree, out l_Data))
            {
                // Use the AICH hash instead.

                l_ReqBuffer.WriteUInt16LE((ushort)P2PHashSets.AICHTree);
                l_ReqBuffer.WriteBytes(ref l_Data);
            }

            return EnqueuePacket(l_OpcodeEx, lc_Flags, l_ReqBuffer);
        }

        public bool RequestUploadStart(uint a_PrimTransferID,
                                       uint a_AuxTransferID)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.None;

            if (c_NoTransferID == a_PrimTransferID)
            {
                return EnqueueFileIDPacket(Opcodes.OP_STARTUPLOADREQ, lc_Flags, f_DlPeerCntxt.f_FileID);
            }

            MemoryIO l_Mem = CreateTransferIDBuffer(a_PrimTransferID, a_AuxTransferID, UInt128.c_ByteSize);

            l_Mem.WriteUInt128BE(f_DlPeerCntxt.f_FileID.f_Val);

            return EnqueuePacket(OpcodesEx.TI_UPLOAD_REQUEST, lc_Flags, l_Mem);
        }

        public bool RequestQueueRank(uint a_TransferID)
        {
            return EnqueueEmptyTransferIDPacket(OpcodesEx.TI_QUEUE_RANK_REQUEST, PacketSendingFlags.None, a_TransferID);
        }

        public bool SendOnQueueNotification(uint a_TransferID)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.None;

            if (c_NoTransferID != a_TransferID)
            {
                return EnqueueEmptyTransferIDPacket(OpcodesEx.TI_REFUSE_UPLOAD_INFO, lc_Flags, a_TransferID);
            }

            if (false == f_PeerInfo.SupportsOP_OUTOFPARTREQS())
            {
                return EnqueueEmptyPacket(OpcodesEx.OP_DESELECT_UPLOAD_FILE, lc_Flags);
            }

            return EnqueueEmptyPacket(Opcodes.OP_OUTOFPARTREQS, lc_Flags);
        }

        public bool SendOnQueueAcknowledgement(uint a_PrimTransferID,
                                               uint a_AuxTransferID)
        {
            MemoryIO l_Mem = CreateTransferIDBuffer(a_PrimTransferID, a_AuxTransferID, sizeof(ushort));

            return EnqueuePacket(OpcodesEx.TI_REFUSE_UPLOAD_ACKNOWLEDGEMENT, PacketSendingFlags.None, l_Mem);
        }

        private bool ProcessOnQueueNotification(Packet a_Packet)
        {
            if (null == f_DlPeerCntxt)
            {
                bool l_Excusable = TryExcuseMissingDlContext();

                return l_Excusable;
            }

            if (false == a_Packet.p_Buffer.EndReached())
            {
                return false;
            }

            if (false == f_DlPeerCntxt.NotifyOnQueueAgain(true))
            {
                return false;
            }

            f_DlHelper.Invalidate();

            return true;
        }

        public bool SendDeselectDlFileMessage(FileID a_FileID)
        {
            return EnqueueFileIDPacket(Opcodes.OP_END_OF_DOWNLOAD, PacketSendingFlags.None, a_FileID);
        }

        private bool ProcessDeselectDlFileMessage(Packet a_Packet)
        {
            // The remote client wants to abort downloading.

            if (Opcodes.OP_CANCELTRANSFER == a_Packet.p_DonkeyOpcode)
            {
                if (false == a_Packet.p_Buffer.EndReached())
                {
                    DebugFatal(230);

                    return false;
                }
            }
            else
            {
                if (null != f_UlPeerCntxt &&
                    (UInt128.c_ByteSize != a_Packet.p_Buffer.GetRemaining() ||
                     false == f_UlPeerCntxt.f_FileID.f_Val.ValueEquals(a_Packet.p_Buffer.ReadUInt128BE())))
                {
                    DebugFatal(231);

                    return false;
                }
            }

            if (false == BeginRemoveUploadContext(false))
            {
                return false;
            }

            return true;
        }

        private bool BeginRemoveUploadContext(bool a_SelfTriggered)
        {
            if (null == f_UlPeerCntxt)
            {
                if (null != f_UlPeerCntxtCreationCookie)
                {
                    f_UlPeerCntxtCreationCookie = null;
                    f_UlPeerCntxtCreationMode = UlPeerCntxtCreationModes.None;

                    return true;
                }

                // IMULETWEAK: iMule can send OP_CANCELTRANSFER without
                // ever having hinted that it wants to download.
                if (true == f_PeerInfo.p_IsiMule)
                {
                    return true;
                }

                return TryExcuseMissingUlContext();
            }

            BeginRemoveUploadContextCore(a_SelfTriggered);

            return true;
        }

        private void BeginRemoveUploadContextCore(bool a_SelfTriggered)
        {
            f_UlPeerCntxt.MarkAsRemovable(false == a_SelfTriggered ? PeerContext.DestructionInfo.RemoveAtPeersRequest : PeerContext.DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);
        }

        public bool SendDeselectUlFileMessage()
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.IncUsersCount;

            if (false == f_PeerInfo.SupportBasicOpcodeExtensions())
            {
                return EnqueueFileIDPacket(Opcodes.OP_FILEREQANSNOFIL, lc_Flags, f_UlPeerCntxt.f_FileID);
            }

            return EnqueueFileIDPacket(OpcodesEx.OP_DESELECT_UPLOAD_FILE, lc_Flags, f_UlPeerCntxt.f_FileID);
        }

        private bool ProcessDeselectUlFileMessage(Packet a_Packet)
        {
            bool l_Excusable;
            if (null == ConfirmPeerContext(a_Packet, PeerContextRequirements.Dl, out l_Excusable))
            {
                DebugFatal(240, !l_Excusable);

                return l_Excusable;
            }

            BeginRemoveDownloadContextCore(false);

            return true;
        }

        private void BeginRemoveDownloadContextCore(bool a_SelfTriggered)
        {
            f_DlPeerCntxt.MarkAsRemovable(SourceStates.PeerEndedUpload, false == a_SelfTriggered ? PeerContext.DestructionInfo.RemoveAtPeersRequest : PeerContext.DestructionInfo.RemoveAtOurRequestPeerNeedsInfo);
        }

        public bool SendQueueRanking(uint a_TransferID,
                                     uint a_Rank)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.IncUsersCount;
            MemoryIO l_Mem;

            if (c_NoTransferID == a_TransferID)
            {
                // We use OP_QUEUERANKING instead of OP_QUEUERANK.

                l_Mem = MemoryIO.CreateWritable(12);

                l_Mem.WriteUInt16LE((ushort)a_Rank);

                return EnqueuePacket(Opcodes.OP_QUEUERANKING, lc_Flags, l_Mem);
            }

            l_Mem = CreateTransferIDBuffer(a_TransferID, sizeof(ushort));

            l_Mem.WriteUInt16LE((ushort)a_Rank);

            return EnqueuePacket(OpcodesEx.TI_QUEUE_RANK_INFO, lc_Flags, l_Mem);
        }

        private bool ProcessQueueRanking(Packet a_Packet)
        {
            // The queue ranking is an auto answer to
            // OP_STARTUPLOADREQ.

            if (null == f_DlPeerCntxt)
            {
                bool l_Excusable = TryExcuseMissingDlContext();

                DebugFatal(250, !l_Excusable);

                return l_Excusable;
            }

            uint l_Rank;

            if (Opcodes.OP_QUEUERANK == a_Packet.p_DonkeyOpcode)
            {
                if (sizeof(uint) != a_Packet.p_Buffer.GetRemaining())
                {
                    DebugFatal(251);

                    return false;
                }

                l_Rank = a_Packet.p_Buffer.ReadUInt32LE();
            }
            else
            {
                uint l_Remaining = a_Packet.p_Buffer.GetRemaining();

                if (sizeof(ushort) > l_Remaining ||
                    12 < l_Remaining)
                {
                    DebugFatal(252);

                    return false;
                }

                l_Rank = a_Packet.p_Buffer.ReadUInt16LE();

                // We will ignore anything after the first ushort.
            }

            if (false == f_DlPeerCntxt.SetQueueRank(l_Rank))
            {
                DebugFatal(253);

                return false;
            }

            NotifyReceive(a_Packet);

            return true;
        }

        public bool SendAcceptUploadNotification(uint a_TransferID)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.IncUsersCount;

            if (c_NoTransferID == a_TransferID)
            {
                return EnqueueEmptyPacket(Opcodes.OP_ACCEPTUPLOADREQ, lc_Flags);
            }

            return EnqueueEmptyTransferIDPacket(OpcodesEx.TI_ACCEPT_UPLOAD_INFO, lc_Flags, a_TransferID);
        }

        private bool SendNoSuchFileNotification(FileID a_FileID)
        {
            return EnqueueFileIDPacket(Opcodes.OP_FILEREQANSNOFIL, PacketSendingFlags.IncUsersCount, a_FileID);
        }

        private bool ProcessFileRequest(Packet a_Packet)
        {
            UlPeerCntxtCreationModes l_EffMode;
            UlPeerCntxtCreationStates l_Status = TryBeginCreateUlPeerContext(a_Packet, out l_EffMode);

            switch (l_Status)
            {
                case UlPeerCntxtCreationStates.AlreadyUp:
                    return SendFileStatusOrComplBitmap(l_EffMode);

                case UlPeerCntxtCreationStates.Pending:
                    return true;

                case UlPeerCntxtCreationStates.NotAllowed:
                    return EnqueueEmptyPacket(OpcodesEx.OP_FILE_SELECTION_NOT_ALLOWED_INFO, PacketSendingFlags.IncUsersCount);

                case UlPeerCntxtCreationStates.Corruption:
                    return false;

                default:
                    throw(new UnexpectedStateException());
            }
        }

        private bool ProcessStartUploadRequest(Packet a_Packet)
        {
            bool l_Excusable;
            if (null == ConfirmPeerContext(a_Packet, PeerContextRequirements.Ul, out l_Excusable))
            {
                DebugFatal(270, !l_Excusable);

                return l_Excusable;
            }

            if (false == f_UlPeerCntxt.NotifyStartUploadRequest(c_NoTransferID, c_NoTransferID))
            {
                return false;
            }

            return true;
        }

        private bool ProcessAcceptUploadRequest(Packet a_Packet)
        {
            if (null == f_DlPeerCntxt)
            {
                bool l_Excusable = TryExcuseMissingDlContext();

                DebugFatal(280, !l_Excusable);

                return l_Excusable;
            }

            if (false == a_Packet.p_Buffer.EndReached())
            {
                DebugFatal(281);

                return false;
            }

            bool l_OK = f_DlPeerCntxt.NotifyUploadRequestAccepted(f_PeerInfo.p_IsiMule);

            DebugFatal(282, !l_OK);

            return l_OK;
        }

        public bool SendEndUploadMessage(uint a_TransferID)
        {
            return EnqueueEmptyTransferIDPacket(OpcodesEx.TI_END_OF_UPLOAD_INFO, PacketSendingFlags.IncUsersCount, a_TransferID);
        }

        private bool ProcessFileRangeRequest(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;

            if (null == ConfirmPeerContext(l_Buffer, PeerContextRequirements.Ul, out l_Remaining, out l_Excusable))
            {
                // IMULETWEAK: After a OP_OUTOFPARTREQS iMule can
                // take a long time to stop requesting ranges.
                if (true == f_PeerInfo.p_IsiMule &&
                    (c_ClassicFileRangeRequestULongs * sizeof(ulong)) == l_Remaining)
                {
                    return true;
                }

                DebugFatal(290, !l_Excusable);

                return l_Excusable;
            }

            return ProcessFileRangeRequestCore(l_Buffer, false);
        }

        private bool ProcessFileRangeRequestCore(BufferIO a_Buffer,
                                                 bool a_NewFormat)
        {
            ulong[] l_Arr;
            uint l_Count;
            uint l_Remaining = a_Buffer.GetRemaining();

            if (false == a_NewFormat)
            {
                if ((c_ClassicFileRangeRequestULongs * sizeof(ulong)) != l_Remaining)
                {
                    DebugFatal(291);

                    return false;
                }

                l_Arr = a_Buffer.ReadBaseTypeArray<ulong>(c_ClassicFileRangeRequestULongs, sizeof(ulong));
                l_Count = c_ClassicFileRangeRequestSlots;
            }
            else
            {
                if ((2 * sizeof(byte)) >= l_Remaining)
                {
                    return false;
                }

                const byte lc_64BitOffsetsFlag = 1;
                byte l_Flags = a_Buffer.ReadByte();
                l_Count = a_Buffer.ReadByte();

                if (0 != (l_Flags & ~lc_64BitOffsetsFlag) ||
                    0 == l_Count ||
                    UlPeerContext.c_MaxRequestedRanges < l_Count)
                {
                    return false;
                }

                bool l_Use64BitOffsets = (0 != (l_Flags & lc_64BitOffsetsFlag));
                uint l_OffsetSize = (false == l_Use64BitOffsets ? (uint)sizeof(uint) : sizeof(ulong));
                uint l_ExpectedSize = (l_Count * (l_OffsetSize + sizeof(uint)));

                if ((l_Remaining - (2 * sizeof(byte)) != l_ExpectedSize))
                {
                    return false;
                }

                l_Arr = new ulong[2 * l_Count];

                for (uint i = 0; i != l_Count; i++)
                {
                    ulong l_Offset = a_Buffer.ReadUIntVarSizeLE(l_OffsetSize);
                    ulong l_FinalOffset = (l_Offset + a_Buffer.ReadUInt32LE());

                    if (l_Offset > l_FinalOffset)
                    {
                        // Overflow.

                        return false;
                    }

                    l_Arr[i] = l_Offset;
                    l_Arr[i + l_Count] = l_FinalOffset;
                }
            }

            if (false == f_UlPeerCntxt.AddRequestedRanges(l_Arr, (int)l_Count))
            {
                return false;
            }

            return true;
        }

        public bool RequestFileRange(uint a_TransferID,
                                     ref FlexArray<DlPeerContext.RequestableRange> a_Ranges,
                                     int a_Index,
                                     int a_Count)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.None;
            DlPeerContext.RequestableRange[] l_Arr = a_Ranges.p_Arr;

            if (c_NoTransferID == a_TransferID)
            {
                // Up to three ranges can be requested at once.
                // iMule won't accept ranges of more than 500 KiB
                // per slot (the larger a chunk, the more bandwidth
                // is wasted on cancelation if compression is used).

                byte[] l_Bytes = new byte[UInt128.c_ByteSize + (c_ClassicFileRangeRequestULongs * sizeof(ulong))];

                f_DlPeerCntxt.f_FileID.f_Val.ToByteArrayBE(l_Bytes, 0);

                for (int i = 0, k = a_Index; i < a_Count; i++, k++)
                {
                    Util.WriteUInt64LE(l_Arr[k].f_Range.f_StartingOffset, l_Bytes, UInt128.c_ByteSize + (sizeof(ulong) * i));
                    Util.WriteUInt64LE(l_Arr[k].f_Range.p_FinalOffset, l_Bytes, UInt128.c_ByteSize + (sizeof(ulong) * 3) + (sizeof(ulong) * i));
                }

                return EnqueuePacket(Opcodes.OP_REQUESTPARTS, lc_Flags, MemoryIO.CreateReadable(l_Bytes));
            }

            bool l_Use64BitOffsets = (uint.MaxValue < f_DlPeerCntxt.f_Known.f_FileSize);
            uint l_OffsetSize = (false == l_Use64BitOffsets ? (uint)sizeof(uint) : sizeof(ulong));
            uint l_ExtraCapacity = ((2u * sizeof(byte)) + ((uint)a_Count * (l_OffsetSize + sizeof(uint))));

            MemoryIO l_Mem = CreateTransferIDBuffer(a_TransferID, l_ExtraCapacity);

            l_Mem.WriteBool(l_Use64BitOffsets);
            l_Mem.WriteByte((byte)a_Count);

            for (int i = a_Index, lim = (a_Index + a_Count); i < lim; i++)
            {
                l_Mem.WriteUIntVarSizeLE(l_OffsetSize, l_Arr[i].f_Range.f_StartingOffset);
                l_Mem.WriteUInt32LE(l_Arr[i].f_Range.f_Length);
            }

            return EnqueuePacket(OpcodesEx.TI_RANGE_REQUEST, lc_Flags, l_Mem);
        }

        public bool SendFileContent(ulong a_StartingOffset,
                                    ref ByteArray a_RawSrc,
                                    Packet.PacketCompletelySentHandler a_SentCompletely,
                                    out uint a_CompressionGain)
        {
            // a_RawSrc.f_Len must not be zero. Returns zero on error.

            bool l_AllowCompression = (0x1 == f_PeerInfo.f_DataCompVer &&
                                       0x1 == f_Inst.f_NetworkMngr.f_SelfInfo.f_DataCompVer);

            if (true == l_AllowCompression)
            {
                ByteArray l_CompBytes;

                if (true == zlib.Compress(ref a_RawSrc, true, out l_CompBytes))
                {
                    uint l_Gain = (uint)(a_RawSrc.f_Len - l_CompBytes.f_Len);

                    if (50 <= l_Gain &&
                        (2 * (a_RawSrc.f_Len / 100)) <= l_Gain)
                    {
                        a_CompressionGain = l_Gain;

                        return SendFileContentCore(a_StartingOffset, ref l_CompBytes, a_SentCompletely, true);
                    }
                }
            }

            a_CompressionGain = 0;

            return SendFileContentCore(a_StartingOffset, ref a_RawSrc, a_SentCompletely, false);
        }

        private bool SendFileContentCore(ulong a_StartingOffset,
                                         ref ByteArray a_Src,
                                         Packet.PacketCompletelySentHandler a_SentCompletely,
                                         bool a_Compressed)
        {
            // a_Src.f_Len must not be zero.

            bool l_Align = (false == a_Compressed &&
                            NB.Content.Internal.PieceBounds.c_PieceSize <= a_Src.f_Len &&
                            true == f_PeerInfo.p_IsThisClient);
            uint l_FirstPieceSize;
            uint l_OtherPieceMaxSize;
            uint l_PieceCount = CalcOutgoingFileContentPieceSize(a_StartingOffset, (uint)a_Src.f_Len, l_Align, out l_FirstPieceSize, out l_OtherPieceMaxSize);

            uint l_Rem = (uint)a_Src.f_Len;
            uint l_BufferOffset = (uint)a_Src.f_Idx;
            ulong l_FileOffset = a_StartingOffset;
            uint l_ClusterNumber = ++f_FileContentClusterNumber;

            for (uint k = 0, lim = (l_PieceCount - 1); k != l_PieceCount; k++)
            {
                uint l_BodySize;

                if (k == lim)
                {
                    // Last piece.

                    l_BodySize = l_Rem;
                }
                else if (0 == k)
                {
                    // First piece.

                    l_BodySize = l_FirstPieceSize;
                }
                else
                {
                    l_BodySize = Math.Min(l_OtherPieceMaxSize, l_Rem);
                }

                const PacketSendingFlags lc_Flags = (PacketSendingFlags.IncBothCounts | PacketSendingFlags.InsertOutpacableAtEndOfQueue | PacketSendingFlags.NoCompression);
                ByteArray l_BodySrc = new ByteArray(a_Src.f_Arr, (int)l_BufferOffset, (int)l_BodySize);
                Packet l_Packet;

                if (false == a_Compressed)
                {
                    MemoryIO l_ResFile = MemoryIO.CreateWritable(32 + l_BodySize);

                    l_ResFile.WriteUInt128BE(f_UlPeerCntxt.f_FileID.f_Val);
                    l_ResFile.WriteUInt64LE(l_FileOffset);

                    l_FileOffset += l_BodySize;

                    l_ResFile.WriteUInt64LE(l_FileOffset);
                    l_ResFile.WriteBytes(ref l_BodySrc);

                    l_Packet = EnqueuePacketCore(Opcodes.OP_SENDINGPART, OpcodesEx.Undefined, lc_Flags, l_ResFile);
                }
                else
                {
                    MemoryIO l_ResFile = MemoryIO.CreateWritable(28 + l_BodySize);

                    l_ResFile.WriteUInt128BE(f_UlPeerCntxt.f_FileID.f_Val);
                    l_ResFile.WriteUInt64LE(a_StartingOffset);
                    l_ResFile.WriteUInt32LE((uint)a_Src.f_Len);
                    l_ResFile.WriteBytes(ref l_BodySrc);

                    l_Packet = EnqueuePacketCore(Opcodes.OP_COMPRESSEDPART, OpcodesEx.Undefined, lc_Flags, l_ResFile);

                    if (null != l_Packet)
                    {
                        uint l_CumComprOffset = (l_BufferOffset - (uint)a_Src.f_Idx);
                        l_Packet.SetCumulativeCompressionOffset(l_CumComprOffset);
                    }
                }

                if (null == l_Packet)
                {
                    return false;
                }

                l_Packet.SetClusterNumber(l_ClusterNumber, k, l_PieceCount);
                l_Packet.e_SentCompletely += a_SentCompletely;

                l_Rem -= l_BodySize;
                l_BufferOffset += l_BodySize;
            }

            return true;
        }

        private static uint CalcOutgoingFileContentPieceSize(ulong a_FileOffset,
                                                             uint a_RangeByteSize,
                                                             bool a_Align,
                                                             out uint a_FirstPieceSize,
                                                             out uint a_OtherPieceMaxSize)
        {
            // a_RangeByteSize must not be zero. Returns the number of
            // pieces necessary.

            if (c_DefaultPieceSize >= a_RangeByteSize)
            {
                goto RETURN_ONE_PIECE;
            }

            uint l_PieceCount;

            if (false == a_Align)
            {
                // Balance the packet size.
                l_PieceCount = ((a_RangeByteSize / c_DefaultPieceSize) + (0 == (a_RangeByteSize % c_DefaultPieceSize) ? 0u : 1u));
                uint l_MaxPacketBodySize = ((a_RangeByteSize / l_PieceCount) + (0 == (a_RangeByteSize % l_PieceCount) ? 0u : 1u));

                a_FirstPieceSize = l_MaxPacketBodySize;
                a_OtherPieceMaxSize = Math.Min(l_MaxPacketBodySize, a_RangeByteSize - l_MaxPacketBodySize);
            }
            else
            {
                uint l_RemRaw = (uint)(a_FileOffset % c_DefaultPieceSize);
                uint l_Remainder = (0 == l_RemRaw ? 0 : (c_DefaultPieceSize - l_RemRaw));

                uint l_FirstSize;

                if (0 == l_Remainder)
                {
                    l_FirstSize = c_DefaultPieceSize;
                }
                else if (c_MinAllowedFragementSize > l_Remainder)
                {
                    l_FirstSize = Math.Min(l_Remainder + c_DefaultPieceSize, a_RangeByteSize);
                }
                else
                {
                    l_FirstSize = l_Remainder;
                }

                uint l_TailSize = (a_RangeByteSize - l_FirstSize);

                if (c_MinAllowedFragementSize > l_TailSize)
                {
                    goto RETURN_ONE_PIECE;
                }

                a_FirstPieceSize = l_FirstSize;

                if (c_DefaultPieceSize > l_TailSize)
                {
                    l_PieceCount = 2;
                    a_OtherPieceMaxSize = l_TailSize;
                }
                else
                {
                    l_PieceCount = (1 + (l_TailSize / c_DefaultPieceSize));
                    a_OtherPieceMaxSize = c_DefaultPieceSize;

                    uint l_TailRem = (l_TailSize % c_DefaultPieceSize);

                    if (c_MinAllowedFragementSize <= l_TailRem)
                    {
                        l_PieceCount++;
                    }
                }
            }

            return l_PieceCount;

        RETURN_ONE_PIECE:

            a_FirstPieceSize = a_RangeByteSize;
            a_OtherPieceMaxSize = 0;

            return 1;
        }

        protected override void NotifyIncomingContent(ref IncomingPacketInfo a_Info,
                                                      uint a_NewlyRcvdBytes)
        {
            if (null == a_Info.f_DlPeerCntxt &&
                UInt128.c_ByteSize <= a_Info.p_BytesDone)
            {
                UInt128 l_UInt128 = UInt128.FromByteArrayBE(a_Info.f_Data, 0);

                if (null != f_DlPeerCntxt &&
                    true == f_DlPeerCntxt.HasFileID(new FileID(l_UInt128)))
                {
                    a_Info.f_DlPeerCntxt = f_DlPeerCntxt;
                    a_NewlyRcvdBytes += a_Info.f_UnrecognizedSize;
                    a_Info.f_UnrecognizedSize = 0;
                }
            }

            if (null == a_Info.f_DlPeerCntxt)
            {
                a_Info.f_UnrecognizedSize += a_NewlyRcvdBytes;
            }
            else
            {
                a_Info.f_DlPeerCntxt.NotifyReceivedContentBytes(a_NewlyRcvdBytes);
            }
        }

        private bool ProcessFileContent(Packet a_Packet,
                                        bool a_Compressed)
        {
            byte l_HeaderSize = (byte)(false == a_Compressed ? (c_PlainContentHeaderSize - UInt128.c_ByteSize) : (c_ComprContentEssentialHeaderSize - UInt128.c_ByteSize));
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;

            if (null == ConfirmPeerContext(l_Buffer, PeerContextRequirements.Dl, out l_Remaining, out l_Excusable))
            {
                DebugFatal(300, !l_Excusable);

                return l_Excusable;
            }

            if (l_HeaderSize >= l_Remaining)
            {
                DebugFatal(301);

                return false;
            }

            ulong l_StartOffset = l_Buffer.ReadUInt64LE();

            if (false == f_DlHelper.IsValid())
            {
                FileRange l_Range;

                if (false == f_DlPeerCntxt.ActivateRequestedRange(l_StartOffset, out l_Range))
                {
                    // We didn't even request this.

                    DebugFatal(302);

                    return false;
                }

                f_DlHelper = new DownloadHelper(f_DlPeerCntxt, ref l_Range);
            }

            if (false == a_Compressed)
            {
                ulong l_ExclEndingOffset = l_Buffer.ReadUInt64LE();

                if (l_ExclEndingOffset <= l_StartOffset)
                {
                    DebugFatal(303);

                    return false;
                }

                ulong l_Len = (l_ExclEndingOffset - l_StartOffset);

                if (f_DlHelper.f_PlainRange.f_Length < l_Len ||
                    l_Buffer.GetRemaining() != l_Len)
                {
                    DebugFatal(304);

                    return false;
                }

                ByteArray l_Body;

                if (false == l_Buffer.ExposeRangeChecked(l_Buffer.p_Position, (uint)l_Len, out l_Body) ||
                    false == f_DlHelper.AddPlainData(l_StartOffset, ref l_Body))
                {
                    DebugFatal(305);

                    return false;
                }
            }
            else
            {
                uint l_BlockSize = l_Buffer.ReadUInt32LE();
                l_Remaining = l_Buffer.GetRemaining();

                if (l_BlockSize < l_Remaining ||
                    0 == l_Remaining)
                {
                    DebugFatal(306);

                    return false;
                }

                ByteArray l_Body;

                if (false == l_Buffer.ExposeRangeChecked(l_Buffer.p_Position, l_Remaining, out l_Body) ||
                    false == f_DlHelper.AddCompressedData(l_StartOffset, l_BlockSize, ref l_Body))
                {
                    DebugFatal(307);

                    return false;
                }
            }

            if (true == f_DlHelper.IsFinished())
            {
                f_DlHelper.Invalidate();
            }

            NotifyReceive(a_Packet);

            return true;
        }

        private bool ProcessRevealSharedObjectsRequest(Packet a_Packet)
        {
            return EnqueueEmptyPacket(Opcodes.OP_ASKSHAREDDENIEDANS, PacketSendingFlags.IncUsersCount);
        }

        private bool RequestVersionSignature()
        {
            if (false == EnqueueEmptyPacket(OpcodesEx.OP_VERSIONSIGNATURE_REQ, PacketSendingFlags.None))
            {
                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.RequestedVersionSignature));

            return true;
        }

        private static System.IO.Stream OpenVersionSignatureResource()
        {
            // The caller must dispose the returned object later.

            System.IO.Stream l_Strm = Software.OpenResource("VerSig.bin", false);

            return l_Strm;
        }

        private bool ProcessVersionSignatureRequest(Packet a_Packet)
        {
            if (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.SentVersionSignature)))
            {
                return false;
            }

            if (false == a_Packet.p_Buffer.EndReached())
            {
                return false;
            }

            // The signature is ebedded into the binary as a resource.

            ByteArray l_Data;

            System.IO.Stream l_Strm = OpenVersionSignatureResource();
            {
                l_Data = new ByteArray(c_VersionSignatureBufferSize);

                if (Util.ReadFromStream(l_Strm, ref l_Data) != c_VersionSignatureBufferSize)
                {
                    return true;
                }
            }
            l_Strm.Dispose();

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.SentVersionSignature));

            return EnqueuePacket(OpcodesEx.OP_VERSIONSIGNATURE_ANS, PacketSendingFlags.IncUsersCount, ref l_Data);
        }

        private bool ProcessVersionSignature(Packet a_Packet)
        {
            if (false == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.RequestedVersionSignature)) ||
                true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.ReceivedVersionSignature)))
            {
                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.ReceivedVersionSignature));

            BufferIO l_Buf = a_Packet.p_Buffer;

            if (true == l_Buf.EndReached())
            {
                return false;
            }

            const byte l_ExpectedKeySerial = 1;
            byte l_KeySerial = l_Buf.ReadByte();

            if (l_ExpectedKeySerial != l_KeySerial)
            {
                if (l_ExpectedKeySerial > l_KeySerial)
                {
                    return true;
                }

                return false;
            }

            if (c_VersionSignatureBufferSize != l_Buf.p_Length)
            {
                return false;
            }

            if (l_Buf.ReadUInt32LE() != f_PeerInfo.EncodeVersion())
            {
                return false;
            }

            const ushort l_SignatureBitLength = 4128;
            const ushort l_SignatureByteSize = (l_SignatureBitLength / Util.c_BitsPerByte);
            const ushort l_VersionSignatureBufferSeedSize = (ushort)(c_VersionSignatureBufferSize - l_SignatureByteSize);
            const ushort l_EncodedPublicKeyByteSize = 525;

            DateTime l_NewerVersionBuildDate;

            if (false == Util.TryBuildDateTime(l_Buf.ReadUInt16LE(), l_Buf.ReadByte(), l_Buf.ReadByte(), DateTimeKind.Utc, out l_NewerVersionBuildDate))
            {
                return false;
            }

            if (Software.s_ReleaseInfo.f_BuildDate > l_NewerVersionBuildDate)
            {
                return false;
            }

            // Hash the data we received.

            ByteArray l_HashableData = new ByteArray();
            l_Buf.ExposeRange(0, l_VersionSignatureBufferSeedSize, out l_HashableData);

            Tiger l_Tiger = new Tiger();
            ByteArray l_HashBytes = new ByteArray(Tiger.c_HashByteSize);

            l_Tiger.Update(ref l_HashableData);

            // Repeat hashing three times, to make it harder to find
            // a collision through brute-force.
            const byte l_HashRepeat = 3;

            for (uint i = 0; l_HashRepeat != i; i++)
            {
                l_Tiger.GetDigest(l_HashBytes.f_Arr, 0);
                l_Tiger.Reset();
                l_Tiger.Update(l_HashBytes.f_Arr, 0, Tiger.c_HashByteSize);
            }

            l_Tiger.GetDigest160(l_HashBytes.f_Arr, 0);

            l_HashBytes = new ByteArray(l_HashBytes.f_Arr, 0, SHA_1.c_HashByteSize);

            // Load public key.

            System.Security.Cryptography.RSACryptoServiceProvider l_RSA;
            RsaPublicKey l_PubKey;

            System.IO.Stream l_Strm = OpenVersionSignatureResource();
            {
                l_RSA = new System.Security.Cryptography.RSACryptoServiceProvider();
                ByteArray l_PubKeyData = new ByteArray(l_EncodedPublicKeyByteSize);

                l_Strm.Position = c_VersionSignatureBufferSize;

                if (Util.ReadFromStream(l_Strm, ref l_PubKeyData) != l_EncodedPublicKeyByteSize)
                {
                    return true;
                }

                bool l_GotKey = RsaPublicKey.FromPortableRsaPublicKeyFormat(l_RSA, ref l_PubKeyData, out l_PubKey);

                if (false == l_GotKey)
                {
                    return true;
                }
            }
            l_Strm.Dispose();

            // Verify signature.

            ByteArray l_Signature;
            l_Buf.ExposeRange(l_VersionSignatureBufferSeedSize, l_SignatureByteSize, out l_Signature);

            RsaPublicKey.SignatureVerificationContext l_VerCntxt = new RsaPublicKey.SignatureVerificationContext(l_RSA, l_HashBytes.ToArray(false), l_Signature.ToArray(false), false, null);

            if (false == l_PubKey.VerifySignature(ref l_VerCntxt))
            {
                return false;
            }

            // Fine.

            Software.NotifyNewerVersionIsAvailable(f_PeerInfo.EncodeVersion(), l_NewerVersionBuildDate);

            return true;
        }

        public bool RequestSources(FileID a_FileID)
        {
            return EnqueueFileIDPacket(Opcodes.OP_REQUESTSOURCES, PacketSendingFlags.None, a_FileID);
        }

        private bool ProcessSourcesRequest(Packet a_Packet)
        {
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(a_Packet, PeerContextRequirements.Any, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                DebugFatal(310, !l_Excusable);

                return l_Excusable;
            }

            Time l_Time = f_Inst.GetTime();

            if (l_Time < f_EarliestNextSourcesRequest)
            {
                // Requests happen too frequently, so ignore.

                return true;
            }

            f_EarliestNextSourcesRequest = l_Time.AddSeconds(DownloadQueue.c_SourceQueryIntervalSecs);
            bool l_RestrictToClassicAddresses = p_RestrictToClassicAddresses;
            KnownFile.Source[] l_Sources = l_PeerCntxt.f_Known.p_ActCntxt.FindSuitableSourcesForPeer(25, l_RestrictToClassicAddresses, l_PeerCntxt);

            if (null == l_Sources)
            {
                return true;
            }

            int l_Count = l_Sources.Length;
            uint l_InitialSize = (UInt128.c_ByteSize + sizeof(ushort) +
                                  ((uint)l_Count * (I2PAddress.c_MaximumByteSize + UInt128.c_ByteSize)));
            MemoryIO l_AnswBuf = MemoryIO.CreateWritableAndGrowable(l_InitialSize);

            l_AnswBuf.WriteUInt128BE(l_PeerCntxt.f_FileID.f_Val);
            l_AnswBuf.WriteUInt16LE((ushort)l_Count);

            foreach (KnownFile.Source l_Source in l_Sources)
            {
                l_Source.f_StreamAddr.ToBufferRaw(l_AnswBuf);
                l_AnswBuf.WriteUInt128BE(l_Source.f_UserCntxt.p_UserID.f_Val);
            }

            return EnqueuePacket(Opcodes.OP_ANSWERSOURCES, PacketSendingFlags.IncUsersCount, l_AnswBuf);
        }

        private bool ProcessSourcesAnswer(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            bool l_Excusable;
            PeerContext l_PeerCntxt = ConfirmPeerContext(l_Buffer, PeerContextRequirements.Any, out l_Remaining, out l_Excusable);

            if (null == l_PeerCntxt)
            {
                // IMULETWEAK: It can happen with iMule 2.3.2 that it
                // takes five minutes before it answers a source
                // request. It probably stores the request.
                if (true == f_PeerInfo.p_IsiMule ||
                    (true == f_PeerInfo.p_IsThisClient &&
                     0x10418 >= f_PeerInfo.EncodeVersion()))
                {
                    l_Excusable = true;
                }

                DebugFatal(320, !l_Excusable);

                return l_Excusable;
            }

            if (sizeof(ushort) > l_Remaining)
            {
                DebugFatal(321);

                return false;
            }

            int l_Count = l_Buffer.ReadUInt16LE();
            l_Remaining -= sizeof(ushort);

            if (0 == l_Count ||
                ((uint)l_Count * (I2PAddress.c_MinimumByteSize + UInt128.c_ByteSize)) > l_Remaining)
            {
                DebugFatal(322);

                return false;
            }

            int l_RealCount = 0;
            I2PAddress[] l_StreamAddrs = new I2PAddress[l_Count];

            for (int i = 0; i < l_Count; i++)
            {
                I2PAddress l_Addr;
                BufferIO.LoadResult l_Res = I2PAddress.TryCreateFromBufferRaw(l_Buffer, out l_Addr);

                if (BufferIO.LoadResult.Success == l_Res)
                {
                    l_StreamAddrs[l_RealCount++] = l_Addr;
                }
                else
                {
                    if (BufferIO.LoadResult.Corruption == l_Res)
                    {
                        return true;
                    }
                }

                if (UInt128.c_ByteSize > l_Buffer.GetRemaining())
                {
                    return false;
                }

                l_Buffer.JumpAhead(UInt128.c_ByteSize);
            }

            if (false == l_Buffer.EndReached())
            {
                return false;
            }

            if (l_Count != l_RealCount)
            {
                if (0 >= l_RealCount)
                {
                    return true;
                }

                Array.Resize(ref l_StreamAddrs, l_RealCount);
            }

            return l_PeerCntxt.NotifySources(l_StreamAddrs);
        }

        private bool ProcessFileNameRequest(Packet a_Packet)
        {
            UlPeerCntxtCreationModes l_Junk;
            UlPeerCntxtCreationStates l_Status = TryBeginCreateUlPeerContext(a_Packet, out l_Junk);

            switch (l_Status)
            {
                case UlPeerCntxtCreationStates.AlreadyUp:
                    return SendFileNameAndComment(f_UlPeerCntxt);

                case UlPeerCntxtCreationStates.Pending:
                    return true;

                case UlPeerCntxtCreationStates.NotAllowed:
                    return EnqueueEmptyPacket(OpcodesEx.OP_FILE_SELECTION_NOT_ALLOWED_INFO, PacketSendingFlags.IncUsersCount);

                case UlPeerCntxtCreationStates.Corruption:
                    return false;

                default:
                    throw (new UnexpectedStateException());
            }
        }

        public bool SendFileNameAndComment(PeerContext a_PeerCntxt)
        {
            return (true == SendFileName(a_PeerCntxt) &&
                    true == SendFileComment(a_PeerCntxt));
        }

        public bool SendFileName(PeerContext a_PeerCntxt)
        {
            MemoryIO l_ResBuffer = MemoryIO.CreateWritable();

            string l_FileName = a_PeerCntxt.f_Known.GetFileNameVolatile(false);

            if (0 == l_FileName.Length)
            {
                // This should rarely ever happen.

                l_FileName = FileNames.c_FallbackFileName;
            }

            l_ResBuffer.WriteUInt128BE(a_PeerCntxt.f_FileID.f_Val);
            WriteString16(l_ResBuffer, l_FileName);

            return EnqueuePacket(Opcodes.OP_REQFILENAMEANSWER, PacketSendingFlags.IncUsersCount, l_ResBuffer);
        }

        private bool ProcessFileName(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            DlPeerContext l_DlPeerCntxt;
            UlPeerContext l_UlPeerCntxt;
            bool l_Excusable;

            if (false == ConfirmPeerContexts(l_Buffer, out l_DlPeerCntxt, out l_UlPeerCntxt, out l_Remaining, out l_Excusable))
            {
                DebugFatal(340, !l_Excusable);

                return l_Excusable;
            }

            if (sizeof(ushort) > l_Remaining)
            {
                DebugFatal(341);

                return false;
            }

            uint l_ByteLen = l_Buffer.ReadUInt16LE();
            l_Remaining -= sizeof(ushort);

            if (l_ByteLen != l_Remaining)
            {
                DebugFatal(342);

                return false;
            }

            string l_FileName;

            if (0 == l_ByteLen)
            {
                l_FileName = string.Empty;
            }
            else if (false == ReadStringUntrusted(l_Buffer, l_ByteLen, out l_FileName))
            {
                DebugFatal(343);

                return false;
            }

            if (null != l_DlPeerCntxt)
            {
                f_SupposedCommentFileID = l_DlPeerCntxt.f_FileID;
                l_DlPeerCntxt.NotifyFileName(l_FileName);
            }

            if (null != l_UlPeerCntxt)
            {
                f_SupposedCommentFileID = l_UlPeerCntxt.f_FileID;
                l_UlPeerCntxt.NotifyFileName(l_FileName);
            }

            NotifyReceive(a_Packet);

            return true;
        }

        private bool SendFileComment(PeerContext a_PeerCntxt)
        {
            if (0 == f_PeerInfo.f_AcceptCommentVer ||
                (FileRatings.NotRated == a_PeerCntxt.f_Known.f_OwnRating &&
                 null == a_PeerCntxt.f_Known.f_OwnComment))
            {
                return true;
            }

            MemoryIO l_Buffer = MemoryIO.CreateWritable();

            l_Buffer.WriteByte((byte)a_PeerCntxt.f_Known.f_OwnRating);
            WriteString32(l_Buffer, a_PeerCntxt.f_Known.f_OwnComment);

            return EnqueuePacket(Opcodes.OP_FILEDESC, PacketSendingFlags.IncUsersCount, l_Buffer);
        }

        private bool ProcessFileComment(Packet a_Packet)
        {
            if (null == f_DlPeerCntxt &&
                null == f_UlPeerCntxt)
            {
                bool l_Excusable = TryExcuseAnyMissingContext();

                DebugFatal(350, !l_Excusable);

                return l_Excusable;
            }

            if (null == f_SupposedCommentFileID.f_Val)
            {
                // We need a file name packet first.

                DebugFatal(351);

                return false;
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining = l_Buffer.GetRemaining();

            if ((sizeof(byte) + sizeof(uint)) > l_Remaining)
            {
                DebugFatal(352);

                return false;
            }

            FileRatings l_Rating = (FileRatings)l_Buffer.ReadByte();

            if (FileRatings.Excellent < l_Rating)
            {
                DebugFatal(353);

                return false;
            }

            uint l_ByteLen = l_Buffer.ReadUInt32LE();
            l_Remaining -= (sizeof(byte) + sizeof(uint));

            if (l_Remaining != l_ByteLen ||
                ushort.MaxValue < l_ByteLen)
            {
                DebugFatal(354);

                return false;
            }

            string l_Description;

            if (0 == l_ByteLen)
            {
                l_Description = null;
            }
            else if (false == ReadStringUntrusted(l_Buffer, l_ByteLen, out l_Description) ||
                     false == f_Inst.f_Kad.TrimAndVerifyComment(l_Description, out l_Description))
            {
                DebugFatal(355);

                return false;
            }

            if (FileRatings.NotRated != l_Rating ||
                null != l_Description)
            {
                if (null != f_DlPeerCntxt &&
                    true == f_DlPeerCntxt.HasFileID(f_SupposedCommentFileID) &&
                    false == f_DlPeerCntxt.NotifyFileComment(l_Rating, l_Description))
                {
                    DebugFatal(356);

                    return false;
                }

                if (null != f_UlPeerCntxt &&
                    true == f_UlPeerCntxt.HasFileID(f_SupposedCommentFileID) &&
                    false == f_UlPeerCntxt.NotifyFileComment(l_Rating, l_Description))
                {
                    DebugFatal(357);

                    return false;
                }

                NotifyReceive(a_Packet);
            }

            return true;
        }

        private bool SendFileStatusOrComplBitmap(UlPeerCntxtCreationModes a_Mode)
        {
            const PacketSendingFlags lc_Flags = PacketSendingFlags.IncUsersCount;
            bool l_Succ;
            KnownFile l_Known = f_UlPeerCntxt.f_Known;

            if (0 != (UlPeerCntxtCreationModes.SendFileStatus & a_Mode))
            {
                bool l_Complete = l_Known.p_eD2kCompleted;
                uint l_Size = (UInt128.c_ByteSize + sizeof(ulong) + (false == l_Complete ? l_Known.f_eD2kBitArrayByteSize : 0));
                MemoryIO l_ResFile = MemoryIO.CreateWritable(l_Size);
                l_ResFile.WriteUInt128BE(f_UlPeerCntxt.f_FileID.f_Val);

                if (false == l_Known.p_eD2kCompleted)
                {
                    l_ResFile.WriteUInt64LE(l_Known.f_PartCountFileStatus);
                    l_Known.p_ActCntxt.ExposeOwnComplBitmap(false).WriteBodyToBuffer(l_ResFile);
                }
                else
                {
                    l_ResFile.ZeroBytes(sizeof(ulong));
                }

                l_Succ = EnqueuePacket(Opcodes.OP_FILESTATUS, lc_Flags, l_ResFile);
            }
            else
            {
                bool l_UseNativeBitmap = (0 != (UlPeerCntxtCreationModes.UseNativeBitmap & a_Mode));
                bool l_IncludeBitmapBody = (0 != (UlPeerCntxtCreationModes.IncludeBitmapBodyIfNontrivial & a_Mode));
                BitMap l_ComplMap = l_Known.p_ActCntxt.ExposeOwnComplBitmap(l_UseNativeBitmap);

                if (true == l_IncludeBitmapBody &&
                    (true == l_ComplMap.p_AllClear ||
                     true == l_ComplMap.p_AllSet))
                {
                    // Do not include the bitmap if its content is obvious.

                    l_IncludeBitmapBody = false;
                }

                // NOTE: We don't write the bitmap length, since we
                // already made sure the requestor knows our file size.
                uint l_Size = (UInt128.c_ByteSize + sizeof(uint) + sizeof(ulong) + (false == l_IncludeBitmapBody ? 0 : l_ComplMap.CalcByteCount()));
                MemoryIO l_ResFile = MemoryIO.CreateWritable(l_Size);
                l_ResFile.WriteUInt128BE(f_UlPeerCntxt.f_FileID.f_Val);
                l_ResFile.WriteUInt32LE(l_ComplMap.f_Granularity);
                l_ResFile.WriteUInt64LE(l_ComplMap.p_SetBitCount);

                if (true == l_IncludeBitmapBody)
                {
                    l_ComplMap.WriteBodyToBuffer(l_ResFile);
                }

                l_Succ = EnqueuePacket(OpcodesEx.OP_COMPLBITMAP_ANS, lc_Flags, l_ResFile);
            }

            return l_Succ;
        }

        private bool ProcessFileStatusOrComplBitmap(Packet a_Packet)
        {
            // Works for up- and downloading.

            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            DlPeerContext l_DlPeerCntxt;
            UlPeerContext l_UlPeerCntxt;
            bool l_Excusable;

            if (false == ConfirmPeerContexts(l_Buffer, out l_DlPeerCntxt, out l_UlPeerCntxt, out l_Remaining, out l_Excusable))
            {
                DebugFatal(360, !l_Excusable);

                return l_Excusable;
            }

            if (sizeof(ulong) > l_Remaining)
            {
                DebugFatal(361);

                return false;
            }

            KnownFile l_Known;

            if (null == l_DlPeerCntxt)
            {
                l_Known = f_UlPeerCntxt.f_Known;
            }
            else
            {
                if (null != l_UlPeerCntxt &&
                    f_DlPeerCntxt.f_Known != f_UlPeerCntxt.f_Known)
                {
                    DebugFatal(362);

                    return false;
                }

                l_Known = f_DlPeerCntxt.f_Known;
            }

            BitMap l_ComplMap;

            if (Protocols.OP_EDONKEYPROT == a_Packet.f_Prot)
            {
                // They use this crazy errornous formula to calculate
                // what they believe is a part count, so beware.
                ulong l_PartCountFileStatus = l_Buffer.ReadUInt64LE();
                l_Remaining -= sizeof(ulong);
                bool l_ExpectBitmap = (0 != l_PartCountFileStatus);
                uint l_ExpSize;

                if (false == l_ExpectBitmap)
                {
                    l_ExpSize = 0;
                }
                else
                {
                    if (l_Known.f_PartCountFileStatus != l_PartCountFileStatus)
                    {
                        // The sender seems to know a different file size,
                        // but that's no reason to close the connection.

                        return (null == l_DlPeerCntxt ||
                                true == f_DlPeerCntxt.NotifyDisputedFileSize());
                    }

                    l_ExpSize = l_Known.f_eD2kBitArrayByteSize;
                }

                if (l_ExpSize != l_Remaining)
                {
                    // There may be one useless extra byte when
                    // a bitmap is present.
                    bool l_JustUselessExtraByte = (0 != l_ExpSize &&
                                                   (l_ExpSize + 1) == l_Remaining);

                    if (false == l_JustUselessExtraByte)
                    {
                        DebugFatal(363);

                        return false;
                    }
                }

                l_ComplMap = new BitMap(BitMap.c_eD2kGranularity, l_Known.f_PartCount);

                if (0 == l_PartCountFileStatus)
                {
                    // Zero means the sender has the complete file.

                    l_ComplMap.SetAll(true);
                }
                else
                {
                    l_ComplMap.ReadBodyFromBuffer(0, l_Known.f_eD2kBitArrayByteSize, l_Buffer);
                }
            }
            else
            {
                uint l_Granularity = l_Buffer.ReadUInt32LE();
                l_Remaining -= sizeof(uint);

                if (0 == l_Granularity ||
                    sizeof(ulong) > l_Remaining)
                {
                    DebugFatal(364);

                    return false;
                }

                ulong l_SetBitCount = l_Buffer.ReadUInt64LE();
                l_Remaining -= sizeof(ulong);
                uint l_BitCount;
                uint l_ByteSize;
                bool l_RescaleToNative = false;

                if (BitMap.c_eD2kGranularity == l_Granularity)
                {
                    l_BitCount = l_Known.f_PartCount;
                    l_ByteSize = l_Known.f_eD2kBitArrayByteSize;
                }
                else if (BitMap.c_NativeGranularity == l_Granularity)
                {
                    l_BitCount = (uint)BitMap.CalcBitCount(BitMap.c_NativeGranularity, l_Known.f_FileSize);
                    l_ByteSize = BitMap.CalcByteCount(l_BitCount);
                }
                else
                {
                    ulong l_BitCount64 = BitMap.CalcBitCount(l_Granularity, l_Known.f_FileSize);

                    if (int.MaxValue < l_BitCount64)
                    {
                        DebugFatal(365);

                        return false;
                    }

                    l_BitCount = (uint)l_BitCount64;
                    l_ByteSize = BitMap.CalcByteCount(l_BitCount);
                    l_RescaleToNative = true;
                }

                if (l_SetBitCount > l_BitCount)
                {
                    DebugFatal(366);

                    return false;
                }

                // NOTE: We assume we requested the bitmap body in
                // case the bitmap is nontrivial.

                bool l_Trivial = (0 == l_SetBitCount ||
                                  l_BitCount == l_SetBitCount);

                if ((false == l_Trivial ? l_ByteSize : 0) != l_Remaining)
                {
                    DebugFatal(367);

                    return false;
                }

                if (false == l_Trivial)
                {
                    l_ComplMap = new BitMap(l_Granularity, l_BitCount);
                    l_ComplMap.ReadBodyFromBuffer(0, l_ByteSize, l_Buffer);

                    if (true == l_RescaleToNative)
                    {
                        BitMap l_NewComplMap = BitMap.CreateFromFileSize(BitMap.c_NativeGranularity, l_Known.f_FileSize);
                        l_NewComplMap.TakeOver(l_ComplMap, BitMap.TakeOverMode.RoundDown);
                        l_ComplMap = l_NewComplMap;
                    }
                }
                else
                {
                    if (true == l_RescaleToNative)
                    {
                        l_Granularity = BitMap.c_NativeGranularity;
                        l_BitCount = (uint)BitMap.CalcBitCount(BitMap.c_NativeGranularity, l_Known.f_FileSize);

                        if (0 != l_SetBitCount)
                        {
                            l_SetBitCount = l_BitCount;
                        }
                    }

                    l_ComplMap = new BitMap(l_Granularity, l_BitCount);

                    if (0 != l_SetBitCount)
                    {
                        l_ComplMap.SetAll(true);
                    }
                }
            }

            Time l_Time = f_Inst.GetTime();
            bool l_OK = ((null == l_DlPeerCntxt ||
                          true == l_DlPeerCntxt.NotifyFileStatus(l_Time, l_ComplMap)) &&
                         (null == l_UlPeerCntxt ||
                          true == l_UlPeerCntxt.NotifyFileStatus(l_Time, l_ComplMap)));

            if (true == l_OK)
            {
                NotifyReceive(a_Packet);
            }

            return l_OK;
        }

        private bool ProcessFileReqAnswerNoFile(Packet a_Packet)
        {
            BufferIO l_Buffer = a_Packet.p_Buffer;
            uint l_Remaining;
            DlPeerContext l_DlPeerCntxt;
            UlPeerContext l_UlPeerCntxt;
            bool l_Excusable;

            if (false == ConfirmPeerContexts(l_Buffer, out l_DlPeerCntxt, out l_UlPeerCntxt, out l_Remaining, out l_Excusable))
            {
                DebugFatal(370, !l_Excusable);

                return l_Excusable;
            }

            if (0 != l_Remaining)
            {
                DebugFatal(371);

                return false;
            }

            if (null != l_DlPeerCntxt &&
                false == l_DlPeerCntxt.NotifyNoSuchFile())
            {
                DebugFatal(372);

                return false;
            }

            if (null != l_UlPeerCntxt &&
                false == l_UlPeerCntxt.NotifyNoSuchFile())
            {
                DebugFatal(373);

                return false;
            }

            NotifyReceive(a_Packet);

            return true;
        }

        private bool ProcessHello(Packet a_Packet)
        {
            if (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.ReceivedHello)))
            {
                DebugFatal(380);

                return false;
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;

            if (Opcodes.OP_HELLO == a_Packet.p_DonkeyOpcode)
            {
                // The OP_HELLO packet has a special size byte
                // that the OP_HELLO_ANSWER lacks.

                if (0 == l_Buffer.GetRemaining())
                {
                    DebugFatal(381);

                    return false;
                }

                byte l_HashSize = l_Buffer.ReadByte();

                if (UInt128.c_ByteSize != l_HashSize)
                {
                    DebugFatal(382);

                    return false;
                }
            }

            if ((UInt128.c_ByteSize + sizeof(uint)) > l_Buffer.GetRemaining())
            {
                DebugFatal(383);

                return false;
            }

            UserID l_UserID = new UserID(l_Buffer.ReadUInt128BE());

            if (false == l_UserID.IsValid())
            {
                DebugFatal(384);

                return false;
            }

            uint l_TagCount = l_Buffer.ReadUInt32LE();

            if (0 == l_TagCount)
            {
                DebugFatal(385);

                return false;
            }

            if (byte.MaxValue < l_TagCount)
            {
                DebugFatal(386);

                return false;
            }

            TagList l_TagList = TagList.ReadListUntrusted(l_Buffer, l_TagCount, PeerInfo.c_ExpectNormalStrsAsUTF8InIngoingHello);

            if (null == l_TagList)
            {
                DebugFatal(387);

                return false;
            }

            if (false == l_Buffer.EndReached())
            {
                DebugFatal(388);

                return false;
            }

            if (false == PeerInfo.CreateFromTagList(l_UserID, l_TagList, out f_PeerInfo))
            {
                return false;
            }

            if (false == f_Inst.f_UserMngr.TrySetUserID(f_UserCntxt, l_UserID))
            {
                return false;
            }

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.ReceivedHello));
            f_ClientStr = f_PeerInfo.BuildClientString(f_Inst.f_TextSet);

            if (Opcodes.OP_HELLOANSWER != a_Packet.p_DonkeyOpcode &&
                false == SendHello())
            {
                DebugFatal(392);

                return false;
            }

            NotifyReceiveIfOutgoing(a_Packet);

            if (false == f_PeerInfo.HasSecIdentSupport())
            {
                if (true == f_UserCntxt.HasPublicKey() &&
                    false == AllowPeerToGoOnWithFailedAuthentication())
                {
                    DebugFatal(393);

                    return false;
                }

                FinishedHandshake(false);

                return true;
            }

            if (true == f_Inst.f_NetworkMngr.f_SelfInfo.HasSecIdentSupport())
            {
                return SendSecIdentState();
            }

            // Should never happen, since we don't ever plan
            // to deactivate our own sec ident support.

            FinishedHandshake(false);

            return true;
        }

        private bool ProcessMuleInfoPacket(Packet a_Packet)
        {
            if (true == f_Flags.CheckFlag(FlagsToUInt(StreamFlags.ReceivedMuleInfo)))
            {
                DebugFatal(400);

                return false;
            }

            BufferIO l_Buffer = a_Packet.p_Buffer;

            if (((2 * sizeof(byte)) + sizeof(uint)) > l_Buffer.GetRemaining())
            {
                DebugFatal(401);

                return false;
            }

            byte l_eMuleVer = l_Buffer.ReadByte();
            byte l_ProtVer = l_Buffer.ReadByte();
            uint l_TagCount = l_Buffer.ReadUInt32LE();

            if (byte.MaxValue < l_TagCount)
            {
                DebugFatal(402);

                return false;
            }

            TagList l_TagList = TagList.ReadListUntrusted(l_Buffer, l_TagCount, p_UnicodeSupport);

            if (null == l_TagList)
            {
                DebugFatal(403);

                return false;
            }

            bool l_NewMule = (0xFF == l_ProtVer);

            if (false == l_NewMule)
            {
                DebugFatal(404);

                return false;
            }

            //string l_ClientOSInfo = l_TagList.GetTagAsString(TagCodes.ET_OS_INFO);

            f_Flags.SetFlag(FlagsToUInt(StreamFlags.ReceivedMuleInfo));

            NotifyReceiveIfOutgoing(a_Packet);

            return true;
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        private struct DownloadHelper
        {
            // Don't use the struct default constructor.

            private static readonly uint s_MaxComprBlockSize;

            public readonly DlPeerContext f_DlPeerCntxt;
            public FileRange f_PlainRange;

            private bool f_Failed;
            private bool f_Compression;
            private uint f_BlocksReceived;
            private uint f_RemLength;

            // For plain data.
            private ulong f_PlainOffset;

            // For compressed data.
            private int f_ComprOffset;
            private uint f_ComprBlockSize;
            private byte[] f_CompressedData;

            static DownloadHelper()
            {
                s_MaxComprBlockSize = zlib.GetMaxCompressedSize(c_MaxRangeLength);
            }

            public DownloadHelper(DlPeerContext a_DlPeerCntxt,
                                  ref FileRange a_PlainRange)
            {
                f_DlPeerCntxt = a_DlPeerCntxt;
                f_PlainRange = a_PlainRange;

                f_Failed = false;
                f_Compression = false;
                f_BlocksReceived = 0;
                f_RemLength = 0;

                f_PlainOffset = 0;

                f_ComprOffset = 0;
                f_ComprBlockSize = 0;
                f_CompressedData = null;
            }

            public bool IsValid()
            {
                return (0 != f_PlainRange.f_Length);
            }

            public void Invalidate()
            {
                f_PlainRange = new FileRange();
            }

            public bool AddPlainData(ulong a_Offset,
                                     ref ByteArray a_Data)
            {
                if (true == f_Failed ||
                    false == AddPlainDataCore(a_Offset, ref a_Data))
                {
                    f_Failed = true;

                    return false;
                }

                return true;
            }

            private bool AddPlainDataCore(ulong a_Offset,
                                          ref ByteArray a_Data)
            {
                // We expect a_Offset to grow with each block.

                if (true == f_Compression)
                {
                    return false;
                }

                if (0 == f_BlocksReceived)
                {
                    f_PlainOffset = f_PlainRange.f_StartingOffset;
                    f_RemLength = f_PlainRange.f_Length;
                }

                if (a_Offset != f_PlainOffset ||
                    a_Data.f_Len > f_RemLength)
                {
                    return false;
                }

                f_RemLength -= (uint)a_Data.f_Len;

                if (0 != f_BlocksReceived &&
                    0 != f_RemLength &&
                    c_MinAllowedFragementSize > a_Data.f_Len)
                {
                    // Reject mini blocks (except for the first and the
                    // last block).

                    return false;
                }

                f_BlocksReceived++;

                if (false == f_DlPeerCntxt.RegisterDownload(f_PlainOffset, ref a_Data, f_RemLength, 0))
                {
                    return false;
                }

                f_PlainOffset += (uint)a_Data.f_Len;

                return true;
            }

            public bool AddCompressedData(ulong a_Offset,
                                          uint a_BlockSize,
                                          ref ByteArray a_Data)
            {
                if (true == f_Failed ||
                    false == AddCompressedDataCore(a_Offset, a_BlockSize, ref a_Data))
                {
                    f_Failed = true;

                    return false;
                }

                return true;
            }

            private bool AddCompressedDataCore(ulong a_Offset,
                                               uint a_BlockSize,
                                               ref ByteArray a_Data)
            {
                // a_Offset is expected to stay the same for each block

                if (f_PlainRange.f_StartingOffset != a_Offset)
                {
                    return false;
                }

                if (0 == f_BlocksReceived)
                {
                    if (0 == a_BlockSize ||
                        s_MaxComprBlockSize < a_BlockSize)
                    {
                        return false;
                    }

                    f_Compression = true;
                    f_RemLength = a_BlockSize;
                    f_ComprOffset = 0;
                    f_ComprBlockSize = a_BlockSize;
                    f_CompressedData = new byte[f_ComprBlockSize];
                }
                else if (false == f_Compression ||
                         a_BlockSize != f_ComprBlockSize)
                {
                    // The block size must stay the same.

                    return false;
                }

                if (a_Data.f_Len > f_RemLength)
                {
                    return false;
                }

                Buffer.BlockCopy(a_Data.f_Arr, a_Data.f_Idx, f_CompressedData, f_ComprOffset, a_Data.f_Len);

                f_BlocksReceived++;
                f_RemLength -= (uint)a_Data.f_Len;
                f_ComprOffset += a_Data.f_Len;

                if (0 == f_RemLength)
                {
                    ByteArray l_CompData = new ByteArray(f_CompressedData);
                    ByteArray l_UncompData;
                    if (false == zlib.Uncompress(ref l_CompData, c_MaxRangeLength, out l_UncompData) ||
                        0 == l_UncompData.f_Len)
                    {
                        return false;
                    }

                    uint l_CompGain = 0;

                    if (l_UncompData.f_Len > f_ComprBlockSize)
                    {
                        l_CompGain = ((uint)l_UncompData.f_Len - f_ComprBlockSize);
                    }

                    if (false == f_DlPeerCntxt.RegisterDownload(f_PlainRange.f_StartingOffset, ref l_UncompData, 0, l_CompGain))
                    {
                        return false;
                    }
                }

                return true;
            }

            public bool IsFinished()
            {
                return (false == f_Failed &&
                        0 != f_BlocksReceived &&
                        0 == f_RemLength);
            }
        }

        private static uint FlagsToUInt(StreamFlags a_Flags)
        {
            return (uint)a_Flags;
        }

        [Flags]
        private enum StreamFlags : uint
        {
            SentHello = 0x1,
            SentOwnPubKey = 0x2,
            SentOwnSignature = 0x4,
            // This also means we have requested the peers key (if
            // not already received) and signature.
            SentSecIdentState = 0x8,
            SentRequestForSharedFiles = 0x10,
            SentRequestForSharedDirs = 0x20,

            ReceivedHello = 0x40,
            ReceivedMuleInfo = 0x80,
            RcvdPeersSecIdentState = 0x80000000,
            // This also indicates a (possibly pending) signature
            // request.
            RcvdPeersChallenge = 0x100,
            RcvdPeersPubKey = 0x200,
            RcvdPeersSignature = 0x400,

            RequestedVersionSignature = 0x800,
            ReceivedVersionSignature = 0x1000,
            SentVersionSignature = 0x2000,

            SecIdentFullyIdentifiedPeer = 0x4000,
            SecIdentAttemptToIdentifyPeerFailed = 0x8000,
            SecIdentFastTrackIndentifiedPeer = 0x10000,
            SecIdentPeerShouldBeDone = 0x20000,
        }

        [Flags]
        private enum FileSelectionFlags : byte
        {
            // If an operation is not allowed, it is ignored and a
            // OP_FILE_SELECTION_NOT_ALLOWED_INFO message is sent to
            // the requestor.

            None = 0,

            // AllowConstruction indicates if it's allowed to select
            // a file when there is no file selected yet.
            AllowConstruction = 0x1,
            // AllowConfirmation indicates if it's allowed to re-select
            // a currently selected file.
            AllowConfirmation = 0x2,
            // AllowConstruction indicates if it's allowed to replace
            // a currently selected file by another.
            AllowReplacement = 0x4,

            All = (AllowConstruction | AllowConfirmation | AllowReplacement),
        }

        public enum AuthenticationState : byte
        {
            PreAuth = 0,

            NoAuth = 1,
            FailedAuth = 2,
            SuccessfulAuth = 3,
        }

        [Flags]
        private enum PacketSendingFlags : byte
        {
            None = 0,

            IncUsersCount = 0x1,
            IncUserCountIfIncoming = 0x2,
            IncFilesCount = 0x4,
            IncBothCounts = (IncUsersCount | IncFilesCount),

            InsertAtEndOfQueue = 0x4,
            MayBeOutpaced = 0x8,
            NoCompression = 0x10,

            InsertOutpacableAtEndOfQueue = (InsertAtEndOfQueue | MayBeOutpaced),
        }

        [Flags]
        public enum AdvFileMetaDataFlags : ushort
        {
            IntegerSize64 = 0x1,
            Range = 0x2,
            Indices = 0x4,
        }

        [Flags]
        public enum UlPeerCntxtCreationModes : byte
        {
            None = 0,

            SendFileStatus = 0x1,
            SendFileNameAndComment = 0x2,
            SendComplBitmap = 0x4,

            // Used with SendComplBitmap.
            UseNativeBitmap = 0x40,
            IncludeBitmapBodyIfNontrivial = 0x80,
        }

        private enum UlPeerCntxtCreationStates : byte
        {
            Pending = 0,
            Corruption = 1,
            AlreadyUp = 2,
            NotAllowed = 3,
        }

        [Flags]
        private enum PeerContextRequirements : byte
        {
            None,
            Dl = 0x1,
            Ul = 0x2,
            Any = 0x3,
        }

#if !MONO
        [StructLayout(LayoutKind.Auto)]
#endif
        public struct ShortPeerInfo
        {
            public readonly string f_UserName;
            public readonly string f_ClientStr;
            public readonly uint f_AuthKeyStrength;
            public readonly byte f_SecIdentVer;

            public ShortPeerInfo(string a_UserName,
                                 string a_ClientStr,
                                 uint a_AuthKeyStrength,
                                 byte a_SecIdentVer)
            {
                f_UserName = a_UserName;
                f_ClientStr = a_ClientStr;
                f_AuthKeyStrength = a_AuthKeyStrength;
                f_SecIdentVer = a_SecIdentVer;
            }
        }

        enum CryptCips : byte
        {
            NONE = 0,
            CRYPT_CIP_REMOTECLIENT = 10,
            CRYPT_CIP_LOCALCLIENT = 20,
            CRYPT_CIP_NONECLIENT = 30,
        }

        enum SecureIdentStates : byte
        {
            IS_UNAVAILABLE = 0,
            IS_SIGNATURENEEDED = 1,
            IS_KEYANDSIGNEEDED = 2,

            // Own extension from NB 1.4.14 on.
            IS_AUTHENTICATED_INSTANTLY = 3,
        }
    }
}

// End of file.
